#include "polyrt.h"
#include "buffer_manager.h"
extern int max_num_elements_y;
extern int max_num_elements_tmp;
extern int max_num_elements_A;
extern int max_num_elements_x;
#include <mpi.h>
#include "tbb/concurrent_priority_queue.h"

#include <omp.h>

#ifdef __DEBUG_FLUSH
#define IF_DEBUG_FLUSH(foo) foo
#else
#define IF_DEBUG_FLUSH(foo) 
#endif

#ifdef __DYNSCHEDULER_MORE_DEBUG_PRINT
#define IF_DYNSCHEDULER_MORE_DEBUG_PRINT(foo) foo
#else
#define IF_DYNSCHEDULER_MORE_DEBUG_PRINT(foo) 
#endif

#ifdef __DYNSCHEDULER_DEBUG_PRINT
#define IF_DYNSCHEDULER_DEBUG_PRINT(foo) foo
#else
#define IF_DYNSCHEDULER_DEBUG_PRINT(foo) 
#endif

extern FILE *__debug_print_fp;

class __Task {
	public:
		int loop;
		int id1;
		int id2;
		int id3;
		int remote_dep_tasks;
		int local_dep_tasks;
		int thread_affinity;
		__Task() {}
		__Task(const int l, const int _id1, const int _id2, const int remote, const int local, const int affinity) :
			loop(l),
			id1(_id1),
			id2(_id2),
			id3(0),
			remote_dep_tasks(remote),
			local_dep_tasks(local),
			thread_affinity(affinity) {}
		__Task(const int l, const int _id1, const int _id2, const int _id3, const int remote, const int local, const int affinity) :
			loop(l),
			id1(_id1),
			id2(_id2),
			id3(_id3),
			remote_dep_tasks(remote),
			local_dep_tasks(local),
			thread_affinity(affinity) {}
		__Task(const __Task &task) :
			loop(task.loop),
			id1(task.id1),
			id2(task.id2),
			id3(task.id3),
			remote_dep_tasks(task.remote_dep_tasks),
			local_dep_tasks(task.local_dep_tasks),
			thread_affinity(task.thread_affinity) {}
		~__Task() {}
};

class __Compare {
	public:
		// returns true only if task2 has higher priority than task1
		bool operator()(const __Task &task1, const __Task &task2) const {
#ifndef __DYNSCHEDULER_NO_PRIORITY
			if (task2.remote_dep_tasks>task1.remote_dep_tasks) return true;
			if (task1.remote_dep_tasks>task2.remote_dep_tasks) return false;
			if (task2.local_dep_tasks>task1.local_dep_tasks) return true;
			if (task1.local_dep_tasks>task2.local_dep_tasks) return false;
			int thread_id = omp_get_thread_num();
			if ((thread_id == task2.thread_affinity) && (thread_id != task1.thread_affinity)) return true;
			if ((thread_id == task1.thread_affinity) && (thread_id != task2.thread_affinity)) return false;
			// lexicographic compare 
			if ( (task2.id1 > task1.id1) 
				|| ( (task2.id1 == task1.id1) && (task2.id2 > task1.id2) ) 
				|| ( (task2.id1 == task1.id1) && (task2.id2 == task1.id2) && (task2.id3 > task1.id3) ) )
				 return true;
#endif
			return false;
		}
};

extern tbb::concurrent_priority_queue<__Task, __Compare> pqueue;

int is_receiver_0(int ts1, int ts2, int _PB_NY, int _PB_NX, int my_rank, int nprocs);
int get_num_remote_dep_tasks_0(int ts1, int ts2, int _PB_NY, int _PB_NX, int my_rank, int nprocs);
void remote_update_dep_tasks_0(int ts1, int ts2, int _PB_NY, int _PB_NX, int my_rank, int nprocs, int num_threads,int *tasks_loop0,int *tasks_loop1,int *tasks_loop2,int *tasks_loop3);
int get_num_local_dep_tasks_0(int ts1, int ts2, int _PB_NY, int _PB_NX, int my_rank, int nprocs);
void local_update_dep_tasks_0(int ts1, int ts2, int _PB_NY, int _PB_NX, int my_rank, int nprocs, int num_threads,int *tasks_loop0,int *tasks_loop1,int *tasks_loop2,int *tasks_loop3);
int get_num_remote_src_tasks_0(int ts1, int ts2, int _PB_NY, int _PB_NX, int my_rank, int nprocs);
int get_num_local_src_tasks_0(int ts1, int ts2, int _PB_NY, int _PB_NX, int my_rank, int nprocs);
extern int max_num_tasks_loop0_dim0;
extern int lb_tasks_loop0_dim0;
extern int *tasks_loop0;
int is_receiver_1(int ts1, int ts2, int _PB_NY, int _PB_NX, int my_rank, int nprocs);
int get_num_remote_dep_tasks_1(int ts1, int ts2, int _PB_NY, int _PB_NX, int my_rank, int nprocs);
void remote_update_dep_tasks_1(int ts1, int ts2, int _PB_NY, int _PB_NX, int my_rank, int nprocs, int num_threads,int *tasks_loop0,int *tasks_loop1,int *tasks_loop2,int *tasks_loop3);
int get_num_local_dep_tasks_1(int ts1, int ts2, int _PB_NY, int _PB_NX, int my_rank, int nprocs);
void local_update_dep_tasks_1(int ts1, int ts2, int _PB_NY, int _PB_NX, int my_rank, int nprocs, int num_threads,int *tasks_loop0,int *tasks_loop1,int *tasks_loop2,int *tasks_loop3);
int get_num_remote_src_tasks_1(int ts1, int ts2, int _PB_NY, int _PB_NX, int my_rank, int nprocs);
int get_num_local_src_tasks_1(int ts1, int ts2, int _PB_NY, int _PB_NX, int my_rank, int nprocs);
extern int max_num_tasks_loop1_dim0;
extern int lb_tasks_loop1_dim0;
extern int *tasks_loop1;
int is_receiver_2(int ts1, int ts2, int _PB_NY, int _PB_NX, int my_rank, int nprocs);
int get_num_remote_dep_tasks_2(int ts1, int ts2, int _PB_NY, int _PB_NX, int my_rank, int nprocs);
void remote_update_dep_tasks_2(int ts1, int ts2, int _PB_NY, int _PB_NX, int my_rank, int nprocs, int num_threads,int *tasks_loop0,int *tasks_loop1,int *tasks_loop2,int *tasks_loop3);
int get_num_local_dep_tasks_2(int ts1, int ts2, int _PB_NY, int _PB_NX, int my_rank, int nprocs);
void local_update_dep_tasks_2(int ts1, int ts2, int _PB_NY, int _PB_NX, int my_rank, int nprocs, int num_threads,int *tasks_loop0,int *tasks_loop1,int *tasks_loop2,int *tasks_loop3);
int get_num_remote_src_tasks_2(int ts1, int ts2, int _PB_NY, int _PB_NX, int my_rank, int nprocs);
int get_num_local_src_tasks_2(int ts1, int ts2, int _PB_NY, int _PB_NX, int my_rank, int nprocs);
extern int max_num_tasks_loop2_dim0;
extern int lb_tasks_loop2_dim0;
extern int *tasks_loop2;
int is_receiver_3(int ts1, int ts2, int ts3, int _PB_NY, int _PB_NX, int my_rank, int nprocs);
int get_num_remote_dep_tasks_3(int ts1, int ts2, int ts3, int _PB_NY, int _PB_NX, int my_rank, int nprocs);
void remote_update_dep_tasks_3(int ts1, int ts2, int ts3, int _PB_NY, int _PB_NX, int my_rank, int nprocs, int num_threads,int *tasks_loop0,int *tasks_loop1,int *tasks_loop2,int *tasks_loop3);
int get_num_local_dep_tasks_3(int ts1, int ts2, int ts3, int _PB_NY, int _PB_NX, int my_rank, int nprocs);
void local_update_dep_tasks_3(int ts1, int ts2, int ts3, int _PB_NY, int _PB_NX, int my_rank, int nprocs, int num_threads,int *tasks_loop0,int *tasks_loop1,int *tasks_loop2,int *tasks_loop3);
int get_num_remote_src_tasks_3(int ts1, int ts2, int ts3, int _PB_NY, int _PB_NX, int my_rank, int nprocs);
int get_num_local_src_tasks_3(int ts1, int ts2, int ts3, int _PB_NY, int _PB_NX, int my_rank, int nprocs);
extern int max_num_tasks_loop3_dim0;
extern int lb_tasks_loop3_dim0;
extern int max_num_tasks_loop3_dim1;
extern int lb_tasks_loop3_dim1;
extern int *tasks_loop3;
void init_tasks(int my_rank,int nprocs,int num_threads,int *tasks_loop0,int *tasks_loop1,int *tasks_loop2,int *tasks_loop3,long int *p_num_tasks_to_execute,long int *p_num_tasks_to_unpack);
void data_tile_ref_count_init_y_0(int ts1,int ts2, double** y_trans, BufferManager *buff_mang_y);
void data_tile_ref_count_init_tmp_1(int ts1,int ts2, double** tmp_trans, BufferManager *buff_mang_tmp);
void data_tile_ref_count_init_tmp_2(int ts1,int ts2, double** tmp_trans, BufferManager *buff_mang_tmp);
void data_tile_ref_count_init_A_2(int ts1,int ts2, double** A_trans,int A_size_1, BufferManager *buff_mang_A);
void data_tile_ref_count_init_x_2(int ts1,int ts2, double** x_trans, BufferManager *buff_mang_x);
void read_in_data_tile_alloc_tmp_2(int ts1,int ts2, double** tmp_trans, BufferManager *buff_mang_tmp);
void read_in_data_tile_alloc_A_2(int ts1,int ts2, double** A_trans,int A_size_1, BufferManager *buff_mang_A);
void read_in_data_tile_alloc_x_2(int ts1,int ts2, double** x_trans, BufferManager *buff_mang_x);
void data_tile_ref_count_init_y_3(int ts1,int ts2,int ts3, double** y_trans, BufferManager *buff_mang_y);
void data_tile_ref_count_init_A_3(int ts1,int ts2,int ts3, double** A_trans,int A_size_1, BufferManager *buff_mang_A);
void data_tile_ref_count_init_tmp_3(int ts1,int ts2,int ts3, double** tmp_trans, BufferManager *buff_mang_tmp);
void read_in_data_tile_alloc_y_3(int ts1,int ts2,int ts3, double** y_trans, BufferManager *buff_mang_y);
void read_in_data_tile_alloc_A_3(int ts1,int ts2,int ts3, double** A_trans,int A_size_1, BufferManager *buff_mang_A);
void read_in_data_tile_alloc_tmp_3(int ts1,int ts2,int ts3, double** tmp_trans, BufferManager *buff_mang_tmp);
int is_receiver_y_0(int ts1, int ts2, int _PB_NY, int _PB_NX, int my_rank, int nprocs);
void sigma_y_1_0(int ts1, int ts2, int _PB_NY, int _PB_NX, int my_rank, int nprocs, int *receiver_list);
int is_receiver_y_1_0(int ts1, int ts2, int _PB_NY, int _PB_NX, int my_rank, int nprocs);
int pack_y_1_0(int ts1,int ts2,double *send_buf_y,int send_counts_y, double** y_trans, BufferManager *buff_mang_y);
int unpack_y_1_0(int ts1,int ts2,double *recv_buf_y,int curr_displs_y, double** y_trans, BufferManager *buff_mang_y);
void pack_send_0(int my_rank,int nprocs,int t1,int t2, double** y_trans, BufferManager *buff_mang_y,std::vector<double *>& send_buf_y,std::vector<MPI_Request>& send_reqs_y,double *p_total_count);
void unpack_buf_y_0(int my_rank,int nprocs, double** y_trans, BufferManager *buff_mang_y,double *recv_buf_y);
int is_receiver_tmp_1(int ts1, int ts2, int _PB_NY, int _PB_NX, int my_rank, int nprocs);
void sigma_tmp_1_1(int ts1, int ts2, int _PB_NY, int _PB_NX, int my_rank, int nprocs, int *receiver_list);
int is_receiver_tmp_1_1(int ts1, int ts2, int _PB_NY, int _PB_NX, int my_rank, int nprocs);
int pack_tmp_1_1(int ts1,int ts2,double *send_buf_tmp,int send_counts_tmp, double** tmp_trans, BufferManager *buff_mang_tmp);
int unpack_tmp_1_1(int ts1,int ts2,double *recv_buf_tmp,int curr_displs_tmp, double** tmp_trans, BufferManager *buff_mang_tmp);
void pack_send_1(int my_rank,int nprocs,int t1,int t2, double** tmp_trans, BufferManager *buff_mang_tmp,std::vector<double *>& send_buf_tmp,std::vector<MPI_Request>& send_reqs_tmp,double *p_total_count);
void unpack_buf_tmp_1(int my_rank,int nprocs, double** tmp_trans, BufferManager *buff_mang_tmp,double *recv_buf_tmp);
int is_receiver_tmp_2(int ts1, int ts2, int _PB_NY, int _PB_NX, int my_rank, int nprocs);
void sigma_tmp_1_2(int ts1, int ts2, int _PB_NY, int _PB_NX, int my_rank, int nprocs, int *receiver_list);
int is_receiver_tmp_1_2(int ts1, int ts2, int _PB_NY, int _PB_NX, int my_rank, int nprocs);
int pack_tmp_1_2(int ts1,int ts2,double *send_buf_tmp,int send_counts_tmp, double** tmp_trans, BufferManager *buff_mang_tmp);
int unpack_tmp_1_2(int ts1,int ts2,double *recv_buf_tmp,int curr_displs_tmp, double** tmp_trans, BufferManager *buff_mang_tmp);
int is_receiver_A_2(int ts1, int ts2, int _PB_NY, int _PB_NX, int my_rank, int nprocs);
void sigma_A_1_2(int ts1, int ts2, int _PB_NY, int _PB_NX, int my_rank, int nprocs, int *receiver_list);
int is_receiver_A_1_2(int ts1, int ts2, int _PB_NY, int _PB_NX, int my_rank, int nprocs);
int is_receiver_x_2(int ts1, int ts2, int _PB_NY, int _PB_NX, int my_rank, int nprocs);
void sigma_x_1_2(int ts1, int ts2, int _PB_NY, int _PB_NX, int my_rank, int nprocs, int *receiver_list);
int is_receiver_x_1_2(int ts1, int ts2, int _PB_NY, int _PB_NX, int my_rank, int nprocs);
void pack_send_2(int my_rank,int nprocs,int t1,int t2, double** tmp_trans, BufferManager *buff_mang_tmp,std::vector<double *>& send_buf_tmp,std::vector<MPI_Request>& send_reqs_tmp, double** A_trans,int A_size_1, BufferManager *buff_mang_A,std::vector<double *>& send_buf_A,std::vector<MPI_Request>& send_reqs_A, double** x_trans, BufferManager *buff_mang_x,std::vector<double *>& send_buf_x,std::vector<MPI_Request>& send_reqs_x,double *p_total_count);
void unpack_buf_tmp_2(int my_rank,int nprocs, double** tmp_trans, BufferManager *buff_mang_tmp,double *recv_buf_tmp);
void unpack_buf_A_2(int my_rank,int nprocs, double** A_trans,int A_size_1, BufferManager *buff_mang_A,double *recv_buf_A);
void unpack_buf_x_2(int my_rank,int nprocs, double** x_trans, BufferManager *buff_mang_x,double *recv_buf_x);
int is_receiver_y_3(int ts1, int ts2, int ts3, int _PB_NY, int _PB_NX, int my_rank, int nprocs);
void sigma_y_1_3(int ts1, int ts2, int ts3, int _PB_NY, int _PB_NX, int my_rank, int nprocs, int *receiver_list);
int is_receiver_y_1_3(int ts1, int ts2, int ts3, int _PB_NY, int _PB_NX, int my_rank, int nprocs);
int pack_y_1_3(int ts1,int ts2,int ts3,double *send_buf_y,int send_counts_y, double** y_trans, BufferManager *buff_mang_y);
int unpack_y_1_3(int ts1,int ts2,int ts3,double *recv_buf_y,int curr_displs_y, double** y_trans, BufferManager *buff_mang_y);
int is_receiver_A_3(int ts1, int ts2, int ts3, int _PB_NY, int _PB_NX, int my_rank, int nprocs);
void sigma_A_1_3(int ts1, int ts2, int ts3, int _PB_NY, int _PB_NX, int my_rank, int nprocs, int *receiver_list);
int is_receiver_A_1_3(int ts1, int ts2, int ts3, int _PB_NY, int _PB_NX, int my_rank, int nprocs);
int is_receiver_tmp_3(int ts1, int ts2, int ts3, int _PB_NY, int _PB_NX, int my_rank, int nprocs);
void sigma_tmp_1_3(int ts1, int ts2, int ts3, int _PB_NY, int _PB_NX, int my_rank, int nprocs, int *receiver_list);
int is_receiver_tmp_1_3(int ts1, int ts2, int ts3, int _PB_NY, int _PB_NX, int my_rank, int nprocs);
void pack_send_3(int my_rank,int nprocs,int t1,int t2,int t3, double** y_trans, BufferManager *buff_mang_y,std::vector<double *>& send_buf_y,std::vector<MPI_Request>& send_reqs_y, double** A_trans,int A_size_1, BufferManager *buff_mang_A,std::vector<double *>& send_buf_A,std::vector<MPI_Request>& send_reqs_A, double** tmp_trans, BufferManager *buff_mang_tmp,std::vector<double *>& send_buf_tmp,std::vector<MPI_Request>& send_reqs_tmp,double *p_total_count);
void unpack_buf_y_3(int my_rank,int nprocs, double** y_trans, BufferManager *buff_mang_y,double *recv_buf_y);
void unpack_buf_A_3(int my_rank,int nprocs, double** A_trans,int A_size_1, BufferManager *buff_mang_A,double *recv_buf_A);
void unpack_buf_tmp_3(int my_rank,int nprocs, double** tmp_trans, BufferManager *buff_mang_tmp,double *recv_buf_tmp);
int pi_0( int t1, int t5, int _PB_NY, int _PB_NX, int nprocs);
int pi_threads_0( int t1, int t5, int _PB_NY, int _PB_NX, int my_rank, int nprocs, int num_threads);
int pi_1( int t1, int t5, int _PB_NY, int _PB_NX, int nprocs);
int pi_threads_1( int t1, int t5, int _PB_NY, int _PB_NX, int my_rank, int nprocs, int num_threads);
int pi_2( int t1, int t5, int _PB_NY, int _PB_NX, int nprocs);
int pi_threads_2( int t1, int t5, int _PB_NY, int _PB_NX, int my_rank, int nprocs, int num_threads);
int pi_3( int t1, int t5, int t7, int _PB_NY, int _PB_NX, int nprocs);
int pi_threads_3( int t1, int t5, int t7, int _PB_NY, int _PB_NX, int my_rank, int nprocs, int num_threads);
void data_tile_alloc_0(int ts1,int ts2, double** y_trans, BufferManager *buff_mang_y);
void data_tile_ref_count_update_0(int ts1,int ts2, double** y_trans, BufferManager *buff_mang_y);
void compute_task_0(int t1,int t2, double** y_trans, BufferManager *buff_mang_y);
void data_tile_alloc_1(int ts1,int ts2, double** tmp_trans, BufferManager *buff_mang_tmp);
void data_tile_ref_count_update_1(int ts1,int ts2, double** tmp_trans, BufferManager *buff_mang_tmp);
void compute_task_1(int t1,int t2, double** tmp_trans, BufferManager *buff_mang_tmp);
void data_tile_alloc_2(int ts1,int ts2, double** tmp_trans, BufferManager *buff_mang_tmp, double** A_trans,int A_size_1, BufferManager *buff_mang_A, double** x_trans, BufferManager *buff_mang_x);
void data_tile_ref_count_update_2(int ts1,int ts2, double** tmp_trans, BufferManager *buff_mang_tmp, double** A_trans,int A_size_1, BufferManager *buff_mang_A, double** x_trans, BufferManager *buff_mang_x);
void compute_task_2(int t1,int t2, double** tmp_trans, BufferManager *buff_mang_tmp, double** A_trans,int A_size_1, BufferManager *buff_mang_A, double** x_trans, BufferManager *buff_mang_x);
void data_tile_alloc_3(int ts1,int ts2,int ts3, double** y_trans, BufferManager *buff_mang_y, double** A_trans,int A_size_1, BufferManager *buff_mang_A, double** tmp_trans, BufferManager *buff_mang_tmp);
void data_tile_ref_count_update_3(int ts1,int ts2,int ts3, double** y_trans, BufferManager *buff_mang_y, double** A_trans,int A_size_1, BufferManager *buff_mang_A, double** tmp_trans, BufferManager *buff_mang_tmp);
void compute_task_3(int t1,int t2,int t3, double** y_trans, BufferManager *buff_mang_y, double** A_trans,int A_size_1, BufferManager *buff_mang_A, double** tmp_trans, BufferManager *buff_mang_tmp);
