#include <stdio.h>
#include <unistd.h>
#include <sys/time.h>
#include <stdlib.h>

#include <mkl_scalapack.h>
#include <mpi.h>

#define NX 1000
#define NY 1000

#define BLOCK_SIZE 100

#define alpha 1
#define beta 1

/*
#pragma declarations
double A[NX][NY];

double x[NY];
double y[NY];
double tmp[NX];

#pragma enddeclarations
*/

#ifdef TIME
#define IF_TIME(foo) foo;
#else
#define IF_TIME(foo)
#endif

int my_rank = 1;


/* Array initialization. */
static
void init_array ()
{
  int i, j;

  for (i = 0; i < NY; i++)
      x[i] = i * M_PI;
  for (i = 0; i < NX; i++)
    for (j = 0; j < NY; j++)
      A[i][j] = ((DATA_TYPE) i*(j+1)) / NX;
}
static
void print_array()

{
  int i;

  for (i = 0; i < NX; i++) {
    fprintf (stderr, DATA_PRINTF_MODIFIER, y[i]);
    if (i % 20 == 0) fprintf (stderr, "\n");
  }
  fprintf (stderr, "\n");
}
void print_array(double *w, int size)
{
    int i, j;

    for (i=0; i<size; i++) {
        fprintf(stderr, "%0.2lf ", w[i]);
        if (i%80 == 20) fprintf(stderr, "\n");
    }
    fprintf(stderr, "\n");
}

void print_1d_array(double *arr, int size){

    int i;
    int stride = size/BLOCK_SIZE;
    //return;
    if(my_rank != 0)
        return;
 
    fprintf(stderr, "\n------------------------------ \n");
    for(i=0;i<stride;i++){
        fprintf(stderr, "%0.2lf ", arr[i]);
            if (i%80 == 20) fprintf(stderr, "\n");
    }
    fprintf(stderr, "\n------------------------------ \n");
}

void print_2d_array(double *arr){
    int i, j;
    int stride = BLOCK_SIZE;

    if(my_rank != 0)
        return;

    fprintf(stderr, "\n------------------------------ \n");
 
    for(i=0;i<stride;i++){
        for(j=0;j<stride;j++){
            fprintf(stderr, "%0.2lf ", arr[i* stride + j]);
            if (i%80 == 20) fprintf(stderr, "\n");
        }
    }

    fprintf(stderr, "\n------------------------------ \n");
}

double rtclock()
{
    struct timezone Tzp;
    struct timeval Tp;
    int stat;
    stat = gettimeofday (&Tp, &Tzp);
    if (stat != 0) printf("Error return from gettimeofday: %d",stat);
    return(Tp.tv_sec + Tp.tv_usec*1.0e-6);
}

typedef MKL_INT MDESC[ 9 ];
#define MAX(x,y) ((x) >(y)? (x) :(y))
/* Parameters */
const double  zero = 0.0E+0, one = 1.0E+0, two = 2.0E+0, oneeighty = 180.0E+0;
const MKL_INT i_zero = 0, i_one = 1, i_two = 2, i_four = 4, i_negone = -1;
const char    trans_N = 'N';
const char    trans_T = 'T';


void get_procs_grid_size(int p, int *nprow, int *npcol ){

    //printf("p in %d\n", p);
    switch(p) {
        case 1:
            *nprow = 1; *npcol = 1;
            return;
        case 2:
            *nprow = 2; *npcol = 1;
            return;
        case 3:
            *nprow = 3; *npcol = 1;
            return;
        case 4:
            *nprow = 2; *npcol = 2;
            return;
        case 5:
            *nprow = 5; *npcol = 1;
            return;
        case 6:
            *nprow = 3; *npcol = 2;
            return;
        case 7:
            *nprow = 7; *npcol = 1;
            return;
        case 8:
            *nprow = 4; *npcol = 2;
            return;
        case 9:
            *nprow = 3; *npcol = 3;
            return;
        case 10:
            *nprow = 5; *npcol = 2;
            return;
        case 16:
            *nprow = 4; *npcol = 4;
            return;
        case 25:
            *nprow = 5; *npcol = 5;
            return;
        case 32:
            *nprow = 8; *npcol = 4;
            return;
        default:
            *nprow = 1; *npcol = 1;
            return;
    }
}

void generate_random_values(double *arr, int size){

    if(arr == NULL) return;
    int i = 0;
    srand(rand());
    for(i =0; i<size; i++)
        arr[i] = rand()% 100;

    return;
}

int main(int argc, char** argv)
{

    double t_start, t_end;

    double *A_local, *B_local, *x_local, *y_local, *w_local, *z_local, *u1_local;
    double *u2_local, *v1_local, *v2_local;

    double *A, *B, *x, *y, *w, *z, *u1, *u2, *v1, *v2;

/*  Matrix descriptors */
    MDESC   desc_A, desc_B, desc_x, desc_y, desc_u1, desc_u2, desc_v1, desc_v2, desc_w, desc_z;
    MDESC   desc_A_local, desc_B_local, desc_x_local, desc_y_local, desc_u1_local, desc_u2_local, desc_v1_local, desc_v2_local, desc_w_local, desc_z_local;

/*  Local scalars */
    MKL_INT iam, nprocs, ictxt, myrow, mycol, nprow, npcol;
    MKL_INT lld, lld_local, info, n, nb, nq, mp, i, j, nq_vec;

    double alpha_mkl = alpha, beta_mkl = beta;
    int     n_int, nb_int;

    int N =  100;
    int num_procs = atoi(argv[1]);
    //N = atoi(argv[2]);
     //problem size
    N = sqrt(num_procs) * N_PROC_SIZE;
    printf("num procs = %d  problem size = %d\n", num_procs, N);
    n_int = N;


    //block size
    nb_int = BLOCK_SIZE;

    n = (MKL_INT) n_int;
    nb = (MKL_INT) nb_int;

    int USE_ONLY_LOCAL_ARRAYS = 1;

    int n_procs_row = 1, n_procs_col = 1;
  
    get_procs_grid_size(atoi(argv[1]), &n_procs_row, &n_procs_col);

    nprow = (MKL_INT) n_procs_row;
    npcol = (MKL_INT) n_procs_col;

    printf("myrow=%d, mycol=%d\n", nprow, npcol);

/*  Init working 2D process grid */
    blacs_get_( &i_negone, &i_zero, &ictxt );
    blacs_gridinit_( &ictxt, "R", &nprow, &npcol );
    blacs_gridinfo_( &ictxt, &nprow, &npcol, &myrow, &mycol );

    printf("myrow=%d, mycol=%d\n", myrow, mycol);

    if(!USE_ONLY_LOCAL_ARRAYS){
        if((myrow == 0) && (mycol == 0)) {

            A = (double*) calloc( N*N, sizeof( double ) );
            B = (double*) calloc( N*N, sizeof( double ) );
            x = (double*) calloc( N, sizeof( double ) );
            y = (double*) calloc( N, sizeof( double ) );
            w = (double*) calloc( N, sizeof( double ) );
            z = (double*) calloc( N, sizeof( double ) );
            u1 = (double*) calloc( N, sizeof( double ) );
            u2 = (double*) calloc( N, sizeof( double ) );
            v1 = (double*) calloc( N, sizeof( double ) );
            v2 = (double*) calloc( N, sizeof( double ) );

            //init_array();
            for (i=0; i<N; i++) {
                u1[i] = i + 1;
                u2[i] = (i+1)/N/2.0 + 1;
                v1[i] = (i+1)/N/4.0 + 1;
                v2[i] = (i+1)/N/6.0 + 1;
                y[i] = (i+1)/N/8.0;
                z[i] = (i+1)/N/9.0;
                x[i] = 0.0;
                w[i] = 0.0;
                for (j=0; j<N; j++) {
                    A[i*N +j] = ((double) i+j)/N;
                }
            }
        }
    }

/*  Compute precise length of local pieces and allocate array on
    each process for parts of distributed vectors */
    mp = numroc_( &n, &nb, &myrow, &i_zero, &nprow );
    nq = numroc_( &n, &nb, &mycol, &i_zero, &npcol );
    nq_vec = numroc_( &i_one, &i_one, &mycol, &i_zero, &npcol );
    printf("myrow=%d, mycol=%d mp=%d, nq = %d, nq_rhs = %d\n", myrow, mycol, mp, nq, nq_vec);

    if( ( myrow == 0 ) && ( mycol == 0 ) ){
        my_rank = 0;
    }

    A_local = (double*) calloc( mp*nq, sizeof( double ) );
    B_local = (double*) calloc( mp*nq, sizeof( double ) );

/*  Only #0 process column contains distributed 1-d vectors*/
    if ( nq_vec != 0 ){
        x_local = (double *) calloc(mp, sizeof( double ));
        y_local = (double *) calloc(mp, sizeof( double ));
        u1_local = (double *) calloc(mp, sizeof( double ));
        u2_local = (double *) calloc(mp, sizeof( double ));
        v1_local = (double *) calloc(mp, sizeof( double ));
        v2_local = (double *) calloc(mp, sizeof( double ));
        w_local = (double *) calloc(mp, sizeof( double ));
        z_local = (double *) calloc(mp, sizeof( double ));

    } else {                                                                                 
        x_local = NULL;
        y_local = NULL;
        u1_local = NULL;
        u2_local = NULL;
        v1_local = NULL;
        v2_local = NULL;
        w_local = NULL;
        z_local = NULL;
    }

   /*  Compute leading dimensions */
    lld_local = MAX( numroc_( &n, &n, &myrow, &i_zero, &nprow ), 1 );
    lld = MAX( mp, 1 );

    /*  Initialize descriptors for initial arrays located on 0 process */
    descinit_( desc_A, &n, &n, &n, &n, &i_zero, &i_zero, &ictxt, &lld_local, &info );
    descinit_( desc_B, &n, &n, &n, &n, &i_zero, &i_zero, &ictxt, &lld_local, &info );

    descinit_( desc_x, &n, &i_one, &n, &i_one, &i_zero, &i_zero, &ictxt, &lld_local, &info );
    descinit_( desc_y, &n, &i_one, &n, &i_one, &i_zero, &i_zero, &ictxt, &lld_local, &info );
    descinit_( desc_w, &n, &i_one, &n, &i_one, &i_zero, &i_zero, &ictxt, &lld_local, &info );
    descinit_( desc_z, &n, &i_one, &n, &i_one, &i_zero, &i_zero, &ictxt, &lld_local, &info );
    descinit_( desc_u1, &n, &i_one, &n, &i_one, &i_zero, &i_zero, &ictxt, &lld_local, &info );
    descinit_( desc_u2, &n, &i_one, &n, &i_one, &i_zero, &i_zero, &ictxt, &lld_local, &info );
    descinit_( desc_v1, &n, &i_one, &n, &i_one, &i_zero, &i_zero, &ictxt, &lld_local, &info );
    descinit_( desc_v2, &n, &i_one, &n, &i_one, &i_zero, &i_zero, &ictxt, &lld_local, &info );
 
/*  Initialize descriptors for distributed arrays */
    descinit_( desc_A_local, &n, &n, &nb, &nb, &i_zero, &i_zero, &ictxt, &lld, &info );
    descinit_( desc_B_local, &n, &n, &nb, &nb, &i_zero, &i_zero, &ictxt, &lld, &info );

    descinit_( desc_x_local, &n, &i_one, &nb, &i_one, &i_zero, &i_zero, &ictxt, &lld, &info );
    descinit_( desc_y_local, &n, &i_one, &nb, &i_one, &i_zero, &i_zero, &ictxt, &lld, &info );
    descinit_( desc_w_local, &n, &i_one, &nb, &i_one, &i_zero, &i_zero, &ictxt, &lld, &info );
    descinit_( desc_z_local, &n, &i_one, &nb, &i_one, &i_zero, &i_zero, &ictxt, &lld, &info );
    descinit_( desc_u1_local, &n, &i_one, &nb, &i_one, &i_zero, &i_zero, &ictxt, &lld, &info );
    descinit_( desc_u2_local, &n, &i_one, &nb, &i_one, &i_zero, &i_zero, &ictxt, &lld, &info );
    descinit_( desc_v1_local, &n, &i_one, &nb, &i_one, &i_zero, &i_zero, &ictxt, &lld, &info );
    descinit_( desc_v2_local, &n, &i_one, &nb, &i_one, &i_zero, &i_zero, &ictxt, &lld, &info );

/*  Distribute vectors and matrices from 0 process over process grid */
    if(USE_ONLY_LOCAL_ARRAYS){
        generate_random_values(A_local, mp*nq);
        generate_random_values(B_local, mp*nq);
        generate_random_values(x_local, mp);
        generate_random_values(y_local, mp);
        generate_random_values(w_local, mp);
        generate_random_values(z_local, mp);
        generate_random_values(u1_local, mp);
        generate_random_values(u2_local, mp);
        generate_random_values(v1_local, mp);
        generate_random_values(v2_local, mp);

    }
    else {
    pdgeadd_( &trans_N, &n, &n, &one, 
            A, &i_one, &i_one, desc_A,
            &zero, 
            A_local, &i_one, &i_one, desc_A_local );

    pdgeadd_( &trans_N, &n, &n, &one, 
            B, &i_one, &i_one, desc_B,
            &zero,
            B_local, &i_one, &i_one, desc_B_local );

    pdgeadd_( &trans_N, &n, &i_one, &one, x, &i_one, &i_one, desc_x, &zero, x_local, &i_one, &i_one, desc_x_local );
    pdgeadd_( &trans_N, &n, &i_one, &one, y, &i_one, &i_one, desc_y, &zero, y_local, &i_one, &i_one, desc_y_local );
    pdgeadd_( &trans_N, &n, &i_one, &one, w, &i_one, &i_one, desc_w, &zero, w_local, &i_one, &i_one, desc_w_local );
    pdgeadd_( &trans_N, &n, &i_one, &one, z, &i_one, &i_one, desc_z, &zero, z_local, &i_one, &i_one, desc_z_local );

    pdgeadd_( &trans_N, &n, &i_one, &one, u1, &i_one, &i_one, desc_u1, &zero, u1_local, &i_one, &i_one, desc_u1_local );
    pdgeadd_( &trans_N, &n, &i_one, &one, u2, &i_one, &i_one, desc_u2, &zero, u2_local, &i_one, &i_one, desc_u2_local );
    pdgeadd_( &trans_N, &n, &i_one, &one, v1, &i_one, &i_one, desc_v1, &zero, v1_local, &i_one, &i_one, desc_v1_local );
    pdgeadd_( &trans_N, &n, &i_one, &one, v2, &i_one, &i_one, desc_v2, &zero, v2_local, &i_one, &i_one, desc_v2_local );

/*  Destroy arrays on 0 process - they are not necessary anymore */
    if( ( myrow == 0 ) && ( mycol == 0 ) ){
        free(A);
        free(B);
        free(x);
        free(y);
        free(z);
        free(u1);
        free(u2);
        free(v1);
        free(v2);
    
    }
    }

    IF_TIME(t_start = rtclock());
/*  B <- A + u1*v1^T + u2*v2 ^ T */
/*  Copy matrix A to B */

    //print_2d_array(A);
    pdgeadd_( &trans_N, &n, &n, &one, 
            A_local, &i_one, &i_one, desc_A_local,
            &zero,
            B_local, &i_one, &i_one, desc_B_local );

    //MKL_INT size_to_copy = n;
    //printf("n = %d\n", size_to_copy * 2);
    //pdcopy_( &size_to_copy, 
            //A_local, &i_one, &i_one, desc_A_local, &i_one,
            //B_local, &i_one, &i_one, desc_B_local, &i_one );

    //print_2d_array(B);

    pdger_(&n, &n, &one, 
            u1_local, &i_one, &i_one, desc_u1_local, &i_one, 
            v1_local, &i_one, &i_one, desc_v1_local, &i_one, 
            B_local, &i_one, &i_one, desc_B_local);

    //print_2d_array(B);

    pdger_(&n, &n, &one, 
            u2_local, &i_one, &i_one, desc_u2_local, &i_one, 
            v2_local, &i_one, &i_one, desc_v2_local, &i_one, 
            B_local, &i_one, &i_one, desc_B_local);
    
     //print_2d_array(B);

/*  x = beta * B^T * y + z*/
/*  Copy vector z to x */
    //print_1d_array(z_local);
    pdcopy_( &n,
            z_local, &i_one, &i_one, desc_z_local, &i_one,
            x_local, &i_one, &i_one, desc_x_local, &i_one);

    //print_1d_array(x_local);

    pdgemv_(&trans_T, &n, &n, &beta_mkl, 
            B_local, &i_one, &i_one, desc_B_local,
            y_local, &i_one, &i_one, desc_y_local, &i_one,
            &one,
            x_local, &i_one, &i_one, desc_x_local, &i_one);

     //print_1d_array(x_local);

/*  w = alpha* B * z */
    pdgemv_(&trans_N, &n, &n, &alpha_mkl, 
            B_local, &i_one, &i_one, desc_B_local,
            x_local, &i_one, &i_one, desc_x_local, &i_one, 
            &one,
            w_local, &i_one, &i_one, desc_w_local, &i_one);

   //print_1d_array(w_local);
    IF_TIME(t_end = rtclock());
    IF_TIME(fprintf(stdout, " myrow = %d mycol = %d time = %0.6lfs\n", myrow, mycol, t_end - t_start));



    if(!USE_ONLY_LOCAL_ARRAYS){
    pdgeadd_(&trans_N, &n, &i_one, &one, 
            w_local, &i_one, &i_one, desc_w_local,
            &zero,
            w, &i_one, &i_one, desc_w);
    }

if(myrow == 0 && mycol == 0)
	printf("----------------------------------------------------\n");
    //print_1d_array(w);
    /*Destroy process grid */    
    blacs_gridexit_( &ictxt );
    blacs_exit_( &i_zero );

    print_1d_array(w_local, N);
    if(!USE_ONLY_LOCAL_ARRAYS){
    if (fopen(".test", "r")) {
#ifdef MPI
        if (my_rank == 0) {
            print_array(w, N);
        }
#else
        print_array(w, N);
#endif
    }
    }

    return 0;
}

