#include "bicg.dist_dynsched.h"
#include "polyrt.h"
#include "pi_defs.h"
#include <assert.h>
#include <omp.h>
#include <mpi.h>
#include <limits.h>
#include <math.h>
#define ceild(n,d)  ceil(((double)(n))/((double)(d)))
#define floord(n,d) floor(((double)(n))/((double)(d)))
#define max(x,y)    ((x) > (y)? (x) : (y))
#define min(x,y)    ((x) < (y)? (x) : (y))

/**
 * bicg.c: This file is part of the PolyBench/C 3.2 test suite.
 *
 *
 * Contact: Louis-Noel Pouchet <pouchet@cse.ohio-state.edu>
 * Web address: http://polybench.sourceforge.net
 */
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <math.h>

/* Include polybench common header. */
#include <polybench.h>

/* Include benchmark-specific header. */
/* Default data type is double, default size is 4000. */
#include "bicg.h"


/* Array initialization. */
static
void init_array (int nx, int ny,
		 DATA_TYPE POLYBENCH_2D(A,NX,NY,nx,ny),
		 DATA_TYPE POLYBENCH_1D(r,NX,nx),
		 DATA_TYPE POLYBENCH_1D(p,NY,ny))
{
  int i, j;

  for (i = 0; i < ny; i++)
    p[i] = i * M_PI;
  for (i = 0; i < nx; i++) {
    r[i] = i * M_PI;
    for (j = 0; j < ny; j++)
      A[i][j] = ((DATA_TYPE) i*(j+1))/nx;
  }
}


/* DCE code. Must scan the entire live-out data.
   Can be used also to check the correctness of the output. */
static
void print_array(int nx, int ny,
		 DATA_TYPE POLYBENCH_1D(s,NY,ny),
		 DATA_TYPE POLYBENCH_1D(q,NX,nx))

{
  int i;

  for (i = 0; i < ny; i++) {
    fprintf (stderr, DATA_PRINTF_MODIFIER, s[i]);
    if (i % 20 == 0) fprintf (stderr, "\n");
  }
  for (i = 0; i < nx; i++) {
    fprintf (stderr, DATA_PRINTF_MODIFIER, q[i]);
    if (i % 20 == 0) fprintf (stderr, "\n");
  }
  fprintf (stderr, "\n");
}

#define __DECLARATION_OF_A DATA_TYPE POLYBENCH_2D(A,NX,NY,nx,ny)
#define __DECLARATION_OF_s DATA_TYPE POLYBENCH_1D(s,NY,ny)
#define __DECLARATION_OF_q DATA_TYPE POLYBENCH_1D(q,NX,nx)
#define __DECLARATION_OF_p DATA_TYPE POLYBENCH_1D(p,NY,ny)
#define __DECLARATION_OF_r DATA_TYPE POLYBENCH_1D(r,NX,nx)

#define __PLACE_TO_INSERT_FORWARD_DECLARATIONS
int max_num_elements_s;
int max_num_elements_q;
FILE *__debug_print_fp;
tbb::concurrent_priority_queue<__Task, __Compare> pqueue;
int max_num_tasks_loop0_dim0;
int lb_tasks_loop0_dim0;
int *tasks_loop0;
int max_num_tasks_loop1_dim0;
int lb_tasks_loop1_dim0;
int *tasks_loop1;
int max_num_tasks_loop2_dim0;
int lb_tasks_loop2_dim0;
int *tasks_loop2;
int max_num_tasks_loop3_dim0;
int lb_tasks_loop3_dim0;
int *tasks_loop3;
int pack_s_1_0(int ts1,int ts2,double *send_buf_s,int send_counts_s, __DECLARATION_OF_s);
int unpack_s_1_0(int ts1,int ts2,double *recv_buf_s,int curr_displs_s, __DECLARATION_OF_s);
int writeout_pack_s_0(int ts1,int ts2,double *lw_buf_s,int lw_count_s, __DECLARATION_OF_s);
int writeout_unpack_s_0(int ts1,int ts2,double *lw_recv_buf_s,int displs_lw_s, __DECLARATION_OF_s);
void pack_send_0(int my_rank,int nprocs,int t1,int t2,__DECLARATION_OF_s,std::vector<double *>& send_buf_s,std::vector<MPI_Request>& send_reqs_s,double *p_total_count);
void unpack_buf_s_0(int my_rank,int nprocs,__DECLARATION_OF_s,double *recv_buf_s);
int pack_q_1_1(int ts1,int ts2,double *send_buf_q,int send_counts_q, __DECLARATION_OF_q);
int unpack_q_1_1(int ts1,int ts2,double *recv_buf_q,int curr_displs_q, __DECLARATION_OF_q);
int writeout_pack_q_1(int ts1,int ts2,double *lw_buf_q,int lw_count_q, __DECLARATION_OF_q);
int writeout_unpack_q_1(int ts1,int ts2,double *lw_recv_buf_q,int displs_lw_q, __DECLARATION_OF_q);
void pack_send_1(int my_rank,int nprocs,int t1,int t2,__DECLARATION_OF_q,std::vector<double *>& send_buf_q,std::vector<MPI_Request>& send_reqs_q,double *p_total_count);
void unpack_buf_q_1(int my_rank,int nprocs,__DECLARATION_OF_q,double *recv_buf_q);
int writeout_pack_s_2(int ts1,int ts2,double *lw_buf_s,int lw_count_s, __DECLARATION_OF_s);
int writeout_unpack_s_2(int ts1,int ts2,double *lw_recv_buf_s,int displs_lw_s, __DECLARATION_OF_s);
void pack_send_2(int my_rank,int nprocs,int t1,int t2,__DECLARATION_OF_s,std::vector<double *>& send_buf_s,std::vector<MPI_Request>& send_reqs_s,double *p_total_count);
void unpack_buf_s_2(int my_rank,int nprocs,__DECLARATION_OF_s,double *recv_buf_s);
int writeout_pack_q_3(int ts1,int ts2,double *lw_buf_q,int lw_count_q, __DECLARATION_OF_q);
int writeout_unpack_q_3(int ts1,int ts2,double *lw_recv_buf_q,int displs_lw_q, __DECLARATION_OF_q);
void pack_send_3(int my_rank,int nprocs,int t1,int t2,__DECLARATION_OF_q,std::vector<double *>& send_buf_q,std::vector<MPI_Request>& send_reqs_q,double *p_total_count);
void unpack_buf_q_3(int my_rank,int nprocs,__DECLARATION_OF_q,double *recv_buf_q);
void write_out(int my_rank,int nprocs,double *lw_buf_s,double *lw_recv_buf_s,int *displs_lw_s,double *lw_buf_q,double *lw_recv_buf_q,int *displs_lw_q, __DECLARATION_OF_s, __DECLARATION_OF_q);
void compute_task_0(int t1,int t2,__DECLARATION_OF_s);
void compute_task_1(int t1,int t2,__DECLARATION_OF_q);
void compute_task_2(int t1,int t2,__DECLARATION_OF_s,__DECLARATION_OF_r,__DECLARATION_OF_A);
void compute_task_3(int t1,int t2,__DECLARATION_OF_q,__DECLARATION_OF_A,__DECLARATION_OF_p);

/* Main computational kernel. The whole function will be timed,
   including the call and return. */
static
void kernel_bicg(int nx, int ny,
		 DATA_TYPE POLYBENCH_2D(A,NX,NY,nx,ny),
		 DATA_TYPE POLYBENCH_1D(s,NY,ny),
		 DATA_TYPE POLYBENCH_1D(q,NX,nx),
		 DATA_TYPE POLYBENCH_1D(p,NY,ny),
		 DATA_TYPE POLYBENCH_1D(r,NX,nx))
{
  int i, j;

  int t1, t2, t3, t4, t5, t6, t7, t8, t9;
 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;
int _num_threads;
#pragma omp parallel
{_num_threads = omp_get_num_threads();}
#ifndef GLOBAL_MY_RANK
 int my_rank;
#endif
#ifndef __MAX_NUM_RECVS
 #define __MAX_NUM_RECVS (nprocs*_num_threads*16)
#endif
#ifndef __MAX_NUM_SENDS
 #define __MAX_NUM_SENDS (nprocs*_num_threads*8)
#endif
 int nprocs, my_start, my_end, _i, __p, proc, recv_proc, distinct_recv, __tid;
 long int _num_tasks_to_execute, _num_tasks_to_unpack;
 int count;
int req_count;
 int _lb_dist, _ub_dist, lbd_t1, ubd_t1, lbd_t2, ubd_t2, lbd_t3, ubd_t3, lbd_t4, ubd_t4, lbd_t5, ubd_t5, lbd_t6, ubd_t6, lbd_t7, ubd_t7, lbd_t8, ubd_t8, lbd_t9, ubd_t9;
 int thread_support_provided;
 MPI_Init_thread(NULL, NULL, MPI_THREAD_MULTIPLE, &thread_support_provided);
 assert(thread_support_provided == MPI_THREAD_MULTIPLE);
 MPI_Comm_rank(MPI_COMM_WORLD, &my_rank);
 MPI_Comm_size(MPI_COMM_WORLD, &nprocs);
 int __tid_receiver;
 int waiting_pops[_num_threads], waiting_pops_threads, waiting_pops_global;
 for (__tid=0; __tid<_num_threads; __tid++) waiting_pops[__tid] = 0;
 double t_tasks_create, t_globaltasks_create;
 double t_tasks_manage_start[_num_threads], t_tasks_manage[_num_threads], t_tasks_manage_threads, t_globaltasks_manage;
 for (__tid=0; __tid<_num_threads; __tid++) t_tasks_manage[__tid] = 0;
 double t_comp_start[_num_threads], t_comp[_num_threads], t_comp_threads_min, t_globalcomp_min, t_comp_threads, t_globalcomp;
 double t_comp_mean, t_comp_sd, t_globalcomp_mean, t_globalcomp_sd;
 for (__tid=0; __tid<_num_threads; __tid++) t_comp[__tid] = 0;
 double t_wait_comp_start[_num_threads], t_wait_comp[_num_threads], t_wait_comp_threads, t_globalwait_comp;
 for (__tid=0; __tid<_num_threads; __tid++) t_wait_comp[__tid] = 0;
 double t_pack_start[_num_threads], t_pack[_num_threads], t_pack_threads, t_globalpack;
 for (__tid=0; __tid<_num_threads; __tid++) t_pack[__tid] = 0;
 double t_writeout = 0.0;
 double t_unpack_start, t_unpack = 0.0, t_globalunpack = 0.0;
 double t_wait_unpack_start, t_wait_unpack = 0.0, t_globalwait_unpack = 0.0;
 double t_local = 0.0, t_global = 0.0;
 double __total_count[_num_threads];
 for (__tid=0; __tid<_num_threads; __tid++) __total_count[__tid] = 0;
 double __total_count_threads = 0;
 double __total_count_all = 0;
std::vector<MPI_Request> send_reqs_s[_num_threads];
std::vector<MPI_Request> recv_reqs_s;
std::vector<int> recv_completed_s;
                        for (__tid=0; __tid<_num_threads; __tid++) {
                        for (__p=0; __p<__MAX_NUM_SENDS; __p++) {
                        send_reqs_s[__tid].push_back(MPI_REQUEST_NULL);
}
}
                    for (__p=0; __p<__MAX_NUM_RECVS; __p++) {
                    recv_reqs_s.push_back(MPI_REQUEST_NULL);
                    recv_completed_s.push_back(0);
                    }
int lw_count_s = 0;
int send_counts_s[__tid][nprocs];
max_num_elements_s = 0;
int recv_counts_s[nprocs];
int lw_recv_counts_s[nprocs];
int displs_s[nprocs];
int displs_lw_s[nprocs];
std::vector<double *> send_buf_s[_num_threads];
std::vector<size_t> send_buf_size_s[_num_threads];
for (__tid=0; __tid<_num_threads; __tid++) {
      for (__p=0; __p<__MAX_NUM_SENDS; __p++) {
      send_buf_s[__tid].push_back(NULL);
      send_buf_size_s[__tid].push_back(0);
}
}std::vector<double *> recv_buf_s;
size_t recv_buf_size_s[__MAX_NUM_RECVS];
for (__p=0; __p<__MAX_NUM_RECVS; __p++) {
                    recv_buf_s.push_back(NULL);
                    recv_buf_size_s[__p] = 0;
}double *lw_buf_s = NULL;
size_t lw_buf_size_s = 0;
double *lw_recv_buf_s = NULL;
size_t lw_recv_buf_size_s = 0;
std::vector<MPI_Request> send_reqs_q[_num_threads];
std::vector<MPI_Request> recv_reqs_q;
std::vector<int> recv_completed_q;
                        for (__tid=0; __tid<_num_threads; __tid++) {
                        for (__p=0; __p<__MAX_NUM_SENDS; __p++) {
                        send_reqs_q[__tid].push_back(MPI_REQUEST_NULL);
}
}
                    for (__p=0; __p<__MAX_NUM_RECVS; __p++) {
                    recv_reqs_q.push_back(MPI_REQUEST_NULL);
                    recv_completed_q.push_back(0);
                    }
int lw_count_q = 0;
int send_counts_q[__tid][nprocs];
max_num_elements_q = 0;
int recv_counts_q[nprocs];
int lw_recv_counts_q[nprocs];
int displs_q[nprocs];
int displs_lw_q[nprocs];
std::vector<double *> send_buf_q[_num_threads];
std::vector<size_t> send_buf_size_q[_num_threads];
for (__tid=0; __tid<_num_threads; __tid++) {
      for (__p=0; __p<__MAX_NUM_SENDS; __p++) {
      send_buf_q[__tid].push_back(NULL);
      send_buf_size_q[__tid].push_back(0);
}
}std::vector<double *> recv_buf_q;
size_t recv_buf_size_q[__MAX_NUM_RECVS];
for (__p=0; __p<__MAX_NUM_RECVS; __p++) {
                    recv_buf_q.push_back(NULL);
                    recv_buf_size_q[__p] = 0;
}double *lw_buf_q = NULL;
size_t lw_buf_size_q = 0;
double *lw_recv_buf_q = NULL;
size_t lw_recv_buf_size_q = 0;
max_num_tasks_loop0_dim0 = ceil(floorf((+1*_PB_NY-1)/(float)32) - 0 + 1);
lb_tasks_loop0_dim0 = floor(0);
tasks_loop0 = (int *)malloc(max_num_tasks_loop0_dim0*sizeof(int));
max_num_tasks_loop1_dim0 = ceil(floorf((+1*_PB_NX-1)/(float)32) - 0 + 1);
lb_tasks_loop1_dim0 = floor(0);
tasks_loop1 = (int *)malloc(max_num_tasks_loop1_dim0*sizeof(int));
max_num_tasks_loop2_dim0 = ceil(floorf((+1*_PB_NY-1)/(float)32) - 0 + 1);
lb_tasks_loop2_dim0 = floor(0);
tasks_loop2 = (int *)malloc(max_num_tasks_loop2_dim0*sizeof(int));
max_num_tasks_loop3_dim0 = ceil(floorf((+1*_PB_NX-1)/(float)32) - 0 + 1);
lb_tasks_loop3_dim0 = floor(0);
tasks_loop3 = (int *)malloc(max_num_tasks_loop3_dim0*sizeof(int));
            max_num_elements_s = max(max_num_elements_s, (+1*(32)) + 3);
                for (__tid=0; __tid<_num_threads; __tid++) { for (__p=0; __p<__MAX_NUM_SENDS; __p++) {
                send_buf_s[__tid][__p] = (double *) polyrt_max_alloc(send_buf_s[__tid][__p], sizeof(double)*max_num_elements_s, &send_buf_size_s[__tid][__p]);
}
}
for (__p=0; __p<__MAX_NUM_RECVS; __p++) { recv_buf_s[__p] = (double *) polyrt_max_alloc(recv_buf_s[__p], sizeof(double)*max_num_elements_s, &recv_buf_size_s[__p]); }
if (fopen(".test", "r")) {lw_buf_s = (double *) polyrt_max_alloc(lw_buf_s, sizeof(double)*((+1*(32))*floorf((1*(floorf((+1*_PB_NY-1)/(float)32) - 0 + 1+nprocs))/(float)nprocs)), &lw_buf_size_s);
lw_recv_buf_s = (double *) polyrt_max_alloc(lw_recv_buf_s, nprocs*lw_buf_size_s, &lw_recv_buf_size_s);
 for (__p=0; __p<nprocs; __p++) { displs_lw_s[__p] = __p*lw_buf_size_s/sizeof(double);}
} max_num_elements_q = max(max_num_elements_q, (+1*(32)) + 3);
                for (__tid=0; __tid<_num_threads; __tid++) { for (__p=0; __p<__MAX_NUM_SENDS; __p++) {
                send_buf_q[__tid][__p] = (double *) polyrt_max_alloc(send_buf_q[__tid][__p], sizeof(double)*max_num_elements_q, &send_buf_size_q[__tid][__p]);
}
}
for (__p=0; __p<__MAX_NUM_RECVS; __p++) { recv_buf_q[__p] = (double *) polyrt_max_alloc(recv_buf_q[__p], sizeof(double)*max_num_elements_q, &recv_buf_size_q[__p]); }
if (fopen(".test", "r")) {lw_buf_q = (double *) polyrt_max_alloc(lw_buf_q, sizeof(double)*((+1*(32))*floorf((1*(floorf((+1*_PB_NX-1)/(float)32) - 0 + 1+nprocs))/(float)nprocs)), &lw_buf_size_q);
lw_recv_buf_q = (double *) polyrt_max_alloc(lw_recv_buf_q, nprocs*lw_buf_size_q, &lw_recv_buf_size_q);
 for (__p=0; __p<nprocs; __p++) { displs_lw_q[__p] = __p*lw_buf_size_q/sizeof(double);}
} max_num_elements_s = max(max_num_elements_s, (+1*(1)) + 3);
                for (__tid=0; __tid<_num_threads; __tid++) { for (__p=0; __p<__MAX_NUM_SENDS; __p++) {
                send_buf_s[__tid][__p] = (double *) polyrt_max_alloc(send_buf_s[__tid][__p], sizeof(double)*max_num_elements_s, &send_buf_size_s[__tid][__p]);
}
}
for (__p=0; __p<__MAX_NUM_RECVS; __p++) { recv_buf_s[__p] = (double *) polyrt_max_alloc(recv_buf_s[__p], sizeof(double)*max_num_elements_s, &recv_buf_size_s[__p]); }
if (fopen(".test", "r")) {lw_buf_s = (double *) polyrt_max_alloc(lw_buf_s, sizeof(double)*((+1*(32))*floorf((1*(floorf((+1*_PB_NY-1)/(float)32) - 0 + 1+nprocs))/(float)nprocs)), &lw_buf_size_s);
lw_recv_buf_s = (double *) polyrt_max_alloc(lw_recv_buf_s, nprocs*lw_buf_size_s, &lw_recv_buf_size_s);
 for (__p=0; __p<nprocs; __p++) { displs_lw_s[__p] = __p*lw_buf_size_s/sizeof(double);}
} max_num_elements_q = max(max_num_elements_q, (+1*(1)) + 3);
                for (__tid=0; __tid<_num_threads; __tid++) { for (__p=0; __p<__MAX_NUM_SENDS; __p++) {
                send_buf_q[__tid][__p] = (double *) polyrt_max_alloc(send_buf_q[__tid][__p], sizeof(double)*max_num_elements_q, &send_buf_size_q[__tid][__p]);
}
}
for (__p=0; __p<__MAX_NUM_RECVS; __p++) { recv_buf_q[__p] = (double *) polyrt_max_alloc(recv_buf_q[__p], sizeof(double)*max_num_elements_q, &recv_buf_size_q[__p]); }
if (fopen(".test", "r")) {lw_buf_q = (double *) polyrt_max_alloc(lw_buf_q, sizeof(double)*((+1*(32))*floorf((1*(floorf((+1*_PB_NX-1)/(float)32) - 0 + 1+nprocs))/(float)nprocs)), &lw_buf_size_q);
lw_recv_buf_q = (double *) polyrt_max_alloc(lw_recv_buf_q, nprocs*lw_buf_size_q, &lw_recv_buf_size_q);
 for (__p=0; __p<nprocs; __p++) { displs_lw_q[__p] = __p*lw_buf_size_q/sizeof(double);}
} polyrt_init_grid_size();
 IF_TIME(t_local = rtclock());
/*Start of dynamic scheduling code*/
#if defined(__DYNSCHEDULER_DEBUG_PRINT) || defined(__DYNSCHEDULER_MORE_DEBUG_PRINT) || defined(__DEBUG_FLUSH)
 char *debug_print_name = (char *)malloc(128*sizeof(char));
 strcpy(debug_print_name, "debug_print");
 sprintf(debug_print_name+strlen(debug_print_name), "_node%d", my_rank);
 __debug_print_fp = fopen(debug_print_name, "w");
 free(debug_print_name);
#endif
 IF_TIME(t_tasks_create = rtclock());
 init_tasks(my_rank,nprocs,_num_threads,tasks_loop0,tasks_loop1,tasks_loop2,tasks_loop3,&_num_tasks_to_execute,&_num_tasks_to_unpack);
 IF_TIME(t_tasks_create = rtclock() - t_tasks_create);
#ifdef __DYNSCHEDULER_MORE_DEBUG_PRINT
 std::vector<int> prev_firing_count_s_0;
 std::vector<int> prev_firing_count_s_1;
 std::vector<int> prev_firing_count_s_2;
 std::vector<int> prev_firing_count_s_3;
 std::vector<int> prev_firing_count_q_0;
 std::vector<int> prev_firing_count_q_1;
 std::vector<int> prev_firing_count_q_2;
 std::vector<int> prev_firing_count_q_3;
 for (__p=0; __p<__MAX_NUM_RECVS; __p++) {
  prev_firing_count_s_0.push_back(-1);
  prev_firing_count_s_1.push_back(-1);
  prev_firing_count_s_2.push_back(-1);
  prev_firing_count_s_3.push_back(-1);
  prev_firing_count_q_0.push_back(-1);
  prev_firing_count_q_1.push_back(-1);
  prev_firing_count_q_2.push_back(-1);
  prev_firing_count_q_3.push_back(-1);
 }
#endif
 IF_TIME(t_unpack_start = rtclock());
 for (__p=0; __p<__MAX_NUM_RECVS; __p++) {
  int __error;
  __error = MPI_Irecv(recv_buf_s[__p], max_num_elements_s, MPI_DOUBLE, MPI_ANY_SOURCE, 3123, MPI_COMM_WORLD, &recv_reqs_s[__p]);
#ifdef __DYNSCHEDULER_DEBUG_PRINT
  if (__error) {
   char *msg=(char*)malloc((MPI_MAX_ERROR_STRING+1)*sizeof(char)); int len;
   MPI_Error_string(__error, msg, &len);
   fprintf(__debug_print_fp, "node %d s recv buf %d error %d %s\n", my_rank, __p, __error, msg);
   free(msg);
  }
#endif
  __error = MPI_Irecv(recv_buf_q[__p], max_num_elements_q, MPI_DOUBLE, MPI_ANY_SOURCE, 5123, MPI_COMM_WORLD, &recv_reqs_q[__p]);
#ifdef __DYNSCHEDULER_DEBUG_PRINT
  if (__error) {
   char *msg=(char*)malloc((MPI_MAX_ERROR_STRING+1)*sizeof(char)); int len;
   MPI_Error_string(__error, msg, &len);
   fprintf(__debug_print_fp, "node %d q recv buf %d error %d %s\n", my_rank, __p, __error, msg);
   free(msg);
  }
#endif
 }
 IF_TIME(t_unpack += rtclock() - t_unpack_start);
#ifdef __DYNSCHEDULER_DEBUG_PRINT
 fprintf(__debug_print_fp, "node %d num tasks to execute %ld num tasks to unpack %ld max recv buffers %d s max elements %d q max elements %d\n", my_rank, _num_tasks_to_execute, _num_tasks_to_unpack, __MAX_NUM_RECVS, max_num_elements_s, max_num_elements_q);
 IF_DEBUG_FLUSH(fflush(__debug_print_fp));
#endif
#pragma omp parallel private(__tid)
{
 __tid = omp_get_thread_num();
 int __flag;
 int __loop;
 int task_id1,task_id2;
#ifdef __DYNSCHEDULER_DEBUG_PRINT
 fprintf(__debug_print_fp, "node %d thread %d\n", my_rank, __tid);
 IF_DEBUG_FLUSH(fflush(__debug_print_fp));
#endif
 IF_TIME(t_wait_comp_start[__tid] = rtclock());
#ifdef __DYNSCHEDULER_DEBUG_PRINT
 int __busy_wait_count = 0;
#endif
#pragma omp single nowait
{ /* receive-unpack */
 __tid_receiver = __tid;
 IF_TIME(t_wait_unpack_start = rtclock());
 while(_num_tasks_to_unpack > 0) {
#ifdef __DYNSCHEDULER_DEBUG_PRINT
#pragma omp atomic
  __busy_wait_count++;
#endif
  for (__p=0; (__p<recv_buf_s.size()) || (__p<recv_buf_q.size()); __p++) {
  if (__p<recv_buf_s.size()) {
   if (recv_reqs_s[__p] != MPI_REQUEST_NULL) {
#ifndef __DYNSCHEDULER_DEBUG_PRINT
    MPI_Test(&recv_reqs_s[__p],&recv_completed_s[__p],MPI_STATUS_IGNORE);
#else
    MPI_Status recv_stats_s;
    recv_stats_s.MPI_ERROR = MPI_Test(&recv_reqs_s[__p],&recv_completed_s[__p],&recv_stats_s);
    if (recv_completed_s[__p]) {
     __loop = (int)recv_buf_s[__p][0];
     task_id1 = (int)recv_buf_s[__p][1];
     task_id2 = (int)recv_buf_s[__p][2];
     int size;
     MPI_Get_count(&recv_stats_s, MPI_DOUBLE, &size);
     if (recv_stats_s.MPI_ERROR != MPI_SUCCESS) {
      char *msg=(char*)malloc((MPI_MAX_ERROR_STRING+1)*sizeof(char)); int len;
      MPI_Error_string(recv_stats_s.MPI_ERROR, msg, &len);
      fprintf(__debug_print_fp, "received s node %d from node %d buf %d size %d loop %d task %d %d error %d %s\n", my_rank, recv_stats_s.MPI_SOURCE, __p, size, __loop, task_id1, task_id2, recv_stats_s.MPI_ERROR, msg);
      free(msg);
     } else {
      fprintf(__debug_print_fp, "received s node %d from node %d buf %d size %d loop %d task %d %d\n", my_rank, recv_stats_s.MPI_SOURCE, __p, size, __loop, task_id1, task_id2);
     }
     IF_DEBUG_FLUSH(fflush(__debug_print_fp));
    }
#endif
   }
   if (recv_completed_s[__p]) {
    __loop = (int)recv_buf_s[__p][0];
    task_id1 = (int)recv_buf_s[__p][1];
    task_id2 = (int)recv_buf_s[__p][2];
    switch(__loop){
     case 0:
#ifdef __DYNSCHEDULER_MORE_DEBUG_PRINT
      if (prev_firing_count_s_0[__p] != tasks_loop0[(task_id2-lb_tasks_loop0_dim0)+0]) {
       prev_firing_count_s_0[__p] = tasks_loop0[(task_id2-lb_tasks_loop0_dim0)+0];
       fprintf(__debug_print_fp, "waiting to unpack s node %d buf %d loop %d task %d %d count %d\n", my_rank, __p, __loop, task_id1, task_id2, tasks_loop0[(task_id2-lb_tasks_loop0_dim0)+0]);
       IF_DEBUG_FLUSH(fflush(__debug_print_fp));
      }
#endif
      if (tasks_loop0[(task_id2-lb_tasks_loop0_dim0)+0] <= 1) {
       IF_TIME(t_wait_unpack += rtclock() - t_wait_unpack_start);
       IF_TIME(t_unpack_start = rtclock());
       if (tasks_loop0[(task_id2-lb_tasks_loop0_dim0)+0] == 1) {
        tasks_loop0[(task_id2-lb_tasks_loop0_dim0)+0] -= !is_receiver_s_0(task_id1,task_id2,_PB_NY,_PB_NX,my_rank,nprocs);
       }
       unpack_buf_s_0(my_rank,nprocs,s,recv_buf_s[__p]);
       tasks_loop0[(task_id2-lb_tasks_loop0_dim0)+0]--;
       recv_completed_s[__p] = 0;
       int __error = MPI_Irecv(recv_buf_s[__p], max_num_elements_s, MPI_DOUBLE, MPI_ANY_SOURCE, 3123, MPI_COMM_WORLD, &recv_reqs_s[__p]);
#ifdef __DYNSCHEDULER_DEBUG_PRINT
       if (__error) {
        char *msg=(char*)malloc((MPI_MAX_ERROR_STRING+1)*sizeof(char)); int len;
        MPI_Error_string(__error, msg, &len);
        fprintf(__debug_print_fp, "node %d s recv buf %d error %d %s\n", my_rank, __p, __error, msg);
        free(msg);
       }
#pragma omp atomic
       __busy_wait_count *= 0;
       fprintf(__debug_print_fp, "unpacked s node %d buf %d loop %d task %d %d count %d\n", my_rank, __p, __loop, task_id1, task_id2, tasks_loop0[(task_id2-lb_tasks_loop0_dim0)+0]);
       IF_DEBUG_FLUSH(fflush(__debug_print_fp));
#ifdef __DYNSCHEDULER_MORE_DEBUG_PRINT
       prev_firing_count_s_0[__p] = -1;
#endif
#endif
       if (tasks_loop0[(task_id2-lb_tasks_loop0_dim0)+0] == 0) {
        remote_update_dep_tasks_0(task_id1,task_id2,_PB_NY,_PB_NX,my_rank,nprocs,_num_threads,tasks_loop0,tasks_loop1,tasks_loop2,tasks_loop3);
        _num_tasks_to_unpack--;
#ifdef __DYNSCHEDULER_DEBUG_PRINT
        fprintf(__debug_print_fp, "remote updated s node %d loop %d task %d %d num tasks left to unpack %ld\n", my_rank, __loop, task_id1, task_id2, _num_tasks_to_unpack);
        IF_DEBUG_FLUSH(fflush(__debug_print_fp));
#endif
       }
       IF_TIME(t_unpack += rtclock() - t_unpack_start);
       IF_TIME(t_wait_unpack_start = rtclock());
      }
      break;
     case 2:
#ifdef __DYNSCHEDULER_MORE_DEBUG_PRINT
      if (prev_firing_count_s_2[__p] != tasks_loop2[(task_id2-lb_tasks_loop2_dim0)+0]) {
       prev_firing_count_s_2[__p] = tasks_loop2[(task_id2-lb_tasks_loop2_dim0)+0];
       fprintf(__debug_print_fp, "waiting to unpack s node %d buf %d loop %d task %d %d count %d\n", my_rank, __p, __loop, task_id1, task_id2, tasks_loop2[(task_id2-lb_tasks_loop2_dim0)+0]);
       IF_DEBUG_FLUSH(fflush(__debug_print_fp));
      }
#endif
      if (tasks_loop2[(task_id2-lb_tasks_loop2_dim0)+0] <= 1) {
       IF_TIME(t_wait_unpack += rtclock() - t_wait_unpack_start);
       IF_TIME(t_unpack_start = rtclock());
       if (tasks_loop2[(task_id2-lb_tasks_loop2_dim0)+0] == 1) {
        tasks_loop2[(task_id2-lb_tasks_loop2_dim0)+0] -= !is_receiver_s_2(task_id1,task_id2,_PB_NY,_PB_NX,my_rank,nprocs);
       }
       unpack_buf_s_2(my_rank,nprocs,s,recv_buf_s[__p]);
       tasks_loop2[(task_id2-lb_tasks_loop2_dim0)+0]--;
       recv_completed_s[__p] = 0;
       int __error = MPI_Irecv(recv_buf_s[__p], max_num_elements_s, MPI_DOUBLE, MPI_ANY_SOURCE, 3123, MPI_COMM_WORLD, &recv_reqs_s[__p]);
#ifdef __DYNSCHEDULER_DEBUG_PRINT
       if (__error) {
        char *msg=(char*)malloc((MPI_MAX_ERROR_STRING+1)*sizeof(char)); int len;
        MPI_Error_string(__error, msg, &len);
        fprintf(__debug_print_fp, "node %d s recv buf %d error %d %s\n", my_rank, __p, __error, msg);
        free(msg);
       }
#pragma omp atomic
       __busy_wait_count *= 0;
       fprintf(__debug_print_fp, "unpacked s node %d buf %d loop %d task %d %d count %d\n", my_rank, __p, __loop, task_id1, task_id2, tasks_loop2[(task_id2-lb_tasks_loop2_dim0)+0]);
       IF_DEBUG_FLUSH(fflush(__debug_print_fp));
#ifdef __DYNSCHEDULER_MORE_DEBUG_PRINT
       prev_firing_count_s_2[__p] = -1;
#endif
#endif
       if (tasks_loop2[(task_id2-lb_tasks_loop2_dim0)+0] == 0) {
        remote_update_dep_tasks_2(task_id1,task_id2,_PB_NY,_PB_NX,my_rank,nprocs,_num_threads,tasks_loop0,tasks_loop1,tasks_loop2,tasks_loop3);
        _num_tasks_to_unpack--;
#ifdef __DYNSCHEDULER_DEBUG_PRINT
        fprintf(__debug_print_fp, "remote updated s node %d loop %d task %d %d num tasks left to unpack %ld\n", my_rank, __loop, task_id1, task_id2, _num_tasks_to_unpack);
        IF_DEBUG_FLUSH(fflush(__debug_print_fp));
#endif
       }
       IF_TIME(t_unpack += rtclock() - t_unpack_start);
       IF_TIME(t_wait_unpack_start = rtclock());
      }
      break;
     default:
      fprintf(stderr, "ERROR: s data cannot be sent by a task of loop %d\n", __loop);
      assert(0);
    }
   }
  }
  if (__p<recv_buf_q.size()) {
   if (recv_reqs_q[__p] != MPI_REQUEST_NULL) {
#ifndef __DYNSCHEDULER_DEBUG_PRINT
    MPI_Test(&recv_reqs_q[__p],&recv_completed_q[__p],MPI_STATUS_IGNORE);
#else
    MPI_Status recv_stats_q;
    recv_stats_q.MPI_ERROR = MPI_Test(&recv_reqs_q[__p],&recv_completed_q[__p],&recv_stats_q);
    if (recv_completed_q[__p]) {
     __loop = (int)recv_buf_q[__p][0];
     task_id1 = (int)recv_buf_q[__p][1];
     task_id2 = (int)recv_buf_q[__p][2];
     int size;
     MPI_Get_count(&recv_stats_q, MPI_DOUBLE, &size);
     if (recv_stats_q.MPI_ERROR != MPI_SUCCESS) {
      char *msg=(char*)malloc((MPI_MAX_ERROR_STRING+1)*sizeof(char)); int len;
      MPI_Error_string(recv_stats_q.MPI_ERROR, msg, &len);
      fprintf(__debug_print_fp, "received q node %d from node %d buf %d size %d loop %d task %d %d error %d %s\n", my_rank, recv_stats_q.MPI_SOURCE, __p, size, __loop, task_id1, task_id2, recv_stats_q.MPI_ERROR, msg);
      free(msg);
     } else {
      fprintf(__debug_print_fp, "received q node %d from node %d buf %d size %d loop %d task %d %d\n", my_rank, recv_stats_q.MPI_SOURCE, __p, size, __loop, task_id1, task_id2);
     }
     IF_DEBUG_FLUSH(fflush(__debug_print_fp));
    }
#endif
   }
   if (recv_completed_q[__p]) {
    __loop = (int)recv_buf_q[__p][0];
    task_id1 = (int)recv_buf_q[__p][1];
    task_id2 = (int)recv_buf_q[__p][2];
    switch(__loop){
     case 1:
#ifdef __DYNSCHEDULER_MORE_DEBUG_PRINT
      if (prev_firing_count_q_1[__p] != tasks_loop1[(task_id2-lb_tasks_loop1_dim0)+0]) {
       prev_firing_count_q_1[__p] = tasks_loop1[(task_id2-lb_tasks_loop1_dim0)+0];
       fprintf(__debug_print_fp, "waiting to unpack q node %d buf %d loop %d task %d %d count %d\n", my_rank, __p, __loop, task_id1, task_id2, tasks_loop1[(task_id2-lb_tasks_loop1_dim0)+0]);
       IF_DEBUG_FLUSH(fflush(__debug_print_fp));
      }
#endif
      if (tasks_loop1[(task_id2-lb_tasks_loop1_dim0)+0] <= 1) {
       IF_TIME(t_wait_unpack += rtclock() - t_wait_unpack_start);
       IF_TIME(t_unpack_start = rtclock());
       if (tasks_loop1[(task_id2-lb_tasks_loop1_dim0)+0] == 1) {
        tasks_loop1[(task_id2-lb_tasks_loop1_dim0)+0] -= !is_receiver_q_1(task_id1,task_id2,_PB_NY,_PB_NX,my_rank,nprocs);
       }
       unpack_buf_q_1(my_rank,nprocs,q,recv_buf_q[__p]);
       tasks_loop1[(task_id2-lb_tasks_loop1_dim0)+0]--;
       recv_completed_q[__p] = 0;
       int __error = MPI_Irecv(recv_buf_q[__p], max_num_elements_q, MPI_DOUBLE, MPI_ANY_SOURCE, 5123, MPI_COMM_WORLD, &recv_reqs_q[__p]);
#ifdef __DYNSCHEDULER_DEBUG_PRINT
       if (__error) {
        char *msg=(char*)malloc((MPI_MAX_ERROR_STRING+1)*sizeof(char)); int len;
        MPI_Error_string(__error, msg, &len);
        fprintf(__debug_print_fp, "node %d q recv buf %d error %d %s\n", my_rank, __p, __error, msg);
        free(msg);
       }
#pragma omp atomic
       __busy_wait_count *= 0;
       fprintf(__debug_print_fp, "unpacked q node %d buf %d loop %d task %d %d count %d\n", my_rank, __p, __loop, task_id1, task_id2, tasks_loop1[(task_id2-lb_tasks_loop1_dim0)+0]);
       IF_DEBUG_FLUSH(fflush(__debug_print_fp));
#ifdef __DYNSCHEDULER_MORE_DEBUG_PRINT
       prev_firing_count_q_1[__p] = -1;
#endif
#endif
       if (tasks_loop1[(task_id2-lb_tasks_loop1_dim0)+0] == 0) {
        remote_update_dep_tasks_1(task_id1,task_id2,_PB_NY,_PB_NX,my_rank,nprocs,_num_threads,tasks_loop0,tasks_loop1,tasks_loop2,tasks_loop3);
        _num_tasks_to_unpack--;
#ifdef __DYNSCHEDULER_DEBUG_PRINT
        fprintf(__debug_print_fp, "remote updated q node %d loop %d task %d %d num tasks left to unpack %ld\n", my_rank, __loop, task_id1, task_id2, _num_tasks_to_unpack);
        IF_DEBUG_FLUSH(fflush(__debug_print_fp));
#endif
       }
       IF_TIME(t_unpack += rtclock() - t_unpack_start);
       IF_TIME(t_wait_unpack_start = rtclock());
      }
      break;
     case 3:
#ifdef __DYNSCHEDULER_MORE_DEBUG_PRINT
      if (prev_firing_count_q_3[__p] != tasks_loop3[(task_id2-lb_tasks_loop3_dim0)+0]) {
       prev_firing_count_q_3[__p] = tasks_loop3[(task_id2-lb_tasks_loop3_dim0)+0];
       fprintf(__debug_print_fp, "waiting to unpack q node %d buf %d loop %d task %d %d count %d\n", my_rank, __p, __loop, task_id1, task_id2, tasks_loop3[(task_id2-lb_tasks_loop3_dim0)+0]);
       IF_DEBUG_FLUSH(fflush(__debug_print_fp));
      }
#endif
      if (tasks_loop3[(task_id2-lb_tasks_loop3_dim0)+0] <= 1) {
       IF_TIME(t_wait_unpack += rtclock() - t_wait_unpack_start);
       IF_TIME(t_unpack_start = rtclock());
       if (tasks_loop3[(task_id2-lb_tasks_loop3_dim0)+0] == 1) {
        tasks_loop3[(task_id2-lb_tasks_loop3_dim0)+0] -= !is_receiver_q_3(task_id1,task_id2,_PB_NY,_PB_NX,my_rank,nprocs);
       }
       unpack_buf_q_3(my_rank,nprocs,q,recv_buf_q[__p]);
       tasks_loop3[(task_id2-lb_tasks_loop3_dim0)+0]--;
       recv_completed_q[__p] = 0;
       int __error = MPI_Irecv(recv_buf_q[__p], max_num_elements_q, MPI_DOUBLE, MPI_ANY_SOURCE, 5123, MPI_COMM_WORLD, &recv_reqs_q[__p]);
#ifdef __DYNSCHEDULER_DEBUG_PRINT
       if (__error) {
        char *msg=(char*)malloc((MPI_MAX_ERROR_STRING+1)*sizeof(char)); int len;
        MPI_Error_string(__error, msg, &len);
        fprintf(__debug_print_fp, "node %d q recv buf %d error %d %s\n", my_rank, __p, __error, msg);
        free(msg);
       }
#pragma omp atomic
       __busy_wait_count *= 0;
       fprintf(__debug_print_fp, "unpacked q node %d buf %d loop %d task %d %d count %d\n", my_rank, __p, __loop, task_id1, task_id2, tasks_loop3[(task_id2-lb_tasks_loop3_dim0)+0]);
       IF_DEBUG_FLUSH(fflush(__debug_print_fp));
#ifdef __DYNSCHEDULER_MORE_DEBUG_PRINT
       prev_firing_count_q_3[__p] = -1;
#endif
#endif
       if (tasks_loop3[(task_id2-lb_tasks_loop3_dim0)+0] == 0) {
        remote_update_dep_tasks_3(task_id1,task_id2,_PB_NY,_PB_NX,my_rank,nprocs,_num_threads,tasks_loop0,tasks_loop1,tasks_loop2,tasks_loop3);
        _num_tasks_to_unpack--;
#ifdef __DYNSCHEDULER_DEBUG_PRINT
        fprintf(__debug_print_fp, "remote updated q node %d loop %d task %d %d num tasks left to unpack %ld\n", my_rank, __loop, task_id1, task_id2, _num_tasks_to_unpack);
        IF_DEBUG_FLUSH(fflush(__debug_print_fp));
#endif
       }
       IF_TIME(t_unpack += rtclock() - t_unpack_start);
       IF_TIME(t_wait_unpack_start = rtclock());
      }
      break;
     default:
      fprintf(stderr, "ERROR: q data cannot be sent by a task of loop %d\n", __loop);
      assert(0);
    }
   }
  }
  }
  for (__p=0; __p<recv_buf_s.size(); __p++) {
   if (recv_reqs_s[__p] != MPI_REQUEST_NULL) break;
  }
  if (__p == recv_buf_s.size()) {
   for (__p=0; __p<__MAX_NUM_RECVS; __p++) {
    double *buf = NULL;
    size_t buf_size = 0;
    buf = (double *) polyrt_max_alloc(buf, sizeof(double)*max_num_elements_s, &buf_size);
    recv_buf_s.push_back(buf);
    recv_reqs_s.push_back(MPI_REQUEST_NULL);
    recv_completed_s.push_back(0);
#ifdef __DYNSCHEDULER_MORE_DEBUG_PRINT
    prev_firing_count_s_0.push_back(-1);
    prev_firing_count_s_1.push_back(-1);
    prev_firing_count_s_2.push_back(-1);
    prev_firing_count_s_3.push_back(-1);
#endif
    MPI_Irecv(buf, max_num_elements_s, MPI_DOUBLE, MPI_ANY_SOURCE, 3123, MPI_COMM_WORLD, &recv_reqs_s[recv_reqs_s.size()-1]);
   }
#ifdef __DYNSCHEDULER_DEBUG_PRINT
   fprintf(__debug_print_fp, "node %d recv thread %d s max recv buffers %lu\n", my_rank, __tid, recv_buf_s.size());
   IF_DEBUG_FLUSH(fflush(__debug_print_fp));
#endif
  }
  for (__p=0; __p<recv_buf_q.size(); __p++) {
   if (recv_reqs_q[__p] != MPI_REQUEST_NULL) break;
  }
  if (__p == recv_buf_q.size()) {
   for (__p=0; __p<__MAX_NUM_RECVS; __p++) {
    double *buf = NULL;
    size_t buf_size = 0;
    buf = (double *) polyrt_max_alloc(buf, sizeof(double)*max_num_elements_q, &buf_size);
    recv_buf_q.push_back(buf);
    recv_reqs_q.push_back(MPI_REQUEST_NULL);
    recv_completed_q.push_back(0);
#ifdef __DYNSCHEDULER_MORE_DEBUG_PRINT
    prev_firing_count_q_0.push_back(-1);
    prev_firing_count_q_1.push_back(-1);
    prev_firing_count_q_2.push_back(-1);
    prev_firing_count_q_3.push_back(-1);
#endif
    MPI_Irecv(buf, max_num_elements_q, MPI_DOUBLE, MPI_ANY_SOURCE, 5123, MPI_COMM_WORLD, &recv_reqs_q[recv_reqs_q.size()-1]);
   }
#ifdef __DYNSCHEDULER_DEBUG_PRINT
   fprintf(__debug_print_fp, "node %d recv thread %d q max recv buffers %lu\n", my_rank, __tid, recv_buf_q.size());
   IF_DEBUG_FLUSH(fflush(__debug_print_fp));
#endif
  }
#ifdef __DYNSCHEDULER_DEBUG_PRINT
  if (__busy_wait_count > 1000000) {
   fprintf(stderr, "\nNO PROGRESS in Node %d\n", my_rank);
   fflush(stderr);
   exit(1);
  }
#endif
#ifndef __DYNSCHEDULER_DEDICATED_RECEIVER
{ /* compute-pack-send */
 while (_num_tasks_to_execute > 0) {
  __Task task;
  __flag = pqueue.try_pop(task);
  if(__flag==0) { IF_TIME(waiting_pops[__tid]++); if (pqueue.size() == 0) {
#ifdef __DYNSCHEDULER_DEBUG_PRINT
   int __i;
   for (__i=0;__i<send_reqs_s[__tid].size();__i++) {
    if (send_reqs_s[__tid][__i] != MPI_REQUEST_NULL) {
    int flag, __error = MPI_Test(&send_reqs_s[__tid][__i], &flag, MPI_STATUS_IGNORE);
     if (__error) {
      char *msg=(char*)malloc((MPI_MAX_ERROR_STRING+1)*sizeof(char)); int len;
      MPI_Error_string(__error, msg, &len);
      fprintf(__debug_print_fp, "node %d s send test buf %d error %d %s\n", my_rank, __i, __error, msg);
      free(msg);
     }
     if (flag) {
      fprintf(__debug_print_fp, "node %d s send completed buf %d\n", my_rank, __i);
     }
    }
   }
   for (__i=0;__i<send_reqs_q[__tid].size();__i++) {
    if (send_reqs_q[__tid][__i] != MPI_REQUEST_NULL) {
    int flag, __error = MPI_Test(&send_reqs_q[__tid][__i], &flag, MPI_STATUS_IGNORE);
     if (__error) {
      char *msg=(char*)malloc((MPI_MAX_ERROR_STRING+1)*sizeof(char)); int len;
      MPI_Error_string(__error, msg, &len);
      fprintf(__debug_print_fp, "node %d q send test buf %d error %d %s\n", my_rank, __i, __error, msg);
      free(msg);
     }
     if (flag) {
      fprintf(__debug_print_fp, "node %d q send completed buf %d\n", my_rank, __i);
     }
    }
   }
   IF_DEBUG_FLUSH(fflush(__debug_print_fp));
#endif
   break;
  } else {
   continue;
  } }
  __loop = task.loop;
  task_id1 = task.id1;
  task_id2 = task.id2;
#ifdef __DYNSCHEDULER_DEBUG_PRINT
  long int captured;
#pragma omp atomic capture
  captured = --_num_tasks_to_execute;
  fprintf(__debug_print_fp, "computing node %d thread %d loop %d task %d %d ready tasks %lu\n", my_rank, __tid, __loop, task_id1, task_id2, pqueue.size());
  IF_DEBUG_FLUSH(fflush(__debug_print_fp));
#else
#pragma omp atomic
  _num_tasks_to_execute--;
#endif
  switch(__loop){
   case 0:
    IF_TIME(t_wait_comp[__tid] += rtclock() - t_wait_comp_start[__tid]);
    IF_TIME(t_comp_start[__tid] = rtclock());
    compute_task_0(task_id1,task_id2,s);
    IF_TIME(t_comp[__tid] += rtclock() - t_comp_start[__tid]);
    IF_TIME(t_pack_start[__tid] = rtclock());
    pack_send_0(my_rank,nprocs,task_id1,task_id2,s,send_buf_s[__tid],send_reqs_s[__tid],&__total_count[__tid]);
    IF_TIME(t_pack[__tid] += rtclock() - t_pack_start[__tid]);
    IF_TIME(t_tasks_manage_start[__tid] = rtclock());
    local_update_dep_tasks_0(task_id1,task_id2,_PB_NY,_PB_NX,my_rank,nprocs,_num_threads,tasks_loop0,tasks_loop1,tasks_loop2,tasks_loop3);
    IF_TIME(t_tasks_manage[__tid] += rtclock() - t_tasks_manage_start[__tid]);
    IF_TIME(t_wait_comp_start[__tid] = rtclock());
    break;
   case 1:
    IF_TIME(t_wait_comp[__tid] += rtclock() - t_wait_comp_start[__tid]);
    IF_TIME(t_comp_start[__tid] = rtclock());
    compute_task_1(task_id1,task_id2,q);
    IF_TIME(t_comp[__tid] += rtclock() - t_comp_start[__tid]);
    IF_TIME(t_pack_start[__tid] = rtclock());
    pack_send_1(my_rank,nprocs,task_id1,task_id2,q,send_buf_q[__tid],send_reqs_q[__tid],&__total_count[__tid]);
    IF_TIME(t_pack[__tid] += rtclock() - t_pack_start[__tid]);
    IF_TIME(t_tasks_manage_start[__tid] = rtclock());
    local_update_dep_tasks_1(task_id1,task_id2,_PB_NY,_PB_NX,my_rank,nprocs,_num_threads,tasks_loop0,tasks_loop1,tasks_loop2,tasks_loop3);
    IF_TIME(t_tasks_manage[__tid] += rtclock() - t_tasks_manage_start[__tid]);
    IF_TIME(t_wait_comp_start[__tid] = rtclock());
    break;
   case 2:
    IF_TIME(t_wait_comp[__tid] += rtclock() - t_wait_comp_start[__tid]);
    IF_TIME(t_comp_start[__tid] = rtclock());
    compute_task_2(task_id1,task_id2,s,r,A);
    IF_TIME(t_comp[__tid] += rtclock() - t_comp_start[__tid]);
    IF_TIME(t_pack_start[__tid] = rtclock());
    pack_send_2(my_rank,nprocs,task_id1,task_id2,s,send_buf_s[__tid],send_reqs_s[__tid],&__total_count[__tid]);
    IF_TIME(t_pack[__tid] += rtclock() - t_pack_start[__tid]);
    IF_TIME(t_tasks_manage_start[__tid] = rtclock());
    local_update_dep_tasks_2(task_id1,task_id2,_PB_NY,_PB_NX,my_rank,nprocs,_num_threads,tasks_loop0,tasks_loop1,tasks_loop2,tasks_loop3);
    IF_TIME(t_tasks_manage[__tid] += rtclock() - t_tasks_manage_start[__tid]);
    IF_TIME(t_wait_comp_start[__tid] = rtclock());
    break;
   case 3:
    IF_TIME(t_wait_comp[__tid] += rtclock() - t_wait_comp_start[__tid]);
    IF_TIME(t_comp_start[__tid] = rtclock());
    compute_task_3(task_id1,task_id2,q,A,p);
    IF_TIME(t_comp[__tid] += rtclock() - t_comp_start[__tid]);
    IF_TIME(t_pack_start[__tid] = rtclock());
    pack_send_3(my_rank,nprocs,task_id1,task_id2,q,send_buf_q[__tid],send_reqs_q[__tid],&__total_count[__tid]);
    IF_TIME(t_pack[__tid] += rtclock() - t_pack_start[__tid]);
    IF_TIME(t_tasks_manage_start[__tid] = rtclock());
    local_update_dep_tasks_3(task_id1,task_id2,_PB_NY,_PB_NX,my_rank,nprocs,_num_threads,tasks_loop0,tasks_loop1,tasks_loop2,tasks_loop3);
    IF_TIME(t_tasks_manage[__tid] += rtclock() - t_tasks_manage_start[__tid]);
    IF_TIME(t_wait_comp_start[__tid] = rtclock());
    break;
   default:
    assert(0);
  }
#ifdef __DYNSCHEDULER_DEBUG_PRINT
  fprintf(__debug_print_fp, "computed node %d thread %d loop %d task %d %d tasks left %ld\n", my_rank, __tid, __loop, task_id1, task_id2, captured);
  IF_DEBUG_FLUSH(fflush(__debug_print_fp));
#pragma omp atomic
  __busy_wait_count *= 0;
#endif
 }
}
#endif
 }
}
{ /* compute-pack-send */
 while (_num_tasks_to_execute > 0) {
  __Task task;
  __flag = pqueue.try_pop(task);
  if(__flag==0) { IF_TIME(waiting_pops[__tid]++); if (pqueue.size() == 0) {
#ifdef __DYNSCHEDULER_DEBUG_PRINT
   int __i;
   for (__i=0;__i<send_reqs_s[__tid].size();__i++) {
    if (send_reqs_s[__tid][__i] != MPI_REQUEST_NULL) {
    int flag, __error = MPI_Test(&send_reqs_s[__tid][__i], &flag, MPI_STATUS_IGNORE);
     if (__error) {
      char *msg=(char*)malloc((MPI_MAX_ERROR_STRING+1)*sizeof(char)); int len;
      MPI_Error_string(__error, msg, &len);
      fprintf(__debug_print_fp, "node %d s send test buf %d error %d %s\n", my_rank, __i, __error, msg);
      free(msg);
     }
     if (flag) {
      fprintf(__debug_print_fp, "node %d s send completed buf %d\n", my_rank, __i);
     }
    }
   }
   for (__i=0;__i<send_reqs_q[__tid].size();__i++) {
    if (send_reqs_q[__tid][__i] != MPI_REQUEST_NULL) {
    int flag, __error = MPI_Test(&send_reqs_q[__tid][__i], &flag, MPI_STATUS_IGNORE);
     if (__error) {
      char *msg=(char*)malloc((MPI_MAX_ERROR_STRING+1)*sizeof(char)); int len;
      MPI_Error_string(__error, msg, &len);
      fprintf(__debug_print_fp, "node %d q send test buf %d error %d %s\n", my_rank, __i, __error, msg);
      free(msg);
     }
     if (flag) {
      fprintf(__debug_print_fp, "node %d q send completed buf %d\n", my_rank, __i);
     }
    }
   }
   IF_DEBUG_FLUSH(fflush(__debug_print_fp));
#endif
  }
  continue;
  }
  __loop = task.loop;
  task_id1 = task.id1;
  task_id2 = task.id2;
#ifdef __DYNSCHEDULER_DEBUG_PRINT
  long int captured;
#pragma omp atomic capture
  captured = --_num_tasks_to_execute;
  fprintf(__debug_print_fp, "computing node %d thread %d loop %d task %d %d ready tasks %lu\n", my_rank, __tid, __loop, task_id1, task_id2, pqueue.size());
  IF_DEBUG_FLUSH(fflush(__debug_print_fp));
#else
#pragma omp atomic
  _num_tasks_to_execute--;
#endif
  switch(__loop){
   case 0:
    IF_TIME(t_wait_comp[__tid] += rtclock() - t_wait_comp_start[__tid]);
    IF_TIME(t_comp_start[__tid] = rtclock());
    compute_task_0(task_id1,task_id2,s);
    IF_TIME(t_comp[__tid] += rtclock() - t_comp_start[__tid]);
    IF_TIME(t_pack_start[__tid] = rtclock());
    pack_send_0(my_rank,nprocs,task_id1,task_id2,s,send_buf_s[__tid],send_reqs_s[__tid],&__total_count[__tid]);
    IF_TIME(t_pack[__tid] += rtclock() - t_pack_start[__tid]);
    IF_TIME(t_tasks_manage_start[__tid] = rtclock());
    local_update_dep_tasks_0(task_id1,task_id2,_PB_NY,_PB_NX,my_rank,nprocs,_num_threads,tasks_loop0,tasks_loop1,tasks_loop2,tasks_loop3);
    IF_TIME(t_tasks_manage[__tid] += rtclock() - t_tasks_manage_start[__tid]);
    IF_TIME(t_wait_comp_start[__tid] = rtclock());
    break;
   case 1:
    IF_TIME(t_wait_comp[__tid] += rtclock() - t_wait_comp_start[__tid]);
    IF_TIME(t_comp_start[__tid] = rtclock());
    compute_task_1(task_id1,task_id2,q);
    IF_TIME(t_comp[__tid] += rtclock() - t_comp_start[__tid]);
    IF_TIME(t_pack_start[__tid] = rtclock());
    pack_send_1(my_rank,nprocs,task_id1,task_id2,q,send_buf_q[__tid],send_reqs_q[__tid],&__total_count[__tid]);
    IF_TIME(t_pack[__tid] += rtclock() - t_pack_start[__tid]);
    IF_TIME(t_tasks_manage_start[__tid] = rtclock());
    local_update_dep_tasks_1(task_id1,task_id2,_PB_NY,_PB_NX,my_rank,nprocs,_num_threads,tasks_loop0,tasks_loop1,tasks_loop2,tasks_loop3);
    IF_TIME(t_tasks_manage[__tid] += rtclock() - t_tasks_manage_start[__tid]);
    IF_TIME(t_wait_comp_start[__tid] = rtclock());
    break;
   case 2:
    IF_TIME(t_wait_comp[__tid] += rtclock() - t_wait_comp_start[__tid]);
    IF_TIME(t_comp_start[__tid] = rtclock());
    compute_task_2(task_id1,task_id2,s,r,A);
    IF_TIME(t_comp[__tid] += rtclock() - t_comp_start[__tid]);
    IF_TIME(t_pack_start[__tid] = rtclock());
    pack_send_2(my_rank,nprocs,task_id1,task_id2,s,send_buf_s[__tid],send_reqs_s[__tid],&__total_count[__tid]);
    IF_TIME(t_pack[__tid] += rtclock() - t_pack_start[__tid]);
    IF_TIME(t_tasks_manage_start[__tid] = rtclock());
    local_update_dep_tasks_2(task_id1,task_id2,_PB_NY,_PB_NX,my_rank,nprocs,_num_threads,tasks_loop0,tasks_loop1,tasks_loop2,tasks_loop3);
    IF_TIME(t_tasks_manage[__tid] += rtclock() - t_tasks_manage_start[__tid]);
    IF_TIME(t_wait_comp_start[__tid] = rtclock());
    break;
   case 3:
    IF_TIME(t_wait_comp[__tid] += rtclock() - t_wait_comp_start[__tid]);
    IF_TIME(t_comp_start[__tid] = rtclock());
    compute_task_3(task_id1,task_id2,q,A,p);
    IF_TIME(t_comp[__tid] += rtclock() - t_comp_start[__tid]);
    IF_TIME(t_pack_start[__tid] = rtclock());
    pack_send_3(my_rank,nprocs,task_id1,task_id2,q,send_buf_q[__tid],send_reqs_q[__tid],&__total_count[__tid]);
    IF_TIME(t_pack[__tid] += rtclock() - t_pack_start[__tid]);
    IF_TIME(t_tasks_manage_start[__tid] = rtclock());
    local_update_dep_tasks_3(task_id1,task_id2,_PB_NY,_PB_NX,my_rank,nprocs,_num_threads,tasks_loop0,tasks_loop1,tasks_loop2,tasks_loop3);
    IF_TIME(t_tasks_manage[__tid] += rtclock() - t_tasks_manage_start[__tid]);
    IF_TIME(t_wait_comp_start[__tid] = rtclock());
    break;
   default:
    assert(0);
  }
#ifdef __DYNSCHEDULER_DEBUG_PRINT
  fprintf(__debug_print_fp, "computed node %d thread %d loop %d task %d %d tasks left %ld\n", my_rank, __tid, __loop, task_id1, task_id2, captured);
  IF_DEBUG_FLUSH(fflush(__debug_print_fp));
#pragma omp atomic
  __busy_wait_count *= 0;
#endif
 }
}
#ifdef __DYNSCHEDULER_DEBUG_PRINT
 fprintf(__debug_print_fp, "node %d thread %d done\n", my_rank, __tid);
 IF_DEBUG_FLUSH(fflush(__debug_print_fp));
#endif
}
 IF_TIME(t_unpack_start = rtclock());
 for (__p=0; __p<recv_buf_s.size(); __p++) {
  if (recv_reqs_s[__p] != MPI_REQUEST_NULL) MPI_Request_free(&recv_reqs_s[__p]);
 }
 for (__p=0; __p<recv_buf_q.size(); __p++) {
  if (recv_reqs_q[__p] != MPI_REQUEST_NULL) MPI_Request_free(&recv_reqs_q[__p]);
 }
 IF_TIME(t_unpack += rtclock() - t_unpack_start);
#ifdef __DYNSCHEDULER_DEBUG_PRINT
 fclose(__debug_print_fp);
#endif
/*End of dynamic scheduling code*/
 IF_TIME(t_writeout = rtclock());
 
#ifndef USE_LOCAL_ARRAYS
 if (fopen(".test", "r")) {
 write_out(my_rank,nprocs,lw_buf_s,lw_recv_buf_s,displs_lw_s,lw_buf_q,lw_recv_buf_q,displs_lw_q,s,q);
} 
#endif
 IF_TIME(t_writeout = rtclock() - t_writeout);
#ifdef TIME
 IF_TIME(t_local = rtclock() - t_local - t_writeout);
 for (__tid=0; __tid<_num_threads; __tid++) __total_count_threads += __total_count[__tid];
 waiting_pops_threads = waiting_pops[0];
                    for (__tid=1; __tid<_num_threads; __tid++) if (waiting_pops_threads < waiting_pops[__tid]) waiting_pops_threads = waiting_pops[__tid];
 t_tasks_manage_threads = t_tasks_manage[0];
                    for (__tid=1; __tid<_num_threads; __tid++) if (t_tasks_manage_threads < t_tasks_manage[__tid]) t_tasks_manage_threads = t_tasks_manage[__tid];
 if (__tid_receiver != 0) t_comp_threads_min = t_comp[0];
                    else t_comp_threads_min = t_comp[1];
                    for (__tid=1; __tid<_num_threads; __tid++) if ((__tid != __tid_receiver) && (t_comp_threads_min > t_comp[__tid])) t_comp_threads_min = t_comp[__tid];
 t_comp_threads = t_comp[0];
                    for (__tid=1; __tid<_num_threads; __tid++) if (t_comp_threads < t_comp[__tid]) t_comp_threads = t_comp[__tid];
 t_comp_mean = 0.0;
                    for (__tid=0; __tid<_num_threads; __tid++) if (__tid != __tid_receiver) t_comp_mean += t_comp[__tid];
                    t_globalcomp_mean = t_comp_mean;
                    t_comp_mean = t_comp_mean/(_num_threads-1);
 t_comp_sd = 0.0;
                    for (__tid=0; __tid<_num_threads; __tid++) if (__tid != __tid_receiver) t_comp_sd += (t_comp[__tid]-t_comp_mean)*(t_comp[__tid]-t_comp_mean);
                    t_comp_sd = sqrt(t_comp_sd/(_num_threads-1));
 t_wait_comp_threads = t_wait_comp[0];
                    for (__tid=1; __tid<_num_threads; __tid++) if (t_wait_comp_threads < t_wait_comp[__tid]) t_wait_comp_threads = t_wait_comp[__tid];
 t_pack_threads = t_pack[0];
                    for (__tid=1; __tid<_num_threads; __tid++) if (t_pack_threads < t_pack[__tid]) t_pack_threads = t_pack[__tid];
 char *buffer = (char *)malloc(4096 * sizeof(char));
 strcpy(buffer, "");
#ifdef __DETAILED_TIME_REPORT
 sprintf(buffer+strlen(buffer), "Node %d\n", my_rank);
 sprintf(buffer+strlen(buffer), "Write-out time: %0.61f s\n", t_writeout);
 sprintf(buffer+strlen(buffer), "Values communicated: %0.0lf\n", __total_count_threads);
 sprintf(buffer+strlen(buffer), "Minimum computation time across all threads: %0.6lf s\n", t_comp_threads_min);
 sprintf(buffer+strlen(buffer), "Maximum computation time across all threads: %0.6lf s\n", t_comp_threads);
 sprintf(buffer+strlen(buffer), "Mean computation time across all threads: %0.6lf s\n", t_comp_mean);
 sprintf(buffer+strlen(buffer), "Standard deviation in computation time across all threads: %0.6lf s\n", t_comp_sd);
 sprintf(buffer+strlen(buffer), "Load imbalance factor (standard-deviation/mean): %0.2lf %%\n", 100*t_comp_sd/t_comp_mean);
 sprintf(buffer+strlen(buffer), "'Waiting for computation' time: %0.6lf s\n", t_wait_comp_threads);
 sprintf(buffer+strlen(buffer), "Tasks creation time: %0.6lf s\n", t_tasks_create);
 sprintf(buffer+strlen(buffer), "Tasks management time: %0.6lf s\n", t_tasks_manage_threads);
 sprintf(buffer+strlen(buffer), "Packing time: %0.6lf s\n", t_pack_threads);
 sprintf(buffer+strlen(buffer), "Unpacking time: %0.6lf s\n", t_unpack);
 sprintf(buffer+strlen(buffer), "'Waiting for unpacking' time: %0.6lf s\n", t_wait_unpack);
 sprintf(buffer+strlen(buffer), "Sum of split times: %0.6lf s\n", t_comp_threads + t_wait_comp_threads + t_tasks_create + t_tasks_manage_threads + t_pack_threads);
 sprintf(buffer+strlen(buffer), "Total time minus write-out time: %0.6lf s\n", t_local);
 sprintf(buffer+strlen(buffer), "-------");
 fprintf(stdout, "%s\n", buffer);
#endif
 MPI_Reduce(&__total_count_threads, &__total_count_all, 1, MPI_DOUBLE, MPI_SUM, 0, MPI_COMM_WORLD);
 MPI_Allreduce(MPI_IN_PLACE, &t_globalcomp_mean, 1, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
 t_globalcomp_mean = t_globalcomp_mean/((_num_threads-1)*nprocs);
 t_globalcomp_sd = 0.0;
                    for (__tid=0; __tid<_num_threads; __tid++) if (__tid != __tid_receiver) t_globalcomp_sd += (t_comp[__tid]-t_globalcomp_mean)*(t_comp[__tid]-t_globalcomp_mean);
                    MPI_Allreduce(MPI_IN_PLACE, &t_globalcomp_sd, 1, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
                    t_globalcomp_sd = sqrt(t_globalcomp_sd/((_num_threads-1)*nprocs));
 MPI_Reduce(&waiting_pops_threads, &waiting_pops_global, 1, MPI_INT, MPI_MAX, 0, MPI_COMM_WORLD);
 MPI_Reduce(&t_comp_threads_min, &t_globalcomp_min, 1, MPI_DOUBLE, MPI_MIN, 0, MPI_COMM_WORLD);
 MPI_Reduce(&t_comp_threads, &t_globalcomp, 1, MPI_DOUBLE, MPI_MAX, 0, MPI_COMM_WORLD);
 MPI_Reduce(&t_wait_comp_threads, &t_globalwait_comp, 1, MPI_DOUBLE, MPI_MAX, 0, MPI_COMM_WORLD);
 MPI_Reduce(&t_tasks_create, &t_globaltasks_create, 1, MPI_DOUBLE, MPI_MAX, 0, MPI_COMM_WORLD);
 MPI_Reduce(&t_tasks_manage_threads, &t_globaltasks_manage, 1, MPI_DOUBLE, MPI_MAX, 0, MPI_COMM_WORLD);
 MPI_Reduce(&t_pack_threads, &t_globalpack, 1, MPI_DOUBLE, MPI_MAX, 0, MPI_COMM_WORLD);
 MPI_Reduce(&t_unpack, &t_globalunpack, 1, MPI_DOUBLE, MPI_MAX, 0, MPI_COMM_WORLD);
 MPI_Reduce(&t_wait_unpack, &t_globalwait_unpack, 1, MPI_DOUBLE, MPI_MAX, 0, MPI_COMM_WORLD);
 MPI_Reduce(&t_local, &t_global, 1, MPI_DOUBLE, MPI_MAX, 0, MPI_COMM_WORLD);
 if (my_rank==0) {
  strcpy(buffer, "SUMMARY\n");
  sprintf(buffer+strlen(buffer), "Write-out time spent in master node: %0.6lf s\n", t_writeout);
  sprintf(buffer+strlen(buffer), "Total communication volume across all nodes: %0.6lf GB\n", __total_count_all*8/((double)1024*1024*1024));
  sprintf(buffer+strlen(buffer), "Maximum computation time spent across all nodes: %0.6lf s\n", t_globalcomp);
  sprintf(buffer+strlen(buffer), "Minimum computation time spent across all nodes: %0.6lf s\n", t_globalcomp_min);
  sprintf(buffer+strlen(buffer), "Mean computation time across all nodes: %0.6lf s\n", t_globalcomp_mean);
  sprintf(buffer+strlen(buffer), "Standard deviation in computation time across all nodes: %0.6lf s\n", t_globalcomp_sd);
  sprintf(buffer+strlen(buffer), "Load imbalance factor (standard-deviation/mean): %0.2lf %%\n", 100*t_globalcomp_sd/t_globalcomp_mean);
  sprintf(buffer+strlen(buffer), "Maximum 'waiting for computation' time spent across all nodes: %0.6lf s\n", t_globalwait_comp);
  sprintf(buffer+strlen(buffer), "Maximum number of waiting/unsuccessful POPs across all nodes: %d\n", waiting_pops_global);
  sprintf(buffer+strlen(buffer), "Maximum tasks creation time spent across all nodes: %0.6lf s\n", t_globaltasks_create);
  sprintf(buffer+strlen(buffer), "Maximum tasks management time spent across all nodes: %0.6lf s\n", t_globaltasks_manage);
  sprintf(buffer+strlen(buffer), "Maximum packing time spent across all nodes: %0.6lf s\n", t_globalpack);
  sprintf(buffer+strlen(buffer), "Maximum unpacking time spent across all nodes: %0.6lf s\n", t_globalunpack);
  sprintf(buffer+strlen(buffer), "Maximum 'waiting for unpacking' time spent across all nodes: %0.6lf s\n", t_globalwait_unpack);
  sprintf(buffer+strlen(buffer), "Maximum total time spent across all nodes: %0.6lf s\n", t_global);
  sprintf(buffer+strlen(buffer), "-------");
  fprintf(stdout, "%s\n", buffer);
 }
 free(buffer);
#endif
 MPI_Barrier(MPI_COMM_WORLD);
 MPI_Finalize();

}


int main(int argc, char** argv)
{
  /* Retrieve problem size. */
  int nx = NX;
  int ny = NY;

  /* Variable declaration/allocation. */
  POLYBENCH_2D_ARRAY_DECL(A, DATA_TYPE, NX, NY, nx, ny);
  POLYBENCH_1D_ARRAY_DECL(s, DATA_TYPE, NY, ny);
  POLYBENCH_1D_ARRAY_DECL(q, DATA_TYPE, NX, nx);
  POLYBENCH_1D_ARRAY_DECL(p, DATA_TYPE, NY, ny);
  POLYBENCH_1D_ARRAY_DECL(r, DATA_TYPE, NX, nx);

  /* Initialize array(s). */
  init_array (nx, ny,
	      POLYBENCH_ARRAY(A),
	      POLYBENCH_ARRAY(r),
	      POLYBENCH_ARRAY(p));

  /* Start timer. */
  polybench_start_instruments;

  /* Run kernel. */
  kernel_bicg (nx, ny,
	       POLYBENCH_ARRAY(A),
	       POLYBENCH_ARRAY(s),
	       POLYBENCH_ARRAY(q),
	       POLYBENCH_ARRAY(p),
	       POLYBENCH_ARRAY(r));

  /* Stop and print timer. */
  polybench_stop_instruments;
  polybench_print_instruments;

  /* Prevent dead-code elimination. All live-out data must be printed
     by the function call in argument. */
  polybench_prevent_dce(print_array(nx, ny, POLYBENCH_ARRAY(s), POLYBENCH_ARRAY(q)));

  /* Be clean. */
  POLYBENCH_FREE_ARRAY(A);
  POLYBENCH_FREE_ARRAY(s);
  POLYBENCH_FREE_ARRAY(q);
  POLYBENCH_FREE_ARRAY(p);
  POLYBENCH_FREE_ARRAY(r);

  return 0;
}
void init_tasks(int my_rank,int nprocs,int num_threads,int *tasks_loop0,int *tasks_loop1,int *tasks_loop2,int *tasks_loop3,long int *p_num_tasks_to_execute,long int *p_num_tasks_to_unpack)
{
int remote_dep_tasks, local_dep_tasks, affinity;
long int _num_tasks_to_execute = 0;
long int _num_tasks_to_unpack = 0;





  int t1, t2, t3;

 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;

/* Start of CLooG code */
if (_PB_NY >= 1) {
  lbp=0;
  ubp=floord(_PB_NY-1,32);
#pragma omp parallel for private(remote_dep_tasks, local_dep_tasks, affinity) reduction(+:_num_tasks_to_execute,_num_tasks_to_unpack)
  for (t3=lbp;t3<=ubp;t3++) {
    if (pi_0(2,t3,_PB_NY,_PB_NX,nprocs) == my_rank) { _num_tasks_to_execute++;tasks_loop0[(t3-lb_tasks_loop0_dim0)+0] = get_num_remote_src_tasks_0(2,t3,_PB_NY,_PB_NX,my_rank,nprocs) + get_num_local_src_tasks_0(2,t3,_PB_NY,_PB_NX,my_rank,nprocs);if (tasks_loop0[(t3-lb_tasks_loop0_dim0)+0] == 0) { remote_dep_tasks = get_num_remote_dep_tasks_0(2,t3,_PB_NY,_PB_NX,my_rank,nprocs); local_dep_tasks = get_num_local_dep_tasks_0(2,t3,_PB_NY,_PB_NX,my_rank,nprocs); if (__is_multi_partitioned[0] || __is_block_cyclic[0]) affinity = -1; else affinity = pi_threads_0(2,t3,_PB_NY,_PB_NX,my_rank,nprocs,num_threads); __Task task(0,2,t3,remote_dep_tasks,local_dep_tasks,affinity); pqueue.push(task); IF_DYNSCHEDULER_MORE_DEBUG_PRINT( fprintf(__debug_print_fp, "added node %d loop 0 task %d %d                 remote_dep_tasks %d local_dep_tasks %d ready tasks %lu\n", my_rank, 2,t3, remote_dep_tasks, local_dep_tasks, pqueue.size())); IF_DEBUG_FLUSH(fflush(__debug_print_fp)); }} else if (is_receiver_0(2,t3,_PB_NY,_PB_NX,my_rank,nprocs)) { _num_tasks_to_unpack++;tasks_loop0[(t3-lb_tasks_loop0_dim0)+0] = 1 + get_num_local_src_tasks_0(2,t3,_PB_NY,_PB_NX,my_rank,nprocs);IF_DYNSCHEDULER_MORE_DEBUG_PRINT( fprintf(__debug_print_fp, "node %d loop 0 task %d %d                 initial count %d\n", my_rank, 2,t3, tasks_loop0[(t3-lb_tasks_loop0_dim0)+0])); IF_DEBUG_FLUSH(fflush(__debug_print_fp));};
  }
}
if (_PB_NX >= 1) {
  lbp=0;
  ubp=floord(_PB_NX-1,32);
#pragma omp parallel for private(remote_dep_tasks, local_dep_tasks, affinity) reduction(+:_num_tasks_to_execute,_num_tasks_to_unpack)
  for (t3=lbp;t3<=ubp;t3++) {
    if (pi_1(0,t3,_PB_NY,_PB_NX,nprocs) == my_rank) { _num_tasks_to_execute++;tasks_loop1[(t3-lb_tasks_loop1_dim0)+0] = get_num_remote_src_tasks_1(0,t3,_PB_NY,_PB_NX,my_rank,nprocs) + get_num_local_src_tasks_1(0,t3,_PB_NY,_PB_NX,my_rank,nprocs);if (tasks_loop1[(t3-lb_tasks_loop1_dim0)+0] == 0) { remote_dep_tasks = get_num_remote_dep_tasks_1(0,t3,_PB_NY,_PB_NX,my_rank,nprocs); local_dep_tasks = get_num_local_dep_tasks_1(0,t3,_PB_NY,_PB_NX,my_rank,nprocs); if (__is_multi_partitioned[1] || __is_block_cyclic[1]) affinity = -1; else affinity = pi_threads_1(0,t3,_PB_NY,_PB_NX,my_rank,nprocs,num_threads); __Task task(1,0,t3,remote_dep_tasks,local_dep_tasks,affinity); pqueue.push(task); IF_DYNSCHEDULER_MORE_DEBUG_PRINT( fprintf(__debug_print_fp, "added node %d loop 1 task %d %d                 remote_dep_tasks %d local_dep_tasks %d ready tasks %lu\n", my_rank, 0,t3, remote_dep_tasks, local_dep_tasks, pqueue.size())); IF_DEBUG_FLUSH(fflush(__debug_print_fp)); }} else if (is_receiver_1(0,t3,_PB_NY,_PB_NX,my_rank,nprocs)) { _num_tasks_to_unpack++;tasks_loop1[(t3-lb_tasks_loop1_dim0)+0] = 1 + get_num_local_src_tasks_1(0,t3,_PB_NY,_PB_NX,my_rank,nprocs);IF_DYNSCHEDULER_MORE_DEBUG_PRINT( fprintf(__debug_print_fp, "node %d loop 1 task %d %d                 initial count %d\n", my_rank, 0,t3, tasks_loop1[(t3-lb_tasks_loop1_dim0)+0])); IF_DEBUG_FLUSH(fflush(__debug_print_fp));};
  }
}
if ((_PB_NX >= 1) && (_PB_NY >= 1)) {
  lbp=0;
  ubp=floord(_PB_NY-1,32);
#pragma omp parallel for private(remote_dep_tasks, local_dep_tasks, affinity) reduction(+:_num_tasks_to_execute,_num_tasks_to_unpack)
  for (t3=lbp;t3<=ubp;t3++) {
    if (pi_2(3,t3,_PB_NY,_PB_NX,nprocs) == my_rank) { _num_tasks_to_execute++;tasks_loop2[(t3-lb_tasks_loop2_dim0)+0] = get_num_remote_src_tasks_2(3,t3,_PB_NY,_PB_NX,my_rank,nprocs) + get_num_local_src_tasks_2(3,t3,_PB_NY,_PB_NX,my_rank,nprocs);if (tasks_loop2[(t3-lb_tasks_loop2_dim0)+0] == 0) { remote_dep_tasks = get_num_remote_dep_tasks_2(3,t3,_PB_NY,_PB_NX,my_rank,nprocs); local_dep_tasks = get_num_local_dep_tasks_2(3,t3,_PB_NY,_PB_NX,my_rank,nprocs); if (__is_multi_partitioned[2] || __is_block_cyclic[2]) affinity = -1; else affinity = pi_threads_2(3,t3,_PB_NY,_PB_NX,my_rank,nprocs,num_threads); __Task task(2,3,t3,remote_dep_tasks,local_dep_tasks,affinity); pqueue.push(task); IF_DYNSCHEDULER_MORE_DEBUG_PRINT( fprintf(__debug_print_fp, "added node %d loop 2 task %d %d                 remote_dep_tasks %d local_dep_tasks %d ready tasks %lu\n", my_rank, 3,t3, remote_dep_tasks, local_dep_tasks, pqueue.size())); IF_DEBUG_FLUSH(fflush(__debug_print_fp)); }} else if (is_receiver_2(3,t3,_PB_NY,_PB_NX,my_rank,nprocs)) { _num_tasks_to_unpack++;tasks_loop2[(t3-lb_tasks_loop2_dim0)+0] = 1 + get_num_local_src_tasks_2(3,t3,_PB_NY,_PB_NX,my_rank,nprocs);IF_DYNSCHEDULER_MORE_DEBUG_PRINT( fprintf(__debug_print_fp, "node %d loop 2 task %d %d                 initial count %d\n", my_rank, 3,t3, tasks_loop2[(t3-lb_tasks_loop2_dim0)+0])); IF_DEBUG_FLUSH(fflush(__debug_print_fp));};
  }
}
if ((_PB_NX >= 1) && (_PB_NY >= 1)) {
  lbp=0;
  ubp=floord(_PB_NX-1,32);
#pragma omp parallel for private(remote_dep_tasks, local_dep_tasks, affinity) reduction(+:_num_tasks_to_execute,_num_tasks_to_unpack)
  for (t3=lbp;t3<=ubp;t3++) {
    if (pi_3(1,t3,_PB_NY,_PB_NX,nprocs) == my_rank) { _num_tasks_to_execute++;tasks_loop3[(t3-lb_tasks_loop3_dim0)+0] = get_num_remote_src_tasks_3(1,t3,_PB_NY,_PB_NX,my_rank,nprocs) + get_num_local_src_tasks_3(1,t3,_PB_NY,_PB_NX,my_rank,nprocs);if (tasks_loop3[(t3-lb_tasks_loop3_dim0)+0] == 0) { remote_dep_tasks = get_num_remote_dep_tasks_3(1,t3,_PB_NY,_PB_NX,my_rank,nprocs); local_dep_tasks = get_num_local_dep_tasks_3(1,t3,_PB_NY,_PB_NX,my_rank,nprocs); if (__is_multi_partitioned[3] || __is_block_cyclic[3]) affinity = -1; else affinity = pi_threads_3(1,t3,_PB_NY,_PB_NX,my_rank,nprocs,num_threads); __Task task(3,1,t3,remote_dep_tasks,local_dep_tasks,affinity); pqueue.push(task); IF_DYNSCHEDULER_MORE_DEBUG_PRINT( fprintf(__debug_print_fp, "added node %d loop 3 task %d %d                 remote_dep_tasks %d local_dep_tasks %d ready tasks %lu\n", my_rank, 1,t3, remote_dep_tasks, local_dep_tasks, pqueue.size())); IF_DEBUG_FLUSH(fflush(__debug_print_fp)); }} else if (is_receiver_3(1,t3,_PB_NY,_PB_NX,my_rank,nprocs)) { _num_tasks_to_unpack++;tasks_loop3[(t3-lb_tasks_loop3_dim0)+0] = 1 + get_num_local_src_tasks_3(1,t3,_PB_NY,_PB_NX,my_rank,nprocs);IF_DYNSCHEDULER_MORE_DEBUG_PRINT( fprintf(__debug_print_fp, "node %d loop 3 task %d %d                 initial count %d\n", my_rank, 1,t3, tasks_loop3[(t3-lb_tasks_loop3_dim0)+0])); IF_DEBUG_FLUSH(fflush(__debug_print_fp));};
  }
}
/* End of CLooG code */




*p_num_tasks_to_execute = _num_tasks_to_execute;
*p_num_tasks_to_unpack = _num_tasks_to_unpack;
}

int pack_s_1_0(int ts1,int ts2,double *send_buf_s,int send_counts_s, __DECLARATION_OF_s){

int recv_proc;


  int t1, t2;

 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;

/* Start of CLooG code */
if ((_PB_NX >= 1) && (ts1 == 2) && (ts2 >= 0) && (32*ts2 <= _PB_NY-1)) {
  for (t1=32*ts2;t1<=min(_PB_NY-1,32*ts2+31);t1++) {
    send_buf_s[send_counts_s++] = s[t1];
  }
}
/* End of CLooG code */

return send_counts_s;


}

int unpack_s_1_0(int ts1,int ts2,double *recv_buf_s,int curr_displs_s, __DECLARATION_OF_s){

int recv_proc;


  int t1, t2;

 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;

/* Start of CLooG code */
if ((_PB_NX >= 1) && (ts1 == 2) && (ts2 >= 0) && (32*ts2 <= _PB_NY-1)) {
  for (t1=32*ts2;t1<=min(_PB_NY-1,32*ts2+31);t1++) {
    s[t1] = recv_buf_s[curr_displs_s++];
  }
}
/* End of CLooG code */

return curr_displs_s;


}

int writeout_pack_s_0(int ts1,int ts2,double *lw_buf_s,int lw_count_s, __DECLARATION_OF_s){

int recv_proc;


  int t1, t2;

 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;

/* Start of CLooG code */
if ((_PB_NX <= 0) && (ts1 == 2) && (ts2 >= 0) && (32*ts2 <= _PB_NY-1)) {
  for (t1=32*ts2;t1<=min(_PB_NY-1,32*ts2+31);t1++) {
    lw_buf_s[lw_count_s++] = s[t1];
  }
}
/* End of CLooG code */

return lw_count_s;


}

int writeout_unpack_s_0(int ts1,int ts2,double *lw_recv_buf_s,int displs_lw_s, __DECLARATION_OF_s){

int recv_proc;


  int t1, t2;

 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;

/* Start of CLooG code */
if ((_PB_NX <= 0) && (ts1 == 2) && (ts2 >= 0) && (32*ts2 <= _PB_NY-1)) {
  for (t1=32*ts2;t1<=min(_PB_NY-1,32*ts2+31);t1++) {
    
#ifndef USE_LOCAL_ARRAYS
 s[t1] = lw_recv_buf_s[displs_lw_s++]; 
#endif
;
  }
}
/* End of CLooG code */

return displs_lw_s;


}

void pack_send_0(int my_rank,int nprocs,int t1,int t2,__DECLARATION_OF_s,std::vector<double *>& send_buf_s,std::vector<MPI_Request>& send_reqs_s,double *p_total_count)
{
int __p, receiver_list[nprocs];
int send_counts_s[nprocs];
for (__p=0;__p<nprocs;__p++) send_counts_s[__p] = 0;
int index_s[nprocs];
int current_index_s = 0;
int flag;
double __total_count = 0;
for (__p=0; __p<nprocs; __p++) { receiver_list[__p] = 0; } sigma_s_1_0(t1,t2,_PB_NY,_PB_NX, my_rank, nprocs, receiver_list); for (__p=0; __p<nprocs; __p++) { if (receiver_list[__p] != 0) { if (send_counts_s[__p] == 0) { for (;current_index_s<send_reqs_s.size();current_index_s++) { if (send_reqs_s[current_index_s] == MPI_REQUEST_NULL) break; int __error = MPI_Test(&send_reqs_s[current_index_s], &flag, MPI_STATUS_IGNORE); IF_DYNSCHEDULER_DEBUG_PRINT( if (flag) { fprintf(__debug_print_fp, "node %d s send completed buf %d\n", my_rank, current_index_s); } if (__error) { char *msg=(char*)malloc((MPI_MAX_ERROR_STRING+1)*sizeof(char)); int len; MPI_Error_string(__error, msg, &len); fprintf(__debug_print_fp, "node %d s send test buf %d error %d %s\n", my_rank, current_index_s, __error, msg); } ) IF_DEBUG_FLUSH(fflush(__debug_print_fp)); if (flag) break; } if (current_index_s==send_reqs_s.size()) { double *buf = NULL; size_t buf_size = 0; buf = (double *) polyrt_max_alloc(buf, sizeof(double)*max_num_elements_s, &buf_size); send_buf_s.push_back(buf); IF_DYNSCHEDULER_DEBUG_PRINT( fprintf(__debug_print_fp, "node %d s max send buffers %d\n", my_rank, current_index_s+1)); IF_DEBUG_FLUSH(fflush(__debug_print_fp)); send_reqs_s.push_back(MPI_REQUEST_NULL); } index_s[__p] = current_index_s++; send_counts_s[__p] = 3; }send_counts_s[__p] = pack_s_1_0(t1,t2, send_buf_s[index_s[__p]], send_counts_s[__p],s); } };
    for (__p=0; __p<nprocs; __p++) { if(send_counts_s[__p] > 0) {send_buf_s[index_s[__p]][0] = 0;send_buf_s[index_s[__p]][1] = t1;send_buf_s[index_s[__p]][2] = t2; IF_DYNSCHEDULER_DEBUG_PRINT( fprintf(__debug_print_fp, "sending node %d buf %d loop 0 task %d %d 		        s size %d to node %d\n", my_rank, index_s[__p], t1,t2, send_counts_s[__p], __p)); IF_DEBUG_FLUSH(fflush(__debug_print_fp));IF_TIME(__total_count += send_counts_s[__p]); int __error = MPI_Isend(send_buf_s[index_s[__p]], send_counts_s[__p], MPI_DOUBLE, __p, 3123, MPI_COMM_WORLD, &send_reqs_s[index_s[__p]]); send_counts_s[__p] = 0; IF_DYNSCHEDULER_DEBUG_PRINT( if (__error) { char *msg=(char*)malloc((MPI_MAX_ERROR_STRING+1)*sizeof(char)); int len; MPI_Error_string(__error, msg, &len); fprintf(__debug_print_fp, "node %d s send buf %d error %d %s\n", my_rank, index_s[__p], __error, msg); } ) }};
*p_total_count += __total_count;
}

void unpack_buf_s_0(int my_rank,int nprocs,__DECLARATION_OF_s,double *recv_buf_s)
{
int curr_displs_s;
curr_displs_s = 3; if (is_receiver_s_1_0((int)recv_buf_s[1],(int)recv_buf_s[2],_PB_NY,_PB_NX,my_rank,nprocs) != 0) { curr_displs_s = unpack_s_1_0((int)recv_buf_s[1],(int)recv_buf_s[2],recv_buf_s,curr_displs_s,s); };
}

int pack_q_1_1(int ts1,int ts2,double *send_buf_q,int send_counts_q, __DECLARATION_OF_q){

int recv_proc;


  int t1, t2;

 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;

/* Start of CLooG code */
if ((_PB_NY >= 1) && (ts1 == 0) && (ts2 >= 0) && (32*ts2 <= _PB_NX-1)) {
  for (t1=32*ts2;t1<=min(_PB_NX-1,32*ts2+31);t1++) {
    send_buf_q[send_counts_q++] = q[t1];
  }
}
/* End of CLooG code */

return send_counts_q;


}

int unpack_q_1_1(int ts1,int ts2,double *recv_buf_q,int curr_displs_q, __DECLARATION_OF_q){

int recv_proc;


  int t1, t2;

 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;

/* Start of CLooG code */
if ((_PB_NY >= 1) && (ts1 == 0) && (ts2 >= 0) && (32*ts2 <= _PB_NX-1)) {
  for (t1=32*ts2;t1<=min(_PB_NX-1,32*ts2+31);t1++) {
    q[t1] = recv_buf_q[curr_displs_q++];
  }
}
/* End of CLooG code */

return curr_displs_q;


}

int writeout_pack_q_1(int ts1,int ts2,double *lw_buf_q,int lw_count_q, __DECLARATION_OF_q){

int recv_proc;


  int t1, t2;

 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;

/* Start of CLooG code */
if ((_PB_NY <= 0) && (ts1 == 0) && (ts2 >= 0) && (32*ts2 <= _PB_NX-1)) {
  for (t1=32*ts2;t1<=min(_PB_NX-1,32*ts2+31);t1++) {
    lw_buf_q[lw_count_q++] = q[t1];
  }
}
/* End of CLooG code */

return lw_count_q;


}

int writeout_unpack_q_1(int ts1,int ts2,double *lw_recv_buf_q,int displs_lw_q, __DECLARATION_OF_q){

int recv_proc;


  int t1, t2;

 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;

/* Start of CLooG code */
if ((_PB_NY <= 0) && (ts1 == 0) && (ts2 >= 0) && (32*ts2 <= _PB_NX-1)) {
  for (t1=32*ts2;t1<=min(_PB_NX-1,32*ts2+31);t1++) {
    
#ifndef USE_LOCAL_ARRAYS
 q[t1] = lw_recv_buf_q[displs_lw_q++]; 
#endif
;
  }
}
/* End of CLooG code */

return displs_lw_q;


}

void pack_send_1(int my_rank,int nprocs,int t1,int t2,__DECLARATION_OF_q,std::vector<double *>& send_buf_q,std::vector<MPI_Request>& send_reqs_q,double *p_total_count)
{
int __p, receiver_list[nprocs];
int send_counts_q[nprocs];
for (__p=0;__p<nprocs;__p++) send_counts_q[__p] = 0;
int index_q[nprocs];
int current_index_q = 0;
int flag;
double __total_count = 0;
for (__p=0; __p<nprocs; __p++) { receiver_list[__p] = 0; } sigma_q_1_1(t1,t2,_PB_NY,_PB_NX, my_rank, nprocs, receiver_list); for (__p=0; __p<nprocs; __p++) { if (receiver_list[__p] != 0) { if (send_counts_q[__p] == 0) { for (;current_index_q<send_reqs_q.size();current_index_q++) { if (send_reqs_q[current_index_q] == MPI_REQUEST_NULL) break; int __error = MPI_Test(&send_reqs_q[current_index_q], &flag, MPI_STATUS_IGNORE); IF_DYNSCHEDULER_DEBUG_PRINT( if (flag) { fprintf(__debug_print_fp, "node %d q send completed buf %d\n", my_rank, current_index_q); } if (__error) { char *msg=(char*)malloc((MPI_MAX_ERROR_STRING+1)*sizeof(char)); int len; MPI_Error_string(__error, msg, &len); fprintf(__debug_print_fp, "node %d q send test buf %d error %d %s\n", my_rank, current_index_q, __error, msg); } ) IF_DEBUG_FLUSH(fflush(__debug_print_fp)); if (flag) break; } if (current_index_q==send_reqs_q.size()) { double *buf = NULL; size_t buf_size = 0; buf = (double *) polyrt_max_alloc(buf, sizeof(double)*max_num_elements_q, &buf_size); send_buf_q.push_back(buf); IF_DYNSCHEDULER_DEBUG_PRINT( fprintf(__debug_print_fp, "node %d q max send buffers %d\n", my_rank, current_index_q+1)); IF_DEBUG_FLUSH(fflush(__debug_print_fp)); send_reqs_q.push_back(MPI_REQUEST_NULL); } index_q[__p] = current_index_q++; send_counts_q[__p] = 3; }send_counts_q[__p] = pack_q_1_1(t1,t2, send_buf_q[index_q[__p]], send_counts_q[__p],q); } };
    for (__p=0; __p<nprocs; __p++) { if(send_counts_q[__p] > 0) {send_buf_q[index_q[__p]][0] = 1;send_buf_q[index_q[__p]][1] = t1;send_buf_q[index_q[__p]][2] = t2; IF_DYNSCHEDULER_DEBUG_PRINT( fprintf(__debug_print_fp, "sending node %d buf %d loop 1 task %d %d 		        q size %d to node %d\n", my_rank, index_q[__p], t1,t2, send_counts_q[__p], __p)); IF_DEBUG_FLUSH(fflush(__debug_print_fp));IF_TIME(__total_count += send_counts_q[__p]); int __error = MPI_Isend(send_buf_q[index_q[__p]], send_counts_q[__p], MPI_DOUBLE, __p, 5123, MPI_COMM_WORLD, &send_reqs_q[index_q[__p]]); send_counts_q[__p] = 0; IF_DYNSCHEDULER_DEBUG_PRINT( if (__error) { char *msg=(char*)malloc((MPI_MAX_ERROR_STRING+1)*sizeof(char)); int len; MPI_Error_string(__error, msg, &len); fprintf(__debug_print_fp, "node %d q send buf %d error %d %s\n", my_rank, index_q[__p], __error, msg); } ) }};
*p_total_count += __total_count;
}

void unpack_buf_q_1(int my_rank,int nprocs,__DECLARATION_OF_q,double *recv_buf_q)
{
int curr_displs_q;
curr_displs_q = 3; if (is_receiver_q_1_1((int)recv_buf_q[1],(int)recv_buf_q[2],_PB_NY,_PB_NX,my_rank,nprocs) != 0) { curr_displs_q = unpack_q_1_1((int)recv_buf_q[1],(int)recv_buf_q[2],recv_buf_q,curr_displs_q,q); };
}

int writeout_pack_s_2(int ts1,int ts2,double *lw_buf_s,int lw_count_s, __DECLARATION_OF_s){

int recv_proc;


  int t1, t2;

 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;

/* Start of CLooG code */
if ((_PB_NX >= 1) && (ts1 == 3) && (ts2 >= 0) && (32*ts2 <= _PB_NY-1)) {
  for (t1=32*ts2;t1<=min(_PB_NY-1,32*ts2+31);t1++) {
    lw_buf_s[lw_count_s++] = s[t1];
  }
}
/* End of CLooG code */

return lw_count_s;


}

int writeout_unpack_s_2(int ts1,int ts2,double *lw_recv_buf_s,int displs_lw_s, __DECLARATION_OF_s){

int recv_proc;


  int t1, t2;

 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;

/* Start of CLooG code */
if ((_PB_NX >= 1) && (ts1 == 3) && (ts2 >= 0) && (32*ts2 <= _PB_NY-1)) {
  for (t1=32*ts2;t1<=min(_PB_NY-1,32*ts2+31);t1++) {
    
#ifndef USE_LOCAL_ARRAYS
 s[t1] = lw_recv_buf_s[displs_lw_s++]; 
#endif
;
  }
}
/* End of CLooG code */

return displs_lw_s;


}

void pack_send_2(int my_rank,int nprocs,int t1,int t2,__DECLARATION_OF_s,std::vector<double *>& send_buf_s,std::vector<MPI_Request>& send_reqs_s,double *p_total_count)
{
int __p, receiver_list[nprocs];
int send_counts_s[nprocs];
for (__p=0;__p<nprocs;__p++) send_counts_s[__p] = 0;
int index_s[nprocs];
int current_index_s = 0;
int flag;
double __total_count = 0;
;
    for (__p=0; __p<nprocs; __p++) { if(send_counts_s[__p] > 0) {send_buf_s[index_s[__p]][0] = 2;send_buf_s[index_s[__p]][1] = t1;send_buf_s[index_s[__p]][2] = t2; IF_DYNSCHEDULER_DEBUG_PRINT( fprintf(__debug_print_fp, "sending node %d buf %d loop 2 task %d %d 		        s size %d to node %d\n", my_rank, index_s[__p], t1,t2, send_counts_s[__p], __p)); IF_DEBUG_FLUSH(fflush(__debug_print_fp));IF_TIME(__total_count += send_counts_s[__p]); int __error = MPI_Isend(send_buf_s[index_s[__p]], send_counts_s[__p], MPI_DOUBLE, __p, 3123, MPI_COMM_WORLD, &send_reqs_s[index_s[__p]]); send_counts_s[__p] = 0; IF_DYNSCHEDULER_DEBUG_PRINT( if (__error) { char *msg=(char*)malloc((MPI_MAX_ERROR_STRING+1)*sizeof(char)); int len; MPI_Error_string(__error, msg, &len); fprintf(__debug_print_fp, "node %d s send buf %d error %d %s\n", my_rank, index_s[__p], __error, msg); } ) }};
*p_total_count += __total_count;
}

void unpack_buf_s_2(int my_rank,int nprocs,__DECLARATION_OF_s,double *recv_buf_s)
{
int curr_displs_s;
curr_displs_s = 3; ;
}

int writeout_pack_q_3(int ts1,int ts2,double *lw_buf_q,int lw_count_q, __DECLARATION_OF_q){

int recv_proc;


  int t1, t2;

 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;

/* Start of CLooG code */
if ((_PB_NY >= 1) && (ts1 == 1) && (ts2 >= 0) && (32*ts2 <= _PB_NX-1)) {
  for (t1=32*ts2;t1<=min(_PB_NX-1,32*ts2+31);t1++) {
    lw_buf_q[lw_count_q++] = q[t1];
  }
}
/* End of CLooG code */

return lw_count_q;


}

int writeout_unpack_q_3(int ts1,int ts2,double *lw_recv_buf_q,int displs_lw_q, __DECLARATION_OF_q){

int recv_proc;


  int t1, t2;

 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;

/* Start of CLooG code */
if ((_PB_NY >= 1) && (ts1 == 1) && (ts2 >= 0) && (32*ts2 <= _PB_NX-1)) {
  for (t1=32*ts2;t1<=min(_PB_NX-1,32*ts2+31);t1++) {
    
#ifndef USE_LOCAL_ARRAYS
 q[t1] = lw_recv_buf_q[displs_lw_q++]; 
#endif
;
  }
}
/* End of CLooG code */

return displs_lw_q;


}

void pack_send_3(int my_rank,int nprocs,int t1,int t2,__DECLARATION_OF_q,std::vector<double *>& send_buf_q,std::vector<MPI_Request>& send_reqs_q,double *p_total_count)
{
int __p, receiver_list[nprocs];
int send_counts_q[nprocs];
for (__p=0;__p<nprocs;__p++) send_counts_q[__p] = 0;
int index_q[nprocs];
int current_index_q = 0;
int flag;
double __total_count = 0;
;
    for (__p=0; __p<nprocs; __p++) { if(send_counts_q[__p] > 0) {send_buf_q[index_q[__p]][0] = 3;send_buf_q[index_q[__p]][1] = t1;send_buf_q[index_q[__p]][2] = t2; IF_DYNSCHEDULER_DEBUG_PRINT( fprintf(__debug_print_fp, "sending node %d buf %d loop 3 task %d %d 		        q size %d to node %d\n", my_rank, index_q[__p], t1,t2, send_counts_q[__p], __p)); IF_DEBUG_FLUSH(fflush(__debug_print_fp));IF_TIME(__total_count += send_counts_q[__p]); int __error = MPI_Isend(send_buf_q[index_q[__p]], send_counts_q[__p], MPI_DOUBLE, __p, 5123, MPI_COMM_WORLD, &send_reqs_q[index_q[__p]]); send_counts_q[__p] = 0; IF_DYNSCHEDULER_DEBUG_PRINT( if (__error) { char *msg=(char*)malloc((MPI_MAX_ERROR_STRING+1)*sizeof(char)); int len; MPI_Error_string(__error, msg, &len); fprintf(__debug_print_fp, "node %d q send buf %d error %d %s\n", my_rank, index_q[__p], __error, msg); } ) }};
*p_total_count += __total_count;
}

void unpack_buf_q_3(int my_rank,int nprocs,__DECLARATION_OF_q,double *recv_buf_q)
{
int curr_displs_q;
curr_displs_q = 3; ;
}

void write_out(int my_rank,int nprocs,double *lw_buf_s,double *lw_recv_buf_s,int *displs_lw_s,double *lw_buf_q,double *lw_recv_buf_q,int *displs_lw_q, __DECLARATION_OF_s, __DECLARATION_OF_q)
{
int __p, proc;
int _lb_dist, _ub_dist, lbd_t1, ubd_t1, lbd_t2, ubd_t2, lbd_t3, ubd_t3, lbd_t4, ubd_t4, lbd_t5, ubd_t5, lbd_t6, ubd_t6, lbd_t7, ubd_t7, lbd_t8, ubd_t8, lbd_t9, ubd_t9;
int lw_recv_counts_s[nprocs];
int curr_displs_lw_s[nprocs];
int lw_count_s = 0;
int lw_recv_counts_q[nprocs];
int curr_displs_lw_q[nprocs];
int lw_count_q = 0;
  int t1, t2, t3, t4, t5, t6;
 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;
/* Start of CLooG code */
for (t6=0;t6<=floord(_PB_NX-1,32);t6++) {
  if (pi_1(0,t6,_PB_NY,_PB_NX,nprocs) == my_rank)lw_count_q = writeout_pack_q_1(0,t6,lw_buf_q,lw_count_q,q);;
}
if (_PB_NX >= 1) {
  assert((nprocs == 1) || (lw_count_q <= displs_lw_q[1])); MPI_Gather(&lw_count_q, 1, MPI_INT, lw_recv_counts_q, 1, MPI_INT, 0, MPI_COMM_WORLD);MPI_Gatherv(lw_buf_q, lw_count_q, MPI_DOUBLE, lw_recv_buf_q, lw_recv_counts_q, displs_lw_q, MPI_DOUBLE, 0, MPI_COMM_WORLD); lw_count_q = 0; for (__p=0; __p<nprocs; __p++) curr_displs_lw_q[__p] = displs_lw_q[__p];;
}
if (my_rank == 0) {
  for (t6=0;t6<=floord(_PB_NX-1,32);t6++) {
    proc = pi_1(0, t6, _PB_NY, _PB_NX, nprocs);if ((my_rank != proc) && (lw_recv_counts_q[proc] > 0)) {curr_displs_lw_q[proc] = writeout_unpack_q_1(0,t6,lw_recv_buf_q,curr_displs_lw_q[proc],q);};
  }
}
if (_PB_NY >= 1) {
  for (t6=0;t6<=floord(_PB_NX-1,32);t6++) {
    if (pi_3(1,t6,_PB_NY,_PB_NX,nprocs) == my_rank)lw_count_q = writeout_pack_q_3(1,t6,lw_buf_q,lw_count_q,q);;
  }
}
if ((_PB_NX >= 1) && (_PB_NY >= 1)) {
  assert((nprocs == 1) || (lw_count_q <= displs_lw_q[1])); MPI_Gather(&lw_count_q, 1, MPI_INT, lw_recv_counts_q, 1, MPI_INT, 0, MPI_COMM_WORLD);MPI_Gatherv(lw_buf_q, lw_count_q, MPI_DOUBLE, lw_recv_buf_q, lw_recv_counts_q, displs_lw_q, MPI_DOUBLE, 0, MPI_COMM_WORLD); lw_count_q = 0; for (__p=0; __p<nprocs; __p++) curr_displs_lw_q[__p] = displs_lw_q[__p];;
}
if ((_PB_NY >= 1) && (my_rank == 0)) {
  for (t6=0;t6<=floord(_PB_NX-1,32);t6++) {
    proc = pi_3(1, t6, _PB_NY, _PB_NX, nprocs);if ((my_rank != proc) && (lw_recv_counts_q[proc] > 0)) {curr_displs_lw_q[proc] = writeout_unpack_q_3(1,t6,lw_recv_buf_q,curr_displs_lw_q[proc],q);};
  }
}
for (t6=0;t6<=floord(_PB_NY-1,32);t6++) {
  if (pi_0(2,t6,_PB_NY,_PB_NX,nprocs) == my_rank)lw_count_s = writeout_pack_s_0(2,t6,lw_buf_s,lw_count_s,s);;
}
if (_PB_NY >= 1) {
  assert((nprocs == 1) || (lw_count_s <= displs_lw_s[1])); MPI_Gather(&lw_count_s, 1, MPI_INT, lw_recv_counts_s, 1, MPI_INT, 0, MPI_COMM_WORLD);MPI_Gatherv(lw_buf_s, lw_count_s, MPI_DOUBLE, lw_recv_buf_s, lw_recv_counts_s, displs_lw_s, MPI_DOUBLE, 0, MPI_COMM_WORLD); lw_count_s = 0; for (__p=0; __p<nprocs; __p++) curr_displs_lw_s[__p] = displs_lw_s[__p];;
}
if (my_rank == 0) {
  for (t6=0;t6<=floord(_PB_NY-1,32);t6++) {
    proc = pi_0(2, t6, _PB_NY, _PB_NX, nprocs);if ((my_rank != proc) && (lw_recv_counts_s[proc] > 0)) {curr_displs_lw_s[proc] = writeout_unpack_s_0(2,t6,lw_recv_buf_s,curr_displs_lw_s[proc],s);};
  }
}
if (_PB_NX >= 1) {
  for (t6=0;t6<=floord(_PB_NY-1,32);t6++) {
    if (pi_2(3,t6,_PB_NY,_PB_NX,nprocs) == my_rank)lw_count_s = writeout_pack_s_2(3,t6,lw_buf_s,lw_count_s,s);;
  }
}
if ((_PB_NX >= 1) && (_PB_NY >= 1)) {
  assert((nprocs == 1) || (lw_count_s <= displs_lw_s[1])); MPI_Gather(&lw_count_s, 1, MPI_INT, lw_recv_counts_s, 1, MPI_INT, 0, MPI_COMM_WORLD);MPI_Gatherv(lw_buf_s, lw_count_s, MPI_DOUBLE, lw_recv_buf_s, lw_recv_counts_s, displs_lw_s, MPI_DOUBLE, 0, MPI_COMM_WORLD); lw_count_s = 0; for (__p=0; __p<nprocs; __p++) curr_displs_lw_s[__p] = displs_lw_s[__p];;
}
if ((_PB_NX >= 1) && (my_rank == 0)) {
  for (t6=0;t6<=floord(_PB_NY-1,32);t6++) {
    proc = pi_2(3, t6, _PB_NY, _PB_NX, nprocs);if ((my_rank != proc) && (lw_recv_counts_s[proc] > 0)) {curr_displs_lw_s[proc] = writeout_unpack_s_2(3,t6,lw_recv_buf_s,curr_displs_lw_s[proc],s);};
  }
}
/* End of CLooG code */
}
void compute_task_0(int t1,int t2,__DECLARATION_OF_s)
{
  int t3, t4, t5, t6, t7, t8, t9;
 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;
/* Start of CLooG code */
lbv=32*t2;
ubv=min(_PB_NY-1,32*t2+31);
#pragma ivdep
#pragma vector always
for (t3=lbv;t3<=ubv;t3++) {
  s[t3] = 0;;
}
/* End of CLooG code */
}
void compute_task_1(int t1,int t2,__DECLARATION_OF_q)
{
  int t3, t4, t5, t6, t7, t8, t9;
 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;
/* Start of CLooG code */
lbv=32*t2;
ubv=min(_PB_NX-1,32*t2+31);
#pragma ivdep
#pragma vector always
for (t3=lbv;t3<=ubv;t3++) {
  q[t3] = 0;;
}
/* End of CLooG code */
}
void compute_task_2(int t1,int t2,__DECLARATION_OF_s,__DECLARATION_OF_r,__DECLARATION_OF_A)
{
  int t3, t4, t5, t6, t7, t8, t9;
 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;
/* Start of CLooG code */
for (t3=0;t3<=floord(_PB_NX-1,32);t3++) {
  for (t4=32*t2;t4<=min(_PB_NY-1,32*t2+31);t4++) {
    lbv=32*t3;
    ubv=min(_PB_NX-1,32*t3+31);
#pragma ivdep
#pragma vector always
    for (t5=lbv;t5<=ubv;t5++) {
      s[t4] = s[t4] + r[t5] * A[t5][t4];;
    }
  }
}
/* End of CLooG code */
}
void compute_task_3(int t1,int t2,__DECLARATION_OF_q,__DECLARATION_OF_A,__DECLARATION_OF_p)
{
  int t3, t4, t5, t6, t7, t8, t9;
 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;
/* Start of CLooG code */
for (t3=0;t3<=floord(_PB_NY-1,32);t3++) {
  for (t4=32*t2;t4<=min(_PB_NX-1,32*t2+31);t4++) {
    lbv=32*t3;
    ubv=min(_PB_NY-1,32*t3+31);
#pragma ivdep
#pragma vector always
    for (t5=lbv;t5<=ubv;t5++) {
      q[t4] = q[t4] + A[t4][t5] * p[t5];;
    }
  }
}
/* End of CLooG code */
}
