#include "polyrt.h"
#include "buffer_manager.h"
extern int max_num_elements_s;
extern int max_num_elements_q;
extern int max_num_elements_r;
extern int max_num_elements_A;
extern int max_num_elements_p;
#include <mpi.h>
#include "tbb/concurrent_priority_queue.h"

#include <omp.h>

#ifdef __DEBUG_FLUSH
#define IF_DEBUG_FLUSH(foo) foo
#else
#define IF_DEBUG_FLUSH(foo) 
#endif

#ifdef __DYNSCHEDULER_MORE_DEBUG_PRINT
#define IF_DYNSCHEDULER_MORE_DEBUG_PRINT(foo) foo
#else
#define IF_DYNSCHEDULER_MORE_DEBUG_PRINT(foo) 
#endif

#ifdef __DYNSCHEDULER_DEBUG_PRINT
#define IF_DYNSCHEDULER_DEBUG_PRINT(foo) foo
#else
#define IF_DYNSCHEDULER_DEBUG_PRINT(foo) 
#endif

extern FILE *__debug_print_fp;

class __Task {
	public:
		int loop;
		int id1;
		int id2;
		int remote_dep_tasks;
		int local_dep_tasks;
		int thread_affinity;
		__Task() {}
		__Task(const int l, const int _id1, const int _id2, const int remote, const int local, const int affinity) :
			loop(l),
			id1(_id1),
			id2(_id2),
			remote_dep_tasks(remote),
			local_dep_tasks(local),
			thread_affinity(affinity) {}
		__Task(const __Task &task) :
			loop(task.loop),
			id1(task.id1),
			id2(task.id2),
			remote_dep_tasks(task.remote_dep_tasks),
			local_dep_tasks(task.local_dep_tasks),
			thread_affinity(task.thread_affinity) {}
		~__Task() {}
};

class __Compare {
	public:
		// returns true only if task2 has higher priority than task1
		bool operator()(const __Task &task1, const __Task &task2) const {
#ifndef __DYNSCHEDULER_NO_PRIORITY
			if (task2.remote_dep_tasks>task1.remote_dep_tasks) return true;
			if (task1.remote_dep_tasks>task2.remote_dep_tasks) return false;
			if (task2.local_dep_tasks>task1.local_dep_tasks) return true;
			if (task1.local_dep_tasks>task2.local_dep_tasks) return false;
			int thread_id = omp_get_thread_num();
			if ((thread_id == task2.thread_affinity) && (thread_id != task1.thread_affinity)) return true;
			if ((thread_id == task1.thread_affinity) && (thread_id != task2.thread_affinity)) return false;
			// lexicographic compare 
			if ( (task2.id1 > task1.id1) 
				|| ( (task2.id1 == task1.id1) && (task2.id2 > task1.id2) ) )
				 return true;
#endif
			return false;
		}
};

extern tbb::concurrent_priority_queue<__Task, __Compare> pqueue;

int is_receiver_0(int ts1, int ts2, int _PB_NY, int _PB_NX, int my_rank, int nprocs);
int get_num_remote_dep_tasks_0(int ts1, int ts2, int _PB_NY, int _PB_NX, int my_rank, int nprocs);
void remote_update_dep_tasks_0(int ts1, int ts2, int _PB_NY, int _PB_NX, int my_rank, int nprocs, int num_threads,int *tasks_loop0,int *tasks_loop1,int *tasks_loop2,int *tasks_loop3);
int get_num_local_dep_tasks_0(int ts1, int ts2, int _PB_NY, int _PB_NX, int my_rank, int nprocs);
void local_update_dep_tasks_0(int ts1, int ts2, int _PB_NY, int _PB_NX, int my_rank, int nprocs, int num_threads,int *tasks_loop0,int *tasks_loop1,int *tasks_loop2,int *tasks_loop3);
int get_num_remote_src_tasks_0(int ts1, int ts2, int _PB_NY, int _PB_NX, int my_rank, int nprocs);
int get_num_local_src_tasks_0(int ts1, int ts2, int _PB_NY, int _PB_NX, int my_rank, int nprocs);
extern int max_num_tasks_loop0_dim0;
extern int lb_tasks_loop0_dim0;
extern int *tasks_loop0;
int is_receiver_1(int ts1, int ts2, int _PB_NY, int _PB_NX, int my_rank, int nprocs);
int get_num_remote_dep_tasks_1(int ts1, int ts2, int _PB_NY, int _PB_NX, int my_rank, int nprocs);
void remote_update_dep_tasks_1(int ts1, int ts2, int _PB_NY, int _PB_NX, int my_rank, int nprocs, int num_threads,int *tasks_loop0,int *tasks_loop1,int *tasks_loop2,int *tasks_loop3);
int get_num_local_dep_tasks_1(int ts1, int ts2, int _PB_NY, int _PB_NX, int my_rank, int nprocs);
void local_update_dep_tasks_1(int ts1, int ts2, int _PB_NY, int _PB_NX, int my_rank, int nprocs, int num_threads,int *tasks_loop0,int *tasks_loop1,int *tasks_loop2,int *tasks_loop3);
int get_num_remote_src_tasks_1(int ts1, int ts2, int _PB_NY, int _PB_NX, int my_rank, int nprocs);
int get_num_local_src_tasks_1(int ts1, int ts2, int _PB_NY, int _PB_NX, int my_rank, int nprocs);
extern int max_num_tasks_loop1_dim0;
extern int lb_tasks_loop1_dim0;
extern int *tasks_loop1;
int is_receiver_2(int ts1, int ts2, int _PB_NY, int _PB_NX, int my_rank, int nprocs);
int get_num_remote_dep_tasks_2(int ts1, int ts2, int _PB_NY, int _PB_NX, int my_rank, int nprocs);
void remote_update_dep_tasks_2(int ts1, int ts2, int _PB_NY, int _PB_NX, int my_rank, int nprocs, int num_threads,int *tasks_loop0,int *tasks_loop1,int *tasks_loop2,int *tasks_loop3);
int get_num_local_dep_tasks_2(int ts1, int ts2, int _PB_NY, int _PB_NX, int my_rank, int nprocs);
void local_update_dep_tasks_2(int ts1, int ts2, int _PB_NY, int _PB_NX, int my_rank, int nprocs, int num_threads,int *tasks_loop0,int *tasks_loop1,int *tasks_loop2,int *tasks_loop3);
int get_num_remote_src_tasks_2(int ts1, int ts2, int _PB_NY, int _PB_NX, int my_rank, int nprocs);
int get_num_local_src_tasks_2(int ts1, int ts2, int _PB_NY, int _PB_NX, int my_rank, int nprocs);
extern int max_num_tasks_loop2_dim0;
extern int lb_tasks_loop2_dim0;
extern int *tasks_loop2;
int is_receiver_3(int ts1, int ts2, int _PB_NY, int _PB_NX, int my_rank, int nprocs);
int get_num_remote_dep_tasks_3(int ts1, int ts2, int _PB_NY, int _PB_NX, int my_rank, int nprocs);
void remote_update_dep_tasks_3(int ts1, int ts2, int _PB_NY, int _PB_NX, int my_rank, int nprocs, int num_threads,int *tasks_loop0,int *tasks_loop1,int *tasks_loop2,int *tasks_loop3);
int get_num_local_dep_tasks_3(int ts1, int ts2, int _PB_NY, int _PB_NX, int my_rank, int nprocs);
void local_update_dep_tasks_3(int ts1, int ts2, int _PB_NY, int _PB_NX, int my_rank, int nprocs, int num_threads,int *tasks_loop0,int *tasks_loop1,int *tasks_loop2,int *tasks_loop3);
int get_num_remote_src_tasks_3(int ts1, int ts2, int _PB_NY, int _PB_NX, int my_rank, int nprocs);
int get_num_local_src_tasks_3(int ts1, int ts2, int _PB_NY, int _PB_NX, int my_rank, int nprocs);
extern int max_num_tasks_loop3_dim0;
extern int lb_tasks_loop3_dim0;
extern int *tasks_loop3;
void init_tasks(int my_rank,int nprocs,int num_threads,int *tasks_loop0,int *tasks_loop1,int *tasks_loop2,int *tasks_loop3,long int *p_num_tasks_to_execute,long int *p_num_tasks_to_unpack);
void data_tile_ref_count_init_s_0(int ts1,int ts2, double** s_trans, BufferManager *buff_mang_s);
void data_tile_ref_count_init_q_1(int ts1,int ts2, double** q_trans, BufferManager *buff_mang_q);
void data_tile_ref_count_init_s_2(int ts1,int ts2, double** s_trans, BufferManager *buff_mang_s);
void data_tile_ref_count_init_r_2(int ts1,int ts2, double** r_trans, BufferManager *buff_mang_r);
void data_tile_ref_count_init_A_2(int ts1,int ts2, double** A_trans,int A_size_1, BufferManager *buff_mang_A);
void read_in_data_tile_alloc_s_2(int ts1,int ts2, double** s_trans, BufferManager *buff_mang_s);
void read_in_data_tile_alloc_r_2(int ts1,int ts2, double** r_trans, BufferManager *buff_mang_r);
void read_in_data_tile_alloc_A_2(int ts1,int ts2, double** A_trans,int A_size_1, BufferManager *buff_mang_A);
void data_tile_ref_count_init_q_3(int ts1,int ts2, double** q_trans, BufferManager *buff_mang_q);
void data_tile_ref_count_init_A_3(int ts1,int ts2, double** A_trans,int A_size_1, BufferManager *buff_mang_A);
void data_tile_ref_count_init_p_3(int ts1,int ts2, double** p_trans, BufferManager *buff_mang_p);
void read_in_data_tile_alloc_q_3(int ts1,int ts2, double** q_trans, BufferManager *buff_mang_q);
void read_in_data_tile_alloc_A_3(int ts1,int ts2, double** A_trans,int A_size_1, BufferManager *buff_mang_A);
void read_in_data_tile_alloc_p_3(int ts1,int ts2, double** p_trans, BufferManager *buff_mang_p);
int is_receiver_s_0(int ts1, int ts2, int _PB_NY, int _PB_NX, int my_rank, int nprocs);
void sigma_s_1_0(int ts1, int ts2, int _PB_NY, int _PB_NX, int my_rank, int nprocs, int *receiver_list);
int is_receiver_s_1_0(int ts1, int ts2, int _PB_NY, int _PB_NX, int my_rank, int nprocs);
int pack_s_1_0(int ts1,int ts2,double *send_buf_s,int send_counts_s, double** s_trans, BufferManager *buff_mang_s);
int unpack_s_1_0(int ts1,int ts2,double *recv_buf_s,int curr_displs_s, double** s_trans, BufferManager *buff_mang_s);
void pack_send_0(int my_rank,int nprocs,int t1,int t2, double** s_trans, BufferManager *buff_mang_s,std::vector<double *>& send_buf_s,std::vector<MPI_Request>& send_reqs_s,double *p_total_count);
void unpack_buf_s_0(int my_rank,int nprocs, double** s_trans, BufferManager *buff_mang_s,double *recv_buf_s);
int is_receiver_q_1(int ts1, int ts2, int _PB_NY, int _PB_NX, int my_rank, int nprocs);
void sigma_q_1_1(int ts1, int ts2, int _PB_NY, int _PB_NX, int my_rank, int nprocs, int *receiver_list);
int is_receiver_q_1_1(int ts1, int ts2, int _PB_NY, int _PB_NX, int my_rank, int nprocs);
int pack_q_1_1(int ts1,int ts2,double *send_buf_q,int send_counts_q, double** q_trans, BufferManager *buff_mang_q);
int unpack_q_1_1(int ts1,int ts2,double *recv_buf_q,int curr_displs_q, double** q_trans, BufferManager *buff_mang_q);
void pack_send_1(int my_rank,int nprocs,int t1,int t2, double** q_trans, BufferManager *buff_mang_q,std::vector<double *>& send_buf_q,std::vector<MPI_Request>& send_reqs_q,double *p_total_count);
void unpack_buf_q_1(int my_rank,int nprocs, double** q_trans, BufferManager *buff_mang_q,double *recv_buf_q);
int is_receiver_s_2(int ts1, int ts2, int _PB_NY, int _PB_NX, int my_rank, int nprocs);
void sigma_s_1_2(int ts1, int ts2, int _PB_NY, int _PB_NX, int my_rank, int nprocs, int *receiver_list);
int is_receiver_s_1_2(int ts1, int ts2, int _PB_NY, int _PB_NX, int my_rank, int nprocs);
int is_receiver_r_2(int ts1, int ts2, int _PB_NY, int _PB_NX, int my_rank, int nprocs);
void sigma_r_1_2(int ts1, int ts2, int _PB_NY, int _PB_NX, int my_rank, int nprocs, int *receiver_list);
int is_receiver_r_1_2(int ts1, int ts2, int _PB_NY, int _PB_NX, int my_rank, int nprocs);
int is_receiver_A_2(int ts1, int ts2, int _PB_NY, int _PB_NX, int my_rank, int nprocs);
void sigma_A_1_2(int ts1, int ts2, int _PB_NY, int _PB_NX, int my_rank, int nprocs, int *receiver_list);
int is_receiver_A_1_2(int ts1, int ts2, int _PB_NY, int _PB_NX, int my_rank, int nprocs);
void pack_send_2(int my_rank,int nprocs,int t1,int t2, double** s_trans, BufferManager *buff_mang_s,std::vector<double *>& send_buf_s,std::vector<MPI_Request>& send_reqs_s, double** r_trans, BufferManager *buff_mang_r,std::vector<double *>& send_buf_r,std::vector<MPI_Request>& send_reqs_r, double** A_trans,int A_size_1, BufferManager *buff_mang_A,std::vector<double *>& send_buf_A,std::vector<MPI_Request>& send_reqs_A,double *p_total_count);
void unpack_buf_s_2(int my_rank,int nprocs, double** s_trans, BufferManager *buff_mang_s,double *recv_buf_s);
void unpack_buf_r_2(int my_rank,int nprocs, double** r_trans, BufferManager *buff_mang_r,double *recv_buf_r);
void unpack_buf_A_2(int my_rank,int nprocs, double** A_trans,int A_size_1, BufferManager *buff_mang_A,double *recv_buf_A);
int is_receiver_q_3(int ts1, int ts2, int _PB_NY, int _PB_NX, int my_rank, int nprocs);
void sigma_q_1_3(int ts1, int ts2, int _PB_NY, int _PB_NX, int my_rank, int nprocs, int *receiver_list);
int is_receiver_q_1_3(int ts1, int ts2, int _PB_NY, int _PB_NX, int my_rank, int nprocs);
int is_receiver_A_3(int ts1, int ts2, int _PB_NY, int _PB_NX, int my_rank, int nprocs);
void sigma_A_1_3(int ts1, int ts2, int _PB_NY, int _PB_NX, int my_rank, int nprocs, int *receiver_list);
int is_receiver_A_1_3(int ts1, int ts2, int _PB_NY, int _PB_NX, int my_rank, int nprocs);
int is_receiver_p_3(int ts1, int ts2, int _PB_NY, int _PB_NX, int my_rank, int nprocs);
void sigma_p_1_3(int ts1, int ts2, int _PB_NY, int _PB_NX, int my_rank, int nprocs, int *receiver_list);
int is_receiver_p_1_3(int ts1, int ts2, int _PB_NY, int _PB_NX, int my_rank, int nprocs);
void pack_send_3(int my_rank,int nprocs,int t1,int t2, double** q_trans, BufferManager *buff_mang_q,std::vector<double *>& send_buf_q,std::vector<MPI_Request>& send_reqs_q, double** A_trans,int A_size_1, BufferManager *buff_mang_A,std::vector<double *>& send_buf_A,std::vector<MPI_Request>& send_reqs_A, double** p_trans, BufferManager *buff_mang_p,std::vector<double *>& send_buf_p,std::vector<MPI_Request>& send_reqs_p,double *p_total_count);
void unpack_buf_q_3(int my_rank,int nprocs, double** q_trans, BufferManager *buff_mang_q,double *recv_buf_q);
void unpack_buf_A_3(int my_rank,int nprocs, double** A_trans,int A_size_1, BufferManager *buff_mang_A,double *recv_buf_A);
void unpack_buf_p_3(int my_rank,int nprocs, double** p_trans, BufferManager *buff_mang_p,double *recv_buf_p);
int pi_0( int t1, int t6, int _PB_NY, int _PB_NX, int nprocs);
int pi_threads_0( int t1, int t6, int _PB_NY, int _PB_NX, int my_rank, int nprocs, int num_threads);
int pi_1( int t1, int t6, int _PB_NY, int _PB_NX, int nprocs);
int pi_threads_1( int t1, int t6, int _PB_NY, int _PB_NX, int my_rank, int nprocs, int num_threads);
int pi_2( int t1, int t6, int _PB_NY, int _PB_NX, int nprocs);
int pi_threads_2( int t1, int t6, int _PB_NY, int _PB_NX, int my_rank, int nprocs, int num_threads);
int pi_3( int t1, int t6, int _PB_NY, int _PB_NX, int nprocs);
int pi_threads_3( int t1, int t6, int _PB_NY, int _PB_NX, int my_rank, int nprocs, int num_threads);
void data_tile_alloc_0(int ts1,int ts2, double** s_trans, BufferManager *buff_mang_s);
void data_tile_ref_count_update_0(int ts1,int ts2, double** s_trans, BufferManager *buff_mang_s);
void compute_task_0(int t1,int t2, double** s_trans, BufferManager *buff_mang_s);
void data_tile_alloc_1(int ts1,int ts2, double** q_trans, BufferManager *buff_mang_q);
void data_tile_ref_count_update_1(int ts1,int ts2, double** q_trans, BufferManager *buff_mang_q);
void compute_task_1(int t1,int t2, double** q_trans, BufferManager *buff_mang_q);
void data_tile_alloc_2(int ts1,int ts2, double** s_trans, BufferManager *buff_mang_s, double** r_trans, BufferManager *buff_mang_r, double** A_trans,int A_size_1, BufferManager *buff_mang_A);
void data_tile_ref_count_update_2(int ts1,int ts2, double** s_trans, BufferManager *buff_mang_s, double** r_trans, BufferManager *buff_mang_r, double** A_trans,int A_size_1, BufferManager *buff_mang_A);
void compute_task_2(int t1,int t2, double** s_trans, BufferManager *buff_mang_s, double** r_trans, BufferManager *buff_mang_r, double** A_trans,int A_size_1, BufferManager *buff_mang_A);
void data_tile_alloc_3(int ts1,int ts2, double** q_trans, BufferManager *buff_mang_q, double** A_trans,int A_size_1, BufferManager *buff_mang_A, double** p_trans, BufferManager *buff_mang_p);
void data_tile_ref_count_update_3(int ts1,int ts2, double** q_trans, BufferManager *buff_mang_q, double** A_trans,int A_size_1, BufferManager *buff_mang_A, double** p_trans, BufferManager *buff_mang_p);
void compute_task_3(int t1,int t2, double** q_trans, BufferManager *buff_mang_q, double** A_trans,int A_size_1, BufferManager *buff_mang_A, double** p_trans, BufferManager *buff_mang_p);
