#include "polyrt.h"
#include "buffer_manager.h"
#include <mpi.h>
#include "tbb/concurrent_priority_queue.h"

#include <omp.h>

#ifdef __DEBUG_FLUSH
#define IF_DEBUG_FLUSH(foo) foo
#else
#define IF_DEBUG_FLUSH(foo) 
#endif

#ifdef __DYNSCHEDULER_MORE_DEBUG_PRINT
#define IF_DYNSCHEDULER_MORE_DEBUG_PRINT(foo) foo
#else
#define IF_DYNSCHEDULER_MORE_DEBUG_PRINT(foo) 
#endif

#ifdef __DYNSCHEDULER_DEBUG_PRINT
#define IF_DYNSCHEDULER_DEBUG_PRINT(foo) foo
#else
#define IF_DYNSCHEDULER_DEBUG_PRINT(foo) 
#endif

extern FILE *__debug_print_fp;

class __Task {
	public:
		int loop;
		int id1;
		int id2;
		int id3;
		int remote_dep_tasks;
		int local_dep_tasks;
		int thread_affinity;
		__Task() {}
		__Task(const int l, const int _id1, const int _id2, const int _id3, const int remote, const int local, const int affinity) :
			loop(l),
			id1(_id1),
			id2(_id2),
			id3(_id3),
			remote_dep_tasks(remote),
			local_dep_tasks(local),
			thread_affinity(affinity) {}
		__Task(const int l, const int _id1, const int _id2, const int remote, const int local, const int affinity) :
			loop(l),
			id1(_id1),
			id2(_id2),
			id3(0),
			remote_dep_tasks(remote),
			local_dep_tasks(local),
			thread_affinity(affinity) {}
		__Task(const __Task &task) :
			loop(task.loop),
			id1(task.id1),
			id2(task.id2),
			id3(task.id3),
			remote_dep_tasks(task.remote_dep_tasks),
			local_dep_tasks(task.local_dep_tasks),
			thread_affinity(task.thread_affinity) {}
		~__Task() {}
};

class __Compare {
	public:
		// returns true only if task2 has higher priority than task1
		bool operator()(const __Task &task1, const __Task &task2) const {
#ifndef __DYNSCHEDULER_NO_PRIORITY
			if (task2.remote_dep_tasks>task1.remote_dep_tasks) return true;
			if (task1.remote_dep_tasks>task2.remote_dep_tasks) return false;
			if (task2.local_dep_tasks>task1.local_dep_tasks) return true;
			if (task1.local_dep_tasks>task2.local_dep_tasks) return false;
			int thread_id = omp_get_thread_num();
			if ((thread_id == task2.thread_affinity) && (thread_id != task1.thread_affinity)) return true;
			if ((thread_id == task1.thread_affinity) && (thread_id != task2.thread_affinity)) return false;
			// lexicographic compare 
			if ( (task2.id1 > task1.id1) 
				|| ( (task2.id1 == task1.id1) && (task2.id2 > task1.id2) ) 
				|| ( (task2.id1 == task1.id1) && (task2.id2 == task1.id2) && (task2.id3 > task1.id3) ) )
				 return true;
#endif
			return false;
		}
};

extern tbb::concurrent_priority_queue<__Task, __Compare> pqueue;

int is_receiver_0(int ts1, int ts2, int ts3, int N, int n, int my_rank, int nprocs);
int get_num_remote_dep_tasks_0(int ts1, int ts2, int ts3, int N, int n, int my_rank, int nprocs);
void remote_update_dep_tasks_0(int ts1, int ts2, int ts3, int N, int n, int my_rank, int nprocs, int num_threads,int *tasks_loop0,int *tasks_loop1,int *tasks_loop2,int *tasks_loop3);
int get_num_local_dep_tasks_0(int ts1, int ts2, int ts3, int N, int n, int my_rank, int nprocs);
void local_update_dep_tasks_0(int ts1, int ts2, int ts3, int N, int n, int my_rank, int nprocs, int num_threads,int *tasks_loop0,int *tasks_loop1,int *tasks_loop2,int *tasks_loop3);
int get_num_remote_src_tasks_0(int ts1, int ts2, int ts3, int N, int n, int my_rank, int nprocs);
int get_num_local_src_tasks_0(int ts1, int ts2, int ts3, int N, int n, int my_rank, int nprocs);
extern int max_num_tasks_loop0_dim0;
extern int lb_tasks_loop0_dim0;
extern int max_num_tasks_loop0_dim1;
extern int lb_tasks_loop0_dim1;
extern int *tasks_loop0;
int is_receiver_1(int ts1, int ts2, int ts3, int N, int n, int my_rank, int nprocs);
int get_num_remote_dep_tasks_1(int ts1, int ts2, int ts3, int N, int n, int my_rank, int nprocs);
void remote_update_dep_tasks_1(int ts1, int ts2, int ts3, int N, int n, int my_rank, int nprocs, int num_threads,int *tasks_loop0,int *tasks_loop1,int *tasks_loop2,int *tasks_loop3);
int get_num_local_dep_tasks_1(int ts1, int ts2, int ts3, int N, int n, int my_rank, int nprocs);
void local_update_dep_tasks_1(int ts1, int ts2, int ts3, int N, int n, int my_rank, int nprocs, int num_threads,int *tasks_loop0,int *tasks_loop1,int *tasks_loop2,int *tasks_loop3);
int get_num_remote_src_tasks_1(int ts1, int ts2, int ts3, int N, int n, int my_rank, int nprocs);
int get_num_local_src_tasks_1(int ts1, int ts2, int ts3, int N, int n, int my_rank, int nprocs);
extern int max_num_tasks_loop1_dim0;
extern int lb_tasks_loop1_dim0;
extern int max_num_tasks_loop1_dim1;
extern int lb_tasks_loop1_dim1;
extern int *tasks_loop1;
int is_receiver_2(int ts1, int ts2, int N, int n, int my_rank, int nprocs);
int get_num_remote_dep_tasks_2(int ts1, int ts2, int N, int n, int my_rank, int nprocs);
void remote_update_dep_tasks_2(int ts1, int ts2, int N, int n, int my_rank, int nprocs, int num_threads,int *tasks_loop0,int *tasks_loop1,int *tasks_loop2,int *tasks_loop3);
int get_num_local_dep_tasks_2(int ts1, int ts2, int N, int n, int my_rank, int nprocs);
void local_update_dep_tasks_2(int ts1, int ts2, int N, int n, int my_rank, int nprocs, int num_threads,int *tasks_loop0,int *tasks_loop1,int *tasks_loop2,int *tasks_loop3);
int get_num_remote_src_tasks_2(int ts1, int ts2, int N, int n, int my_rank, int nprocs);
int get_num_local_src_tasks_2(int ts1, int ts2, int N, int n, int my_rank, int nprocs);
extern int max_num_tasks_loop2_dim0;
extern int lb_tasks_loop2_dim0;
extern int *tasks_loop2;
int is_receiver_3(int ts1, int ts2, int ts3, int N, int n, int my_rank, int nprocs);
int get_num_remote_dep_tasks_3(int ts1, int ts2, int ts3, int N, int n, int my_rank, int nprocs);
void remote_update_dep_tasks_3(int ts1, int ts2, int ts3, int N, int n, int my_rank, int nprocs, int num_threads,int *tasks_loop0,int *tasks_loop1,int *tasks_loop2,int *tasks_loop3);
int get_num_local_dep_tasks_3(int ts1, int ts2, int ts3, int N, int n, int my_rank, int nprocs);
void local_update_dep_tasks_3(int ts1, int ts2, int ts3, int N, int n, int my_rank, int nprocs, int num_threads,int *tasks_loop0,int *tasks_loop1,int *tasks_loop2,int *tasks_loop3);
int get_num_remote_src_tasks_3(int ts1, int ts2, int ts3, int N, int n, int my_rank, int nprocs);
int get_num_local_src_tasks_3(int ts1, int ts2, int ts3, int N, int n, int my_rank, int nprocs);
extern int max_num_tasks_loop3_dim0;
extern int lb_tasks_loop3_dim0;
extern int max_num_tasks_loop3_dim1;
extern int lb_tasks_loop3_dim1;
extern int *tasks_loop3;
void init_tasks(int my_rank,int nprocs,int num_threads,int *tasks_loop0,int *tasks_loop1,int *tasks_loop2,int *tasks_loop3,long int *p_num_tasks_to_execute,long int *p_num_tasks_to_unpack);
void data_tile_ref_count_init_B_0(int ts1,int ts2,int ts3, double** B_trans,int B_size_1, BufferManager *buff_mang_B);
void data_tile_ref_count_init_A_0(int ts1,int ts2,int ts3, double** A_trans,int A_size_1, BufferManager *buff_mang_A);
void data_tile_ref_count_init_u1_0(int ts1,int ts2,int ts3, double** u1_trans, BufferManager *buff_mang_u1);
void data_tile_ref_count_init_v1_0(int ts1,int ts2,int ts3, double** v1_trans, BufferManager *buff_mang_v1);
void data_tile_ref_count_init_u2_0(int ts1,int ts2,int ts3, double** u2_trans, BufferManager *buff_mang_u2);
void data_tile_ref_count_init_v2_0(int ts1,int ts2,int ts3, double** v2_trans, BufferManager *buff_mang_v2);
void read_in_data_tile_alloc_A_0(int ts1,int ts2,int ts3, double** A_trans,int A_size_1, BufferManager *buff_mang_A);
void read_in_copy_A_0(int ts1,int ts2,int ts3, double** A_trans,int A_size_1, BufferManager *buff_mang_A);
void read_in_data_tile_alloc_u1_0(int ts1,int ts2,int ts3, double** u1_trans, BufferManager *buff_mang_u1);
void read_in_copy_u1_0(int ts1,int ts2,int ts3, double** u1_trans, BufferManager *buff_mang_u1);
void read_in_data_tile_alloc_v1_0(int ts1,int ts2,int ts3, double** v1_trans, BufferManager *buff_mang_v1);
void read_in_copy_v1_0(int ts1,int ts2,int ts3, double** v1_trans, BufferManager *buff_mang_v1);
void read_in_data_tile_alloc_u2_0(int ts1,int ts2,int ts3, double** u2_trans, BufferManager *buff_mang_u2);
void read_in_copy_u2_0(int ts1,int ts2,int ts3, double** u2_trans, BufferManager *buff_mang_u2);
void read_in_data_tile_alloc_v2_0(int ts1,int ts2,int ts3, double** v2_trans, BufferManager *buff_mang_v2);
void read_in_copy_v2_0(int ts1,int ts2,int ts3, double** v2_trans, BufferManager *buff_mang_v2);
void data_tile_ref_count_init_x_1(int ts1,int ts2,int ts3, double** x_trans, BufferManager *buff_mang_x);
void data_tile_ref_count_init_B_1(int ts1,int ts2,int ts3, double** B_trans,int B_size_1, BufferManager *buff_mang_B);
void data_tile_ref_count_init_y_1(int ts1,int ts2,int ts3, double** y_trans, BufferManager *buff_mang_y);
void read_in_data_tile_alloc_x_1(int ts1,int ts2,int ts3, double** x_trans, BufferManager *buff_mang_x);
void read_in_copy_x_1(int ts1,int ts2,int ts3, double** x_trans, BufferManager *buff_mang_x);
void read_in_data_tile_alloc_B_1(int ts1,int ts2,int ts3, double** B_trans,int B_size_1, BufferManager *buff_mang_B);
void read_in_copy_B_1(int ts1,int ts2,int ts3, double** B_trans,int B_size_1, BufferManager *buff_mang_B);
void read_in_data_tile_alloc_y_1(int ts1,int ts2,int ts3, double** y_trans, BufferManager *buff_mang_y);
void read_in_copy_y_1(int ts1,int ts2,int ts3, double** y_trans, BufferManager *buff_mang_y);
void data_tile_ref_count_init_x_2(int ts1,int ts2, double** x_trans, BufferManager *buff_mang_x);
void data_tile_ref_count_init_z_2(int ts1,int ts2, double** z_trans, BufferManager *buff_mang_z);
void read_in_data_tile_alloc_x_2(int ts1,int ts2, double** x_trans, BufferManager *buff_mang_x);
void read_in_copy_x_2(int ts1,int ts2, double** x_trans, BufferManager *buff_mang_x);
void read_in_data_tile_alloc_z_2(int ts1,int ts2, double** z_trans, BufferManager *buff_mang_z);
void read_in_copy_z_2(int ts1,int ts2, double** z_trans, BufferManager *buff_mang_z);
void data_tile_ref_count_init_w_3(int ts1,int ts2,int ts3, double** w_trans, BufferManager *buff_mang_w);
void data_tile_ref_count_init_B_3(int ts1,int ts2,int ts3, double** B_trans,int B_size_1, BufferManager *buff_mang_B);
void data_tile_ref_count_init_x_3(int ts1,int ts2,int ts3, double** x_trans, BufferManager *buff_mang_x);
void read_in_data_tile_alloc_w_3(int ts1,int ts2,int ts3, double** w_trans, BufferManager *buff_mang_w);
void read_in_copy_w_3(int ts1,int ts2,int ts3, double** w_trans, BufferManager *buff_mang_w);
void read_in_data_tile_alloc_B_3(int ts1,int ts2,int ts3, double** B_trans,int B_size_1, BufferManager *buff_mang_B);
void read_in_copy_B_3(int ts1,int ts2,int ts3, double** B_trans,int B_size_1, BufferManager *buff_mang_B);
void read_in_data_tile_alloc_x_3(int ts1,int ts2,int ts3, double** x_trans, BufferManager *buff_mang_x);
void read_in_copy_x_3(int ts1,int ts2,int ts3, double** x_trans, BufferManager *buff_mang_x);
void data_tile_init_0(int my_rank, int nprocs, double** B_trans,int B_size_1, BufferManager *buff_mang_B, double** A_trans,int A_size_1, BufferManager *buff_mang_A, double** u1_trans, BufferManager *buff_mang_u1, double** v1_trans, BufferManager *buff_mang_v1, double** u2_trans, BufferManager *buff_mang_u2, double** v2_trans, BufferManager *buff_mang_v2, double** x_trans, BufferManager *buff_mang_x, double** y_trans, BufferManager *buff_mang_y, double** z_trans, BufferManager *buff_mang_z, double** w_trans, BufferManager *buff_mang_w);
int is_receiver_A_0(int ts1, int ts2, int ts3, int N, int n, int my_rank, int nprocs);
void sigma_A_1_0(int ts1, int ts2, int ts3, int N, int n, int my_rank, int nprocs, int *receiver_list);
int is_receiver_A_1_0(int ts1, int ts2, int ts3, int N, int n, int my_rank, int nprocs);
int is_receiver_u1_0(int ts1, int ts2, int ts3, int N, int n, int my_rank, int nprocs);
void sigma_u1_1_0(int ts1, int ts2, int ts3, int N, int n, int my_rank, int nprocs, int *receiver_list);
int is_receiver_u1_1_0(int ts1, int ts2, int ts3, int N, int n, int my_rank, int nprocs);
int is_receiver_v1_0(int ts1, int ts2, int ts3, int N, int n, int my_rank, int nprocs);
void sigma_v1_1_0(int ts1, int ts2, int ts3, int N, int n, int my_rank, int nprocs, int *receiver_list);
int is_receiver_v1_1_0(int ts1, int ts2, int ts3, int N, int n, int my_rank, int nprocs);
int is_receiver_u2_0(int ts1, int ts2, int ts3, int N, int n, int my_rank, int nprocs);
void sigma_u2_1_0(int ts1, int ts2, int ts3, int N, int n, int my_rank, int nprocs, int *receiver_list);
int is_receiver_u2_1_0(int ts1, int ts2, int ts3, int N, int n, int my_rank, int nprocs);
int is_receiver_v2_0(int ts1, int ts2, int ts3, int N, int n, int my_rank, int nprocs);
void sigma_v2_1_0(int ts1, int ts2, int ts3, int N, int n, int my_rank, int nprocs, int *receiver_list);
int is_receiver_v2_1_0(int ts1, int ts2, int ts3, int N, int n, int my_rank, int nprocs);
int is_receiver_B_0(int ts1, int ts2, int ts3, int N, int n, int my_rank, int nprocs);
void sigma_B_1_0(int ts1, int ts2, int ts3, int N, int n, int my_rank, int nprocs, int *receiver_list);
int is_receiver_B_1_0(int ts1, int ts2, int ts3, int N, int n, int my_rank, int nprocs);
int pack_B_1_0(int ts1,int ts2,int ts3,double *send_buf_B,int send_counts_B, double** B_trans,int B_size_1, BufferManager *buff_mang_B);
int unpack_B_1_0(int ts1,int ts2,int ts3,double *recv_buf_B,int curr_displs_B, double** B_trans,int B_size_1, BufferManager *buff_mang_B);
void data_tile_copy_back_B_0(int ts1,int ts2,int ts3, double** B_trans,int B_size_1, BufferManager *buff_mang_B);
int writeout_pack_B_0(int ts1,int ts2,int ts3,double *lw_buf_B,int lw_count_B, double** B_trans,int B_size_1, BufferManager *buff_mang_B);
int writeout_unpack_B_0(int ts1,int ts2,int ts3,double *lw_recv_buf_B,int displs_lw_B, double** B_trans,int B_size_1, BufferManager *buff_mang_B);
void pack_send_0(int my_rank,int nprocs,int t1,int t2,int t3, double** B_trans,int B_size_1, BufferManager *buff_mang_B,double **send_buf_B,MPI_Request *send_reqs_B,MPI_Status *send_stats_B, double** A_trans,int A_size_1, BufferManager *buff_mang_A,double **send_buf_A,MPI_Request *send_reqs_A,MPI_Status *send_stats_A, double** u1_trans, BufferManager *buff_mang_u1,double **send_buf_u1,MPI_Request *send_reqs_u1,MPI_Status *send_stats_u1, double** v1_trans, BufferManager *buff_mang_v1,double **send_buf_v1,MPI_Request *send_reqs_v1,MPI_Status *send_stats_v1, double** u2_trans, BufferManager *buff_mang_u2,double **send_buf_u2,MPI_Request *send_reqs_u2,MPI_Status *send_stats_u2, double** v2_trans, BufferManager *buff_mang_v2,double **send_buf_v2,MPI_Request *send_reqs_v2,MPI_Status *send_stats_v2,int max_num_sends,double *p_total_count);
void unpack_buf_A_0(int my_rank,int nprocs, double** A_trans,int A_size_1, BufferManager *buff_mang_A,double *recv_buf_A);
void unpack_buf_u1_0(int my_rank,int nprocs, double** u1_trans, BufferManager *buff_mang_u1,double *recv_buf_u1);
void unpack_buf_v1_0(int my_rank,int nprocs, double** v1_trans, BufferManager *buff_mang_v1,double *recv_buf_v1);
void unpack_buf_u2_0(int my_rank,int nprocs, double** u2_trans, BufferManager *buff_mang_u2,double *recv_buf_u2);
void unpack_buf_v2_0(int my_rank,int nprocs, double** v2_trans, BufferManager *buff_mang_v2,double *recv_buf_v2);
void unpack_buf_B_0(int my_rank,int nprocs, double** B_trans,int B_size_1, BufferManager *buff_mang_B,double *recv_buf_B);
int is_receiver_x_1(int ts1, int ts2, int ts3, int N, int n, int my_rank, int nprocs);
void sigma_x_1_1(int ts1, int ts2, int ts3, int N, int n, int my_rank, int nprocs, int *receiver_list);
int is_receiver_x_1_1(int ts1, int ts2, int ts3, int N, int n, int my_rank, int nprocs);
void sigma_x_2_1(int ts1, int ts2, int ts3, int N, int n, int my_rank, int nprocs, int *receiver_list);
int is_receiver_x_2_1(int ts1, int ts2, int ts3, int N, int n, int my_rank, int nprocs);
void sigma_x_3_1(int ts1, int ts2, int ts3, int N, int n, int my_rank, int nprocs, int *receiver_list);
int is_receiver_x_3_1(int ts1, int ts2, int ts3, int N, int n, int my_rank, int nprocs);
int pack_x_1_1(int ts1,int ts2,int ts3,double *send_buf_x,int send_counts_x, double** x_trans, BufferManager *buff_mang_x);
int unpack_x_1_1(int ts1,int ts2,int ts3,double *recv_buf_x,int curr_displs_x, double** x_trans, BufferManager *buff_mang_x);
int pack_x_2_1(int ts1,int ts2,int ts3,double *send_buf_x,int send_counts_x, double** x_trans, BufferManager *buff_mang_x);
int unpack_x_2_1(int ts1,int ts2,int ts3,double *recv_buf_x,int curr_displs_x, double** x_trans, BufferManager *buff_mang_x);
int pack_x_3_1(int ts1,int ts2,int ts3,double *send_buf_x,int send_counts_x, double** x_trans, BufferManager *buff_mang_x);
int unpack_x_3_1(int ts1,int ts2,int ts3,double *recv_buf_x,int curr_displs_x, double** x_trans, BufferManager *buff_mang_x);
int is_receiver_B_1(int ts1, int ts2, int ts3, int N, int n, int my_rank, int nprocs);
void sigma_B_1_1(int ts1, int ts2, int ts3, int N, int n, int my_rank, int nprocs, int *receiver_list);
int is_receiver_B_1_1(int ts1, int ts2, int ts3, int N, int n, int my_rank, int nprocs);
int is_receiver_y_1(int ts1, int ts2, int ts3, int N, int n, int my_rank, int nprocs);
void sigma_y_1_1(int ts1, int ts2, int ts3, int N, int n, int my_rank, int nprocs, int *receiver_list);
int is_receiver_y_1_1(int ts1, int ts2, int ts3, int N, int n, int my_rank, int nprocs);
void data_tile_copy_back_x_1(int ts1,int ts2,int ts3, double** x_trans, BufferManager *buff_mang_x);
int writeout_pack_x_1(int ts1,int ts2,int ts3,double *lw_buf_x,int lw_count_x, double** x_trans, BufferManager *buff_mang_x);
int writeout_unpack_x_1(int ts1,int ts2,int ts3,double *lw_recv_buf_x,int displs_lw_x, double** x_trans, BufferManager *buff_mang_x);
void pack_send_1(int my_rank,int nprocs,int t1,int t2,int t3, double** x_trans, BufferManager *buff_mang_x,double **send_buf_x,MPI_Request *send_reqs_x,MPI_Status *send_stats_x, double** B_trans,int B_size_1, BufferManager *buff_mang_B,double **send_buf_B,MPI_Request *send_reqs_B,MPI_Status *send_stats_B, double** y_trans, BufferManager *buff_mang_y,double **send_buf_y,MPI_Request *send_reqs_y,MPI_Status *send_stats_y,int max_num_sends,double *p_total_count);
void unpack_buf_x_1(int my_rank,int nprocs, double** x_trans, BufferManager *buff_mang_x,double *recv_buf_x);
void unpack_buf_B_1(int my_rank,int nprocs, double** B_trans,int B_size_1, BufferManager *buff_mang_B,double *recv_buf_B);
void unpack_buf_y_1(int my_rank,int nprocs, double** y_trans, BufferManager *buff_mang_y,double *recv_buf_y);
int is_receiver_x_2(int ts1, int ts2, int N, int n, int my_rank, int nprocs);
void sigma_x_1_2(int ts1, int ts2, int N, int n, int my_rank, int nprocs, int *receiver_list);
int is_receiver_x_1_2(int ts1, int ts2, int N, int n, int my_rank, int nprocs);
int pack_x_1_2(int ts1,int ts2,double *send_buf_x,int send_counts_x, double** x_trans, BufferManager *buff_mang_x);
int unpack_x_1_2(int ts1,int ts2,double *recv_buf_x,int curr_displs_x, double** x_trans, BufferManager *buff_mang_x);
int is_receiver_z_2(int ts1, int ts2, int N, int n, int my_rank, int nprocs);
void sigma_z_1_2(int ts1, int ts2, int N, int n, int my_rank, int nprocs, int *receiver_list);
int is_receiver_z_1_2(int ts1, int ts2, int N, int n, int my_rank, int nprocs);
void data_tile_copy_back_x_2(int ts1,int ts2, double** x_trans, BufferManager *buff_mang_x);
int writeout_pack_x_2(int ts1,int ts2,double *lw_buf_x,int lw_count_x, double** x_trans, BufferManager *buff_mang_x);
int writeout_unpack_x_2(int ts1,int ts2,double *lw_recv_buf_x,int displs_lw_x, double** x_trans, BufferManager *buff_mang_x);
void pack_send_2(int my_rank,int nprocs,int t1,int t2, double** x_trans, BufferManager *buff_mang_x,double **send_buf_x,MPI_Request *send_reqs_x,MPI_Status *send_stats_x, double** z_trans, BufferManager *buff_mang_z,double **send_buf_z,MPI_Request *send_reqs_z,MPI_Status *send_stats_z,int max_num_sends,double *p_total_count);
void unpack_buf_x_2(int my_rank,int nprocs, double** x_trans, BufferManager *buff_mang_x,double *recv_buf_x);
void unpack_buf_z_2(int my_rank,int nprocs, double** z_trans, BufferManager *buff_mang_z,double *recv_buf_z);
int is_receiver_w_3(int ts1, int ts2, int ts3, int N, int n, int my_rank, int nprocs);
void sigma_w_1_3(int ts1, int ts2, int ts3, int N, int n, int my_rank, int nprocs, int *receiver_list);
int is_receiver_w_1_3(int ts1, int ts2, int ts3, int N, int n, int my_rank, int nprocs);
int pack_w_1_3(int ts1,int ts2,int ts3,double *send_buf_w,int send_counts_w, double** w_trans, BufferManager *buff_mang_w);
int unpack_w_1_3(int ts1,int ts2,int ts3,double *recv_buf_w,int curr_displs_w, double** w_trans, BufferManager *buff_mang_w);
int is_receiver_B_3(int ts1, int ts2, int ts3, int N, int n, int my_rank, int nprocs);
void sigma_B_1_3(int ts1, int ts2, int ts3, int N, int n, int my_rank, int nprocs, int *receiver_list);
int is_receiver_B_1_3(int ts1, int ts2, int ts3, int N, int n, int my_rank, int nprocs);
int is_receiver_x_3(int ts1, int ts2, int ts3, int N, int n, int my_rank, int nprocs);
void sigma_x_1_3(int ts1, int ts2, int ts3, int N, int n, int my_rank, int nprocs, int *receiver_list);
int is_receiver_x_1_3(int ts1, int ts2, int ts3, int N, int n, int my_rank, int nprocs);
void data_tile_copy_back_w_3(int ts1,int ts2,int ts3, double** w_trans, BufferManager *buff_mang_w);
int writeout_pack_w_3(int ts1,int ts2,int ts3,double *lw_buf_w,int lw_count_w, double** w_trans, BufferManager *buff_mang_w);
int writeout_unpack_w_3(int ts1,int ts2,int ts3,double *lw_recv_buf_w,int displs_lw_w, double** w_trans, BufferManager *buff_mang_w);
void pack_send_3(int my_rank,int nprocs,int t1,int t2,int t3, double** w_trans, BufferManager *buff_mang_w,double **send_buf_w,MPI_Request *send_reqs_w,MPI_Status *send_stats_w, double** B_trans,int B_size_1, BufferManager *buff_mang_B,double **send_buf_B,MPI_Request *send_reqs_B,MPI_Status *send_stats_B, double** x_trans, BufferManager *buff_mang_x,double **send_buf_x,MPI_Request *send_reqs_x,MPI_Status *send_stats_x,int max_num_sends,double *p_total_count);
void unpack_buf_w_3(int my_rank,int nprocs, double** w_trans, BufferManager *buff_mang_w,double *recv_buf_w);
void unpack_buf_B_3(int my_rank,int nprocs, double** B_trans,int B_size_1, BufferManager *buff_mang_B,double *recv_buf_B);
void unpack_buf_x_3(int my_rank,int nprocs, double** x_trans, BufferManager *buff_mang_x,double *recv_buf_x);
void copy_back_writeout_0(int my_rank, int nprocs, double** B_trans,int B_size_1, BufferManager *buff_mang_B, double** x_trans, BufferManager *buff_mang_x, double** w_trans, BufferManager *buff_mang_w);
int pi_0( int t1, int t5, int t7, int N, int n, int nprocs);
int pi_threads_0( int t1, int t5, int t7, int N, int n, int my_rank, int nprocs, int num_threads);
int pi_1( int t1, int t5, int t7, int N, int n, int nprocs);
int pi_threads_1( int t1, int t5, int t7, int N, int n, int my_rank, int nprocs, int num_threads);
int pi_2( int t1, int t5, int N, int n, int nprocs);
int pi_threads_2( int t1, int t5, int N, int n, int my_rank, int nprocs, int num_threads);
int pi_3( int t1, int t5, int t7, int N, int n, int nprocs);
int pi_threads_3( int t1, int t5, int t7, int N, int n, int my_rank, int nprocs, int num_threads);
void write_out(int my_rank,int nprocs,double *lw_buf_B,double *lw_recv_buf_B,int *displs_lw_B,double *lw_buf_x,double *lw_recv_buf_x,int *displs_lw_x,double *lw_buf_w,double *lw_recv_buf_w,int *displs_lw_w, double** B_trans,int B_size_1, BufferManager *buff_mang_B, double** x_trans, BufferManager *buff_mang_x, double** w_trans, BufferManager *buff_mang_w);
void data_tile_alloc_0(int ts1,int ts2,int ts3, double** B_trans,int B_size_1, BufferManager *buff_mang_B, double** A_trans,int A_size_1, BufferManager *buff_mang_A, double** u1_trans, BufferManager *buff_mang_u1, double** v1_trans, BufferManager *buff_mang_v1, double** u2_trans, BufferManager *buff_mang_u2, double** v2_trans, BufferManager *buff_mang_v2);
void data_tile_ref_count_update_0(int ts1,int ts2,int ts3, double** B_trans,int B_size_1, BufferManager *buff_mang_B, double** A_trans,int A_size_1, BufferManager *buff_mang_A, double** u1_trans, BufferManager *buff_mang_u1, double** v1_trans, BufferManager *buff_mang_v1, double** u2_trans, BufferManager *buff_mang_u2, double** v2_trans, BufferManager *buff_mang_v2);
void compute_task_0(int t1,int t2,int t3, double** B_trans,int B_size_1, BufferManager *buff_mang_B, double** A_trans,int A_size_1, BufferManager *buff_mang_A, double** u1_trans, BufferManager *buff_mang_u1, double** v1_trans, BufferManager *buff_mang_v1, double** u2_trans, BufferManager *buff_mang_u2, double** v2_trans, BufferManager *buff_mang_v2);
void data_tile_alloc_1(int ts1,int ts2,int ts3, double** x_trans, BufferManager *buff_mang_x, double** B_trans,int B_size_1, BufferManager *buff_mang_B, double** y_trans, BufferManager *buff_mang_y);
void data_tile_ref_count_update_1(int ts1,int ts2,int ts3, double** x_trans, BufferManager *buff_mang_x, double** B_trans,int B_size_1, BufferManager *buff_mang_B, double** y_trans, BufferManager *buff_mang_y);
void compute_task_1(int t1,int t2,int t3, double** x_trans, BufferManager *buff_mang_x, double** B_trans,int B_size_1, BufferManager *buff_mang_B, double** y_trans, BufferManager *buff_mang_y);
void data_tile_alloc_2(int ts1,int ts2, double** x_trans, BufferManager *buff_mang_x, double** z_trans, BufferManager *buff_mang_z);
void data_tile_ref_count_update_2(int ts1,int ts2, double** x_trans, BufferManager *buff_mang_x, double** z_trans, BufferManager *buff_mang_z);
void compute_task_2(int t1,int t2, double** x_trans, BufferManager *buff_mang_x, double** z_trans, BufferManager *buff_mang_z);
void data_tile_alloc_3(int ts1,int ts2,int ts3, double** w_trans, BufferManager *buff_mang_w, double** B_trans,int B_size_1, BufferManager *buff_mang_B, double** x_trans, BufferManager *buff_mang_x);
void data_tile_ref_count_update_3(int ts1,int ts2,int ts3, double** w_trans, BufferManager *buff_mang_w, double** B_trans,int B_size_1, BufferManager *buff_mang_B, double** x_trans, BufferManager *buff_mang_x);
void compute_task_3(int t1,int t2,int t3, double** w_trans, BufferManager *buff_mang_w, double** B_trans,int B_size_1, BufferManager *buff_mang_B, double** x_trans, BufferManager *buff_mang_x);
