#include "polyrt.h"
#include <mpi.h>
#include "tbb/concurrent_priority_queue.h"

#include <omp.h>

#ifdef __DEBUG_FLUSH
#define IF_DEBUG_FLUSH(foo) foo
#else
#define IF_DEBUG_FLUSH(foo) 
#endif

#ifdef __DYNSCHEDULER_MORE_DEBUG_PRINT
#define IF_DYNSCHEDULER_MORE_DEBUG_PRINT(foo) foo
#else
#define IF_DYNSCHEDULER_MORE_DEBUG_PRINT(foo) 
#endif

#ifdef __DYNSCHEDULER_DEBUG_PRINT
#define IF_DYNSCHEDULER_DEBUG_PRINT(foo) foo
#else
#define IF_DYNSCHEDULER_DEBUG_PRINT(foo) 
#endif

extern FILE *__debug_print_fp;

class __Task {
	public:
		int loop;
		int id1;
		int id2;
		int remote_dep_tasks;
		int local_dep_tasks;
		int thread_affinity;
		__Task() {}
		__Task(const int l, const int _id1, const int _id2, const int remote, const int local, const int affinity) :
			loop(l),
			id1(_id1),
			id2(_id2),
			remote_dep_tasks(remote),
			local_dep_tasks(local),
			thread_affinity(affinity) {}
		__Task(const __Task &task) :
			loop(task.loop),
			id1(task.id1),
			id2(task.id2),
			remote_dep_tasks(task.remote_dep_tasks),
			local_dep_tasks(task.local_dep_tasks),
			thread_affinity(task.thread_affinity) {}
		~__Task() {}
};

class __Compare {
	public:
		// returns true only if task2 has higher priority than task1
		bool operator()(const __Task &task1, const __Task &task2) const {
#ifndef __DYNSCHEDULER_NO_PRIORITY
			if (task2.remote_dep_tasks>task1.remote_dep_tasks) return true;
			if (task1.remote_dep_tasks>task2.remote_dep_tasks) return false;
			if (task2.local_dep_tasks>task1.local_dep_tasks) return true;
			if (task1.local_dep_tasks>task2.local_dep_tasks) return false;
			int thread_id = omp_get_thread_num();
			if ((thread_id == task2.thread_affinity) && (thread_id != task1.thread_affinity)) return true;
			if ((thread_id == task1.thread_affinity) && (thread_id != task2.thread_affinity)) return false;
			// lexicographic compare 
			if ( (task2.id1 > task1.id1) 
				|| ( (task2.id1 == task1.id1) && (task2.id2 > task1.id2) ) )
				 return true;
#endif
			return false;
		}
};

extern tbb::concurrent_priority_queue<__Task, __Compare> pqueue;

int is_receiver_0(int ts1, int ts2, int _PB_N, int my_rank, int nprocs);
int get_num_remote_dep_tasks_0(int ts1, int ts2, int _PB_N, int my_rank, int nprocs);
void remote_update_dep_tasks_0(int ts1, int ts2, int _PB_N, int my_rank, int nprocs, int num_threads,int *tasks_loop0,int *tasks_loop1);
int get_num_local_dep_tasks_0(int ts1, int ts2, int _PB_N, int my_rank, int nprocs);
void local_update_dep_tasks_0(int ts1, int ts2, int _PB_N, int my_rank, int nprocs, int num_threads,int *tasks_loop0,int *tasks_loop1);
int get_num_remote_src_tasks_0(int ts1, int ts2, int _PB_N, int my_rank, int nprocs);
int get_num_local_src_tasks_0(int ts1, int ts2, int _PB_N, int my_rank, int nprocs);
extern int max_num_tasks_loop0_dim0;
extern int lb_tasks_loop0_dim0;
extern int *tasks_loop0;
int is_receiver_1(int ts1, int ts2, int _PB_N, int my_rank, int nprocs);
int get_num_remote_dep_tasks_1(int ts1, int ts2, int _PB_N, int my_rank, int nprocs);
void remote_update_dep_tasks_1(int ts1, int ts2, int _PB_N, int my_rank, int nprocs, int num_threads,int *tasks_loop0,int *tasks_loop1);
int get_num_local_dep_tasks_1(int ts1, int ts2, int _PB_N, int my_rank, int nprocs);
void local_update_dep_tasks_1(int ts1, int ts2, int _PB_N, int my_rank, int nprocs, int num_threads,int *tasks_loop0,int *tasks_loop1);
int get_num_remote_src_tasks_1(int ts1, int ts2, int _PB_N, int my_rank, int nprocs);
int get_num_local_src_tasks_1(int ts1, int ts2, int _PB_N, int my_rank, int nprocs);
extern int max_num_tasks_loop1_dim0;
extern int lb_tasks_loop1_dim0;
extern int *tasks_loop1;
void init_tasks(int my_rank,int nprocs,int num_threads,int *tasks_loop0,int *tasks_loop1,long int *p_num_tasks_to_execute,long int *p_num_tasks_to_unpack);
int is_receiver_tmp_0(int ts1, int ts2, int _PB_N, int my_rank, int nprocs);
void sigma_tmp_1_0(int ts1, int ts2, int _PB_N, int my_rank, int nprocs, int *receiver_list);
int is_receiver_tmp_1_0(int ts1, int ts2, int _PB_N, int my_rank, int nprocs);
int is_receiver_y_0(int ts1, int ts2, int _PB_N, int my_rank, int nprocs);
void sigma_y_1_0(int ts1, int ts2, int _PB_N, int my_rank, int nprocs, int *receiver_list);
int is_receiver_y_1_0(int ts1, int ts2, int _PB_N, int my_rank, int nprocs);
int is_receiver_y_1(int ts1, int ts2, int _PB_N, int my_rank, int nprocs);
void sigma_y_1_1(int ts1, int ts2, int _PB_N, int my_rank, int nprocs, int *receiver_list);
int is_receiver_y_1_1(int ts1, int ts2, int _PB_N, int my_rank, int nprocs);
int pi_0( int t1, int t4, int _PB_N, int nprocs);
int pi_threads_0( int t1, int t4, int _PB_N, int my_rank, int nprocs, int num_threads);
int pi_1( int t1, int t4, int _PB_N, int nprocs);
int pi_threads_1( int t1, int t4, int _PB_N, int my_rank, int nprocs, int num_threads);
