#include "gesummv.dist_dynsched_data_dist.h"
#include "polyrt.h"
#include "buffer_manager.h"
#include "pi_defs.h"
#include <assert.h>
#include <omp.h>
#include <mpi.h>
#include <limits.h>
#include <math.h>
#define ceild(n,d)  ceil(((double)(n))/((double)(d)))
#define floord(n,d) floor(((double)(n))/((double)(d)))
#define max(x,y)    ((x) > (y)? (x) : (y))
#define min(x,y)    ((x) < (y)? (x) : (y))

/**
 * gesummv.c: This file is part of the PolyBench/C 3.2 test suite.
 *
 *
 * Contact: Louis-Noel Pouchet <pouchet@cse.ohio-state.edu>
 * Web address: http://polybench.sourceforge.net
 */
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <math.h>

/* Include polybench common header. */
#include <polybench.h>

/* Include benchmark-specific header. */
/* Default data type is double, default size is 4000. */
#define N 30000

#include "gesummv.h"

  DATA_TYPE aplha_s = 43532;
  DATA_TYPE beta = 12313;
  int n = N;

/* Array initialization. */
static
void init_array(int n,
		DATA_TYPE POLYBENCH_2D(A,N,N,n,n),
		DATA_TYPE POLYBENCH_2D(B,N,N,n,n),
		DATA_TYPE POLYBENCH_1D(x,N,n),
        DATA_TYPE POLYBENCH_1D(tmp,N,n),
        DATA_TYPE POLYBENCH_1D(y,N,n)
        )
{
  int i, j;

  aplha_s = 43532;
  beta = 12313;
  for (i = 0; i < n; i++)
    {
      x[i] = ((DATA_TYPE) i) / n;
      tmp[i] = 0;
      y[i] = 0;
      for (j = 0; j < n; j++) {
	A[i][j] = ((DATA_TYPE) i*j) / n;
	B[i][j] = ((DATA_TYPE) i*j) / n;
      }
    }
}


/* DCE code. Must scan the entire live-out data.
   Can be used also to check the correctness of the output. */
static
void print_array(int n,
		 DATA_TYPE POLYBENCH_1D(y,N,n))

{
  int i;

  for (i = 0; i < n; i++) {
    fprintf (stderr, DATA_PRINTF_MODIFIER, y[i]);
    if (i % 20 == 0) fprintf (stderr, "\n");
  }
}

#define __DECLARATION_OF_aplha_s DATA_TYPE aplha_s
#define __DECLARATION_OF_beta DATA_TYPE beta
#define __DECLARATION_OF_A DATA_TYPE POLYBENCH_2D(A,N,N,n,n)
#define __DECLARATION_OF_B DATA_TYPE POLYBENCH_2D(B,N,N,n,n)
#define __DECLARATION_OF_tmp DATA_TYPE POLYBENCH_1D(tmp,N,n)
#define __DECLARATION_OF_x DATA_TYPE POLYBENCH_1D(x,N,n)
#define __DECLARATION_OF_y DATA_TYPE POLYBENCH_1D(y,N,n)

#define __PLACE_TO_INSERT_FORWARD_DECLARATIONS
FILE *__debug_print_fp;
tbb::concurrent_priority_queue<__Task, __Compare> pqueue;
int max_num_tasks_loop0_dim0;
int lb_tasks_loop0_dim0;
int *tasks_loop0;
int max_num_tasks_loop1_dim0;
int lb_tasks_loop1_dim0;
int *tasks_loop1;
void read_in_copy_A_0(int ts1,int ts2 
#ifndef USE_LOCAL_ARRAYS
 ,__DECLARATION_OF_A 
#endif
 , double** A_trans,int A_size_1, BufferManager *buff_mang_A);
void read_in_copy_x_0(int ts1,int ts2 
#ifndef USE_LOCAL_ARRAYS
 ,__DECLARATION_OF_x 
#endif
 , double** x_trans, BufferManager *buff_mang_x);
void read_in_copy_tmp_0(int ts1,int ts2 
#ifndef USE_LOCAL_ARRAYS
 ,__DECLARATION_OF_tmp 
#endif
 , double** tmp_trans, BufferManager *buff_mang_tmp);
void read_in_copy_B_0(int ts1,int ts2 
#ifndef USE_LOCAL_ARRAYS
 ,__DECLARATION_OF_B 
#endif
 , double** B_trans,int B_size_1, BufferManager *buff_mang_B);
void read_in_copy_y_0(int ts1,int ts2 
#ifndef USE_LOCAL_ARRAYS
 ,__DECLARATION_OF_y 
#endif
 , double** y_trans, BufferManager *buff_mang_y);
void read_in_copy_tmp_1(int ts1,int ts2 
#ifndef USE_LOCAL_ARRAYS
 ,__DECLARATION_OF_tmp 
#endif
 , double** tmp_trans, BufferManager *buff_mang_tmp);
void read_in_copy_y_1(int ts1,int ts2 
#ifndef USE_LOCAL_ARRAYS
 ,__DECLARATION_OF_y 
#endif
 , double** y_trans, BufferManager *buff_mang_y);
void data_tile_init_0(int my_rank, int nprocs 
#ifndef USE_LOCAL_ARRAYS
 ,__DECLARATION_OF_tmp,__DECLARATION_OF_A,__DECLARATION_OF_x,__DECLARATION_OF_y,__DECLARATION_OF_B 
#endif
 , double** tmp_trans, BufferManager *buff_mang_tmp, double** A_trans,int A_size_1, BufferManager *buff_mang_A, double** x_trans, BufferManager *buff_mang_x, double** y_trans, BufferManager *buff_mang_y, double** B_trans,int B_size_1, BufferManager *buff_mang_B);
void data_tile_copy_back_tmp_0(int ts1,int ts2 
#ifndef USE_LOCAL_ARRAYS
 ,__DECLARATION_OF_tmp 
#endif
 , double** tmp_trans, BufferManager *buff_mang_tmp);
void data_tile_copy_back_y_0(int ts1,int ts2 
#ifndef USE_LOCAL_ARRAYS
 ,__DECLARATION_OF_y 
#endif
 , double** y_trans, BufferManager *buff_mang_y);
int writeout_pack_tmp_0(int ts1,int ts2,double *lw_buf_tmp,int lw_count_tmp, __DECLARATION_OF_tmp, double** tmp_trans, BufferManager *buff_mang_tmp);
int writeout_unpack_tmp_0(int ts1,int ts2,double *lw_recv_buf_tmp,int displs_lw_tmp, __DECLARATION_OF_tmp, double** tmp_trans, BufferManager *buff_mang_tmp);
int writeout_pack_y_0(int ts1,int ts2,double *lw_buf_y,int lw_count_y, __DECLARATION_OF_y, double** y_trans, BufferManager *buff_mang_y);
int writeout_unpack_y_0(int ts1,int ts2,double *lw_recv_buf_y,int displs_lw_y, __DECLARATION_OF_y, double** y_trans, BufferManager *buff_mang_y);
void data_tile_copy_back_y_1(int ts1,int ts2 
#ifndef USE_LOCAL_ARRAYS
 ,__DECLARATION_OF_y 
#endif
 , double** y_trans, BufferManager *buff_mang_y);
int writeout_pack_y_1(int ts1,int ts2,double *lw_buf_y,int lw_count_y, __DECLARATION_OF_y, double** y_trans, BufferManager *buff_mang_y);
int writeout_unpack_y_1(int ts1,int ts2,double *lw_recv_buf_y,int displs_lw_y, __DECLARATION_OF_y, double** y_trans, BufferManager *buff_mang_y);
void copy_back_writeout_0(int my_rank, int nprocs 
#ifndef USE_LOCAL_ARRAYS
 ,__DECLARATION_OF_tmp,__DECLARATION_OF_y 
#endif
 , double** tmp_trans, BufferManager *buff_mang_tmp, double** y_trans, BufferManager *buff_mang_y);
void write_out(int my_rank,int nprocs,double *lw_buf_tmp,double *lw_recv_buf_tmp,int *displs_lw_tmp,double *lw_buf_y,double *lw_recv_buf_y,int *displs_lw_y, __DECLARATION_OF_tmp, __DECLARATION_OF_y, double** tmp_trans, BufferManager *buff_mang_tmp, double** y_trans, BufferManager *buff_mang_y);

/* Main computational kernel. The whole function will be timed,
   including the call and return. */
static
void kernel_gesummv(int n
#ifndef USE_LOCAL_ARRAYS
		    ,DATA_TYPE POLYBENCH_2D(A,N,N,n,n),
		    DATA_TYPE POLYBENCH_2D(B,N,N,n,n),
		    DATA_TYPE POLYBENCH_1D(tmp,N,n),
		    DATA_TYPE POLYBENCH_1D(x,N,n),
		    DATA_TYPE POLYBENCH_1D(y,N,n)
#endif
    )
{
  int i, j;

int i_2_div = 0;int i_2_mod = 0;double *y_i_2 = NULL;double *tmp_i_2 = NULL; int t1, t2, t3, t4, t5, t6;
 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;
int _num_threads;
#pragma omp parallel
{_num_threads = omp_get_num_threads();}
 double t_data_init_start, t_data_init= 0.0, t_globaldata_init= 0.0;
#ifndef GLOBAL_MY_RANK
 int my_rank;
#endif
#ifndef __MAX_NUM_RECVS
 #define __MAX_NUM_RECVS (nprocs*_num_threads*4)
#endif
#ifndef __MAX_NUM_SENDS
 #define __MAX_NUM_SENDS (nprocs*_num_threads*4)
#endif
 int nprocs, my_start, my_end, _i, __p, proc, recv_proc, distinct_recv, __tid;
 long int _num_tasks_to_execute, _num_tasks_to_unpack;
 int count;
int req_count;
 int _lb_dist, _ub_dist, lbd_t1, ubd_t1, lbd_t2, ubd_t2, lbd_t3, ubd_t3, lbd_t4, ubd_t4, lbd_t5, ubd_t5, lbd_t6, ubd_t6;
 int thread_support_provided;
 MPI_Init_thread(NULL, NULL, MPI_THREAD_MULTIPLE, &thread_support_provided);
 assert(thread_support_provided == MPI_THREAD_MULTIPLE);
 MPI_Comm_rank(MPI_COMM_WORLD, &my_rank);
 MPI_Comm_size(MPI_COMM_WORLD, &nprocs);
 int __tid_receiver;
 int waiting_pops[_num_threads], waiting_pops_threads, waiting_pops_global;
 for (__tid=0; __tid<_num_threads; __tid++) waiting_pops[__tid] = 0;
 double t_tasks_create, t_globaltasks_create;
 double t_tasks_manage_start[_num_threads], t_tasks_manage[_num_threads], t_tasks_manage_threads, t_globaltasks_manage;
 for (__tid=0; __tid<_num_threads; __tid++) t_tasks_manage[__tid] = 0;
 double t_comp_start[_num_threads], t_comp[_num_threads], t_comp_threads_min, t_globalcomp_min, t_comp_threads, t_globalcomp;
 for (__tid=0; __tid<_num_threads; __tid++) t_comp[__tid] = 0;
 double t_wait_comp_start[_num_threads], t_wait_comp[_num_threads], t_wait_comp_threads, t_globalwait_comp;
 for (__tid=0; __tid<_num_threads; __tid++) t_wait_comp[__tid] = 0;
 double t_data_mang_start[_num_threads], t_data_mang[_num_threads], t_data_mang_threads, t_globaldata_mang;
 for (__tid=0; __tid<_num_threads; __tid++) t_data_mang[__tid] = 0;
 double t_pack_start[_num_threads], t_pack[_num_threads], t_pack_threads, t_globalpack;
 for (__tid=0; __tid<_num_threads; __tid++) t_pack[__tid] = 0;
 double t_writeout = 0.0;
 double t_unpack_start, t_unpack = 0.0, t_globalunpack = 0.0;
 double t_wait_unpack_start, t_wait_unpack = 0.0, t_globalwait_unpack = 0.0;
 double t_local = 0.0, t_global = 0.0;
 double __total_count[_num_threads];
 for (__tid=0; __tid<_num_threads; __tid++) __total_count[__tid] = 0;
 double __total_count_threads = 0;
 double __total_count_all = 0;
MPI_Request *send_reqs_tmp[_num_threads];
                    for (__tid=0; __tid<_num_threads; __tid++) {
                    send_reqs_tmp[__tid] = (MPI_Request *)malloc(__MAX_NUM_SENDS*sizeof(MPI_Request));
                    }
std::vector<MPI_Request> recv_reqs_tmp;
MPI_Status *send_stats_tmp[_num_threads];
                    for (__tid=0; __tid<_num_threads; __tid++) {
                    send_stats_tmp[__tid] = (MPI_Status *)malloc(nprocs*sizeof(MPI_Status));
                    }
MPI_Status *recv_stats_tmp = (MPI_Status *)malloc(sizeof(MPI_Status));
std::vector<int> recv_completed_tmp;
                    for (__tid=0; __tid<_num_threads; __tid++) {
                    for (__p=0; __p<__MAX_NUM_SENDS; __p++) {
                    send_reqs_tmp[__tid][__p] = MPI_REQUEST_NULL;
}
}
                    for (__p=0; __p<__MAX_NUM_RECVS; __p++) {
                    recv_reqs_tmp.push_back(MPI_REQUEST_NULL);
                    recv_completed_tmp.push_back(0);
                    }
int lw_count_tmp = 0;
int send_counts_tmp[__tid][nprocs];
int max_num_elements_tmp = 0;
int recv_counts_tmp[nprocs];
int lw_recv_counts_tmp[nprocs];
int displs_tmp[nprocs];
int displs_lw_tmp[nprocs];
double *send_buf_tmp[_num_threads][__MAX_NUM_SENDS];
size_t send_buf_size_tmp[_num_threads][__MAX_NUM_SENDS];
for (__tid=0; __tid<_num_threads; __tid++) {
      for (__p=0; __p<__MAX_NUM_SENDS; __p++) {
      send_buf_tmp[__tid][__p] = NULL;
      send_buf_size_tmp[__tid][__p] = 0;
}
}std::vector<double *> recv_buf_tmp;
size_t recv_buf_size_tmp[__MAX_NUM_RECVS];
for (__p=0; __p<__MAX_NUM_RECVS; __p++) {
                    recv_buf_tmp.push_back(NULL);
                    recv_buf_size_tmp[__p] = 0;
}double *lw_buf_tmp = NULL;
size_t lw_buf_size_tmp = 0;
double *lw_recv_buf_tmp = NULL;
size_t lw_recv_buf_size_tmp = 0;
MPI_Request *send_reqs_A[_num_threads];
                    for (__tid=0; __tid<_num_threads; __tid++) {
                    send_reqs_A[__tid] = (MPI_Request *)malloc(__MAX_NUM_SENDS*sizeof(MPI_Request));
                    }
std::vector<MPI_Request> recv_reqs_A;
MPI_Status *send_stats_A[_num_threads];
                    for (__tid=0; __tid<_num_threads; __tid++) {
                    send_stats_A[__tid] = (MPI_Status *)malloc(nprocs*sizeof(MPI_Status));
                    }
MPI_Status *recv_stats_A = (MPI_Status *)malloc(sizeof(MPI_Status));
std::vector<int> recv_completed_A;
                    for (__tid=0; __tid<_num_threads; __tid++) {
                    for (__p=0; __p<__MAX_NUM_SENDS; __p++) {
                    send_reqs_A[__tid][__p] = MPI_REQUEST_NULL;
}
}
                    for (__p=0; __p<__MAX_NUM_RECVS; __p++) {
                    recv_reqs_A.push_back(MPI_REQUEST_NULL);
                    recv_completed_A.push_back(0);
                    }
int lw_count_A = 0;
int send_counts_A[__tid][nprocs];
int max_num_elements_A = 0;
int recv_counts_A[nprocs];
int lw_recv_counts_A[nprocs];
int displs_A[nprocs];
int displs_lw_A[nprocs];
double *send_buf_A[_num_threads][__MAX_NUM_SENDS];
size_t send_buf_size_A[_num_threads][__MAX_NUM_SENDS];
for (__tid=0; __tid<_num_threads; __tid++) {
      for (__p=0; __p<__MAX_NUM_SENDS; __p++) {
      send_buf_A[__tid][__p] = NULL;
      send_buf_size_A[__tid][__p] = 0;
}
}std::vector<double *> recv_buf_A;
size_t recv_buf_size_A[__MAX_NUM_RECVS];
for (__p=0; __p<__MAX_NUM_RECVS; __p++) {
                    recv_buf_A.push_back(NULL);
                    recv_buf_size_A[__p] = 0;
}double *lw_buf_A = NULL;
size_t lw_buf_size_A = 0;
double *lw_recv_buf_A = NULL;
size_t lw_recv_buf_size_A = 0;
MPI_Request *send_reqs_x[_num_threads];
                    for (__tid=0; __tid<_num_threads; __tid++) {
                    send_reqs_x[__tid] = (MPI_Request *)malloc(__MAX_NUM_SENDS*sizeof(MPI_Request));
                    }
std::vector<MPI_Request> recv_reqs_x;
MPI_Status *send_stats_x[_num_threads];
                    for (__tid=0; __tid<_num_threads; __tid++) {
                    send_stats_x[__tid] = (MPI_Status *)malloc(nprocs*sizeof(MPI_Status));
                    }
MPI_Status *recv_stats_x = (MPI_Status *)malloc(sizeof(MPI_Status));
std::vector<int> recv_completed_x;
                    for (__tid=0; __tid<_num_threads; __tid++) {
                    for (__p=0; __p<__MAX_NUM_SENDS; __p++) {
                    send_reqs_x[__tid][__p] = MPI_REQUEST_NULL;
}
}
                    for (__p=0; __p<__MAX_NUM_RECVS; __p++) {
                    recv_reqs_x.push_back(MPI_REQUEST_NULL);
                    recv_completed_x.push_back(0);
                    }
int lw_count_x = 0;
int send_counts_x[__tid][nprocs];
int max_num_elements_x = 0;
int recv_counts_x[nprocs];
int lw_recv_counts_x[nprocs];
int displs_x[nprocs];
int displs_lw_x[nprocs];
double *send_buf_x[_num_threads][__MAX_NUM_SENDS];
size_t send_buf_size_x[_num_threads][__MAX_NUM_SENDS];
for (__tid=0; __tid<_num_threads; __tid++) {
      for (__p=0; __p<__MAX_NUM_SENDS; __p++) {
      send_buf_x[__tid][__p] = NULL;
      send_buf_size_x[__tid][__p] = 0;
}
}std::vector<double *> recv_buf_x;
size_t recv_buf_size_x[__MAX_NUM_RECVS];
for (__p=0; __p<__MAX_NUM_RECVS; __p++) {
                    recv_buf_x.push_back(NULL);
                    recv_buf_size_x[__p] = 0;
}double *lw_buf_x = NULL;
size_t lw_buf_size_x = 0;
double *lw_recv_buf_x = NULL;
size_t lw_recv_buf_size_x = 0;
MPI_Request *send_reqs_y[_num_threads];
                    for (__tid=0; __tid<_num_threads; __tid++) {
                    send_reqs_y[__tid] = (MPI_Request *)malloc(__MAX_NUM_SENDS*sizeof(MPI_Request));
                    }
std::vector<MPI_Request> recv_reqs_y;
MPI_Status *send_stats_y[_num_threads];
                    for (__tid=0; __tid<_num_threads; __tid++) {
                    send_stats_y[__tid] = (MPI_Status *)malloc(nprocs*sizeof(MPI_Status));
                    }
MPI_Status *recv_stats_y = (MPI_Status *)malloc(sizeof(MPI_Status));
std::vector<int> recv_completed_y;
                    for (__tid=0; __tid<_num_threads; __tid++) {
                    for (__p=0; __p<__MAX_NUM_SENDS; __p++) {
                    send_reqs_y[__tid][__p] = MPI_REQUEST_NULL;
}
}
                    for (__p=0; __p<__MAX_NUM_RECVS; __p++) {
                    recv_reqs_y.push_back(MPI_REQUEST_NULL);
                    recv_completed_y.push_back(0);
                    }
int lw_count_y = 0;
int send_counts_y[__tid][nprocs];
int max_num_elements_y = 0;
int recv_counts_y[nprocs];
int lw_recv_counts_y[nprocs];
int displs_y[nprocs];
int displs_lw_y[nprocs];
double *send_buf_y[_num_threads][__MAX_NUM_SENDS];
size_t send_buf_size_y[_num_threads][__MAX_NUM_SENDS];
for (__tid=0; __tid<_num_threads; __tid++) {
      for (__p=0; __p<__MAX_NUM_SENDS; __p++) {
      send_buf_y[__tid][__p] = NULL;
      send_buf_size_y[__tid][__p] = 0;
}
}std::vector<double *> recv_buf_y;
size_t recv_buf_size_y[__MAX_NUM_RECVS];
for (__p=0; __p<__MAX_NUM_RECVS; __p++) {
                    recv_buf_y.push_back(NULL);
                    recv_buf_size_y[__p] = 0;
}double *lw_buf_y = NULL;
size_t lw_buf_size_y = 0;
double *lw_recv_buf_y = NULL;
size_t lw_recv_buf_size_y = 0;
MPI_Request *send_reqs_B[_num_threads];
                    for (__tid=0; __tid<_num_threads; __tid++) {
                    send_reqs_B[__tid] = (MPI_Request *)malloc(__MAX_NUM_SENDS*sizeof(MPI_Request));
                    }
std::vector<MPI_Request> recv_reqs_B;
MPI_Status *send_stats_B[_num_threads];
                    for (__tid=0; __tid<_num_threads; __tid++) {
                    send_stats_B[__tid] = (MPI_Status *)malloc(nprocs*sizeof(MPI_Status));
                    }
MPI_Status *recv_stats_B = (MPI_Status *)malloc(sizeof(MPI_Status));
std::vector<int> recv_completed_B;
                    for (__tid=0; __tid<_num_threads; __tid++) {
                    for (__p=0; __p<__MAX_NUM_SENDS; __p++) {
                    send_reqs_B[__tid][__p] = MPI_REQUEST_NULL;
}
}
                    for (__p=0; __p<__MAX_NUM_RECVS; __p++) {
                    recv_reqs_B.push_back(MPI_REQUEST_NULL);
                    recv_completed_B.push_back(0);
                    }
int lw_count_B = 0;
int send_counts_B[__tid][nprocs];
int max_num_elements_B = 0;
int recv_counts_B[nprocs];
int lw_recv_counts_B[nprocs];
int displs_B[nprocs];
int displs_lw_B[nprocs];
double *send_buf_B[_num_threads][__MAX_NUM_SENDS];
size_t send_buf_size_B[_num_threads][__MAX_NUM_SENDS];
for (__tid=0; __tid<_num_threads; __tid++) {
      for (__p=0; __p<__MAX_NUM_SENDS; __p++) {
      send_buf_B[__tid][__p] = NULL;
      send_buf_size_B[__tid][__p] = 0;
}
}std::vector<double *> recv_buf_B;
size_t recv_buf_size_B[__MAX_NUM_RECVS];
for (__p=0; __p<__MAX_NUM_RECVS; __p++) {
                    recv_buf_B.push_back(NULL);
                    recv_buf_size_B[__p] = 0;
}double *lw_buf_B = NULL;
size_t lw_buf_size_B = 0;
double *lw_recv_buf_B = NULL;
size_t lw_recv_buf_size_B = 0;
max_num_tasks_loop0_dim0 = ceil(floorf((+1*n-1)/(float)128) - 0 + 1);
lb_tasks_loop0_dim0 = floor(0);
tasks_loop0 = (int *)malloc(max_num_tasks_loop0_dim0*sizeof(int));
max_num_tasks_loop1_dim0 = ceil(floorf((+1*n-1)/(float)128) - 0 + 1);
lb_tasks_loop1_dim0 = floor(0);
tasks_loop1 = (int *)malloc(max_num_tasks_loop1_dim0*sizeof(int));
int B_size_0 = (+1*(floorf((+1*n-1)/(float)128) - 0 + 1));
int B_size_1 = (+1*(floorf((+1*n-1)/(float)128) - 0 + 1));
double **B_trans = (double **)malloc((B_size_0 * B_size_1) * sizeof(double));
BufferManager *buff_mang_B = init_buffer_manager(B_size_0 * B_size_1);
int y_size_0 = (+1*(floorf((+1*n-1)/(float)128) - 0 + 1));
double **y_trans = (double **)malloc((y_size_0) * sizeof(double));
BufferManager *buff_mang_y = init_buffer_manager(y_size_0);
int x_size_0 = (+1*(floorf((+1*n-1)/(float)128) - 0 + 1));
double **x_trans = (double **)malloc((x_size_0) * sizeof(double));
BufferManager *buff_mang_x = init_buffer_manager(x_size_0);
int A_size_0 = (+1*(floorf((+1*n-1)/(float)128) - 0 + 1));
int A_size_1 = (+1*(floorf((+1*n-1)/(float)128) - 0 + 1));
double **A_trans = (double **)malloc((A_size_0 * A_size_1) * sizeof(double));
BufferManager *buff_mang_A = init_buffer_manager(A_size_0 * A_size_1);
int tmp_size_0 = (+1*(floorf((+1*n-1)/(float)128) - 0 + 1));
double **tmp_trans = (double **)malloc((tmp_size_0) * sizeof(double));
BufferManager *buff_mang_tmp = init_buffer_manager(tmp_size_0);
            max_num_elements_A = max(max_num_elements_A, (+1*(1)*(1)) + 3);
                for (__tid=0; __tid<_num_threads; __tid++) { for (__p=0; __p<__MAX_NUM_SENDS; __p++) {
                send_buf_A[__tid][__p] = (double *) polyrt_max_alloc(send_buf_A[__tid][__p], sizeof(double)*max_num_elements_A, &send_buf_size_A[__tid][__p]);
}
}
for (__p=0; __p<__MAX_NUM_RECVS; __p++) { recv_buf_A[__p] = (double *) polyrt_max_alloc(recv_buf_A[__p], sizeof(double)*max_num_elements_A, &recv_buf_size_A[__p]); }
            max_num_elements_x = max(max_num_elements_x, (+1*(1)) + 3);
                for (__tid=0; __tid<_num_threads; __tid++) { for (__p=0; __p<__MAX_NUM_SENDS; __p++) {
                send_buf_x[__tid][__p] = (double *) polyrt_max_alloc(send_buf_x[__tid][__p], sizeof(double)*max_num_elements_x, &send_buf_size_x[__tid][__p]);
}
}
for (__p=0; __p<__MAX_NUM_RECVS; __p++) { recv_buf_x[__p] = (double *) polyrt_max_alloc(recv_buf_x[__p], sizeof(double)*max_num_elements_x, &recv_buf_size_x[__p]); }
            max_num_elements_tmp = max(max_num_elements_tmp, (+1*(128)) + 3);
                for (__tid=0; __tid<_num_threads; __tid++) { for (__p=0; __p<__MAX_NUM_SENDS; __p++) {
                send_buf_tmp[__tid][__p] = (double *) polyrt_max_alloc(send_buf_tmp[__tid][__p], sizeof(double)*max_num_elements_tmp, &send_buf_size_tmp[__tid][__p]);
}
}
for (__p=0; __p<__MAX_NUM_RECVS; __p++) { recv_buf_tmp[__p] = (double *) polyrt_max_alloc(recv_buf_tmp[__p], sizeof(double)*max_num_elements_tmp, &recv_buf_size_tmp[__p]); }
            max_num_elements_B = max(max_num_elements_B, (+1*(1)*(1)) + 3);
                for (__tid=0; __tid<_num_threads; __tid++) { for (__p=0; __p<__MAX_NUM_SENDS; __p++) {
                send_buf_B[__tid][__p] = (double *) polyrt_max_alloc(send_buf_B[__tid][__p], sizeof(double)*max_num_elements_B, &send_buf_size_B[__tid][__p]);
}
}
for (__p=0; __p<__MAX_NUM_RECVS; __p++) { recv_buf_B[__p] = (double *) polyrt_max_alloc(recv_buf_B[__p], sizeof(double)*max_num_elements_B, &recv_buf_size_B[__p]); }
            max_num_elements_y = max(max_num_elements_y, (+1*(128)) + 3);
                for (__tid=0; __tid<_num_threads; __tid++) { for (__p=0; __p<__MAX_NUM_SENDS; __p++) {
                send_buf_y[__tid][__p] = (double *) polyrt_max_alloc(send_buf_y[__tid][__p], sizeof(double)*max_num_elements_y, &send_buf_size_y[__tid][__p]);
}
}
for (__p=0; __p<__MAX_NUM_RECVS; __p++) { recv_buf_y[__p] = (double *) polyrt_max_alloc(recv_buf_y[__p], sizeof(double)*max_num_elements_y, &recv_buf_size_y[__p]); }
if (fopen(".test", "r")) {lw_buf_tmp = (double *) polyrt_max_alloc(lw_buf_tmp, sizeof(double)*((+1*(128))*floorf((1*(floorf((+1*n-1)/(float)128) - 0 + 1+nprocs))/(float)nprocs)), &lw_buf_size_tmp);
lw_recv_buf_tmp = (double *) polyrt_max_alloc(lw_recv_buf_tmp, nprocs*lw_buf_size_tmp, &lw_recv_buf_size_tmp);
 for (__p=0; __p<nprocs; __p++) { displs_lw_tmp[__p] = __p*lw_buf_size_tmp/sizeof(double);}
}if (fopen(".test", "r")) {lw_buf_y = (double *) polyrt_max_alloc(lw_buf_y, sizeof(double)*((+1*(1))*floorf((1*(floorf((+1*n-1)/(float)128) - 0 + 1+nprocs))/(float)nprocs)), &lw_buf_size_y);
lw_recv_buf_y = (double *) polyrt_max_alloc(lw_recv_buf_y, nprocs*lw_buf_size_y, &lw_recv_buf_size_y);
 for (__p=0; __p<nprocs; __p++) { displs_lw_y[__p] = __p*lw_buf_size_y/sizeof(double);}
} max_num_elements_tmp = max(max_num_elements_tmp, (+1*(1)) + 3);
                for (__tid=0; __tid<_num_threads; __tid++) { for (__p=0; __p<__MAX_NUM_SENDS; __p++) {
                send_buf_tmp[__tid][__p] = (double *) polyrt_max_alloc(send_buf_tmp[__tid][__p], sizeof(double)*max_num_elements_tmp, &send_buf_size_tmp[__tid][__p]);
}
}
for (__p=0; __p<__MAX_NUM_RECVS; __p++) { recv_buf_tmp[__p] = (double *) polyrt_max_alloc(recv_buf_tmp[__p], sizeof(double)*max_num_elements_tmp, &recv_buf_size_tmp[__p]); }
            max_num_elements_y = max(max_num_elements_y, (+1*(1)) + 3);
                for (__tid=0; __tid<_num_threads; __tid++) { for (__p=0; __p<__MAX_NUM_SENDS; __p++) {
                send_buf_y[__tid][__p] = (double *) polyrt_max_alloc(send_buf_y[__tid][__p], sizeof(double)*max_num_elements_y, &send_buf_size_y[__tid][__p]);
}
}
for (__p=0; __p<__MAX_NUM_RECVS; __p++) { recv_buf_y[__p] = (double *) polyrt_max_alloc(recv_buf_y[__p], sizeof(double)*max_num_elements_y, &recv_buf_size_y[__p]); }
if (fopen(".test", "r")) {lw_buf_y = (double *) polyrt_max_alloc(lw_buf_y, sizeof(double)*((+1*(128))*floorf((1*(floorf((+1*n-1)/(float)128) - 0 + 1+nprocs))/(float)nprocs)), &lw_buf_size_y);
lw_recv_buf_y = (double *) polyrt_max_alloc(lw_recv_buf_y, nprocs*lw_buf_size_y, &lw_recv_buf_size_y);
 for (__p=0; __p<nprocs; __p++) { displs_lw_y[__p] = __p*lw_buf_size_y/sizeof(double);}
} polyrt_init_grid_size();
 IF_TIME(t_data_init_start = rtclock());
data_tile_init_0(my_rank, nprocs 
#ifndef USE_LOCAL_ARRAYS
 , tmp , A , x , y , B 
#endif
 , tmp_trans, buff_mang_tmp, A_trans, A_size_1, buff_mang_A, x_trans, buff_mang_x, y_trans, buff_mang_y, B_trans, B_size_1, buff_mang_B);
 IF_TIME(t_data_init += rtclock() - t_data_init_start)
 IF_TIME(t_local = rtclock());
/*Start of dynamic scheduling code*/
#if defined(__DYNSCHEDULER_DEBUG_PRINT) || defined(__DYNSCHEDULER_MORE_DEBUG_PRINT) || defined(__DEBUG_FLUSH)
 char *debug_print_name = (char *)malloc(128*sizeof(char));
 strcpy(debug_print_name, "debug_print");
 sprintf(debug_print_name+strlen(debug_print_name), "_node%d", my_rank);
 __debug_print_fp = fopen(debug_print_name, "w");
 free(debug_print_name);
#endif
 IF_TIME(t_tasks_create = rtclock());
 init_tasks(my_rank,nprocs,_num_threads,tasks_loop0,tasks_loop1,&_num_tasks_to_execute,&_num_tasks_to_unpack);
 IF_TIME(t_tasks_create = rtclock() - t_tasks_create);
#ifdef __DYNSCHEDULER_MORE_DEBUG_PRINT
 std::vector<int> prev_firing_count_tmp_0;
 std::vector<int> prev_firing_count_tmp_1;
 std::vector<int> prev_firing_count_y_0;
 std::vector<int> prev_firing_count_y_1;
 for (__p=0; __p<__MAX_NUM_RECVS; __p++) {
  prev_firing_count_tmp_0.push_back(-1);
  prev_firing_count_tmp_1.push_back(-1);
  prev_firing_count_y_0.push_back(-1);
  prev_firing_count_y_1.push_back(-1);
 }
#endif
#pragma omp parallel private(__tid)
{
 __tid = omp_get_thread_num();
 int __flag;
 int __loop;
 int task_id1,task_id2;
#pragma omp single nowait
{ /* receive thread */
 __tid_receiver = __tid;
 IF_TIME(t_unpack_start = rtclock());
 for (__p=0; __p<__MAX_NUM_RECVS; __p++) {
  MPI_Irecv(recv_buf_tmp[__p], max_num_elements_tmp, MPI_DOUBLE, MPI_ANY_SOURCE, 2123, MPI_COMM_WORLD, &recv_reqs_tmp[__p]);
  MPI_Irecv(recv_buf_y[__p], max_num_elements_y, MPI_DOUBLE, MPI_ANY_SOURCE, 5123, MPI_COMM_WORLD, &recv_reqs_y[__p]);
 }
 IF_TIME(t_unpack += rtclock() - t_unpack_start);
#ifdef __DYNSCHEDULER_DEBUG_PRINT
 fprintf(__debug_print_fp, "node %d recv thread %d num tasks to unpack %ld max recv buffers %d tmp max elements %d y max elements %d\n", my_rank, __tid, _num_tasks_to_unpack, __MAX_NUM_RECVS, max_num_elements_tmp, max_num_elements_y);
 IF_DEBUG_FLUSH(fflush(__debug_print_fp));
#endif
 IF_TIME(t_wait_unpack_start = rtclock());
 while(_num_tasks_to_unpack > 0) {
  for (__p=0; (__p<recv_buf_tmp.size()) || (__p<recv_buf_y.size()); __p++) {
  if (__p<recv_buf_tmp.size()) {
   if (recv_reqs_tmp[__p] != MPI_REQUEST_NULL) {
    MPI_Test(&recv_reqs_tmp[__p],&recv_completed_tmp[__p],recv_stats_tmp);
#ifdef __DYNSCHEDULER_DEBUG_PRINT
    if (recv_completed_tmp[__p]) {
     __loop = (int)recv_buf_tmp[__p][0];
     task_id1 = (int)recv_buf_tmp[__p][1];
     task_id2 = (int)recv_buf_tmp[__p][2];
     int size;
     MPI_Get_count(recv_stats_tmp, MPI_DOUBLE, &size);
     fprintf(__debug_print_fp, "received tmp node %d from node %d buf %d size %d loop %d task %d %d\n", my_rank, recv_stats_tmp->MPI_SOURCE, __p, size, __loop, task_id1, task_id2);
     IF_DEBUG_FLUSH(fflush(__debug_print_fp));
    }
#endif
   }
   if (recv_completed_tmp[__p]) {
    __loop = (int)recv_buf_tmp[__p][0];
    task_id1 = (int)recv_buf_tmp[__p][1];
    task_id2 = (int)recv_buf_tmp[__p][2];
    switch(__loop){
     case 0:
#ifdef __DYNSCHEDULER_MORE_DEBUG_PRINT
      if (prev_firing_count_tmp_0[__p] != tasks_loop0[(task_id2-lb_tasks_loop0_dim0)+0]) {
       prev_firing_count_tmp_0[__p] = tasks_loop0[(task_id2-lb_tasks_loop0_dim0)+0];
       fprintf(__debug_print_fp, "waiting to unpack tmp node %d loop %d task %d %d count %d\n", my_rank, __loop, task_id1, task_id2, tasks_loop0[(task_id2-lb_tasks_loop0_dim0)+0]);
       IF_DEBUG_FLUSH(fflush(__debug_print_fp));
      }
#endif
      if (tasks_loop0[(task_id2-lb_tasks_loop0_dim0)+0] <= 2) {
       IF_TIME(t_wait_unpack += rtclock() - t_wait_unpack_start);
       IF_TIME(t_unpack_start = rtclock());
       if (tasks_loop0[(task_id2-lb_tasks_loop0_dim0)+0] == 2) {
        tasks_loop0[(task_id2-lb_tasks_loop0_dim0)+0] -= !is_receiver_tmp_0(task_id1,task_id2,n,my_rank,nprocs);
        tasks_loop0[(task_id2-lb_tasks_loop0_dim0)+0] -= !is_receiver_y_0(task_id1,task_id2,n,my_rank,nprocs);
       }
       unpack_buf_tmp_0(my_rank,nprocs, tmp_trans, buff_mang_tmp,recv_buf_tmp[__p]);
       tasks_loop0[(task_id2-lb_tasks_loop0_dim0)+0]--;
       recv_completed_tmp[__p] = 0;
       MPI_Irecv(recv_buf_tmp[__p], max_num_elements_tmp, MPI_DOUBLE, MPI_ANY_SOURCE, 2123, MPI_COMM_WORLD, &recv_reqs_tmp[__p]);
#ifdef __DYNSCHEDULER_MORE_DEBUG_PRINT
       prev_firing_count_tmp_0[__p] = -1;
#endif
#ifdef __DYNSCHEDULER_DEBUG_PRINT
       fprintf(__debug_print_fp, "unpacked tmp node %d loop %d task %d %d count %d\n", my_rank, __loop, task_id1, task_id2, tasks_loop0[(task_id2-lb_tasks_loop0_dim0)+0]);
       IF_DEBUG_FLUSH(fflush(__debug_print_fp));
#endif
       if (tasks_loop0[(task_id2-lb_tasks_loop0_dim0)+0] == 0) {
        remote_update_dep_tasks_0(task_id1,task_id2,n,my_rank,nprocs,_num_threads,tasks_loop0,tasks_loop1);
        _num_tasks_to_unpack--;
#ifdef __DYNSCHEDULER_DEBUG_PRINT
        fprintf(__debug_print_fp, "remote updated tmp node %d loop %d task %d %d num tasks left to unpack %ld\n", my_rank, __loop, task_id1, task_id2, _num_tasks_to_unpack);
        IF_DEBUG_FLUSH(fflush(__debug_print_fp));
#endif
       }
       IF_TIME(t_unpack += rtclock() - t_unpack_start);
       IF_TIME(t_wait_unpack_start = rtclock());
      }
      break;
     default:
      fprintf(stderr, "ERROR: tmp data cannot be sent by a task of loop %d\n", __loop);
      assert(0);
    }
   }
  }
  if (__p<recv_buf_y.size()) {
   if (recv_reqs_y[__p] != MPI_REQUEST_NULL) {
    MPI_Test(&recv_reqs_y[__p],&recv_completed_y[__p],recv_stats_y);
#ifdef __DYNSCHEDULER_DEBUG_PRINT
    if (recv_completed_y[__p]) {
     __loop = (int)recv_buf_y[__p][0];
     task_id1 = (int)recv_buf_y[__p][1];
     task_id2 = (int)recv_buf_y[__p][2];
     int size;
     MPI_Get_count(recv_stats_y, MPI_DOUBLE, &size);
     fprintf(__debug_print_fp, "received y node %d from node %d buf %d size %d loop %d task %d %d\n", my_rank, recv_stats_y->MPI_SOURCE, __p, size, __loop, task_id1, task_id2);
     IF_DEBUG_FLUSH(fflush(__debug_print_fp));
    }
#endif
   }
   if (recv_completed_y[__p]) {
    __loop = (int)recv_buf_y[__p][0];
    task_id1 = (int)recv_buf_y[__p][1];
    task_id2 = (int)recv_buf_y[__p][2];
    switch(__loop){
     case 0:
#ifdef __DYNSCHEDULER_MORE_DEBUG_PRINT
      if (prev_firing_count_y_0[__p] != tasks_loop0[(task_id2-lb_tasks_loop0_dim0)+0]) {
       prev_firing_count_y_0[__p] = tasks_loop0[(task_id2-lb_tasks_loop0_dim0)+0];
       fprintf(__debug_print_fp, "waiting to unpack y node %d loop %d task %d %d count %d\n", my_rank, __loop, task_id1, task_id2, tasks_loop0[(task_id2-lb_tasks_loop0_dim0)+0]);
       IF_DEBUG_FLUSH(fflush(__debug_print_fp));
      }
#endif
      if (tasks_loop0[(task_id2-lb_tasks_loop0_dim0)+0] <= 2) {
       IF_TIME(t_wait_unpack += rtclock() - t_wait_unpack_start);
       IF_TIME(t_unpack_start = rtclock());
       if (tasks_loop0[(task_id2-lb_tasks_loop0_dim0)+0] == 2) {
        tasks_loop0[(task_id2-lb_tasks_loop0_dim0)+0] -= !is_receiver_tmp_0(task_id1,task_id2,n,my_rank,nprocs);
        tasks_loop0[(task_id2-lb_tasks_loop0_dim0)+0] -= !is_receiver_y_0(task_id1,task_id2,n,my_rank,nprocs);
       }
       unpack_buf_y_0(my_rank,nprocs, y_trans, buff_mang_y,recv_buf_y[__p]);
       tasks_loop0[(task_id2-lb_tasks_loop0_dim0)+0]--;
       recv_completed_y[__p] = 0;
       MPI_Irecv(recv_buf_y[__p], max_num_elements_y, MPI_DOUBLE, MPI_ANY_SOURCE, 5123, MPI_COMM_WORLD, &recv_reqs_y[__p]);
#ifdef __DYNSCHEDULER_MORE_DEBUG_PRINT
       prev_firing_count_y_0[__p] = -1;
#endif
#ifdef __DYNSCHEDULER_DEBUG_PRINT
       fprintf(__debug_print_fp, "unpacked y node %d loop %d task %d %d count %d\n", my_rank, __loop, task_id1, task_id2, tasks_loop0[(task_id2-lb_tasks_loop0_dim0)+0]);
       IF_DEBUG_FLUSH(fflush(__debug_print_fp));
#endif
       if (tasks_loop0[(task_id2-lb_tasks_loop0_dim0)+0] == 0) {
        remote_update_dep_tasks_0(task_id1,task_id2,n,my_rank,nprocs,_num_threads,tasks_loop0,tasks_loop1);
        _num_tasks_to_unpack--;
#ifdef __DYNSCHEDULER_DEBUG_PRINT
        fprintf(__debug_print_fp, "remote updated y node %d loop %d task %d %d num tasks left to unpack %ld\n", my_rank, __loop, task_id1, task_id2, _num_tasks_to_unpack);
        IF_DEBUG_FLUSH(fflush(__debug_print_fp));
#endif
       }
       IF_TIME(t_unpack += rtclock() - t_unpack_start);
       IF_TIME(t_wait_unpack_start = rtclock());
      }
      break;
     case 1:
#ifdef __DYNSCHEDULER_MORE_DEBUG_PRINT
      if (prev_firing_count_y_1[__p] != tasks_loop1[(task_id2-lb_tasks_loop1_dim0)+0]) {
       prev_firing_count_y_1[__p] = tasks_loop1[(task_id2-lb_tasks_loop1_dim0)+0];
       fprintf(__debug_print_fp, "waiting to unpack y node %d loop %d task %d %d count %d\n", my_rank, __loop, task_id1, task_id2, tasks_loop1[(task_id2-lb_tasks_loop1_dim0)+0]);
       IF_DEBUG_FLUSH(fflush(__debug_print_fp));
      }
#endif
      if (tasks_loop1[(task_id2-lb_tasks_loop1_dim0)+0] <= 1) {
       IF_TIME(t_wait_unpack += rtclock() - t_wait_unpack_start);
       IF_TIME(t_unpack_start = rtclock());
       if (tasks_loop1[(task_id2-lb_tasks_loop1_dim0)+0] == 1) {
        tasks_loop1[(task_id2-lb_tasks_loop1_dim0)+0] -= !is_receiver_y_1(task_id1,task_id2,n,my_rank,nprocs);
       }
       unpack_buf_y_1(my_rank,nprocs, y_trans, buff_mang_y,recv_buf_y[__p]);
       tasks_loop1[(task_id2-lb_tasks_loop1_dim0)+0]--;
       recv_completed_y[__p] = 0;
       MPI_Irecv(recv_buf_y[__p], max_num_elements_y, MPI_DOUBLE, MPI_ANY_SOURCE, 5123, MPI_COMM_WORLD, &recv_reqs_y[__p]);
#ifdef __DYNSCHEDULER_MORE_DEBUG_PRINT
       prev_firing_count_y_1[__p] = -1;
#endif
#ifdef __DYNSCHEDULER_DEBUG_PRINT
       fprintf(__debug_print_fp, "unpacked y node %d loop %d task %d %d count %d\n", my_rank, __loop, task_id1, task_id2, tasks_loop1[(task_id2-lb_tasks_loop1_dim0)+0]);
       IF_DEBUG_FLUSH(fflush(__debug_print_fp));
#endif
       if (tasks_loop1[(task_id2-lb_tasks_loop1_dim0)+0] == 0) {
        remote_update_dep_tasks_1(task_id1,task_id2,n,my_rank,nprocs,_num_threads,tasks_loop0,tasks_loop1);
        _num_tasks_to_unpack--;
#ifdef __DYNSCHEDULER_DEBUG_PRINT
        fprintf(__debug_print_fp, "remote updated y node %d loop %d task %d %d num tasks left to unpack %ld\n", my_rank, __loop, task_id1, task_id2, _num_tasks_to_unpack);
        IF_DEBUG_FLUSH(fflush(__debug_print_fp));
#endif
       }
       IF_TIME(t_unpack += rtclock() - t_unpack_start);
       IF_TIME(t_wait_unpack_start = rtclock());
      }
      break;
     default:
      fprintf(stderr, "ERROR: y data cannot be sent by a task of loop %d\n", __loop);
      assert(0);
    }
   }
  }
  }
  for (__p=0; __p<recv_buf_tmp.size(); __p++) {
   if (recv_reqs_tmp[__p] != MPI_REQUEST_NULL) break;
  }
  if (__p == recv_buf_tmp.size()) {
   for (__p=0; __p<nprocs; __p++) {
    double *buf = NULL;
    size_t buf_size = 0;
    buf = (double *) polyrt_max_alloc(buf, sizeof(double)*max_num_elements_tmp, &buf_size);
    recv_buf_tmp.push_back(buf);
    recv_reqs_tmp.push_back(MPI_REQUEST_NULL);
    recv_completed_tmp.push_back(0);
#ifdef __DYNSCHEDULER_MORE_DEBUG_PRINT
    prev_firing_count_tmp_0.push_back(-1);
    prev_firing_count_tmp_1.push_back(-1);
#endif
    MPI_Irecv(buf, max_num_elements_tmp, MPI_DOUBLE, MPI_ANY_SOURCE, 2123, MPI_COMM_WORLD, &recv_reqs_tmp[recv_reqs_tmp.size()-1]);
   }
  }
  for (__p=0; __p<recv_buf_y.size(); __p++) {
   if (recv_reqs_y[__p] != MPI_REQUEST_NULL) break;
  }
  if (__p == recv_buf_y.size()) {
   for (__p=0; __p<nprocs; __p++) {
    double *buf = NULL;
    size_t buf_size = 0;
    buf = (double *) polyrt_max_alloc(buf, sizeof(double)*max_num_elements_y, &buf_size);
    recv_buf_y.push_back(buf);
    recv_reqs_y.push_back(MPI_REQUEST_NULL);
    recv_completed_y.push_back(0);
#ifdef __DYNSCHEDULER_MORE_DEBUG_PRINT
    prev_firing_count_y_0.push_back(-1);
    prev_firing_count_y_1.push_back(-1);
#endif
    MPI_Irecv(buf, max_num_elements_y, MPI_DOUBLE, MPI_ANY_SOURCE, 5123, MPI_COMM_WORLD, &recv_reqs_y[recv_reqs_y.size()-1]);
   }
  }
 }
 IF_TIME(t_unpack_start = rtclock());
 for (__p=0; __p<recv_buf_tmp.size(); __p++) {
  if (recv_reqs_tmp[__p] != MPI_REQUEST_NULL) MPI_Request_free(&recv_reqs_tmp[__p]);
 }
 for (__p=0; __p<recv_buf_y.size(); __p++) {
  if (recv_reqs_y[__p] != MPI_REQUEST_NULL) MPI_Request_free(&recv_reqs_y[__p]);
 }
 IF_TIME(t_unpack += rtclock() - t_unpack_start);
#ifdef __DYNSCHEDULER_DEBUG_PRINT
 fprintf(__debug_print_fp, "node %d receive done\n", my_rank);
 IF_DEBUG_FLUSH(fflush(__debug_print_fp));
#endif
}
{ /* compute-pack-send threads */
#ifdef __DYNSCHEDULER_DEBUG_PRINT
 fprintf(__debug_print_fp, "node %d compute thread %d num tasks to execute %ld\n", my_rank, __tid, _num_tasks_to_execute);
 IF_DEBUG_FLUSH(fflush(__debug_print_fp));
#endif
 IF_TIME(t_wait_comp_start[__tid] = rtclock());
 while (_num_tasks_to_execute > 0) {
  __Task task;
  __flag = pqueue.try_pop(task);
  if(__flag==0) { IF_TIME(waiting_pops[__tid]++); continue; }
  __loop = task.loop;
  task_id1 = task.id1;
  task_id2 = task.id2;
#ifdef __DYNSCHEDULER_DEBUG_PRINT
  long int captured;
#pragma omp atomic capture
  captured = --_num_tasks_to_execute;
  fprintf(__debug_print_fp, "computing node %d thread %d loop %d task %d %d ready tasks %lu\n", my_rank, __tid, __loop, task_id1, task_id2, pqueue.size());
  IF_DEBUG_FLUSH(fflush(__debug_print_fp));
#else
#pragma omp atomic
  _num_tasks_to_execute--;
#endif
  switch(__loop){
   case 0:
    IF_TIME(t_wait_comp[__tid] += rtclock() - t_wait_comp_start[__tid]);
    IF_TIME(t_data_mang_start[__tid] = rtclock());
    data_tile_alloc_0(task_id1,task_id2, tmp_trans, buff_mang_tmp, A_trans, A_size_1, buff_mang_A, x_trans, buff_mang_x, y_trans, buff_mang_y, B_trans, B_size_1, buff_mang_B);
    IF_TIME(t_data_mang[__tid] += rtclock() - t_data_mang_start[__tid]);
    IF_TIME(t_comp_start[__tid] = rtclock());
    compute_task_0(task_id1,task_id2, tmp_trans, buff_mang_tmp, A_trans, A_size_1, buff_mang_A, x_trans, buff_mang_x, y_trans, buff_mang_y, B_trans, B_size_1, buff_mang_B);
    IF_TIME(t_comp[__tid] += rtclock() - t_comp_start[__tid]);
    IF_TIME(t_data_mang_start[__tid] = rtclock());
    data_tile_ref_count_update_0(task_id1,task_id2, tmp_trans, buff_mang_tmp, A_trans, A_size_1, buff_mang_A, x_trans, buff_mang_x, y_trans, buff_mang_y, B_trans, B_size_1, buff_mang_B);
    IF_TIME(t_data_mang[__tid] += rtclock() - t_data_mang_start[__tid]);
    IF_TIME(t_pack_start[__tid] = rtclock());
    pack_send_0(my_rank,nprocs,task_id1,task_id2, tmp_trans, buff_mang_tmp,send_buf_tmp[__tid],send_reqs_tmp[__tid],send_stats_tmp[__tid], A_trans, A_size_1, buff_mang_A,send_buf_A[__tid],send_reqs_A[__tid],send_stats_A[__tid], x_trans, buff_mang_x,send_buf_x[__tid],send_reqs_x[__tid],send_stats_x[__tid], y_trans, buff_mang_y,send_buf_y[__tid],send_reqs_y[__tid],send_stats_y[__tid], B_trans, B_size_1, buff_mang_B,send_buf_B[__tid],send_reqs_B[__tid],send_stats_B[__tid],__MAX_NUM_SENDS,&__total_count[__tid]);
    IF_TIME(t_pack[__tid] += rtclock() - t_pack_start[__tid]);
    IF_TIME(t_tasks_manage_start[__tid] = rtclock());
    local_update_dep_tasks_0(task_id1,task_id2,n,my_rank,nprocs,_num_threads,tasks_loop0,tasks_loop1);
    IF_TIME(t_tasks_manage[__tid] += rtclock() - t_tasks_manage_start[__tid]);
    IF_TIME(t_wait_comp_start[__tid] = rtclock());
    break;
   case 1:
    IF_TIME(t_wait_comp[__tid] += rtclock() - t_wait_comp_start[__tid]);
    IF_TIME(t_data_mang_start[__tid] = rtclock());
    data_tile_alloc_1(task_id1,task_id2, y_trans, buff_mang_y, tmp_trans, buff_mang_tmp);
    IF_TIME(t_data_mang[__tid] += rtclock() - t_data_mang_start[__tid]);
    IF_TIME(t_comp_start[__tid] = rtclock());
    compute_task_1(task_id1,task_id2, y_trans, buff_mang_y, tmp_trans, buff_mang_tmp);
    IF_TIME(t_comp[__tid] += rtclock() - t_comp_start[__tid]);
    IF_TIME(t_data_mang_start[__tid] = rtclock());
    data_tile_ref_count_update_1(task_id1,task_id2, y_trans, buff_mang_y, tmp_trans, buff_mang_tmp);
    IF_TIME(t_data_mang[__tid] += rtclock() - t_data_mang_start[__tid]);
    IF_TIME(t_pack_start[__tid] = rtclock());
    pack_send_1(my_rank,nprocs,task_id1,task_id2, y_trans, buff_mang_y,send_buf_y[__tid],send_reqs_y[__tid],send_stats_y[__tid], tmp_trans, buff_mang_tmp,send_buf_tmp[__tid],send_reqs_tmp[__tid],send_stats_tmp[__tid],__MAX_NUM_SENDS,&__total_count[__tid]);
    IF_TIME(t_pack[__tid] += rtclock() - t_pack_start[__tid]);
    IF_TIME(t_tasks_manage_start[__tid] = rtclock());
    local_update_dep_tasks_1(task_id1,task_id2,n,my_rank,nprocs,_num_threads,tasks_loop0,tasks_loop1);
    IF_TIME(t_tasks_manage[__tid] += rtclock() - t_tasks_manage_start[__tid]);
    IF_TIME(t_wait_comp_start[__tid] = rtclock());
    break;
   default:
    assert(0);
  }
#ifdef __DYNSCHEDULER_MORE_DEBUG_PRINT
  fprintf(__debug_print_fp, "computed node %d thread %d loop %d task %d %d tasks left %ld\n", my_rank, __tid, __loop, task_id1, task_id2, captured);
  IF_DEBUG_FLUSH(fflush(__debug_print_fp));
#endif
 }
#ifdef __DYNSCHEDULER_DEBUG_PRINT
 fprintf(__debug_print_fp, "node %d compute thread %d done\n", my_rank, __tid);
 IF_DEBUG_FLUSH(fflush(__debug_print_fp));
#endif
}
}
#ifdef __DYNSCHEDULER_DEBUG_PRINT
 fclose(__debug_print_fp);
#endif
/*End of dynamic scheduling code*/
 IF_TIME(t_writeout = rtclock());
 
#ifndef USE_LOCAL_ARRAYS
 if (fopen(".test", "r")) {
 copy_back_writeout_0(my_rank, nprocs 
#ifndef USE_LOCAL_ARRAYS
 , tmp , y 
#endif
 , tmp_trans, buff_mang_tmp, y_trans, buff_mang_y);
write_out(my_rank,nprocs,lw_buf_tmp,lw_recv_buf_tmp,displs_lw_tmp,lw_buf_y,lw_recv_buf_y,displs_lw_y,tmp,y, tmp_trans, buff_mang_tmp, y_trans, buff_mang_y);
} 
#endif
 IF_TIME(t_writeout = rtclock() - t_writeout);
free_buffer_mang(buff_mang_tmp);
free_buffer_mang(buff_mang_A);
free_buffer_mang(buff_mang_x);
free_buffer_mang(buff_mang_y);
free_buffer_mang(buff_mang_B);
#ifdef TIME
 IF_TIME(t_local = rtclock() - t_local - t_writeout);
 for (__tid=0; __tid<_num_threads; __tid++) __total_count_threads += __total_count[__tid];
 waiting_pops_threads = waiting_pops[0];
                    for (__tid=1; __tid<_num_threads; __tid++) if (waiting_pops_threads < waiting_pops[__tid]) waiting_pops_threads = waiting_pops[__tid];
 t_tasks_manage_threads = t_tasks_manage[0];
                    for (__tid=1; __tid<_num_threads; __tid++) if (t_tasks_manage_threads < t_tasks_manage[__tid]) t_tasks_manage_threads = t_tasks_manage[__tid];
 if (__tid_receiver != 0) t_comp_threads_min = t_comp[0];
                    else t_comp_threads_min = t_comp[1];
                    for (__tid=1; __tid<_num_threads; __tid++) if ((__tid != __tid_receiver) && (t_comp_threads_min > t_comp[__tid])) t_comp_threads_min = t_comp[__tid];
 t_comp_threads = t_comp[0];
                    for (__tid=1; __tid<_num_threads; __tid++) if (t_comp_threads < t_comp[__tid]) t_comp_threads = t_comp[__tid];
 t_wait_comp_threads = t_wait_comp[0];
                    for (__tid=1; __tid<_num_threads; __tid++) if (t_wait_comp_threads < t_wait_comp[__tid]) t_wait_comp_threads = t_wait_comp[__tid];
 t_data_mang_threads = t_data_mang[0];
      for (__tid=1; __tid<_num_threads; __tid++) if (t_data_mang_threads < t_data_mang[__tid]) t_data_mang_threads = t_data_mang[__tid];
 t_pack_threads = t_pack[0];
                    for (__tid=1; __tid<_num_threads; __tid++) if (t_pack_threads < t_pack[__tid]) t_pack_threads = t_pack[__tid];
 char *buffer = (char *)malloc(4096 * sizeof(char));
 strcpy(buffer, "");
#ifdef __DETAILED_TIME_REPORT
 sprintf(buffer+strlen(buffer), "Node %d\n", my_rank);
 sprintf(buffer+strlen(buffer), "Write-out time: %lf\n", t_writeout);
 sprintf(buffer+strlen(buffer), "Values communicated: %0.0lf\n", __total_count_threads);
 sprintf(buffer+strlen(buffer), "Data init time: %lf\n", t_data_init);
 sprintf(buffer+strlen(buffer), "Minimum computation time across all threads: %lf\n", t_comp_threads_min);
 sprintf(buffer+strlen(buffer), "Maximum computation time across all threads: %lf\n", t_comp_threads);
 sprintf(buffer+strlen(buffer), "'Waiting for computation' time: %lf\n", t_wait_comp_threads);
 sprintf(buffer+strlen(buffer), "Data mang time: %lf\n", t_data_mang_threads);
 sprintf(buffer+strlen(buffer), "Tasks creation time: %lf\n", t_tasks_create);
 sprintf(buffer+strlen(buffer), "Tasks management time: %lf\n", t_tasks_manage_threads);
 sprintf(buffer+strlen(buffer), "Packing time: %lf\n", t_pack_threads);
 sprintf(buffer+strlen(buffer), "Unpacking time: %lf\n", t_unpack);
 sprintf(buffer+strlen(buffer), "'Waiting for unpacking' time: %lf\n", t_wait_unpack);
 sprintf(buffer+strlen(buffer), "Sum of split times: %lf\n", t_comp_threads + t_wait_comp_threads + t_tasks_create + t_tasks_manage_threads + t_pack_threads);
 sprintf(buffer+strlen(buffer), "Total time minus write-out time: %lf\n", t_local);
 sprintf(buffer+strlen(buffer), "-------");
 fprintf(stdout, "%s\n", buffer);
#endif
 MPI_Reduce(&__total_count_threads, &__total_count_all, 1, MPI_DOUBLE, MPI_SUM, 0, MPI_COMM_WORLD);
 MPI_Reduce(&waiting_pops_threads, &waiting_pops_global, 1, MPI_INT, MPI_MAX, 0, MPI_COMM_WORLD);
 MPI_Reduce(&t_comp_threads_min, &t_globalcomp_min, 1, MPI_DOUBLE, MPI_MIN, 0, MPI_COMM_WORLD);
 MPI_Reduce(&t_comp_threads, &t_globalcomp, 1, MPI_DOUBLE, MPI_MAX, 0, MPI_COMM_WORLD);
 MPI_Reduce(&t_wait_comp_threads, &t_globalwait_comp, 1, MPI_DOUBLE, MPI_MAX, 0, MPI_COMM_WORLD);
 MPI_Reduce(&t_tasks_create, &t_globaltasks_create, 1, MPI_DOUBLE, MPI_MAX, 0, MPI_COMM_WORLD);
 MPI_Reduce(&t_tasks_manage_threads, &t_globaltasks_manage, 1, MPI_DOUBLE, MPI_MAX, 0, MPI_COMM_WORLD);
 MPI_Reduce(&t_pack_threads, &t_globalpack, 1, MPI_DOUBLE, MPI_MAX, 0, MPI_COMM_WORLD);
 MPI_Reduce(&t_unpack, &t_globalunpack, 1, MPI_DOUBLE, MPI_MAX, 0, MPI_COMM_WORLD);
 MPI_Reduce(&t_wait_unpack, &t_globalwait_unpack, 1, MPI_DOUBLE, MPI_MAX, 0, MPI_COMM_WORLD);
 MPI_Reduce(&t_local, &t_global, 1, MPI_DOUBLE, MPI_MAX, 0, MPI_COMM_WORLD);
 if (my_rank==0) {
  strcpy(buffer, "SUMMARY\n");
  sprintf(buffer+strlen(buffer), "Write-out time spent in master node: %0.6lf s\n", t_writeout);
  sprintf(buffer+strlen(buffer), "Total communication volume across all nodes: %0.6lf GB\n", __total_count_all*8/(1024*1024*1024));
  sprintf(buffer+strlen(buffer), "Maximum computation time spent across all nodes: %0.6lf s\n", t_globalcomp);
  sprintf(buffer+strlen(buffer), "Minimum computation time spent across all nodes: %0.6lf s\n", t_globalcomp_min);
  sprintf(buffer+strlen(buffer), "Maximum 'waiting for computation' time spent across all nodes: %0.6lf s\n", t_globalwait_comp);
  sprintf(buffer+strlen(buffer), "Maximum number of waiting/unsuccessful POPs across all nodes: %d\n", waiting_pops_global);
  sprintf(buffer+strlen(buffer), "Maximum tasks creation time spent across all nodes: %0.6lf s\n", t_globaltasks_create);
  sprintf(buffer+strlen(buffer), "Maximum tasks management time spent across all nodes: %0.6lf s\n", t_globaltasks_manage);
  sprintf(buffer+strlen(buffer), "Maximum packing time spent across all nodes: %0.6lf s\n", t_globalpack);
  sprintf(buffer+strlen(buffer), "Maximum unpacking time spent across all nodes: %0.6lf s\n", t_globalunpack);
  sprintf(buffer+strlen(buffer), "Maximum 'waiting for unpacking' time spent across all nodes: %0.6lf s\n", t_globalwait_unpack);
  sprintf(buffer+strlen(buffer), "Maximum total time spent across all nodes: %0.6lf s\n", t_global);
  sprintf(buffer+strlen(buffer), "Maximum data init time spent across all nodes: %0.6lf s\n", t_globaldata_init);
  sprintf(buffer+strlen(buffer), "-------");
  fprintf(stdout, "%s\n", buffer);
 }
#endif
 MPI_Barrier(MPI_COMM_WORLD);
 MPI_Finalize();

}


int main(int argc, char** argv)
{
  /* Retrieve problem size. */

  if(argc == 3){
      int np = atoi(argv[2]);
      int nprocs = atoi(argv[1]);
      n = np * sqrt(nprocs);
  }


#ifndef USE_LOCAL_ARRAYS
  /* Variable declaration/allocation. */
  POLYBENCH_2D_ARRAY_DECL(A, DATA_TYPE, N, N, n, n);
  POLYBENCH_2D_ARRAY_DECL(B, DATA_TYPE, N, N, n, n);
  POLYBENCH_1D_ARRAY_DECL(tmp, DATA_TYPE, N, n);
  POLYBENCH_1D_ARRAY_DECL(x, DATA_TYPE, N, n);
  POLYBENCH_1D_ARRAY_DECL(y, DATA_TYPE, N, n);


  /* Initialize array(s). */
  init_array (n, 
	      POLYBENCH_ARRAY(A),
	      POLYBENCH_ARRAY(B),
	      POLYBENCH_ARRAY(x),
	      POLYBENCH_ARRAY(tmp),
	      POLYBENCH_ARRAY(y));

  /* Start timer. */
  polybench_start_instruments;

  /* Run kernel. */
  kernel_gesummv (n, 
		  POLYBENCH_ARRAY(A),
		  POLYBENCH_ARRAY(B),
		  POLYBENCH_ARRAY(tmp),
		  POLYBENCH_ARRAY(x),
		  POLYBENCH_ARRAY(y));

  /* Stop and print timer. */
  polybench_stop_instruments;
  polybench_print_instruments;

  /* Prevent dead-code elimination. All live-out data must be printed
     by the function call in argument. */
  polybench_prevent_dce(print_array(n, POLYBENCH_ARRAY(y)));

  /* Be clean. */
  POLYBENCH_FREE_ARRAY(A);
  POLYBENCH_FREE_ARRAY(B);
  POLYBENCH_FREE_ARRAY(tmp);
  POLYBENCH_FREE_ARRAY(x);
  POLYBENCH_FREE_ARRAY(y);
#else

  /* Run kernel. */
  kernel_gesummv (n);

#endif
  return 0;
}


void init_tasks(int my_rank,int nprocs,int num_threads,int *tasks_loop0,int *tasks_loop1,long int *p_num_tasks_to_execute,long int *p_num_tasks_to_unpack)
{
int remote_dep_tasks, local_dep_tasks, affinity;
long int _num_tasks_to_execute = 0;
long int _num_tasks_to_unpack = 0;



  int t1, t2, t3;

 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;

/* Start of CLooG code */
if (n >= 1) {
  lbp=0;
  ubp=floord(n-1,128);
#pragma omp parallel for private(remote_dep_tasks, local_dep_tasks, affinity) reduction(+:_num_tasks_to_execute,_num_tasks_to_unpack)
  for (t3=lbp;t3<=ubp;t3++) {
    if (pi_0(0,t3,n,nprocs) == my_rank) { _num_tasks_to_execute++;tasks_loop0[(t3-lb_tasks_loop0_dim0)+0] = get_num_remote_src_tasks_0(0,t3,n,my_rank,nprocs) + get_num_local_src_tasks_0(0,t3,n,my_rank,nprocs);if (tasks_loop0[(t3-lb_tasks_loop0_dim0)+0] == 0) { remote_dep_tasks = get_num_remote_dep_tasks_0(0,t3,n,my_rank,nprocs); local_dep_tasks = get_num_local_dep_tasks_0(0,t3,n,my_rank,nprocs); if (__is_multi_partitioned[0] || __is_block_cyclic[0]) affinity = -1; else affinity = pi_threads_0(0,t3,n,my_rank,nprocs,num_threads); __Task task(0,0,t3,remote_dep_tasks,local_dep_tasks,affinity); pqueue.push(task); IF_DYNSCHEDULER_MORE_DEBUG_PRINT( fprintf(__debug_print_fp, "added node %d loop 0 task %d %d                 remote_dep_tasks %d local_dep_tasks %d ready tasks %lu\n", my_rank, 0,t3, remote_dep_tasks, local_dep_tasks, pqueue.size())); IF_DEBUG_FLUSH(fflush(__debug_print_fp)); }} else if (is_receiver_0(0,t3,n,my_rank,nprocs)) { _num_tasks_to_unpack++;tasks_loop0[(t3-lb_tasks_loop0_dim0)+0] = 2 + get_num_local_src_tasks_0(0,t3,n,my_rank,nprocs);};
  }
  lbp=0;
  ubp=floord(n-1,128);
#pragma omp parallel for private(remote_dep_tasks, local_dep_tasks, affinity) reduction(+:_num_tasks_to_execute,_num_tasks_to_unpack)
  for (t3=lbp;t3<=ubp;t3++) {
    if (pi_1(1,t3,n,nprocs) == my_rank) { _num_tasks_to_execute++;tasks_loop1[(t3-lb_tasks_loop1_dim0)+0] = get_num_remote_src_tasks_1(1,t3,n,my_rank,nprocs) + get_num_local_src_tasks_1(1,t3,n,my_rank,nprocs);if (tasks_loop1[(t3-lb_tasks_loop1_dim0)+0] == 0) { remote_dep_tasks = get_num_remote_dep_tasks_1(1,t3,n,my_rank,nprocs); local_dep_tasks = get_num_local_dep_tasks_1(1,t3,n,my_rank,nprocs); if (__is_multi_partitioned[1] || __is_block_cyclic[1]) affinity = -1; else affinity = pi_threads_1(1,t3,n,my_rank,nprocs,num_threads); __Task task(1,1,t3,remote_dep_tasks,local_dep_tasks,affinity); pqueue.push(task); IF_DYNSCHEDULER_MORE_DEBUG_PRINT( fprintf(__debug_print_fp, "added node %d loop 1 task %d %d                 remote_dep_tasks %d local_dep_tasks %d ready tasks %lu\n", my_rank, 1,t3, remote_dep_tasks, local_dep_tasks, pqueue.size())); IF_DEBUG_FLUSH(fflush(__debug_print_fp)); }} else if (is_receiver_1(1,t3,n,my_rank,nprocs)) { _num_tasks_to_unpack++;tasks_loop1[(t3-lb_tasks_loop1_dim0)+0] = 1 + get_num_local_src_tasks_1(1,t3,n,my_rank,nprocs);};
  }
}
/* End of CLooG code */


*p_num_tasks_to_execute = _num_tasks_to_execute;
*p_num_tasks_to_unpack = _num_tasks_to_unpack;
}

void data_tile_ref_count_init_tmp_0(int ts1,int ts2, double** tmp_trans, BufferManager *buff_mang_tmp)
{


  int t1, t2;

 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;

/* Start of CLooG code */
if ((ts1 == 0) && (ts2 >= 0) && (128*ts2 <= n-1)) {
  increment_ref_count(buff_mang_tmp, (ts2));;
}
/* End of CLooG code */

}

void data_tile_ref_count_init_A_0(int ts1,int ts2, double** A_trans,int A_size_1, BufferManager *buff_mang_A)
{


  int t1, t2, t3;

 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;

/* Start of CLooG code */
if ((ts1 == 0) && (ts2 >= 0) && (128*ts2 <= n-1)) {
  for (t3=0;t3<=floord(n-1,128);t3++) {
    increment_ref_count(buff_mang_A, (ts2)* ( A_size_1 ) + (t3));;
  }
}
/* End of CLooG code */

}

void data_tile_ref_count_init_x_0(int ts1,int ts2, double** x_trans, BufferManager *buff_mang_x)
{


  int t1, t2;

 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;

/* Start of CLooG code */
if ((ts1 == 0) && (ts2 >= 0) && (128*ts2 <= n-1)) {
  for (t2=0;t2<=floord(n-1,128);t2++) {
    increment_ref_count(buff_mang_x, (t2));;
  }
}
/* End of CLooG code */

}

void data_tile_ref_count_init_y_0(int ts1,int ts2, double** y_trans, BufferManager *buff_mang_y)
{


  int t1, t2;

 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;

/* Start of CLooG code */
if ((ts1 == 0) && (ts2 >= 0) && (128*ts2 <= n-1)) {
  increment_ref_count(buff_mang_y, (ts2));;
}
/* End of CLooG code */

}

void data_tile_ref_count_init_B_0(int ts1,int ts2, double** B_trans,int B_size_1, BufferManager *buff_mang_B)
{


  int t1, t2, t3;

 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;

/* Start of CLooG code */
if ((ts1 == 0) && (ts2 >= 0) && (128*ts2 <= n-1)) {
  for (t3=0;t3<=floord(n-1,128);t3++) {
    increment_ref_count(buff_mang_B, (ts2)* ( B_size_1 ) + (t3));;
  }
}
/* End of CLooG code */

}

void read_in_data_tile_alloc_A_0(int ts1,int ts2, double** A_trans,int A_size_1, BufferManager *buff_mang_A)
{


  int t1, t2, t3;

 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;

/* Start of CLooG code */
if ((ts1 == 0) && (ts2 >= 0) && (128*ts2 <= n-1)) {
  for (t3=0;t3<=floord(n-1,128);t3++) {
    if(A_trans[(ts2) * ((A_size_1)) + (t3)] == NULL) A_trans[(ts2)* ( A_size_1 ) + (t3)] = (double *) buffer_alloc_atomic((ts2)* ( A_size_1 ) + (t3), 128*128 * sizeof(double), buff_mang_A);;
  }
}
/* End of CLooG code */

}

void read_in_copy_A_0(int ts1,int ts2 
#ifndef USE_LOCAL_ARRAYS
 ,__DECLARATION_OF_A 
#endif
 , double** A_trans,int A_size_1, BufferManager *buff_mang_A)
{


  int t1, t2, t3;

 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;

/* Start of CLooG code */
if ((ts1 == 0) && (ts2 >= 0) && (128*ts2 <= n-1)) {
  for (t2=128*ts2;t2<=min(n-1,128*ts2+127);t2++) {
    for (t3=0;t3<=n-1;t3++) {
      
#ifndef USE_LOCAL_ARRAYS
 A_trans[((1)*( t2 ))/128* ( A_size_1 ) + ((1)*( t3 ))/128][((1)*( t2 ))%128* (128) + ((1)*( t3 ))%128] = A[t2][t3]; 
#else
 A_trans[((1)*( t2 ))/128* ( A_size_1 ) + ((1)*( t3 ))/128][((1)*( t2 ))%128* (128) + ((1)*( t3 ))%128] = ((DATA_TYPE) t2*t3) / N; 
#endif
;
    }
  }
}
/* End of CLooG code */

}

void read_in_data_tile_alloc_x_0(int ts1,int ts2, double** x_trans, BufferManager *buff_mang_x)
{


  int t1, t2;

 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;

/* Start of CLooG code */
if ((ts1 == 0) && (ts2 >= 0) && (128*ts2 <= n-1)) {
  for (t2=0;t2<=floord(n-1,128);t2++) {
    if(x_trans[(t2)] == NULL) x_trans[(t2)] = (double *) buffer_alloc_atomic((t2), 128 * sizeof(double), buff_mang_x);;
  }
}
/* End of CLooG code */

}

void read_in_copy_x_0(int ts1,int ts2 
#ifndef USE_LOCAL_ARRAYS
 ,__DECLARATION_OF_x 
#endif
 , double** x_trans, BufferManager *buff_mang_x)
{


  int t1, t2;

 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;

/* Start of CLooG code */
if ((ts1 == 0) && (ts2 >= 0) && (128*ts2 <= n-1)) {
  for (t2=0;t2<=n-1;t2++) {
    
#ifndef USE_LOCAL_ARRAYS
 x_trans[((1)*( t2 ))/128][((1)*( t2 ))%128] = x[t2]; 
#else
 x_trans[((1)*( t2 ))/128][((1)*( t2 ))%128] = ((DATA_TYPE) t2) / N; 
#endif
;
  }
}
/* End of CLooG code */

}

void read_in_data_tile_alloc_tmp_0(int ts1,int ts2, double** tmp_trans, BufferManager *buff_mang_tmp)
{


  int t1, t2;

 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;

/* Start of CLooG code */
if ((ts1 == 0) && (ts2 >= 0) && (128*ts2 <= n-1)) {
  if(tmp_trans[(ts2)] == NULL) tmp_trans[(ts2)] = (double *) buffer_alloc_atomic((ts2), 128 * sizeof(double), buff_mang_tmp);;
}
/* End of CLooG code */

}

void read_in_copy_tmp_0(int ts1,int ts2 
#ifndef USE_LOCAL_ARRAYS
 ,__DECLARATION_OF_tmp 
#endif
 , double** tmp_trans, BufferManager *buff_mang_tmp)
{


  int t1, t2;

 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;

/* Start of CLooG code */
if ((ts1 == 0) && (ts2 >= 0) && (128*ts2 <= n-1)) {
  for (t2=128*ts2;t2<=min(n-1,128*ts2+127);t2++) {
    
#ifndef USE_LOCAL_ARRAYS
 tmp_trans[((1)*( t2 ))/128][((1)*( t2 ))%128] = tmp[t2]; 
#else
 tmp_trans[((1)*( t2 ))/128][((1)*( t2 ))%128] = 0; 
#endif
;
  }
}
/* End of CLooG code */

}

void read_in_data_tile_alloc_B_0(int ts1,int ts2, double** B_trans,int B_size_1, BufferManager *buff_mang_B)
{


  int t1, t2, t3;

 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;

/* Start of CLooG code */
if ((ts1 == 0) && (ts2 >= 0) && (128*ts2 <= n-1)) {
  for (t3=0;t3<=floord(n-1,128);t3++) {
    if(B_trans[(ts2) * ((B_size_1)) + (t3)] == NULL) B_trans[(ts2)* ( B_size_1 ) + (t3)] = (double *) buffer_alloc_atomic((ts2)* ( B_size_1 ) + (t3), 128*128 * sizeof(double), buff_mang_B);;
  }
}
/* End of CLooG code */

}

void read_in_copy_B_0(int ts1,int ts2 
#ifndef USE_LOCAL_ARRAYS
 ,__DECLARATION_OF_B 
#endif
 , double** B_trans,int B_size_1, BufferManager *buff_mang_B)
{


  int t1, t2, t3;

 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;

/* Start of CLooG code */
if ((ts1 == 0) && (ts2 >= 0) && (128*ts2 <= n-1)) {
  for (t2=128*ts2;t2<=min(n-1,128*ts2+127);t2++) {
    for (t3=0;t3<=n-1;t3++) {
      
#ifndef USE_LOCAL_ARRAYS
 B_trans[((1)*( t2 ))/128* ( B_size_1 ) + ((1)*( t3 ))/128][((1)*( t2 ))%128* (128) + ((1)*( t3 ))%128] = B[t2][t3]; 
#else
 B_trans[((1)*( t2 ))/128* ( B_size_1 ) + ((1)*( t3 ))/128][((1)*( t2 ))%128* (128) + ((1)*( t3 ))%128] = ((DATA_TYPE) t2*t3) / N; 
#endif
;
    }
  }
}
/* End of CLooG code */

}

void read_in_data_tile_alloc_y_0(int ts1,int ts2, double** y_trans, BufferManager *buff_mang_y)
{


  int t1, t2;

 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;

/* Start of CLooG code */
if ((ts1 == 0) && (ts2 >= 0) && (128*ts2 <= n-1)) {
  if(y_trans[(ts2)] == NULL) y_trans[(ts2)] = (double *) buffer_alloc_atomic((ts2), 128 * sizeof(double), buff_mang_y);;
}
/* End of CLooG code */

}

void read_in_copy_y_0(int ts1,int ts2 
#ifndef USE_LOCAL_ARRAYS
 ,__DECLARATION_OF_y 
#endif
 , double** y_trans, BufferManager *buff_mang_y)
{


  int t1, t2;

 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;

/* Start of CLooG code */
if ((ts1 == 0) && (ts2 >= 0) && (128*ts2 <= n-1)) {
  for (t2=128*ts2;t2<=min(n-1,128*ts2+127);t2++) {
    
#ifndef USE_LOCAL_ARRAYS
 y_trans[((1)*( t2 ))/128][((1)*( t2 ))%128] = y[t2]; 
#else
 y_trans[((1)*( t2 ))/128][((1)*( t2 ))%128] = 0; 
#endif
;
  }
}
/* End of CLooG code */

}

void data_tile_ref_count_init_y_1(int ts1,int ts2, double** y_trans, BufferManager *buff_mang_y)
{


  int t1, t2;

 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;

/* Start of CLooG code */
if ((ts1 == 1) && (ts2 >= 0) && (128*ts2 <= n-1)) {
  increment_ref_count(buff_mang_y, (ts2));;
}
/* End of CLooG code */

}

void data_tile_ref_count_init_tmp_1(int ts1,int ts2, double** tmp_trans, BufferManager *buff_mang_tmp)
{


  int t1, t2;

 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;

/* Start of CLooG code */
if ((ts1 == 1) && (ts2 >= 0) && (128*ts2 <= n-1)) {
  increment_ref_count(buff_mang_tmp, (ts2));;
}
/* End of CLooG code */

}

void read_in_data_tile_alloc_tmp_1(int ts1,int ts2, double** tmp_trans, BufferManager *buff_mang_tmp)
{


  int t1, t2;

 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;

/* Start of CLooG code */
if ((ts1 == 1) && (ts2 >= 0) && (128*ts2 <= n-1)) {
  if(tmp_trans[(ts2)] == NULL) tmp_trans[(ts2)] = (double *) buffer_alloc_atomic((ts2), 128 * sizeof(double), buff_mang_tmp);;
}
/* End of CLooG code */

}

void read_in_copy_tmp_1(int ts1,int ts2 
#ifndef USE_LOCAL_ARRAYS
 ,__DECLARATION_OF_tmp 
#endif
 , double** tmp_trans, BufferManager *buff_mang_tmp)
{


  int t1, t2;

 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;

/* Start of CLooG code */
if ((ts1 == 1) && (ts2 >= 0) && (128*ts2 <= n-1)) {
  for (t2=128*ts2;t2<=min(n-1,128*ts2+127);t2++) {
    
#ifndef USE_LOCAL_ARRAYS
 tmp_trans[((1)*( t2 ))/128][((1)*( t2 ))%128] = tmp[t2]; 
#else
 tmp_trans[((1)*( t2 ))/128][((1)*( t2 ))%128] = 0; 
#endif
;
  }
}
/* End of CLooG code */

}

void read_in_data_tile_alloc_y_1(int ts1,int ts2, double** y_trans, BufferManager *buff_mang_y)
{


  int t1, t2;

 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;

/* Start of CLooG code */
if ((ts1 == 1) && (ts2 >= 0) && (128*ts2 <= n-1)) {
  if(y_trans[(ts2)] == NULL) y_trans[(ts2)] = (double *) buffer_alloc_atomic((ts2), 128 * sizeof(double), buff_mang_y);;
}
/* End of CLooG code */

}

void read_in_copy_y_1(int ts1,int ts2 
#ifndef USE_LOCAL_ARRAYS
 ,__DECLARATION_OF_y 
#endif
 , double** y_trans, BufferManager *buff_mang_y)
{


  int t1, t2;

 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;

/* Start of CLooG code */
if ((ts1 == 1) && (ts2 >= 0) && (128*ts2 <= n-1)) {
  for (t2=128*ts2;t2<=min(n-1,128*ts2+127);t2++) {
    
#ifndef USE_LOCAL_ARRAYS
 y_trans[((1)*( t2 ))/128][((1)*( t2 ))%128] = y[t2]; 
#else
 y_trans[((1)*( t2 ))/128][((1)*( t2 ))%128] = 0; 
#endif
;
  }
}
/* End of CLooG code */

}

void data_tile_init_0(int my_rank, int nprocs 
#ifndef USE_LOCAL_ARRAYS
 ,__DECLARATION_OF_tmp,__DECLARATION_OF_A,__DECLARATION_OF_x,__DECLARATION_OF_y,__DECLARATION_OF_B 
#endif
 , double** tmp_trans, BufferManager *buff_mang_tmp, double** A_trans,int A_size_1, BufferManager *buff_mang_A, double** x_trans, BufferManager *buff_mang_x, double** y_trans, BufferManager *buff_mang_y, double** B_trans,int B_size_1, BufferManager *buff_mang_B)
{
  int t1, t2, t3;
 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;
/* Start of CLooG code */
if (n >= 1) {
  for (t2=0;t2<=floord(n-1,128);t2++) {
    for (t3=0;t3<=floord(n-1,128);t3++) {
      B_trans[(t2)* ( B_size_1 ) + (t3)] = NULL;;
    }
  }
  for (t2=0;t2<=floord(n-1,128);t2++) {
    y_trans[(t2)] = NULL;;
  }
  for (t2=0;t2<=floord(n-1,128);t2++) {
    x_trans[(t2)] = NULL;;
  }
  for (t2=0;t2<=floord(n-1,128);t2++) {
    for (t3=0;t3<=floord(n-1,128);t3++) {
      A_trans[(t2)* ( A_size_1 ) + (t3)] = NULL;;
    }
  }
  for (t2=0;t2<=floord(n-1,128);t2++) {
    tmp_trans[(t2)] = NULL;;
  }
  for (t3=0;t3<=floord(n-1,128);t3++) {
    if(pi_0(0,t3,n,nprocs) == my_rank) data_tile_ref_count_init_A_0(0, t3, A_trans, A_size_1,buff_mang_A);;
  }
  for (t3=0;t3<=floord(n-1,128);t3++) {
    if(pi_0(0,t3,n,nprocs) == my_rank) data_tile_ref_count_init_x_0(0, t3, x_trans,buff_mang_x);;
  }
  for (t3=0;t3<=floord(n-1,128);t3++) {
    if(pi_0(0,t3,n,nprocs) == my_rank) data_tile_ref_count_init_tmp_0(0, t3, tmp_trans,buff_mang_tmp);;
  }
  for (t3=0;t3<=floord(n-1,128);t3++) {
    if(pi_0(0,t3,n,nprocs) == my_rank) data_tile_ref_count_init_B_0(0, t3, B_trans, B_size_1,buff_mang_B);;
  }
  for (t3=0;t3<=floord(n-1,128);t3++) {
    if(pi_0(0,t3,n,nprocs) == my_rank) data_tile_ref_count_init_y_0(0, t3, y_trans,buff_mang_y);;
  }
  for (t3=0;t3<=floord(n-1,128);t3++) {
    if(pi_0(0,t3,n,nprocs) == my_rank) data_tile_ref_count_init_tmp_0(0, t3, tmp_trans,buff_mang_tmp);;
  }
  for (t3=0;t3<=floord(n-1,128);t3++) {
    if(pi_0(0,t3,n,nprocs) == my_rank) data_tile_ref_count_init_y_0(0, t3, y_trans,buff_mang_y);;
  }
  for (t3=0;t3<=floord(n-1,128);t3++) {
    if(pi_0(0,t3,n,nprocs) == my_rank){ read_in_data_tile_alloc_A_0(0, t3, A_trans, A_size_1,buff_mang_A);read_in_copy_A_0(0, t3 
#ifndef USE_LOCAL_ARRAYS
, A 
#endif
, A_trans, A_size_1,buff_mang_A);};
  }
  for (t3=0;t3<=floord(n-1,128);t3++) {
    if(pi_0(0,t3,n,nprocs) == my_rank){ read_in_data_tile_alloc_x_0(0, t3, x_trans,buff_mang_x);read_in_copy_x_0(0, t3 
#ifndef USE_LOCAL_ARRAYS
, x 
#endif
, x_trans,buff_mang_x);};
  }
  for (t3=0;t3<=floord(n-1,128);t3++) {
    if(pi_0(0,t3,n,nprocs) == my_rank){ read_in_data_tile_alloc_tmp_0(0, t3, tmp_trans,buff_mang_tmp);read_in_copy_tmp_0(0, t3 
#ifndef USE_LOCAL_ARRAYS
, tmp 
#endif
, tmp_trans,buff_mang_tmp);};
  }
  for (t3=0;t3<=floord(n-1,128);t3++) {
    if(pi_0(0,t3,n,nprocs) == my_rank){ read_in_data_tile_alloc_B_0(0, t3, B_trans, B_size_1,buff_mang_B);read_in_copy_B_0(0, t3 
#ifndef USE_LOCAL_ARRAYS
, B 
#endif
, B_trans, B_size_1,buff_mang_B);};
  }
  for (t3=0;t3<=floord(n-1,128);t3++) {
    if(pi_0(0,t3,n,nprocs) == my_rank){ read_in_data_tile_alloc_y_0(0, t3, y_trans,buff_mang_y);read_in_copy_y_0(0, t3 
#ifndef USE_LOCAL_ARRAYS
, y 
#endif
, y_trans,buff_mang_y);};
  }
  for (t3=0;t3<=floord(n-1,128);t3++) {
    if(pi_1(1,t3,n,nprocs) == my_rank) data_tile_ref_count_init_tmp_1(1, t3, tmp_trans,buff_mang_tmp);;
  }
  for (t3=0;t3<=floord(n-1,128);t3++) {
    if(pi_1(1,t3,n,nprocs) == my_rank) data_tile_ref_count_init_y_1(1, t3, y_trans,buff_mang_y);;
  }
  for (t3=0;t3<=floord(n-1,128);t3++) {
    if(pi_1(1,t3,n,nprocs) == my_rank) data_tile_ref_count_init_y_1(1, t3, y_trans,buff_mang_y);;
  }
  for (t3=0;t3<=floord(n-1,128);t3++) {
    if(pi_1(1,t3,n,nprocs) == my_rank){ read_in_data_tile_alloc_tmp_1(1, t3, tmp_trans,buff_mang_tmp);read_in_copy_tmp_1(1, t3 
#ifndef USE_LOCAL_ARRAYS
, tmp 
#endif
, tmp_trans,buff_mang_tmp);};
  }
  for (t3=0;t3<=floord(n-1,128);t3++) {
    if(pi_1(1,t3,n,nprocs) == my_rank){ read_in_data_tile_alloc_y_1(1, t3, y_trans,buff_mang_y);read_in_copy_y_1(1, t3 
#ifndef USE_LOCAL_ARRAYS
, y 
#endif
, y_trans,buff_mang_y);};
  }
}
/* End of CLooG code */
}
int pack_tmp_1_0(int ts1,int ts2,double *send_buf_tmp,int send_counts_tmp, double** tmp_trans, BufferManager *buff_mang_tmp){
int recv_proc;
  int t1, t2, t3;
 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;
/* Start of CLooG code */
if ((ts1 == 0) && (ts2 >= 0) && (128*ts2 <= n-1)) {
  if(tmp_trans[(ts2)] == NULL) tmp_trans[(ts2)] = (double *) buffer_alloc_atomic((ts2), 128 * sizeof(double), buff_mang_tmp);;
  for (t1=128*ts2+2;t1<=min(n+1,128*ts2+129);t1++) {
    send_buf_tmp[send_counts_tmp++] = tmp_trans[((1)*( t1-2 ))/128][((1)*( t1-2 ))%128];
  }
}
/* End of CLooG code */
return send_counts_tmp;
}
int unpack_tmp_1_0(int ts1,int ts2,double *recv_buf_tmp,int curr_displs_tmp, double** tmp_trans, BufferManager *buff_mang_tmp){
int recv_proc;
  int t1, t2, t3;
 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;
/* Start of CLooG code */
if ((ts1 == 0) && (ts2 >= 0) && (128*ts2 <= n-1)) {
  if(tmp_trans[(ts2)] == NULL) tmp_trans[(ts2)] = (double *) buffer_alloc_atomic((ts2), 128 * sizeof(double), buff_mang_tmp);;
  for (t1=128*ts2+2;t1<=min(n+1,128*ts2+129);t1++) {
    tmp_trans[((1)*( t1-2 ))/128][((1)*( t1-2 ))%128] = recv_buf_tmp[curr_displs_tmp++];
  }
}
/* End of CLooG code */
return curr_displs_tmp;
}
int pack_y_1_0(int ts1,int ts2,double *send_buf_y,int send_counts_y, double** y_trans, BufferManager *buff_mang_y){
int recv_proc;
  int t1, t2, t3;
 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;
/* Start of CLooG code */
if ((ts1 == 0) && (ts2 >= 0) && (128*ts2 <= n-1)) {
  if(y_trans[(ts2)] == NULL) y_trans[(ts2)] = (double *) buffer_alloc_atomic((ts2), 128 * sizeof(double), buff_mang_y);;
  for (t1=128*ts2+2;t1<=min(n+1,128*ts2+129);t1++) {
    send_buf_y[send_counts_y++] = y_trans[((1)*( t1-2 ))/128][((1)*( t1-2 ))%128];
  }
}
/* End of CLooG code */
return send_counts_y;
}
int unpack_y_1_0(int ts1,int ts2,double *recv_buf_y,int curr_displs_y, double** y_trans, BufferManager *buff_mang_y){
int recv_proc;
  int t1, t2, t3;
 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;
/* Start of CLooG code */
if ((ts1 == 0) && (ts2 >= 0) && (128*ts2 <= n-1)) {
  if(y_trans[(ts2)] == NULL) y_trans[(ts2)] = (double *) buffer_alloc_atomic((ts2), 128 * sizeof(double), buff_mang_y);;
  for (t1=128*ts2+2;t1<=min(n+1,128*ts2+129);t1++) {
    y_trans[((1)*( t1-2 ))/128][((1)*( t1-2 ))%128] = recv_buf_y[curr_displs_y++];
  }
}
/* End of CLooG code */
return curr_displs_y;
}
void data_tile_copy_back_tmp_0(int ts1,int ts2 
#ifndef USE_LOCAL_ARRAYS
 ,__DECLARATION_OF_tmp 
#endif
 , double** tmp_trans, BufferManager *buff_mang_tmp)
{
#ifndef USE_LOCAL_ARRAYS
  int t1, t2;
 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;
/* Start of CLooG code */
if ((ts1 == 0) && (ts2 >= 0) && (128*ts2 <= n-1)) {
  for (t2=128*ts2;t2<=min(n-1,128*ts2+127);t2++) {
    tmp[t2] = tmp_trans[((1)*( t2 ))/128][((1)*( t2 ))%128];;
  }
}
/* End of CLooG code */
#endif
}
void data_tile_copy_back_y_0(int ts1,int ts2 
#ifndef USE_LOCAL_ARRAYS
 ,__DECLARATION_OF_y 
#endif
 , double** y_trans, BufferManager *buff_mang_y)
{
#ifndef USE_LOCAL_ARRAYS
  int t1, t2;
 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;
/* Start of CLooG code */
if ((ts1 == 0) && (ts2 >= 0) && (128*ts2 <= n-1)) {
  for (t2=128*ts2;t2<=min(n-1,128*ts2+127);t2++) {
    y[t2] = y_trans[((1)*( t2 ))/128][((1)*( t2 ))%128];;
  }
}
/* End of CLooG code */
#endif
}
int writeout_pack_tmp_0(int ts1,int ts2,double *lw_buf_tmp,int lw_count_tmp, __DECLARATION_OF_tmp, double** tmp_trans, BufferManager *buff_mang_tmp){
int recv_proc;
  int t1, t2, t3;
 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;
/* Start of CLooG code */
if ((ts1 == 0) && (ts2 >= 0) && (128*ts2 <= n-1)) {
  if(tmp_trans[(ts2)] == NULL) tmp_trans[(ts2)] = (double *) buffer_alloc_atomic((ts2), 128 * sizeof(double), buff_mang_tmp);;
  for (t1=128*ts2+2;t1<=min(n+1,128*ts2+129);t1++) {
    lw_buf_tmp[lw_count_tmp++] = tmp_trans[((1)*( t1-2 ))/128][((1)*( t1-2 ))%128];
  }
}
/* End of CLooG code */
return lw_count_tmp;
}
int writeout_unpack_tmp_0(int ts1,int ts2,double *lw_recv_buf_tmp,int displs_lw_tmp, __DECLARATION_OF_tmp, double** tmp_trans, BufferManager *buff_mang_tmp){
int recv_proc;
  int t1, t2;
 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;
/* Start of CLooG code */
if ((ts1 == 0) && (ts2 >= 0) && (128*ts2 <= n-1)) {
  for (t1=128*ts2;t1<=min(n-1,128*ts2+127);t1++) {
    
#ifndef USE_LOCAL_ARRAYS
 tmp[t1] = lw_recv_buf_tmp[displs_lw_tmp++]; 
#endif
;
  }
}
/* End of CLooG code */
return displs_lw_tmp;
}
int writeout_pack_y_0(int ts1,int ts2,double *lw_buf_y,int lw_count_y, __DECLARATION_OF_y, double** y_trans, BufferManager *buff_mang_y){
int recv_proc;
  int t1, t2, t3;
 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;
/* Start of CLooG code */
/* End of CLooG code */
return lw_count_y;
}
int writeout_unpack_y_0(int ts1,int ts2,double *lw_recv_buf_y,int displs_lw_y, __DECLARATION_OF_y, double** y_trans, BufferManager *buff_mang_y){
int recv_proc;
  int t1, t2;
 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;
/* Start of CLooG code */
/* End of CLooG code */
return displs_lw_y;
}
void pack_send_0(int my_rank,int nprocs,int t1,int t2, double** tmp_trans, BufferManager *buff_mang_tmp,double **send_buf_tmp,MPI_Request *send_reqs_tmp,MPI_Status *send_stats_tmp, double** A_trans,int A_size_1, BufferManager *buff_mang_A,double **send_buf_A,MPI_Request *send_reqs_A,MPI_Status *send_stats_A, double** x_trans, BufferManager *buff_mang_x,double **send_buf_x,MPI_Request *send_reqs_x,MPI_Status *send_stats_x, double** y_trans, BufferManager *buff_mang_y,double **send_buf_y,MPI_Request *send_reqs_y,MPI_Status *send_stats_y, double** B_trans,int B_size_1, BufferManager *buff_mang_B,double **send_buf_B,MPI_Request *send_reqs_B,MPI_Status *send_stats_B,int max_num_sends,double *p_total_count)
{
int __p, receiver_list[nprocs];
int send_counts_A[nprocs];
for (__p=0;__p<nprocs;__p++) send_counts_A[__p] = 0;
int index_A[nprocs];
int current_index_A = 0;
int send_counts_x[nprocs];
for (__p=0;__p<nprocs;__p++) send_counts_x[__p] = 0;
int index_x[nprocs];
int current_index_x = 0;
int send_counts_tmp[nprocs];
for (__p=0;__p<nprocs;__p++) send_counts_tmp[__p] = 0;
int index_tmp[nprocs];
int current_index_tmp = 0;
int send_counts_B[nprocs];
for (__p=0;__p<nprocs;__p++) send_counts_B[__p] = 0;
int index_B[nprocs];
int current_index_B = 0;
int send_counts_y[nprocs];
for (__p=0;__p<nprocs;__p++) send_counts_y[__p] = 0;
int index_y[nprocs];
int current_index_y = 0;
int _i;
double __total_count = 0;
;
    for (__p=0; __p<nprocs; __p++) { if(send_counts_A[__p] > 0) {send_buf_A[index_A[__p]][0] = 0;send_buf_A[index_A[__p]][1] = t1;send_buf_A[index_A[__p]][2] = t2; IF_DYNSCHEDULER_DEBUG_PRINT( fprintf(__debug_print_fp, "sending node %d loop 0 task %d %d 		        size %d to node %d\n", my_rank, t1,t2, send_counts_A[__p], __p)); IF_DEBUG_FLUSH(fflush(__debug_print_fp));IF_TIME(__total_count += send_counts_A[__p]); MPI_Isend(send_buf_A[index_A[__p]], send_counts_A[__p], MPI_DOUBLE, __p, 3123, MPI_COMM_WORLD, &send_reqs_A[index_A[__p]]); send_counts_A[__p] = 0;}};
;
    for (__p=0; __p<nprocs; __p++) { if(send_counts_x[__p] > 0) {send_buf_x[index_x[__p]][0] = 0;send_buf_x[index_x[__p]][1] = t1;send_buf_x[index_x[__p]][2] = t2; IF_DYNSCHEDULER_DEBUG_PRINT( fprintf(__debug_print_fp, "sending node %d loop 0 task %d %d 		        size %d to node %d\n", my_rank, t1,t2, send_counts_x[__p], __p)); IF_DEBUG_FLUSH(fflush(__debug_print_fp));IF_TIME(__total_count += send_counts_x[__p]); MPI_Isend(send_buf_x[index_x[__p]], send_counts_x[__p], MPI_DOUBLE, __p, 4123, MPI_COMM_WORLD, &send_reqs_x[index_x[__p]]); send_counts_x[__p] = 0;}};
for (__p=0; __p<nprocs; __p++) { receiver_list[__p] = 0; } sigma_tmp_1_0(t1,t2,n, my_rank, nprocs, receiver_list); for (__p=0; __p<nprocs; __p++) { if (receiver_list[__p] != 0) { if (send_counts_tmp[__p] == 0) { for (;current_index_tmp<max_num_sends;current_index_tmp++) { if (send_reqs_tmp[current_index_tmp] == MPI_REQUEST_NULL) { _i = current_index_tmp; break; } } if (current_index_tmp<max_num_sends) { current_index_tmp++; } else { MPI_Waitany(max_num_sends, send_reqs_tmp, &_i, &send_stats_tmp[__p]); assert(_i != MPI_UNDEFINED); } index_tmp[__p] = _i; send_counts_tmp[__p] = 3; }send_counts_tmp[__p] = pack_tmp_1_0(t1,t2, send_buf_tmp[index_tmp[__p]], send_counts_tmp[__p], tmp_trans,buff_mang_tmp); } };
    for (__p=0; __p<nprocs; __p++) { if(send_counts_tmp[__p] > 0) {send_buf_tmp[index_tmp[__p]][0] = 0;send_buf_tmp[index_tmp[__p]][1] = t1;send_buf_tmp[index_tmp[__p]][2] = t2; IF_DYNSCHEDULER_DEBUG_PRINT( fprintf(__debug_print_fp, "sending node %d loop 0 task %d %d 		        size %d to node %d\n", my_rank, t1,t2, send_counts_tmp[__p], __p)); IF_DEBUG_FLUSH(fflush(__debug_print_fp));IF_TIME(__total_count += send_counts_tmp[__p]); MPI_Isend(send_buf_tmp[index_tmp[__p]], send_counts_tmp[__p], MPI_DOUBLE, __p, 2123, MPI_COMM_WORLD, &send_reqs_tmp[index_tmp[__p]]); send_counts_tmp[__p] = 0;}};
;
    for (__p=0; __p<nprocs; __p++) { if(send_counts_B[__p] > 0) {send_buf_B[index_B[__p]][0] = 0;send_buf_B[index_B[__p]][1] = t1;send_buf_B[index_B[__p]][2] = t2; IF_DYNSCHEDULER_DEBUG_PRINT( fprintf(__debug_print_fp, "sending node %d loop 0 task %d %d 		        size %d to node %d\n", my_rank, t1,t2, send_counts_B[__p], __p)); IF_DEBUG_FLUSH(fflush(__debug_print_fp));IF_TIME(__total_count += send_counts_B[__p]); MPI_Isend(send_buf_B[index_B[__p]], send_counts_B[__p], MPI_DOUBLE, __p, 6123, MPI_COMM_WORLD, &send_reqs_B[index_B[__p]]); send_counts_B[__p] = 0;}};
for (__p=0; __p<nprocs; __p++) { receiver_list[__p] = 0; } sigma_y_1_0(t1,t2,n, my_rank, nprocs, receiver_list); for (__p=0; __p<nprocs; __p++) { if (receiver_list[__p] != 0) { if (send_counts_y[__p] == 0) { for (;current_index_y<max_num_sends;current_index_y++) { if (send_reqs_y[current_index_y] == MPI_REQUEST_NULL) { _i = current_index_y; break; } } if (current_index_y<max_num_sends) { current_index_y++; } else { MPI_Waitany(max_num_sends, send_reqs_y, &_i, &send_stats_y[__p]); assert(_i != MPI_UNDEFINED); } index_y[__p] = _i; send_counts_y[__p] = 3; }send_counts_y[__p] = pack_y_1_0(t1,t2, send_buf_y[index_y[__p]], send_counts_y[__p], y_trans,buff_mang_y); } };
    for (__p=0; __p<nprocs; __p++) { if(send_counts_y[__p] > 0) {send_buf_y[index_y[__p]][0] = 0;send_buf_y[index_y[__p]][1] = t1;send_buf_y[index_y[__p]][2] = t2; IF_DYNSCHEDULER_DEBUG_PRINT( fprintf(__debug_print_fp, "sending node %d loop 0 task %d %d 		        size %d to node %d\n", my_rank, t1,t2, send_counts_y[__p], __p)); IF_DEBUG_FLUSH(fflush(__debug_print_fp));IF_TIME(__total_count += send_counts_y[__p]); MPI_Isend(send_buf_y[index_y[__p]], send_counts_y[__p], MPI_DOUBLE, __p, 5123, MPI_COMM_WORLD, &send_reqs_y[index_y[__p]]); send_counts_y[__p] = 0;}};
*p_total_count += __total_count;
}
void unpack_buf_A_0(int my_rank,int nprocs, double** A_trans,int A_size_1, BufferManager *buff_mang_A,double *recv_buf_A)
{
int curr_displs_A;
curr_displs_A = 3; ;
}
void unpack_buf_x_0(int my_rank,int nprocs, double** x_trans, BufferManager *buff_mang_x,double *recv_buf_x)
{
int curr_displs_x;
curr_displs_x = 3; ;
}
void unpack_buf_tmp_0(int my_rank,int nprocs, double** tmp_trans, BufferManager *buff_mang_tmp,double *recv_buf_tmp)
{
int curr_displs_tmp;
curr_displs_tmp = 3; if (is_receiver_tmp_1_0((int)recv_buf_tmp[1],(int)recv_buf_tmp[2],n,my_rank,nprocs) != 0) { curr_displs_tmp = unpack_tmp_1_0((int)recv_buf_tmp[1],(int)recv_buf_tmp[2],recv_buf_tmp,curr_displs_tmp, tmp_trans,buff_mang_tmp); };
}
void unpack_buf_B_0(int my_rank,int nprocs, double** B_trans,int B_size_1, BufferManager *buff_mang_B,double *recv_buf_B)
{
int curr_displs_B;
curr_displs_B = 3; ;
}
void unpack_buf_y_0(int my_rank,int nprocs, double** y_trans, BufferManager *buff_mang_y,double *recv_buf_y)
{
int curr_displs_y;
curr_displs_y = 3; if (is_receiver_y_1_0((int)recv_buf_y[1],(int)recv_buf_y[2],n,my_rank,nprocs) != 0) { curr_displs_y = unpack_y_1_0((int)recv_buf_y[1],(int)recv_buf_y[2],recv_buf_y,curr_displs_y, y_trans,buff_mang_y); };
}
void data_tile_copy_back_y_1(int ts1,int ts2 
#ifndef USE_LOCAL_ARRAYS
 ,__DECLARATION_OF_y 
#endif
 , double** y_trans, BufferManager *buff_mang_y)
{
#ifndef USE_LOCAL_ARRAYS
  int t1, t2;
 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;
/* Start of CLooG code */
if ((ts1 == 1) && (ts2 >= 0) && (128*ts2 <= n-1)) {
  for (t2=128*ts2;t2<=min(n-1,128*ts2+127);t2++) {
    y[t2] = y_trans[((1)*( t2 ))/128][((1)*( t2 ))%128];;
  }
}
/* End of CLooG code */
#endif
}
int writeout_pack_y_1(int ts1,int ts2,double *lw_buf_y,int lw_count_y, __DECLARATION_OF_y, double** y_trans, BufferManager *buff_mang_y){
int recv_proc;
  int t1, t2, t3;
 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;
/* Start of CLooG code */
if ((ts1 == 1) && (ts2 >= 0) && (128*ts2 <= n-1)) {
  if(y_trans[(ts2)] == NULL) y_trans[(ts2)] = (double *) buffer_alloc_atomic((ts2), 128 * sizeof(double), buff_mang_y);;
  for (t1=128*ts2+2;t1<=min(n+1,128*ts2+129);t1++) {
    lw_buf_y[lw_count_y++] = y_trans[((1)*( t1-2 ))/128][((1)*( t1-2 ))%128];
  }
}
/* End of CLooG code */
return lw_count_y;
}
int writeout_unpack_y_1(int ts1,int ts2,double *lw_recv_buf_y,int displs_lw_y, __DECLARATION_OF_y, double** y_trans, BufferManager *buff_mang_y){
int recv_proc;
  int t1, t2;
 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;
/* Start of CLooG code */
if ((ts1 == 1) && (ts2 >= 0) && (128*ts2 <= n-1)) {
  for (t1=128*ts2;t1<=min(n-1,128*ts2+127);t1++) {
    
#ifndef USE_LOCAL_ARRAYS
 y[t1] = lw_recv_buf_y[displs_lw_y++]; 
#endif
;
  }
}
/* End of CLooG code */
return displs_lw_y;
}
void pack_send_1(int my_rank,int nprocs,int t1,int t2, double** y_trans, BufferManager *buff_mang_y,double **send_buf_y,MPI_Request *send_reqs_y,MPI_Status *send_stats_y, double** tmp_trans, BufferManager *buff_mang_tmp,double **send_buf_tmp,MPI_Request *send_reqs_tmp,MPI_Status *send_stats_tmp,int max_num_sends,double *p_total_count)
{
int __p, receiver_list[nprocs];
int send_counts_tmp[nprocs];
for (__p=0;__p<nprocs;__p++) send_counts_tmp[__p] = 0;
int index_tmp[nprocs];
int current_index_tmp = 0;
int send_counts_y[nprocs];
for (__p=0;__p<nprocs;__p++) send_counts_y[__p] = 0;
int index_y[nprocs];
int current_index_y = 0;
int _i;
double __total_count = 0;
;
    for (__p=0; __p<nprocs; __p++) { if(send_counts_tmp[__p] > 0) {send_buf_tmp[index_tmp[__p]][0] = 1;send_buf_tmp[index_tmp[__p]][1] = t1;send_buf_tmp[index_tmp[__p]][2] = t2; IF_DYNSCHEDULER_DEBUG_PRINT( fprintf(__debug_print_fp, "sending node %d loop 1 task %d %d 		        size %d to node %d\n", my_rank, t1,t2, send_counts_tmp[__p], __p)); IF_DEBUG_FLUSH(fflush(__debug_print_fp));IF_TIME(__total_count += send_counts_tmp[__p]); MPI_Isend(send_buf_tmp[index_tmp[__p]], send_counts_tmp[__p], MPI_DOUBLE, __p, 2123, MPI_COMM_WORLD, &send_reqs_tmp[index_tmp[__p]]); send_counts_tmp[__p] = 0;}};
;
    for (__p=0; __p<nprocs; __p++) { if(send_counts_y[__p] > 0) {send_buf_y[index_y[__p]][0] = 1;send_buf_y[index_y[__p]][1] = t1;send_buf_y[index_y[__p]][2] = t2; IF_DYNSCHEDULER_DEBUG_PRINT( fprintf(__debug_print_fp, "sending node %d loop 1 task %d %d 		        size %d to node %d\n", my_rank, t1,t2, send_counts_y[__p], __p)); IF_DEBUG_FLUSH(fflush(__debug_print_fp));IF_TIME(__total_count += send_counts_y[__p]); MPI_Isend(send_buf_y[index_y[__p]], send_counts_y[__p], MPI_DOUBLE, __p, 5123, MPI_COMM_WORLD, &send_reqs_y[index_y[__p]]); send_counts_y[__p] = 0;}};
*p_total_count += __total_count;
}
void unpack_buf_tmp_1(int my_rank,int nprocs, double** tmp_trans, BufferManager *buff_mang_tmp,double *recv_buf_tmp)
{
int curr_displs_tmp;
curr_displs_tmp = 3; ;
}
void unpack_buf_y_1(int my_rank,int nprocs, double** y_trans, BufferManager *buff_mang_y,double *recv_buf_y)
{
int curr_displs_y;
curr_displs_y = 3; ;
}
void copy_back_writeout_0(int my_rank, int nprocs 
#ifndef USE_LOCAL_ARRAYS
 ,__DECLARATION_OF_tmp,__DECLARATION_OF_y 
#endif
 , double** tmp_trans, BufferManager *buff_mang_tmp, double** y_trans, BufferManager *buff_mang_y)
{
#ifndef USE_LOCAL_ARRAYS
  int t1, t2, t3;
 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;
/* Start of CLooG code */
if (n >= 1) {
  for (t3=0;t3<=floord(n-1,128);t3++) {
    if(pi_0(0,t3,n,nprocs) == my_rank) data_tile_copy_back_tmp_0(0,t3, tmp, tmp_trans,buff_mang_tmp);;
  }
  for (t3=0;t3<=floord(n-1,128);t3++) {
    if(pi_0(0,t3,n,nprocs) == my_rank) data_tile_copy_back_y_0(0,t3, y, y_trans,buff_mang_y);;
  }
  for (t3=0;t3<=floord(n-1,128);t3++) {
    if(pi_1(1,t3,n,nprocs) == my_rank) data_tile_copy_back_y_1(1,t3, y, y_trans,buff_mang_y);;
  }
}
/* End of CLooG code */
#endif
}
void write_out(int my_rank,int nprocs,double *lw_buf_tmp,double *lw_recv_buf_tmp,int *displs_lw_tmp,double *lw_buf_y,double *lw_recv_buf_y,int *displs_lw_y, __DECLARATION_OF_tmp, __DECLARATION_OF_y, double** tmp_trans, BufferManager *buff_mang_tmp, double** y_trans, BufferManager *buff_mang_y)
{
#ifndef USE_LOCAL_ARRAYS
int __p, proc;
int _lb_dist, _ub_dist, lbd_t1, ubd_t1, lbd_t2, ubd_t2, lbd_t3, ubd_t3, lbd_t4, ubd_t4, lbd_t5, ubd_t5, lbd_t6, ubd_t6;
int lw_recv_counts_tmp[nprocs];
int curr_displs_lw_tmp[nprocs];
int lw_count_tmp = 0;
int lw_recv_counts_y[nprocs];
int curr_displs_lw_y[nprocs];
int lw_count_y = 0;
  int t1, t2, t3, t4;
 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;
/* Start of CLooG code */
if (n >= 1) {
  for (t4=0;t4<=floord(n-1,128);t4++) {
    if (pi_0(0,t4,n,nprocs) == my_rank)lw_count_tmp = writeout_pack_tmp_0(0,t4,lw_buf_tmp,lw_count_tmp,tmp, tmp_trans,buff_mang_tmp);;
  }
  assert((nprocs == 1) || (lw_count_tmp <= displs_lw_tmp[1])); MPI_Gather(&lw_count_tmp, 1, MPI_INT, lw_recv_counts_tmp, 1, MPI_INT, 0, MPI_COMM_WORLD);MPI_Gatherv(lw_buf_tmp, lw_count_tmp, MPI_DOUBLE, lw_recv_buf_tmp, lw_recv_counts_tmp, displs_lw_tmp, MPI_DOUBLE, 0, MPI_COMM_WORLD); lw_count_tmp = 0; for (__p=0; __p<nprocs; __p++) curr_displs_lw_tmp[__p] = displs_lw_tmp[__p];;
  if (my_rank == 0) {
    for (t4=0;t4<=floord(n-1,128);t4++) {
      proc = pi_0(0, t4, n, nprocs);if ((my_rank != proc) && (lw_recv_counts_tmp[proc] > 0)) {curr_displs_lw_tmp[proc] = writeout_unpack_tmp_0(0,t4,lw_recv_buf_tmp,curr_displs_lw_tmp[proc],tmp, tmp_trans,buff_mang_tmp);};
    }
  }
  for (t4=0;t4<=floord(n-1,128);t4++) {
    if (pi_0(0,t4,n,nprocs) == my_rank)lw_count_y = writeout_pack_y_0(0,t4,lw_buf_y,lw_count_y,y, y_trans,buff_mang_y);;
  }
  assert((nprocs == 1) || (lw_count_y <= displs_lw_y[1])); MPI_Gather(&lw_count_y, 1, MPI_INT, lw_recv_counts_y, 1, MPI_INT, 0, MPI_COMM_WORLD);MPI_Gatherv(lw_buf_y, lw_count_y, MPI_DOUBLE, lw_recv_buf_y, lw_recv_counts_y, displs_lw_y, MPI_DOUBLE, 0, MPI_COMM_WORLD); lw_count_y = 0; for (__p=0; __p<nprocs; __p++) curr_displs_lw_y[__p] = displs_lw_y[__p];;
  if (my_rank == 0) {
    for (t4=0;t4<=floord(n-1,128);t4++) {
      proc = pi_0(0, t4, n, nprocs);if ((my_rank != proc) && (lw_recv_counts_y[proc] > 0)) {curr_displs_lw_y[proc] = writeout_unpack_y_0(0,t4,lw_recv_buf_y,curr_displs_lw_y[proc],y, y_trans,buff_mang_y);};
    }
  }
  for (t4=0;t4<=floord(n-1,128);t4++) {
    if (pi_1(1,t4,n,nprocs) == my_rank)lw_count_y = writeout_pack_y_1(1,t4,lw_buf_y,lw_count_y,y, y_trans,buff_mang_y);;
  }
  assert((nprocs == 1) || (lw_count_y <= displs_lw_y[1])); MPI_Gather(&lw_count_y, 1, MPI_INT, lw_recv_counts_y, 1, MPI_INT, 0, MPI_COMM_WORLD);MPI_Gatherv(lw_buf_y, lw_count_y, MPI_DOUBLE, lw_recv_buf_y, lw_recv_counts_y, displs_lw_y, MPI_DOUBLE, 0, MPI_COMM_WORLD); lw_count_y = 0; for (__p=0; __p<nprocs; __p++) curr_displs_lw_y[__p] = displs_lw_y[__p];;
  if (my_rank == 0) {
    for (t4=0;t4<=floord(n-1,128);t4++) {
      proc = pi_1(1, t4, n, nprocs);if ((my_rank != proc) && (lw_recv_counts_y[proc] > 0)) {curr_displs_lw_y[proc] = writeout_unpack_y_1(1,t4,lw_recv_buf_y,curr_displs_lw_y[proc],y, y_trans,buff_mang_y);};
    }
  }
}
/* End of CLooG code */
#endif
}
void data_tile_alloc_0(int ts1,int ts2, double** tmp_trans, BufferManager *buff_mang_tmp, double** A_trans,int A_size_1, BufferManager *buff_mang_A, double** x_trans, BufferManager *buff_mang_x, double** y_trans, BufferManager *buff_mang_y, double** B_trans,int B_size_1, BufferManager *buff_mang_B)
{
  int t1, t2, t3;
 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;
/* Start of CLooG code */
if ((ts1 == 0) && (ts2 >= 0) && (128*ts2 <= n-1)) {
  if(tmp_trans[(ts2)] == NULL) tmp_trans[(ts2)] = (double *) buffer_alloc_atomic((ts2), 128 * sizeof(double), buff_mang_tmp);;
  for (t3=0;t3<=floord(n-1,128);t3++) {
    if(A_trans[(ts2) * ((A_size_1)) + (t3)] == NULL) A_trans[(ts2)* ( A_size_1 ) + (t3)] = (double *) buffer_alloc_atomic((ts2)* ( A_size_1 ) + (t3), 128*128 * sizeof(double), buff_mang_A);;
  }
  for (t2=0;t2<=floord(n-1,128);t2++) {
    if(x_trans[(t2)] == NULL) x_trans[(t2)] = (double *) buffer_alloc_atomic((t2), 128 * sizeof(double), buff_mang_x);;
  }
  if(y_trans[(ts2)] == NULL) y_trans[(ts2)] = (double *) buffer_alloc_atomic((ts2), 128 * sizeof(double), buff_mang_y);;
  for (t3=0;t3<=floord(n-1,128);t3++) {
    if(B_trans[(ts2) * ((B_size_1)) + (t3)] == NULL) B_trans[(ts2)* ( B_size_1 ) + (t3)] = (double *) buffer_alloc_atomic((ts2)* ( B_size_1 ) + (t3), 128*128 * sizeof(double), buff_mang_B);;
  }
}
/* End of CLooG code */
}
void data_tile_ref_count_update_0(int ts1,int ts2, double** tmp_trans, BufferManager *buff_mang_tmp, double** A_trans,int A_size_1, BufferManager *buff_mang_A, double** x_trans, BufferManager *buff_mang_x, double** y_trans, BufferManager *buff_mang_y, double** B_trans,int B_size_1, BufferManager *buff_mang_B)
{
  int t1, t2, t3;
 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;
/* Start of CLooG code */
if ((ts1 == 0) && (ts2 >= 0) && (128*ts2 <= n-1)) {
  decrement_ref_count(buff_mang_tmp, (ts2));;
  for (t3=0;t3<=floord(n-1,128);t3++) {
    decrement_ref_count(buff_mang_A, (ts2)* ( A_size_1 ) + (t3));;
  }
  for (t2=0;t2<=floord(n-1,128);t2++) {
    decrement_ref_count(buff_mang_x, (t2));;
  }
  decrement_ref_count(buff_mang_y, (ts2));;
  for (t3=0;t3<=floord(n-1,128);t3++) {
    decrement_ref_count(buff_mang_B, (ts2)* ( B_size_1 ) + (t3));;
  }
}
/* End of CLooG code */
}
void compute_task_0(int t1,int t2, double** tmp_trans, BufferManager *buff_mang_tmp, double** A_trans,int A_size_1, BufferManager *buff_mang_A, double** x_trans, BufferManager *buff_mang_x, double** y_trans, BufferManager *buff_mang_y, double** B_trans,int B_size_1, BufferManager *buff_mang_B)
{
int i_0_div = 0;int i_0_mod = 0;double *tmp_i_0 = NULL;int i_j_0_div = 0;int i_j_0_mod = 0;double *A_i_j_0 = NULL;int j_0_div = 0;int j_0_mod = 0;double *x_j_0 = NULL;int i_1_div = 0;int i_1_mod = 0;double *y_i_1 = NULL;int i_j_1_div = 0;int i_j_1_mod = 0;double *B_i_j_1 = NULL;int j_1_div = 0;int j_1_mod = 0;double *x_j_1 = NULL; int t3, t4, t5, t6;
 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;
/* Start of CLooG code */
for (t3=0;t3<=floord(n-1,128);t3++) {
  for (t4=128*t2;t4<=min(n-1,128*t2+127);t4++) {
    lbv=128*t3;
    ubv=min(n-1,128*t3+127);
    #if (defined __DATA_DIST_DECLS || defined USE_LOCAL_ARRAYS)
    i_0_div = ((1)*( t4 ))/128;i_0_mod = ((1)*( t4 ))%128;tmp_i_0 = tmp_trans[i_0_div];i_j_0_div = ((1)*( t4 ))/128* ( A_size_1 ) + ((1)*( lbv ))/128;i_j_0_mod = ((1)*( t4 ))%128* (128) + ((1)*( lbv ))%128;A_i_j_0 = A_trans[i_j_0_div];j_0_div = ((1)*( lbv ))/128;j_0_mod = ((1)*( lbv ))%128;x_j_0 = x_trans[j_0_div];;
    i_1_div = ((1)*( t4 ))/128;i_1_mod = ((1)*( t4 ))%128;y_i_1 = y_trans[i_1_div];i_j_1_div = ((1)*( t4 ))/128* ( B_size_1 ) + ((1)*( lbv ))/128;i_j_1_mod = ((1)*( t4 ))%128* (128) + ((1)*( lbv ))%128;B_i_j_1 = B_trans[i_j_1_div];j_1_div = ((1)*( lbv ))/128;j_1_mod = ((1)*( lbv ))%128;x_j_1 = x_trans[j_1_div];;
    #endif
#pragma ivdep
#pragma vector always
    for (t5=lbv;t5<=ubv;t5++) {
      tmp_i_0[i_0_mod]=A_i_j_0[i_j_0_mod + t5 - lbv]*x_j_0[j_0_mod + t5 - lbv]+tmp_i_0[i_0_mod];;
      y_i_1[i_1_mod]=B_i_j_1[i_j_1_mod + t5 - lbv]*x_j_1[j_1_mod + t5 - lbv]+y_i_1[i_1_mod];;
    }
  }
}
/* End of CLooG code */
}
void data_tile_alloc_1(int ts1,int ts2, double** y_trans, BufferManager *buff_mang_y, double** tmp_trans, BufferManager *buff_mang_tmp)
{
  int t1, t2;
 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;
/* Start of CLooG code */
if ((ts1 == 1) && (ts2 >= 0) && (128*ts2 <= n-1)) {
  if(y_trans[(ts2)] == NULL) y_trans[(ts2)] = (double *) buffer_alloc_atomic((ts2), 128 * sizeof(double), buff_mang_y);;
  if(tmp_trans[(ts2)] == NULL) tmp_trans[(ts2)] = (double *) buffer_alloc_atomic((ts2), 128 * sizeof(double), buff_mang_tmp);;
}
/* End of CLooG code */
}
void data_tile_ref_count_update_1(int ts1,int ts2, double** y_trans, BufferManager *buff_mang_y, double** tmp_trans, BufferManager *buff_mang_tmp)
{
  int t1, t2;
 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;
/* Start of CLooG code */
if ((ts1 == 1) && (ts2 >= 0) && (128*ts2 <= n-1)) {
  decrement_ref_count(buff_mang_y, (ts2));;
  decrement_ref_count(buff_mang_tmp, (ts2));;
}
/* End of CLooG code */
}
void compute_task_1(int t1,int t2, double** y_trans, BufferManager *buff_mang_y, double** tmp_trans, BufferManager *buff_mang_tmp)
{
int i_0_div = 0;int i_0_mod = 0;double *y_i_0 = NULL;double *tmp_i_0 = NULL; int t3, t4, t5, t6;
 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;
/* Start of CLooG code */
lbv=128*t2;
ubv=min(n-1,128*t2+127);
#if (defined __DATA_DIST_DECLS || defined USE_LOCAL_ARRAYS)
i_0_div = ((1)*( lbv ))/128;i_0_mod = ((1)*( lbv ))%128;y_i_0 = y_trans[i_0_div];tmp_i_0 = tmp_trans[i_0_div];;
#endif
#pragma ivdep
#pragma vector always
for (t6=lbv;t6<=ubv;t6++) {
  y_i_0[i_0_mod + t6 - lbv]=aplha_s*tmp_i_0[i_0_mod + t6 - lbv]+beta*y_i_0[i_0_mod + t6 - lbv];;
}
/* End of CLooG code */
}
