/**
 * bicg.c: This file is part of the PolyBench/C 3.2 test suite.
 *
 *
 * Contact: Louis-Noel Pouchet <pouchet@cse.ohio-state.edu>
 * Web address: http://polybench.sourceforge.net
 */
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <math.h>

/* Include polybench common header. */
#include <polybench.h>

#include "scalapack_util.h"
#define BLOCK_SIZE 100
#define N 30000


/* Include benchmark-specific header. */
/* Default data type is double, default size is 4000. */

#include "gesummv.h"

/* Array initialization. */
static
void init_array(int n,
		DATA_TYPE *alpha,
		DATA_TYPE *beta,
		DATA_TYPE POLYBENCH_2D(A,N,N,n,n),
		DATA_TYPE POLYBENCH_2D(B,N,N,n,n),
		DATA_TYPE POLYBENCH_1D(x,N,n))
{
  int i, j;

  *alpha = 43532;
  *beta = 12313;
  for (i = 0; i < n; i++)
    {
      x[i] = ((DATA_TYPE) i) / n;
      for (j = 0; j < n; j++) {
	A[i][j] = ((DATA_TYPE) i*j) / n;
	B[i][j] = ((DATA_TYPE) i*j) / n;
      }
    }
}


/* DCE code. Must scan the entire live-out data.
   Can be used also to check the correctness of the output. */
static
void print_array(int n,
		 DATA_TYPE POLYBENCH_1D(y,N,n))

{
  int i;

  for (i = 0; i < n; i++) {
    fprintf (stderr, DATA_PRINTF_MODIFIER, y[i]);
    if (i % 20 == 0) fprintf (stderr, "\n");
  }
}

#define __DECLARATION_OF_alpha DATA_TYPE alpha
#define __DECLARATION_OF_beta DATA_TYPE beta
#define __DECLARATION_OF_A DATA_TYPE POLYBENCH_2D(A,N,N,n,n)
#define __DECLARATION_OF_B DATA_TYPE POLYBENCH_2D(B,N,N,n,n)
#define __DECLARATION_OF_tmp DATA_TYPE POLYBENCH_1D(tmp,N,n)
#define __DECLARATION_OF_x DATA_TYPE POLYBENCH_1D(x,N,n)
#define __DECLARATION_OF_y DATA_TYPE POLYBENCH_1D(y,N,n)

#define __PLACE_TO_INSERT_FORWARD_DECLARATIONS

/* Main computational kernel. The whole function will be timed,
   including the call and return. */
static
void kernel_gesummv(int n,
		    DATA_TYPE alpha,
		    DATA_TYPE beta,
		    DATA_TYPE POLYBENCH_2D(A,N,N,n,n),
		    DATA_TYPE POLYBENCH_2D(B,N,N,n,n),
		    DATA_TYPE POLYBENCH_1D(tmp,N,n),
		    DATA_TYPE POLYBENCH_1D(x,N,n),
		    DATA_TYPE POLYBENCH_1D(y,N,n))
{
  int i, j;

#pragma scop
  for (i = 0; i < _PB_N; i++)
    {
      tmp[i] = 0;
      y[i] = 0;
      for (j = 0; j < _PB_N; j++)
	{
	  tmp[i] = A[i][j] * x[j] + tmp[i];
	  y[i] = B[i][j] * x[j] + y[i];
	}
      y[i] = alpha * tmp[i] + beta * y[i];
    }
#pragma endscop

}

/* Main scalapack computational kernel. The whole function will be timed,
   including the call and return. */
static
void scalapack_kernel_gesummv(int nx, int ny,
         int nprocs)
{
  int i, j;

double *A_local, *B_local, *x_local, *y_local, *tmp_local;

MDESC desc_A, desc_B, desc_x, desc_y, desc_tmp;
MDESC desc_A_local, desc_B_local, desc_x_local, desc_y_local, desc_tmp_local;

MKL_INT iam,  ictxt, myrow, mycol, nprow, npcol;
MKL_INT lld, lld_local, info, nx_mkl, ny_mkl, nb, nq, mp, nq_vec, mp_vec;

    int n_procs_row = 1, n_procs_col = 1;

    get_procs_grid_size(nprocs , &n_procs_row, &n_procs_col);

    nprow = (MKL_INT) n_procs_row;
    npcol = (MKL_INT) n_procs_col;

    nx_mkl = (MKL_INT) nx;
    ny_mkl = (MKL_INT) ny;

    nb = (MKL_INT) BLOCK_SIZE;

/*  Init working 2D process grid */
    blacs_get_( &i_negone, &i_zero, &ictxt );
    blacs_gridinit_( &ictxt, "R", &nprow, &npcol );
    blacs_gridinfo_( &ictxt, &nprow, &npcol, &myrow, &mycol );

/*  Compute precise length of local pieces and allocate array on
    each process for parts of distributed vectors */
    mp = numroc_( &nx_mkl, &nb, &myrow, &i_zero, &nprow );
    nq = numroc_( &ny_mkl, &nb, &mycol, &i_zero, &npcol );

    mp_vec = numroc_( &i_one, &i_one, &myrow, &i_zero, &nprow );
    nq_vec = numroc_( &i_one, &i_one, &mycol, &i_zero, &npcol );
    printf("myrow=%d, mycol=%d mp=%d, nq = %d, nq_rhs = %d\n", myrow, mycol, mp, nq, nq_vec);
    if( ( myrow == 0 ) && ( mycol == 0 ) ){
        my_rank = 0;
    }

    A_local = (double*) calloc( mp*nq, sizeof( double ) );
    B_local = (double*) calloc( mp*nq, sizeof( double ) );

/*  Only #0 process column contains distributed 1-d vectors*/
    if ( nq_vec != 0 ){
        x_local = (double *) calloc(mp, sizeof( double ));
        y_local = (double *) calloc(mp, sizeof( double ));
        tmp_local = (double *) calloc(nq, sizeof( double ));
    } else {                                                                                 
        x_local = NULL;
        y_local = NULL;
        tmp_local = NULL;
    }

   /*  Compute leading dimensions */
    lld_local = MAX( numroc_( &nx_mkl, &ny_mkl, &myrow, &i_zero, &nprow ), 1 );
    lld = MAX( mp, 1 );

    /*  Initialize descriptors for initial arrays located on 0 process */
    descinit_( desc_A, &nx_mkl, &ny_mkl, &nx_mkl, &ny_mkl, &i_zero, &i_zero, &ictxt, &lld_local, &info );
    descinit_( desc_B, &nx_mkl, &ny_mkl, &nx_mkl, &ny_mkl, &i_zero, &i_zero, &ictxt, &lld_local, &info );

    descinit_( desc_x, &nx_mkl, &i_one, &nx_mkl, &i_one, &i_zero, &i_zero, &ictxt, &lld_local, &info );
    descinit_( desc_y, &nx_mkl, &i_one, &nx_mkl, &i_one, &i_zero, &i_zero, &ictxt, &lld_local, &info );
    descinit_( desc_tmp, &ny_mkl, &i_one, &ny_mkl, &i_one, &i_zero, &i_zero, &ictxt, &lld_local, &info );

/*  Initialize descriptors for distributed arrays */

    descinit_( desc_A_local, &nx_mkl, &ny_mkl, &nb, &nb, &i_zero, &i_zero, &ictxt, &lld, &info );
    descinit_( desc_B_local, &nx_mkl, &ny_mkl, &nb, &nb, &i_zero, &i_zero, &ictxt, &lld, &info );
    descinit_( desc_x_local, &nx_mkl, &i_one, &nb, &i_one, &i_zero, &i_zero, &ictxt, &lld, &info );
    descinit_( desc_y_local, &nx_mkl, &i_one, &nb, &i_one, &i_zero, &i_zero, &ictxt, &lld, &info );
    descinit_( desc_tmp_local, &ny_mkl, &i_one, &nb, &i_one, &i_zero, &i_zero, &ictxt, &lld, &info );

  DATA_TYPE alpha;
  DATA_TYPE beta;

/*  Distribute vectors and matrices from 0 process over process grid */
#ifdef USE_LOCAL_ARRAYS
        generate_random_values(A_local, mp*nq);
        generate_random_values(B_local, mp*nq);
        generate_random_values(x_local, mp);
        generate_random_values(y_local, mp);
        generate_random_values(tmp_local, nq);
 
#else
      /* Variable declaration/allocation. */
      POLYBENCH_2D_ARRAY_DECL(A, DATA_TYPE, N, N, nx, ny);
      POLYBENCH_2D_ARRAY_DECL(B, DATA_TYPE, N, N, nx, ny);
      POLYBENCH_1D_ARRAY_DECL(x, DATA_TYPE, N, ny);
      POLYBENCH_1D_ARRAY_DECL(y, DATA_TYPE, N, ny);
      POLYBENCH_1D_ARRAY_DECL(tmp, DATA_TYPE, N, nx);

      /* Initialize array(s). */
      init_array (nx, &alpha, &beta,
	      POLYBENCH_ARRAY(A),
	      POLYBENCH_ARRAY(B),
	      POLYBENCH_ARRAY(x));


       pdgeadd_( &trans_N, &nx_mkl, &ny_mkl, &one, 
            A, &i_one, &i_one, desc_A,
            &zero, 
            A_local, &i_one, &i_one, desc_A_local );

       pdgeadd_( &trans_N, &nx_mkl, &ny_mkl, &one, 
            B, &i_one, &i_one, desc_B,
            &zero, 
            B_local, &i_one, &i_one, desc_B_local );

    pdgeadd_( &trans_N, &nx_mkl, &i_one, &one, x, &i_one, 
        &i_one, desc_x, &zero, x_local, &i_one, &i_one, desc_x_local );
    
#endif

  /* Start timer. */
  polybench_start_instruments;

  // y = a * (A * x)
    pdgemv_(&trans_N, &nx_mkl, &ny_mkl, &alpha, 
            A_local, &i_one, &i_one, desc_A_local,
            x_local, &i_one, &i_one, desc_x_local, &i_one,
            &zero,
            y_local, &i_one, &i_one, desc_y_local, &i_one);

        // y = b * (B * x) + y 
    pdgemv_(&trans_T, &nx_mkl, &ny_mkl, &beta, 
            B_local, &i_one, &i_one, desc_B_local,
            x_local, &i_one, &i_one, desc_x_local, &i_one, 
            &one,
            y_local, &i_one, &i_one, desc_y_local, &i_one);

  /* Stop and print timer. */
  polybench_stop_instruments;
  polybench_print_instruments;


#ifndef USE_LOCAL_ARRAYS
    pdgeadd_(&trans_N, &ny_mkl ,&i_one, &one, 
            y_local, &i_one, &i_one, desc_y_local,
            &zero,
            y, &i_one, &i_one, desc_y);

if(my_rank == 0) {
      /* Prevent dead-code elimination. All live-out data must be printed
         by the function call in argument. */
      polybench_prevent_dce(print_array(ny, POLYBENCH_ARRAY(y)));

}

      /* Be clean. */
      POLYBENCH_FREE_ARRAY(A);
      POLYBENCH_FREE_ARRAY(B);
      POLYBENCH_FREE_ARRAY(tmp);
      POLYBENCH_FREE_ARRAY(x);
      POLYBENCH_FREE_ARRAY(y);


#endif

    blacs_gridexit_( &ictxt );
    blacs_exit_( &i_zero );

}
int main(int argc, char** argv)
{
  
  int nx, ny;
  int num_procs = atoi(argv[1]);

#ifndef USE_LOCAL_ARRAYS
      nx = N;
      ny = N;
#else
      nx = sqrt(num_procs) * N;
      ny = sqrt(num_procs) * N;
#endif

  /* Run kernel. */
  scalapack_kernel_gesummv (nx, ny,
           num_procs);

  return 0;
}
