#include "mvt.dist_dynsched_data_dist.h"
#include "polyrt.h"
#include "buffer_manager.h"
#include "pi_defs.h"
#include <assert.h>
#include <omp.h>
#include <mpi.h>
#include <limits.h>
#include <math.h>
#define ceild(n,d)  ceil(((double)(n))/((double)(d)))
#define floord(n,d) floor(((double)(n))/((double)(d)))
#define max(x,y)    ((x) > (y)? (x) : (y))
#define min(x,y)    ((x) < (y)? (x) : (y))

/**
 * mvt.c: This file is part of the PolyBench/C 3.2 test suite.
 *
 *
 * Contact: Louis-Noel Pouchet <pouchet@cse.ohio-state.edu>
 * Web address: http://polybench.sourceforge.net
 */
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <math.h>

/* Include polybench common header. */
#include <polybench.h>

#define N 30000

/* Include benchmark-specific header. */
/* Default data type is double, default size is 4000. */
#include "mvt.h"

int n = N;

/* Array initialization. */
static
void init_array(int n,
		DATA_TYPE POLYBENCH_1D(x1,N,n),
		DATA_TYPE POLYBENCH_1D(x2,N,n),
		DATA_TYPE POLYBENCH_1D(y_1,N,n),
		DATA_TYPE POLYBENCH_1D(y_2,N,n),
		DATA_TYPE POLYBENCH_2D(A,N,N,n,n))
{
  int i, j;

  for (i = 0; i < n; i++)
    {
      x1[i] = ((DATA_TYPE) i) / n;
      x2[i] = ((DATA_TYPE) i + 1) / n;
      y_1[i] = ((DATA_TYPE) i + 3) / n;
      y_2[i] = ((DATA_TYPE) i + 4) / n;
      for (j = 0; j < n; j++)
	A[i][j] = ((DATA_TYPE) i*j) / n;
    }
}


/* DCE code. Must scan the entire live-out data.
   Can be used also to check the correctness of the output. */
static
void print_array(int n,
		 DATA_TYPE POLYBENCH_1D(x1,N,n),
		 DATA_TYPE POLYBENCH_1D(x2,N,n))

{
  int i;

  for (i = 0; i < n; i++) {
    fprintf (stderr, DATA_PRINTF_MODIFIER, x1[i]);
    fprintf (stderr, DATA_PRINTF_MODIFIER, x2[i]);
    if (i % 20 == 0) fprintf (stderr, "\n");
  }
}

#define __DECLARATION_OF_x1 DATA_TYPE POLYBENCH_1D(x1,N,n)
#define __DECLARATION_OF_x2 DATA_TYPE POLYBENCH_1D(x2,N,n)
#define __DECLARATION_OF_y_1 DATA_TYPE POLYBENCH_1D(y_1,N,n)
#define __DECLARATION_OF_y_2 DATA_TYPE POLYBENCH_1D(y_2,N,n)
#define __DECLARATION_OF_A DATA_TYPE POLYBENCH_2D(A,N,N,n,n)

#define __PLACE_TO_INSERT_FORWARD_DECLARATIONS
FILE *__debug_print_fp;
tbb::concurrent_priority_queue<__Task, __Compare> pqueue;
int max_num_tasks_loop0_dim0;
int lb_tasks_loop0_dim0;
int *tasks_loop0;
void read_in_copy_x1_0(int ts1 
#ifndef USE_LOCAL_ARRAYS
 ,__DECLARATION_OF_x1 
#endif
 , double** x1_trans, BufferManager *buff_mang_x1);
void read_in_copy_A_0(int ts1 
#ifndef USE_LOCAL_ARRAYS
 ,__DECLARATION_OF_A 
#endif
 , double** A_trans,int A_size_1, BufferManager *buff_mang_A);
void read_in_copy_y_1_0(int ts1 
#ifndef USE_LOCAL_ARRAYS
 ,__DECLARATION_OF_y_1 
#endif
 , double** y_1_trans, BufferManager *buff_mang_y_1);
void read_in_copy_x2_0(int ts1 
#ifndef USE_LOCAL_ARRAYS
 ,__DECLARATION_OF_x2 
#endif
 , double** x2_trans, BufferManager *buff_mang_x2);
void read_in_copy_y_2_0(int ts1 
#ifndef USE_LOCAL_ARRAYS
 ,__DECLARATION_OF_y_2 
#endif
 , double** y_2_trans, BufferManager *buff_mang_y_2);
void data_tile_init_0(int my_rank, int nprocs 
#ifndef USE_LOCAL_ARRAYS
 ,__DECLARATION_OF_x1,__DECLARATION_OF_A,__DECLARATION_OF_y_1,__DECLARATION_OF_x2,__DECLARATION_OF_y_2 
#endif
 , double** x1_trans, BufferManager *buff_mang_x1, double** A_trans,int A_size_1, BufferManager *buff_mang_A, double** y_1_trans, BufferManager *buff_mang_y_1, double** x2_trans, BufferManager *buff_mang_x2, double** y_2_trans, BufferManager *buff_mang_y_2);
void data_tile_copy_back_x1_0(int ts1 
#ifndef USE_LOCAL_ARRAYS
 ,__DECLARATION_OF_x1 
#endif
 , double** x1_trans, BufferManager *buff_mang_x1);
void data_tile_copy_back_x2_0(int ts1 
#ifndef USE_LOCAL_ARRAYS
 ,__DECLARATION_OF_x2 
#endif
 , double** x2_trans, BufferManager *buff_mang_x2);
int writeout_pack_x1_0(int ts1,double *lw_buf_x1,int lw_count_x1, __DECLARATION_OF_x1, double** x1_trans, BufferManager *buff_mang_x1);
int writeout_unpack_x1_0(int ts1,double *lw_recv_buf_x1,int displs_lw_x1, __DECLARATION_OF_x1, double** x1_trans, BufferManager *buff_mang_x1);
int writeout_pack_x2_0(int ts1,double *lw_buf_x2,int lw_count_x2, __DECLARATION_OF_x2, double** x2_trans, BufferManager *buff_mang_x2);
int writeout_unpack_x2_0(int ts1,double *lw_recv_buf_x2,int displs_lw_x2, __DECLARATION_OF_x2, double** x2_trans, BufferManager *buff_mang_x2);
void copy_back_writeout_0(int my_rank, int nprocs 
#ifndef USE_LOCAL_ARRAYS
 ,__DECLARATION_OF_x1,__DECLARATION_OF_x2 
#endif
 , double** x1_trans, BufferManager *buff_mang_x1, double** x2_trans, BufferManager *buff_mang_x2);
void write_out(int my_rank,int nprocs,double *lw_buf_x1,double *lw_recv_buf_x1,int *displs_lw_x1,double *lw_buf_x2,double *lw_recv_buf_x2,int *displs_lw_x2, __DECLARATION_OF_x1, __DECLARATION_OF_x2, double** x1_trans, BufferManager *buff_mang_x1, double** x2_trans, BufferManager *buff_mang_x2);

/* Main computational kernel. The whole function will be timed,
   including the call and return. */
static
void kernel_mvt(int n
#ifndef USE_LOCAL_ARRAYS
		,DATA_TYPE POLYBENCH_1D(x1,N,n),
		DATA_TYPE POLYBENCH_1D(x2,N,n),
		DATA_TYPE POLYBENCH_1D(y_1,N,n),
		DATA_TYPE POLYBENCH_1D(y_2,N,n),
		DATA_TYPE POLYBENCH_2D(A,N,N,n,n)
#endif
    )
{
  int i, j;

  int t1, t2, t3, t4;
 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;
int _num_threads;
#pragma omp parallel
{_num_threads = omp_get_num_threads();}
 double t_data_init_start, t_data_init= 0.0, t_globaldata_init= 0.0;
#ifndef GLOBAL_MY_RANK
 int my_rank;
#endif
#ifndef __MAX_NUM_RECVS
 #define __MAX_NUM_RECVS (nprocs*_num_threads*4)
#endif
#ifndef __MAX_NUM_SENDS
 #define __MAX_NUM_SENDS (nprocs*_num_threads*4)
#endif
 int nprocs, my_start, my_end, _i, __p, proc, recv_proc, distinct_recv, __tid;
 long int _num_tasks_to_execute, _num_tasks_to_unpack;
 int count;
int req_count;
 int _lb_dist, _ub_dist, lbd_t1, ubd_t1, lbd_t2, ubd_t2, lbd_t3, ubd_t3, lbd_t4, ubd_t4;
 int thread_support_provided;
 MPI_Init_thread(NULL, NULL, MPI_THREAD_MULTIPLE, &thread_support_provided);
 assert(thread_support_provided == MPI_THREAD_MULTIPLE);
 MPI_Comm_rank(MPI_COMM_WORLD, &my_rank);
 MPI_Comm_size(MPI_COMM_WORLD, &nprocs);
 int __tid_receiver;
 int waiting_pops[_num_threads], waiting_pops_threads, waiting_pops_global;
 for (__tid=0; __tid<_num_threads; __tid++) waiting_pops[__tid] = 0;
 double t_tasks_create, t_globaltasks_create;
 double t_tasks_manage_start[_num_threads], t_tasks_manage[_num_threads], t_tasks_manage_threads, t_globaltasks_manage;
 for (__tid=0; __tid<_num_threads; __tid++) t_tasks_manage[__tid] = 0;
 double t_comp_start[_num_threads], t_comp[_num_threads], t_comp_threads_min, t_globalcomp_min, t_comp_threads, t_globalcomp;
 for (__tid=0; __tid<_num_threads; __tid++) t_comp[__tid] = 0;
 double t_wait_comp_start[_num_threads], t_wait_comp[_num_threads], t_wait_comp_threads, t_globalwait_comp;
 for (__tid=0; __tid<_num_threads; __tid++) t_wait_comp[__tid] = 0;
 double t_data_mang_start[_num_threads], t_data_mang[_num_threads], t_data_mang_threads, t_globaldata_mang;
 for (__tid=0; __tid<_num_threads; __tid++) t_data_mang[__tid] = 0;
 double t_pack_start[_num_threads], t_pack[_num_threads], t_pack_threads, t_globalpack;
 for (__tid=0; __tid<_num_threads; __tid++) t_pack[__tid] = 0;
 double t_writeout = 0.0;
 double t_unpack_start, t_unpack = 0.0, t_globalunpack = 0.0;
 double t_wait_unpack_start, t_wait_unpack = 0.0, t_globalwait_unpack = 0.0;
 double t_local = 0.0, t_global = 0.0;
 double __total_count[_num_threads];
 for (__tid=0; __tid<_num_threads; __tid++) __total_count[__tid] = 0;
 double __total_count_threads = 0;
 double __total_count_all = 0;
MPI_Request *send_reqs_x1[_num_threads];
                    for (__tid=0; __tid<_num_threads; __tid++) {
                    send_reqs_x1[__tid] = (MPI_Request *)malloc(__MAX_NUM_SENDS*sizeof(MPI_Request));
                    }
std::vector<MPI_Request> recv_reqs_x1;
MPI_Status *send_stats_x1[_num_threads];
                    for (__tid=0; __tid<_num_threads; __tid++) {
                    send_stats_x1[__tid] = (MPI_Status *)malloc(nprocs*sizeof(MPI_Status));
                    }
MPI_Status *recv_stats_x1 = (MPI_Status *)malloc(sizeof(MPI_Status));
std::vector<int> recv_completed_x1;
                    for (__tid=0; __tid<_num_threads; __tid++) {
                    for (__p=0; __p<__MAX_NUM_SENDS; __p++) {
                    send_reqs_x1[__tid][__p] = MPI_REQUEST_NULL;
}
}
                    for (__p=0; __p<__MAX_NUM_RECVS; __p++) {
                    recv_reqs_x1.push_back(MPI_REQUEST_NULL);
                    recv_completed_x1.push_back(0);
                    }
int lw_count_x1 = 0;
int send_counts_x1[__tid][nprocs];
int max_num_elements_x1 = 0;
int recv_counts_x1[nprocs];
int lw_recv_counts_x1[nprocs];
int displs_x1[nprocs];
int displs_lw_x1[nprocs];
double *send_buf_x1[_num_threads][__MAX_NUM_SENDS];
size_t send_buf_size_x1[_num_threads][__MAX_NUM_SENDS];
for (__tid=0; __tid<_num_threads; __tid++) {
      for (__p=0; __p<__MAX_NUM_SENDS; __p++) {
      send_buf_x1[__tid][__p] = NULL;
      send_buf_size_x1[__tid][__p] = 0;
}
}std::vector<double *> recv_buf_x1;
size_t recv_buf_size_x1[__MAX_NUM_RECVS];
for (__p=0; __p<__MAX_NUM_RECVS; __p++) {
                    recv_buf_x1.push_back(NULL);
                    recv_buf_size_x1[__p] = 0;
}double *lw_buf_x1 = NULL;
size_t lw_buf_size_x1 = 0;
double *lw_recv_buf_x1 = NULL;
size_t lw_recv_buf_size_x1 = 0;
MPI_Request *send_reqs_A[_num_threads];
                    for (__tid=0; __tid<_num_threads; __tid++) {
                    send_reqs_A[__tid] = (MPI_Request *)malloc(__MAX_NUM_SENDS*sizeof(MPI_Request));
                    }
std::vector<MPI_Request> recv_reqs_A;
MPI_Status *send_stats_A[_num_threads];
                    for (__tid=0; __tid<_num_threads; __tid++) {
                    send_stats_A[__tid] = (MPI_Status *)malloc(nprocs*sizeof(MPI_Status));
                    }
MPI_Status *recv_stats_A = (MPI_Status *)malloc(sizeof(MPI_Status));
std::vector<int> recv_completed_A;
                    for (__tid=0; __tid<_num_threads; __tid++) {
                    for (__p=0; __p<__MAX_NUM_SENDS; __p++) {
                    send_reqs_A[__tid][__p] = MPI_REQUEST_NULL;
}
}
                    for (__p=0; __p<__MAX_NUM_RECVS; __p++) {
                    recv_reqs_A.push_back(MPI_REQUEST_NULL);
                    recv_completed_A.push_back(0);
                    }
int lw_count_A = 0;
int send_counts_A[__tid][nprocs];
int max_num_elements_A = 0;
int recv_counts_A[nprocs];
int lw_recv_counts_A[nprocs];
int displs_A[nprocs];
int displs_lw_A[nprocs];
double *send_buf_A[_num_threads][__MAX_NUM_SENDS];
size_t send_buf_size_A[_num_threads][__MAX_NUM_SENDS];
for (__tid=0; __tid<_num_threads; __tid++) {
      for (__p=0; __p<__MAX_NUM_SENDS; __p++) {
      send_buf_A[__tid][__p] = NULL;
      send_buf_size_A[__tid][__p] = 0;
}
}std::vector<double *> recv_buf_A;
size_t recv_buf_size_A[__MAX_NUM_RECVS];
for (__p=0; __p<__MAX_NUM_RECVS; __p++) {
                    recv_buf_A.push_back(NULL);
                    recv_buf_size_A[__p] = 0;
}double *lw_buf_A = NULL;
size_t lw_buf_size_A = 0;
double *lw_recv_buf_A = NULL;
size_t lw_recv_buf_size_A = 0;
MPI_Request *send_reqs_y_1[_num_threads];
                    for (__tid=0; __tid<_num_threads; __tid++) {
                    send_reqs_y_1[__tid] = (MPI_Request *)malloc(__MAX_NUM_SENDS*sizeof(MPI_Request));
                    }
std::vector<MPI_Request> recv_reqs_y_1;
MPI_Status *send_stats_y_1[_num_threads];
                    for (__tid=0; __tid<_num_threads; __tid++) {
                    send_stats_y_1[__tid] = (MPI_Status *)malloc(nprocs*sizeof(MPI_Status));
                    }
MPI_Status *recv_stats_y_1 = (MPI_Status *)malloc(sizeof(MPI_Status));
std::vector<int> recv_completed_y_1;
                    for (__tid=0; __tid<_num_threads; __tid++) {
                    for (__p=0; __p<__MAX_NUM_SENDS; __p++) {
                    send_reqs_y_1[__tid][__p] = MPI_REQUEST_NULL;
}
}
                    for (__p=0; __p<__MAX_NUM_RECVS; __p++) {
                    recv_reqs_y_1.push_back(MPI_REQUEST_NULL);
                    recv_completed_y_1.push_back(0);
                    }
int lw_count_y_1 = 0;
int send_counts_y_1[__tid][nprocs];
int max_num_elements_y_1 = 0;
int recv_counts_y_1[nprocs];
int lw_recv_counts_y_1[nprocs];
int displs_y_1[nprocs];
int displs_lw_y_1[nprocs];
double *send_buf_y_1[_num_threads][__MAX_NUM_SENDS];
size_t send_buf_size_y_1[_num_threads][__MAX_NUM_SENDS];
for (__tid=0; __tid<_num_threads; __tid++) {
      for (__p=0; __p<__MAX_NUM_SENDS; __p++) {
      send_buf_y_1[__tid][__p] = NULL;
      send_buf_size_y_1[__tid][__p] = 0;
}
}std::vector<double *> recv_buf_y_1;
size_t recv_buf_size_y_1[__MAX_NUM_RECVS];
for (__p=0; __p<__MAX_NUM_RECVS; __p++) {
                    recv_buf_y_1.push_back(NULL);
                    recv_buf_size_y_1[__p] = 0;
}double *lw_buf_y_1 = NULL;
size_t lw_buf_size_y_1 = 0;
double *lw_recv_buf_y_1 = NULL;
size_t lw_recv_buf_size_y_1 = 0;
MPI_Request *send_reqs_x2[_num_threads];
                    for (__tid=0; __tid<_num_threads; __tid++) {
                    send_reqs_x2[__tid] = (MPI_Request *)malloc(__MAX_NUM_SENDS*sizeof(MPI_Request));
                    }
std::vector<MPI_Request> recv_reqs_x2;
MPI_Status *send_stats_x2[_num_threads];
                    for (__tid=0; __tid<_num_threads; __tid++) {
                    send_stats_x2[__tid] = (MPI_Status *)malloc(nprocs*sizeof(MPI_Status));
                    }
MPI_Status *recv_stats_x2 = (MPI_Status *)malloc(sizeof(MPI_Status));
std::vector<int> recv_completed_x2;
                    for (__tid=0; __tid<_num_threads; __tid++) {
                    for (__p=0; __p<__MAX_NUM_SENDS; __p++) {
                    send_reqs_x2[__tid][__p] = MPI_REQUEST_NULL;
}
}
                    for (__p=0; __p<__MAX_NUM_RECVS; __p++) {
                    recv_reqs_x2.push_back(MPI_REQUEST_NULL);
                    recv_completed_x2.push_back(0);
                    }
int lw_count_x2 = 0;
int send_counts_x2[__tid][nprocs];
int max_num_elements_x2 = 0;
int recv_counts_x2[nprocs];
int lw_recv_counts_x2[nprocs];
int displs_x2[nprocs];
int displs_lw_x2[nprocs];
double *send_buf_x2[_num_threads][__MAX_NUM_SENDS];
size_t send_buf_size_x2[_num_threads][__MAX_NUM_SENDS];
for (__tid=0; __tid<_num_threads; __tid++) {
      for (__p=0; __p<__MAX_NUM_SENDS; __p++) {
      send_buf_x2[__tid][__p] = NULL;
      send_buf_size_x2[__tid][__p] = 0;
}
}std::vector<double *> recv_buf_x2;
size_t recv_buf_size_x2[__MAX_NUM_RECVS];
for (__p=0; __p<__MAX_NUM_RECVS; __p++) {
                    recv_buf_x2.push_back(NULL);
                    recv_buf_size_x2[__p] = 0;
}double *lw_buf_x2 = NULL;
size_t lw_buf_size_x2 = 0;
double *lw_recv_buf_x2 = NULL;
size_t lw_recv_buf_size_x2 = 0;
MPI_Request *send_reqs_y_2[_num_threads];
                    for (__tid=0; __tid<_num_threads; __tid++) {
                    send_reqs_y_2[__tid] = (MPI_Request *)malloc(__MAX_NUM_SENDS*sizeof(MPI_Request));
                    }
std::vector<MPI_Request> recv_reqs_y_2;
MPI_Status *send_stats_y_2[_num_threads];
                    for (__tid=0; __tid<_num_threads; __tid++) {
                    send_stats_y_2[__tid] = (MPI_Status *)malloc(nprocs*sizeof(MPI_Status));
                    }
MPI_Status *recv_stats_y_2 = (MPI_Status *)malloc(sizeof(MPI_Status));
std::vector<int> recv_completed_y_2;
                    for (__tid=0; __tid<_num_threads; __tid++) {
                    for (__p=0; __p<__MAX_NUM_SENDS; __p++) {
                    send_reqs_y_2[__tid][__p] = MPI_REQUEST_NULL;
}
}
                    for (__p=0; __p<__MAX_NUM_RECVS; __p++) {
                    recv_reqs_y_2.push_back(MPI_REQUEST_NULL);
                    recv_completed_y_2.push_back(0);
                    }
int lw_count_y_2 = 0;
int send_counts_y_2[__tid][nprocs];
int max_num_elements_y_2 = 0;
int recv_counts_y_2[nprocs];
int lw_recv_counts_y_2[nprocs];
int displs_y_2[nprocs];
int displs_lw_y_2[nprocs];
double *send_buf_y_2[_num_threads][__MAX_NUM_SENDS];
size_t send_buf_size_y_2[_num_threads][__MAX_NUM_SENDS];
for (__tid=0; __tid<_num_threads; __tid++) {
      for (__p=0; __p<__MAX_NUM_SENDS; __p++) {
      send_buf_y_2[__tid][__p] = NULL;
      send_buf_size_y_2[__tid][__p] = 0;
}
}std::vector<double *> recv_buf_y_2;
size_t recv_buf_size_y_2[__MAX_NUM_RECVS];
for (__p=0; __p<__MAX_NUM_RECVS; __p++) {
                    recv_buf_y_2.push_back(NULL);
                    recv_buf_size_y_2[__p] = 0;
}double *lw_buf_y_2 = NULL;
size_t lw_buf_size_y_2 = 0;
double *lw_recv_buf_y_2 = NULL;
size_t lw_recv_buf_size_y_2 = 0;
max_num_tasks_loop0_dim0 = ceil(floorf((+1*n-1)/(float)128) - 0 + 1);
lb_tasks_loop0_dim0 = floor(0);
tasks_loop0 = (int *)malloc(max_num_tasks_loop0_dim0*sizeof(int));
int y_2_size_0 = (+1*(floorf((+1*n-1)/(float)128) - 0 + 1));
double **y_2_trans = (double **)malloc((y_2_size_0) * sizeof(double));
BufferManager *buff_mang_y_2 = init_buffer_manager(y_2_size_0);
int x2_size_0 = (+1*(floorf((+1*n-1)/(float)128) - 0 + 1));
double **x2_trans = (double **)malloc((x2_size_0) * sizeof(double));
BufferManager *buff_mang_x2 = init_buffer_manager(x2_size_0);
int y_1_size_0 = (+1*(floorf((+1*n-1)/(float)128) - 0 + 1));
double **y_1_trans = (double **)malloc((y_1_size_0) * sizeof(double));
BufferManager *buff_mang_y_1 = init_buffer_manager(y_1_size_0);
int A_size_0 = (+1*(floorf((+1*n-1)/(float)128) - 0 + 1)+1*(floorf((+1*n-1)/(float)128) - 0 + 1));
int A_size_1 = (+1*(floorf((+1*n-1)/(float)128) - 0 + 1)+1*(floorf((+1*n-1)/(float)128) - 0 + 1));
double **A_trans = (double **)malloc((A_size_0 * A_size_1) * sizeof(double));
BufferManager *buff_mang_A = init_buffer_manager(A_size_0 * A_size_1);
int x1_size_0 = (+1*(floorf((+1*n-1)/(float)128) - 0 + 1));
double **x1_trans = (double **)malloc((x1_size_0) * sizeof(double));
BufferManager *buff_mang_x1 = init_buffer_manager(x1_size_0);
            max_num_elements_x1 = max(max_num_elements_x1, (+1*(1)) + 2);
                for (__tid=0; __tid<_num_threads; __tid++) { for (__p=0; __p<__MAX_NUM_SENDS; __p++) {
                send_buf_x1[__tid][__p] = (double *) polyrt_max_alloc(send_buf_x1[__tid][__p], sizeof(double)*max_num_elements_x1, &send_buf_size_x1[__tid][__p]);
}
}
for (__p=0; __p<__MAX_NUM_RECVS; __p++) { recv_buf_x1[__p] = (double *) polyrt_max_alloc(recv_buf_x1[__p], sizeof(double)*max_num_elements_x1, &recv_buf_size_x1[__p]); }
            max_num_elements_A = max(max_num_elements_A, (+1*(1)*(1)) + 2);
                for (__tid=0; __tid<_num_threads; __tid++) { for (__p=0; __p<__MAX_NUM_SENDS; __p++) {
                send_buf_A[__tid][__p] = (double *) polyrt_max_alloc(send_buf_A[__tid][__p], sizeof(double)*max_num_elements_A, &send_buf_size_A[__tid][__p]);
}
}
for (__p=0; __p<__MAX_NUM_RECVS; __p++) { recv_buf_A[__p] = (double *) polyrt_max_alloc(recv_buf_A[__p], sizeof(double)*max_num_elements_A, &recv_buf_size_A[__p]); }
            max_num_elements_y_1 = max(max_num_elements_y_1, (+1*(1)) + 2);
                for (__tid=0; __tid<_num_threads; __tid++) { for (__p=0; __p<__MAX_NUM_SENDS; __p++) {
                send_buf_y_1[__tid][__p] = (double *) polyrt_max_alloc(send_buf_y_1[__tid][__p], sizeof(double)*max_num_elements_y_1, &send_buf_size_y_1[__tid][__p]);
}
}
for (__p=0; __p<__MAX_NUM_RECVS; __p++) { recv_buf_y_1[__p] = (double *) polyrt_max_alloc(recv_buf_y_1[__p], sizeof(double)*max_num_elements_y_1, &recv_buf_size_y_1[__p]); }
            max_num_elements_x2 = max(max_num_elements_x2, (+1*(1)) + 2);
                for (__tid=0; __tid<_num_threads; __tid++) { for (__p=0; __p<__MAX_NUM_SENDS; __p++) {
                send_buf_x2[__tid][__p] = (double *) polyrt_max_alloc(send_buf_x2[__tid][__p], sizeof(double)*max_num_elements_x2, &send_buf_size_x2[__tid][__p]);
}
}
for (__p=0; __p<__MAX_NUM_RECVS; __p++) { recv_buf_x2[__p] = (double *) polyrt_max_alloc(recv_buf_x2[__p], sizeof(double)*max_num_elements_x2, &recv_buf_size_x2[__p]); }
            max_num_elements_y_2 = max(max_num_elements_y_2, (+1*(1)) + 2);
                for (__tid=0; __tid<_num_threads; __tid++) { for (__p=0; __p<__MAX_NUM_SENDS; __p++) {
                send_buf_y_2[__tid][__p] = (double *) polyrt_max_alloc(send_buf_y_2[__tid][__p], sizeof(double)*max_num_elements_y_2, &send_buf_size_y_2[__tid][__p]);
}
}
for (__p=0; __p<__MAX_NUM_RECVS; __p++) { recv_buf_y_2[__p] = (double *) polyrt_max_alloc(recv_buf_y_2[__p], sizeof(double)*max_num_elements_y_2, &recv_buf_size_y_2[__p]); }
if (fopen(".test", "r")) {lw_buf_x1 = (double *) polyrt_max_alloc(lw_buf_x1, sizeof(double)*((+1*(128))*floorf((1*(floorf((+1*n-1)/(float)128) - 0 + 1+nprocs))/(float)nprocs)), &lw_buf_size_x1);
lw_recv_buf_x1 = (double *) polyrt_max_alloc(lw_recv_buf_x1, nprocs*lw_buf_size_x1, &lw_recv_buf_size_x1);
 for (__p=0; __p<nprocs; __p++) { displs_lw_x1[__p] = __p*lw_buf_size_x1/sizeof(double);}
}if (fopen(".test", "r")) {lw_buf_x2 = (double *) polyrt_max_alloc(lw_buf_x2, sizeof(double)*((+1*(128))*floorf((1*(floorf((+1*n-1)/(float)128) - 0 + 1+nprocs))/(float)nprocs)), &lw_buf_size_x2);
lw_recv_buf_x2 = (double *) polyrt_max_alloc(lw_recv_buf_x2, nprocs*lw_buf_size_x2, &lw_recv_buf_size_x2);
 for (__p=0; __p<nprocs; __p++) { displs_lw_x2[__p] = __p*lw_buf_size_x2/sizeof(double);}
} polyrt_init_grid_size();
 IF_TIME(t_data_init_start = rtclock());
data_tile_init_0(my_rank, nprocs 
#ifndef USE_LOCAL_ARRAYS
 , x1 , A , y_1 , x2 , y_2 
#endif
 , x1_trans, buff_mang_x1, A_trans, A_size_1, buff_mang_A, y_1_trans, buff_mang_y_1, x2_trans, buff_mang_x2, y_2_trans, buff_mang_y_2);
 IF_TIME(t_data_init += rtclock() - t_data_init_start)
 IF_TIME(t_local = rtclock());
/*Start of dynamic scheduling code*/
#if defined(__DYNSCHEDULER_DEBUG_PRINT) || defined(__DYNSCHEDULER_MORE_DEBUG_PRINT) || defined(__DEBUG_FLUSH)
 char *debug_print_name = (char *)malloc(128*sizeof(char));
 strcpy(debug_print_name, "debug_print");
 sprintf(debug_print_name+strlen(debug_print_name), "_node%d", my_rank);
 __debug_print_fp = fopen(debug_print_name, "w");
 free(debug_print_name);
#endif
 IF_TIME(t_tasks_create = rtclock());
 init_tasks(my_rank,nprocs,_num_threads,tasks_loop0,&_num_tasks_to_execute,&_num_tasks_to_unpack);
 IF_TIME(t_tasks_create = rtclock() - t_tasks_create);
#ifdef __DYNSCHEDULER_MORE_DEBUG_PRINT
 std::vector<int> prev_firing_count_x1_0;
 std::vector<int> prev_firing_count_x2_0;
 for (__p=0; __p<__MAX_NUM_RECVS; __p++) {
  prev_firing_count_x1_0.push_back(-1);
  prev_firing_count_x2_0.push_back(-1);
 }
#endif
#pragma omp parallel private(__tid)
{
 __tid = omp_get_thread_num();
 int __flag;
 int __loop;
 int task_id1;
#pragma omp single nowait
{ /* receive thread */
 __tid_receiver = __tid;
 IF_TIME(t_unpack_start = rtclock());
 for (__p=0; __p<__MAX_NUM_RECVS; __p++) {
  MPI_Irecv(recv_buf_x1[__p], max_num_elements_x1, MPI_DOUBLE, MPI_ANY_SOURCE, 2123, MPI_COMM_WORLD, &recv_reqs_x1[__p]);
  MPI_Irecv(recv_buf_x2[__p], max_num_elements_x2, MPI_DOUBLE, MPI_ANY_SOURCE, 5123, MPI_COMM_WORLD, &recv_reqs_x2[__p]);
 }
 IF_TIME(t_unpack += rtclock() - t_unpack_start);
#ifdef __DYNSCHEDULER_DEBUG_PRINT
 fprintf(__debug_print_fp, "node %d recv thread %d num tasks to unpack %ld max recv buffers %d x1 max elements %d x2 max elements %d\n", my_rank, __tid, _num_tasks_to_unpack, __MAX_NUM_RECVS, max_num_elements_x1, max_num_elements_x2);
 IF_DEBUG_FLUSH(fflush(__debug_print_fp));
#endif
 IF_TIME(t_wait_unpack_start = rtclock());
 while(_num_tasks_to_unpack > 0) {
  for (__p=0; (__p<recv_buf_x1.size()) || (__p<recv_buf_x2.size()); __p++) {
  if (__p<recv_buf_x1.size()) {
   if (recv_reqs_x1[__p] != MPI_REQUEST_NULL) {
    MPI_Test(&recv_reqs_x1[__p],&recv_completed_x1[__p],recv_stats_x1);
#ifdef __DYNSCHEDULER_DEBUG_PRINT
    if (recv_completed_x1[__p]) {
     __loop = (int)recv_buf_x1[__p][0];
     task_id1 = (int)recv_buf_x1[__p][1];
     int size;
     MPI_Get_count(recv_stats_x1, MPI_DOUBLE, &size);
     fprintf(__debug_print_fp, "received x1 node %d from node %d buf %d size %d loop %d task %d\n", my_rank, recv_stats_x1->MPI_SOURCE, __p, size, __loop, task_id1);
     IF_DEBUG_FLUSH(fflush(__debug_print_fp));
    }
#endif
   }
   if (recv_completed_x1[__p]) {
    __loop = (int)recv_buf_x1[__p][0];
    task_id1 = (int)recv_buf_x1[__p][1];
    switch(__loop){
     case 0:
#ifdef __DYNSCHEDULER_MORE_DEBUG_PRINT
      if (prev_firing_count_x1_0[__p] != tasks_loop0[(task_id1-lb_tasks_loop0_dim0)+0]) {
       prev_firing_count_x1_0[__p] = tasks_loop0[(task_id1-lb_tasks_loop0_dim0)+0];
       fprintf(__debug_print_fp, "waiting to unpack x1 node %d loop %d task %d count %d\n", my_rank, __loop, task_id1, tasks_loop0[(task_id1-lb_tasks_loop0_dim0)+0]);
       IF_DEBUG_FLUSH(fflush(__debug_print_fp));
      }
#endif
      if (tasks_loop0[(task_id1-lb_tasks_loop0_dim0)+0] <= 2) {
       IF_TIME(t_wait_unpack += rtclock() - t_wait_unpack_start);
       IF_TIME(t_unpack_start = rtclock());
       if (tasks_loop0[(task_id1-lb_tasks_loop0_dim0)+0] == 2) {
        tasks_loop0[(task_id1-lb_tasks_loop0_dim0)+0] -= !is_receiver_x1_0(task_id1,n,my_rank,nprocs);
        tasks_loop0[(task_id1-lb_tasks_loop0_dim0)+0] -= !is_receiver_x2_0(task_id1,n,my_rank,nprocs);
       }
       unpack_buf_x1_0(my_rank,nprocs, x1_trans, buff_mang_x1,recv_buf_x1[__p]);
       tasks_loop0[(task_id1-lb_tasks_loop0_dim0)+0]--;
       recv_completed_x1[__p] = 0;
       MPI_Irecv(recv_buf_x1[__p], max_num_elements_x1, MPI_DOUBLE, MPI_ANY_SOURCE, 2123, MPI_COMM_WORLD, &recv_reqs_x1[__p]);
#ifdef __DYNSCHEDULER_MORE_DEBUG_PRINT
       prev_firing_count_x1_0[__p] = -1;
#endif
#ifdef __DYNSCHEDULER_DEBUG_PRINT
       fprintf(__debug_print_fp, "unpacked x1 node %d loop %d task %d count %d\n", my_rank, __loop, task_id1, tasks_loop0[(task_id1-lb_tasks_loop0_dim0)+0]);
       IF_DEBUG_FLUSH(fflush(__debug_print_fp));
#endif
       if (tasks_loop0[(task_id1-lb_tasks_loop0_dim0)+0] == 0) {
        remote_update_dep_tasks_0(task_id1,n,my_rank,nprocs,_num_threads,tasks_loop0);
        _num_tasks_to_unpack--;
#ifdef __DYNSCHEDULER_DEBUG_PRINT
        fprintf(__debug_print_fp, "remote updated x1 node %d loop %d task %d num tasks left to unpack %ld\n", my_rank, __loop, task_id1, _num_tasks_to_unpack);
        IF_DEBUG_FLUSH(fflush(__debug_print_fp));
#endif
       }
       IF_TIME(t_unpack += rtclock() - t_unpack_start);
       IF_TIME(t_wait_unpack_start = rtclock());
      }
      break;
     default:
      fprintf(stderr, "ERROR: x1 data cannot be sent by a task of loop %d\n", __loop);
      assert(0);
    }
   }
  }
  if (__p<recv_buf_x2.size()) {
   if (recv_reqs_x2[__p] != MPI_REQUEST_NULL) {
    MPI_Test(&recv_reqs_x2[__p],&recv_completed_x2[__p],recv_stats_x2);
#ifdef __DYNSCHEDULER_DEBUG_PRINT
    if (recv_completed_x2[__p]) {
     __loop = (int)recv_buf_x2[__p][0];
     task_id1 = (int)recv_buf_x2[__p][1];
     int size;
     MPI_Get_count(recv_stats_x2, MPI_DOUBLE, &size);
     fprintf(__debug_print_fp, "received x2 node %d from node %d buf %d size %d loop %d task %d\n", my_rank, recv_stats_x2->MPI_SOURCE, __p, size, __loop, task_id1);
     IF_DEBUG_FLUSH(fflush(__debug_print_fp));
    }
#endif
   }
   if (recv_completed_x2[__p]) {
    __loop = (int)recv_buf_x2[__p][0];
    task_id1 = (int)recv_buf_x2[__p][1];
    switch(__loop){
     case 0:
#ifdef __DYNSCHEDULER_MORE_DEBUG_PRINT
      if (prev_firing_count_x2_0[__p] != tasks_loop0[(task_id1-lb_tasks_loop0_dim0)+0]) {
       prev_firing_count_x2_0[__p] = tasks_loop0[(task_id1-lb_tasks_loop0_dim0)+0];
       fprintf(__debug_print_fp, "waiting to unpack x2 node %d loop %d task %d count %d\n", my_rank, __loop, task_id1, tasks_loop0[(task_id1-lb_tasks_loop0_dim0)+0]);
       IF_DEBUG_FLUSH(fflush(__debug_print_fp));
      }
#endif
      if (tasks_loop0[(task_id1-lb_tasks_loop0_dim0)+0] <= 2) {
       IF_TIME(t_wait_unpack += rtclock() - t_wait_unpack_start);
       IF_TIME(t_unpack_start = rtclock());
       if (tasks_loop0[(task_id1-lb_tasks_loop0_dim0)+0] == 2) {
        tasks_loop0[(task_id1-lb_tasks_loop0_dim0)+0] -= !is_receiver_x1_0(task_id1,n,my_rank,nprocs);
        tasks_loop0[(task_id1-lb_tasks_loop0_dim0)+0] -= !is_receiver_x2_0(task_id1,n,my_rank,nprocs);
       }
       unpack_buf_x2_0(my_rank,nprocs, x2_trans, buff_mang_x2,recv_buf_x2[__p]);
       tasks_loop0[(task_id1-lb_tasks_loop0_dim0)+0]--;
       recv_completed_x2[__p] = 0;
       MPI_Irecv(recv_buf_x2[__p], max_num_elements_x2, MPI_DOUBLE, MPI_ANY_SOURCE, 5123, MPI_COMM_WORLD, &recv_reqs_x2[__p]);
#ifdef __DYNSCHEDULER_MORE_DEBUG_PRINT
       prev_firing_count_x2_0[__p] = -1;
#endif
#ifdef __DYNSCHEDULER_DEBUG_PRINT
       fprintf(__debug_print_fp, "unpacked x2 node %d loop %d task %d count %d\n", my_rank, __loop, task_id1, tasks_loop0[(task_id1-lb_tasks_loop0_dim0)+0]);
       IF_DEBUG_FLUSH(fflush(__debug_print_fp));
#endif
       if (tasks_loop0[(task_id1-lb_tasks_loop0_dim0)+0] == 0) {
        remote_update_dep_tasks_0(task_id1,n,my_rank,nprocs,_num_threads,tasks_loop0);
        _num_tasks_to_unpack--;
#ifdef __DYNSCHEDULER_DEBUG_PRINT
        fprintf(__debug_print_fp, "remote updated x2 node %d loop %d task %d num tasks left to unpack %ld\n", my_rank, __loop, task_id1, _num_tasks_to_unpack);
        IF_DEBUG_FLUSH(fflush(__debug_print_fp));
#endif
       }
       IF_TIME(t_unpack += rtclock() - t_unpack_start);
       IF_TIME(t_wait_unpack_start = rtclock());
      }
      break;
     default:
      fprintf(stderr, "ERROR: x2 data cannot be sent by a task of loop %d\n", __loop);
      assert(0);
    }
   }
  }
  }
  for (__p=0; __p<recv_buf_x1.size(); __p++) {
   if (recv_reqs_x1[__p] != MPI_REQUEST_NULL) break;
  }
  if (__p == recv_buf_x1.size()) {
   for (__p=0; __p<nprocs; __p++) {
    double *buf = NULL;
    size_t buf_size = 0;
    buf = (double *) polyrt_max_alloc(buf, sizeof(double)*max_num_elements_x1, &buf_size);
    recv_buf_x1.push_back(buf);
    recv_reqs_x1.push_back(MPI_REQUEST_NULL);
    recv_completed_x1.push_back(0);
#ifdef __DYNSCHEDULER_MORE_DEBUG_PRINT
    prev_firing_count_x1_0.push_back(-1);
#endif
    MPI_Irecv(buf, max_num_elements_x1, MPI_DOUBLE, MPI_ANY_SOURCE, 2123, MPI_COMM_WORLD, &recv_reqs_x1[recv_reqs_x1.size()-1]);
   }
  }
  for (__p=0; __p<recv_buf_x2.size(); __p++) {
   if (recv_reqs_x2[__p] != MPI_REQUEST_NULL) break;
  }
  if (__p == recv_buf_x2.size()) {
   for (__p=0; __p<nprocs; __p++) {
    double *buf = NULL;
    size_t buf_size = 0;
    buf = (double *) polyrt_max_alloc(buf, sizeof(double)*max_num_elements_x2, &buf_size);
    recv_buf_x2.push_back(buf);
    recv_reqs_x2.push_back(MPI_REQUEST_NULL);
    recv_completed_x2.push_back(0);
#ifdef __DYNSCHEDULER_MORE_DEBUG_PRINT
    prev_firing_count_x2_0.push_back(-1);
#endif
    MPI_Irecv(buf, max_num_elements_x2, MPI_DOUBLE, MPI_ANY_SOURCE, 5123, MPI_COMM_WORLD, &recv_reqs_x2[recv_reqs_x2.size()-1]);
   }
  }
 }
 IF_TIME(t_unpack_start = rtclock());
 for (__p=0; __p<recv_buf_x1.size(); __p++) {
  if (recv_reqs_x1[__p] != MPI_REQUEST_NULL) MPI_Request_free(&recv_reqs_x1[__p]);
 }
 for (__p=0; __p<recv_buf_x2.size(); __p++) {
  if (recv_reqs_x2[__p] != MPI_REQUEST_NULL) MPI_Request_free(&recv_reqs_x2[__p]);
 }
 IF_TIME(t_unpack += rtclock() - t_unpack_start);
#ifdef __DYNSCHEDULER_DEBUG_PRINT
 fprintf(__debug_print_fp, "node %d receive done\n", my_rank);
 IF_DEBUG_FLUSH(fflush(__debug_print_fp));
#endif
}
{ /* compute-pack-send threads */
#ifdef __DYNSCHEDULER_DEBUG_PRINT
 fprintf(__debug_print_fp, "node %d compute thread %d num tasks to execute %ld\n", my_rank, __tid, _num_tasks_to_execute);
 IF_DEBUG_FLUSH(fflush(__debug_print_fp));
#endif
 IF_TIME(t_wait_comp_start[__tid] = rtclock());
 while (_num_tasks_to_execute > 0) {
  __Task task;
  __flag = pqueue.try_pop(task);
  if(__flag==0) { IF_TIME(waiting_pops[__tid]++); continue; }
  __loop = task.loop;
  task_id1 = task.id1;
#ifdef __DYNSCHEDULER_DEBUG_PRINT
  long int captured;
#pragma omp atomic capture
  captured = --_num_tasks_to_execute;
  fprintf(__debug_print_fp, "computing node %d thread %d loop %d task %d ready tasks %lu\n", my_rank, __tid, __loop, task_id1, pqueue.size());
  IF_DEBUG_FLUSH(fflush(__debug_print_fp));
#else
#pragma omp atomic
  _num_tasks_to_execute--;
#endif
  switch(__loop){
   case 0:
    IF_TIME(t_wait_comp[__tid] += rtclock() - t_wait_comp_start[__tid]);
    IF_TIME(t_data_mang_start[__tid] = rtclock());
    data_tile_alloc_0(task_id1, x1_trans, buff_mang_x1, A_trans, A_size_1, buff_mang_A, y_1_trans, buff_mang_y_1, x2_trans, buff_mang_x2, y_2_trans, buff_mang_y_2);
    IF_TIME(t_data_mang[__tid] += rtclock() - t_data_mang_start[__tid]);
    IF_TIME(t_comp_start[__tid] = rtclock());
    compute_task_0(task_id1, x1_trans, buff_mang_x1, A_trans, A_size_1, buff_mang_A, y_1_trans, buff_mang_y_1, x2_trans, buff_mang_x2, y_2_trans, buff_mang_y_2);
    IF_TIME(t_comp[__tid] += rtclock() - t_comp_start[__tid]);
    IF_TIME(t_data_mang_start[__tid] = rtclock());
    data_tile_ref_count_update_0(task_id1, x1_trans, buff_mang_x1, A_trans, A_size_1, buff_mang_A, y_1_trans, buff_mang_y_1, x2_trans, buff_mang_x2, y_2_trans, buff_mang_y_2);
    IF_TIME(t_data_mang[__tid] += rtclock() - t_data_mang_start[__tid]);
    IF_TIME(t_pack_start[__tid] = rtclock());
    pack_send_0(my_rank,nprocs,task_id1, x1_trans, buff_mang_x1,send_buf_x1[__tid],send_reqs_x1[__tid],send_stats_x1[__tid], A_trans, A_size_1, buff_mang_A,send_buf_A[__tid],send_reqs_A[__tid],send_stats_A[__tid], y_1_trans, buff_mang_y_1,send_buf_y_1[__tid],send_reqs_y_1[__tid],send_stats_y_1[__tid], x2_trans, buff_mang_x2,send_buf_x2[__tid],send_reqs_x2[__tid],send_stats_x2[__tid], y_2_trans, buff_mang_y_2,send_buf_y_2[__tid],send_reqs_y_2[__tid],send_stats_y_2[__tid],__MAX_NUM_SENDS,&__total_count[__tid]);
    IF_TIME(t_pack[__tid] += rtclock() - t_pack_start[__tid]);
    IF_TIME(t_tasks_manage_start[__tid] = rtclock());
    local_update_dep_tasks_0(task_id1,n,my_rank,nprocs,_num_threads,tasks_loop0);
    IF_TIME(t_tasks_manage[__tid] += rtclock() - t_tasks_manage_start[__tid]);
    IF_TIME(t_wait_comp_start[__tid] = rtclock());
    break;
   default:
    assert(0);
  }
#ifdef __DYNSCHEDULER_MORE_DEBUG_PRINT
  fprintf(__debug_print_fp, "computed node %d thread %d loop %d task %d tasks left %ld\n", my_rank, __tid, __loop, task_id1, captured);
  IF_DEBUG_FLUSH(fflush(__debug_print_fp));
#endif
 }
#ifdef __DYNSCHEDULER_DEBUG_PRINT
 fprintf(__debug_print_fp, "node %d compute thread %d done\n", my_rank, __tid);
 IF_DEBUG_FLUSH(fflush(__debug_print_fp));
#endif
}
}
#ifdef __DYNSCHEDULER_DEBUG_PRINT
 fclose(__debug_print_fp);
#endif
/*End of dynamic scheduling code*/
 IF_TIME(t_writeout = rtclock());
 
#ifndef USE_LOCAL_ARRAYS
 if (fopen(".test", "r")) {
 copy_back_writeout_0(my_rank, nprocs 
#ifndef USE_LOCAL_ARRAYS
 , x1 , x2 
#endif
 , x1_trans, buff_mang_x1, x2_trans, buff_mang_x2);
write_out(my_rank,nprocs,lw_buf_x1,lw_recv_buf_x1,displs_lw_x1,lw_buf_x2,lw_recv_buf_x2,displs_lw_x2,x1,x2, x1_trans, buff_mang_x1, x2_trans, buff_mang_x2);
} 
#endif
 IF_TIME(t_writeout = rtclock() - t_writeout);
free_buffer_mang(buff_mang_x1);
free_buffer_mang(buff_mang_A);
free_buffer_mang(buff_mang_y_1);
free_buffer_mang(buff_mang_x2);
free_buffer_mang(buff_mang_y_2);
#ifdef TIME
 IF_TIME(t_local = rtclock() - t_local - t_writeout);
 for (__tid=0; __tid<_num_threads; __tid++) __total_count_threads += __total_count[__tid];
 waiting_pops_threads = waiting_pops[0];
                    for (__tid=1; __tid<_num_threads; __tid++) if (waiting_pops_threads < waiting_pops[__tid]) waiting_pops_threads = waiting_pops[__tid];
 t_tasks_manage_threads = t_tasks_manage[0];
                    for (__tid=1; __tid<_num_threads; __tid++) if (t_tasks_manage_threads < t_tasks_manage[__tid]) t_tasks_manage_threads = t_tasks_manage[__tid];
 if (__tid_receiver != 0) t_comp_threads_min = t_comp[0];
                    else t_comp_threads_min = t_comp[1];
                    for (__tid=1; __tid<_num_threads; __tid++) if ((__tid != __tid_receiver) && (t_comp_threads_min > t_comp[__tid])) t_comp_threads_min = t_comp[__tid];
 t_comp_threads = t_comp[0];
                    for (__tid=1; __tid<_num_threads; __tid++) if (t_comp_threads < t_comp[__tid]) t_comp_threads = t_comp[__tid];
 t_wait_comp_threads = t_wait_comp[0];
                    for (__tid=1; __tid<_num_threads; __tid++) if (t_wait_comp_threads < t_wait_comp[__tid]) t_wait_comp_threads = t_wait_comp[__tid];
 t_data_mang_threads = t_data_mang[0];
      for (__tid=1; __tid<_num_threads; __tid++) if (t_data_mang_threads < t_data_mang[__tid]) t_data_mang_threads = t_data_mang[__tid];
 t_pack_threads = t_pack[0];
                    for (__tid=1; __tid<_num_threads; __tid++) if (t_pack_threads < t_pack[__tid]) t_pack_threads = t_pack[__tid];
 char *buffer = (char *)malloc(4096 * sizeof(char));
 strcpy(buffer, "");
#ifdef __DETAILED_TIME_REPORT
 sprintf(buffer+strlen(buffer), "Node %d\n", my_rank);
 sprintf(buffer+strlen(buffer), "Write-out time: %lf\n", t_writeout);
 sprintf(buffer+strlen(buffer), "Values communicated: %0.0lf\n", __total_count_threads);
 sprintf(buffer+strlen(buffer), "Data init time: %lf\n", t_data_init);
 sprintf(buffer+strlen(buffer), "Minimum computation time across all threads: %lf\n", t_comp_threads_min);
 sprintf(buffer+strlen(buffer), "Maximum computation time across all threads: %lf\n", t_comp_threads);
 sprintf(buffer+strlen(buffer), "'Waiting for computation' time: %lf\n", t_wait_comp_threads);
 sprintf(buffer+strlen(buffer), "Data mang time: %lf\n", t_data_mang_threads);
 sprintf(buffer+strlen(buffer), "Tasks creation time: %lf\n", t_tasks_create);
 sprintf(buffer+strlen(buffer), "Tasks management time: %lf\n", t_tasks_manage_threads);
 sprintf(buffer+strlen(buffer), "Packing time: %lf\n", t_pack_threads);
 sprintf(buffer+strlen(buffer), "Unpacking time: %lf\n", t_unpack);
 sprintf(buffer+strlen(buffer), "'Waiting for unpacking' time: %lf\n", t_wait_unpack);
 sprintf(buffer+strlen(buffer), "Sum of split times: %lf\n", t_comp_threads + t_wait_comp_threads + t_tasks_create + t_tasks_manage_threads + t_pack_threads);
 sprintf(buffer+strlen(buffer), "Total time minus write-out time: %lf\n", t_local);
 sprintf(buffer+strlen(buffer), "-------");
 fprintf(stdout, "%s\n", buffer);
#endif
 MPI_Reduce(&__total_count_threads, &__total_count_all, 1, MPI_DOUBLE, MPI_SUM, 0, MPI_COMM_WORLD);
 MPI_Reduce(&waiting_pops_threads, &waiting_pops_global, 1, MPI_INT, MPI_MAX, 0, MPI_COMM_WORLD);
 MPI_Reduce(&t_comp_threads_min, &t_globalcomp_min, 1, MPI_DOUBLE, MPI_MIN, 0, MPI_COMM_WORLD);
 MPI_Reduce(&t_comp_threads, &t_globalcomp, 1, MPI_DOUBLE, MPI_MAX, 0, MPI_COMM_WORLD);
 MPI_Reduce(&t_wait_comp_threads, &t_globalwait_comp, 1, MPI_DOUBLE, MPI_MAX, 0, MPI_COMM_WORLD);
 MPI_Reduce(&t_tasks_create, &t_globaltasks_create, 1, MPI_DOUBLE, MPI_MAX, 0, MPI_COMM_WORLD);
 MPI_Reduce(&t_tasks_manage_threads, &t_globaltasks_manage, 1, MPI_DOUBLE, MPI_MAX, 0, MPI_COMM_WORLD);
 MPI_Reduce(&t_pack_threads, &t_globalpack, 1, MPI_DOUBLE, MPI_MAX, 0, MPI_COMM_WORLD);
 MPI_Reduce(&t_unpack, &t_globalunpack, 1, MPI_DOUBLE, MPI_MAX, 0, MPI_COMM_WORLD);
 MPI_Reduce(&t_wait_unpack, &t_globalwait_unpack, 1, MPI_DOUBLE, MPI_MAX, 0, MPI_COMM_WORLD);
 MPI_Reduce(&t_local, &t_global, 1, MPI_DOUBLE, MPI_MAX, 0, MPI_COMM_WORLD);
 if (my_rank==0) {
  strcpy(buffer, "SUMMARY\n");
  sprintf(buffer+strlen(buffer), "Write-out time spent in master node: %0.6lf s\n", t_writeout);
  sprintf(buffer+strlen(buffer), "Total communication volume across all nodes: %0.6lf GB\n", __total_count_all*8/(1024*1024*1024));
  sprintf(buffer+strlen(buffer), "Maximum computation time spent across all nodes: %0.6lf s\n", t_globalcomp);
  sprintf(buffer+strlen(buffer), "Minimum computation time spent across all nodes: %0.6lf s\n", t_globalcomp_min);
  sprintf(buffer+strlen(buffer), "Maximum 'waiting for computation' time spent across all nodes: %0.6lf s\n", t_globalwait_comp);
  sprintf(buffer+strlen(buffer), "Maximum number of waiting/unsuccessful POPs across all nodes: %d\n", waiting_pops_global);
  sprintf(buffer+strlen(buffer), "Maximum tasks creation time spent across all nodes: %0.6lf s\n", t_globaltasks_create);
  sprintf(buffer+strlen(buffer), "Maximum tasks management time spent across all nodes: %0.6lf s\n", t_globaltasks_manage);
  sprintf(buffer+strlen(buffer), "Maximum packing time spent across all nodes: %0.6lf s\n", t_globalpack);
  sprintf(buffer+strlen(buffer), "Maximum unpacking time spent across all nodes: %0.6lf s\n", t_globalunpack);
  sprintf(buffer+strlen(buffer), "Maximum 'waiting for unpacking' time spent across all nodes: %0.6lf s\n", t_globalwait_unpack);
  sprintf(buffer+strlen(buffer), "Maximum total time spent across all nodes: %0.6lf s\n", t_global);
  sprintf(buffer+strlen(buffer), "Maximum data init time spent across all nodes: %0.6lf s\n", t_globaldata_init);
  sprintf(buffer+strlen(buffer), "-------");
  fprintf(stdout, "%s\n", buffer);
 }
#endif
 MPI_Barrier(MPI_COMM_WORLD);
 MPI_Finalize();

}


int main(int argc, char** argv)
{
  /* Retrieve problem size. */

  if(argc == 3){
      int np = atoi(argv[2]);
      int nprocs = atoi(argv[1]);
      n = np * sqrt(nprocs);
  }

#ifndef USE_LOCAL_ARRAYS
  /* Variable declaration/allocation. */
  POLYBENCH_2D_ARRAY_DECL(A, DATA_TYPE, N, N, n, n);
  POLYBENCH_1D_ARRAY_DECL(x1, DATA_TYPE, N, n);
  POLYBENCH_1D_ARRAY_DECL(x2, DATA_TYPE, N, n);
  POLYBENCH_1D_ARRAY_DECL(y_1, DATA_TYPE, N, n);
  POLYBENCH_1D_ARRAY_DECL(y_2, DATA_TYPE, N, n);


  /* Initialize array(s). */
  init_array (n,
	      POLYBENCH_ARRAY(x1),
	      POLYBENCH_ARRAY(x2),
	      POLYBENCH_ARRAY(y_1),
	      POLYBENCH_ARRAY(y_2),
	      POLYBENCH_ARRAY(A));

  /* Start timer. */
  polybench_start_instruments;

  /* Run kernel. */
  kernel_mvt (n,
	      POLYBENCH_ARRAY(x1),
	      POLYBENCH_ARRAY(x2),
	      POLYBENCH_ARRAY(y_1),
	      POLYBENCH_ARRAY(y_2),
	      POLYBENCH_ARRAY(A));

  /* Stop and print timer. */
  polybench_stop_instruments;
  polybench_print_instruments;

  /* Prevent dead-code elimination. All live-out data must be printed
     by the function call in argument. */
  polybench_prevent_dce(print_array(n, POLYBENCH_ARRAY(x1), POLYBENCH_ARRAY(x2)));

  /* Be clean. */
  POLYBENCH_FREE_ARRAY(A);
  POLYBENCH_FREE_ARRAY(x1);
  POLYBENCH_FREE_ARRAY(x2);
  POLYBENCH_FREE_ARRAY(y_1);
  POLYBENCH_FREE_ARRAY(y_2);
#else

  /* Run kernel. */
  kernel_mvt (n);

#endif

  return 0;
}


void init_tasks(int my_rank,int nprocs,int num_threads,int *tasks_loop0,long int *p_num_tasks_to_execute,long int *p_num_tasks_to_unpack)
{
int remote_dep_tasks, local_dep_tasks, affinity;
long int _num_tasks_to_execute = 0;
long int _num_tasks_to_unpack = 0;


  int t1, t2;

 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;

/* Start of CLooG code */
if (n >= 1) {
  lbp=0;
  ubp=floord(n-1,128);
#pragma omp parallel for private(remote_dep_tasks, local_dep_tasks, affinity) reduction(+:_num_tasks_to_execute,_num_tasks_to_unpack)
  for (t2=lbp;t2<=ubp;t2++) {
    if (pi_0(t2,n,nprocs) == my_rank) { _num_tasks_to_execute++;tasks_loop0[(t2-lb_tasks_loop0_dim0)+0] = get_num_remote_src_tasks_0(t2,n,my_rank,nprocs) + get_num_local_src_tasks_0(t2,n,my_rank,nprocs);if (tasks_loop0[(t2-lb_tasks_loop0_dim0)+0] == 0) { remote_dep_tasks = get_num_remote_dep_tasks_0(t2,n,my_rank,nprocs); local_dep_tasks = get_num_local_dep_tasks_0(t2,n,my_rank,nprocs); if (__is_multi_partitioned[0] || __is_block_cyclic[0]) affinity = -1; else affinity = pi_threads_0(t2,n,my_rank,nprocs,num_threads); __Task task(0,t2,remote_dep_tasks,local_dep_tasks,affinity); pqueue.push(task); IF_DYNSCHEDULER_MORE_DEBUG_PRINT( fprintf(__debug_print_fp, "added node %d loop 0 task %d                 remote_dep_tasks %d local_dep_tasks %d ready tasks %lu\n", my_rank, t2, remote_dep_tasks, local_dep_tasks, pqueue.size())); IF_DEBUG_FLUSH(fflush(__debug_print_fp)); }} else if (is_receiver_0(t2,n,my_rank,nprocs)) { _num_tasks_to_unpack++;tasks_loop0[(t2-lb_tasks_loop0_dim0)+0] = 2 + get_num_local_src_tasks_0(t2,n,my_rank,nprocs);};
  }
}
/* End of CLooG code */

*p_num_tasks_to_execute = _num_tasks_to_execute;
*p_num_tasks_to_unpack = _num_tasks_to_unpack;
}

void data_tile_ref_count_init_x1_0(int ts1, double** x1_trans, BufferManager *buff_mang_x1)
{


  int t1, t2;

 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;

/* Start of CLooG code */
if ((ts1 >= 0) && (128*ts1 <= n-1)) {
  increment_ref_count(buff_mang_x1, (ts1));;
}
/* End of CLooG code */

}

void data_tile_ref_count_init_A_0(int ts1, double** A_trans,int A_size_1, BufferManager *buff_mang_A)
{


  int t1, t2, t3;

 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;

/* Start of CLooG code */
if ((ts1 >= 0) && (128*ts1 <= n-1)) {
  for (t2=0;t2<=ts1-1;t2++) {
    increment_ref_count(buff_mang_A, (t2)* ( A_size_1 ) + (ts1));;
  }
  for (t3=0;t3<=floord(n-1,128);t3++) {
    increment_ref_count(buff_mang_A, (ts1)* ( A_size_1 ) + (t3));;
  }
  for (t2=ts1+1;t2<=floord(n-1,128);t2++) {
    increment_ref_count(buff_mang_A, (t2)* ( A_size_1 ) + (ts1));;
  }
}
/* End of CLooG code */

}

void data_tile_ref_count_init_y_1_0(int ts1, double** y_1_trans, BufferManager *buff_mang_y_1)
{


  int t1, t2;

 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;

/* Start of CLooG code */
if ((ts1 >= 0) && (128*ts1 <= n-1)) {
  for (t2=0;t2<=floord(n-1,128);t2++) {
    increment_ref_count(buff_mang_y_1, (t2));;
  }
}
/* End of CLooG code */

}

void data_tile_ref_count_init_x2_0(int ts1, double** x2_trans, BufferManager *buff_mang_x2)
{


  int t1, t2;

 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;

/* Start of CLooG code */
if ((ts1 >= 0) && (128*ts1 <= n-1)) {
  increment_ref_count(buff_mang_x2, (ts1));;
}
/* End of CLooG code */

}

void data_tile_ref_count_init_y_2_0(int ts1, double** y_2_trans, BufferManager *buff_mang_y_2)
{


  int t1, t2;

 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;

/* Start of CLooG code */
if ((ts1 >= 0) && (128*ts1 <= n-1)) {
  for (t2=0;t2<=floord(n-1,128);t2++) {
    increment_ref_count(buff_mang_y_2, (t2));;
  }
}
/* End of CLooG code */

}

void read_in_data_tile_alloc_x1_0(int ts1, double** x1_trans, BufferManager *buff_mang_x1)
{


  int t1, t2;

 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;

/* Start of CLooG code */
if ((ts1 >= 0) && (128*ts1 <= n-1)) {
  if(x1_trans[(ts1)] == NULL) x1_trans[(ts1)] = (double *) buffer_alloc_atomic((ts1), 128 * sizeof(double), buff_mang_x1);;
}
/* End of CLooG code */

}

void read_in_copy_x1_0(int ts1 
#ifndef USE_LOCAL_ARRAYS
 ,__DECLARATION_OF_x1 
#endif
 , double** x1_trans, BufferManager *buff_mang_x1)
{


  int t1, t2;

 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;

/* Start of CLooG code */
if ((ts1 >= 0) && (128*ts1 <= n-1)) {
  for (t2=128*ts1;t2<=min(n-1,128*ts1+127);t2++) {
    
#ifndef USE_LOCAL_ARRAYS
 x1_trans[((1)*( t2 ))/128][((1)*( t2 ))%128] = x1[t2]; 
#else
 x1_trans[((1)*( t2 ))/128][((1)*( t2 ))%128] = ((DATA_TYPE) t2) / n; 
#endif
;
  }
}
/* End of CLooG code */

}

void read_in_data_tile_alloc_A_0(int ts1, double** A_trans,int A_size_1, BufferManager *buff_mang_A)
{


  int t1, t2, t3;

 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;

/* Start of CLooG code */
if ((ts1 >= 0) && (128*ts1 <= n-1)) {
  for (t2=0;t2<=ts1-1;t2++) {
    if(A_trans[(t2) * ((A_size_1)) + (ts1)] == NULL) A_trans[(t2)* ( A_size_1 ) + (ts1)] = (double *) buffer_alloc_atomic((t2)* ( A_size_1 ) + (ts1), 128*128 * sizeof(double), buff_mang_A);;
  }
  for (t3=0;t3<=floord(n-1,128);t3++) {
    if(A_trans[(ts1) * ((A_size_1)) + (t3)] == NULL) A_trans[(ts1)* ( A_size_1 ) + (t3)] = (double *) buffer_alloc_atomic((ts1)* ( A_size_1 ) + (t3), 128*128 * sizeof(double), buff_mang_A);;
  }
  for (t2=ts1+1;t2<=floord(n-1,128);t2++) {
    if(A_trans[(t2) * ((A_size_1)) + (ts1)] == NULL) A_trans[(t2)* ( A_size_1 ) + (ts1)] = (double *) buffer_alloc_atomic((t2)* ( A_size_1 ) + (ts1), 128*128 * sizeof(double), buff_mang_A);;
  }
}
/* End of CLooG code */

}

void read_in_copy_A_0(int ts1 
#ifndef USE_LOCAL_ARRAYS
 ,__DECLARATION_OF_A 
#endif
 , double** A_trans,int A_size_1, BufferManager *buff_mang_A)
{


  int t1, t2, t3;

 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;

/* Start of CLooG code */
if ((ts1 >= 0) && (128*ts1 <= n-1)) {
  for (t2=0;t2<=128*ts1-1;t2++) {
    for (t3=128*ts1;t3<=min(n-1,128*ts1+127);t3++) {
      
#ifndef USE_LOCAL_ARRAYS
 A_trans[((1)*( t3 ))/128* ( A_size_1 ) + ((1)*( t2 ))/128][((1)*( t3 ))%128* (128) + ((1)*( t2 ))%128] = A[t2][t3]; 
#else
 A_trans[((1)*( t3 ))/128* ( A_size_1 ) + ((1)*( t2 ))/128][((1)*( t3 ))%128* (128) + ((1)*( t2 ))%128] = ((DATA_TYPE) t2*t3) / n; 
#endif
;
    }
  }
  if ((n >= 129) && (ts1 == 0)) {
    for (t2=0;t2<=127;t2++) {
      for (t3=0;t3<=127;t3++) {
        
#ifndef USE_LOCAL_ARRAYS
 A_trans[((1)*( t3 ))/128* ( A_size_1 ) + ((1)*( t2 ))/128][((1)*( t3 ))%128* (128) + ((1)*( t2 ))%128] = A[t2][t3]; 
#else
 A_trans[((1)*( t3 ))/128* ( A_size_1 ) + ((1)*( t2 ))/128][((1)*( t3 ))%128* (128) + ((1)*( t2 ))%128] = ((DATA_TYPE) t2*t3) / n; 
#endif
;
      }
      for (t3=128;t3<=n-1;t3++) {
        
#ifndef USE_LOCAL_ARRAYS
 A_trans[((1)*( t3 ))/128* ( A_size_1 ) + ((1)*( t2 ))/128][((1)*( t3 ))%128* (128) + ((1)*( t2 ))%128] = A[t2][t3]; 
#else
 A_trans[((1)*( t3 ))/128* ( A_size_1 ) + ((1)*( t2 ))/128][((1)*( t3 ))%128* (128) + ((1)*( t2 ))%128] = ((DATA_TYPE) t2*t3) / n; 
#endif
;
      }
    }
  }
  if ((n <= 128) && (ts1 == 0)) {
    for (t2=0;t2<=n-1;t2++) {
      for (t3=0;t3<=n-1;t3++) {
        
#ifndef USE_LOCAL_ARRAYS
 A_trans[((1)*( t3 ))/128* ( A_size_1 ) + ((1)*( t2 ))/128][((1)*( t3 ))%128* (128) + ((1)*( t2 ))%128] = A[t2][t3]; 
#else
 A_trans[((1)*( t3 ))/128* ( A_size_1 ) + ((1)*( t2 ))/128][((1)*( t3 ))%128* (128) + ((1)*( t2 ))%128] = ((DATA_TYPE) t2*t3) / n; 
#endif
;
      }
    }
  }
  if (ts1 == 0) {
    for (t2=128;t2<=n-1;t2++) {
      for (t3=0;t3<=127;t3++) {
        
#ifndef USE_LOCAL_ARRAYS
 A_trans[((1)*( t3 ))/128* ( A_size_1 ) + ((1)*( t2 ))/128][((1)*( t3 ))%128* (128) + ((1)*( t2 ))%128] = A[t2][t3]; 
#else
 A_trans[((1)*( t3 ))/128* ( A_size_1 ) + ((1)*( t2 ))/128][((1)*( t3 ))%128* (128) + ((1)*( t2 ))%128] = ((DATA_TYPE) t2*t3) / n; 
#endif
;
      }
    }
  }
  if ((ts1 >= 1) && (ts1 <= floord(n-129,128))) {
    for (t2=128*ts1;t2<=128*ts1+127;t2++) {
      for (t3=0;t3<=128*ts1-1;t3++) {
        
#ifndef USE_LOCAL_ARRAYS
 A_trans[((1)*( t3 ))/128* ( A_size_1 ) + ((1)*( t2 ))/128][((1)*( t3 ))%128* (128) + ((1)*( t2 ))%128] = A[t2][t3]; 
#else
 A_trans[((1)*( t3 ))/128* ( A_size_1 ) + ((1)*( t2 ))/128][((1)*( t3 ))%128* (128) + ((1)*( t2 ))%128] = ((DATA_TYPE) t2*t3) / n; 
#endif
;
      }
      for (t3=128*ts1;t3<=128*ts1+127;t3++) {
        
#ifndef USE_LOCAL_ARRAYS
 A_trans[((1)*( t3 ))/128* ( A_size_1 ) + ((1)*( t2 ))/128][((1)*( t3 ))%128* (128) + ((1)*( t2 ))%128] = A[t2][t3]; 
#else
 A_trans[((1)*( t3 ))/128* ( A_size_1 ) + ((1)*( t2 ))/128][((1)*( t3 ))%128* (128) + ((1)*( t2 ))%128] = ((DATA_TYPE) t2*t3) / n; 
#endif
;
      }
      for (t3=128*ts1+128;t3<=n-1;t3++) {
        
#ifndef USE_LOCAL_ARRAYS
 A_trans[((1)*( t3 ))/128* ( A_size_1 ) + ((1)*( t2 ))/128][((1)*( t3 ))%128* (128) + ((1)*( t2 ))%128] = A[t2][t3]; 
#else
 A_trans[((1)*( t3 ))/128* ( A_size_1 ) + ((1)*( t2 ))/128][((1)*( t3 ))%128* (128) + ((1)*( t2 ))%128] = ((DATA_TYPE) t2*t3) / n; 
#endif
;
      }
    }
  }
  if (ts1 >= max(1,ceild(n-128,128))) {
    for (t2=128*ts1;t2<=n-1;t2++) {
      for (t3=0;t3<=128*ts1-1;t3++) {
        
#ifndef USE_LOCAL_ARRAYS
 A_trans[((1)*( t3 ))/128* ( A_size_1 ) + ((1)*( t2 ))/128][((1)*( t3 ))%128* (128) + ((1)*( t2 ))%128] = A[t2][t3]; 
#else
 A_trans[((1)*( t3 ))/128* ( A_size_1 ) + ((1)*( t2 ))/128][((1)*( t3 ))%128* (128) + ((1)*( t2 ))%128] = ((DATA_TYPE) t2*t3) / n; 
#endif
;
      }
      for (t3=128*ts1;t3<=n-1;t3++) {
        
#ifndef USE_LOCAL_ARRAYS
 A_trans[((1)*( t3 ))/128* ( A_size_1 ) + ((1)*( t2 ))/128][((1)*( t3 ))%128* (128) + ((1)*( t2 ))%128] = A[t2][t3]; 
#else
 A_trans[((1)*( t3 ))/128* ( A_size_1 ) + ((1)*( t2 ))/128][((1)*( t3 ))%128* (128) + ((1)*( t2 ))%128] = ((DATA_TYPE) t2*t3) / n; 
#endif
;
      }
    }
  }
  if (ts1 >= 1) {
    for (t2=128*ts1+128;t2<=n-1;t2++) {
      for (t3=128*ts1;t3<=128*ts1+127;t3++) {
        
#ifndef USE_LOCAL_ARRAYS
 A_trans[((1)*( t3 ))/128* ( A_size_1 ) + ((1)*( t2 ))/128][((1)*( t3 ))%128* (128) + ((1)*( t2 ))%128] = A[t2][t3]; 
#else
 A_trans[((1)*( t3 ))/128* ( A_size_1 ) + ((1)*( t2 ))/128][((1)*( t3 ))%128* (128) + ((1)*( t2 ))%128] = ((DATA_TYPE) t2*t3) / n; 
#endif
;
      }
    }
  }
}
/* End of CLooG code */

}

void read_in_data_tile_alloc_y_1_0(int ts1, double** y_1_trans, BufferManager *buff_mang_y_1)
{


  int t1, t2;

 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;

/* Start of CLooG code */
if ((ts1 >= 0) && (128*ts1 <= n-1)) {
  for (t2=0;t2<=floord(n-1,128);t2++) {
    if(y_1_trans[(t2)] == NULL) y_1_trans[(t2)] = (double *) buffer_alloc_atomic((t2), 128 * sizeof(double), buff_mang_y_1);;
  }
}
/* End of CLooG code */

}

void read_in_copy_y_1_0(int ts1 
#ifndef USE_LOCAL_ARRAYS
 ,__DECLARATION_OF_y_1 
#endif
 , double** y_1_trans, BufferManager *buff_mang_y_1)
{


  int t1, t2;

 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;

/* Start of CLooG code */
if ((ts1 >= 0) && (128*ts1 <= n-1)) {
  for (t2=0;t2<=n-1;t2++) {
    
#ifndef USE_LOCAL_ARRAYS
 y_1_trans[((1)*( t2 ))/128][((1)*( t2 ))%128] = y_1[t2]; 
#else
 y_1_trans[((1)*( t2 ))/128][((1)*( t2 ))%128] = ((DATA_TYPE) t2 + 3) / n; 
#endif
;
  }
}
/* End of CLooG code */

}

void read_in_data_tile_alloc_x2_0(int ts1, double** x2_trans, BufferManager *buff_mang_x2)
{


  int t1, t2;

 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;

/* Start of CLooG code */
if ((ts1 >= 0) && (128*ts1 <= n-1)) {
  if(x2_trans[(ts1)] == NULL) x2_trans[(ts1)] = (double *) buffer_alloc_atomic((ts1), 128 * sizeof(double), buff_mang_x2);;
}
/* End of CLooG code */

}

void read_in_copy_x2_0(int ts1 
#ifndef USE_LOCAL_ARRAYS
 ,__DECLARATION_OF_x2 
#endif
 , double** x2_trans, BufferManager *buff_mang_x2)
{


  int t1, t2;

 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;

/* Start of CLooG code */
if ((ts1 >= 0) && (128*ts1 <= n-1)) {
  for (t2=128*ts1;t2<=min(n-1,128*ts1+127);t2++) {
    
#ifndef USE_LOCAL_ARRAYS
 x2_trans[((1)*( t2 ))/128][((1)*( t2 ))%128] = x2[t2]; 
#else
 x2_trans[((1)*( t2 ))/128][((1)*( t2 ))%128] = ((DATA_TYPE) t2 + 1) / n; 
#endif
;
  }
}
/* End of CLooG code */

}

void read_in_data_tile_alloc_y_2_0(int ts1, double** y_2_trans, BufferManager *buff_mang_y_2)
{


  int t1, t2;

 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;

/* Start of CLooG code */
if ((ts1 >= 0) && (128*ts1 <= n-1)) {
  for (t2=0;t2<=floord(n-1,128);t2++) {
    if(y_2_trans[(t2)] == NULL) y_2_trans[(t2)] = (double *) buffer_alloc_atomic((t2), 128 * sizeof(double), buff_mang_y_2);;
  }
}
/* End of CLooG code */

}

void read_in_copy_y_2_0(int ts1 
#ifndef USE_LOCAL_ARRAYS
 ,__DECLARATION_OF_y_2 
#endif
 , double** y_2_trans, BufferManager *buff_mang_y_2)
{


  int t1, t2;

 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;

/* Start of CLooG code */
if ((ts1 >= 0) && (128*ts1 <= n-1)) {
  for (t2=0;t2<=n-1;t2++) {
    
#ifndef USE_LOCAL_ARRAYS
 y_2_trans[((1)*( t2 ))/128][((1)*( t2 ))%128] = y_2[t2]; 
#else
 y_2_trans[((1)*( t2 ))/128][((1)*( t2 ))%128] = ((DATA_TYPE) t2 + 4) / n; 
#endif
;
  }
}
/* End of CLooG code */

}

void data_tile_init_0(int my_rank, int nprocs 
#ifndef USE_LOCAL_ARRAYS
 ,__DECLARATION_OF_x1,__DECLARATION_OF_A,__DECLARATION_OF_y_1,__DECLARATION_OF_x2,__DECLARATION_OF_y_2 
#endif
 , double** x1_trans, BufferManager *buff_mang_x1, double** A_trans,int A_size_1, BufferManager *buff_mang_A, double** y_1_trans, BufferManager *buff_mang_y_1, double** x2_trans, BufferManager *buff_mang_x2, double** y_2_trans, BufferManager *buff_mang_y_2)
{
  int t1, t2, t3;
 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;
/* Start of CLooG code */
if (n >= 1) {
  for (t2=0;t2<=floord(n-1,128);t2++) {
    y_2_trans[(t2)] = NULL;;
  }
  for (t2=0;t2<=floord(n-1,128);t2++) {
    x2_trans[(t2)] = NULL;;
  }
  for (t2=0;t2<=floord(n-1,128);t2++) {
    y_1_trans[(t2)] = NULL;;
  }
  for (t2=0;t2<=floord(n-1,128);t2++) {
    for (t3=0;t3<=floord(n-1,128);t3++) {
      A_trans[(t2)* ( A_size_1 ) + (t3)] = NULL;;
    }
  }
  for (t2=0;t2<=floord(n-1,128);t2++) {
    x1_trans[(t2)] = NULL;;
  }
  for (t2=0;t2<=floord(n-1,128);t2++) {
    if(pi_0(t2,n,nprocs) == my_rank) data_tile_ref_count_init_x1_0(t2, x1_trans,buff_mang_x1);;
  }
  for (t2=0;t2<=floord(n-1,128);t2++) {
    if(pi_0(t2,n,nprocs) == my_rank) data_tile_ref_count_init_A_0(t2, A_trans, A_size_1,buff_mang_A);;
  }
  for (t2=0;t2<=floord(n-1,128);t2++) {
    if(pi_0(t2,n,nprocs) == my_rank) data_tile_ref_count_init_y_1_0(t2, y_1_trans,buff_mang_y_1);;
  }
  for (t2=0;t2<=floord(n-1,128);t2++) {
    if(pi_0(t2,n,nprocs) == my_rank) data_tile_ref_count_init_x2_0(t2, x2_trans,buff_mang_x2);;
  }
  for (t2=0;t2<=floord(n-1,128);t2++) {
    if(pi_0(t2,n,nprocs) == my_rank) data_tile_ref_count_init_y_2_0(t2, y_2_trans,buff_mang_y_2);;
  }
  for (t2=0;t2<=floord(n-1,128);t2++) {
    if(pi_0(t2,n,nprocs) == my_rank) data_tile_ref_count_init_x1_0(t2, x1_trans,buff_mang_x1);;
  }
  for (t2=0;t2<=floord(n-1,128);t2++) {
    if(pi_0(t2,n,nprocs) == my_rank) data_tile_ref_count_init_x2_0(t2, x2_trans,buff_mang_x2);;
  }
  for (t2=0;t2<=floord(n-1,128);t2++) {
    if(pi_0(t2,n,nprocs) == my_rank){ read_in_data_tile_alloc_x1_0(t2, x1_trans,buff_mang_x1);read_in_copy_x1_0(t2 
#ifndef USE_LOCAL_ARRAYS
, x1 
#endif
, x1_trans,buff_mang_x1);};
  }
  for (t2=0;t2<=floord(n-1,128);t2++) {
    if(pi_0(t2,n,nprocs) == my_rank){ read_in_data_tile_alloc_A_0(t2, A_trans, A_size_1,buff_mang_A);read_in_copy_A_0(t2 
#ifndef USE_LOCAL_ARRAYS
, A 
#endif
, A_trans, A_size_1,buff_mang_A);};
  }
  for (t2=0;t2<=floord(n-1,128);t2++) {
    if(pi_0(t2,n,nprocs) == my_rank){ read_in_data_tile_alloc_y_1_0(t2, y_1_trans,buff_mang_y_1);read_in_copy_y_1_0(t2 
#ifndef USE_LOCAL_ARRAYS
, y_1 
#endif
, y_1_trans,buff_mang_y_1);};
  }
  for (t2=0;t2<=floord(n-1,128);t2++) {
    if(pi_0(t2,n,nprocs) == my_rank){ read_in_data_tile_alloc_x2_0(t2, x2_trans,buff_mang_x2);read_in_copy_x2_0(t2 
#ifndef USE_LOCAL_ARRAYS
, x2 
#endif
, x2_trans,buff_mang_x2);};
  }
  for (t2=0;t2<=floord(n-1,128);t2++) {
    if(pi_0(t2,n,nprocs) == my_rank){ read_in_data_tile_alloc_y_2_0(t2, y_2_trans,buff_mang_y_2);read_in_copy_y_2_0(t2 
#ifndef USE_LOCAL_ARRAYS
, y_2 
#endif
, y_2_trans,buff_mang_y_2);};
  }
}
/* End of CLooG code */
}
void data_tile_copy_back_x1_0(int ts1 
#ifndef USE_LOCAL_ARRAYS
 ,__DECLARATION_OF_x1 
#endif
 , double** x1_trans, BufferManager *buff_mang_x1)
{
#ifndef USE_LOCAL_ARRAYS
  int t1, t2;
 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;
/* Start of CLooG code */
if ((ts1 >= 0) && (128*ts1 <= n-1)) {
  for (t2=128*ts1;t2<=min(n-1,128*ts1+127);t2++) {
    x1[t2] = x1_trans[((1)*( t2 ))/128][((1)*( t2 ))%128];;
  }
}
/* End of CLooG code */
#endif
}
void data_tile_copy_back_x2_0(int ts1 
#ifndef USE_LOCAL_ARRAYS
 ,__DECLARATION_OF_x2 
#endif
 , double** x2_trans, BufferManager *buff_mang_x2)
{
#ifndef USE_LOCAL_ARRAYS
  int t1, t2;
 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;
/* Start of CLooG code */
if ((ts1 >= 0) && (128*ts1 <= n-1)) {
  for (t2=128*ts1;t2<=min(n-1,128*ts1+127);t2++) {
    x2[t2] = x2_trans[((1)*( t2 ))/128][((1)*( t2 ))%128];;
  }
}
/* End of CLooG code */
#endif
}
int writeout_pack_x1_0(int ts1,double *lw_buf_x1,int lw_count_x1, __DECLARATION_OF_x1, double** x1_trans, BufferManager *buff_mang_x1){
int recv_proc;
  int t1, t2;
 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;
/* Start of CLooG code */
if ((ts1 >= 0) && (128*ts1 <= n-1)) {
  if(x1_trans[(ts1)] == NULL) x1_trans[(ts1)] = (double *) buffer_alloc_atomic((ts1), 128 * sizeof(double), buff_mang_x1);;
  for (t1=128*ts1+2;t1<=min(n+1,128*ts1+129);t1++) {
    lw_buf_x1[lw_count_x1++] = x1_trans[((1)*( t1-2 ))/128][((1)*( t1-2 ))%128];
  }
}
/* End of CLooG code */
return lw_count_x1;
}
int writeout_unpack_x1_0(int ts1,double *lw_recv_buf_x1,int displs_lw_x1, __DECLARATION_OF_x1, double** x1_trans, BufferManager *buff_mang_x1){
int recv_proc;
  int t1;
 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;
/* Start of CLooG code */
if ((ts1 >= 0) && (128*ts1 <= n-1)) {
  for (t1=128*ts1;t1<=min(n-1,128*ts1+127);t1++) {
    
#ifndef USE_LOCAL_ARRAYS
 x1[t1] = lw_recv_buf_x1[displs_lw_x1++]; 
#endif
;
  }
}
/* End of CLooG code */
return displs_lw_x1;
}
int writeout_pack_x2_0(int ts1,double *lw_buf_x2,int lw_count_x2, __DECLARATION_OF_x2, double** x2_trans, BufferManager *buff_mang_x2){
int recv_proc;
  int t1, t2;
 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;
/* Start of CLooG code */
if ((ts1 >= 0) && (128*ts1 <= n-1)) {
  if(x2_trans[(ts1)] == NULL) x2_trans[(ts1)] = (double *) buffer_alloc_atomic((ts1), 128 * sizeof(double), buff_mang_x2);;
  for (t1=128*ts1+2;t1<=min(n+1,128*ts1+129);t1++) {
    lw_buf_x2[lw_count_x2++] = x2_trans[((1)*( t1-2 ))/128][((1)*( t1-2 ))%128];
  }
}
/* End of CLooG code */
return lw_count_x2;
}
int writeout_unpack_x2_0(int ts1,double *lw_recv_buf_x2,int displs_lw_x2, __DECLARATION_OF_x2, double** x2_trans, BufferManager *buff_mang_x2){
int recv_proc;
  int t1;
 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;
/* Start of CLooG code */
if ((ts1 >= 0) && (128*ts1 <= n-1)) {
  for (t1=128*ts1;t1<=min(n-1,128*ts1+127);t1++) {
    
#ifndef USE_LOCAL_ARRAYS
 x2[t1] = lw_recv_buf_x2[displs_lw_x2++]; 
#endif
;
  }
}
/* End of CLooG code */
return displs_lw_x2;
}
void pack_send_0(int my_rank,int nprocs,int t1, double** x1_trans, BufferManager *buff_mang_x1,double **send_buf_x1,MPI_Request *send_reqs_x1,MPI_Status *send_stats_x1, double** A_trans,int A_size_1, BufferManager *buff_mang_A,double **send_buf_A,MPI_Request *send_reqs_A,MPI_Status *send_stats_A, double** y_1_trans, BufferManager *buff_mang_y_1,double **send_buf_y_1,MPI_Request *send_reqs_y_1,MPI_Status *send_stats_y_1, double** x2_trans, BufferManager *buff_mang_x2,double **send_buf_x2,MPI_Request *send_reqs_x2,MPI_Status *send_stats_x2, double** y_2_trans, BufferManager *buff_mang_y_2,double **send_buf_y_2,MPI_Request *send_reqs_y_2,MPI_Status *send_stats_y_2,int max_num_sends,double *p_total_count)
{
int __p, receiver_list[nprocs];
int send_counts_x1[nprocs];
for (__p=0;__p<nprocs;__p++) send_counts_x1[__p] = 0;
int index_x1[nprocs];
int current_index_x1 = 0;
int send_counts_A[nprocs];
for (__p=0;__p<nprocs;__p++) send_counts_A[__p] = 0;
int index_A[nprocs];
int current_index_A = 0;
int send_counts_y_1[nprocs];
for (__p=0;__p<nprocs;__p++) send_counts_y_1[__p] = 0;
int index_y_1[nprocs];
int current_index_y_1 = 0;
int send_counts_x2[nprocs];
for (__p=0;__p<nprocs;__p++) send_counts_x2[__p] = 0;
int index_x2[nprocs];
int current_index_x2 = 0;
int send_counts_y_2[nprocs];
for (__p=0;__p<nprocs;__p++) send_counts_y_2[__p] = 0;
int index_y_2[nprocs];
int current_index_y_2 = 0;
int _i;
double __total_count = 0;
;
    for (__p=0; __p<nprocs; __p++) { if(send_counts_x1[__p] > 0) {send_buf_x1[index_x1[__p]][0] = 0;send_buf_x1[index_x1[__p]][1] = t1; IF_DYNSCHEDULER_DEBUG_PRINT( fprintf(__debug_print_fp, "sending node %d loop 0 task %d 		        size %d to node %d\n", my_rank, t1, send_counts_x1[__p], __p)); IF_DEBUG_FLUSH(fflush(__debug_print_fp));IF_TIME(__total_count += send_counts_x1[__p]); MPI_Isend(send_buf_x1[index_x1[__p]], send_counts_x1[__p], MPI_DOUBLE, __p, 2123, MPI_COMM_WORLD, &send_reqs_x1[index_x1[__p]]); send_counts_x1[__p] = 0;}};
;
    for (__p=0; __p<nprocs; __p++) { if(send_counts_A[__p] > 0) {send_buf_A[index_A[__p]][0] = 0;send_buf_A[index_A[__p]][1] = t1; IF_DYNSCHEDULER_DEBUG_PRINT( fprintf(__debug_print_fp, "sending node %d loop 0 task %d 		        size %d to node %d\n", my_rank, t1, send_counts_A[__p], __p)); IF_DEBUG_FLUSH(fflush(__debug_print_fp));IF_TIME(__total_count += send_counts_A[__p]); MPI_Isend(send_buf_A[index_A[__p]], send_counts_A[__p], MPI_DOUBLE, __p, 3123, MPI_COMM_WORLD, &send_reqs_A[index_A[__p]]); send_counts_A[__p] = 0;}};
;
    for (__p=0; __p<nprocs; __p++) { if(send_counts_y_1[__p] > 0) {send_buf_y_1[index_y_1[__p]][0] = 0;send_buf_y_1[index_y_1[__p]][1] = t1; IF_DYNSCHEDULER_DEBUG_PRINT( fprintf(__debug_print_fp, "sending node %d loop 0 task %d 		        size %d to node %d\n", my_rank, t1, send_counts_y_1[__p], __p)); IF_DEBUG_FLUSH(fflush(__debug_print_fp));IF_TIME(__total_count += send_counts_y_1[__p]); MPI_Isend(send_buf_y_1[index_y_1[__p]], send_counts_y_1[__p], MPI_DOUBLE, __p, 4123, MPI_COMM_WORLD, &send_reqs_y_1[index_y_1[__p]]); send_counts_y_1[__p] = 0;}};
;
    for (__p=0; __p<nprocs; __p++) { if(send_counts_x2[__p] > 0) {send_buf_x2[index_x2[__p]][0] = 0;send_buf_x2[index_x2[__p]][1] = t1; IF_DYNSCHEDULER_DEBUG_PRINT( fprintf(__debug_print_fp, "sending node %d loop 0 task %d 		        size %d to node %d\n", my_rank, t1, send_counts_x2[__p], __p)); IF_DEBUG_FLUSH(fflush(__debug_print_fp));IF_TIME(__total_count += send_counts_x2[__p]); MPI_Isend(send_buf_x2[index_x2[__p]], send_counts_x2[__p], MPI_DOUBLE, __p, 5123, MPI_COMM_WORLD, &send_reqs_x2[index_x2[__p]]); send_counts_x2[__p] = 0;}};
;
    for (__p=0; __p<nprocs; __p++) { if(send_counts_y_2[__p] > 0) {send_buf_y_2[index_y_2[__p]][0] = 0;send_buf_y_2[index_y_2[__p]][1] = t1; IF_DYNSCHEDULER_DEBUG_PRINT( fprintf(__debug_print_fp, "sending node %d loop 0 task %d 		        size %d to node %d\n", my_rank, t1, send_counts_y_2[__p], __p)); IF_DEBUG_FLUSH(fflush(__debug_print_fp));IF_TIME(__total_count += send_counts_y_2[__p]); MPI_Isend(send_buf_y_2[index_y_2[__p]], send_counts_y_2[__p], MPI_DOUBLE, __p, 6123, MPI_COMM_WORLD, &send_reqs_y_2[index_y_2[__p]]); send_counts_y_2[__p] = 0;}};
*p_total_count += __total_count;
}
void unpack_buf_x1_0(int my_rank,int nprocs, double** x1_trans, BufferManager *buff_mang_x1,double *recv_buf_x1)
{
int curr_displs_x1;
curr_displs_x1 = 2; ;
}
void unpack_buf_A_0(int my_rank,int nprocs, double** A_trans,int A_size_1, BufferManager *buff_mang_A,double *recv_buf_A)
{
int curr_displs_A;
curr_displs_A = 2; ;
}
void unpack_buf_y_1_0(int my_rank,int nprocs, double** y_1_trans, BufferManager *buff_mang_y_1,double *recv_buf_y_1)
{
int curr_displs_y_1;
curr_displs_y_1 = 2; ;
}
void unpack_buf_x2_0(int my_rank,int nprocs, double** x2_trans, BufferManager *buff_mang_x2,double *recv_buf_x2)
{
int curr_displs_x2;
curr_displs_x2 = 2; ;
}
void unpack_buf_y_2_0(int my_rank,int nprocs, double** y_2_trans, BufferManager *buff_mang_y_2,double *recv_buf_y_2)
{
int curr_displs_y_2;
curr_displs_y_2 = 2; ;
}
void copy_back_writeout_0(int my_rank, int nprocs 
#ifndef USE_LOCAL_ARRAYS
 ,__DECLARATION_OF_x1,__DECLARATION_OF_x2 
#endif
 , double** x1_trans, BufferManager *buff_mang_x1, double** x2_trans, BufferManager *buff_mang_x2)
{
#ifndef USE_LOCAL_ARRAYS
  int t1, t2;
 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;
/* Start of CLooG code */
if (n >= 1) {
  for (t2=0;t2<=floord(n-1,128);t2++) {
    if(pi_0(t2,n,nprocs) == my_rank) data_tile_copy_back_x1_0(t2, x1, x1_trans,buff_mang_x1);;
  }
  for (t2=0;t2<=floord(n-1,128);t2++) {
    if(pi_0(t2,n,nprocs) == my_rank) data_tile_copy_back_x2_0(t2, x2, x2_trans,buff_mang_x2);;
  }
}
/* End of CLooG code */
#endif
}
void write_out(int my_rank,int nprocs,double *lw_buf_x1,double *lw_recv_buf_x1,int *displs_lw_x1,double *lw_buf_x2,double *lw_recv_buf_x2,int *displs_lw_x2, __DECLARATION_OF_x1, __DECLARATION_OF_x2, double** x1_trans, BufferManager *buff_mang_x1, double** x2_trans, BufferManager *buff_mang_x2)
{
#ifndef USE_LOCAL_ARRAYS
int __p, proc;
int _lb_dist, _ub_dist, lbd_t1, ubd_t1, lbd_t2, ubd_t2, lbd_t3, ubd_t3, lbd_t4, ubd_t4;
int lw_recv_counts_x1[nprocs];
int curr_displs_lw_x1[nprocs];
int lw_count_x1 = 0;
int lw_recv_counts_x2[nprocs];
int curr_displs_lw_x2[nprocs];
int lw_count_x2 = 0;
  int t1, t2;
 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;
/* Start of CLooG code */
if (n >= 1) {
  for (t2=0;t2<=floord(n-1,128);t2++) {
    if (pi_0(t2,n,nprocs) == my_rank)lw_count_x1 = writeout_pack_x1_0(t2,lw_buf_x1,lw_count_x1,x1, x1_trans,buff_mang_x1);;
  }
  assert((nprocs == 1) || (lw_count_x1 <= displs_lw_x1[1])); MPI_Gather(&lw_count_x1, 1, MPI_INT, lw_recv_counts_x1, 1, MPI_INT, 0, MPI_COMM_WORLD);MPI_Gatherv(lw_buf_x1, lw_count_x1, MPI_DOUBLE, lw_recv_buf_x1, lw_recv_counts_x1, displs_lw_x1, MPI_DOUBLE, 0, MPI_COMM_WORLD); lw_count_x1 = 0; for (__p=0; __p<nprocs; __p++) curr_displs_lw_x1[__p] = displs_lw_x1[__p];;
  if (my_rank == 0) {
    for (t2=0;t2<=floord(n-1,128);t2++) {
      proc = pi_0(t2, n, nprocs);if ((my_rank != proc) && (lw_recv_counts_x1[proc] > 0)) {curr_displs_lw_x1[proc] = writeout_unpack_x1_0(t2,lw_recv_buf_x1,curr_displs_lw_x1[proc],x1, x1_trans,buff_mang_x1);};
    }
  }
  for (t2=0;t2<=floord(n-1,128);t2++) {
    if (pi_0(t2,n,nprocs) == my_rank)lw_count_x2 = writeout_pack_x2_0(t2,lw_buf_x2,lw_count_x2,x2, x2_trans,buff_mang_x2);;
  }
  assert((nprocs == 1) || (lw_count_x2 <= displs_lw_x2[1])); MPI_Gather(&lw_count_x2, 1, MPI_INT, lw_recv_counts_x2, 1, MPI_INT, 0, MPI_COMM_WORLD);MPI_Gatherv(lw_buf_x2, lw_count_x2, MPI_DOUBLE, lw_recv_buf_x2, lw_recv_counts_x2, displs_lw_x2, MPI_DOUBLE, 0, MPI_COMM_WORLD); lw_count_x2 = 0; for (__p=0; __p<nprocs; __p++) curr_displs_lw_x2[__p] = displs_lw_x2[__p];;
  if (my_rank == 0) {
    for (t2=0;t2<=floord(n-1,128);t2++) {
      proc = pi_0(t2, n, nprocs);if ((my_rank != proc) && (lw_recv_counts_x2[proc] > 0)) {curr_displs_lw_x2[proc] = writeout_unpack_x2_0(t2,lw_recv_buf_x2,curr_displs_lw_x2[proc],x2, x2_trans,buff_mang_x2);};
    }
  }
}
/* End of CLooG code */
#endif
}
void data_tile_alloc_0(int ts1, double** x1_trans, BufferManager *buff_mang_x1, double** A_trans,int A_size_1, BufferManager *buff_mang_A, double** y_1_trans, BufferManager *buff_mang_y_1, double** x2_trans, BufferManager *buff_mang_x2, double** y_2_trans, BufferManager *buff_mang_y_2)
{
  int t1, t2, t3;
 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;
/* Start of CLooG code */
if ((ts1 >= 0) && (128*ts1 <= n-1)) {
  if(x1_trans[(ts1)] == NULL) x1_trans[(ts1)] = (double *) buffer_alloc_atomic((ts1), 128 * sizeof(double), buff_mang_x1);;
  for (t2=0;t2<=ts1-1;t2++) {
    if(A_trans[(t2) * ((A_size_1)) + (ts1)] == NULL) A_trans[(t2)* ( A_size_1 ) + (ts1)] = (double *) buffer_alloc_atomic((t2)* ( A_size_1 ) + (ts1), 128*128 * sizeof(double), buff_mang_A);;
  }
  for (t3=0;t3<=floord(n-1,128);t3++) {
    if(A_trans[(ts1) * ((A_size_1)) + (t3)] == NULL) A_trans[(ts1)* ( A_size_1 ) + (t3)] = (double *) buffer_alloc_atomic((ts1)* ( A_size_1 ) + (t3), 128*128 * sizeof(double), buff_mang_A);;
  }
  for (t2=ts1+1;t2<=floord(n-1,128);t2++) {
    if(A_trans[(t2) * ((A_size_1)) + (ts1)] == NULL) A_trans[(t2)* ( A_size_1 ) + (ts1)] = (double *) buffer_alloc_atomic((t2)* ( A_size_1 ) + (ts1), 128*128 * sizeof(double), buff_mang_A);;
  }
  for (t2=0;t2<=floord(n-1,128);t2++) {
    if(y_1_trans[(t2)] == NULL) y_1_trans[(t2)] = (double *) buffer_alloc_atomic((t2), 128 * sizeof(double), buff_mang_y_1);;
  }
  if(x2_trans[(ts1)] == NULL) x2_trans[(ts1)] = (double *) buffer_alloc_atomic((ts1), 128 * sizeof(double), buff_mang_x2);;
  for (t2=0;t2<=floord(n-1,128);t2++) {
    if(y_2_trans[(t2)] == NULL) y_2_trans[(t2)] = (double *) buffer_alloc_atomic((t2), 128 * sizeof(double), buff_mang_y_2);;
  }
}
/* End of CLooG code */
}
void data_tile_ref_count_update_0(int ts1, double** x1_trans, BufferManager *buff_mang_x1, double** A_trans,int A_size_1, BufferManager *buff_mang_A, double** y_1_trans, BufferManager *buff_mang_y_1, double** x2_trans, BufferManager *buff_mang_x2, double** y_2_trans, BufferManager *buff_mang_y_2)
{
  int t1, t2, t3;
 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;
/* Start of CLooG code */
if ((ts1 >= 0) && (128*ts1 <= n-1)) {
  decrement_ref_count(buff_mang_x1, (ts1));;
  for (t2=0;t2<=ts1-1;t2++) {
    decrement_ref_count(buff_mang_A, (t2)* ( A_size_1 ) + (ts1));;
  }
  for (t3=0;t3<=floord(n-1,128);t3++) {
    decrement_ref_count(buff_mang_A, (ts1)* ( A_size_1 ) + (t3));;
  }
  for (t2=ts1+1;t2<=floord(n-1,128);t2++) {
    decrement_ref_count(buff_mang_A, (t2)* ( A_size_1 ) + (ts1));;
  }
  for (t2=0;t2<=floord(n-1,128);t2++) {
    decrement_ref_count(buff_mang_y_1, (t2));;
  }
  decrement_ref_count(buff_mang_x2, (ts1));;
  for (t2=0;t2<=floord(n-1,128);t2++) {
    decrement_ref_count(buff_mang_y_2, (t2));;
  }
}
/* End of CLooG code */
}
void compute_task_0(int t1, double** x1_trans, BufferManager *buff_mang_x1, double** A_trans,int A_size_1, BufferManager *buff_mang_A, double** y_1_trans, BufferManager *buff_mang_y_1, double** x2_trans, BufferManager *buff_mang_x2, double** y_2_trans, BufferManager *buff_mang_y_2)
{
int i_0_div = 0;int i_0_mod = 0;double *x1_i_0 = NULL;int j_i_0_div = 0;int j_i_0_mod = 0;double *A_j_i_0 = NULL;int j_0_div = 0;int j_0_mod = 0;double *y_1_j_0 = NULL;int i_1_div = 0;int i_1_mod = 0;double *x2_i_1 = NULL;int i_j_1_div = 0;int i_j_1_mod = 0;double *A_i_j_1 = NULL;int j_1_div = 0;int j_1_mod = 0;double *y_2_j_1 = NULL; int t2, t3, t4;
 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;
/* Start of CLooG code */
for (t2=0;t2<=floord(n-1,128);t2++) {
  for (t3=128*t1;t3<=min(n-1,128*t1+127);t3++) {
    lbv=128*t2;
    ubv=min(n-1,128*t2+127);
    #if (defined __DATA_DIST_DECLS || defined USE_LOCAL_ARRAYS)
    i_0_div = ((1)*( t3 ))/128;i_0_mod = ((1)*( t3 ))%128;x1_i_0 = x1_trans[i_0_div];j_i_0_div = ((1)*( lbv ))/128* ( A_size_1 ) + ((1)*( t3 ))/128;j_i_0_mod = ((1)*( lbv ))%128* (128) + ((1)*( t3 ))%128;A_j_i_0 = A_trans[j_i_0_div];j_0_div = ((1)*( lbv ))/128;j_0_mod = ((1)*( lbv ))%128;y_1_j_0 = y_1_trans[j_0_div];;
    i_1_div = ((1)*( t3 ))/128;i_1_mod = ((1)*( t3 ))%128;x2_i_1 = x2_trans[i_1_div];i_j_1_div = ((1)*( t3 ))/128* ( A_size_1 ) + ((1)*( lbv ))/128;i_j_1_mod = ((1)*( t3 ))%128* (128) + ((1)*( lbv ))%128;A_i_j_1 = A_trans[i_j_1_div];j_1_div = ((1)*( lbv ))/128;j_1_mod = ((1)*( lbv ))%128;y_2_j_1 = y_2_trans[j_1_div];;
    #endif
#pragma ivdep
#pragma vector always
    for (t4=lbv;t4<=ubv;t4++) {
      x1_i_0[i_0_mod]=x1_i_0[i_0_mod]+A_j_i_0[j_i_0_mod + (t4 - lbv) * 128]*y_1_j_0[j_0_mod + t4 - lbv];;
      x2_i_1[i_1_mod]=x2_i_1[i_1_mod]+A_i_j_1[i_j_1_mod + t4 - lbv]*y_2_j_1[j_1_mod + t4 - lbv];;
    }
  }
}
/* End of CLooG code */
}
