#include "polyrt.h"
#include "buffer_manager.h"
#include <mpi.h>
#include "tbb/concurrent_priority_queue.h"

#include <omp.h>

#ifdef __DEBUG_FLUSH
#define IF_DEBUG_FLUSH(foo) foo
#else
#define IF_DEBUG_FLUSH(foo) 
#endif

#ifdef __DYNSCHEDULER_MORE_DEBUG_PRINT
#define IF_DYNSCHEDULER_MORE_DEBUG_PRINT(foo) foo
#else
#define IF_DYNSCHEDULER_MORE_DEBUG_PRINT(foo) 
#endif

#ifdef __DYNSCHEDULER_DEBUG_PRINT
#define IF_DYNSCHEDULER_DEBUG_PRINT(foo) foo
#else
#define IF_DYNSCHEDULER_DEBUG_PRINT(foo) 
#endif

extern FILE *__debug_print_fp;

class __Task {
	public:
		int loop;
		int id1;
		int remote_dep_tasks;
		int local_dep_tasks;
		int thread_affinity;
		__Task() {}
		__Task(const int l, const int _id1, const int remote, const int local, const int affinity) :
			loop(l),
			id1(_id1),
			remote_dep_tasks(remote),
			local_dep_tasks(local),
			thread_affinity(affinity) {}
		__Task(const __Task &task) :
			loop(task.loop),
			id1(task.id1),
			remote_dep_tasks(task.remote_dep_tasks),
			local_dep_tasks(task.local_dep_tasks),
			thread_affinity(task.thread_affinity) {}
		~__Task() {}
};

class __Compare {
	public:
		// returns true only if task2 has higher priority than task1
		bool operator()(const __Task &task1, const __Task &task2) const {
#ifndef __DYNSCHEDULER_NO_PRIORITY
			if (task2.remote_dep_tasks>task1.remote_dep_tasks) return true;
			if (task1.remote_dep_tasks>task2.remote_dep_tasks) return false;
			if (task2.local_dep_tasks>task1.local_dep_tasks) return true;
			if (task1.local_dep_tasks>task2.local_dep_tasks) return false;
			int thread_id = omp_get_thread_num();
			if ((thread_id == task2.thread_affinity) && (thread_id != task1.thread_affinity)) return true;
			if ((thread_id == task1.thread_affinity) && (thread_id != task2.thread_affinity)) return false;
			// lexicographic compare 
			if ( (task2.id1 > task1.id1) )
				 return true;
#endif
			return false;
		}
};

extern tbb::concurrent_priority_queue<__Task, __Compare> pqueue;

int is_receiver_0(int ts1, int n, int my_rank, int nprocs);
int get_num_remote_dep_tasks_0(int ts1, int n, int my_rank, int nprocs);
void remote_update_dep_tasks_0(int ts1, int n, int my_rank, int nprocs, int num_threads,int *tasks_loop0);
int get_num_local_dep_tasks_0(int ts1, int n, int my_rank, int nprocs);
void local_update_dep_tasks_0(int ts1, int n, int my_rank, int nprocs, int num_threads,int *tasks_loop0);
int get_num_remote_src_tasks_0(int ts1, int n, int my_rank, int nprocs);
int get_num_local_src_tasks_0(int ts1, int n, int my_rank, int nprocs);
extern int max_num_tasks_loop0_dim0;
extern int lb_tasks_loop0_dim0;
extern int *tasks_loop0;
void init_tasks(int my_rank,int nprocs,int num_threads,int *tasks_loop0,long int *p_num_tasks_to_execute,long int *p_num_tasks_to_unpack);
void data_tile_ref_count_init_x1_0(int ts1, double** x1_trans, BufferManager *buff_mang_x1);
void data_tile_ref_count_init_A_0(int ts1, double** A_trans,int A_size_1, BufferManager *buff_mang_A);
void data_tile_ref_count_init_y_1_0(int ts1, double** y_1_trans, BufferManager *buff_mang_y_1);
void data_tile_ref_count_init_x2_0(int ts1, double** x2_trans, BufferManager *buff_mang_x2);
void data_tile_ref_count_init_y_2_0(int ts1, double** y_2_trans, BufferManager *buff_mang_y_2);
void read_in_data_tile_alloc_x1_0(int ts1, double** x1_trans, BufferManager *buff_mang_x1);
void read_in_data_tile_alloc_A_0(int ts1, double** A_trans,int A_size_1, BufferManager *buff_mang_A);
void read_in_data_tile_alloc_y_1_0(int ts1, double** y_1_trans, BufferManager *buff_mang_y_1);
void read_in_data_tile_alloc_x2_0(int ts1, double** x2_trans, BufferManager *buff_mang_x2);
void read_in_data_tile_alloc_y_2_0(int ts1, double** y_2_trans, BufferManager *buff_mang_y_2);
int is_receiver_x1_0(int ts1, int n, int my_rank, int nprocs);
void sigma_x1_1_0(int ts1, int n, int my_rank, int nprocs, int *receiver_list);
int is_receiver_x1_1_0(int ts1, int n, int my_rank, int nprocs);
int is_receiver_A_0(int ts1, int n, int my_rank, int nprocs);
void sigma_A_1_0(int ts1, int n, int my_rank, int nprocs, int *receiver_list);
int is_receiver_A_1_0(int ts1, int n, int my_rank, int nprocs);
int is_receiver_y_1_0(int ts1, int n, int my_rank, int nprocs);
void sigma_y_1_1_0(int ts1, int n, int my_rank, int nprocs, int *receiver_list);
int is_receiver_y_1_1_0(int ts1, int n, int my_rank, int nprocs);
int is_receiver_x2_0(int ts1, int n, int my_rank, int nprocs);
void sigma_x2_1_0(int ts1, int n, int my_rank, int nprocs, int *receiver_list);
int is_receiver_x2_1_0(int ts1, int n, int my_rank, int nprocs);
int is_receiver_y_2_0(int ts1, int n, int my_rank, int nprocs);
void sigma_y_2_1_0(int ts1, int n, int my_rank, int nprocs, int *receiver_list);
int is_receiver_y_2_1_0(int ts1, int n, int my_rank, int nprocs);
void pack_send_0(int my_rank,int nprocs,int t1, double** x1_trans, BufferManager *buff_mang_x1,double **send_buf_x1,MPI_Request *send_reqs_x1,MPI_Status *send_stats_x1, double** A_trans,int A_size_1, BufferManager *buff_mang_A,double **send_buf_A,MPI_Request *send_reqs_A,MPI_Status *send_stats_A, double** y_1_trans, BufferManager *buff_mang_y_1,double **send_buf_y_1,MPI_Request *send_reqs_y_1,MPI_Status *send_stats_y_1, double** x2_trans, BufferManager *buff_mang_x2,double **send_buf_x2,MPI_Request *send_reqs_x2,MPI_Status *send_stats_x2, double** y_2_trans, BufferManager *buff_mang_y_2,double **send_buf_y_2,MPI_Request *send_reqs_y_2,MPI_Status *send_stats_y_2,int max_num_sends,double *p_total_count);
void unpack_buf_x1_0(int my_rank,int nprocs, double** x1_trans, BufferManager *buff_mang_x1,double *recv_buf_x1);
void unpack_buf_A_0(int my_rank,int nprocs, double** A_trans,int A_size_1, BufferManager *buff_mang_A,double *recv_buf_A);
void unpack_buf_y_1_0(int my_rank,int nprocs, double** y_1_trans, BufferManager *buff_mang_y_1,double *recv_buf_y_1);
void unpack_buf_x2_0(int my_rank,int nprocs, double** x2_trans, BufferManager *buff_mang_x2,double *recv_buf_x2);
void unpack_buf_y_2_0(int my_rank,int nprocs, double** y_2_trans, BufferManager *buff_mang_y_2,double *recv_buf_y_2);
int pi_0( int t2, int n, int nprocs);
int pi_threads_0( int t2, int n, int my_rank, int nprocs, int num_threads);
void data_tile_alloc_0(int ts1, double** x1_trans, BufferManager *buff_mang_x1, double** A_trans,int A_size_1, BufferManager *buff_mang_A, double** y_1_trans, BufferManager *buff_mang_y_1, double** x2_trans, BufferManager *buff_mang_x2, double** y_2_trans, BufferManager *buff_mang_y_2);
void data_tile_ref_count_update_0(int ts1, double** x1_trans, BufferManager *buff_mang_x1, double** A_trans,int A_size_1, BufferManager *buff_mang_A, double** y_1_trans, BufferManager *buff_mang_y_1, double** x2_trans, BufferManager *buff_mang_x2, double** y_2_trans, BufferManager *buff_mang_y_2);
void compute_task_0(int t1, double** x1_trans, BufferManager *buff_mang_x1, double** A_trans,int A_size_1, BufferManager *buff_mang_A, double** y_1_trans, BufferManager *buff_mang_y_1, double** x2_trans, BufferManager *buff_mang_x2, double** y_2_trans, BufferManager *buff_mang_y_2);
