/**
 * bicg.c: This file is part of the PolyBench/C 3.2 test suite.
 *
 *
 * Contact: Louis-Noel Pouchet <pouchet@cse.ohio-state.edu>
 * Web address: http://polybench.sourceforge.net
 */
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <math.h>

/* Include polybench common header. */
#include <polybench.h>

#include "scalapack_util.h"
#define BLOCK_SIZE 100

#define N 30000

/* Include benchmark-specific header. */
/* Default data type is double, default size is 4000. */
#include "mvt.h"


/* Array initialization. */
static
void init_array(int n,
		DATA_TYPE POLYBENCH_1D(x1,N,n),
		DATA_TYPE POLYBENCH_1D(x2,N,n),
		DATA_TYPE POLYBENCH_1D(y_1,N,n),
		DATA_TYPE POLYBENCH_1D(y_2,N,n),
		DATA_TYPE POLYBENCH_2D(A,N,N,n,n))
{
  int i, j;

  for (i = 0; i < n; i++)
    {
      x1[i] = ((DATA_TYPE) i) / n;
      x2[i] = ((DATA_TYPE) i + 1) / n;
      y_1[i] = ((DATA_TYPE) i + 3) / n;
      y_2[i] = ((DATA_TYPE) i + 4) / n;
      for (j = 0; j < n; j++)
	A[i][j] = ((DATA_TYPE) i*j) / N;
    }
}


/* DCE code. Must scan the entire live-out data.
   Can be used also to check the correctness of the output. */
static
void print_array(int n,
		 DATA_TYPE POLYBENCH_1D(x1,N,n),
		 DATA_TYPE POLYBENCH_1D(x2,N,n))

{
  int i;

  for (i = 0; i < n; i++) {
    fprintf (stderr, DATA_PRINTF_MODIFIER, x1[i]);
    fprintf (stderr, DATA_PRINTF_MODIFIER, x2[i]);
    if (i % 20 == 0) fprintf (stderr, "\n");
  }
}

#define __DECLARATION_OF_x1 DATA_TYPE POLYBENCH_1D(x1,N,n)
#define __DECLARATION_OF_x2 DATA_TYPE POLYBENCH_1D(x2,N,n)
#define __DECLARATION_OF_y_1 DATA_TYPE POLYBENCH_1D(y_1,N,n)
#define __DECLARATION_OF_y_2 DATA_TYPE POLYBENCH_1D(y_2,N,n)
#define __DECLARATION_OF_A DATA_TYPE POLYBENCH_2D(A,N,N,n,n)

#define __PLACE_TO_INSERT_FORWARD_DECLARATIONS

/* Main computational kernel. The whole function will be timed,
   including the call and return. */
static
void kernel_mvt(int n,
		DATA_TYPE POLYBENCH_1D(x1,N,n),
		DATA_TYPE POLYBENCH_1D(x2,N,n),
		DATA_TYPE POLYBENCH_1D(y_1,N,n),
		DATA_TYPE POLYBENCH_1D(y_2,N,n),
		DATA_TYPE POLYBENCH_2D(A,N,N,n,n))
{
  int i, j;

#pragma scop
  for (i = 0; i < _PB_N; i++)
    for (j = 0; j < _PB_N; j++)
      x1[i] = x1[i] + A[i][j] * y_1[j];
  for (i = 0; i < _PB_N; i++)
    for (j = 0; j < _PB_N; j++)
      x2[i] = x2[i] + A[j][i] * y_2[j];
#pragma endscop

}



/* Main scalapack computational kernel. The whole function will be timed,
   including the call and return. */
static
void scalapack_kernel_mvt(int nx, int ny,
         int nprocs)
{
  int i, j;

double *A_local, *x1_local, *x2_local, *y1_local, *y2_local;

MDESC desc_A, desc_x1, desc_x2, desc_y1, desc_y2;
MDESC desc_A_local, desc_x1_local, desc_x2_local, desc_y1_local, desc_y2_local;

MKL_INT iam,  ictxt, myrow, mycol, nprow, npcol;
MKL_INT lld, lld_local, info, nx_mkl, ny_mkl, nb, nq, mp, nq_vec, mp_vec;

    int n_procs_row = 1, n_procs_col = 1;

    get_procs_grid_size(nprocs , &n_procs_row, &n_procs_col);

    nprow = (MKL_INT) n_procs_row;
    npcol = (MKL_INT) n_procs_col;

    nx_mkl = (MKL_INT) nx;
    ny_mkl = (MKL_INT) ny;

    nb = (MKL_INT) BLOCK_SIZE;

/*  Init working 2D process grid */
    blacs_get_( &i_negone, &i_zero, &ictxt );
    blacs_gridinit_( &ictxt, "R", &nprow, &npcol );
    blacs_gridinfo_( &ictxt, &nprow, &npcol, &myrow, &mycol );

/*  Compute precise length of local pieces and allocate array on
    each process for parts of distributed vectors */
    mp = numroc_( &nx_mkl, &nb, &myrow, &i_zero, &nprow );
    nq = numroc_( &ny_mkl, &nb, &mycol, &i_zero, &npcol );

    mp_vec = numroc_( &i_one, &i_one, &myrow, &i_zero, &nprow );
    nq_vec = numroc_( &i_one, &i_one, &mycol, &i_zero, &npcol );
    //printf("myrow=%d, mycol=%d mp=%d, nq = %d, nq_rhs = %d\n", myrow, mycol, mp, nq, nq_vec);
    if( ( myrow == 0 ) && ( mycol == 0 ) ){
        my_rank = 0;
    }


    A_local = (double*) calloc( mp*nq, sizeof( double ) );

/*  Only #0 process column contains distributed 1-d vectors*/
    if ( nq_vec != 0 ){
        x1_local = (double *) calloc(mp, sizeof( double ));
        x2_local = (double *) calloc(mp, sizeof( double ));
        y1_local = (double *) calloc(mp, sizeof( double ));
        y2_local = (double *) calloc(mp, sizeof( double ));
    } else {                                                                                 
        x1_local = NULL;
        x2_local = NULL;
        y1_local = NULL;
        y2_local = NULL;
    }

   /*  Compute leading dimensions */
    lld_local = MAX( numroc_( &nx_mkl, &ny_mkl, &myrow, &i_zero, &nprow ), 1 );
    lld = MAX( mp, 1 );

    /*  Initialize descriptors for initial arrays located on 0 process */
    descinit_( desc_A, &nx_mkl, &ny_mkl, &nx_mkl, &ny_mkl, &i_zero, &i_zero, &ictxt, &lld_local, &info );

    descinit_( desc_x1, &nx_mkl, &i_one, &nx_mkl, &i_one, &i_zero, &i_zero, &ictxt, &lld_local, &info );
    descinit_( desc_x2, &nx_mkl, &i_one, &nx_mkl, &i_one, &i_zero, &i_zero, &ictxt, &lld_local, &info );
    descinit_( desc_y1, &nx_mkl, &i_one, &nx_mkl, &i_one, &i_zero, &i_zero, &ictxt, &lld_local, &info );
    descinit_( desc_y2, &nx_mkl, &i_one, &nx_mkl, &i_one, &i_zero, &i_zero, &ictxt, &lld_local, &info );

/*  Initialize descriptors for distributed arrays */

    descinit_( desc_A_local, &nx_mkl, &ny_mkl, &nb, &nb, &i_zero, &i_zero, &ictxt, &lld, &info );
    descinit_( desc_x1_local, &nx_mkl, &i_one, &nb, &i_one, &i_zero, &i_zero, &ictxt, &lld, &info );
    descinit_( desc_x2_local, &nx_mkl, &i_one, &nb, &i_one, &i_zero, &i_zero, &ictxt, &lld, &info );
    descinit_( desc_y1_local, &nx_mkl, &i_one, &nb, &i_one, &i_zero, &i_zero, &ictxt, &lld, &info );
    descinit_( desc_y2_local, &nx_mkl, &i_one, &nb, &i_one, &i_zero, &i_zero, &ictxt, &lld, &info );

  DATA_TYPE alpha;
  DATA_TYPE beta;

/*  Distribute vectors and matrices from 0 process over process grid */
#ifdef USE_LOCAL_ARRAYS
        generate_random_values(A_local, mp*nq);
        generate_random_values(x1_local, mp);
        generate_random_values(x2_local, mp);
        generate_random_values(y1_local, mp);
        generate_random_values(y2_local, mp);
 
#else

int n = nx;
  /* Variable declaration/allocation. */
  POLYBENCH_2D_ARRAY_DECL(A, DATA_TYPE, N, N, n, n);
  POLYBENCH_1D_ARRAY_DECL(x1, DATA_TYPE, N, n);
  POLYBENCH_1D_ARRAY_DECL(x2, DATA_TYPE, N, n);
  POLYBENCH_1D_ARRAY_DECL(y1, DATA_TYPE, N, n);
  POLYBENCH_1D_ARRAY_DECL(y2, DATA_TYPE, N, n);

  /* Initialize array(s). */
  init_array (n,
	      POLYBENCH_ARRAY(x1),
	      POLYBENCH_ARRAY(x2),
	      POLYBENCH_ARRAY(y1),
	      POLYBENCH_ARRAY(y2),
	      POLYBENCH_ARRAY(A));

       pdgeadd_( &trans_N, &nx_mkl, &ny_mkl, &one, 
            A, &i_one, &i_one, desc_A,
            &zero, 
            A_local, &i_one, &i_one, desc_A_local );

    pdgeadd_( &trans_N, &nx_mkl, &i_one, &one, x1, &i_one, 
        &i_one, desc_x1, &zero, x1_local, &i_one, &i_one, desc_x1_local );
    
    pdgeadd_( &trans_N, &nx_mkl, &i_one, &one, x2, &i_one, 
        &i_one, desc_x2, &zero, x2_local, &i_one, &i_one, desc_x2_local );

    pdgeadd_( &trans_N, &nx_mkl, &i_one, &one, y1, &i_one, 
        &i_one, desc_y1, &zero, y1_local, &i_one, &i_one, desc_y1_local );
 
    pdgeadd_( &trans_N, &nx_mkl, &i_one, &one, y2, &i_one, 
        &i_one, desc_y2, &zero, y2_local, &i_one, &i_one, desc_y2_local );
    
#endif

  /* Start timer. */
  polybench_start_instruments;

  // x1  = x1 +  (A * y1)
    pdgemv_(&trans_N, &nx_mkl, &ny_mkl, &one, 
            A_local, &i_one, &i_one, desc_A_local,
            y1_local, &i_one, &i_one, desc_y1_local, &i_one,
            &one,
            x1_local, &i_one, &i_one, desc_x1_local, &i_one);

  // x2  = x2 +  (A` * y2)
    pdgemv_(&trans_T, &nx_mkl, &ny_mkl, &one, 
            A_local, &i_one, &i_one, desc_A_local,
            y2_local, &i_one, &i_one, desc_y2_local, &i_one,
            &one,
            x2_local, &i_one, &i_one, desc_x2_local, &i_one);

  /* Stop and print timer. */
  polybench_stop_instruments;
  polybench_print_instruments;


#ifndef USE_LOCAL_ARRAYS
    pdgeadd_(&trans_N, &ny_mkl ,&i_one, &one, 
            x1_local, &i_one, &i_one, desc_x1_local,
            &zero,
            x1, &i_one, &i_one, desc_x1);

    pdgeadd_(&trans_N, &ny_mkl ,&i_one, &one, 
            x2_local, &i_one, &i_one, desc_x2_local,
            &zero,
            x2, &i_one, &i_one, desc_x2);

if(my_rank == 0) {
      /* Prevent dead-code elimination. All live-out data must be printed
         by the function call in argument. */
      polybench_prevent_dce(print_array(n, POLYBENCH_ARRAY(x1), POLYBENCH_ARRAY(x2)));

}
  /* Be clean. */
  POLYBENCH_FREE_ARRAY(A);
  POLYBENCH_FREE_ARRAY(x1);
  POLYBENCH_FREE_ARRAY(x2);
  POLYBENCH_FREE_ARRAY(y1);
  POLYBENCH_FREE_ARRAY(y2);


#endif

    blacs_gridexit_( &ictxt );
    blacs_exit_( &i_zero );

}
int main(int argc, char** argv)
{
  
  int nx, ny;
  int num_procs = atoi(argv[1]);

#ifndef USE_LOCAL_ARRAYS
      nx = N;
      ny = N;
#else
      nx = sqrt(num_procs) * N;
      ny = sqrt(num_procs) * N;
#endif

  /* Run kernel. */
  scalapack_kernel_mvt (nx, ny,
           num_procs);

  return 0;
}
