#include "3mm.dyn_data_dist.h"
#include <assert.h>
#include <omp.h>
#include <math.h>
#define ceild(n,d)  ceil(((double)(n))/((double)(d)))
#define floord(n,d) floor(((double)(n))/((double)(d)))
#define max(x,y)    ((x) > (y)? (x) : (y))
#define min(x,y)    ((x) < (y)? (x) : (y))

/**
 * 3mm.c: This file is part of the PolyBench/C 3.2 test suite.
 *
 *
 * Contact: Louis-Noel Pouchet <pouchet@cse.ohio-state.edu>
 * Web address: http://polybench.sourceforge.net
 */
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <math.h>

/* Include polybench common header. */
#include <polybench.h>

#define NI 2048
#define NJ 2048
#define NK 2048
#define NM 2048
#define NL 2048

/* Include benchmark-specific header. */
/* Default data type is double, default size is 4000. */
#include "3mm.h"


/* Array initialization. */
static
void init_array(int ni, int nj, int nk, int nl, int nm,
		DATA_TYPE POLYBENCH_2D(A,NI,NK,ni,nk),
		DATA_TYPE POLYBENCH_2D(B,NK,NJ,nk,nj),
		DATA_TYPE POLYBENCH_2D(C,NJ,NM,nj,nm),
		DATA_TYPE POLYBENCH_2D(D,NM,NL,nm,nl))
{
  int i, j;

  for (i = 0; i < ni; i++)
    for (j = 0; j < nk; j++)
      A[i][j] = ((DATA_TYPE) i*j) / ni;
  for (i = 0; i < nk; i++)
    for (j = 0; j < nj; j++)
      B[i][j] = ((DATA_TYPE) i*(j+1)) / nj;
  for (i = 0; i < nj; i++)
    for (j = 0; j < nm; j++)
      C[i][j] = ((DATA_TYPE) i*(j+3)) / nl;
  for (i = 0; i < nm; i++)
    for (j = 0; j < nl; j++)
      D[i][j] = ((DATA_TYPE) i*(j+2)) / nk;
}


/* DCE code. Must scan the entire live-out data.
   Can be used also to check the correctness of the output. */
static
void print_array(int ni, int nl,
		 DATA_TYPE POLYBENCH_2D(G,NI,NL,ni,nl))
{
  int i, j;

  for (i = 0; i < ni; i++)
    for (j = 0; j < nl; j++) {
	fprintf (stderr, DATA_PRINTF_MODIFIER, G[i][j]);
	if ((i * ni + j) % 20 == 0) fprintf (stderr, "\n");
    }
  fprintf (stderr, "\n");
}

#define __DECLARATION_OF_E DATA_TYPE POLYBENCH_2D(E,NI,NJ,ni,nj)
#define __DECLARATION_OF_A DATA_TYPE POLYBENCH_2D(A,NI,NK,ni,nk)
#define __DECLARATION_OF_B DATA_TYPE POLYBENCH_2D(B,NK,NJ,nk,nj)
#define __DECLARATION_OF_F DATA_TYPE POLYBENCH_2D(F,NJ,NL,nj,nl)
#define __DECLARATION_OF_C DATA_TYPE POLYBENCH_2D(C,NJ,NM,nj,nm)
#define __DECLARATION_OF_D DATA_TYPE POLYBENCH_2D(D,NM,NL,nm,nl)
#define __DECLARATION_OF_G DATA_TYPE POLYBENCH_2D(G,NI,NL,ni,nl)

#define __PLACE_TO_INSERT_FORWARD_DECLARATIONS
void read_in_copy_F_2(int ts1,int ts2 
#ifndef USE_LOCAL_ARRAYS
 ,__DECLARATION_OF_F 
#endif
 , double** F_trans,int F_size_1, BufferManager *buff_mang_F);
void read_in_copy_C_2(int ts1,int ts2 
#ifndef USE_LOCAL_ARRAYS
 ,__DECLARATION_OF_C 
#endif
 , double** C_trans,int C_size_1, BufferManager *buff_mang_C);
void read_in_copy_D_2(int ts1,int ts2 
#ifndef USE_LOCAL_ARRAYS
 ,__DECLARATION_OF_D 
#endif
 , double** D_trans,int D_size_1, BufferManager *buff_mang_D);
void read_in_copy_E_3(int ts1,int ts2,int ts3 
#ifndef USE_LOCAL_ARRAYS
 ,__DECLARATION_OF_E 
#endif
 , double** E_trans,int E_size_1, BufferManager *buff_mang_E);
void read_in_copy_A_3(int ts1,int ts2,int ts3 
#ifndef USE_LOCAL_ARRAYS
 ,__DECLARATION_OF_A 
#endif
 , double** A_trans,int A_size_1, BufferManager *buff_mang_A);
void read_in_copy_B_3(int ts1,int ts2,int ts3 
#ifndef USE_LOCAL_ARRAYS
 ,__DECLARATION_OF_B 
#endif
 , double** B_trans,int B_size_1, BufferManager *buff_mang_B);
void read_in_copy_G_3(int ts1,int ts2,int ts3 
#ifndef USE_LOCAL_ARRAYS
 ,__DECLARATION_OF_G 
#endif
 , double** G_trans,int G_size_1, BufferManager *buff_mang_G);
void read_in_copy_F_3(int ts1,int ts2,int ts3 
#ifndef USE_LOCAL_ARRAYS
 ,__DECLARATION_OF_F 
#endif
 , double** F_trans,int F_size_1, BufferManager *buff_mang_F);
void data_tile_init_0(int my_rank, int nprocs 
#ifndef USE_LOCAL_ARRAYS
 ,__DECLARATION_OF_E,__DECLARATION_OF_A,__DECLARATION_OF_B,__DECLARATION_OF_F,__DECLARATION_OF_C,__DECLARATION_OF_D,__DECLARATION_OF_G 
#endif
 , double** E_trans,int E_size_1, BufferManager *buff_mang_E, double** A_trans,int A_size_1, BufferManager *buff_mang_A, double** B_trans,int B_size_1, BufferManager *buff_mang_B, double** F_trans,int F_size_1, BufferManager *buff_mang_F, double** C_trans,int C_size_1, BufferManager *buff_mang_C, double** D_trans,int D_size_1, BufferManager *buff_mang_D, double** G_trans,int G_size_1, BufferManager *buff_mang_G);
FILE *__debug_print_fp;
tbb::concurrent_priority_queue<__Task, __Compare> pqueue;
int max_num_tasks_loop0_dim0;
int lb_tasks_loop0_dim0;
int *tasks_loop0;
int max_num_tasks_loop1_dim0;
int lb_tasks_loop1_dim0;
int *tasks_loop1;
int max_num_tasks_loop2_dim0;
int lb_tasks_loop2_dim0;
int *tasks_loop2;
int max_num_tasks_loop3_dim0;
int lb_tasks_loop3_dim0;
int max_num_tasks_loop3_dim1;
int lb_tasks_loop3_dim1;
int *tasks_loop3;
void data_tile_copy_back_E_0(int ts1,int ts2 
#ifndef USE_LOCAL_ARRAYS
 ,__DECLARATION_OF_E 
#endif
 , double** E_trans,int E_size_1, BufferManager *buff_mang_E);
void data_tile_copy_back_F_1(int ts1,int ts2 
#ifndef USE_LOCAL_ARRAYS
 ,__DECLARATION_OF_F 
#endif
 , double** F_trans,int F_size_1, BufferManager *buff_mang_F);
void data_tile_copy_back_G_1(int ts1,int ts2 
#ifndef USE_LOCAL_ARRAYS
 ,__DECLARATION_OF_G 
#endif
 , double** G_trans,int G_size_1, BufferManager *buff_mang_G);
void data_tile_copy_back_F_2(int ts1,int ts2 
#ifndef USE_LOCAL_ARRAYS
 ,__DECLARATION_OF_F 
#endif
 , double** F_trans,int F_size_1, BufferManager *buff_mang_F);
void data_tile_copy_back_E_3(int ts1,int ts2,int ts3 
#ifndef USE_LOCAL_ARRAYS
 ,__DECLARATION_OF_E 
#endif
 , double** E_trans,int E_size_1, BufferManager *buff_mang_E);
void data_tile_copy_back_G_3(int ts1,int ts2,int ts3 
#ifndef USE_LOCAL_ARRAYS
 ,__DECLARATION_OF_G 
#endif
 , double** G_trans,int G_size_1, BufferManager *buff_mang_G);
void copy_back_writeout_0(int my_rank, int nprocs 
#ifndef USE_LOCAL_ARRAYS
 ,__DECLARATION_OF_E,__DECLARATION_OF_F,__DECLARATION_OF_G 
#endif
 , double** E_trans,int E_size_1, BufferManager *buff_mang_E, double** F_trans,int F_size_1, BufferManager *buff_mang_F, double** G_trans,int G_size_1, BufferManager *buff_mang_G);

/* Main computational kernel. The whole function will be timed,
   including the call and return. */
static
void kernel_3mm(int ni, int nj, int nk, int nl, int nm,
		DATA_TYPE POLYBENCH_2D(E,NI,NJ,ni,nj),
		DATA_TYPE POLYBENCH_2D(A,NI,NK,ni,nk),
		DATA_TYPE POLYBENCH_2D(B,NK,NJ,nk,nj),
		DATA_TYPE POLYBENCH_2D(F,NJ,NL,nj,nl),
		DATA_TYPE POLYBENCH_2D(C,NJ,NM,nj,nm),
		DATA_TYPE POLYBENCH_2D(D,NM,NL,nm,nl),
		DATA_TYPE POLYBENCH_2D(G,NI,NL,ni,nl))
{
  int i, j, k;


 int _num_threads;
#pragma omp parallel
{_num_threads = omp_get_num_threads();}
 int __tid;
 long int _num_tasks_to_execute;
 int waiting_pops_start[_num_threads], waiting_pops[_num_threads], waiting_pops_threads;
 for (__tid=0; __tid<_num_threads; __tid++) waiting_pops[__tid] = 0;
 double t_local;
 double t_tasks_create;
 double t_tasks_manage_start[_num_threads], t_tasks_manage[_num_threads], t_tasks_manage_threads;
 for (__tid=0; __tid<_num_threads; __tid++) t_tasks_manage[__tid] = 0;
 double t_comp_start[_num_threads], t_comp[_num_threads], t_comp_threads_min, t_comp_threads;
 for (__tid=0; __tid<_num_threads; __tid++) t_comp[__tid] = 0;
 double t_wait_comp_start[_num_threads], t_wait_comp[_num_threads], t_wait_comp_threads;
 for (__tid=0; __tid<_num_threads; __tid++) t_wait_comp[__tid] = 0;
 double t_data_mang_start[_num_threads], t_data_mang[_num_threads], t_data_mang_threads;
 for (__tid=0; __tid<_num_threads; __tid++) t_data_mang[__tid] = 0;
 double t_writeout_start, t_writeout = 0.0;
 double t_data_init_start, t_data_init = 0.0;
int G_size_0 = (+1*(floorf((+1*_PB_NI-1)/(float)32) - 0 + 1));
int G_size_1 = (+1*(floorf((+1*_PB_NL-1)/(float)32) - 0 + 1));
double **G_trans = (double **)malloc((G_size_0 * G_size_1) * sizeof(double));
BufferManager *buff_mang_G = init_buffer_manager(G_size_0 * G_size_1);
int D_size_0 = (+1*(floorf((+1*_PB_NM-1)/(float)32) - 0 + 1));
int D_size_1 = (+1*(floorf((+1*_PB_NL-1)/(float)32) - 0 + 1));
double **D_trans = (double **)malloc((D_size_0 * D_size_1) * sizeof(double));
BufferManager *buff_mang_D = init_buffer_manager(D_size_0 * D_size_1);
int C_size_0 = (+1*(floorf((+1*_PB_NJ-1)/(float)32) - 0 + 1));
int C_size_1 = (+1*(floorf((+1*_PB_NM-1)/(float)32) - 0 + 1));
double **C_trans = (double **)malloc((C_size_0 * C_size_1) * sizeof(double));
BufferManager *buff_mang_C = init_buffer_manager(C_size_0 * C_size_1);
int F_size_0 = (+1*(floorf((+1*_PB_NJ-1)/(float)32) - 0 + 1));
int F_size_1 = (+1*(floorf((+1*_PB_NL-1)/(float)32) - 0 + 1));
double **F_trans = (double **)malloc((F_size_0 * F_size_1) * sizeof(double));
BufferManager *buff_mang_F = init_buffer_manager(F_size_0 * F_size_1);
int B_size_0 = (+1*(floorf((+1*_PB_NK-1)/(float)32) - 0 + 1));
int B_size_1 = (+1*(floorf((+1*_PB_NJ-1)/(float)32) - 0 + 1));
double **B_trans = (double **)malloc((B_size_0 * B_size_1) * sizeof(double));
BufferManager *buff_mang_B = init_buffer_manager(B_size_0 * B_size_1);
int A_size_0 = (+1*(floorf((+1*_PB_NI-1)/(float)32) - 0 + 1));
int A_size_1 = (+1*(floorf((+1*_PB_NK-1)/(float)32) - 0 + 1));
double **A_trans = (double **)malloc((A_size_0 * A_size_1) * sizeof(double));
BufferManager *buff_mang_A = init_buffer_manager(A_size_0 * A_size_1);
int E_size_0 = (+1*(floorf((+1*_PB_NI-1)/(float)32) - 0 + 1));
int E_size_1 = (+1*(floorf((+1*_PB_NJ-1)/(float)32) - 0 + 1));
double **E_trans = (double **)malloc((E_size_0 * E_size_1) * sizeof(double));
BufferManager *buff_mang_E = init_buffer_manager(E_size_0 * E_size_1);
max_num_tasks_loop0_dim0 = ceil(floorf((+1*_PB_NJ-1)/(float)32) - 0 + 1);
lb_tasks_loop0_dim0 = floor(0);
tasks_loop0 = (int *)malloc(max_num_tasks_loop0_dim0*sizeof(int));
max_num_tasks_loop1_dim0 = ceil(floorf((+1*_PB_NJ-1)/(float)32) - 0 + 1);
lb_tasks_loop1_dim0 = floor(0);
tasks_loop1 = (int *)malloc(max_num_tasks_loop1_dim0*sizeof(int));
max_num_tasks_loop2_dim0 = ceil(floorf((+1*_PB_NJ-1)/(float)32) - 0 + 1);
lb_tasks_loop2_dim0 = floor(0);
tasks_loop2 = (int *)malloc(max_num_tasks_loop2_dim0*sizeof(int));
max_num_tasks_loop3_dim0 = ceil(floorf((+1*_PB_NI+1*_PB_NJ-2)/(float)32) - -1 + 1);
lb_tasks_loop3_dim0 = floor(-1);
max_num_tasks_loop3_dim1 = ceil(floorf((+1*_PB_NI-1)/(float)32) - 0 + 1);
lb_tasks_loop3_dim1 = floor(0);
tasks_loop3 = (int *)malloc(max_num_tasks_loop3_dim0*max_num_tasks_loop3_dim1*sizeof(int));
 IF_TIME(t_data_init_start = rtclock());
data_tile_init_0(0, 0 
#ifndef USE_LOCAL_ARRAYS
 , E , A , B , F , C , D , G 
#endif
 , E_trans, E_size_1, buff_mang_E, A_trans, A_size_1, buff_mang_A, B_trans, B_size_1, buff_mang_B, F_trans, F_size_1, buff_mang_F, C_trans, C_size_1, buff_mang_C, D_trans, D_size_1, buff_mang_D, G_trans, G_size_1, buff_mang_G);
 IF_TIME(t_data_init += rtclock() - t_data_init_start)



/*Start of dynamic scheduling code*/
 IF_TIME(t_local = rtclock());


#if defined(__DYNSCHEDULER_DEBUG_PRINT) || defined(__DYNSCHEDULER_MORE_DEBUG_PRINT) || defined(__DEBUG_FLUSH)
 char *debug_print_name = (char *)malloc(128*sizeof(char));
 strcpy(debug_print_name, "debug_print");
 sprintf(debug_print_name+strlen(debug_print_name), "_node%d", my_rank);
 __debug_print_fp = fopen(debug_print_name, "w");
 free(debug_print_name);
#endif
 IF_TIME(t_tasks_create = rtclock());
 init_tasks(_num_threads,&_num_tasks_to_execute,tasks_loop0,tasks_loop1,tasks_loop2,tasks_loop3);
 IF_TIME(t_tasks_create = rtclock() - t_tasks_create);
#ifdef __DYNSCHEDULER_MORE_DEBUG_PRINT
 std::vector<int> prev_firing_count_E_0;
 std::vector<int> prev_firing_count_E_1;
 std::vector<int> prev_firing_count_E_2;
 std::vector<int> prev_firing_count_E_3;
 std::vector<int> prev_firing_count_F_0;
 std::vector<int> prev_firing_count_F_1;
 std::vector<int> prev_firing_count_F_2;
 std::vector<int> prev_firing_count_F_3;
 std::vector<int> prev_firing_count_G_0;
 std::vector<int> prev_firing_count_G_1;
 std::vector<int> prev_firing_count_G_2;
 std::vector<int> prev_firing_count_G_3;
 for (__p=0; __p<__MAX_NUM_RECVS; __p++) {
  prev_firing_count_E_0.push_back(-1);
  prev_firing_count_E_1.push_back(-1);
  prev_firing_count_E_2.push_back(-1);
  prev_firing_count_E_3.push_back(-1);
  prev_firing_count_F_0.push_back(-1);
  prev_firing_count_F_1.push_back(-1);
  prev_firing_count_F_2.push_back(-1);
  prev_firing_count_F_3.push_back(-1);
  prev_firing_count_G_0.push_back(-1);
  prev_firing_count_G_1.push_back(-1);
  prev_firing_count_G_2.push_back(-1);
  prev_firing_count_G_3.push_back(-1);
 }
#endif

#pragma omp parallel private(__tid)
{
 __tid = omp_get_thread_num();
 int __flag;
 int __loop;
 int task_id1,task_id2,task_id3;

{ /* compute-pack-send threads */
#ifdef __DYNSCHEDULER_DEBUG_PRINT
 fprintf(__debug_print_fp, "node %d compute thread %d num tasks to execute %ld\n", my_rank, __tid, _num_tasks_to_execute);
 IF_DEBUG_FLUSH(fflush(__debug_print_fp));
#endif
 IF_TIME(t_wait_comp_start[__tid] = rtclock());
 while (_num_tasks_to_execute > 0) {
  __Task task;
  __flag = pqueue.try_pop(task);
  if(__flag==0) { IF_TIME(waiting_pops[__tid]++); continue; }
  __loop = task.loop;
  task_id1 = task.id1;
  task_id2 = task.id2;
  task_id3 = task.id3;
#ifdef __DYNSCHEDULER_DEBUG_PRINT
  long int captured;
#pragma omp atomic capture
  captured = --_num_tasks_to_execute;
  fprintf(__debug_print_fp, "computing node %d thread %d loop %d task %d %d %d ready tasks %lu\n", my_rank, __tid, __loop, task_id1, task_id2, task_id3, pqueue.size());
  IF_DEBUG_FLUSH(fflush(__debug_print_fp));
#else
#pragma omp atomic
  _num_tasks_to_execute--;
#endif
  switch(__loop){
   case 0:
    IF_TIME(t_wait_comp[__tid] += rtclock() - t_wait_comp_start[__tid]);
    IF_TIME(t_data_mang_start[__tid] = rtclock());
    data_tile_alloc_0(task_id1,task_id2, E_trans, E_size_1, buff_mang_E);
    IF_TIME(t_data_mang[__tid] += rtclock() - t_data_mang_start[__tid]);
    IF_TIME(t_comp_start[__tid] = rtclock());
    compute_task_0(task_id1,task_id2, E_trans, E_size_1, buff_mang_E);
    IF_TIME(t_comp[__tid] += rtclock() - t_comp_start[__tid]);
    IF_TIME(t_data_mang_start[__tid] = rtclock());
    data_tile_ref_count_update_0(task_id1,task_id2, E_trans, E_size_1, buff_mang_E);
    IF_TIME(t_data_mang[__tid] += rtclock() - t_data_mang_start[__tid]);
    IF_TIME(t_tasks_manage_start[__tid] = rtclock());
    local_update_dep_tasks_0(task_id1,task_id2,_PB_NI,_PB_NJ,_PB_NK,_PB_NL,_PB_NM,_num_threads,tasks_loop0,tasks_loop1,tasks_loop2,tasks_loop3);
    IF_TIME(t_tasks_manage[__tid] += rtclock() - t_tasks_manage_start[__tid]);
    IF_TIME(t_wait_comp_start[__tid] = rtclock());
    break;
   case 1:
    IF_TIME(t_wait_comp[__tid] += rtclock() - t_wait_comp_start[__tid]);
    IF_TIME(t_data_mang_start[__tid] = rtclock());
    data_tile_alloc_1(task_id1,task_id2, F_trans, F_size_1, buff_mang_F, G_trans, G_size_1, buff_mang_G);
    IF_TIME(t_data_mang[__tid] += rtclock() - t_data_mang_start[__tid]);
    IF_TIME(t_comp_start[__tid] = rtclock());
    compute_task_1(task_id1,task_id2, F_trans, F_size_1, buff_mang_F, G_trans, G_size_1, buff_mang_G);
    IF_TIME(t_comp[__tid] += rtclock() - t_comp_start[__tid]);
    IF_TIME(t_data_mang_start[__tid] = rtclock());
    data_tile_ref_count_update_1(task_id1,task_id2, F_trans, F_size_1, buff_mang_F, G_trans, G_size_1, buff_mang_G);
    IF_TIME(t_data_mang[__tid] += rtclock() - t_data_mang_start[__tid]);
    IF_TIME(t_tasks_manage_start[__tid] = rtclock());
    local_update_dep_tasks_1(task_id1,task_id2,_PB_NI,_PB_NJ,_PB_NK,_PB_NL,_PB_NM,_num_threads,tasks_loop0,tasks_loop1,tasks_loop2,tasks_loop3);
    IF_TIME(t_tasks_manage[__tid] += rtclock() - t_tasks_manage_start[__tid]);
    IF_TIME(t_wait_comp_start[__tid] = rtclock());
    break;
   case 2:
    IF_TIME(t_wait_comp[__tid] += rtclock() - t_wait_comp_start[__tid]);
    IF_TIME(t_data_mang_start[__tid] = rtclock());
    data_tile_alloc_2(task_id1,task_id2, F_trans, F_size_1, buff_mang_F, C_trans, C_size_1, buff_mang_C, D_trans, D_size_1, buff_mang_D);
    IF_TIME(t_data_mang[__tid] += rtclock() - t_data_mang_start[__tid]);
    IF_TIME(t_comp_start[__tid] = rtclock());
    compute_task_2(task_id1,task_id2, F_trans, F_size_1, buff_mang_F, C_trans, C_size_1, buff_mang_C, D_trans, D_size_1, buff_mang_D);
    IF_TIME(t_comp[__tid] += rtclock() - t_comp_start[__tid]);
    IF_TIME(t_data_mang_start[__tid] = rtclock());
    data_tile_ref_count_update_2(task_id1,task_id2, F_trans, F_size_1, buff_mang_F, C_trans, C_size_1, buff_mang_C, D_trans, D_size_1, buff_mang_D);
    IF_TIME(t_data_mang[__tid] += rtclock() - t_data_mang_start[__tid]);
    IF_TIME(t_tasks_manage_start[__tid] = rtclock());
    local_update_dep_tasks_2(task_id1,task_id2,_PB_NI,_PB_NJ,_PB_NK,_PB_NL,_PB_NM,_num_threads,tasks_loop0,tasks_loop1,tasks_loop2,tasks_loop3);
    IF_TIME(t_tasks_manage[__tid] += rtclock() - t_tasks_manage_start[__tid]);
    IF_TIME(t_wait_comp_start[__tid] = rtclock());
    break;
   case 3:
    IF_TIME(t_wait_comp[__tid] += rtclock() - t_wait_comp_start[__tid]);
    IF_TIME(t_data_mang_start[__tid] = rtclock());
    data_tile_alloc_3(task_id1,task_id2,task_id3, E_trans, E_size_1, buff_mang_E, A_trans, A_size_1, buff_mang_A, B_trans, B_size_1, buff_mang_B, G_trans, G_size_1, buff_mang_G, F_trans, F_size_1, buff_mang_F);
    IF_TIME(t_data_mang[__tid] += rtclock() - t_data_mang_start[__tid]);
    IF_TIME(t_comp_start[__tid] = rtclock());
    compute_task_3(task_id1,task_id2,task_id3, E_trans, E_size_1, buff_mang_E, A_trans, A_size_1, buff_mang_A, B_trans, B_size_1, buff_mang_B, G_trans, G_size_1, buff_mang_G, F_trans, F_size_1, buff_mang_F);
    IF_TIME(t_comp[__tid] += rtclock() - t_comp_start[__tid]);
    IF_TIME(t_data_mang_start[__tid] = rtclock());
    data_tile_ref_count_update_3(task_id1,task_id2,task_id3, E_trans, E_size_1, buff_mang_E, A_trans, A_size_1, buff_mang_A, B_trans, B_size_1, buff_mang_B, G_trans, G_size_1, buff_mang_G, F_trans, F_size_1, buff_mang_F);
    IF_TIME(t_data_mang[__tid] += rtclock() - t_data_mang_start[__tid]);
    IF_TIME(t_tasks_manage_start[__tid] = rtclock());
    local_update_dep_tasks_3(task_id1,task_id2,task_id3,_PB_NI,_PB_NJ,_PB_NK,_PB_NL,_PB_NM,_num_threads,tasks_loop0,tasks_loop1,tasks_loop2,tasks_loop3);
    IF_TIME(t_tasks_manage[__tid] += rtclock() - t_tasks_manage_start[__tid]);
    IF_TIME(t_wait_comp_start[__tid] = rtclock());
    break;
   default:
    assert(0);
  }
#ifdef __DYNSCHEDULER_MORE_DEBUG_PRINT
  fprintf(__debug_print_fp, "computed node %d thread %d loop %d task %d %d %d tasks left %ld\n", my_rank, __tid, __loop, task_id1, task_id2, task_id3, captured);
  IF_DEBUG_FLUSH(fflush(__debug_print_fp));
#endif
 }

#ifdef __DYNSCHEDULER_DEBUG_PRINT
 fprintf(__debug_print_fp, "node %d compute thread %d done\n", my_rank, __tid);
 IF_DEBUG_FLUSH(fflush(__debug_print_fp));
#endif
}

}

#ifdef __DYNSCHEDULER_DEBUG_PRINT
 fclose(__debug_print_fp);
#endif



/*End of dynamic scheduling code*/


 IF_TIME(t_writeout = rtclock());
 
#ifndef USE_LOCAL_ARRAYS
 if (fopen(".test", "r")) {
 copy_back_writeout_0(0, 0 
#ifndef USE_LOCAL_ARRAYS
 , E , F , G 
#endif
 , E_trans, E_size_1, buff_mang_E, F_trans, F_size_1, buff_mang_F, G_trans, G_size_1, buff_mang_G);

} 
#endif
 IF_TIME(t_writeout = rtclock() - t_writeout);
free_buffer_mang(buff_mang_E);
free_buffer_mang(buff_mang_A);
free_buffer_mang(buff_mang_B);
free_buffer_mang(buff_mang_F);
free_buffer_mang(buff_mang_C);
free_buffer_mang(buff_mang_D);
free_buffer_mang(buff_mang_G);
#ifdef TIME
 waiting_pops_threads = waiting_pops[0];
                for (__tid=1; __tid<_num_threads; __tid++) if (waiting_pops_threads < waiting_pops[__tid]) waiting_pops_threads = waiting_pops[__tid];
 t_local = rtclock() - t_local;
 t_tasks_manage_threads = t_tasks_manage[0];
                for (__tid=1; __tid<_num_threads; __tid++) if (t_tasks_manage_threads < t_tasks_manage[__tid]) t_tasks_manage_threads = t_tasks_manage[__tid];
 t_comp_threads_min = t_comp[0];
                for (__tid=1; __tid<_num_threads; __tid++) if (t_comp_threads_min > t_comp[__tid]) t_comp_threads_min = t_comp[__tid];
 t_comp_threads = t_comp[0];
                for (__tid=1; __tid<_num_threads; __tid++) if (t_comp_threads < t_comp[__tid]) t_comp_threads = t_comp[__tid];
 t_wait_comp_threads = t_wait_comp[0];
                for (__tid=1; __tid<_num_threads; __tid++) if (t_wait_comp_threads < t_wait_comp[__tid]) t_wait_comp_threads = t_wait_comp[__tid];
 char buffer[4096];
 strcpy(buffer, "");
 sprintf(buffer+strlen(buffer), "Minimum computation time across all threads: %lf\n", t_comp_threads_min);

 sprintf(buffer+strlen(buffer), "Maximum computation time across all threads: %lf\n", t_comp_threads);

 sprintf(buffer+strlen(buffer), "Maximum 'waiting for computation' time across all threads: %lf\n", t_wait_comp_threads);

 sprintf(buffer+strlen(buffer), "Maximum number of waiting/unsuccessful POPs across all threads: %d\n", waiting_pops_threads);

 sprintf(buffer+strlen(buffer), "Maximum tasks management time across all threads: %lf\n", t_tasks_manage_threads);

 sprintf(buffer+strlen(buffer), "Tasks creation time: %lf\n", t_tasks_create);

 sprintf(buffer+strlen(buffer), "Sum of split times: %lf\n", t_comp_threads+t_wait_comp_threads+t_tasks_manage_threads+t_tasks_create);

 sprintf(buffer+strlen(buffer), "Total time: %lf\n", t_local);

 fprintf(stdout, "%s\n", buffer);
#endif

}


int main(int argc, char** argv)
{
  /* Retrieve problem size. */
  int ni = NI;
  int nj = NJ;
  int nk = NK;
  int nl = NL;
  int nm = NM;

  /* Variable declaration/allocation. */
  POLYBENCH_2D_ARRAY_DECL(E, DATA_TYPE, NI, NJ, ni, nj);
  POLYBENCH_2D_ARRAY_DECL(A, DATA_TYPE, NI, NK, ni, nk);
  POLYBENCH_2D_ARRAY_DECL(B, DATA_TYPE, NK, NJ, nk, nj);
  POLYBENCH_2D_ARRAY_DECL(F, DATA_TYPE, NJ, NL, nj, nl);
  POLYBENCH_2D_ARRAY_DECL(C, DATA_TYPE, NJ, NM, nj, nm);
  POLYBENCH_2D_ARRAY_DECL(D, DATA_TYPE, NM, NL, nm, nl);
  POLYBENCH_2D_ARRAY_DECL(G, DATA_TYPE, NI, NL, ni, nl);

  /* Initialize array(s). */
  init_array (ni, nj, nk, nl, nm,
	      POLYBENCH_ARRAY(A),
	      POLYBENCH_ARRAY(B),
	      POLYBENCH_ARRAY(C),
	      POLYBENCH_ARRAY(D));

  /* Start timer. */
  polybench_start_instruments;

  /* Run kernel. */
  kernel_3mm (ni, nj, nk, nl, nm,
	      POLYBENCH_ARRAY(E),
	      POLYBENCH_ARRAY(A),
	      POLYBENCH_ARRAY(B),
	      POLYBENCH_ARRAY(F),
	      POLYBENCH_ARRAY(C),
	      POLYBENCH_ARRAY(D),
	      POLYBENCH_ARRAY(G));

  /* Stop and print timer. */
  polybench_stop_instruments;
  polybench_print_instruments;

  /* Prevent dead-code elimination. All live-out data must be printed
     by the function call in argument. */
  polybench_prevent_dce(print_array(ni, nl,  POLYBENCH_ARRAY(G)));

  /* Be clean. */
  POLYBENCH_FREE_ARRAY(E);
  POLYBENCH_FREE_ARRAY(A);
  POLYBENCH_FREE_ARRAY(B);
  POLYBENCH_FREE_ARRAY(F);
  POLYBENCH_FREE_ARRAY(C);
  POLYBENCH_FREE_ARRAY(D);
  POLYBENCH_FREE_ARRAY(G);

  return 0;
}
void data_tile_ref_count_init_E_0(int ts1,int ts2, double** E_trans,int E_size_1, BufferManager *buff_mang_E)
{


  int t1, t2, t3;

 int lbp, ubp;
 register int lbv, ubv;

/* Start of CLooG code */
if ((_PB_NI >= 1) && (ts1 == 2) && (ts2 >= 0) && (32*ts2 <= _PB_NJ-1)) {
  for (t2=0;t2<=floord(_PB_NI-1,32);t2++) {
    increment_ref_count(buff_mang_E, (t2)* ( E_size_1 ) + (ts2));;
  }
}
/* End of CLooG code */

}

void data_tile_ref_count_init_F_1(int ts1,int ts2, double** F_trans,int F_size_1, BufferManager *buff_mang_F)
{


  int t1, t2, t3;

 int lbp, ubp;
 register int lbv, ubv;

/* Start of CLooG code */
if ((_PB_NL >= 1) && (ts1 == 0) && (ts2 >= 0) && (32*ts2 <= _PB_NJ-1)) {
  for (t3=0;t3<=floord(_PB_NL-1,32);t3++) {
    increment_ref_count(buff_mang_F, (ts2)* ( F_size_1 ) + (t3));;
  }
}
/* End of CLooG code */

}

void data_tile_ref_count_init_G_1(int ts1,int ts2, double** G_trans,int G_size_1, BufferManager *buff_mang_G)
{


  int t1, t2, t3;

 int lbp, ubp;
 register int lbv, ubv;

/* Start of CLooG code */
if ((_PB_NL >= 1) && (ts1 == 0) && (ts2 >= 0) && (32*ts2 <= _PB_NI-1)) {
  for (t3=0;t3<=floord(_PB_NL-1,32);t3++) {
    increment_ref_count(buff_mang_G, (ts2)* ( G_size_1 ) + (t3));;
  }
}
/* End of CLooG code */

}

void data_tile_ref_count_init_F_2(int ts1,int ts2, double** F_trans,int F_size_1, BufferManager *buff_mang_F)
{


  int t1, t2, t3;

 int lbp, ubp;
 register int lbv, ubv;

/* Start of CLooG code */
if ((_PB_NL >= 1) && (_PB_NM >= 1) && (ts1 == 1) && (ts2 >= 0) && (32*ts2 <= _PB_NJ-1)) {
  for (t3=0;t3<=floord(_PB_NL-1,32);t3++) {
    increment_ref_count(buff_mang_F, (ts2)* ( F_size_1 ) + (t3));;
  }
}
/* End of CLooG code */

}

void data_tile_ref_count_init_C_2(int ts1,int ts2, double** C_trans,int C_size_1, BufferManager *buff_mang_C)
{


  int t1, t2, t3;

 int lbp, ubp;
 register int lbv, ubv;

/* Start of CLooG code */
if ((_PB_NL >= 1) && (_PB_NM >= 1) && (ts1 == 1) && (ts2 >= 0) && (32*ts2 <= _PB_NJ-1)) {
  for (t3=0;t3<=floord(_PB_NM-1,32);t3++) {
    increment_ref_count(buff_mang_C, (ts2)* ( C_size_1 ) + (t3));;
  }
}
/* End of CLooG code */

}

void data_tile_ref_count_init_D_2(int ts1,int ts2, double** D_trans,int D_size_1, BufferManager *buff_mang_D)
{


  int t1, t2, t3;

 int lbp, ubp;
 register int lbv, ubv;

/* Start of CLooG code */
if ((_PB_NL >= 1) && (_PB_NM >= 1) && (ts1 == 1) && (ts2 >= 0) && (32*ts2 <= _PB_NJ-1)) {
  for (t2=0;t2<=floord(_PB_NM-1,32);t2++) {
    for (t3=0;t3<=floord(_PB_NL-1,32);t3++) {
      increment_ref_count(buff_mang_D, (t2)* ( D_size_1 ) + (t3));;
    }
  }
}
/* End of CLooG code */

}

void read_in_data_tile_alloc_F_2(int ts1,int ts2, double** F_trans,int F_size_1, BufferManager *buff_mang_F)
{


  int t1, t2, t3;

 int lbp, ubp;
 register int lbv, ubv;

/* Start of CLooG code */
if ((_PB_NL >= 1) && (_PB_NM >= 1) && (ts1 == 1) && (ts2 >= 0) && (32*ts2 <= _PB_NJ-1)) {
  for (t3=0;t3<=floord(_PB_NL-1,32);t3++) {
    if(F_trans[(ts2) * ((F_size_1)) + (t3)] == NULL) F_trans[(ts2)* ( F_size_1 ) + (t3)] = (double *) buffer_alloc_atomic((ts2)* ( F_size_1 ) + (t3), 32*32 * sizeof(double), buff_mang_F);;
  }
}
/* End of CLooG code */

}

void read_in_copy_F_2(int ts1,int ts2 
#ifndef USE_LOCAL_ARRAYS
 ,__DECLARATION_OF_F 
#endif
 , double** F_trans,int F_size_1, BufferManager *buff_mang_F)
{


  int t1, t2, t3;

 int lbp, ubp;
 register int lbv, ubv;

/* Start of CLooG code */
if ((_PB_NL >= 1) && (_PB_NM >= 1) && (ts1 == 1) && (ts2 >= 0) && (32*ts2 <= _PB_NJ-1)) {
  for (t2=32*ts2;t2<=min(_PB_NJ-1,32*ts2+31);t2++) {
    for (t3=0;t3<=_PB_NL-1;t3++) {
      
#ifndef USE_LOCAL_ARRAYS
 F_trans[((1)*( t2 ))/32* ( F_size_1 ) + ((1)*( t3 ))/32][((1)*( t2 ))%32* (32) + ((1)*( t3 ))%32] = F[t2][t3]; 
#else
 
#endif
;
    }
  }
}
/* End of CLooG code */

}

void read_in_data_tile_alloc_C_2(int ts1,int ts2, double** C_trans,int C_size_1, BufferManager *buff_mang_C)
{


  int t1, t2, t3;

 int lbp, ubp;
 register int lbv, ubv;

/* Start of CLooG code */
if ((_PB_NL >= 1) && (_PB_NM >= 1) && (ts1 == 1) && (ts2 >= 0) && (32*ts2 <= _PB_NJ-1)) {
  for (t3=0;t3<=floord(_PB_NM-1,32);t3++) {
    if(C_trans[(ts2) * ((C_size_1)) + (t3)] == NULL) C_trans[(ts2)* ( C_size_1 ) + (t3)] = (double *) buffer_alloc_atomic((ts2)* ( C_size_1 ) + (t3), 32*32 * sizeof(double), buff_mang_C);;
  }
}
/* End of CLooG code */

}

void read_in_copy_C_2(int ts1,int ts2 
#ifndef USE_LOCAL_ARRAYS
 ,__DECLARATION_OF_C 
#endif
 , double** C_trans,int C_size_1, BufferManager *buff_mang_C)
{


  int t1, t2, t3;

 int lbp, ubp;
 register int lbv, ubv;

/* Start of CLooG code */
if ((_PB_NL >= 1) && (_PB_NM >= 1) && (ts1 == 1) && (ts2 >= 0) && (32*ts2 <= _PB_NJ-1)) {
  for (t2=32*ts2;t2<=min(_PB_NJ-1,32*ts2+31);t2++) {
    for (t3=0;t3<=_PB_NM-1;t3++) {
      
#ifndef USE_LOCAL_ARRAYS
 C_trans[((1)*( t2 ))/32* ( C_size_1 ) + ((1)*( t3 ))/32][((1)*( t2 ))%32* (32) + ((1)*( t3 ))%32] = C[t2][t3]; 
#else
 
#endif
;
    }
  }
}
/* End of CLooG code */

}

void read_in_data_tile_alloc_D_2(int ts1,int ts2, double** D_trans,int D_size_1, BufferManager *buff_mang_D)
{


  int t1, t2, t3;

 int lbp, ubp;
 register int lbv, ubv;

/* Start of CLooG code */
if ((_PB_NL >= 1) && (_PB_NM >= 1) && (ts1 == 1) && (ts2 >= 0) && (32*ts2 <= _PB_NJ-1)) {
  for (t2=0;t2<=floord(_PB_NM-1,32);t2++) {
    for (t3=0;t3<=floord(_PB_NL-1,32);t3++) {
      if(D_trans[(t2) * ((D_size_1)) + (t3)] == NULL) D_trans[(t2)* ( D_size_1 ) + (t3)] = (double *) buffer_alloc_atomic((t2)* ( D_size_1 ) + (t3), 32*32 * sizeof(double), buff_mang_D);;
    }
  }
}
/* End of CLooG code */

}

void read_in_copy_D_2(int ts1,int ts2 
#ifndef USE_LOCAL_ARRAYS
 ,__DECLARATION_OF_D 
#endif
 , double** D_trans,int D_size_1, BufferManager *buff_mang_D)
{


  int t1, t2, t3;

 int lbp, ubp;
 register int lbv, ubv;

/* Start of CLooG code */
if ((_PB_NL >= 1) && (_PB_NM >= 1) && (ts1 == 1) && (ts2 >= 0) && (32*ts2 <= _PB_NJ-1)) {
  for (t2=0;t2<=_PB_NM-1;t2++) {
    for (t3=0;t3<=_PB_NL-1;t3++) {
      
#ifndef USE_LOCAL_ARRAYS
 D_trans[((1)*( t2 ))/32* ( D_size_1 ) + ((1)*( t3 ))/32][((1)*( t2 ))%32* (32) + ((1)*( t3 ))%32] = D[t2][t3]; 
#else
 
#endif
;
    }
  }
}
/* End of CLooG code */

}

void data_tile_ref_count_init_E_3(int ts1,int ts2,int ts3, double** E_trans,int E_size_1, BufferManager *buff_mang_E)
{


  int t1, t2, t3;

 int lbp, ubp;
 register int lbv, ubv;

/* Start of CLooG code */
if ((ts1 == 3) && (ts2 >= ts3) && (32*ts2 <= 32*ts3+_PB_NJ-1) && (ts3 >= 0) && (32*ts3 <= _PB_NI-1)) {
  if (_PB_NL >= 1) {
    increment_ref_count(buff_mang_E, (ts3)* ( E_size_1 ) + (ts2-ts3));;
  }
  if ((_PB_NK >= 1) && (_PB_NL <= 0)) {
    increment_ref_count(buff_mang_E, (ts3)* ( E_size_1 ) + (ts2-ts3));;
  }
}
/* End of CLooG code */

}

void data_tile_ref_count_init_A_3(int ts1,int ts2,int ts3, double** A_trans,int A_size_1, BufferManager *buff_mang_A)
{


  int t1, t2, t3;

 int lbp, ubp;
 register int lbv, ubv;

/* Start of CLooG code */
if ((_PB_NK >= 1) && (ts1 == 3) && (ts2 >= ts3) && (32*ts2 <= 32*ts3+_PB_NJ-1) && (ts3 >= 0) && (32*ts3 <= _PB_NI-1)) {
  for (t3=0;t3<=floord(_PB_NK-1,32);t3++) {
    increment_ref_count(buff_mang_A, (ts3)* ( A_size_1 ) + (t3));;
  }
}
/* End of CLooG code */

}

void data_tile_ref_count_init_B_3(int ts1,int ts2,int ts3, double** B_trans,int B_size_1, BufferManager *buff_mang_B)
{


  int t1, t2, t3;

 int lbp, ubp;
 register int lbv, ubv;

/* Start of CLooG code */
if ((_PB_NK >= 1) && (ts1 == 3) && (ts2 >= ts3) && (32*ts2 <= 32*ts3+_PB_NJ-1) && (ts3 >= 0) && (32*ts3 <= _PB_NI-1)) {
  for (t2=0;t2<=floord(_PB_NK-1,32);t2++) {
    increment_ref_count(buff_mang_B, (t2)* ( B_size_1 ) + (ts2-ts3));;
  }
}
/* End of CLooG code */

}

void data_tile_ref_count_init_G_3(int ts1,int ts2,int ts3, double** G_trans,int G_size_1, BufferManager *buff_mang_G)
{


  int t1, t2, t3;

 int lbp, ubp;
 register int lbv, ubv;

/* Start of CLooG code */
if ((_PB_NL >= 1) && (ts1 == 3) && (ts2 >= ts3) && (32*ts2 <= 32*ts3+_PB_NJ-1) && (ts3 >= 0) && (32*ts3 <= _PB_NI-1)) {
  for (t3=0;t3<=floord(_PB_NL-1,32);t3++) {
    increment_ref_count(buff_mang_G, (ts3)* ( G_size_1 ) + (t3));;
  }
}
/* End of CLooG code */

}

void data_tile_ref_count_init_F_3(int ts1,int ts2,int ts3, double** F_trans,int F_size_1, BufferManager *buff_mang_F)
{


  int t1, t2, t3;

 int lbp, ubp;
 register int lbv, ubv;

/* Start of CLooG code */
if ((_PB_NL >= 1) && (ts1 == 3) && (ts2 >= ts3) && (32*ts2 <= 32*ts3+_PB_NJ-1) && (ts3 >= 0) && (32*ts3 <= _PB_NI-1)) {
  for (t3=0;t3<=floord(_PB_NL-1,32);t3++) {
    increment_ref_count(buff_mang_F, (ts2-ts3)* ( F_size_1 ) + (t3));;
  }
}
/* End of CLooG code */

}

void read_in_data_tile_alloc_E_3(int ts1,int ts2,int ts3, double** E_trans,int E_size_1, BufferManager *buff_mang_E)
{


  int t1, t2, t3;

 int lbp, ubp;
 register int lbv, ubv;

/* Start of CLooG code */
if ((ts1 == 3) && (ts2 >= ts3) && (32*ts2 <= 32*ts3+_PB_NJ-1) && (ts3 >= 0) && (32*ts3 <= _PB_NI-1)) {
  if (_PB_NL >= 1) {
    if(E_trans[(ts3) * ((E_size_1)) + (ts2-ts3)] == NULL) E_trans[(ts3)* ( E_size_1 ) + (ts2-ts3)] = (double *) buffer_alloc_atomic((ts3)* ( E_size_1 ) + (ts2-ts3), 32*32 * sizeof(double), buff_mang_E);;
  }
  if ((_PB_NK >= 1) && (_PB_NL <= 0)) {
    if(E_trans[(ts3) * ((E_size_1)) + (ts2-ts3)] == NULL) E_trans[(ts3)* ( E_size_1 ) + (ts2-ts3)] = (double *) buffer_alloc_atomic((ts3)* ( E_size_1 ) + (ts2-ts3), 32*32 * sizeof(double), buff_mang_E);;
  }
}
/* End of CLooG code */

}

void read_in_copy_E_3(int ts1,int ts2,int ts3 
#ifndef USE_LOCAL_ARRAYS
 ,__DECLARATION_OF_E 
#endif
 , double** E_trans,int E_size_1, BufferManager *buff_mang_E)
{


  int t1, t2, t3;

 int lbp, ubp;
 register int lbv, ubv;

/* Start of CLooG code */
if ((ts1 == 3) && (ts2 >= ts3) && (32*ts2 <= 32*ts3+_PB_NJ-1) && (ts3 >= 0) && (32*ts3 <= _PB_NI-1)) {
  if (_PB_NL >= 1) {
    for (t2=32*ts3;t2<=min(_PB_NI-1,32*ts3+31);t2++) {
      for (t3=32*ts2-32*ts3;t3<=min(_PB_NJ-1,32*ts2-32*ts3+31);t3++) {
        
#ifndef USE_LOCAL_ARRAYS
 E_trans[((1)*( t2 ))/32* ( E_size_1 ) + ((1)*( t3 ))/32][((1)*( t2 ))%32* (32) + ((1)*( t3 ))%32] = E[t2][t3]; 
#else
 
#endif
;
      }
    }
  }
  if ((_PB_NK >= 1) && (_PB_NL <= 0)) {
    for (t2=32*ts3;t2<=min(_PB_NI-1,32*ts3+31);t2++) {
      for (t3=32*ts2-32*ts3;t3<=min(_PB_NJ-1,32*ts2-32*ts3+31);t3++) {
        
#ifndef USE_LOCAL_ARRAYS
 E_trans[((1)*( t2 ))/32* ( E_size_1 ) + ((1)*( t3 ))/32][((1)*( t2 ))%32* (32) + ((1)*( t3 ))%32] = E[t2][t3]; 
#else
 
#endif
;
      }
    }
  }
}
/* End of CLooG code */

}

void read_in_data_tile_alloc_A_3(int ts1,int ts2,int ts3, double** A_trans,int A_size_1, BufferManager *buff_mang_A)
{


  int t1, t2, t3;

 int lbp, ubp;
 register int lbv, ubv;

/* Start of CLooG code */
if ((_PB_NK >= 1) && (ts1 == 3) && (ts2 >= ts3) && (32*ts2 <= 32*ts3+_PB_NJ-1) && (ts3 >= 0) && (32*ts3 <= _PB_NI-1)) {
  for (t3=0;t3<=floord(_PB_NK-1,32);t3++) {
    if(A_trans[(ts3) * ((A_size_1)) + (t3)] == NULL) A_trans[(ts3)* ( A_size_1 ) + (t3)] = (double *) buffer_alloc_atomic((ts3)* ( A_size_1 ) + (t3), 32*32 * sizeof(double), buff_mang_A);;
  }
}
/* End of CLooG code */

}

void read_in_copy_A_3(int ts1,int ts2,int ts3 
#ifndef USE_LOCAL_ARRAYS
 ,__DECLARATION_OF_A 
#endif
 , double** A_trans,int A_size_1, BufferManager *buff_mang_A)
{


  int t1, t2, t3;

 int lbp, ubp;
 register int lbv, ubv;

/* Start of CLooG code */
if ((_PB_NK >= 1) && (ts1 == 3) && (ts2 >= ts3) && (32*ts2 <= 32*ts3+_PB_NJ-1) && (ts3 >= 0) && (32*ts3 <= _PB_NI-1)) {
  for (t2=32*ts3;t2<=min(_PB_NI-1,32*ts3+31);t2++) {
    for (t3=0;t3<=_PB_NK-1;t3++) {
      
#ifndef USE_LOCAL_ARRAYS
 A_trans[((1)*( t2 ))/32* ( A_size_1 ) + ((1)*( t3 ))/32][((1)*( t2 ))%32* (32) + ((1)*( t3 ))%32] = A[t2][t3]; 
#else
 
#endif
;
    }
  }
}
/* End of CLooG code */

}

void read_in_data_tile_alloc_B_3(int ts1,int ts2,int ts3, double** B_trans,int B_size_1, BufferManager *buff_mang_B)
{


  int t1, t2, t3;

 int lbp, ubp;
 register int lbv, ubv;

/* Start of CLooG code */
if ((_PB_NK >= 1) && (ts1 == 3) && (ts2 >= ts3) && (32*ts2 <= 32*ts3+_PB_NJ-1) && (ts3 >= 0) && (32*ts3 <= _PB_NI-1)) {
  for (t2=0;t2<=floord(_PB_NK-1,32);t2++) {
    if(B_trans[(t2) * ((B_size_1)) + (ts2-ts3)] == NULL) B_trans[(t2)* ( B_size_1 ) + (ts2-ts3)] = (double *) buffer_alloc_atomic((t2)* ( B_size_1 ) + (ts2-ts3), 32*32 * sizeof(double), buff_mang_B);;
  }
}
/* End of CLooG code */

}

void read_in_copy_B_3(int ts1,int ts2,int ts3 
#ifndef USE_LOCAL_ARRAYS
 ,__DECLARATION_OF_B 
#endif
 , double** B_trans,int B_size_1, BufferManager *buff_mang_B)
{


  int t1, t2, t3;

 int lbp, ubp;
 register int lbv, ubv;

/* Start of CLooG code */
if ((_PB_NK >= 1) && (ts1 == 3) && (ts2 >= ts3) && (32*ts2 <= 32*ts3+_PB_NJ-1) && (ts3 >= 0) && (32*ts3 <= _PB_NI-1)) {
  for (t2=0;t2<=_PB_NK-1;t2++) {
    for (t3=32*ts2-32*ts3;t3<=min(_PB_NJ-1,32*ts2-32*ts3+31);t3++) {
      
#ifndef USE_LOCAL_ARRAYS
 B_trans[((1)*( t2 ))/32* ( B_size_1 ) + ((1)*( t3 ))/32][((1)*( t2 ))%32* (32) + ((1)*( t3 ))%32] = B[t2][t3]; 
#else
 
#endif
;
    }
  }
}
/* End of CLooG code */

}

void read_in_data_tile_alloc_G_3(int ts1,int ts2,int ts3, double** G_trans,int G_size_1, BufferManager *buff_mang_G)
{


  int t1, t2, t3;

 int lbp, ubp;
 register int lbv, ubv;

/* Start of CLooG code */
if ((_PB_NL >= 1) && (ts1 == 3) && (ts2 >= ts3) && (32*ts2 <= 32*ts3+_PB_NJ-1) && (ts3 >= 0) && (32*ts3 <= _PB_NI-1)) {
  for (t3=0;t3<=floord(_PB_NL-1,32);t3++) {
    if(G_trans[(ts3) * ((G_size_1)) + (t3)] == NULL) G_trans[(ts3)* ( G_size_1 ) + (t3)] = (double *) buffer_alloc_atomic((ts3)* ( G_size_1 ) + (t3), 32*32 * sizeof(double), buff_mang_G);;
  }
}
/* End of CLooG code */

}

void read_in_copy_G_3(int ts1,int ts2,int ts3 
#ifndef USE_LOCAL_ARRAYS
 ,__DECLARATION_OF_G 
#endif
 , double** G_trans,int G_size_1, BufferManager *buff_mang_G)
{


  int t1, t2, t3;

 int lbp, ubp;
 register int lbv, ubv;

/* Start of CLooG code */
if ((_PB_NL >= 1) && (ts1 == 3) && (ts2 >= ts3) && (32*ts2 <= 32*ts3+_PB_NJ-1) && (ts3 >= 0) && (32*ts3 <= _PB_NI-1)) {
  for (t2=32*ts3;t2<=min(_PB_NI-1,32*ts3+31);t2++) {
    for (t3=0;t3<=_PB_NL-1;t3++) {
      
#ifndef USE_LOCAL_ARRAYS
 G_trans[((1)*( t2 ))/32* ( G_size_1 ) + ((1)*( t3 ))/32][((1)*( t2 ))%32* (32) + ((1)*( t3 ))%32] = G[t2][t3]; 
#else
 
#endif
;
    }
  }
}
/* End of CLooG code */

}

void read_in_data_tile_alloc_F_3(int ts1,int ts2,int ts3, double** F_trans,int F_size_1, BufferManager *buff_mang_F)
{


  int t1, t2, t3;

 int lbp, ubp;
 register int lbv, ubv;

/* Start of CLooG code */
if ((_PB_NL >= 1) && (ts1 == 3) && (ts2 >= ts3) && (32*ts2 <= 32*ts3+_PB_NJ-1) && (ts3 >= 0) && (32*ts3 <= _PB_NI-1)) {
  for (t3=0;t3<=floord(_PB_NL-1,32);t3++) {
    if(F_trans[(ts2-ts3) * ((F_size_1)) + (t3)] == NULL) F_trans[(ts2-ts3)* ( F_size_1 ) + (t3)] = (double *) buffer_alloc_atomic((ts2-ts3)* ( F_size_1 ) + (t3), 32*32 * sizeof(double), buff_mang_F);;
  }
}
/* End of CLooG code */

}

void read_in_copy_F_3(int ts1,int ts2,int ts3 
#ifndef USE_LOCAL_ARRAYS
 ,__DECLARATION_OF_F 
#endif
 , double** F_trans,int F_size_1, BufferManager *buff_mang_F)
{


  int t1, t2, t3;

 int lbp, ubp;
 register int lbv, ubv;

/* Start of CLooG code */
if ((_PB_NL >= 1) && (ts1 == 3) && (ts2 >= ts3) && (32*ts2 <= 32*ts3+_PB_NJ-1) && (ts3 >= 0) && (32*ts3 <= _PB_NI-1)) {
  for (t2=32*ts2-32*ts3;t2<=min(_PB_NJ-1,32*ts2-32*ts3+31);t2++) {
    for (t3=0;t3<=_PB_NL-1;t3++) {
      
#ifndef USE_LOCAL_ARRAYS
 F_trans[((1)*( t2 ))/32* ( F_size_1 ) + ((1)*( t3 ))/32][((1)*( t2 ))%32* (32) + ((1)*( t3 ))%32] = F[t2][t3]; 
#else
 
#endif
;
    }
  }
}
/* End of CLooG code */

}

void data_tile_init_0(int my_rank, int nprocs 
#ifndef USE_LOCAL_ARRAYS
 ,__DECLARATION_OF_E,__DECLARATION_OF_A,__DECLARATION_OF_B,__DECLARATION_OF_F,__DECLARATION_OF_C,__DECLARATION_OF_D,__DECLARATION_OF_G 
#endif
 , double** E_trans,int E_size_1, BufferManager *buff_mang_E, double** A_trans,int A_size_1, BufferManager *buff_mang_A, double** B_trans,int B_size_1, BufferManager *buff_mang_B, double** F_trans,int F_size_1, BufferManager *buff_mang_F, double** C_trans,int C_size_1, BufferManager *buff_mang_C, double** D_trans,int D_size_1, BufferManager *buff_mang_D, double** G_trans,int G_size_1, BufferManager *buff_mang_G)
{
  int t1, t2, t3, t4;
 int lbp, ubp;
 register int lbv, ubv;
/* Start of CLooG code */
if (_PB_NL >= 1) {
  for (t2=0;t2<=floord(_PB_NI-1,32);t2++) {
    for (t3=0;t3<=floord(_PB_NL-1,32);t3++) {
      G_trans[(t2)* ( G_size_1 ) + (t3)] = NULL;;
    }
  }
}
if ((_PB_NJ >= 1) && (_PB_NL >= 1)) {
  for (t2=0;t2<=floord(_PB_NM-1,32);t2++) {
    for (t3=0;t3<=floord(_PB_NL-1,32);t3++) {
      D_trans[(t2)* ( D_size_1 ) + (t3)] = NULL;;
    }
  }
}
if ((_PB_NL >= 1) && (_PB_NM >= 1)) {
  for (t2=0;t2<=floord(_PB_NJ-1,32);t2++) {
    for (t3=0;t3<=floord(_PB_NM-1,32);t3++) {
      C_trans[(t2)* ( C_size_1 ) + (t3)] = NULL;;
    }
  }
}
if (_PB_NL >= 1) {
  for (t2=0;t2<=floord(_PB_NJ-1,32);t2++) {
    for (t3=0;t3<=floord(_PB_NL-1,32);t3++) {
      F_trans[(t2)* ( F_size_1 ) + (t3)] = NULL;;
    }
  }
}
if ((_PB_NI >= 1) && (_PB_NJ >= 1)) {
  for (t2=0;t2<=floord(_PB_NK-1,32);t2++) {
    for (t3=0;t3<=floord(_PB_NJ-1,32);t3++) {
      B_trans[(t2)* ( B_size_1 ) + (t3)] = NULL;;
    }
  }
}
if ((_PB_NJ >= 1) && (_PB_NK >= 1)) {
  for (t2=0;t2<=floord(_PB_NI-1,32);t2++) {
    for (t3=0;t3<=floord(_PB_NK-1,32);t3++) {
      A_trans[(t2)* ( A_size_1 ) + (t3)] = NULL;;
    }
  }
}
if (_PB_NJ >= 1) {
  for (t2=0;t2<=floord(_PB_NI-1,32);t2++) {
    for (t3=0;t3<=floord(_PB_NJ-1,32);t3++) {
      E_trans[(t2)* ( E_size_1 ) + (t3)] = NULL;;
    }
  }
}
if ((_PB_NI >= 1) && (_PB_NJ >= 1)) {
  for (t3=0;t3<=floord(_PB_NJ-1,32);t3++) {
    data_tile_ref_count_init_E_0(2, t3, E_trans, E_size_1,buff_mang_E);;
  }
}
if ((_PB_NI >= 1) && (_PB_NJ >= 1)) {
  for (t3=0;t3<=floord(_PB_NJ-1,32);t3++) {
    data_tile_ref_count_init_E_0(2, t3, E_trans, E_size_1,buff_mang_E);;
  }
}
if ((_PB_NJ >= 1) && (_PB_NL >= 1)) {
  for (t3=0;t3<=floord(_PB_NJ-1,32);t3++) {
    data_tile_ref_count_init_F_1(0, t3, F_trans, F_size_1,buff_mang_F);;
  }
}
if ((_PB_NI >= 1) && (_PB_NL >= 1)) {
  for (t3=0;t3<=floord(_PB_NI-1,32);t3++) {
    data_tile_ref_count_init_G_1(0, t3, G_trans, G_size_1,buff_mang_G);;
  }
}
if ((_PB_NJ >= 1) && (_PB_NL >= 1)) {
  for (t3=0;t3<=floord(_PB_NJ-1,32);t3++) {
    data_tile_ref_count_init_F_1(0, t3, F_trans, F_size_1,buff_mang_F);;
  }
}
if ((_PB_NI >= 1) && (_PB_NL >= 1)) {
  for (t3=0;t3<=floord(_PB_NI-1,32);t3++) {
    data_tile_ref_count_init_G_1(0, t3, G_trans, G_size_1,buff_mang_G);;
  }
}
if ((_PB_NJ >= 1) && (_PB_NL >= 1) && (_PB_NM >= 1)) {
  for (t3=0;t3<=floord(_PB_NJ-1,32);t3++) {
    data_tile_ref_count_init_F_2(1, t3, F_trans, F_size_1,buff_mang_F);;
  }
}
if ((_PB_NJ >= 1) && (_PB_NL >= 1) && (_PB_NM >= 1)) {
  for (t3=0;t3<=floord(_PB_NJ-1,32);t3++) {
    data_tile_ref_count_init_C_2(1, t3, C_trans, C_size_1,buff_mang_C);;
  }
}
if ((_PB_NJ >= 1) && (_PB_NL >= 1) && (_PB_NM >= 1)) {
  for (t3=0;t3<=floord(_PB_NJ-1,32);t3++) {
    data_tile_ref_count_init_D_2(1, t3, D_trans, D_size_1,buff_mang_D);;
  }
}
if ((_PB_NJ >= 1) && (_PB_NL >= 1) && (_PB_NM >= 1)) {
  for (t3=0;t3<=floord(_PB_NJ-1,32);t3++) {
    data_tile_ref_count_init_F_2(1, t3, F_trans, F_size_1,buff_mang_F);;
  }
}
if ((_PB_NJ >= 1) && (_PB_NL >= 1) && (_PB_NM >= 1)) {
  for (t3=0;t3<=floord(_PB_NJ-1,32);t3++) {
    read_in_data_tile_alloc_F_2(1, t3, F_trans, F_size_1,buff_mang_F);read_in_copy_F_2(1, t3 
#ifndef USE_LOCAL_ARRAYS
, F 
#endif
, F_trans, F_size_1,buff_mang_F);;
  }
}
if ((_PB_NJ >= 1) && (_PB_NL >= 1) && (_PB_NM >= 1)) {
  for (t3=0;t3<=floord(_PB_NJ-1,32);t3++) {
    read_in_data_tile_alloc_C_2(1, t3, C_trans, C_size_1,buff_mang_C);read_in_copy_C_2(1, t3 
#ifndef USE_LOCAL_ARRAYS
, C 
#endif
, C_trans, C_size_1,buff_mang_C);;
  }
}
if ((_PB_NJ >= 1) && (_PB_NL >= 1) && (_PB_NM >= 1)) {
  for (t3=0;t3<=floord(_PB_NJ-1,32);t3++) {
    read_in_data_tile_alloc_D_2(1, t3, D_trans, D_size_1,buff_mang_D);read_in_copy_D_2(1, t3 
#ifndef USE_LOCAL_ARRAYS
, D 
#endif
, D_trans, D_size_1,buff_mang_D);;
  }
}
if ((_PB_NI >= 1) && (_PB_NJ >= 1) && (_PB_NK >= 1)) {
  for (t3=0;t3<=floord(_PB_NI+_PB_NJ-2,32);t3++) {
    for (t4=max(0,ceild(32*t3-_PB_NJ+1,32));t4<=min(floord(_PB_NI-1,32),t3);t4++) {
      data_tile_ref_count_init_E_3(3, t3, t4, E_trans, E_size_1,buff_mang_E);;
    }
  }
}
if ((_PB_NI >= 1) && (_PB_NJ >= 1) && (_PB_NK >= 1)) {
  for (t3=0;t3<=floord(_PB_NI+_PB_NJ-2,32);t3++) {
    for (t4=max(0,ceild(32*t3-_PB_NJ+1,32));t4<=min(floord(_PB_NI-1,32),t3);t4++) {
      data_tile_ref_count_init_A_3(3, t3, t4, A_trans, A_size_1,buff_mang_A);;
    }
  }
}
if ((_PB_NI >= 1) && (_PB_NJ >= 1) && (_PB_NK >= 1)) {
  for (t3=0;t3<=floord(_PB_NI+_PB_NJ-2,32);t3++) {
    for (t4=max(0,ceild(32*t3-_PB_NJ+1,32));t4<=min(floord(_PB_NI-1,32),t3);t4++) {
      data_tile_ref_count_init_B_3(3, t3, t4, B_trans, B_size_1,buff_mang_B);;
    }
  }
}
if ((_PB_NI >= 1) && (_PB_NJ >= 1) && (_PB_NL >= 1)) {
  for (t3=0;t3<=floord(_PB_NI+_PB_NJ-2,32);t3++) {
    for (t4=max(0,ceild(32*t3-_PB_NJ+1,32));t4<=min(floord(_PB_NI-1,32),t3);t4++) {
      data_tile_ref_count_init_G_3(3, t3, t4, G_trans, G_size_1,buff_mang_G);;
    }
  }
}
if ((_PB_NI >= 1) && (_PB_NJ >= 1) && (_PB_NL >= 1)) {
  for (t3=0;t3<=floord(_PB_NI+_PB_NJ-2,32);t3++) {
    for (t4=max(0,ceild(32*t3-_PB_NJ+1,32));t4<=min(floord(_PB_NI-1,32),t3);t4++) {
      data_tile_ref_count_init_F_3(3, t3, t4, F_trans, F_size_1,buff_mang_F);;
    }
  }
}
if ((_PB_NI >= 1) && (_PB_NJ >= 1) && (_PB_NK >= 1)) {
  for (t3=0;t3<=floord(_PB_NI+_PB_NJ-2,32);t3++) {
    for (t4=max(0,ceild(32*t3-_PB_NJ+1,32));t4<=min(floord(_PB_NI-1,32),t3);t4++) {
      data_tile_ref_count_init_E_3(3, t3, t4, E_trans, E_size_1,buff_mang_E);;
    }
  }
}
if ((_PB_NI >= 1) && (_PB_NJ >= 1) && (_PB_NL >= 1)) {
  for (t3=0;t3<=floord(_PB_NI+_PB_NJ-2,32);t3++) {
    for (t4=max(0,ceild(32*t3-_PB_NJ+1,32));t4<=min(floord(_PB_NI-1,32),t3);t4++) {
      data_tile_ref_count_init_G_3(3, t3, t4, G_trans, G_size_1,buff_mang_G);;
    }
  }
}
if ((_PB_NI >= 1) && (_PB_NJ >= 1) && (_PB_NK >= 1)) {
  for (t3=0;t3<=floord(_PB_NI+_PB_NJ-2,32);t3++) {
    for (t4=max(0,ceild(32*t3-_PB_NJ+1,32));t4<=min(floord(_PB_NI-1,32),t3);t4++) {
      read_in_data_tile_alloc_E_3(3, t3, t4, E_trans, E_size_1,buff_mang_E);read_in_copy_E_3(3, t3, t4 
#ifndef USE_LOCAL_ARRAYS
, E 
#endif
, E_trans, E_size_1,buff_mang_E);;
    }
  }
}
if ((_PB_NI >= 1) && (_PB_NJ >= 1) && (_PB_NK >= 1)) {
  for (t3=0;t3<=floord(_PB_NI+_PB_NJ-2,32);t3++) {
    for (t4=max(0,ceild(32*t3-_PB_NJ+1,32));t4<=min(floord(_PB_NI-1,32),t3);t4++) {
      read_in_data_tile_alloc_A_3(3, t3, t4, A_trans, A_size_1,buff_mang_A);read_in_copy_A_3(3, t3, t4 
#ifndef USE_LOCAL_ARRAYS
, A 
#endif
, A_trans, A_size_1,buff_mang_A);;
    }
  }
}
if ((_PB_NI >= 1) && (_PB_NJ >= 1) && (_PB_NK >= 1)) {
  for (t3=0;t3<=floord(_PB_NI+_PB_NJ-2,32);t3++) {
    for (t4=max(0,ceild(32*t3-_PB_NJ+1,32));t4<=min(floord(_PB_NI-1,32),t3);t4++) {
      read_in_data_tile_alloc_B_3(3, t3, t4, B_trans, B_size_1,buff_mang_B);read_in_copy_B_3(3, t3, t4 
#ifndef USE_LOCAL_ARRAYS
, B 
#endif
, B_trans, B_size_1,buff_mang_B);;
    }
  }
}
if ((_PB_NI >= 1) && (_PB_NJ >= 1) && (_PB_NL >= 1)) {
  for (t3=0;t3<=floord(_PB_NI+_PB_NJ-2,32);t3++) {
    for (t4=max(0,ceild(32*t3-_PB_NJ+1,32));t4<=min(floord(_PB_NI-1,32),t3);t4++) {
      read_in_data_tile_alloc_G_3(3, t3, t4, G_trans, G_size_1,buff_mang_G);read_in_copy_G_3(3, t3, t4 
#ifndef USE_LOCAL_ARRAYS
, G 
#endif
, G_trans, G_size_1,buff_mang_G);;
    }
  }
}
if ((_PB_NI >= 1) && (_PB_NJ >= 1) && (_PB_NL >= 1)) {
  for (t3=0;t3<=floord(_PB_NI+_PB_NJ-2,32);t3++) {
    for (t4=max(0,ceild(32*t3-_PB_NJ+1,32));t4<=min(floord(_PB_NI-1,32),t3);t4++) {
      read_in_data_tile_alloc_F_3(3, t3, t4, F_trans, F_size_1,buff_mang_F);read_in_copy_F_3(3, t3, t4 
#ifndef USE_LOCAL_ARRAYS
, F 
#endif
, F_trans, F_size_1,buff_mang_F);;
    }
  }
}
/* End of CLooG code */
}
void init_tasks(int num_threads,long int *p_num_tasks_to_execute,int *tasks_loop0,int *tasks_loop1,int *tasks_loop2,int *tasks_loop3)
{
int local_dep_tasks, affinity;
long int _num_tasks_to_execute = 0;
  int t1, t2, t3, t4;
 int lbp, ubp;
 register int lbv, ubv;
/* Start of CLooG code */
if ((_PB_NI >= 1) && (_PB_NJ >= 1)) {
  lbp=0;
  ubp=floord(_PB_NJ-1,32);
#pragma omp parallel for private(local_dep_tasks, affinity, t4) reduction(+:_num_tasks_to_execute)
  for (t3=lbp;t3<=ubp;t3++) {
    _num_tasks_to_execute++;tasks_loop0[(t3-lb_tasks_loop0_dim0)+0] = get_num_local_src_tasks_0(2,t3,_PB_NI,_PB_NJ,_PB_NK,_PB_NL,_PB_NM);if (tasks_loop0[(t3-lb_tasks_loop0_dim0)+0] == 0) { local_dep_tasks = get_num_local_dep_tasks_0(2,t3,_PB_NI,_PB_NJ,_PB_NK,_PB_NL,_PB_NM); affinity = pi_0(2,t3,_PB_NI,_PB_NJ,_PB_NK,_PB_NL,_PB_NM,num_threads); __Task task(0,2,t3,local_dep_tasks,affinity); pqueue.push(task); IF_DYNSCHEDULER_MORE_DEBUG_PRINT( fprintf(__debug_print_fp, "added node %d loop 0 task %d %d                 local_dep_tasks %d ready tasks %lu\n", my_rank, 2,t3, local_dep_tasks, pqueue.size())); IF_DEBUG_FLUSH(fflush(__debug_print_fp)); };
  }
}
if (_PB_NL >= 1) {
  lbp=0;
  ubp=floord(_PB_NI-1,32);
#pragma omp parallel for private(local_dep_tasks, affinity, t4) reduction(+:_num_tasks_to_execute)
  for (t3=lbp;t3<=ubp;t3++) {
    _num_tasks_to_execute++;tasks_loop1[(t3-lb_tasks_loop1_dim0)+0] = get_num_local_src_tasks_1(0,t3,_PB_NI,_PB_NJ,_PB_NK,_PB_NL,_PB_NM);if (tasks_loop1[(t3-lb_tasks_loop1_dim0)+0] == 0) { local_dep_tasks = get_num_local_dep_tasks_1(0,t3,_PB_NI,_PB_NJ,_PB_NK,_PB_NL,_PB_NM); affinity = pi_1(0,t3,_PB_NI,_PB_NJ,_PB_NK,_PB_NL,_PB_NM,num_threads); __Task task(1,0,t3,local_dep_tasks,affinity); pqueue.push(task); IF_DYNSCHEDULER_MORE_DEBUG_PRINT( fprintf(__debug_print_fp, "added node %d loop 1 task %d %d                 local_dep_tasks %d ready tasks %lu\n", my_rank, 0,t3, local_dep_tasks, pqueue.size())); IF_DEBUG_FLUSH(fflush(__debug_print_fp)); };
  }
  lbp=max(0,ceild(_PB_NI,32));
  ubp=floord(_PB_NJ-1,32);
#pragma omp parallel for private(local_dep_tasks, affinity) reduction(+:_num_tasks_to_execute)
  for (t3=lbp;t3<=ubp;t3++) {
    _num_tasks_to_execute++;tasks_loop1[(t3-lb_tasks_loop1_dim0)+0] = get_num_local_src_tasks_1(0,t3,_PB_NI,_PB_NJ,_PB_NK,_PB_NL,_PB_NM);if (tasks_loop1[(t3-lb_tasks_loop1_dim0)+0] == 0) { local_dep_tasks = get_num_local_dep_tasks_1(0,t3,_PB_NI,_PB_NJ,_PB_NK,_PB_NL,_PB_NM); affinity = pi_1(0,t3,_PB_NI,_PB_NJ,_PB_NK,_PB_NL,_PB_NM,num_threads); __Task task(1,0,t3,local_dep_tasks,affinity); pqueue.push(task); IF_DYNSCHEDULER_MORE_DEBUG_PRINT( fprintf(__debug_print_fp, "added node %d loop 1 task %d %d                 local_dep_tasks %d ready tasks %lu\n", my_rank, 0,t3, local_dep_tasks, pqueue.size())); IF_DEBUG_FLUSH(fflush(__debug_print_fp)); };
  }
}
if ((_PB_NJ >= 1) && (_PB_NL >= 1) && (_PB_NM >= 1)) {
  lbp=0;
  ubp=floord(_PB_NJ-1,32);
#pragma omp parallel for private(local_dep_tasks, affinity, t4) reduction(+:_num_tasks_to_execute)
  for (t3=lbp;t3<=ubp;t3++) {
    _num_tasks_to_execute++;tasks_loop2[(t3-lb_tasks_loop2_dim0)+0] = get_num_local_src_tasks_2(1,t3,_PB_NI,_PB_NJ,_PB_NK,_PB_NL,_PB_NM);if (tasks_loop2[(t3-lb_tasks_loop2_dim0)+0] == 0) { local_dep_tasks = get_num_local_dep_tasks_2(1,t3,_PB_NI,_PB_NJ,_PB_NK,_PB_NL,_PB_NM); affinity = pi_2(1,t3,_PB_NI,_PB_NJ,_PB_NK,_PB_NL,_PB_NM,num_threads); __Task task(2,1,t3,local_dep_tasks,affinity); pqueue.push(task); IF_DYNSCHEDULER_MORE_DEBUG_PRINT( fprintf(__debug_print_fp, "added node %d loop 2 task %d %d                 local_dep_tasks %d ready tasks %lu\n", my_rank, 1,t3, local_dep_tasks, pqueue.size())); IF_DEBUG_FLUSH(fflush(__debug_print_fp)); };
  }
}
if ((_PB_NI >= 1) && (_PB_NJ >= 1)) {
  if (_PB_NL >= 1) {
    lbp=0;
    ubp=floord(_PB_NI+_PB_NJ-2,32);
#pragma omp parallel for private(local_dep_tasks, affinity, t4) reduction(+:_num_tasks_to_execute)
    for (t3=lbp;t3<=ubp;t3++) {
      for (t4=max(0,ceild(32*t3-_PB_NJ+1,32));t4<=min(floord(_PB_NI-1,32),t3);t4++) {
        _num_tasks_to_execute++;tasks_loop3[(t3-lb_tasks_loop3_dim0)*max_num_tasks_loop3_dim1+(t4-lb_tasks_loop3_dim1)+0] = get_num_local_src_tasks_3(3,t3,t4,_PB_NI,_PB_NJ,_PB_NK,_PB_NL,_PB_NM);if (tasks_loop3[(t3-lb_tasks_loop3_dim0)*max_num_tasks_loop3_dim1+(t4-lb_tasks_loop3_dim1)+0] == 0) { local_dep_tasks = get_num_local_dep_tasks_3(3,t3,t4,_PB_NI,_PB_NJ,_PB_NK,_PB_NL,_PB_NM); affinity = pi_3(3,t3,t4,_PB_NI,_PB_NJ,_PB_NK,_PB_NL,_PB_NM,num_threads); __Task task(3,3,t3,t4,local_dep_tasks,affinity); pqueue.push(task); IF_DYNSCHEDULER_MORE_DEBUG_PRINT( fprintf(__debug_print_fp, "added node %d loop 3 task %d %d %d                 local_dep_tasks %d ready tasks %lu\n", my_rank, 3,t3,t4, local_dep_tasks, pqueue.size())); IF_DEBUG_FLUSH(fflush(__debug_print_fp)); };
      }
    }
  }
  if ((_PB_NK >= 1) && (_PB_NL <= 0)) {
    lbp=0;
    ubp=floord(_PB_NI+_PB_NJ-2,32);
#pragma omp parallel for private(local_dep_tasks, affinity) reduction(+:_num_tasks_to_execute)
    for (t3=lbp;t3<=ubp;t3++) {
      for (t4=max(0,ceild(32*t3-_PB_NJ+1,32));t4<=min(floord(_PB_NI-1,32),t3);t4++) {
        _num_tasks_to_execute++;tasks_loop3[(t3-lb_tasks_loop3_dim0)*max_num_tasks_loop3_dim1+(t4-lb_tasks_loop3_dim1)+0] = get_num_local_src_tasks_3(3,t3,t4,_PB_NI,_PB_NJ,_PB_NK,_PB_NL,_PB_NM);if (tasks_loop3[(t3-lb_tasks_loop3_dim0)*max_num_tasks_loop3_dim1+(t4-lb_tasks_loop3_dim1)+0] == 0) { local_dep_tasks = get_num_local_dep_tasks_3(3,t3,t4,_PB_NI,_PB_NJ,_PB_NK,_PB_NL,_PB_NM); affinity = pi_3(3,t3,t4,_PB_NI,_PB_NJ,_PB_NK,_PB_NL,_PB_NM,num_threads); __Task task(3,3,t3,t4,local_dep_tasks,affinity); pqueue.push(task); IF_DYNSCHEDULER_MORE_DEBUG_PRINT( fprintf(__debug_print_fp, "added node %d loop 3 task %d %d %d                 local_dep_tasks %d ready tasks %lu\n", my_rank, 3,t3,t4, local_dep_tasks, pqueue.size())); IF_DEBUG_FLUSH(fflush(__debug_print_fp)); };
      }
    }
  }
}
/* End of CLooG code */
*p_num_tasks_to_execute = _num_tasks_to_execute;
}
void data_tile_alloc_0(int ts1,int ts2, double** E_trans,int E_size_1, BufferManager *buff_mang_E)
{
  int t1, t2, t3;
 int lbp, ubp;
 register int lbv, ubv;
/* Start of CLooG code */
if ((_PB_NI >= 1) && (ts1 == 2) && (ts2 >= 0) && (32*ts2 <= _PB_NJ-1)) {
  for (t2=0;t2<=floord(_PB_NI-1,32);t2++) {
    if(E_trans[(t2) * ((E_size_1)) + (ts2)] == NULL) E_trans[(t2)* ( E_size_1 ) + (ts2)] = (double *) buffer_alloc_atomic((t2)* ( E_size_1 ) + (ts2), 32*32 * sizeof(double), buff_mang_E);;
  }
}
/* End of CLooG code */
}
void data_tile_ref_count_update_0(int ts1,int ts2, double** E_trans,int E_size_1, BufferManager *buff_mang_E)
{
  int t1, t2, t3;
 int lbp, ubp;
 register int lbv, ubv;
/* Start of CLooG code */
if ((_PB_NI >= 1) && (ts1 == 2) && (ts2 >= 0) && (32*ts2 <= _PB_NJ-1)) {
  for (t2=0;t2<=floord(_PB_NI-1,32);t2++) {
    decrement_ref_count(buff_mang_E, (t2)* ( E_size_1 ) + (ts2));;
  }
}
/* End of CLooG code */
}
void compute_task_0(int t1,int t2, double** E_trans,int E_size_1, BufferManager *buff_mang_E)
{
int i_j_0_div = 0;int i_j_0_mod = 0;double *E_i_j_0 = NULL; int t3, t4, t5, t6, t7, t8, t9, t10, t11, t12, t13, t14, t15, t16, t17;
 int lbp, ubp;
 register int lbv, ubv;
/* Start of CLooG code */
for (t3=0;t3<=floord(_PB_NI-1,32);t3++) {
  for (t4=32*t3;t4<=min(_PB_NI-1,32*t3+31);t4++) {
    lbv=32*t2;
    ubv=min(_PB_NJ-1,32*t2+31);
    #if (defined __DATA_DIST_DECLS || defined USE_LOCAL_ARRAYS)
    i_j_0_div = ((1)*( t4 ))/32* ( E_size_1 ) + ((1)*( lbv ))/32;i_j_0_mod = ((1)*( t4 ))%32* (32) + ((1)*( lbv ))%32;E_i_j_0 = E_trans[i_j_0_div];;
    #endif
#pragma ivdep
#pragma vector always
    for (t17=lbv;t17<=ubv;t17++) {
      E_i_j_0[i_j_0_mod + t17 - lbv]=0;;
    }
  }
}
/* End of CLooG code */
}
void data_tile_alloc_1(int ts1,int ts2, double** F_trans,int F_size_1, BufferManager *buff_mang_F, double** G_trans,int G_size_1, BufferManager *buff_mang_G)
{
  int t1, t2, t3;
 int lbp, ubp;
 register int lbv, ubv;
/* Start of CLooG code */
if ((_PB_NL >= 1) && (ts1 == 0) && (ts2 >= 0)) {
  if (ts2 <= floord(_PB_NJ-1,32)) {
    for (t3=0;t3<=floord(_PB_NL-1,32);t3++) {
      if(F_trans[(ts2) * ((F_size_1)) + (t3)] == NULL) F_trans[(ts2)* ( F_size_1 ) + (t3)] = (double *) buffer_alloc_atomic((ts2)* ( F_size_1 ) + (t3), 32*32 * sizeof(double), buff_mang_F);;
    }
  }
  if (ts2 <= floord(_PB_NI-1,32)) {
    for (t3=0;t3<=floord(_PB_NL-1,32);t3++) {
      if(G_trans[(ts2) * ((G_size_1)) + (t3)] == NULL) G_trans[(ts2)* ( G_size_1 ) + (t3)] = (double *) buffer_alloc_atomic((ts2)* ( G_size_1 ) + (t3), 32*32 * sizeof(double), buff_mang_G);;
    }
  }
}
/* End of CLooG code */
}
void data_tile_ref_count_update_1(int ts1,int ts2, double** F_trans,int F_size_1, BufferManager *buff_mang_F, double** G_trans,int G_size_1, BufferManager *buff_mang_G)
{
  int t1, t2, t3;
 int lbp, ubp;
 register int lbv, ubv;
/* Start of CLooG code */
if ((_PB_NL >= 1) && (ts1 == 0) && (ts2 >= 0)) {
  if (ts2 <= floord(_PB_NJ-1,32)) {
    for (t3=0;t3<=floord(_PB_NL-1,32);t3++) {
      decrement_ref_count(buff_mang_F, (ts2)* ( F_size_1 ) + (t3));;
    }
  }
  if (ts2 <= floord(_PB_NI-1,32)) {
    for (t3=0;t3<=floord(_PB_NL-1,32);t3++) {
      decrement_ref_count(buff_mang_G, (ts2)* ( G_size_1 ) + (t3));;
    }
  }
}
/* End of CLooG code */
}
void compute_task_1(int t1,int t2, double** F_trans,int F_size_1, BufferManager *buff_mang_F, double** G_trans,int G_size_1, BufferManager *buff_mang_G)
{
int i_j_0_div = 0;int i_j_0_mod = 0;double *F_i_j_0 = NULL;int i_j_1_div = 0;int i_j_1_mod = 0;double *G_i_j_1 = NULL; int t3, t4, t5, t6, t7, t8, t9, t10, t11, t12, t13, t14, t15, t16, t17;
 int lbp, ubp;
 register int lbv, ubv;
/* Start of CLooG code */
for (t3=0;t3<=floord(_PB_NL-1,32);t3++) {
  for (t4=32*t2;t4<=min(min(_PB_NI-1,_PB_NJ-1),32*t2+31);t4++) {
    lbv=32*t3;
    ubv=min(_PB_NL-1,32*t3+31);
    #if (defined __DATA_DIST_DECLS || defined USE_LOCAL_ARRAYS)
    i_j_0_div = ((1)*( t4 ))/32* ( F_size_1 ) + ((1)*( lbv ))/32;i_j_0_mod = ((1)*( t4 ))%32* (32) + ((1)*( lbv ))%32;F_i_j_0 = F_trans[i_j_0_div];;
    i_j_1_div = ((1)*( t4 ))/32* ( G_size_1 ) + ((1)*( lbv ))/32;i_j_1_mod = ((1)*( t4 ))%32* (32) + ((1)*( lbv ))%32;G_i_j_1 = G_trans[i_j_1_div];;
    #endif
#pragma ivdep
#pragma vector always
    for (t17=lbv;t17<=ubv;t17++) {
      F_i_j_0[i_j_0_mod + t17 - lbv]=0;;
      G_i_j_1[i_j_1_mod + t17 - lbv]=0;;
    }
  }
  for (t4=max(_PB_NI,32*t2);t4<=min(_PB_NJ-1,32*t2+31);t4++) {
    lbv=32*t3;
    ubv=min(_PB_NL-1,32*t3+31);
    #if (defined __DATA_DIST_DECLS || defined USE_LOCAL_ARRAYS)
    i_j_0_div = ((1)*( t4 ))/32* ( F_size_1 ) + ((1)*( lbv ))/32;i_j_0_mod = ((1)*( t4 ))%32* (32) + ((1)*( lbv ))%32;F_i_j_0 = F_trans[i_j_0_div];;
    #endif
#pragma ivdep
#pragma vector always
    for (t17=lbv;t17<=ubv;t17++) {
      F_i_j_0[i_j_0_mod + t17 - lbv]=0;;
    }
  }
  for (t4=max(_PB_NJ,32*t2);t4<=min(_PB_NI-1,32*t2+31);t4++) {
    lbv=32*t3;
    ubv=min(_PB_NL-1,32*t3+31);
    #if (defined __DATA_DIST_DECLS || defined USE_LOCAL_ARRAYS)
    i_j_1_div = ((1)*( t4 ))/32* ( G_size_1 ) + ((1)*( lbv ))/32;i_j_1_mod = ((1)*( t4 ))%32* (32) + ((1)*( lbv ))%32;G_i_j_1 = G_trans[i_j_1_div];;
    #endif
#pragma ivdep
#pragma vector always
    for (t17=lbv;t17<=ubv;t17++) {
      G_i_j_1[i_j_1_mod + t17 - lbv]=0;;
    }
  }
}
/* End of CLooG code */
}
void data_tile_alloc_2(int ts1,int ts2, double** F_trans,int F_size_1, BufferManager *buff_mang_F, double** C_trans,int C_size_1, BufferManager *buff_mang_C, double** D_trans,int D_size_1, BufferManager *buff_mang_D)
{
  int t1, t2, t3;
 int lbp, ubp;
 register int lbv, ubv;
/* Start of CLooG code */
if ((_PB_NL >= 1) && (_PB_NM >= 1) && (ts1 == 1) && (ts2 >= 0) && (32*ts2 <= _PB_NJ-1)) {
  for (t3=0;t3<=floord(_PB_NL-1,32);t3++) {
    if(F_trans[(ts2) * ((F_size_1)) + (t3)] == NULL) F_trans[(ts2)* ( F_size_1 ) + (t3)] = (double *) buffer_alloc_atomic((ts2)* ( F_size_1 ) + (t3), 32*32 * sizeof(double), buff_mang_F);;
  }
  for (t3=0;t3<=floord(_PB_NM-1,32);t3++) {
    if(C_trans[(ts2) * ((C_size_1)) + (t3)] == NULL) C_trans[(ts2)* ( C_size_1 ) + (t3)] = (double *) buffer_alloc_atomic((ts2)* ( C_size_1 ) + (t3), 32*32 * sizeof(double), buff_mang_C);;
  }
  for (t2=0;t2<=floord(_PB_NM-1,32);t2++) {
    for (t3=0;t3<=floord(_PB_NL-1,32);t3++) {
      if(D_trans[(t2) * ((D_size_1)) + (t3)] == NULL) D_trans[(t2)* ( D_size_1 ) + (t3)] = (double *) buffer_alloc_atomic((t2)* ( D_size_1 ) + (t3), 32*32 * sizeof(double), buff_mang_D);;
    }
  }
}
/* End of CLooG code */
}
void data_tile_ref_count_update_2(int ts1,int ts2, double** F_trans,int F_size_1, BufferManager *buff_mang_F, double** C_trans,int C_size_1, BufferManager *buff_mang_C, double** D_trans,int D_size_1, BufferManager *buff_mang_D)
{
  int t1, t2, t3;
 int lbp, ubp;
 register int lbv, ubv;
/* Start of CLooG code */
if ((_PB_NL >= 1) && (_PB_NM >= 1) && (ts1 == 1) && (ts2 >= 0) && (32*ts2 <= _PB_NJ-1)) {
  for (t3=0;t3<=floord(_PB_NL-1,32);t3++) {
    decrement_ref_count(buff_mang_F, (ts2)* ( F_size_1 ) + (t3));;
  }
  for (t3=0;t3<=floord(_PB_NM-1,32);t3++) {
    decrement_ref_count(buff_mang_C, (ts2)* ( C_size_1 ) + (t3));;
  }
  for (t2=0;t2<=floord(_PB_NM-1,32);t2++) {
    for (t3=0;t3<=floord(_PB_NL-1,32);t3++) {
      decrement_ref_count(buff_mang_D, (t2)* ( D_size_1 ) + (t3));;
    }
  }
}
/* End of CLooG code */
}
void compute_task_2(int t1,int t2, double** F_trans,int F_size_1, BufferManager *buff_mang_F, double** C_trans,int C_size_1, BufferManager *buff_mang_C, double** D_trans,int D_size_1, BufferManager *buff_mang_D)
{
int i_j_0_div = 0;int i_j_0_mod = 0;double *F_i_j_0 = NULL;int i_k_0_div = 0;int i_k_0_mod = 0;double *C_i_k_0 = NULL;int k_j_0_div = 0;int k_j_0_mod = 0;double *D_k_j_0 = NULL; int t3, t4, t5, t6, t7, t8, t9, t10, t11, t12, t13, t14, t15, t16, t17;
 int lbp, ubp;
 register int lbv, ubv;
/* Start of CLooG code */
for (t3=0;t3<=floord(_PB_NL-1,32);t3++) {
  for (t5=0;t5<=floord(_PB_NM-1,32);t5++) {
    for (t6=32*t2;t6<=min(_PB_NJ-1,32*t2+31);t6++) {
      for (t8=32*t5;t8<=min(_PB_NM-1,32*t5+31);t8++) {
        lbv=32*t3;
        ubv=min(_PB_NL-1,32*t3+31);
        #if (defined __DATA_DIST_DECLS || defined USE_LOCAL_ARRAYS)
        i_j_0_div = ((1)*( t6 ))/32* ( F_size_1 ) + ((1)*( lbv ))/32;i_j_0_mod = ((1)*( t6 ))%32* (32) + ((1)*( lbv ))%32;F_i_j_0 = F_trans[i_j_0_div];i_k_0_div = ((1)*( t6 ))/32* ( C_size_1 ) + ((1)*( t8 ))/32;i_k_0_mod = ((1)*( t6 ))%32* (32) + ((1)*( t8 ))%32;C_i_k_0 = C_trans[i_k_0_div];k_j_0_div = ((1)*( t8 ))/32* ( D_size_1 ) + ((1)*( lbv ))/32;k_j_0_mod = ((1)*( t8 ))%32* (32) + ((1)*( lbv ))%32;D_k_j_0 = D_trans[k_j_0_div];;
        #endif
#pragma ivdep
#pragma vector always
        for (t17=lbv;t17<=ubv;t17++) {
          F_i_j_0[i_j_0_mod + t17 - lbv]+=C_i_k_0[i_k_0_mod]*D_k_j_0[k_j_0_mod + t17 - lbv];;
        }
      }
    }
  }
}
/* End of CLooG code */
}
void data_tile_alloc_3(int ts1,int ts2,int ts3, double** E_trans,int E_size_1, BufferManager *buff_mang_E, double** A_trans,int A_size_1, BufferManager *buff_mang_A, double** B_trans,int B_size_1, BufferManager *buff_mang_B, double** G_trans,int G_size_1, BufferManager *buff_mang_G, double** F_trans,int F_size_1, BufferManager *buff_mang_F)
{
  int t1, t2, t3;
 int lbp, ubp;
 register int lbv, ubv;
/* Start of CLooG code */
if ((ts1 == 3) && (ts2 >= ts3) && (32*ts2 <= 32*ts3+_PB_NJ-1) && (ts3 >= 0) && (32*ts3 <= _PB_NI-1)) {
  if (_PB_NK >= 1) {
    if(E_trans[(ts3) * ((E_size_1)) + (ts2-ts3)] == NULL) E_trans[(ts3)* ( E_size_1 ) + (ts2-ts3)] = (double *) buffer_alloc_atomic((ts3)* ( E_size_1 ) + (ts2-ts3), 32*32 * sizeof(double), buff_mang_E);;
  }
  if ((_PB_NK <= 0) && (_PB_NL >= 1)) {
    if(E_trans[(ts3) * ((E_size_1)) + (ts2-ts3)] == NULL) E_trans[(ts3)* ( E_size_1 ) + (ts2-ts3)] = (double *) buffer_alloc_atomic((ts3)* ( E_size_1 ) + (ts2-ts3), 32*32 * sizeof(double), buff_mang_E);;
  }
  if (_PB_NK >= 1) {
    for (t3=0;t3<=floord(_PB_NK-1,32);t3++) {
      if(A_trans[(ts3) * ((A_size_1)) + (t3)] == NULL) A_trans[(ts3)* ( A_size_1 ) + (t3)] = (double *) buffer_alloc_atomic((ts3)* ( A_size_1 ) + (t3), 32*32 * sizeof(double), buff_mang_A);;
    }
  }
  for (t2=0;t2<=floord(_PB_NK-1,32);t2++) {
    if(B_trans[(t2) * ((B_size_1)) + (ts2-ts3)] == NULL) B_trans[(t2)* ( B_size_1 ) + (ts2-ts3)] = (double *) buffer_alloc_atomic((t2)* ( B_size_1 ) + (ts2-ts3), 32*32 * sizeof(double), buff_mang_B);;
  }
  if (_PB_NL >= 1) {
    for (t3=0;t3<=floord(_PB_NL-1,32);t3++) {
      if(G_trans[(ts3) * ((G_size_1)) + (t3)] == NULL) G_trans[(ts3)* ( G_size_1 ) + (t3)] = (double *) buffer_alloc_atomic((ts3)* ( G_size_1 ) + (t3), 32*32 * sizeof(double), buff_mang_G);;
    }
  }
  if (_PB_NL >= 1) {
    for (t3=0;t3<=floord(_PB_NL-1,32);t3++) {
      if(F_trans[(ts2-ts3) * ((F_size_1)) + (t3)] == NULL) F_trans[(ts2-ts3)* ( F_size_1 ) + (t3)] = (double *) buffer_alloc_atomic((ts2-ts3)* ( F_size_1 ) + (t3), 32*32 * sizeof(double), buff_mang_F);;
    }
  }
}
/* End of CLooG code */
}
void data_tile_ref_count_update_3(int ts1,int ts2,int ts3, double** E_trans,int E_size_1, BufferManager *buff_mang_E, double** A_trans,int A_size_1, BufferManager *buff_mang_A, double** B_trans,int B_size_1, BufferManager *buff_mang_B, double** G_trans,int G_size_1, BufferManager *buff_mang_G, double** F_trans,int F_size_1, BufferManager *buff_mang_F)
{
  int t1, t2, t3;
 int lbp, ubp;
 register int lbv, ubv;
/* Start of CLooG code */
if ((ts1 == 3) && (ts2 >= ts3) && (32*ts2 <= 32*ts3+_PB_NJ-1) && (ts3 >= 0) && (32*ts3 <= _PB_NI-1)) {
  if (_PB_NK >= 1) {
    decrement_ref_count(buff_mang_E, (ts3)* ( E_size_1 ) + (ts2-ts3));;
  }
  if ((_PB_NK <= 0) && (_PB_NL >= 1)) {
    decrement_ref_count(buff_mang_E, (ts3)* ( E_size_1 ) + (ts2-ts3));;
  }
  if (_PB_NK >= 1) {
    for (t3=0;t3<=floord(_PB_NK-1,32);t3++) {
      decrement_ref_count(buff_mang_A, (ts3)* ( A_size_1 ) + (t3));;
    }
  }
  for (t2=0;t2<=floord(_PB_NK-1,32);t2++) {
    decrement_ref_count(buff_mang_B, (t2)* ( B_size_1 ) + (ts2-ts3));;
  }
  if (_PB_NL >= 1) {
    for (t3=0;t3<=floord(_PB_NL-1,32);t3++) {
      decrement_ref_count(buff_mang_G, (ts3)* ( G_size_1 ) + (t3));;
    }
  }
  if (_PB_NL >= 1) {
    for (t3=0;t3<=floord(_PB_NL-1,32);t3++) {
      decrement_ref_count(buff_mang_F, (ts2-ts3)* ( F_size_1 ) + (t3));;
    }
  }
}
/* End of CLooG code */
}
void compute_task_3(int t1,int t2,int t3, double** E_trans,int E_size_1, BufferManager *buff_mang_E, double** A_trans,int A_size_1, BufferManager *buff_mang_A, double** B_trans,int B_size_1, BufferManager *buff_mang_B, double** G_trans,int G_size_1, BufferManager *buff_mang_G, double** F_trans,int F_size_1, BufferManager *buff_mang_F)
{
int i_j_0_div = 0;int i_j_0_mod = 0;double *E_i_j_0 = NULL;int i_k_0_div = 0;int i_k_0_mod = 0;double *A_i_k_0 = NULL;int k_j_0_div = 0;int k_j_0_mod = 0;double *B_k_j_0 = NULL;int i_j_1_div = 0;int i_j_1_mod = 0;double *G_i_j_1 = NULL;int i_k_1_div = 0;int i_k_1_mod = 0;double *E_i_k_1 = NULL;int k_j_1_div = 0;int k_j_1_mod = 0;double *F_k_j_1 = NULL; int t4, t5, t6, t7, t8, t9, t10, t11, t12, t13, t14, t15, t16, t17;
 int lbp, ubp;
 register int lbv, ubv;
/* Start of CLooG code */
for (t5=0;t5<=floord(_PB_NK-1,32);t5++) {
  for (t6=32*t2-32*t3;t6<=min(_PB_NJ-1,32*t2-32*t3+31);t6++) {
    for (t7=32*t3;t7<=min(_PB_NI-1,32*t3+31);t7++) {
      lbv=32*t5;
      ubv=min(_PB_NK-1,32*t5+31);
      #if (defined __DATA_DIST_DECLS || defined USE_LOCAL_ARRAYS)
      i_j_0_div = ((1)*( t7 ))/32* ( E_size_1 ) + ((1)*( t6 ))/32;i_j_0_mod = ((1)*( t7 ))%32* (32) + ((1)*( t6 ))%32;E_i_j_0 = E_trans[i_j_0_div];i_k_0_div = ((1)*( t7 ))/32* ( A_size_1 ) + ((1)*( lbv ))/32;i_k_0_mod = ((1)*( t7 ))%32* (32) + ((1)*( lbv ))%32;A_i_k_0 = A_trans[i_k_0_div];k_j_0_div = ((1)*( lbv ))/32* ( B_size_1 ) + ((1)*( t6 ))/32;k_j_0_mod = ((1)*( lbv ))%32* (32) + ((1)*( t6 ))%32;B_k_j_0 = B_trans[k_j_0_div];;
      #endif
#pragma ivdep
#pragma vector always
      for (t9=lbv;t9<=ubv;t9++) {
        E_i_j_0[i_j_0_mod]+=A_i_k_0[i_k_0_mod + t9 - lbv]*B_k_j_0[k_j_0_mod + (t9 - lbv) * 32];;
      }
    }
  }
}
for (t5=0;t5<=floord(_PB_NL-1,32);t5++) {
  for (t6=32*t2-32*t3;t6<=min(_PB_NJ-1,32*t2-32*t3+31);t6++) {
    for (t7=32*t3;t7<=min(_PB_NI-1,32*t3+31);t7++) {
      lbv=32*t5;
      ubv=min(_PB_NL-1,32*t5+31);
      #if (defined __DATA_DIST_DECLS || defined USE_LOCAL_ARRAYS)
      i_j_1_div = ((1)*( t7 ))/32* ( G_size_1 ) + ((1)*( lbv ))/32;i_j_1_mod = ((1)*( t7 ))%32* (32) + ((1)*( lbv ))%32;G_i_j_1 = G_trans[i_j_1_div];i_k_1_div = ((1)*( t7 ))/32* ( E_size_1 ) + ((1)*( t6 ))/32;i_k_1_mod = ((1)*( t7 ))%32* (32) + ((1)*( t6 ))%32;E_i_k_1 = E_trans[i_k_1_div];k_j_1_div = ((1)*( t6 ))/32* ( F_size_1 ) + ((1)*( lbv ))/32;k_j_1_mod = ((1)*( t6 ))%32* (32) + ((1)*( lbv ))%32;F_k_j_1 = F_trans[k_j_1_div];;
      #endif
#pragma ivdep
#pragma vector always
      for (t17=lbv;t17<=ubv;t17++) {
        G_i_j_1[i_j_1_mod + t17 - lbv]+=E_i_k_1[i_k_1_mod]*F_k_j_1[k_j_1_mod + t17 - lbv];;
      }
    }
  }
}
/* End of CLooG code */
}
void data_tile_copy_back_E_0(int ts1,int ts2 
#ifndef USE_LOCAL_ARRAYS
 ,__DECLARATION_OF_E 
#endif
 , double** E_trans,int E_size_1, BufferManager *buff_mang_E)
{
#ifndef USE_LOCAL_ARRAYS
  int t1, t2, t3;
 int lbp, ubp;
 register int lbv, ubv;
/* Start of CLooG code */
if ((_PB_NI >= 1) && (ts1 == 2) && (ts2 >= 0) && (32*ts2 <= _PB_NJ-1)) {
  for (t2=0;t2<=_PB_NI-1;t2++) {
    for (t3=32*ts2;t3<=min(_PB_NJ-1,32*ts2+31);t3++) {
      E[t2][t3] = E_trans[((1)*( t2 ))/32* ( E_size_1 ) + ((1)*( t3 ))/32][((1)*( t2 ))%32* (32) + ((1)*( t3 ))%32];;
    }
  }
}
/* End of CLooG code */
#endif
}
void data_tile_copy_back_F_1(int ts1,int ts2 
#ifndef USE_LOCAL_ARRAYS
 ,__DECLARATION_OF_F 
#endif
 , double** F_trans,int F_size_1, BufferManager *buff_mang_F)
{
#ifndef USE_LOCAL_ARRAYS
  int t1, t2, t3;
 int lbp, ubp;
 register int lbv, ubv;
/* Start of CLooG code */
if ((_PB_NL >= 1) && (ts1 == 0) && (ts2 >= 0) && (32*ts2 <= _PB_NJ-1)) {
  for (t2=32*ts2;t2<=min(_PB_NJ-1,32*ts2+31);t2++) {
    for (t3=0;t3<=_PB_NL-1;t3++) {
      F[t2][t3] = F_trans[((1)*( t2 ))/32* ( F_size_1 ) + ((1)*( t3 ))/32][((1)*( t2 ))%32* (32) + ((1)*( t3 ))%32];;
    }
  }
}
/* End of CLooG code */
#endif
}
void data_tile_copy_back_G_1(int ts1,int ts2 
#ifndef USE_LOCAL_ARRAYS
 ,__DECLARATION_OF_G 
#endif
 , double** G_trans,int G_size_1, BufferManager *buff_mang_G)
{
#ifndef USE_LOCAL_ARRAYS
  int t1, t2, t3;
 int lbp, ubp;
 register int lbv, ubv;
/* Start of CLooG code */
if ((_PB_NL >= 1) && (ts1 == 0) && (ts2 >= 0) && (32*ts2 <= _PB_NI-1)) {
  for (t2=32*ts2;t2<=min(_PB_NI-1,32*ts2+31);t2++) {
    for (t3=0;t3<=_PB_NL-1;t3++) {
      G[t2][t3] = G_trans[((1)*( t2 ))/32* ( G_size_1 ) + ((1)*( t3 ))/32][((1)*( t2 ))%32* (32) + ((1)*( t3 ))%32];;
    }
  }
}
/* End of CLooG code */
#endif
}
void data_tile_copy_back_F_2(int ts1,int ts2 
#ifndef USE_LOCAL_ARRAYS
 ,__DECLARATION_OF_F 
#endif
 , double** F_trans,int F_size_1, BufferManager *buff_mang_F)
{
#ifndef USE_LOCAL_ARRAYS
  int t1, t2, t3;
 int lbp, ubp;
 register int lbv, ubv;
/* Start of CLooG code */
if ((_PB_NL >= 1) && (_PB_NM >= 1) && (ts1 == 1) && (ts2 >= 0) && (32*ts2 <= _PB_NJ-1)) {
  for (t2=32*ts2;t2<=min(_PB_NJ-1,32*ts2+31);t2++) {
    for (t3=0;t3<=_PB_NL-1;t3++) {
      F[t2][t3] = F_trans[((1)*( t2 ))/32* ( F_size_1 ) + ((1)*( t3 ))/32][((1)*( t2 ))%32* (32) + ((1)*( t3 ))%32];;
    }
  }
}
/* End of CLooG code */
#endif
}
void data_tile_copy_back_E_3(int ts1,int ts2,int ts3 
#ifndef USE_LOCAL_ARRAYS
 ,__DECLARATION_OF_E 
#endif
 , double** E_trans,int E_size_1, BufferManager *buff_mang_E)
{
#ifndef USE_LOCAL_ARRAYS
  int t1, t2, t3;
 int lbp, ubp;
 register int lbv, ubv;
/* Start of CLooG code */
if ((_PB_NK >= 1) && (ts1 == 3) && (ts2 >= ts3) && (32*ts2 <= 32*ts3+_PB_NJ-1) && (ts3 >= 0) && (32*ts3 <= _PB_NI-1)) {
  for (t2=32*ts3;t2<=min(_PB_NI-1,32*ts3+31);t2++) {
    for (t3=32*ts2-32*ts3;t3<=min(_PB_NJ-1,32*ts2-32*ts3+31);t3++) {
      E[t2][t3] = E_trans[((1)*( t2 ))/32* ( E_size_1 ) + ((1)*( t3 ))/32][((1)*( t2 ))%32* (32) + ((1)*( t3 ))%32];;
    }
  }
}
/* End of CLooG code */
#endif
}
void data_tile_copy_back_G_3(int ts1,int ts2,int ts3 
#ifndef USE_LOCAL_ARRAYS
 ,__DECLARATION_OF_G 
#endif
 , double** G_trans,int G_size_1, BufferManager *buff_mang_G)
{
#ifndef USE_LOCAL_ARRAYS
  int t1, t2, t3;
 int lbp, ubp;
 register int lbv, ubv;
/* Start of CLooG code */
if ((_PB_NL >= 1) && (ts1 == 3) && (ts2 >= ts3) && (32*ts2 <= 32*ts3+_PB_NJ-1) && (ts3 >= 0) && (32*ts3 <= _PB_NI-1)) {
  for (t2=32*ts3;t2<=min(_PB_NI-1,32*ts3+31);t2++) {
    for (t3=0;t3<=_PB_NL-1;t3++) {
      G[t2][t3] = G_trans[((1)*( t2 ))/32* ( G_size_1 ) + ((1)*( t3 ))/32][((1)*( t2 ))%32* (32) + ((1)*( t3 ))%32];;
    }
  }
}
/* End of CLooG code */
#endif
}
void copy_back_writeout_0(int my_rank, int nprocs 
#ifndef USE_LOCAL_ARRAYS
 ,__DECLARATION_OF_E,__DECLARATION_OF_F,__DECLARATION_OF_G 
#endif
 , double** E_trans,int E_size_1, BufferManager *buff_mang_E, double** F_trans,int F_size_1, BufferManager *buff_mang_F, double** G_trans,int G_size_1, BufferManager *buff_mang_G)
{
#ifndef USE_LOCAL_ARRAYS
  int t1, t2, t3, t4;
 int lbp, ubp;
 register int lbv, ubv;
/* Start of CLooG code */
if ((_PB_NI >= 1) && (_PB_NJ >= 1)) {
  for (t3=0;t3<=floord(_PB_NJ-1,32);t3++) {
    data_tile_copy_back_E_0(2,t3, E, E_trans, E_size_1,buff_mang_E);;
  }
}
if ((_PB_NJ >= 1) && (_PB_NL >= 1)) {
  for (t3=0;t3<=floord(_PB_NJ-1,32);t3++) {
    data_tile_copy_back_F_1(0,t3, F, F_trans, F_size_1,buff_mang_F);;
  }
}
if ((_PB_NI >= 1) && (_PB_NL >= 1)) {
  for (t3=0;t3<=floord(_PB_NI-1,32);t3++) {
    data_tile_copy_back_G_1(0,t3, G, G_trans, G_size_1,buff_mang_G);;
  }
}
if ((_PB_NJ >= 1) && (_PB_NL >= 1) && (_PB_NM >= 1)) {
  for (t3=0;t3<=floord(_PB_NJ-1,32);t3++) {
    data_tile_copy_back_F_2(1,t3, F, F_trans, F_size_1,buff_mang_F);;
  }
}
if ((_PB_NI >= 1) && (_PB_NJ >= 1) && (_PB_NK >= 1)) {
  for (t3=0;t3<=floord(_PB_NI+_PB_NJ-2,32);t3++) {
    for (t4=max(0,ceild(32*t3-_PB_NJ+1,32));t4<=min(floord(_PB_NI-1,32),t3);t4++) {
      data_tile_copy_back_E_3(3,t3,t4, E, E_trans, E_size_1,buff_mang_E);;
    }
  }
}
if ((_PB_NI >= 1) && (_PB_NJ >= 1) && (_PB_NL >= 1)) {
  for (t3=0;t3<=floord(_PB_NI+_PB_NJ-2,32);t3++) {
    for (t4=max(0,ceild(32*t3-_PB_NJ+1,32));t4<=min(floord(_PB_NI-1,32),t3);t4++) {
      data_tile_copy_back_G_3(3,t3,t4, G, G_trans, G_size_1,buff_mang_G);;
    }
  }
}
/* End of CLooG code */
#endif
}
