#include "cholesky.dyn.h"
#include <assert.h>
#include <omp.h>
#include <math.h>
#define ceild(n,d)  ceil(((double)(n))/((double)(d)))
#define floord(n,d) floor(((double)(n))/((double)(d)))
#define max(x,y)    ((x) > (y)? (x) : (y))
#define min(x,y)    ((x) < (y)? (x) : (y))

/**
 * cholesky.c: This file is part of the PolyBench/C 3.2 test suite.
 *
 *
 * Contact: Louis-Noel Pouchet <pouchet@cse.ohio-state.edu>
 * Web address: http://polybench.sourceforge.net
 */
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <math.h>

/* Include polybench common header. */
#include <polybench.h>

#define N 8192
/* Include benchmark-specific header. */
/* Default data type is double, default size is 4000. */
#include "cholesky.h"


/* Array initialization. */
static
void init_array(int n,
		DATA_TYPE POLYBENCH_1D(p,N,n),
		DATA_TYPE POLYBENCH_2D(A,N,N,n,n))
{
  int i, j;

  srand(42);
  for (i = 0; i < n; i++)
    {
      p[i] = ((DATA_TYPE) rand()) / n;
      for (j = 0; j < n; j++)
	A[i][j] = ((DATA_TYPE) rand()) /n; 
    }


}


/* DCE code. Must scan the entire live-out data.
   Can be used also to check the correctness of the output. */
static
void print_array(int n,
		 DATA_TYPE POLYBENCH_2D(A,N,N,n,n))

{
  int i, j;

  for (i = 0; i < n; i++)
    for (j = 0; j < n; j++) {
    fprintf (stderr, DATA_PRINTF_MODIFIER, A[i][j]);
    if ((i * N + j) % 20 == 0) fprintf (stderr, "\n");
  }
}

#define __DECLARATION_OF_p DATA_TYPE POLYBENCH_1D(p,N,n)
#define __DECLARATION_OF_POINTER_TO_p(p) DATA_TYPE (*p)
#define __DECLARATION_OF_A DATA_TYPE POLYBENCH_2D(A,N,N,n,n)
#define __DECLARATION_OF_POINTER_TO_A(A) DATA_TYPE (*A)[N]

#define __PLACE_TO_INSERT_FORWARD_DECLARATIONS
FILE *__debug_print_fp;
tbb::concurrent_priority_queue<__Task, __Compare> pqueue;
int max_num_tasks_loop0_dim0;
int lb_tasks_loop0_dim0;
int *tasks_loop0;
int max_num_tasks_loop1_dim0;
int lb_tasks_loop1_dim0;
int *tasks_loop1;
int max_num_tasks_loop2_dim0;
int lb_tasks_loop2_dim0;
int max_num_tasks_loop2_dim1;
int lb_tasks_loop2_dim1;
int *tasks_loop2;
void compute_task_0(int t1,int t2,__DECLARATION_OF_p,__DECLARATION_OF_A);
void compute_task_1(int t1,int t2,__DECLARATION_OF_A);
void compute_task_2(int t1,int t2,int t3,__DECLARATION_OF_p,__DECLARATION_OF_A);

/* Main computational kernel. The whole function will be timed,
   including the call and return. */
static
void kernel_cholesky(int n
#ifndef USE_LOCAL_ARRAYS
		     ,DATA_TYPE POLYBENCH_1D(p,N,n),
		     DATA_TYPE POLYBENCH_2D(A,N,N,n,n)
#endif
)
{
  int i, j, k;


 int _num_threads;
#pragma omp parallel
{_num_threads = omp_get_num_threads();}
 int __tid;
 long int _num_tasks_to_execute;
 int waiting_pops_start[_num_threads], waiting_pops[_num_threads], waiting_pops_threads;
 for (__tid=0; __tid<_num_threads; __tid++) waiting_pops[__tid] = 0;
 double t_local;
 double t_tasks_create;
 double t_tasks_manage_start[_num_threads], t_tasks_manage[_num_threads], t_tasks_manage_threads;
 for (__tid=0; __tid<_num_threads; __tid++) t_tasks_manage[__tid] = 0;
 double t_comp_start[_num_threads], t_comp[_num_threads], t_comp_threads_min, t_comp_threads;
 for (__tid=0; __tid<_num_threads; __tid++) t_comp[__tid] = 0;
 double t_wait_comp_start[_num_threads], t_wait_comp[_num_threads], t_wait_comp_threads;
 for (__tid=0; __tid<_num_threads; __tid++) t_wait_comp[__tid] = 0;
max_num_tasks_loop0_dim0 = ceil(floorf((+1*_PB_N-1)/(float)64) - 0 + 1);
lb_tasks_loop0_dim0 = floor(0);
tasks_loop0 = (int *)malloc(max_num_tasks_loop0_dim0*sizeof(int));
max_num_tasks_loop1_dim0 = ceil(floorf((+1*_PB_N-2)/(float)64) - 0 + 1);
lb_tasks_loop1_dim0 = floor(0);
tasks_loop1 = (int *)malloc(max_num_tasks_loop1_dim0*sizeof(int));
max_num_tasks_loop2_dim0 = ceil(floorf((+2*_PB_N-3)/(float)64) - -1 + 1);
lb_tasks_loop2_dim0 = floor(-1);
max_num_tasks_loop2_dim1 = ceil(floorf((+1*_PB_N-1)/(float)64) - 0 + 1);
lb_tasks_loop2_dim1 = floor(0);
tasks_loop2 = (int *)malloc(max_num_tasks_loop2_dim0*max_num_tasks_loop2_dim1*sizeof(int));
 IF_TIME(t_local = rtclock());



/*Start of dynamic scheduling code*/


#if defined(__DYNSCHEDULER_DEBUG_PRINT) || defined(__DYNSCHEDULER_MORE_DEBUG_PRINT) || defined(__DEBUG_FLUSH)
 char *debug_print_name = (char *)malloc(128*sizeof(char));
 strcpy(debug_print_name, "debug_print");
 sprintf(debug_print_name+strlen(debug_print_name), "_node%d", my_rank);
 __debug_print_fp = fopen(debug_print_name, "w");
 free(debug_print_name);
#endif
 IF_TIME(t_tasks_create = rtclock());
 init_tasks(_num_threads,&_num_tasks_to_execute,tasks_loop0,tasks_loop1,tasks_loop2);
 IF_TIME(t_tasks_create = rtclock() - t_tasks_create);
#ifdef __DYNSCHEDULER_MORE_DEBUG_PRINT
 std::vector<int> prev_firing_count_p_0;
 std::vector<int> prev_firing_count_p_1;
 std::vector<int> prev_firing_count_p_2;
 std::vector<int> prev_firing_count_A_0;
 std::vector<int> prev_firing_count_A_1;
 std::vector<int> prev_firing_count_A_2;
 for (__p=0; __p<__MAX_NUM_RECVS; __p++) {
  prev_firing_count_p_0.push_back(-1);
  prev_firing_count_p_1.push_back(-1);
  prev_firing_count_p_2.push_back(-1);
  prev_firing_count_A_0.push_back(-1);
  prev_firing_count_A_1.push_back(-1);
  prev_firing_count_A_2.push_back(-1);
 }
#endif

#pragma omp parallel private(__tid)
{
 __tid = omp_get_thread_num();
 int __flag;
 int __loop;
 int task_id1,task_id2,task_id3;
#ifdef __DYNSCHEDULER_DEBUG_PRINT
 fprintf(__debug_print_fp, "node %d thread %d\n", my_rank, __tid);
 IF_DEBUG_FLUSH(fflush(__debug_print_fp));
#endif
 IF_TIME(t_wait_comp_start[__tid] = rtclock());

{ /* compute-pack-send */
 while (_num_tasks_to_execute > 0) {
  __Task task;
  __flag = pqueue.try_pop(task);
  if(__flag==0) { IF_TIME(waiting_pops[__tid]++); if (pqueue.size() == 0) {
#ifdef __DYNSCHEDULER_DEBUG_PRINT
   int __i;
   for (__i=0;__i<send_reqs_p[__tid].size();__i++) {
    if (send_reqs_p[__tid][__i] != MPI_REQUEST_NULL) {
    int flag, __error = MPI_Test(&send_reqs_p[__tid][__i], &flag, MPI_STATUS_IGNORE);
     if (__error) {
      char *msg=(char*)malloc((MPI_MAX_ERROR_STRING+1)*sizeof(char)); int len;
      MPI_Error_string(__error, msg, &len);
      fprintf(__debug_print_fp, "node %d p send test buf %d error %d %s\n", my_rank, __i, __error, msg);
      free(msg);
     }
     if (flag) {
      fprintf(__debug_print_fp, "node %d p send completed buf %d\n", my_rank, __i);
     }
    }
   }
   for (__i=0;__i<send_reqs_A[__tid].size();__i++) {
    if (send_reqs_A[__tid][__i] != MPI_REQUEST_NULL) {
    int flag, __error = MPI_Test(&send_reqs_A[__tid][__i], &flag, MPI_STATUS_IGNORE);
     if (__error) {
      char *msg=(char*)malloc((MPI_MAX_ERROR_STRING+1)*sizeof(char)); int len;
      MPI_Error_string(__error, msg, &len);
      fprintf(__debug_print_fp, "node %d A send test buf %d error %d %s\n", my_rank, __i, __error, msg);
      free(msg);
     }
     if (flag) {
      fprintf(__debug_print_fp, "node %d A send completed buf %d\n", my_rank, __i);
     }
    }
   }
   IF_DEBUG_FLUSH(fflush(__debug_print_fp));
#endif
  }
  continue;
  }
  __loop = task.loop;
  task_id1 = task.id1;
  task_id2 = task.id2;
  task_id3 = task.id3;
#ifdef __DYNSCHEDULER_DEBUG_PRINT
  long int captured;
#pragma omp atomic capture
  captured = --_num_tasks_to_execute;
  fprintf(__debug_print_fp, "computing node %d thread %d loop %d task %d %d %d ready tasks %lu\n", my_rank, __tid, __loop, task_id1, task_id2, task_id3, pqueue.size());
  IF_DEBUG_FLUSH(fflush(__debug_print_fp));
#else
#pragma omp atomic
  _num_tasks_to_execute--;
#endif
  switch(__loop){
   case 0:
    IF_TIME(t_wait_comp[__tid] += rtclock() - t_wait_comp_start[__tid]);
    IF_TIME(t_comp_start[__tid] = rtclock());
    compute_task_0(task_id1,task_id2,p,A);
    IF_TIME(t_comp[__tid] += rtclock() - t_comp_start[__tid]);
    IF_TIME(t_tasks_manage_start[__tid] = rtclock());
    local_update_dep_tasks_0(task_id1,task_id2,_PB_N,_num_threads,tasks_loop0,tasks_loop1,tasks_loop2);
    IF_TIME(t_tasks_manage[__tid] += rtclock() - t_tasks_manage_start[__tid]);
    IF_TIME(t_wait_comp_start[__tid] = rtclock());
    break;
   case 1:
    IF_TIME(t_wait_comp[__tid] += rtclock() - t_wait_comp_start[__tid]);
    IF_TIME(t_comp_start[__tid] = rtclock());
    compute_task_1(task_id1,task_id2,A);
    IF_TIME(t_comp[__tid] += rtclock() - t_comp_start[__tid]);
    IF_TIME(t_tasks_manage_start[__tid] = rtclock());
    local_update_dep_tasks_1(task_id1,task_id2,_PB_N,_num_threads,tasks_loop0,tasks_loop1,tasks_loop2);
    IF_TIME(t_tasks_manage[__tid] += rtclock() - t_tasks_manage_start[__tid]);
    IF_TIME(t_wait_comp_start[__tid] = rtclock());
    break;
   case 2:
    IF_TIME(t_wait_comp[__tid] += rtclock() - t_wait_comp_start[__tid]);
    IF_TIME(t_comp_start[__tid] = rtclock());
    compute_task_2(task_id1,task_id2,task_id3,p,A);
    IF_TIME(t_comp[__tid] += rtclock() - t_comp_start[__tid]);
    IF_TIME(t_tasks_manage_start[__tid] = rtclock());
    local_update_dep_tasks_2(task_id1,task_id2,task_id3,_PB_N,_num_threads,tasks_loop0,tasks_loop1,tasks_loop2);
    IF_TIME(t_tasks_manage[__tid] += rtclock() - t_tasks_manage_start[__tid]);
    IF_TIME(t_wait_comp_start[__tid] = rtclock());
    break;
   default:
    assert(0);
  }
#ifdef __DYNSCHEDULER_DEBUG_PRINT
  fprintf(__debug_print_fp, "computed node %d thread %d loop %d task %d %d %d tasks left %ld\n", my_rank, __tid, __loop, task_id1, task_id2, task_id3, captured);
  IF_DEBUG_FLUSH(fflush(__debug_print_fp));
#pragma omp atomic
  __busy_wait_count *= 0;
#endif
 }
}

#ifdef __DYNSCHEDULER_DEBUG_PRINT
 fprintf(__debug_print_fp, "node %d thread %d done\n", my_rank, __tid);
 IF_DEBUG_FLUSH(fflush(__debug_print_fp));
#endif

}

#ifdef __DYNSCHEDULER_DEBUG_PRINT
 fclose(__debug_print_fp);
#endif



/*End of dynamic scheduling code*/


 t_local = rtclock() - t_local;
#ifdef TIME
 waiting_pops_threads = waiting_pops[0];
                for (__tid=1; __tid<_num_threads; __tid++) if (waiting_pops_threads < waiting_pops[__tid]) waiting_pops_threads = waiting_pops[__tid];
 t_tasks_manage_threads = t_tasks_manage[0];
                for (__tid=1; __tid<_num_threads; __tid++) if (t_tasks_manage_threads < t_tasks_manage[__tid]) t_tasks_manage_threads = t_tasks_manage[__tid];
 t_comp_threads_min = t_comp[0];
                for (__tid=1; __tid<_num_threads; __tid++) if (t_comp_threads_min > t_comp[__tid]) t_comp_threads_min = t_comp[__tid];
 t_comp_threads = t_comp[0];
                for (__tid=1; __tid<_num_threads; __tid++) if (t_comp_threads < t_comp[__tid]) t_comp_threads = t_comp[__tid];
 t_wait_comp_threads = t_wait_comp[0];
                for (__tid=1; __tid<_num_threads; __tid++) if (t_wait_comp_threads < t_wait_comp[__tid]) t_wait_comp_threads = t_wait_comp[__tid];
 char buffer[4096];
 strcpy(buffer, "");
 sprintf(buffer+strlen(buffer), "Minimum computation time across all threads: %lf\n", t_comp_threads_min);

 sprintf(buffer+strlen(buffer), "Maximum computation time across all threads: %lf\n", t_comp_threads);

 sprintf(buffer+strlen(buffer), "Maximum 'waiting for computation' time across all threads: %lf\n", t_wait_comp_threads);

 sprintf(buffer+strlen(buffer), "Maximum number of waiting/unsuccessful POPs across all threads: %d\n", waiting_pops_threads);

 sprintf(buffer+strlen(buffer), "Maximum tasks management time across all threads: %lf\n", t_tasks_manage_threads);

 sprintf(buffer+strlen(buffer), "Tasks creation time: %lf\n", t_tasks_create);

 sprintf(buffer+strlen(buffer), "Sum of split times: %lf\n", t_comp_threads+t_wait_comp_threads+t_tasks_manage_threads+t_tasks_create);

 sprintf(buffer+strlen(buffer), "Total time: %lf\n", t_local);

 fprintf(stdout, "%s\n", buffer);
#endif

}


int main(int argc, char** argv)
{
  /* Retrieve problem size. */
  int n = N;

#ifndef USE_LOCAL_ARRAYS

  /* Variable declaration/allocation. */
  POLYBENCH_2D_ARRAY_DECL(A, DATA_TYPE, N, N, n, n);
  POLYBENCH_1D_ARRAY_DECL(p, DATA_TYPE, N, n);


  /* Initialize array(s). */
  init_array (n, POLYBENCH_ARRAY(p), POLYBENCH_ARRAY(A));

  /* Start timer. */
  polybench_start_instruments;

  /* Run kernel. */
  kernel_cholesky (n, POLYBENCH_ARRAY(p), POLYBENCH_ARRAY(A));

  /* Stop and print timer. */
  polybench_stop_instruments;
  polybench_print_instruments;

  /* Prevent dead-code elimination. All live-out data must be printed
     by the function call in argument. */
  polybench_prevent_dce(print_array(n, POLYBENCH_ARRAY(A)));

  /* Be clean. */
  POLYBENCH_FREE_ARRAY(A);
  POLYBENCH_FREE_ARRAY(p);
#else

  kernel_cholesky (n);

#endif
  return 0;
}
void init_tasks(int num_threads,long int *p_num_tasks_to_execute,int *tasks_loop0,int *tasks_loop1,int *tasks_loop2)
{
int local_dep_tasks, affinity;
long int _num_tasks_to_execute = 0;




  int t1, t2, t3, t4;

 int lbp, ubp;
 register int lbv, ubv;

/* Start of CLooG code */
if (_PB_N >= 1) {
  lbp=0;
  ubp=floord(_PB_N-1,64);
#pragma omp parallel for private(local_dep_tasks, affinity, t4) reduction(+:_num_tasks_to_execute)
  for (t3=lbp;t3<=ubp;t3++) {
    _num_tasks_to_execute++;tasks_loop0[(t3-lb_tasks_loop0_dim0)+0] = get_num_local_src_tasks_0(1,t3,_PB_N);if (tasks_loop0[(t3-lb_tasks_loop0_dim0)+0] == 0) { local_dep_tasks = get_num_local_dep_tasks_0(1,t3,_PB_N); affinity = pi_0(1,t3,_PB_N,num_threads); __Task task(0,1,t3,local_dep_tasks,affinity); pqueue.push(task); IF_DYNSCHEDULER_MORE_DEBUG_PRINT( fprintf(__debug_print_fp, "added node %d loop 0 task %d %d                 local_dep_tasks %d ready tasks %lu\n", my_rank, 1,t3, local_dep_tasks, pqueue.size())); IF_DEBUG_FLUSH(fflush(__debug_print_fp)); };
  }
  if (_PB_N >= 2) {
    lbp=0;
    ubp=floord(_PB_N-2,64);
#pragma omp parallel for private(local_dep_tasks, affinity, t4) reduction(+:_num_tasks_to_execute)
    for (t3=lbp;t3<=ubp;t3++) {
      _num_tasks_to_execute++;tasks_loop1[(t3-lb_tasks_loop1_dim0)+0] = get_num_local_src_tasks_1(0,t3,_PB_N);if (tasks_loop1[(t3-lb_tasks_loop1_dim0)+0] == 0) { local_dep_tasks = get_num_local_dep_tasks_1(0,t3,_PB_N); affinity = pi_1(0,t3,_PB_N,num_threads); __Task task(1,0,t3,local_dep_tasks,affinity); pqueue.push(task); IF_DYNSCHEDULER_MORE_DEBUG_PRINT( fprintf(__debug_print_fp, "added node %d loop 1 task %d %d                 local_dep_tasks %d ready tasks %lu\n", my_rank, 0,t3, local_dep_tasks, pqueue.size())); IF_DEBUG_FLUSH(fflush(__debug_print_fp)); };
    }
  }
  lbp=0;
  ubp=floord(_PB_N-1,32);
#pragma omp parallel for private(local_dep_tasks, affinity, t4) reduction(+:_num_tasks_to_execute)
  for (t3=lbp;t3<=ubp;t3++) {
    for (t4=ceild(t3,2);t4<=min(floord(_PB_N-1,64),t3);t4++) {
      _num_tasks_to_execute++;tasks_loop2[(t3-lb_tasks_loop2_dim0)*max_num_tasks_loop2_dim1+(t4-lb_tasks_loop2_dim1)+0] = get_num_local_src_tasks_2(2,t3,t4,_PB_N);if (tasks_loop2[(t3-lb_tasks_loop2_dim0)*max_num_tasks_loop2_dim1+(t4-lb_tasks_loop2_dim1)+0] == 0) { local_dep_tasks = get_num_local_dep_tasks_2(2,t3,t4,_PB_N); affinity = pi_2(2,t3,t4,_PB_N,num_threads); __Task task(2,2,t3,t4,local_dep_tasks,affinity); pqueue.push(task); IF_DYNSCHEDULER_MORE_DEBUG_PRINT( fprintf(__debug_print_fp, "added node %d loop 2 task %d %d %d                 local_dep_tasks %d ready tasks %lu\n", my_rank, 2,t3,t4, local_dep_tasks, pqueue.size())); IF_DEBUG_FLUSH(fflush(__debug_print_fp)); };
    }
  }
}
/* End of CLooG code */



*p_num_tasks_to_execute = _num_tasks_to_execute;
}

void compute_task_0(int t1,int t2,__DECLARATION_OF_p,__DECLARATION_OF_A)
{



  int t3, t4, t5, t6, t7, t8, t9, t10;

 int lbp, ubp;
 register int lbv, ubv;

/* Start of CLooG code */
lbv=64*t2;
ubv=min(_PB_N-1,64*t2+63);

#if (defined __DATA_DIST_DECLS || defined USE_LOCAL_ARRAYS)
;


#endif
#pragma ivdep
#pragma vector always
for (t3=lbv;t3<=ubv;t3++) {
  p[t3]=A[t3][t3];;
}
/* End of CLooG code */


}

void compute_task_1(int t1,int t2,__DECLARATION_OF_A)
{



  int t3, t4, t5, t6, t7, t8, t9, t10;

 int lbp, ubp;
 register int lbv, ubv;

/* Start of CLooG code */
for (t3=t2;t3<=floord(_PB_N-1,64);t3++) {
  for (t4=64*t2;t4<=min(min(_PB_N-2,64*t2+63),64*t3+62);t4++) {
    lbv=max(64*t3,t4+1);
    ubv=min(_PB_N-1,64*t3+63);

#if (defined __DATA_DIST_DECLS || defined USE_LOCAL_ARRAYS)
    ;


#endif
#pragma ivdep
#pragma vector always
    for (t5=lbv;t5<=ubv;t5++) {
      A[t5][t4]=A[t4][t5];;
    }
  }
}
/* End of CLooG code */


}

void compute_task_2(int t1,int t2,int t3,__DECLARATION_OF_p,__DECLARATION_OF_A)
{






  int t4, t5, t6, t7, t8, t9, t10;

 int lbp, ubp;
 register int lbv, ubv;

/* Start of CLooG code */
for (t4=t2-t3;t4<=t3;t4++) {
  if ((t2 == 2*t3) && (t2 == 2*t4)) {
    for (t5=32*t2;t5<=min(_PB_N-3,32*t2+61);t5++) {
      if (t2%2 == 0) {
        p[t5]=1.0/sqrt(p[t5]);;
      }
      if (t2%2 == 0) {
        A[t5+1][t5]=A[t5+1][t5]*p[t5];;
      }
      if (t2%2 == 0) {
        p[t5+1]=p[t5+1]-A[t5+1][t5]*A[t5+1][t5];;
      }
      for (t6=t5+2;t6<=min(_PB_N-1,32*t2+63);t6++) {
        if (t2%2 == 0) {
          A[t6][t5]=A[t6][t5]*p[t5];;
        }
        lbv=t5+1;
        ubv=t6-1;

#if (defined __DATA_DIST_DECLS || defined USE_LOCAL_ARRAYS)
          ;


#endif
#pragma ivdep
#pragma vector always
        for (t7=lbv;t7<=ubv;t7++) {
          if (t2%2 == 0) {
            A[t6][t7]=A[t6][t7]-A[t6][t5]*A[t7][t5];;
          }
        }
        if (t2%2 == 0) {
          p[t6]=p[t6]-A[t6][t5]*A[t6][t5];;
        }
      }
    }
  }
  if ((t2 == 2*t3) && (t2 == 2*t4) && (t2 <= floord(_PB_N-65,32))) {
    if (t2%2 == 0) {
      p[32*t2+62]=1.0/sqrt(p[32*t2+62]);;
    }
    if (t2%2 == 0) {
      A[32*t2+63][32*t2+62]=A[32*t2+63][32*t2+62]*p[32*t2+62];;
    }
    if (t2%2 == 0) {
      p[32*t2+63]=p[32*t2+63]-A[32*t2+63][32*t2+62]*A[32*t2+63][32*t2+62];;
    }
  }
  if ((t2 == 2*t3) && (t2 == 2*t4) && (t2 <= floord(_PB_N-2,32)) && (t2 >= ceild(_PB_N-64,32))) {
    if (t2%2 == 0) {
      p[_PB_N-2]=1.0/sqrt(p[_PB_N-2]);;
    }
    if (t2%2 == 0) {
      A[_PB_N-1][_PB_N-2]=A[_PB_N-1][_PB_N-2]*p[_PB_N-2];;
    }
    if (t2%2 == 0) {
      p[_PB_N-1]=p[_PB_N-1]-A[_PB_N-1][_PB_N-2]*A[_PB_N-1][_PB_N-2];;
    }
  }
  if ((t2 <= 2*t3-1) && (t3 == t4) && (t3 <= floord(_PB_N-2,64))) {
    for (t5=64*t2-64*t3;t5<=64*t2-64*t3+63;t5++) {
      p[64*t3]=p[64*t3]-A[64*t3][t5]*A[64*t3][t5];;
      for (t6=64*t3+1;t6<=min(_PB_N-1,64*t3+63);t6++) {
        lbv=64*t3;
        ubv=t6-1;

#if (defined __DATA_DIST_DECLS || defined USE_LOCAL_ARRAYS)
        ;


#endif
#pragma ivdep
#pragma vector always
        for (t7=lbv;t7<=ubv;t7++) {
          A[t6][t7]=A[t6][t7]-A[t6][t5]*A[t7][t5];;
        }
        p[t6]=p[t6]-A[t6][t5]*A[t6][t5];;
      }
    }
  }
  if ((t2 <= floord(_PB_N-33,32)) && (64*t3 == _PB_N-1) && (64*t4 == _PB_N-1)) {
    for (t5=64*t2-_PB_N+1;t5<=64*t2-_PB_N+64;t5++) {
      if ((_PB_N+63)%64 == 0) {
        p[_PB_N-1]=p[_PB_N-1]-A[_PB_N-1][t5]*A[_PB_N-1][t5];;
      }
    }
  }
  if ((t2 == 2*t3) && (t2 == 2*t4) && (t2 <= floord(_PB_N-65,32))) {
    if (t2%2 == 0) {
      p[32*t2+63]=1.0/sqrt(p[32*t2+63]);;
    }
  }
  if ((t2 == 2*t3) && (t2 == 2*t4) && (t2 >= ceild(_PB_N-64,32))) {
    if (t2%2 == 0) {
      p[_PB_N-1]=1.0/sqrt(p[_PB_N-1]);;
    }
  }
  if ((t2 == t3+t4) && (t2 <= 2*t3-1)) {
    for (t5=64*t2-64*t3;t5<=64*t2-64*t3+62;t5++) {
      for (t6=64*t3;t6<=min(_PB_N-1,64*t3+63);t6++) {
        A[t6][t5]=A[t6][t5]*p[t5];;
        lbv=t5+1;
        ubv=64*t2-64*t3+63;

#if (defined __DATA_DIST_DECLS || defined USE_LOCAL_ARRAYS)
        ;


#endif
#pragma ivdep
#pragma vector always
        for (t7=lbv;t7<=ubv;t7++) {
          A[t6][t7]=A[t6][t7]-A[t6][t5]*A[t7][t5];;
        }
      }
    }
  }
  if ((t2 <= t3+t4-1) && (t3 >= t4+1)) {
    for (t5=64*t2-64*t3;t5<=64*t2-64*t3+63;t5++) {
      for (t6=64*t3;t6<=min(_PB_N-1,64*t3+63);t6++) {
        lbv=64*t4;
        ubv=64*t4+63;

#if (defined __DATA_DIST_DECLS || defined USE_LOCAL_ARRAYS)
        ;


#endif
#pragma ivdep
#pragma vector always
        for (t7=lbv;t7<=ubv;t7++) {
          A[t6][t7]=A[t6][t7]-A[t6][t5]*A[t7][t5];;
        }
      }
    }
  }
  if ((t2 == t3+t4) && (t2 <= 2*t3-1)) {
    for (t6=64*t3;t6<=min(_PB_N-1,64*t3+63);t6++) {
      A[t6][64*t2-64*t3+63]=A[t6][64*t2-64*t3+63]*p[64*t2-64*t3+63];;
    }
  }
}
/* End of CLooG code */
}
