#include <omp.h>
#include <math.h>
#define ceild(n,d)  ceil(((double)(n))/((double)(d)))
#define floord(n,d) floor(((double)(n))/((double)(d)))
#define max(x,y)    ((x) > (y)? (x) : (y))
#define min(x,y)    ((x) < (y)? (x) : (y))

/**
 * cholesky.c: This file is part of the PolyBench/C 3.2 test suite.
 *
 *
 * Contact: Louis-Noel Pouchet <pouchet@cse.ohio-state.edu>
 * Web address: http://polybench.sourceforge.net
 */
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <math.h>

/* Include polybench common header. */
#include <polybench.h>

#define N 4096
/* Include benchmark-specific header. */
/* Default data type is double, default size is 4000. */
#include "cholesky.h"


/* Array initialization. */
static
void init_array(int n,
		DATA_TYPE POLYBENCH_1D(p,N,n),
		DATA_TYPE POLYBENCH_2D(A,N,N,n,n))
{
  int i, j;

  srand(42);
  for (i = 0; i < n; i++)
    {
      p[i] = ((DATA_TYPE) rand()) / n;
      for (j = 0; j < n; j++)
	A[i][j] = ((DATA_TYPE) rand()) /n; 
    }


}


/* DCE code. Must scan the entire live-out data.
   Can be used also to check the correctness of the output. */
static
void print_array(int n,
		 DATA_TYPE POLYBENCH_2D(A,N,N,n,n))

{
  int i, j;

  for (i = 0; i < n; i++)
    for (j = 0; j < n; j++) {
    fprintf (stderr, DATA_PRINTF_MODIFIER, A[i][j]);
    if ((i * N + j) % 20 == 0) fprintf (stderr, "\n");
  }
}

#define __DECLARATION_OF_p DATA_TYPE POLYBENCH_1D(p,N,n)
#define __DECLARATION_OF_POINTER_TO_p(p) DATA_TYPE (*p)
#define __DECLARATION_OF_A DATA_TYPE POLYBENCH_2D(A,N,N,n,n)
#define __DECLARATION_OF_POINTER_TO_A(A) DATA_TYPE (*A)[N]

#define __PLACE_TO_INSERT_FORWARD_DECLARATIONS

/* Main computational kernel. The whole function will be timed,
   including the call and return. */
static
void kernel_cholesky(int n
#ifndef USE_LOCAL_ARRAYS
		     ,DATA_TYPE POLYBENCH_1D(p,N,n),
		     DATA_TYPE POLYBENCH_2D(A,N,N,n,n)
#endif
)
{
  int i, j, k;

  int t1, t2, t3, t4, t5, t6, t7, t8, t9, t10;
 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;
/* Start of CLooG code */
if (_PB_N >= 1) {
  lbp=0;
  ubp=floord(_PB_N-2,64);
#pragma omp parallel for private(lbv,ubv,t3,t4,t5,t6,t7,t8,t9,t10)
  for (t2=lbp;t2<=ubp;t2++) {
    for (t3=t2;t3<=floord(_PB_N-1,64);t3++) {
      for (t4=64*t2;t4<=min(min(_PB_N-2,64*t2+63),64*t3+62);t4++) {
        lbv=max(64*t3,t4+1);
        ubv=min(_PB_N-1,64*t3+63);
#if (defined __DATA_DIST_DECLS || defined USE_LOCAL_ARRAYS)
        ;
#endif
#pragma ivdep
#pragma vector always
        for (t5=lbv;t5<=ubv;t5++) {
          A[t5][t4]=A[t4][t5];;
        }
      }
    }
  }
  lbp=0;
  ubp=floord(_PB_N-1,64);
#pragma omp parallel for private(lbv,ubv,t3,t4,t5,t6,t7,t8,t9,t10)
  for (t2=lbp;t2<=ubp;t2++) {
    lbv=64*t2;
    ubv=min(_PB_N-1,64*t2+63);
#if (defined __DATA_DIST_DECLS || defined USE_LOCAL_ARRAYS)
    ;
#endif
#pragma ivdep
#pragma vector always
    for (t3=lbv;t3<=ubv;t3++) {
      p[t3]=A[t3][t3];;
    }
  }
  for (t2=0;t2<=floord(_PB_N-1,32);t2++) {
    lbp=ceild(t2,2);
    ubp=min(floord(_PB_N-1,64),t2);
#pragma omp parallel for private(lbv,ubv,t4,t5,t6,t7,t8,t9,t10)
    for (t3=lbp;t3<=ubp;t3++) {
      for (t4=t2-t3;t4<=t3;t4++) {
        if ((t2 == 2*t3) && (t2 == 2*t4)) {
          for (t5=32*t2;t5<=min(_PB_N-3,32*t2+61);t5++) {
            if (t2%2 == 0) {
              p[t5]=1.0/sqrt(p[t5]);;
            }
            if (t2%2 == 0) {
              A[t5+1][t5]=A[t5+1][t5]*p[t5];;
            }
            if (t2%2 == 0) {
              p[t5+1]=p[t5+1]-A[t5+1][t5]*A[t5+1][t5];;
            }
            for (t6=t5+2;t6<=min(_PB_N-1,32*t2+63);t6++) {
              if (t2%2 == 0) {
                A[t6][t5]=A[t6][t5]*p[t5];;
              }
              for (t7=t5+1;t7<=t6-1;t7++) {
                if (t2%2 == 0) {
                  A[t6][t7]=A[t6][t7]-A[t6][t5]*A[t7][t5];;
                }
              }
              if (t2%2 == 0) {
                p[t6]=p[t6]-A[t6][t5]*A[t6][t5];;
              }
            }
          }
        }
        if ((t2 == 2*t3) && (t2 == 2*t4) && (t2 <= floord(_PB_N-65,32))) {
          if (t2%2 == 0) {
            p[32*t2+62]=1.0/sqrt(p[32*t2+62]);;
          }
          if (t2%2 == 0) {
            A[32*t2+63][32*t2+62]=A[32*t2+63][32*t2+62]*p[32*t2+62];;
          }
          if (t2%2 == 0) {
            p[32*t2+63]=p[32*t2+63]-A[32*t2+63][32*t2+62]*A[32*t2+63][32*t2+62];;
          }
        }
        if ((t2 == 2*t3) && (t2 == 2*t4) && (t2 <= floord(_PB_N-2,32)) && (t2 >= ceild(_PB_N-64,32))) {
          if (t2%2 == 0) {
            p[_PB_N-2]=1.0/sqrt(p[_PB_N-2]);;
          }
          if (t2%2 == 0) {
            A[_PB_N-1][_PB_N-2]=A[_PB_N-1][_PB_N-2]*p[_PB_N-2];;
          }
          if (t2%2 == 0) {
            p[_PB_N-1]=p[_PB_N-1]-A[_PB_N-1][_PB_N-2]*A[_PB_N-1][_PB_N-2];;
          }
        }
        if ((t2 <= 2*t3-1) && (t3 == t4) && (t3 <= floord(_PB_N-2,64))) {
          for (t5=64*t2-64*t3;t5<=64*t2-64*t3+63;t5++) {
            p[64*t3]=p[64*t3]-A[64*t3][t5]*A[64*t3][t5];;
            for (t6=64*t3+1;t6<=min(_PB_N-1,64*t3+63);t6++) {
              for (t7=64*t3;t7<=t6-1;t7++) {
                A[t6][t7]=A[t6][t7]-A[t6][t5]*A[t7][t5];;
              }
              p[t6]=p[t6]-A[t6][t5]*A[t6][t5];;
            }
          }
        }
        if ((t2 <= floord(_PB_N-33,32)) && (64*t3 == _PB_N-1) && (64*t4 == _PB_N-1)) {
          for (t5=64*t2-_PB_N+1;t5<=64*t2-_PB_N+64;t5++) {
            if ((_PB_N+63)%64 == 0) {
              p[_PB_N-1]=p[_PB_N-1]-A[_PB_N-1][t5]*A[_PB_N-1][t5];;
            }
          }
        }
        if ((t2 == 2*t3) && (t2 == 2*t4) && (t2 <= floord(_PB_N-65,32))) {
          if (t2%2 == 0) {
            p[32*t2+63]=1.0/sqrt(p[32*t2+63]);;
          }
        }
        if ((t2 == 2*t3) && (t2 == 2*t4) && (t2 >= ceild(_PB_N-64,32))) {
          if (t2%2 == 0) {
            p[_PB_N-1]=1.0/sqrt(p[_PB_N-1]);;
          }
        }
        if ((t2 == t3+t4) && (t2 <= 2*t3-1)) {
          for (t5=64*t2-64*t3;t5<=64*t2-64*t3+62;t5++) {
            for (t6=64*t3;t6<=min(_PB_N-1,64*t3+63);t6++) {
              A[t6][t5]=A[t6][t5]*p[t5];;
              for (t7=t5+1;t7<=64*t2-64*t3+63;t7++) {
                A[t6][t7]=A[t6][t7]-A[t6][t5]*A[t7][t5];;
              }
            }
          }
        }
        if ((t2 <= t3+t4-1) && (t3 >= t4+1)) {
          for (t5=64*t2-64*t3;t5<=64*t2-64*t3+63;t5++) {
            for (t6=64*t3;t6<=min(_PB_N-1,64*t3+63);t6++) {
              for (t7=64*t4;t7<=64*t4+63;t7++) {
                A[t6][t7]=A[t6][t7]-A[t6][t5]*A[t7][t5];;
              }
            }
          }
        }
        if ((t2 == t3+t4) && (t2 <= 2*t3-1)) {
          for (t6=64*t3;t6<=min(_PB_N-1,64*t3+63);t6++) {
            A[t6][64*t2-64*t3+63]=A[t6][64*t2-64*t3+63]*p[64*t2-64*t3+63];;
          }
        }
      }
    }
  }
}
/* End of CLooG code */

}


int main(int argc, char** argv)
{
  /* Retrieve problem size. */
  int n = N;

#ifndef USE_LOCAL_ARRAYS

  /* Variable declaration/allocation. */
  POLYBENCH_2D_ARRAY_DECL(A, DATA_TYPE, N, N, n, n);
  POLYBENCH_1D_ARRAY_DECL(p, DATA_TYPE, N, n);


  /* Initialize array(s). */
  init_array (n, POLYBENCH_ARRAY(p), POLYBENCH_ARRAY(A));

  /* Start timer. */
  polybench_start_instruments;

  /* Run kernel. */
  kernel_cholesky (n, POLYBENCH_ARRAY(p), POLYBENCH_ARRAY(A));

  /* Stop and print timer. */
  polybench_stop_instruments;
  polybench_print_instruments;

  /* Prevent dead-code elimination. All live-out data must be printed
     by the function call in argument. */
  polybench_prevent_dce(print_array(n, POLYBENCH_ARRAY(A)));

  /* Be clean. */
  POLYBENCH_FREE_ARRAY(A);
  POLYBENCH_FREE_ARRAY(p);
#else

  kernel_cholesky (n);

#endif
  return 0;
}
