#
# common Makefile
#
# Gets included after the local Makefile in an example sub-directory
#
BASEDIR=$(dir $(lastword $(MAKEFILE_LIST)))

CC=icc
#CC=gcc

NPROCS=4
NTHREADS=32
NTHREADS_WITH_MPI=32
NTHREADS_LIST = 1 2 4 8 16 32
NODES_ARRAY[1]= 0
NODES_ARRAY[2]= 0
NODES_ARRAY[4]= 0
NODES_ARRAY[8]= 0
NODES_ARRAY[16]= 0-2
NODES_ARRAY[32]= all
NPROCS_LIST = 1 2 4 8 16 32
TILE_SIZES = 16 32 64 128 256
POLYRTINCDIR=../polyrt
POLYBENCHINCDIR=../polybench/utilities
POLYBENCHSRC=../polybench/utilities/polybench.c
HELPER_SRC=#$(POLYRTINCDIR)/buffer_manager.c
HOSTS_FILE=$(BASEDIR)hosts

PLC=$(BASEDIR)../polycc

# Intel MKL and AMD ACML library paths
MKL=/opt/intel/mkl
ACML=/usr/local/acml

NUMACTL = numactl

#Intel scalapack library paths and compiler flags
S_CFLAGS +=  -DMKL_ILP64 -I$(MKLROOT)/include 
S_LFLAGS +=  $(MKLROOT)/lib/intel64/libmkl_scalapack_ilp64.a -Wl,--start-group  $(MKLROOT)/lib/intel64/libmkl_cdft_core.a $(MKLROOT)/lib/intel64/libmkl_intel_ilp64.a $(MKLROOT)/lib/intel64/libmkl_intel_thread.a $(MKLROOT)/lib/intel64/libmkl_core.a $(MKLROOT)/lib/intel64/libmkl_blacs_intelmpi_ilp64.a -Wl,--end-group -lpthread -lm  


ifeq ($(CC), icc)
	CXX           := icpc
	OPT_FLAGS     := -O3 -fp-model precise -ansi-alias -ipo #-fno-alias -no-alias-const -restrict 
	PAR_FLAGS     := -parallel
	OMP_FLAGS     := -openmp
	# OPENMPI
	MPICC         := OMPI_CC=icc mpicc -D__MPI
	MPICXX        := OMPI_CC=icpc mpicc -D__MPI
	# MVAPICH/MPICH - uncomment the following lines and comment the previous two lines when running on fist
	#MPICC        := mpicc -cc=icc -D__MPI
	#MPICXX       := mpicc -cc=icpc -D__MPI
	# Intel MPI
	#MPICC        := mpiicc -D__MPI
	#MPICXX       := mpiicpc -D__MPI -DMPICH_IGNORE_CXX_SEEK -DMPICH_SKIP_MPICXX
else
	# for gcc
	CXX           := g++
	OPT_FLAGS     := -O3 -ftree-vectorize -msse3 
	PAR_FLAGS     := -ftree-parallelize-loops=4
	OMP_FLAGS     := -fopenmp
	# OPENMPI
	MPICC         := OMPI_CC=gcc mpicc -D__MPI
	MPICXX        := OMPI_CC=g++ mpicc -D__MPI
	# MVAPICH/MPICH - uncomment the following line and comment the previous two lines when running on fist
	#MPICC        := mpicc -cc=gcc -D__MPI
	#MPICXX       := mpicc -cc=g++ -D__MPI
endif

CFLAGS += -DTIME
LDFLAGS += -lm
PLCFLAGS += --isldep --lastwriter
DISTOPT_FLAGS += --cloogsh --commreport
TILEFLAGS += 

#PERFCTR=perfctr

ifdef PERFCTR
	CFLAGS += -DPERFCTR -L/usr/local/lib64 -lpapi
endif

ifdef POLYBENCH
	CFLAGS += -DPOLYBENCH_USE_SCALAR_LB -DPOLYBENCH_TIME -I $(POLYBENCHINCDIR) $(POLYBENCHSRC)
	PLCFLAGS += --variables_not_global
endif
dyn: $(SRC).dyn.c sigma_$(SRC).dyn.c pi_$(SRC).dyn.c
	$(CXX) $(OPT_FLAGS) $(CFLAGS) $(OMP_FLAGS) $(SRC).dyn.c sigma_$(SRC).dyn.c pi_$(SRC).dyn.c \
		$(HELPER_SRC) $(POLYRTINCDIR)/polyrt.c -o $@ -I $(POLYRTINCDIR) $(LDFLAGS) -ltbb

dyn_data_dist: $(SRC).dyn_data_dist.c sigma_$(SRC).dyn_data_dist.c pi_$(SRC).dyn_data_dist.c
	$(CXX) $(OPT_FLAGS) $(CFLAGS) $(OMP_FLAGS) -DUSE_LOCAL_ARRAYS $(SRC).dyn_data_dist.c sigma_$(SRC).dyn_data_dist.c pi_$(SRC).dyn_data_dist.c \
		$(POLYRTINCDIR)/polyrt.c $(POLYRTINCDIR)/buffer_manager.c -o $@ -I $(POLYRTINCDIR) $(LDFLAGS) -ltbb

make dyn_run: dyn
	OMP_NUM_THREADS=$(NTHREADS) GOMP_CPU_AFFINITY=0-`expr $(NTHREADS) - 1` $(NUMACTL) --interleave=${NODES_ARRAY[$(NTHREADS)]}  ./dyn 2> out_dyn

dyn_data_dist_run: dyn_data_dist
	OMP_NUM_THREADS=$(NTHREADS) GOMP_CPU_AFFINITY=0-`expr $(NTHREADS) - 1` $(NUMACTL) --interleave=${NODES_ARRAY[$(NTHREADS)]}  ./dyn_data_dist 2> out_dyn_data_dist

clean:
	rm -f out_* dyn dyn_data_disr 

