#include "floyd-warshall.dyn.h"
#include <assert.h>
#include <omp.h>
#include <math.h>
#define ceild(n,d)  ceil(((double)(n))/((double)(d)))
#define floord(n,d) floor(((double)(n))/((double)(d)))
#define max(x,y)    ((x) > (y)? (x) : (y))
#define min(x,y)    ((x) < (y)? (x) : (y))

/**
 * floyd-warshall.c: This file is part of the PolyBench/C 3.2 test suite.
 *
 *
 * Contact: Louis-Noel Pouchet <pouchet@cse.ohio-state.edu>
 * Web address: http://polybench.sourceforge.net
 */
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <math.h>

/* Include polybench common header. */
#include <polybench.h>


#define N 4096
/* Include benchmark-specific header. */
/* Default data type is double, default size is 1024. */
#include "floyd-warshall.h"


/* Array initialization. */
static
void init_array (int n,
		 DATA_TYPE POLYBENCH_2D(path,N,N,n,n))
{
  int i, j;

  srand(42);
  for (i = 0; i < n; i++)
    for (j = 0; j < n; j++)
        path[i][j] = ((DATA_TYPE) rand()) / n;
      /*path[i][j] = ((DATA_TYPE) (i+1)*(j+1)) / n;*/
}


/* DCE code. Must scan the entire live-out data.
   Can be used also to check the correctness of the output. */
static
void print_array(int n,
		 DATA_TYPE POLYBENCH_2D(path,N,N,n,n))

{
  int i, j;

  for (i = 0; i < n; i++)
    for (j = 0; j < n; j++) {
      fprintf (stderr, DATA_PRINTF_MODIFIER, path[i][j]);
      if ((i * n + j) % 20 == 0) fprintf (stderr, "\n");
    }
  fprintf (stderr, "\n");
}

#define __DECLARATION_OF_path DATA_TYPE POLYBENCH_2D(path,N,N,n,n)
#define __DECLARATION_OF_POINTER_TO_path(path) DATA_TYPE (*path)[N]

#define __PLACE_TO_INSERT_FORWARD_DECLARATIONS
FILE *__debug_print_fp;
tbb::concurrent_priority_queue<__Task, __Compare> pqueue;
int max_num_tasks_loop0_dim0;
int lb_tasks_loop0_dim0;
int max_num_tasks_loop0_dim1;
int lb_tasks_loop0_dim1;
int *tasks_loop0;
void compute_task_0(int t1,int t2,__DECLARATION_OF_path);

/* Main computational kernel. The whole function will be timed,
   including the call and return. */
static
void kernel_floyd_warshall(int n,
			   DATA_TYPE POLYBENCH_2D(path,N,N,n,n))
{
  int i, j, k;


 int _num_threads;
#pragma omp parallel
{_num_threads = omp_get_num_threads();}
 int __tid;
 long int _num_tasks_to_execute;
 int waiting_pops_start[_num_threads], waiting_pops[_num_threads], waiting_pops_threads;
 for (__tid=0; __tid<_num_threads; __tid++) waiting_pops[__tid] = 0;
 double t_local;
 double t_tasks_create;
 double t_tasks_manage_start[_num_threads], t_tasks_manage[_num_threads], t_tasks_manage_threads;
 for (__tid=0; __tid<_num_threads; __tid++) t_tasks_manage[__tid] = 0;
 double t_comp_start[_num_threads], t_comp[_num_threads], t_comp_threads_min, t_comp_threads;
 for (__tid=0; __tid<_num_threads; __tid++) t_comp[__tid] = 0;
 double t_wait_comp_start[_num_threads], t_wait_comp[_num_threads], t_wait_comp_threads;
 for (__tid=0; __tid<_num_threads; __tid++) t_wait_comp[__tid] = 0;
max_num_tasks_loop0_dim0 = ceil(+1*_PB_N-1 - 0 + 1);
lb_tasks_loop0_dim0 = floor(0);
max_num_tasks_loop0_dim1 = ceil(floorf((+1*_PB_N-1)/(float)32) - 0 + 1);
lb_tasks_loop0_dim1 = floor(0);
tasks_loop0 = (int *)malloc(max_num_tasks_loop0_dim0*max_num_tasks_loop0_dim1*sizeof(int));



/*Start of dynamic scheduling code*/

 IF_TIME(t_local = rtclock());

#if defined(__DYNSCHEDULER_DEBUG_PRINT) || defined(__DYNSCHEDULER_MORE_DEBUG_PRINT) || defined(__DEBUG_FLUSH)
 char *debug_print_name = (char *)malloc(128*sizeof(char));
 strcpy(debug_print_name, "debug_print");
 sprintf(debug_print_name+strlen(debug_print_name), "_node%d", my_rank);
 __debug_print_fp = fopen(debug_print_name, "w");
 free(debug_print_name);
#endif
 IF_TIME(t_tasks_create = rtclock());
 init_tasks(_num_threads,&_num_tasks_to_execute,tasks_loop0);
 IF_TIME(t_tasks_create = rtclock() - t_tasks_create);
#ifdef __DYNSCHEDULER_MORE_DEBUG_PRINT
 std::vector<int> prev_firing_count_path_0;
 for (__p=0; __p<__MAX_NUM_RECVS; __p++) {
  prev_firing_count_path_0.push_back(-1);
 }
#endif

#pragma omp parallel private(__tid)
{
 __tid = omp_get_thread_num();
 int __flag;
 int __loop;
 int task_id1,task_id2;

{ /* compute-pack-send threads */
#ifdef __DYNSCHEDULER_DEBUG_PRINT
 fprintf(__debug_print_fp, "node %d compute thread %d num tasks to execute %ld\n", my_rank, __tid, _num_tasks_to_execute);
 IF_DEBUG_FLUSH(fflush(__debug_print_fp));
#endif
 IF_TIME(t_wait_comp_start[__tid] = rtclock());
 while (_num_tasks_to_execute > 0) {
  __Task task;
  __flag = pqueue.try_pop(task);
  if(__flag==0) { IF_TIME(waiting_pops[__tid]++); continue; }
  __loop = task.loop;
  task_id1 = task.id1;
  task_id2 = task.id2;
#ifdef __DYNSCHEDULER_DEBUG_PRINT
  long int captured;
#pragma omp atomic capture
  captured = --_num_tasks_to_execute;
  fprintf(__debug_print_fp, "computing node %d thread %d loop %d task %d %d ready tasks %lu\n", my_rank, __tid, __loop, task_id1, task_id2, pqueue.size());
  IF_DEBUG_FLUSH(fflush(__debug_print_fp));
#else
#pragma omp atomic
  _num_tasks_to_execute--;
#endif
  switch(__loop){
   case 0:
    IF_TIME(t_wait_comp[__tid] += rtclock() - t_wait_comp_start[__tid]);
    IF_TIME(t_comp_start[__tid] = rtclock());
    compute_task_0(task_id1,task_id2,path);
    IF_TIME(t_comp[__tid] += rtclock() - t_comp_start[__tid]);
    IF_TIME(t_tasks_manage_start[__tid] = rtclock());
    local_update_dep_tasks_0(task_id1,task_id2,_PB_N,_num_threads,tasks_loop0);
    IF_TIME(t_tasks_manage[__tid] += rtclock() - t_tasks_manage_start[__tid]);
    IF_TIME(t_wait_comp_start[__tid] = rtclock());
    break;
   default:
    assert(0);
  }
#ifdef __DYNSCHEDULER_MORE_DEBUG_PRINT
  fprintf(__debug_print_fp, "computed node %d thread %d loop %d task %d %d tasks left %ld\n", my_rank, __tid, __loop, task_id1, task_id2, captured);
  IF_DEBUG_FLUSH(fflush(__debug_print_fp));
#endif
 }

#ifdef __DYNSCHEDULER_DEBUG_PRINT
 fprintf(__debug_print_fp, "node %d compute thread %d done\n", my_rank, __tid);
 IF_DEBUG_FLUSH(fflush(__debug_print_fp));
#endif
}

}

#ifdef __DYNSCHEDULER_DEBUG_PRINT
 fclose(__debug_print_fp);
#endif



/*End of dynamic scheduling code*/


#ifdef TIME
 waiting_pops_threads = waiting_pops[0];
                for (__tid=1; __tid<_num_threads; __tid++) if (waiting_pops_threads < waiting_pops[__tid]) waiting_pops_threads = waiting_pops[__tid];
 t_local = rtclock() - t_local;
 t_tasks_manage_threads = t_tasks_manage[0];
                for (__tid=1; __tid<_num_threads; __tid++) if (t_tasks_manage_threads < t_tasks_manage[__tid]) t_tasks_manage_threads = t_tasks_manage[__tid];
 t_comp_threads_min = t_comp[0];
                for (__tid=1; __tid<_num_threads; __tid++) if (t_comp_threads_min > t_comp[__tid]) t_comp_threads_min = t_comp[__tid];
 t_comp_threads = t_comp[0];
                for (__tid=1; __tid<_num_threads; __tid++) if (t_comp_threads < t_comp[__tid]) t_comp_threads = t_comp[__tid];
 t_wait_comp_threads = t_wait_comp[0];
                for (__tid=1; __tid<_num_threads; __tid++) if (t_wait_comp_threads < t_wait_comp[__tid]) t_wait_comp_threads = t_wait_comp[__tid];
 char buffer[4096];
 strcpy(buffer, "");
 sprintf(buffer+strlen(buffer), "Minimum computation time across all threads: %lf\n", t_comp_threads_min);

 sprintf(buffer+strlen(buffer), "Maximum computation time across all threads: %lf\n", t_comp_threads);

 sprintf(buffer+strlen(buffer), "Maximum 'waiting for computation' time across all threads: %lf\n", t_wait_comp_threads);

 sprintf(buffer+strlen(buffer), "Maximum number of waiting/unsuccessful POPs across all threads: %d\n", waiting_pops_threads);

 sprintf(buffer+strlen(buffer), "Maximum tasks management time across all threads: %lf\n", t_tasks_manage_threads);

 sprintf(buffer+strlen(buffer), "Tasks creation time: %lf\n", t_tasks_create);

 sprintf(buffer+strlen(buffer), "Sum of split times: %lf\n", t_comp_threads+t_wait_comp_threads+t_tasks_manage_threads+t_tasks_create);

 sprintf(buffer+strlen(buffer), "Total time: %lf\n", t_local);

 fprintf(stdout, "%s\n", buffer);
#endif

}


int main(int argc, char** argv)
{
  /* Retrieve problem size. */
  int n = N;

  /* Variable declaration/allocation. */
  POLYBENCH_2D_ARRAY_DECL(path, DATA_TYPE, N, N, n, n);


  /* Initialize array(s). */
  init_array (n, POLYBENCH_ARRAY(path));

  /* Start timer. */
  polybench_start_instruments;

  /* Run kernel. */
  kernel_floyd_warshall (n, POLYBENCH_ARRAY(path));

  /* Stop and print timer. */
  polybench_stop_instruments;
  polybench_print_instruments;

  /* Prevent dead-code elimination. All live-out data must be printed
     by the function call in argument. */
  polybench_prevent_dce(print_array(n, POLYBENCH_ARRAY(path)));

  /* Be clean. */
  POLYBENCH_FREE_ARRAY(path);

  return 0;
}
void init_tasks(int num_threads,long int *p_num_tasks_to_execute,int *tasks_loop0)
{
int local_dep_tasks, affinity;
long int _num_tasks_to_execute = 0;


  int t1, t2, t3;

 int lbp, ubp;
 register int lbv, ubv;

/* Start of CLooG code */
if (_PB_N >= 1) {
  lbp=0;
  ubp=_PB_N-1;
#pragma omp parallel for private(local_dep_tasks, affinity, t3) reduction(+:_num_tasks_to_execute)
  for (t2=lbp;t2<=ubp;t2++) {
    for (t3=0;t3<=floord(_PB_N-1,32);t3++) {
      _num_tasks_to_execute++;tasks_loop0[(t2-lb_tasks_loop0_dim0)*max_num_tasks_loop0_dim1+(t3-lb_tasks_loop0_dim1)+0] = get_num_local_src_tasks_0(t2,t3,_PB_N);if (tasks_loop0[(t2-lb_tasks_loop0_dim0)*max_num_tasks_loop0_dim1+(t3-lb_tasks_loop0_dim1)+0] == 0) { local_dep_tasks = get_num_local_dep_tasks_0(t2,t3,_PB_N); affinity = pi_0(t2,t3,_PB_N,num_threads); __Task task(0,t2,t3,local_dep_tasks,affinity); pqueue.push(task); IF_DYNSCHEDULER_MORE_DEBUG_PRINT( fprintf(__debug_print_fp, "added node %d loop 0 task %d %d                 local_dep_tasks %d ready tasks %lu\n", my_rank, t2,t3, local_dep_tasks, pqueue.size())); IF_DEBUG_FLUSH(fflush(__debug_print_fp)); };
    }
  }
}
/* End of CLooG code */

*p_num_tasks_to_execute = _num_tasks_to_execute;
}

void compute_task_0(int t1,int t2,__DECLARATION_OF_path)
{



  int t3, t4, t5;

 int lbp, ubp;
 register int lbv, ubv;

/* Start of CLooG code */
for (t3=0;t3<=floord(_PB_N-1,32);t3++) {
  for (t4=32*t2;t4<=min(_PB_N-1,32*t2+31);t4++) {
    lbv=32*t3;
    ubv=min(_PB_N-1,32*t3+31);
    #if (defined __DATA_DIST_DECLS || defined USE_LOCAL_ARRAYS)
    ;
    #endif
#pragma ivdep
#pragma vector always
    for (t5=lbv;t5<=ubv;t5++) {
      path[t4][t5]=path[t4][t5]<path[t4][t1]+path[t1][t5]?path[t4][t5]:path[t4][t1]+path[t1][t5];;
    }
  }
}
/* End of CLooG code */


}
