#include "floyd-warshall.dyn_data_dist.h"
#include <assert.h>
#include <omp.h>
#include <math.h>
#define ceild(n,d)  ceil(((double)(n))/((double)(d)))
#define floord(n,d) floor(((double)(n))/((double)(d)))
#define max(x,y)    ((x) > (y)? (x) : (y))
#define min(x,y)    ((x) < (y)? (x) : (y))

/**
 * floyd-warshall.c: This file is part of the PolyBench/C 3.2 test suite.
 *
 *
 * Contact: Louis-Noel Pouchet <pouchet@cse.ohio-state.edu>
 * Web address: http://polybench.sourceforge.net
 */
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <math.h>

/* Include polybench common header. */
#include <polybench.h>

#define N 4096
/* Include benchmark-specific header. */
/* Default data type is double, default size is 1024. */
#include "floyd-warshall.h"


/* Array initialization. */
static
void init_array (int n,
		 DATA_TYPE POLYBENCH_2D(path,N,N,n,n))
{
  int i, j;

  srand(42);
  for (i = 0; i < n; i++)
    for (j = 0; j < n; j++)
        path[i][j] = ((DATA_TYPE) rand()) / n;
      /*path[i][j] = ((DATA_TYPE) (i+1)*(j+1)) / n;*/
}


/* DCE code. Must scan the entire live-out data.
   Can be used also to check the correctness of the output. */
static
void print_array(int n,
		 DATA_TYPE POLYBENCH_2D(path,N,N,n,n))

{
  int i, j;

  for (i = 0; i < n; i++)
    for (j = 0; j < n; j++) {
      fprintf (stderr, DATA_PRINTF_MODIFIER, path[i][j]);
      if ((i * n + j) % 20 == 0) fprintf (stderr, "\n");
    }
  fprintf (stderr, "\n");
}

#define __DECLARATION_OF_path DATA_TYPE POLYBENCH_2D(path,N,N,n,n)
#define __DECLARATION_OF_POINTER_TO_path(path) DATA_TYPE (*path)[N]

#define __PLACE_TO_INSERT_FORWARD_DECLARATIONS
void read_in_copy_path_0(int ts1,int ts2 
#ifndef USE_LOCAL_ARRAYS
 ,__DECLARATION_OF_path 
#endif
 , double** path_trans,int path_size_1, BufferManager *buff_mang_path);
void data_tile_init_0(int my_rank, int nprocs 
#ifndef USE_LOCAL_ARRAYS
 ,__DECLARATION_OF_path 
#endif
 , double** path_trans,int path_size_1, BufferManager *buff_mang_path);
FILE *__debug_print_fp;
tbb::concurrent_priority_queue<__Task, __Compare> pqueue;
int max_num_tasks_loop0_dim0;
int lb_tasks_loop0_dim0;
int max_num_tasks_loop0_dim1;
int lb_tasks_loop0_dim1;
int *tasks_loop0;
void data_tile_copy_back_path_0(int ts1,int ts2 
#ifndef USE_LOCAL_ARRAYS
 ,__DECLARATION_OF_path 
#endif
 , double** path_trans,int path_size_1, BufferManager *buff_mang_path);
void copy_back_writeout_0(int my_rank, int nprocs 
#ifndef USE_LOCAL_ARRAYS
 ,__DECLARATION_OF_path 
#endif
 , double** path_trans,int path_size_1, BufferManager *buff_mang_path);

/* Main computational kernel. The whole function will be timed,
   including the call and return. */
static
void kernel_floyd_warshall(int n,
			   DATA_TYPE POLYBENCH_2D(path,N,N,n,n))
{
  int i, j, k;


 int _num_threads;
#pragma omp parallel
{_num_threads = omp_get_num_threads();}
 int __tid;
 long int _num_tasks_to_execute;
 int waiting_pops_start[_num_threads], waiting_pops[_num_threads], waiting_pops_threads;
 for (__tid=0; __tid<_num_threads; __tid++) waiting_pops[__tid] = 0;
 double t_local;
 double t_tasks_create;
 double t_tasks_manage_start[_num_threads], t_tasks_manage[_num_threads], t_tasks_manage_threads;
 for (__tid=0; __tid<_num_threads; __tid++) t_tasks_manage[__tid] = 0;
 double t_comp_start[_num_threads], t_comp[_num_threads], t_comp_threads_min, t_comp_threads;
 for (__tid=0; __tid<_num_threads; __tid++) t_comp[__tid] = 0;
 double t_wait_comp_start[_num_threads], t_wait_comp[_num_threads], t_wait_comp_threads;
 for (__tid=0; __tid<_num_threads; __tid++) t_wait_comp[__tid] = 0;
 double t_data_mang_start[_num_threads], t_data_mang[_num_threads], t_data_mang_threads;
 for (__tid=0; __tid<_num_threads; __tid++) t_data_mang[__tid] = 0;
 double t_writeout_start, t_writeout = 0.0;
 double t_data_init_start, t_data_init = 0.0;
int path_size_0 = (+1*(floorf((+1*_PB_N-1)/(float)32) - 0 + 1)+1*(floorf((+1*_PB_N-1)/(float)32) - 0 + 1));
int path_size_1 = (+1*(floorf((+1*_PB_N-1)/(float)32) - 0 + 1)+1*(floorf((+1*_PB_N-1)/(float)32) - 0 + 1));
double **path_trans = (double **)malloc((path_size_0 * path_size_1) * sizeof(double));
BufferManager *buff_mang_path = init_buffer_manager(path_size_0 * path_size_1);
max_num_tasks_loop0_dim0 = ceil(+1*_PB_N-1 - 0 + 1);
lb_tasks_loop0_dim0 = floor(0);
max_num_tasks_loop0_dim1 = ceil(floorf((+1*_PB_N-1)/(float)32) - 0 + 1);
lb_tasks_loop0_dim1 = floor(0);
tasks_loop0 = (int *)malloc(max_num_tasks_loop0_dim0*max_num_tasks_loop0_dim1*sizeof(int));
 IF_TIME(t_data_init_start = rtclock());
data_tile_init_0(0, 0 
#ifndef USE_LOCAL_ARRAYS
 , path 
#endif
 , path_trans, path_size_1, buff_mang_path);
 IF_TIME(t_data_init += rtclock() - t_data_init_start)



/*Start of dynamic scheduling code*/


 IF_TIME(t_local = rtclock());
#if defined(__DYNSCHEDULER_DEBUG_PRINT) || defined(__DYNSCHEDULER_MORE_DEBUG_PRINT) || defined(__DEBUG_FLUSH)
 char *debug_print_name = (char *)malloc(128*sizeof(char));
 strcpy(debug_print_name, "debug_print");
 sprintf(debug_print_name+strlen(debug_print_name), "_node%d", my_rank);
 __debug_print_fp = fopen(debug_print_name, "w");
 free(debug_print_name);
#endif
 IF_TIME(t_tasks_create = rtclock());
 init_tasks(_num_threads,&_num_tasks_to_execute,tasks_loop0);
 IF_TIME(t_tasks_create = rtclock() - t_tasks_create);
#ifdef __DYNSCHEDULER_MORE_DEBUG_PRINT
 std::vector<int> prev_firing_count_path_0;
 for (__p=0; __p<__MAX_NUM_RECVS; __p++) {
  prev_firing_count_path_0.push_back(-1);
 }
#endif

#pragma omp parallel private(__tid)
{
 __tid = omp_get_thread_num();
 int __flag;
 int __loop;
 int task_id1,task_id2;

{ /* compute-pack-send threads */
#ifdef __DYNSCHEDULER_DEBUG_PRINT
 fprintf(__debug_print_fp, "node %d compute thread %d num tasks to execute %ld\n", my_rank, __tid, _num_tasks_to_execute);
 IF_DEBUG_FLUSH(fflush(__debug_print_fp));
#endif
 IF_TIME(t_wait_comp_start[__tid] = rtclock());
 while (_num_tasks_to_execute > 0) {
  __Task task;
  __flag = pqueue.try_pop(task);
  if(__flag==0) { IF_TIME(waiting_pops[__tid]++); continue; }
  __loop = task.loop;
  task_id1 = task.id1;
  task_id2 = task.id2;
#ifdef __DYNSCHEDULER_DEBUG_PRINT
  long int captured;
#pragma omp atomic capture
  captured = --_num_tasks_to_execute;
  fprintf(__debug_print_fp, "computing node %d thread %d loop %d task %d %d ready tasks %lu\n", my_rank, __tid, __loop, task_id1, task_id2, pqueue.size());
  IF_DEBUG_FLUSH(fflush(__debug_print_fp));
#else
#pragma omp atomic
  _num_tasks_to_execute--;
#endif
  switch(__loop){
   case 0:
    IF_TIME(t_wait_comp[__tid] += rtclock() - t_wait_comp_start[__tid]);
    IF_TIME(t_data_mang_start[__tid] = rtclock());
    data_tile_alloc_0(task_id1,task_id2, path_trans, path_size_1, buff_mang_path);
    IF_TIME(t_data_mang[__tid] += rtclock() - t_data_mang_start[__tid]);
    IF_TIME(t_comp_start[__tid] = rtclock());
    compute_task_0(task_id1,task_id2, path_trans, path_size_1, buff_mang_path);
    IF_TIME(t_comp[__tid] += rtclock() - t_comp_start[__tid]);
    IF_TIME(t_data_mang_start[__tid] = rtclock());
    data_tile_ref_count_update_0(task_id1,task_id2, path_trans, path_size_1, buff_mang_path);
    IF_TIME(t_data_mang[__tid] += rtclock() - t_data_mang_start[__tid]);
    IF_TIME(t_tasks_manage_start[__tid] = rtclock());
    local_update_dep_tasks_0(task_id1,task_id2,_PB_N,_num_threads,tasks_loop0);
    IF_TIME(t_tasks_manage[__tid] += rtclock() - t_tasks_manage_start[__tid]);
    IF_TIME(t_wait_comp_start[__tid] = rtclock());
    break;
   default:
    assert(0);
  }
#ifdef __DYNSCHEDULER_MORE_DEBUG_PRINT
  fprintf(__debug_print_fp, "computed node %d thread %d loop %d task %d %d tasks left %ld\n", my_rank, __tid, __loop, task_id1, task_id2, captured);
  IF_DEBUG_FLUSH(fflush(__debug_print_fp));
#endif
 }

#ifdef __DYNSCHEDULER_DEBUG_PRINT
 fprintf(__debug_print_fp, "node %d compute thread %d done\n", my_rank, __tid);
 IF_DEBUG_FLUSH(fflush(__debug_print_fp));
#endif
}

}

#ifdef __DYNSCHEDULER_DEBUG_PRINT
 fclose(__debug_print_fp);
#endif



/*End of dynamic scheduling code*/


 IF_TIME(t_writeout = rtclock());
 
#ifndef USE_LOCAL_ARRAYS
 if (fopen(".test", "r")) {
 copy_back_writeout_0(0, 0 
#ifndef USE_LOCAL_ARRAYS
 , path 
#endif
 , path_trans, path_size_1, buff_mang_path);

} 
#endif
 IF_TIME(t_writeout = rtclock() - t_writeout);
free_buffer_mang(buff_mang_path);
#ifdef TIME
 waiting_pops_threads = waiting_pops[0];
                for (__tid=1; __tid<_num_threads; __tid++) if (waiting_pops_threads < waiting_pops[__tid]) waiting_pops_threads = waiting_pops[__tid];
 t_local = rtclock() - t_local;
 t_tasks_manage_threads = t_tasks_manage[0];
                for (__tid=1; __tid<_num_threads; __tid++) if (t_tasks_manage_threads < t_tasks_manage[__tid]) t_tasks_manage_threads = t_tasks_manage[__tid];
 t_comp_threads_min = t_comp[0];
                for (__tid=1; __tid<_num_threads; __tid++) if (t_comp_threads_min > t_comp[__tid]) t_comp_threads_min = t_comp[__tid];
 t_comp_threads = t_comp[0];
                for (__tid=1; __tid<_num_threads; __tid++) if (t_comp_threads < t_comp[__tid]) t_comp_threads = t_comp[__tid];
 t_wait_comp_threads = t_wait_comp[0];
                for (__tid=1; __tid<_num_threads; __tid++) if (t_wait_comp_threads < t_wait_comp[__tid]) t_wait_comp_threads = t_wait_comp[__tid];
 char buffer[4096];
 strcpy(buffer, "");
 sprintf(buffer+strlen(buffer), "Minimum computation time across all threads: %lf\n", t_comp_threads_min);

 sprintf(buffer+strlen(buffer), "Maximum computation time across all threads: %lf\n", t_comp_threads);

 sprintf(buffer+strlen(buffer), "Maximum 'waiting for computation' time across all threads: %lf\n", t_wait_comp_threads);

 sprintf(buffer+strlen(buffer), "Maximum number of waiting/unsuccessful POPs across all threads: %d\n", waiting_pops_threads);

 sprintf(buffer+strlen(buffer), "Maximum tasks management time across all threads: %lf\n", t_tasks_manage_threads);

 sprintf(buffer+strlen(buffer), "Tasks creation time: %lf\n", t_tasks_create);

 sprintf(buffer+strlen(buffer), "Sum of split times: %lf\n", t_comp_threads+t_wait_comp_threads+t_tasks_manage_threads+t_tasks_create);

 sprintf(buffer+strlen(buffer), "Total time: %lf\n", t_local);

 fprintf(stdout, "%s\n", buffer);
#endif

}


int main(int argc, char** argv)
{
  /* Retrieve problem size. */
  int n = N;

  /* Variable declaration/allocation. */
  POLYBENCH_2D_ARRAY_DECL(path, DATA_TYPE, N, N, n, n);


  /* Initialize array(s). */
  init_array (n, POLYBENCH_ARRAY(path));

  /* Start timer. */
  polybench_start_instruments;

  /* Run kernel. */
  kernel_floyd_warshall (n, POLYBENCH_ARRAY(path));

  /* Stop and print timer. */
  polybench_stop_instruments;
  polybench_print_instruments;

  /* Prevent dead-code elimination. All live-out data must be printed
     by the function call in argument. */
  polybench_prevent_dce(print_array(n, POLYBENCH_ARRAY(path)));

  /* Be clean. */
  POLYBENCH_FREE_ARRAY(path);

  return 0;
}
void data_tile_ref_count_init_path_0(int ts1,int ts2, double** path_trans,int path_size_1, BufferManager *buff_mang_path)
{


  int t1, t2, t3;

 int lbp, ubp;
 register int lbv, ubv;

/* Start of CLooG code */
if ((ts1 >= 0) && (ts1 <= _PB_N-1) && (ts2 >= 0) && (32*ts2 <= _PB_N-1)) {
  for (t2=ceild(ts1-31,32);t2<=min(floord(ts1,32),ts2-1);t2++) {
    for (t3=0;t3<=floord(_PB_N-1,32);t3++) {
      increment_ref_count(buff_mang_path, (t2)* ( path_size_1 ) + (t3));;
    }
  }
  for (t3=0;t3<=floord(_PB_N-1,32);t3++) {
    increment_ref_count(buff_mang_path, (ts2)* ( path_size_1 ) + (t3));;
  }
  for (t2=max(ceild(ts1-31,32),ts2+1);t2<=floord(ts1,32);t2++) {
    for (t3=0;t3<=floord(_PB_N-1,32);t3++) {
      increment_ref_count(buff_mang_path, (t2)* ( path_size_1 ) + (t3));;
    }
  }
}
/* End of CLooG code */

}

void read_in_data_tile_alloc_path_0(int ts1,int ts2, double** path_trans,int path_size_1, BufferManager *buff_mang_path)
{


  int t1, t2, t3;

 int lbp, ubp;
 register int lbv, ubv;

/* Start of CLooG code */
if ((ts1 >= 0) && (ts1 <= _PB_N-1) && (ts2 >= 0) && (32*ts2 <= _PB_N-1)) {
  if (ts1 >= 32*ts2+32) {
    for (t3=0;t3<=floord(_PB_N-1,32);t3++) {
      if(path_trans[(ts2) * ((path_size_1)) + (t3)] == NULL) path_trans[(ts2)* ( path_size_1 ) + (t3)] = (double *) buffer_alloc_atomic((ts2)* ( path_size_1 ) + (t3), 32*32 * sizeof(double), buff_mang_path);;
    }
  }
  t2 = floord(ts1,32);
  for (t3=0;t3<=floord(_PB_N-1,32);t3++) {
    if(path_trans[(t2) * ((path_size_1)) + (t3)] == NULL) path_trans[(t2)* ( path_size_1 ) + (t3)] = (double *) buffer_alloc_atomic((t2)* ( path_size_1 ) + (t3), 32*32 * sizeof(double), buff_mang_path);;
  }
  if (ts1 <= 32*ts2-1) {
    for (t3=0;t3<=floord(_PB_N-1,32);t3++) {
      if(path_trans[(ts2) * ((path_size_1)) + (t3)] == NULL) path_trans[(ts2)* ( path_size_1 ) + (t3)] = (double *) buffer_alloc_atomic((ts2)* ( path_size_1 ) + (t3), 32*32 * sizeof(double), buff_mang_path);;
    }
  }
}
/* End of CLooG code */

}

void read_in_copy_path_0(int ts1,int ts2 
#ifndef USE_LOCAL_ARRAYS
 ,__DECLARATION_OF_path 
#endif
 , double** path_trans,int path_size_1, BufferManager *buff_mang_path)
{


  int t1, t2, t3;

 int lbp, ubp;
 register int lbv, ubv;

/* Start of CLooG code */
if ((ts1 >= 0) && (ts1 <= _PB_N-1) && (ts2 >= 0) && (32*ts2 <= _PB_N-1)) {
  for (t2=32*ts2;t2<=min(ts1-1,32*ts2+31);t2++) {
    for (t3=0;t3<=_PB_N-1;t3++) {
      
#ifndef USE_LOCAL_ARRAYS
 path_trans[((1)*( t2 ))/32* ( path_size_1 ) + ((1)*( t3 ))/32][((1)*( t2 ))%32* (32) + ((1)*( t3 ))%32] = path[t2][t3]; 
#else
 
#endif
;
    }
  }
  for (t3=0;t3<=_PB_N-1;t3++) {
    
#ifndef USE_LOCAL_ARRAYS
 path_trans[((1)*( ts1 ))/32* ( path_size_1 ) + ((1)*( t3 ))/32][((1)*( ts1 ))%32* (32) + ((1)*( t3 ))%32] = path[ts1][t3]; 
#else
 
#endif
;
  }
  for (t2=max(32*ts2,ts1+1);t2<=min(_PB_N-1,32*ts2+31);t2++) {
    for (t3=0;t3<=_PB_N-1;t3++) {
      
#ifndef USE_LOCAL_ARRAYS
 path_trans[((1)*( t2 ))/32* ( path_size_1 ) + ((1)*( t3 ))/32][((1)*( t2 ))%32* (32) + ((1)*( t3 ))%32] = path[t2][t3]; 
#else
 
#endif
;
    }
  }
}
/* End of CLooG code */

}

void data_tile_init_0(int my_rank, int nprocs 
#ifndef USE_LOCAL_ARRAYS
 ,__DECLARATION_OF_path 
#endif
 , double** path_trans,int path_size_1, BufferManager *buff_mang_path)
{





  int t1, t2, t3;

 int lbp, ubp;
 register int lbv, ubv;

/* Start of CLooG code */
if (_PB_N >= 1) {
  for (t2=0;t2<=floord(_PB_N-1,32);t2++) {
    for (t3=0;t3<=floord(_PB_N-1,32);t3++) {
      path_trans[(t2)* ( path_size_1 ) + (t3)] = NULL;;
    }
  }
  for (t2=0;t2<=_PB_N-1;t2++) {
    for (t3=0;t3<=floord(_PB_N-1,32);t3++) {
      data_tile_ref_count_init_path_0(t2, t3, path_trans, path_size_1,buff_mang_path);;
    }
  }
  for (t2=0;t2<=_PB_N-1;t2++) {
    for (t3=0;t3<=floord(_PB_N-1,32);t3++) {
      data_tile_ref_count_init_path_0(t2, t3, path_trans, path_size_1,buff_mang_path);;
    }
  }
  for (t2=0;t2<=_PB_N-1;t2++) {
    for (t3=0;t3<=floord(_PB_N-1,32);t3++) {
      read_in_data_tile_alloc_path_0(t2, t3, path_trans, path_size_1,buff_mang_path);read_in_copy_path_0(t2, t3 
#ifndef USE_LOCAL_ARRAYS
, path 
#endif
, path_trans, path_size_1,buff_mang_path);;
    }
  }
}
/* End of CLooG code */




}

void init_tasks(int num_threads,long int *p_num_tasks_to_execute,int *tasks_loop0)
{
int local_dep_tasks, affinity;
long int _num_tasks_to_execute = 0;


  int t1, t2, t3;

 int lbp, ubp;
 register int lbv, ubv;

/* Start of CLooG code */
if (_PB_N >= 1) {
  lbp=0;
  ubp=_PB_N-1;
#pragma omp parallel for private(local_dep_tasks, affinity, t3) reduction(+:_num_tasks_to_execute)
  for (t2=lbp;t2<=ubp;t2++) {
    for (t3=0;t3<=floord(_PB_N-1,32);t3++) {
      _num_tasks_to_execute++;tasks_loop0[(t2-lb_tasks_loop0_dim0)*max_num_tasks_loop0_dim1+(t3-lb_tasks_loop0_dim1)+0] = get_num_local_src_tasks_0(t2,t3,_PB_N);if (tasks_loop0[(t2-lb_tasks_loop0_dim0)*max_num_tasks_loop0_dim1+(t3-lb_tasks_loop0_dim1)+0] == 0) { local_dep_tasks = get_num_local_dep_tasks_0(t2,t3,_PB_N); affinity = pi_0(t2,t3,_PB_N,num_threads); __Task task(0,t2,t3,local_dep_tasks,affinity); pqueue.push(task); IF_DYNSCHEDULER_MORE_DEBUG_PRINT( fprintf(__debug_print_fp, "added node %d loop 0 task %d %d                 local_dep_tasks %d ready tasks %lu\n", my_rank, t2,t3, local_dep_tasks, pqueue.size())); IF_DEBUG_FLUSH(fflush(__debug_print_fp)); };
    }
  }
}
/* End of CLooG code */

*p_num_tasks_to_execute = _num_tasks_to_execute;
}

void data_tile_alloc_0(int ts1,int ts2, double** path_trans,int path_size_1, BufferManager *buff_mang_path)
{


  int t1, t2, t3;

 int lbp, ubp;
 register int lbv, ubv;

/* Start of CLooG code */
if ((ts1 >= 0) && (ts1 <= _PB_N-1) && (ts2 >= 0) && (32*ts2 <= _PB_N-1)) {
  for (t2=ceild(ts1-31,32);t2<=min(floord(ts1,32),ts2-1);t2++) {
    for (t3=0;t3<=floord(_PB_N-1,32);t3++) {
      if(path_trans[(t2) * ((path_size_1)) + (t3)] == NULL) path_trans[(t2)* ( path_size_1 ) + (t3)] = (double *) buffer_alloc_atomic((t2)* ( path_size_1 ) + (t3), 32*32 * sizeof(double), buff_mang_path);;
    }
  }
  for (t3=0;t3<=floord(_PB_N-1,32);t3++) {
    if(path_trans[(ts2) * ((path_size_1)) + (t3)] == NULL) path_trans[(ts2)* ( path_size_1 ) + (t3)] = (double *) buffer_alloc_atomic((ts2)* ( path_size_1 ) + (t3), 32*32 * sizeof(double), buff_mang_path);;
  }
  for (t2=max(ceild(ts1-31,32),ts2+1);t2<=floord(ts1,32);t2++) {
    for (t3=0;t3<=floord(_PB_N-1,32);t3++) {
      if(path_trans[(t2) * ((path_size_1)) + (t3)] == NULL) path_trans[(t2)* ( path_size_1 ) + (t3)] = (double *) buffer_alloc_atomic((t2)* ( path_size_1 ) + (t3), 32*32 * sizeof(double), buff_mang_path);;
    }
  }
}
/* End of CLooG code */

}

void data_tile_ref_count_update_0(int ts1,int ts2, double** path_trans,int path_size_1, BufferManager *buff_mang_path)
{


  int t1, t2, t3;

 int lbp, ubp;
 register int lbv, ubv;

/* Start of CLooG code */
if ((ts1 >= 0) && (ts1 <= _PB_N-1) && (ts2 >= 0) && (32*ts2 <= _PB_N-1)) {
  for (t2=ceild(ts1-31,32);t2<=min(floord(ts1,32),ts2-1);t2++) {
    for (t3=0;t3<=floord(_PB_N-1,32);t3++) {
      decrement_ref_count(buff_mang_path, (t2)* ( path_size_1 ) + (t3));;
    }
  }
  for (t3=0;t3<=floord(_PB_N-1,32);t3++) {
    decrement_ref_count(buff_mang_path, (ts2)* ( path_size_1 ) + (t3));;
  }
  for (t2=max(ceild(ts1-31,32),ts2+1);t2<=floord(ts1,32);t2++) {
    for (t3=0;t3<=floord(_PB_N-1,32);t3++) {
      decrement_ref_count(buff_mang_path, (t2)* ( path_size_1 ) + (t3));;
    }
  }
}
/* End of CLooG code */

}

void compute_task_0(int t1,int t2, double** path_trans,int path_size_1, BufferManager *buff_mang_path)
{



int i_j_0_div = 0;int i_j_0_mod = 0;double *path_i_j_0 = NULL;int i_k_0_div = 0;int i_k_0_mod = 0;double *path_i_k_0 = NULL;int k_j_0_div = 0;int k_j_0_mod = 0;double *path_k_j_0 = NULL; int t3, t4, t5;

 int lbp, ubp;
 register int lbv, ubv;

/* Start of CLooG code */
for (t3=0;t3<=floord(_PB_N-1,32);t3++) {
  for (t4=32*t2;t4<=min(_PB_N-1,32*t2+31);t4++) {
    lbv=32*t3;
    ubv=min(_PB_N-1,32*t3+31);
    #if (defined __DATA_DIST_DECLS || defined USE_LOCAL_ARRAYS)
    i_j_0_div = ((1)*( t4 ))/32* ( path_size_1 ) + ((1)*( lbv ))/32;i_j_0_mod = ((1)*( t4 ))%32* (32) + ((1)*( lbv ))%32;path_i_j_0 = path_trans[i_j_0_div];i_k_0_div = ((1)*( t4 ))/32* ( path_size_1 ) + ((1)*( t1 ))/32;i_k_0_mod = ((1)*( t4 ))%32* (32) + ((1)*( t1 ))%32;path_i_k_0 = path_trans[i_k_0_div];k_j_0_div = ((1)*( t1 ))/32* ( path_size_1 ) + ((1)*( lbv ))/32;k_j_0_mod = ((1)*( t1 ))%32* (32) + ((1)*( lbv ))%32;path_k_j_0 = path_trans[k_j_0_div];;
    #endif
#pragma ivdep
#pragma vector always
    for (t5=lbv;t5<=ubv;t5++) {
      path_i_j_0[i_j_0_mod + t5 - lbv]=path_i_j_0[i_j_0_mod + t5 - lbv]<path_i_k_0[i_k_0_mod]+path_k_j_0[k_j_0_mod + t5 - lbv]?path_i_j_0[i_j_0_mod + t5 - lbv]:path_i_k_0[i_k_0_mod]+path_k_j_0[k_j_0_mod + t5 - lbv];;
    }
  }
}
/* End of CLooG code */


}

void data_tile_copy_back_path_0(int ts1,int ts2 
#ifndef USE_LOCAL_ARRAYS
 ,__DECLARATION_OF_path 
#endif
 , double** path_trans,int path_size_1, BufferManager *buff_mang_path)
{
#ifndef USE_LOCAL_ARRAYS


  int t1, t2, t3;

 int lbp, ubp;
 register int lbv, ubv;

/* Start of CLooG code */
if ((ts1 >= 0) && (ts1 <= _PB_N-1) && (ts2 >= 0) && (32*ts2 <= _PB_N-1)) {
  for (t2=32*ts2;t2<=min(_PB_N-1,32*ts2+31);t2++) {
    for (t3=0;t3<=_PB_N-1;t3++) {
      path[t2][t3] = path_trans[((1)*( t2 ))/32* ( path_size_1 ) + ((1)*( t3 ))/32][((1)*( t2 ))%32* (32) + ((1)*( t3 ))%32];;
    }
  }
}
/* End of CLooG code */
#endif

}

void copy_back_writeout_0(int my_rank, int nprocs 
#ifndef USE_LOCAL_ARRAYS
 ,__DECLARATION_OF_path 
#endif
 , double** path_trans,int path_size_1, BufferManager *buff_mang_path)
{
#ifndef USE_LOCAL_ARRAYS


  int t1, t2, t3;

 int lbp, ubp;
 register int lbv, ubv;

/* Start of CLooG code */
if (_PB_N >= 1) {
  for (t2=0;t2<=_PB_N-1;t2++) {
    for (t3=0;t3<=floord(_PB_N-1,32);t3++) {
      data_tile_copy_back_path_0(t2,t3, path, path_trans, path_size_1,buff_mang_path);;
    }
  }
}
/* End of CLooG code */
#endif

}
