/**
 * lu.c: This file is part of the PolyBench/C 3.2 test suite.
 *
 *
 * Contact: Louis-Noel Pouchet <pouchet@cse.ohio-state.edu>
 * Web address: http://polybench.sourceforge.net
 */
#include <math.h>
#include <limits.h>
#define ceild(n,d)  ceil(((double)(n))/((double)(d)))
#define floord(n,d) floor(((double)(n))/((double)(d)))
#define min(x,y)    ((x) < (y)? (x) : (y))
#include <upc_relaxed.h>
#include <bupc_collectivev.h>
#include <upc.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <math.h>

/* Include polybench common header. */
#include <polybench.h>

/* Include benchmark-specific header. */
/* Default data type is double, default size is 1024. */
#include "lu.h"


/* Array initialization. */
static
void init_array (int n,
                 DATA_TYPE POLYBENCH_2D(A,N,N/THREADS,n,n/THREADS))
{
    int i, j, jj;

    for (i = 0; i < n; i++)
        for (j = 0; j < n/THREADS; j++) {
            jj = (j*THREADS)+MYTHREAD;
            A[i][j] = ((DATA_TYPE) (i+1)*(jj+1)) / n;
            //A[i][j] = ((DATA_TYPE) (3*i-5*jj+19)*(13*jj+11*i+17)) / 37;
        }
}


/* DCE code. Must scan the entire live-out data.
   Can be used also to check the correctness of the output. */
static
void print_array(int n,
                 DATA_TYPE POLYBENCH_2D(A,N,N/THREADS,n,n/THREADS))

{
    int i, j;

    char name[32];
    strcpy(name, "");
    sprintf(name+strlen(name), "out_%d", MYTHREAD);
    FILE *file = fopen(name, "w");
    for (i = 0; i < n; i++) {
        for (j = 0; j < (n/THREADS); j++) {
            fprintf (file, DATA_PRINTF_MODIFIER, A[i][j]);
            //if ((i * n + j) % 32 == 0) fprintf (out, "\n");
        }
        fprintf (file, "\n");
    }
    fprintf (file, "\n");
    fclose(file);
    upc_barrier;
}

#define __DECLARATION_OF_A DATA_TYPE POLYBENCH_2D(A,N,N,n,n)

#define BLOCKSIZE 1024

/* Main computational kernel. The whole function will be timed,
   including the call and return. */
static
void kernel_lu(int n,
               DATA_TYPE POLYBENCH_2D(A,N,N/THREADS,n,n/THREADS), shared [BLOCKSIZE] DATA_TYPE globalA[N], DATA_TYPE localA[N])
{
    int i, j, k, activeColumn, ownerOfColumn, startOfBlock;

    double t_local = 0.0, t_global = 0.0;

    upc_barrier;
    t_local = rtclock();

    activeColumn = 0;
    if (MYTHREAD == 0) {
        for(i = 0; i < _PB_N; i++)
            localA[i] = A[i][activeColumn];
        for(i = 0; i < _PB_N; i+=BLOCKSIZE)
            upc_memput(&globalA[i], &localA[i], BLOCKSIZE*sizeof(DATA_TYPE));
        activeColumn++;
    }
    upc_barrier;
    if (MYTHREAD != 0) {
        for(i = 0; i < _PB_N; i+=BLOCKSIZE)
            upc_memget(&localA[i], &globalA[i], BLOCKSIZE*sizeof(DATA_TYPE));
    }
    upc_barrier;
    for (k = 0; k < _PB_N; k++)
    {
        for (j = activeColumn; j < (_PB_N/THREADS); j++)
            A[k][j] = A[k][j] / localA[k];
        for(i = k + 1; i < _PB_N; i++)
            for (j = activeColumn; j < (_PB_N/THREADS); j++)
                A[i][j] = A[i][j] - localA[i] * A[k][j];

        startOfBlock = ceild((k+1),BLOCKSIZE)*BLOCKSIZE;
        ownerOfColumn = (k+1)%THREADS;
        if (ownerOfColumn == MYTHREAD) {
            for(i = k+1; i < _PB_N; i++)
                localA[i] = A[i][activeColumn];
            if (startOfBlock > (k+1))
                upc_memput(&globalA[k+1], &localA[k+1], (startOfBlock-k-1)*sizeof(DATA_TYPE));
            for(i = startOfBlock; i < _PB_N; i+=BLOCKSIZE)
                upc_memput(&globalA[i], &localA[i], BLOCKSIZE*sizeof(DATA_TYPE));
            activeColumn++;
        }

        upc_barrier;
        if (ownerOfColumn != MYTHREAD) {
            if (startOfBlock > (k+1))
                upc_memget(&localA[k+1], &globalA[k+1], (startOfBlock-k-1)*sizeof(DATA_TYPE));
            for(i = startOfBlock; i < _PB_N; i+=BLOCKSIZE)
                upc_memget(&localA[i], &globalA[i], BLOCKSIZE*sizeof(DATA_TYPE));
        }
        upc_barrier;

    }

    t_local = rtclock() - t_local;
    upc_barrier;

    double *t_localAll;
    t_localAll = (double *) malloc(THREADS*sizeof(double));
    t_global = bupc_allv_reduce(double, t_local, 0, UPC_MAX);
    t_localAll = bupc_allv_gather(double, t_local, 0, t_localAll);
    if (MYTHREAD==0) {
        char buffer[4096];
        int threadId;
        strcpy(buffer, "");
        for (threadId=0; threadId<THREADS; threadId++) {
            sprintf(buffer+strlen(buffer), "Node %d\n", threadId);
            sprintf(buffer+strlen(buffer), "Time: %lf\n", t_localAll[threadId]);
            sprintf(buffer+strlen(buffer), "-------\n");
        }
        sprintf(buffer+strlen(buffer), "SUMMARY\n");
        sprintf(buffer+strlen(buffer), "Maximum time spent across all nodes: %0.6lf s\n", t_global);
        sprintf(buffer+strlen(buffer), "-------");
        fprintf(stdout, "%s\n", buffer);
        fflush(stdout);
    }
    free(t_localAll);
    upc_barrier;
}

int main(int argc, char** argv)
{
    /* Retrieve problem size. */
    int n = N;

    /* Variable declaration/allocation. */
    POLYBENCH_2D_ARRAY_DECL(A, DATA_TYPE, N, N/THREADS, n, n/THREADS);

    shared [BLOCKSIZE] DATA_TYPE (*globalA)[N];
    globalA = upc_all_alloc(N/BLOCKSIZE, BLOCKSIZE*sizeof(DATA_TYPE));

    DATA_TYPE (*localA)[N];
    localA = malloc (N*sizeof(DATA_TYPE));

    /* Initialize array(s). */
    init_array (n, POLYBENCH_ARRAY(A));

    /* Start timer. */
    //polybench_start_instruments;

    /* Run kernel. */
    kernel_lu (n, POLYBENCH_ARRAY(A), *globalA, *localA);

    /* Stop and print timer. */
    //polybench_stop_instruments;
    //polybench_print_instruments;

    /* Prevent dead-code elimination. All live-out data must be printed
       by the function call in argument. */
    polybench_prevent_dce(print_array(n, POLYBENCH_ARRAY(A)));

    /* Be clean. */
    POLYBENCH_FREE_ARRAY(A);
    if (MYTHREAD == 0)
        upc_free(globalA);
    free(localA);

    return 0;
}
