#include "gemver.dist_dynsched_data_dist.h"
#include "polyrt.h"
#include "buffer_manager.h"
#include "pi_defs.h"
#include <assert.h>
#include <omp.h>
#include <mpi.h>
#include <limits.h>
#include <math.h>
#define ceild(n,d)  ceil(((double)(n))/((double)(d)))
#define floord(n,d) floor(((double)(n))/((double)(d)))
#define max(x,y)    ((x) > (y)? (x) : (y))
#define min(x,y)    ((x) < (y)? (x) : (y))

#include <stdio.h>
#include <unistd.h>
#include <sys/time.h>
#include <stdlib.h>

#define N 30000 
int n = N;

#define alpha 1
#define beta 1

#ifndef USE_LOCAL_ARRAYS
#pragma declarations
double A[N][N];
double B[N][N];

double x[N];
double u1[N];
double u2[N];
double v2[N];
double v1[N];
double w[N];
double y[N];
double z[N];
#pragma enddeclarations
#endif

#ifdef TIME
#define IF_TIME(foo) foo;
#else
#define IF_TIME(foo)
#endif

#define __PLACE_TO_INSERT_FORWARD_DECLARATIONS
FILE *__debug_print_fp;
tbb::concurrent_priority_queue<__Task, __Compare> pqueue;
int max_num_tasks_loop0_dim0;
int lb_tasks_loop0_dim0;
int max_num_tasks_loop0_dim1;
int lb_tasks_loop0_dim1;
int *tasks_loop0;
int max_num_tasks_loop1_dim0;
int lb_tasks_loop1_dim0;
int max_num_tasks_loop1_dim1;
int lb_tasks_loop1_dim1;
int *tasks_loop1;
int max_num_tasks_loop2_dim0;
int lb_tasks_loop2_dim0;
int *tasks_loop2;
int max_num_tasks_loop3_dim0;
int lb_tasks_loop3_dim0;
int max_num_tasks_loop3_dim1;
int lb_tasks_loop3_dim1;
int *tasks_loop3;

void init_array()
{
#ifndef USE_LOCAL_ARRAYS
    int i, j;

    for (i=0; i<N; i++) {
        u1[i] = i;
        u2[i] = (i+1)/N/2.0;
        v1[i] = (i+1)/N/4.0;
        v2[i] = (i+1)/N/6.0;
        y[i] = (i+1)/N/8.0;
        z[i] = (i+1)/N/9.0;
        x[i] = 0.0;
        w[i] = 0.0;
        for (j=0; j<N; j++) {
            A[i][j] = ((double) i*j)/N;
        }
    }
#endif
}

void print_array()
{
#ifndef USE_LOCAL_ARRAYS
    int i, j;

    for (i=0; i<N; i++) {
        fprintf(stderr, "%0.2lf ", w[i]);
        if (i%80 == 20) fprintf(stderr, "\n");
    }
    fprintf(stderr, "\n");
#endif
}


double rtclock()
{
    struct timezone Tzp;
    struct timeval Tp;
    int stat;
    stat = gettimeofday (&Tp, &Tzp);
    if (stat != 0) printf("Error return from gettimeofday: %d",stat);
    return(Tp.tv_sec + Tp.tv_usec*1.0e-6);
}

int main(int argc, char** argv)
{
   /* Retrieve problem size. */
  if(argc == 3){
    int np = atoi(argv[2]);
    int nprocs = atoi(argv[1]);
    n = np * pow(nprocs, 1.0/2.0);
  }
 
    double t_start, t_end;
    int i, j;

    init_array();

    IF_TIME(t_start = rtclock());

int i_j_0_div = 0;int i_j_0_mod = 0;double *B_i_j_0 = NULL;double *A_i_j_0 = NULL;int i_0_div = 0;int i_0_mod = 0;double *u1_i_0 = NULL;int j_0_div = 0;int j_0_mod = 0;double *v1_j_0 = NULL;double *u2_i_0 = NULL;double *v2_j_0 = NULL;int i_1_div = 0;int i_1_mod = 0;double *x_i_1 = NULL;int j_i_1_div = 0;int j_i_1_mod = 0;double *B_j_i_1 = NULL;int j_1_div = 0;int j_1_mod = 0;double *y_j_1 = NULL;int i_2_div = 0;int i_2_mod = 0;double *x_i_2 = NULL;double *z_i_2 = NULL;int i_3_div = 0;int i_3_mod = 0;double *w_i_3 = NULL;int i_j_3_div = 0;int i_j_3_mod = 0;double *B_i_j_3 = NULL;int j_3_div = 0;int j_3_mod = 0;double *x_j_3 = NULL; int t1, t2, t3, t4, t5, t6, t7, t8, t9, t10;
 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;
int _num_threads;
#pragma omp parallel
{_num_threads = omp_get_num_threads();}
 double t_data_init_start, t_data_init= 0.0, t_globaldata_init= 0.0;
#ifndef GLOBAL_MY_RANK
 int my_rank;
#endif
#ifndef __MAX_NUM_RECVS
 #define __MAX_NUM_RECVS (nprocs*_num_threads*4)
#endif
#ifndef __MAX_NUM_SENDS
 #define __MAX_NUM_SENDS (nprocs*_num_threads*4)
#endif
 int nprocs, my_start, my_end, _i, __p, proc, recv_proc, distinct_recv, __tid;
 long int _num_tasks_to_execute, _num_tasks_to_unpack;
 int count;
int req_count;
 int _lb_dist, _ub_dist, lbd_t1, ubd_t1, lbd_t2, ubd_t2, lbd_t3, ubd_t3, lbd_t4, ubd_t4, lbd_t5, ubd_t5, lbd_t6, ubd_t6, lbd_t7, ubd_t7, lbd_t8, ubd_t8, lbd_t9, ubd_t9, lbd_t10, ubd_t10;
 int thread_support_provided;
 MPI_Init_thread(NULL, NULL, MPI_THREAD_MULTIPLE, &thread_support_provided);
 assert(thread_support_provided == MPI_THREAD_MULTIPLE);
 MPI_Comm_rank(MPI_COMM_WORLD, &my_rank);
 MPI_Comm_size(MPI_COMM_WORLD, &nprocs);
 int __tid_receiver;
 int waiting_pops[_num_threads], waiting_pops_threads, waiting_pops_global;
 for (__tid=0; __tid<_num_threads; __tid++) waiting_pops[__tid] = 0;
 double t_tasks_create, t_globaltasks_create;
 double t_tasks_manage_start[_num_threads], t_tasks_manage[_num_threads], t_tasks_manage_threads, t_globaltasks_manage;
 for (__tid=0; __tid<_num_threads; __tid++) t_tasks_manage[__tid] = 0;
 double t_comp_start[_num_threads], t_comp[_num_threads], t_comp_threads_min, t_globalcomp_min, t_comp_threads, t_globalcomp;
 for (__tid=0; __tid<_num_threads; __tid++) t_comp[__tid] = 0;
 double t_wait_comp_start[_num_threads], t_wait_comp[_num_threads], t_wait_comp_threads, t_globalwait_comp;
 for (__tid=0; __tid<_num_threads; __tid++) t_wait_comp[__tid] = 0;
 double t_data_mang_start[_num_threads], t_data_mang[_num_threads], t_data_mang_threads, t_globaldata_mang;
 for (__tid=0; __tid<_num_threads; __tid++) t_data_mang[__tid] = 0;
 double t_pack_start[_num_threads], t_pack[_num_threads], t_pack_threads, t_globalpack;
 for (__tid=0; __tid<_num_threads; __tid++) t_pack[__tid] = 0;
 double t_writeout = 0.0;
 double t_unpack_start, t_unpack = 0.0, t_globalunpack = 0.0;
 double t_wait_unpack_start, t_wait_unpack = 0.0, t_globalwait_unpack = 0.0;
 double t_local = 0.0, t_global = 0.0;
 double __total_count[_num_threads];
 for (__tid=0; __tid<_num_threads; __tid++) __total_count[__tid] = 0;
 double __total_count_threads = 0;
 double __total_count_all = 0;
MPI_Request *send_reqs_B[_num_threads];
                    for (__tid=0; __tid<_num_threads; __tid++) {
                    send_reqs_B[__tid] = (MPI_Request *)malloc(__MAX_NUM_SENDS*sizeof(MPI_Request));
                    }
std::vector<MPI_Request> recv_reqs_B;
MPI_Status *send_stats_B[_num_threads];
                    for (__tid=0; __tid<_num_threads; __tid++) {
                    send_stats_B[__tid] = (MPI_Status *)malloc(nprocs*sizeof(MPI_Status));
                    }
MPI_Status *recv_stats_B = (MPI_Status *)malloc(sizeof(MPI_Status));
std::vector<int> recv_completed_B;
                    for (__tid=0; __tid<_num_threads; __tid++) {
                    for (__p=0; __p<__MAX_NUM_SENDS; __p++) {
                    send_reqs_B[__tid][__p] = MPI_REQUEST_NULL;
}
}
                    for (__p=0; __p<__MAX_NUM_RECVS; __p++) {
                    recv_reqs_B.push_back(MPI_REQUEST_NULL);
                    recv_completed_B.push_back(0);
                    }
int lw_count_B = 0;
int send_counts_B[__tid][nprocs];
int max_num_elements_B = 0;
int recv_counts_B[nprocs];
int lw_recv_counts_B[nprocs];
int displs_B[nprocs];
int displs_lw_B[nprocs];
double *send_buf_B[_num_threads][__MAX_NUM_SENDS];
size_t send_buf_size_B[_num_threads][__MAX_NUM_SENDS];
for (__tid=0; __tid<_num_threads; __tid++) {
      for (__p=0; __p<__MAX_NUM_SENDS; __p++) {
      send_buf_B[__tid][__p] = NULL;
      send_buf_size_B[__tid][__p] = 0;
}
}std::vector<double *> recv_buf_B;
size_t recv_buf_size_B[__MAX_NUM_RECVS];
for (__p=0; __p<__MAX_NUM_RECVS; __p++) {
                    recv_buf_B.push_back(NULL);
                    recv_buf_size_B[__p] = 0;
}double *lw_buf_B = NULL;
size_t lw_buf_size_B = 0;
double *lw_recv_buf_B = NULL;
size_t lw_recv_buf_size_B = 0;
MPI_Request *send_reqs_A[_num_threads];
                    for (__tid=0; __tid<_num_threads; __tid++) {
                    send_reqs_A[__tid] = (MPI_Request *)malloc(__MAX_NUM_SENDS*sizeof(MPI_Request));
                    }
std::vector<MPI_Request> recv_reqs_A;
MPI_Status *send_stats_A[_num_threads];
                    for (__tid=0; __tid<_num_threads; __tid++) {
                    send_stats_A[__tid] = (MPI_Status *)malloc(nprocs*sizeof(MPI_Status));
                    }
MPI_Status *recv_stats_A = (MPI_Status *)malloc(sizeof(MPI_Status));
std::vector<int> recv_completed_A;
                    for (__tid=0; __tid<_num_threads; __tid++) {
                    for (__p=0; __p<__MAX_NUM_SENDS; __p++) {
                    send_reqs_A[__tid][__p] = MPI_REQUEST_NULL;
}
}
                    for (__p=0; __p<__MAX_NUM_RECVS; __p++) {
                    recv_reqs_A.push_back(MPI_REQUEST_NULL);
                    recv_completed_A.push_back(0);
                    }
int lw_count_A = 0;
int send_counts_A[__tid][nprocs];
int max_num_elements_A = 0;
int recv_counts_A[nprocs];
int lw_recv_counts_A[nprocs];
int displs_A[nprocs];
int displs_lw_A[nprocs];
double *send_buf_A[_num_threads][__MAX_NUM_SENDS];
size_t send_buf_size_A[_num_threads][__MAX_NUM_SENDS];
for (__tid=0; __tid<_num_threads; __tid++) {
      for (__p=0; __p<__MAX_NUM_SENDS; __p++) {
      send_buf_A[__tid][__p] = NULL;
      send_buf_size_A[__tid][__p] = 0;
}
}std::vector<double *> recv_buf_A;
size_t recv_buf_size_A[__MAX_NUM_RECVS];
for (__p=0; __p<__MAX_NUM_RECVS; __p++) {
                    recv_buf_A.push_back(NULL);
                    recv_buf_size_A[__p] = 0;
}double *lw_buf_A = NULL;
size_t lw_buf_size_A = 0;
double *lw_recv_buf_A = NULL;
size_t lw_recv_buf_size_A = 0;
MPI_Request *send_reqs_u1[_num_threads];
                    for (__tid=0; __tid<_num_threads; __tid++) {
                    send_reqs_u1[__tid] = (MPI_Request *)malloc(__MAX_NUM_SENDS*sizeof(MPI_Request));
                    }
std::vector<MPI_Request> recv_reqs_u1;
MPI_Status *send_stats_u1[_num_threads];
                    for (__tid=0; __tid<_num_threads; __tid++) {
                    send_stats_u1[__tid] = (MPI_Status *)malloc(nprocs*sizeof(MPI_Status));
                    }
MPI_Status *recv_stats_u1 = (MPI_Status *)malloc(sizeof(MPI_Status));
std::vector<int> recv_completed_u1;
                    for (__tid=0; __tid<_num_threads; __tid++) {
                    for (__p=0; __p<__MAX_NUM_SENDS; __p++) {
                    send_reqs_u1[__tid][__p] = MPI_REQUEST_NULL;
}
}
                    for (__p=0; __p<__MAX_NUM_RECVS; __p++) {
                    recv_reqs_u1.push_back(MPI_REQUEST_NULL);
                    recv_completed_u1.push_back(0);
                    }
int lw_count_u1 = 0;
int send_counts_u1[__tid][nprocs];
int max_num_elements_u1 = 0;
int recv_counts_u1[nprocs];
int lw_recv_counts_u1[nprocs];
int displs_u1[nprocs];
int displs_lw_u1[nprocs];
double *send_buf_u1[_num_threads][__MAX_NUM_SENDS];
size_t send_buf_size_u1[_num_threads][__MAX_NUM_SENDS];
for (__tid=0; __tid<_num_threads; __tid++) {
      for (__p=0; __p<__MAX_NUM_SENDS; __p++) {
      send_buf_u1[__tid][__p] = NULL;
      send_buf_size_u1[__tid][__p] = 0;
}
}std::vector<double *> recv_buf_u1;
size_t recv_buf_size_u1[__MAX_NUM_RECVS];
for (__p=0; __p<__MAX_NUM_RECVS; __p++) {
                    recv_buf_u1.push_back(NULL);
                    recv_buf_size_u1[__p] = 0;
}double *lw_buf_u1 = NULL;
size_t lw_buf_size_u1 = 0;
double *lw_recv_buf_u1 = NULL;
size_t lw_recv_buf_size_u1 = 0;
MPI_Request *send_reqs_v1[_num_threads];
                    for (__tid=0; __tid<_num_threads; __tid++) {
                    send_reqs_v1[__tid] = (MPI_Request *)malloc(__MAX_NUM_SENDS*sizeof(MPI_Request));
                    }
std::vector<MPI_Request> recv_reqs_v1;
MPI_Status *send_stats_v1[_num_threads];
                    for (__tid=0; __tid<_num_threads; __tid++) {
                    send_stats_v1[__tid] = (MPI_Status *)malloc(nprocs*sizeof(MPI_Status));
                    }
MPI_Status *recv_stats_v1 = (MPI_Status *)malloc(sizeof(MPI_Status));
std::vector<int> recv_completed_v1;
                    for (__tid=0; __tid<_num_threads; __tid++) {
                    for (__p=0; __p<__MAX_NUM_SENDS; __p++) {
                    send_reqs_v1[__tid][__p] = MPI_REQUEST_NULL;
}
}
                    for (__p=0; __p<__MAX_NUM_RECVS; __p++) {
                    recv_reqs_v1.push_back(MPI_REQUEST_NULL);
                    recv_completed_v1.push_back(0);
                    }
int lw_count_v1 = 0;
int send_counts_v1[__tid][nprocs];
int max_num_elements_v1 = 0;
int recv_counts_v1[nprocs];
int lw_recv_counts_v1[nprocs];
int displs_v1[nprocs];
int displs_lw_v1[nprocs];
double *send_buf_v1[_num_threads][__MAX_NUM_SENDS];
size_t send_buf_size_v1[_num_threads][__MAX_NUM_SENDS];
for (__tid=0; __tid<_num_threads; __tid++) {
      for (__p=0; __p<__MAX_NUM_SENDS; __p++) {
      send_buf_v1[__tid][__p] = NULL;
      send_buf_size_v1[__tid][__p] = 0;
}
}std::vector<double *> recv_buf_v1;
size_t recv_buf_size_v1[__MAX_NUM_RECVS];
for (__p=0; __p<__MAX_NUM_RECVS; __p++) {
                    recv_buf_v1.push_back(NULL);
                    recv_buf_size_v1[__p] = 0;
}double *lw_buf_v1 = NULL;
size_t lw_buf_size_v1 = 0;
double *lw_recv_buf_v1 = NULL;
size_t lw_recv_buf_size_v1 = 0;
MPI_Request *send_reqs_u2[_num_threads];
                    for (__tid=0; __tid<_num_threads; __tid++) {
                    send_reqs_u2[__tid] = (MPI_Request *)malloc(__MAX_NUM_SENDS*sizeof(MPI_Request));
                    }
std::vector<MPI_Request> recv_reqs_u2;
MPI_Status *send_stats_u2[_num_threads];
                    for (__tid=0; __tid<_num_threads; __tid++) {
                    send_stats_u2[__tid] = (MPI_Status *)malloc(nprocs*sizeof(MPI_Status));
                    }
MPI_Status *recv_stats_u2 = (MPI_Status *)malloc(sizeof(MPI_Status));
std::vector<int> recv_completed_u2;
                    for (__tid=0; __tid<_num_threads; __tid++) {
                    for (__p=0; __p<__MAX_NUM_SENDS; __p++) {
                    send_reqs_u2[__tid][__p] = MPI_REQUEST_NULL;
}
}
                    for (__p=0; __p<__MAX_NUM_RECVS; __p++) {
                    recv_reqs_u2.push_back(MPI_REQUEST_NULL);
                    recv_completed_u2.push_back(0);
                    }
int lw_count_u2 = 0;
int send_counts_u2[__tid][nprocs];
int max_num_elements_u2 = 0;
int recv_counts_u2[nprocs];
int lw_recv_counts_u2[nprocs];
int displs_u2[nprocs];
int displs_lw_u2[nprocs];
double *send_buf_u2[_num_threads][__MAX_NUM_SENDS];
size_t send_buf_size_u2[_num_threads][__MAX_NUM_SENDS];
for (__tid=0; __tid<_num_threads; __tid++) {
      for (__p=0; __p<__MAX_NUM_SENDS; __p++) {
      send_buf_u2[__tid][__p] = NULL;
      send_buf_size_u2[__tid][__p] = 0;
}
}std::vector<double *> recv_buf_u2;
size_t recv_buf_size_u2[__MAX_NUM_RECVS];
for (__p=0; __p<__MAX_NUM_RECVS; __p++) {
                    recv_buf_u2.push_back(NULL);
                    recv_buf_size_u2[__p] = 0;
}double *lw_buf_u2 = NULL;
size_t lw_buf_size_u2 = 0;
double *lw_recv_buf_u2 = NULL;
size_t lw_recv_buf_size_u2 = 0;
MPI_Request *send_reqs_v2[_num_threads];
                    for (__tid=0; __tid<_num_threads; __tid++) {
                    send_reqs_v2[__tid] = (MPI_Request *)malloc(__MAX_NUM_SENDS*sizeof(MPI_Request));
                    }
std::vector<MPI_Request> recv_reqs_v2;
MPI_Status *send_stats_v2[_num_threads];
                    for (__tid=0; __tid<_num_threads; __tid++) {
                    send_stats_v2[__tid] = (MPI_Status *)malloc(nprocs*sizeof(MPI_Status));
                    }
MPI_Status *recv_stats_v2 = (MPI_Status *)malloc(sizeof(MPI_Status));
std::vector<int> recv_completed_v2;
                    for (__tid=0; __tid<_num_threads; __tid++) {
                    for (__p=0; __p<__MAX_NUM_SENDS; __p++) {
                    send_reqs_v2[__tid][__p] = MPI_REQUEST_NULL;
}
}
                    for (__p=0; __p<__MAX_NUM_RECVS; __p++) {
                    recv_reqs_v2.push_back(MPI_REQUEST_NULL);
                    recv_completed_v2.push_back(0);
                    }
int lw_count_v2 = 0;
int send_counts_v2[__tid][nprocs];
int max_num_elements_v2 = 0;
int recv_counts_v2[nprocs];
int lw_recv_counts_v2[nprocs];
int displs_v2[nprocs];
int displs_lw_v2[nprocs];
double *send_buf_v2[_num_threads][__MAX_NUM_SENDS];
size_t send_buf_size_v2[_num_threads][__MAX_NUM_SENDS];
for (__tid=0; __tid<_num_threads; __tid++) {
      for (__p=0; __p<__MAX_NUM_SENDS; __p++) {
      send_buf_v2[__tid][__p] = NULL;
      send_buf_size_v2[__tid][__p] = 0;
}
}std::vector<double *> recv_buf_v2;
size_t recv_buf_size_v2[__MAX_NUM_RECVS];
for (__p=0; __p<__MAX_NUM_RECVS; __p++) {
                    recv_buf_v2.push_back(NULL);
                    recv_buf_size_v2[__p] = 0;
}double *lw_buf_v2 = NULL;
size_t lw_buf_size_v2 = 0;
double *lw_recv_buf_v2 = NULL;
size_t lw_recv_buf_size_v2 = 0;
MPI_Request *send_reqs_x[_num_threads];
                    for (__tid=0; __tid<_num_threads; __tid++) {
                    send_reqs_x[__tid] = (MPI_Request *)malloc(__MAX_NUM_SENDS*sizeof(MPI_Request));
                    }
std::vector<MPI_Request> recv_reqs_x;
MPI_Status *send_stats_x[_num_threads];
                    for (__tid=0; __tid<_num_threads; __tid++) {
                    send_stats_x[__tid] = (MPI_Status *)malloc(nprocs*sizeof(MPI_Status));
                    }
MPI_Status *recv_stats_x = (MPI_Status *)malloc(sizeof(MPI_Status));
std::vector<int> recv_completed_x;
                    for (__tid=0; __tid<_num_threads; __tid++) {
                    for (__p=0; __p<__MAX_NUM_SENDS; __p++) {
                    send_reqs_x[__tid][__p] = MPI_REQUEST_NULL;
}
}
                    for (__p=0; __p<__MAX_NUM_RECVS; __p++) {
                    recv_reqs_x.push_back(MPI_REQUEST_NULL);
                    recv_completed_x.push_back(0);
                    }
int lw_count_x = 0;
int send_counts_x[__tid][nprocs];
int max_num_elements_x = 0;
int recv_counts_x[nprocs];
int lw_recv_counts_x[nprocs];
int displs_x[nprocs];
int displs_lw_x[nprocs];
double *send_buf_x[_num_threads][__MAX_NUM_SENDS];
size_t send_buf_size_x[_num_threads][__MAX_NUM_SENDS];
for (__tid=0; __tid<_num_threads; __tid++) {
      for (__p=0; __p<__MAX_NUM_SENDS; __p++) {
      send_buf_x[__tid][__p] = NULL;
      send_buf_size_x[__tid][__p] = 0;
}
}std::vector<double *> recv_buf_x;
size_t recv_buf_size_x[__MAX_NUM_RECVS];
for (__p=0; __p<__MAX_NUM_RECVS; __p++) {
                    recv_buf_x.push_back(NULL);
                    recv_buf_size_x[__p] = 0;
}double *lw_buf_x = NULL;
size_t lw_buf_size_x = 0;
double *lw_recv_buf_x = NULL;
size_t lw_recv_buf_size_x = 0;
MPI_Request *send_reqs_y[_num_threads];
                    for (__tid=0; __tid<_num_threads; __tid++) {
                    send_reqs_y[__tid] = (MPI_Request *)malloc(__MAX_NUM_SENDS*sizeof(MPI_Request));
                    }
std::vector<MPI_Request> recv_reqs_y;
MPI_Status *send_stats_y[_num_threads];
                    for (__tid=0; __tid<_num_threads; __tid++) {
                    send_stats_y[__tid] = (MPI_Status *)malloc(nprocs*sizeof(MPI_Status));
                    }
MPI_Status *recv_stats_y = (MPI_Status *)malloc(sizeof(MPI_Status));
std::vector<int> recv_completed_y;
                    for (__tid=0; __tid<_num_threads; __tid++) {
                    for (__p=0; __p<__MAX_NUM_SENDS; __p++) {
                    send_reqs_y[__tid][__p] = MPI_REQUEST_NULL;
}
}
                    for (__p=0; __p<__MAX_NUM_RECVS; __p++) {
                    recv_reqs_y.push_back(MPI_REQUEST_NULL);
                    recv_completed_y.push_back(0);
                    }
int lw_count_y = 0;
int send_counts_y[__tid][nprocs];
int max_num_elements_y = 0;
int recv_counts_y[nprocs];
int lw_recv_counts_y[nprocs];
int displs_y[nprocs];
int displs_lw_y[nprocs];
double *send_buf_y[_num_threads][__MAX_NUM_SENDS];
size_t send_buf_size_y[_num_threads][__MAX_NUM_SENDS];
for (__tid=0; __tid<_num_threads; __tid++) {
      for (__p=0; __p<__MAX_NUM_SENDS; __p++) {
      send_buf_y[__tid][__p] = NULL;
      send_buf_size_y[__tid][__p] = 0;
}
}std::vector<double *> recv_buf_y;
size_t recv_buf_size_y[__MAX_NUM_RECVS];
for (__p=0; __p<__MAX_NUM_RECVS; __p++) {
                    recv_buf_y.push_back(NULL);
                    recv_buf_size_y[__p] = 0;
}double *lw_buf_y = NULL;
size_t lw_buf_size_y = 0;
double *lw_recv_buf_y = NULL;
size_t lw_recv_buf_size_y = 0;
MPI_Request *send_reqs_z[_num_threads];
                    for (__tid=0; __tid<_num_threads; __tid++) {
                    send_reqs_z[__tid] = (MPI_Request *)malloc(__MAX_NUM_SENDS*sizeof(MPI_Request));
                    }
std::vector<MPI_Request> recv_reqs_z;
MPI_Status *send_stats_z[_num_threads];
                    for (__tid=0; __tid<_num_threads; __tid++) {
                    send_stats_z[__tid] = (MPI_Status *)malloc(nprocs*sizeof(MPI_Status));
                    }
MPI_Status *recv_stats_z = (MPI_Status *)malloc(sizeof(MPI_Status));
std::vector<int> recv_completed_z;
                    for (__tid=0; __tid<_num_threads; __tid++) {
                    for (__p=0; __p<__MAX_NUM_SENDS; __p++) {
                    send_reqs_z[__tid][__p] = MPI_REQUEST_NULL;
}
}
                    for (__p=0; __p<__MAX_NUM_RECVS; __p++) {
                    recv_reqs_z.push_back(MPI_REQUEST_NULL);
                    recv_completed_z.push_back(0);
                    }
int lw_count_z = 0;
int send_counts_z[__tid][nprocs];
int max_num_elements_z = 0;
int recv_counts_z[nprocs];
int lw_recv_counts_z[nprocs];
int displs_z[nprocs];
int displs_lw_z[nprocs];
double *send_buf_z[_num_threads][__MAX_NUM_SENDS];
size_t send_buf_size_z[_num_threads][__MAX_NUM_SENDS];
for (__tid=0; __tid<_num_threads; __tid++) {
      for (__p=0; __p<__MAX_NUM_SENDS; __p++) {
      send_buf_z[__tid][__p] = NULL;
      send_buf_size_z[__tid][__p] = 0;
}
}std::vector<double *> recv_buf_z;
size_t recv_buf_size_z[__MAX_NUM_RECVS];
for (__p=0; __p<__MAX_NUM_RECVS; __p++) {
                    recv_buf_z.push_back(NULL);
                    recv_buf_size_z[__p] = 0;
}double *lw_buf_z = NULL;
size_t lw_buf_size_z = 0;
double *lw_recv_buf_z = NULL;
size_t lw_recv_buf_size_z = 0;
MPI_Request *send_reqs_w[_num_threads];
                    for (__tid=0; __tid<_num_threads; __tid++) {
                    send_reqs_w[__tid] = (MPI_Request *)malloc(__MAX_NUM_SENDS*sizeof(MPI_Request));
                    }
std::vector<MPI_Request> recv_reqs_w;
MPI_Status *send_stats_w[_num_threads];
                    for (__tid=0; __tid<_num_threads; __tid++) {
                    send_stats_w[__tid] = (MPI_Status *)malloc(nprocs*sizeof(MPI_Status));
                    }
MPI_Status *recv_stats_w = (MPI_Status *)malloc(sizeof(MPI_Status));
std::vector<int> recv_completed_w;
                    for (__tid=0; __tid<_num_threads; __tid++) {
                    for (__p=0; __p<__MAX_NUM_SENDS; __p++) {
                    send_reqs_w[__tid][__p] = MPI_REQUEST_NULL;
}
}
                    for (__p=0; __p<__MAX_NUM_RECVS; __p++) {
                    recv_reqs_w.push_back(MPI_REQUEST_NULL);
                    recv_completed_w.push_back(0);
                    }
int lw_count_w = 0;
int send_counts_w[__tid][nprocs];
int max_num_elements_w = 0;
int recv_counts_w[nprocs];
int lw_recv_counts_w[nprocs];
int displs_w[nprocs];
int displs_lw_w[nprocs];
double *send_buf_w[_num_threads][__MAX_NUM_SENDS];
size_t send_buf_size_w[_num_threads][__MAX_NUM_SENDS];
for (__tid=0; __tid<_num_threads; __tid++) {
      for (__p=0; __p<__MAX_NUM_SENDS; __p++) {
      send_buf_w[__tid][__p] = NULL;
      send_buf_size_w[__tid][__p] = 0;
}
}std::vector<double *> recv_buf_w;
size_t recv_buf_size_w[__MAX_NUM_RECVS];
for (__p=0; __p<__MAX_NUM_RECVS; __p++) {
                    recv_buf_w.push_back(NULL);
                    recv_buf_size_w[__p] = 0;
}double *lw_buf_w = NULL;
size_t lw_buf_size_w = 0;
double *lw_recv_buf_w = NULL;
size_t lw_recv_buf_size_w = 0;
#ifdef __USE_BLOCK_CYCLIC
__is_block_cyclic[0] = 1;
#endif
__is_multi_partitioned[1] = 1;
#ifdef __USE_BLOCK_CYCLIC
__is_block_cyclic[2] = 1;
#endif
#ifdef __USE_BLOCK_CYCLIC
__is_block_cyclic[3] = 1;
#endif
max_num_tasks_loop0_dim0 = ceil(floorf((+1*n-1)/(float)128) - 0 + 1);
lb_tasks_loop0_dim0 = floor(0);
max_num_tasks_loop0_dim1 = ceil(floorf((+1*n-1)/(float)128) - 0 + 1);
lb_tasks_loop0_dim1 = floor(0);
tasks_loop0 = (int *)malloc(max_num_tasks_loop0_dim0*max_num_tasks_loop0_dim1*sizeof(int));
max_num_tasks_loop1_dim0 = ceil(floorf((+1*n-1)/(float)128) - 0 + 1);
lb_tasks_loop1_dim0 = floor(0);
max_num_tasks_loop1_dim1 = ceil(floorf((+1*n-1)/(float)128) - 0 + 1);
lb_tasks_loop1_dim1 = floor(0);
tasks_loop1 = (int *)malloc(max_num_tasks_loop1_dim0*max_num_tasks_loop1_dim1*sizeof(int));
max_num_tasks_loop2_dim0 = ceil(floorf((+1*n-1)/(float)128) - 0 + 1);
lb_tasks_loop2_dim0 = floor(0);
tasks_loop2 = (int *)malloc(max_num_tasks_loop2_dim0*sizeof(int));
max_num_tasks_loop3_dim0 = ceil(floorf((+1*n-1)/(float)128) - 0 + 1);
lb_tasks_loop3_dim0 = floor(0);
max_num_tasks_loop3_dim1 = ceil(floorf((+1*n-1)/(float)128) - 0 + 1);
lb_tasks_loop3_dim1 = floor(0);
tasks_loop3 = (int *)malloc(max_num_tasks_loop3_dim0*max_num_tasks_loop3_dim1*sizeof(int));
int w_size_0 = (+1*(floorf((+1*n-1)/(float)128) - 0 + 1));
double **w_trans = (double **)malloc((w_size_0) * sizeof(double));
BufferManager *buff_mang_w = init_buffer_manager(w_size_0);
int z_size_0 = (+1*(floorf((+1*n-1)/(float)128) - 0 + 1));
double **z_trans = (double **)malloc((z_size_0) * sizeof(double));
BufferManager *buff_mang_z = init_buffer_manager(z_size_0);
int y_size_0 = (+1*(floorf((+1*n-1)/(float)128) - 0 + 1));
double **y_trans = (double **)malloc((y_size_0) * sizeof(double));
BufferManager *buff_mang_y = init_buffer_manager(y_size_0);
int x_size_0 = (+1*(floorf((+1*n-1)/(float)128) - 0 + 1));
double **x_trans = (double **)malloc((x_size_0) * sizeof(double));
BufferManager *buff_mang_x = init_buffer_manager(x_size_0);
int v2_size_0 = (+1*(floorf((+1*n-1)/(float)128) - 0 + 1));
double **v2_trans = (double **)malloc((v2_size_0) * sizeof(double));
BufferManager *buff_mang_v2 = init_buffer_manager(v2_size_0);
int u2_size_0 = (+1*(floorf((+1*n-1)/(float)128) - 0 + 1));
double **u2_trans = (double **)malloc((u2_size_0) * sizeof(double));
BufferManager *buff_mang_u2 = init_buffer_manager(u2_size_0);
int v1_size_0 = (+1*(floorf((+1*n-1)/(float)128) - 0 + 1));
double **v1_trans = (double **)malloc((v1_size_0) * sizeof(double));
BufferManager *buff_mang_v1 = init_buffer_manager(v1_size_0);
int u1_size_0 = (+1*(floorf((+1*n-1)/(float)128) - 0 + 1));
double **u1_trans = (double **)malloc((u1_size_0) * sizeof(double));
BufferManager *buff_mang_u1 = init_buffer_manager(u1_size_0);
int A_size_0 = (+1*(floorf((+1*n-1)/(float)128) - 0 + 1));
int A_size_1 = (+1*(floorf((+1*n-1)/(float)128) - 0 + 1));
double **A_trans = (double **)malloc((A_size_0 * A_size_1) * sizeof(double));
BufferManager *buff_mang_A = init_buffer_manager(A_size_0 * A_size_1);
int B_size_0 = (+1*(floorf((+1*n-1)/(float)128) - 0 + 1));
int B_size_1 = (+1*(floorf((+1*n-1)/(float)128) - 0 + 1));
double **B_trans = (double **)malloc((B_size_0 * B_size_1) * sizeof(double));
BufferManager *buff_mang_B = init_buffer_manager(B_size_0 * B_size_1);
            max_num_elements_A = max(max_num_elements_A, (+1*(1)*(1)) + 4);
                for (__tid=0; __tid<_num_threads; __tid++) { for (__p=0; __p<__MAX_NUM_SENDS; __p++) {
                send_buf_A[__tid][__p] = (double *) polyrt_max_alloc(send_buf_A[__tid][__p], sizeof(double)*max_num_elements_A, &send_buf_size_A[__tid][__p]);
}
}
for (__p=0; __p<__MAX_NUM_RECVS; __p++) { recv_buf_A[__p] = (double *) polyrt_max_alloc(recv_buf_A[__p], sizeof(double)*max_num_elements_A, &recv_buf_size_A[__p]); }
            max_num_elements_u1 = max(max_num_elements_u1, (+1*(1)) + 4);
                for (__tid=0; __tid<_num_threads; __tid++) { for (__p=0; __p<__MAX_NUM_SENDS; __p++) {
                send_buf_u1[__tid][__p] = (double *) polyrt_max_alloc(send_buf_u1[__tid][__p], sizeof(double)*max_num_elements_u1, &send_buf_size_u1[__tid][__p]);
}
}
for (__p=0; __p<__MAX_NUM_RECVS; __p++) { recv_buf_u1[__p] = (double *) polyrt_max_alloc(recv_buf_u1[__p], sizeof(double)*max_num_elements_u1, &recv_buf_size_u1[__p]); }
            max_num_elements_v1 = max(max_num_elements_v1, (+1*(1)) + 4);
                for (__tid=0; __tid<_num_threads; __tid++) { for (__p=0; __p<__MAX_NUM_SENDS; __p++) {
                send_buf_v1[__tid][__p] = (double *) polyrt_max_alloc(send_buf_v1[__tid][__p], sizeof(double)*max_num_elements_v1, &send_buf_size_v1[__tid][__p]);
}
}
for (__p=0; __p<__MAX_NUM_RECVS; __p++) { recv_buf_v1[__p] = (double *) polyrt_max_alloc(recv_buf_v1[__p], sizeof(double)*max_num_elements_v1, &recv_buf_size_v1[__p]); }
            max_num_elements_u2 = max(max_num_elements_u2, (+1*(1)) + 4);
                for (__tid=0; __tid<_num_threads; __tid++) { for (__p=0; __p<__MAX_NUM_SENDS; __p++) {
                send_buf_u2[__tid][__p] = (double *) polyrt_max_alloc(send_buf_u2[__tid][__p], sizeof(double)*max_num_elements_u2, &send_buf_size_u2[__tid][__p]);
}
}
for (__p=0; __p<__MAX_NUM_RECVS; __p++) { recv_buf_u2[__p] = (double *) polyrt_max_alloc(recv_buf_u2[__p], sizeof(double)*max_num_elements_u2, &recv_buf_size_u2[__p]); }
            max_num_elements_v2 = max(max_num_elements_v2, (+1*(1)) + 4);
                for (__tid=0; __tid<_num_threads; __tid++) { for (__p=0; __p<__MAX_NUM_SENDS; __p++) {
                send_buf_v2[__tid][__p] = (double *) polyrt_max_alloc(send_buf_v2[__tid][__p], sizeof(double)*max_num_elements_v2, &send_buf_size_v2[__tid][__p]);
}
}
for (__p=0; __p<__MAX_NUM_RECVS; __p++) { recv_buf_v2[__p] = (double *) polyrt_max_alloc(recv_buf_v2[__p], sizeof(double)*max_num_elements_v2, &recv_buf_size_v2[__p]); }
            max_num_elements_B = max(max_num_elements_B, (+1*(128)*(128)) + 4);
                for (__tid=0; __tid<_num_threads; __tid++) { for (__p=0; __p<__MAX_NUM_SENDS; __p++) {
                send_buf_B[__tid][__p] = (double *) polyrt_max_alloc(send_buf_B[__tid][__p], sizeof(double)*max_num_elements_B, &send_buf_size_B[__tid][__p]);
}
}
for (__p=0; __p<__MAX_NUM_RECVS; __p++) { recv_buf_B[__p] = (double *) polyrt_max_alloc(recv_buf_B[__p], sizeof(double)*max_num_elements_B, &recv_buf_size_B[__p]); }
if (fopen(".test", "r")) {lw_buf_B = (double *) polyrt_max_alloc(lw_buf_B, sizeof(double)*((+1*(128)*(128))*floorf((1*(floorf((+1*n-1)/(float)128) - 0 + 1+nprocs)*(floorf((+1*n-1)/(float)128) - 0 + 1+nprocs))/(float)nprocs)), &lw_buf_size_B);
lw_recv_buf_B = (double *) polyrt_max_alloc(lw_recv_buf_B, nprocs*lw_buf_size_B, &lw_recv_buf_size_B);
 for (__p=0; __p<nprocs; __p++) { displs_lw_B[__p] = __p*lw_buf_size_B/sizeof(double);}
} max_num_elements_x = max(max_num_elements_x, (+1*(128)+1*(128)) + 4);
                for (__tid=0; __tid<_num_threads; __tid++) { for (__p=0; __p<__MAX_NUM_SENDS; __p++) {
                send_buf_x[__tid][__p] = (double *) polyrt_max_alloc(send_buf_x[__tid][__p], sizeof(double)*max_num_elements_x, &send_buf_size_x[__tid][__p]);
}
}
for (__p=0; __p<__MAX_NUM_RECVS; __p++) { recv_buf_x[__p] = (double *) polyrt_max_alloc(recv_buf_x[__p], sizeof(double)*max_num_elements_x, &recv_buf_size_x[__p]); }
            max_num_elements_B = max(max_num_elements_B, (+1*(1)*(1)) + 4);
                for (__tid=0; __tid<_num_threads; __tid++) { for (__p=0; __p<__MAX_NUM_SENDS; __p++) {
                send_buf_B[__tid][__p] = (double *) polyrt_max_alloc(send_buf_B[__tid][__p], sizeof(double)*max_num_elements_B, &send_buf_size_B[__tid][__p]);
}
}
for (__p=0; __p<__MAX_NUM_RECVS; __p++) { recv_buf_B[__p] = (double *) polyrt_max_alloc(recv_buf_B[__p], sizeof(double)*max_num_elements_B, &recv_buf_size_B[__p]); }
            max_num_elements_y = max(max_num_elements_y, (+1*(1)) + 4);
                for (__tid=0; __tid<_num_threads; __tid++) { for (__p=0; __p<__MAX_NUM_SENDS; __p++) {
                send_buf_y[__tid][__p] = (double *) polyrt_max_alloc(send_buf_y[__tid][__p], sizeof(double)*max_num_elements_y, &send_buf_size_y[__tid][__p]);
}
}
for (__p=0; __p<__MAX_NUM_RECVS; __p++) { recv_buf_y[__p] = (double *) polyrt_max_alloc(recv_buf_y[__p], sizeof(double)*max_num_elements_y, &recv_buf_size_y[__p]); }
if (fopen(".test", "r")) {lw_buf_x = (double *) polyrt_max_alloc(lw_buf_x, sizeof(double)*((+1*(1))*floorf((1*(floorf((+1*n-1)/(float)128) - 0 + 1+nprocs)*(floorf((+1*n-1)/(float)128) - 0 + 1+nprocs))/(float)nprocs)), &lw_buf_size_x);
lw_recv_buf_x = (double *) polyrt_max_alloc(lw_recv_buf_x, nprocs*lw_buf_size_x, &lw_recv_buf_size_x);
 for (__p=0; __p<nprocs; __p++) { displs_lw_x[__p] = __p*lw_buf_size_x/sizeof(double);}
} max_num_elements_x = max(max_num_elements_x, (+1*(128)) + 4);
                for (__tid=0; __tid<_num_threads; __tid++) { for (__p=0; __p<__MAX_NUM_SENDS; __p++) {
                send_buf_x[__tid][__p] = (double *) polyrt_max_alloc(send_buf_x[__tid][__p], sizeof(double)*max_num_elements_x, &send_buf_size_x[__tid][__p]);
}
}
for (__p=0; __p<__MAX_NUM_RECVS; __p++) { recv_buf_x[__p] = (double *) polyrt_max_alloc(recv_buf_x[__p], sizeof(double)*max_num_elements_x, &recv_buf_size_x[__p]); }
            max_num_elements_z = max(max_num_elements_z, (+1*(1)) + 4);
                for (__tid=0; __tid<_num_threads; __tid++) { for (__p=0; __p<__MAX_NUM_SENDS; __p++) {
                send_buf_z[__tid][__p] = (double *) polyrt_max_alloc(send_buf_z[__tid][__p], sizeof(double)*max_num_elements_z, &send_buf_size_z[__tid][__p]);
}
}
for (__p=0; __p<__MAX_NUM_RECVS; __p++) { recv_buf_z[__p] = (double *) polyrt_max_alloc(recv_buf_z[__p], sizeof(double)*max_num_elements_z, &recv_buf_size_z[__p]); }
if (fopen(".test", "r")) {lw_buf_x = (double *) polyrt_max_alloc(lw_buf_x, sizeof(double)*((+1*(128))*floorf((1*(floorf((+1*n-1)/(float)128) - 0 + 1+nprocs))/(float)nprocs)), &lw_buf_size_x);
lw_recv_buf_x = (double *) polyrt_max_alloc(lw_recv_buf_x, nprocs*lw_buf_size_x, &lw_recv_buf_size_x);
 for (__p=0; __p<nprocs; __p++) { displs_lw_x[__p] = __p*lw_buf_size_x/sizeof(double);}
} max_num_elements_w = max(max_num_elements_w, (+1*(128)) + 4);
                for (__tid=0; __tid<_num_threads; __tid++) { for (__p=0; __p<__MAX_NUM_SENDS; __p++) {
                send_buf_w[__tid][__p] = (double *) polyrt_max_alloc(send_buf_w[__tid][__p], sizeof(double)*max_num_elements_w, &send_buf_size_w[__tid][__p]);
}
}
for (__p=0; __p<__MAX_NUM_RECVS; __p++) { recv_buf_w[__p] = (double *) polyrt_max_alloc(recv_buf_w[__p], sizeof(double)*max_num_elements_w, &recv_buf_size_w[__p]); }
            max_num_elements_B = max(max_num_elements_B, (+1*(1)*(1)) + 4);
                for (__tid=0; __tid<_num_threads; __tid++) { for (__p=0; __p<__MAX_NUM_SENDS; __p++) {
                send_buf_B[__tid][__p] = (double *) polyrt_max_alloc(send_buf_B[__tid][__p], sizeof(double)*max_num_elements_B, &send_buf_size_B[__tid][__p]);
}
}
for (__p=0; __p<__MAX_NUM_RECVS; __p++) { recv_buf_B[__p] = (double *) polyrt_max_alloc(recv_buf_B[__p], sizeof(double)*max_num_elements_B, &recv_buf_size_B[__p]); }
            max_num_elements_x = max(max_num_elements_x, (+1*(1)) + 4);
                for (__tid=0; __tid<_num_threads; __tid++) { for (__p=0; __p<__MAX_NUM_SENDS; __p++) {
                send_buf_x[__tid][__p] = (double *) polyrt_max_alloc(send_buf_x[__tid][__p], sizeof(double)*max_num_elements_x, &send_buf_size_x[__tid][__p]);
}
}
for (__p=0; __p<__MAX_NUM_RECVS; __p++) { recv_buf_x[__p] = (double *) polyrt_max_alloc(recv_buf_x[__p], sizeof(double)*max_num_elements_x, &recv_buf_size_x[__p]); }
if (fopen(".test", "r")) {lw_buf_w = (double *) polyrt_max_alloc(lw_buf_w, sizeof(double)*((+1*(128))*floorf((1*(floorf((+1*n-1)/(float)128) - 0 + 1+nprocs))/(float)nprocs)), &lw_buf_size_w);
lw_recv_buf_w = (double *) polyrt_max_alloc(lw_recv_buf_w, nprocs*lw_buf_size_w, &lw_recv_buf_size_w);
 for (__p=0; __p<nprocs; __p++) { displs_lw_w[__p] = __p*lw_buf_size_w/sizeof(double);}
} polyrt_init_grid_size();
 IF_TIME(t_data_init_start = rtclock());
data_tile_init_0(my_rank, nprocs, B_trans, B_size_1, buff_mang_B, A_trans, A_size_1, buff_mang_A, u1_trans, buff_mang_u1, v1_trans, buff_mang_v1, u2_trans, buff_mang_u2, v2_trans, buff_mang_v2, x_trans, buff_mang_x, y_trans, buff_mang_y, z_trans, buff_mang_z, w_trans, buff_mang_w);
 IF_TIME(t_data_init += rtclock() - t_data_init_start)
 IF_TIME(t_local = rtclock());
/*Start of dynamic scheduling code*/
#if defined(__DYNSCHEDULER_DEBUG_PRINT) || defined(__DYNSCHEDULER_MORE_DEBUG_PRINT) || defined(__DEBUG_FLUSH)
 char *debug_print_name = (char *)malloc(128*sizeof(char));
 strcpy(debug_print_name, "debug_print");
 sprintf(debug_print_name+strlen(debug_print_name), "_node%d", my_rank);
 __debug_print_fp = fopen(debug_print_name, "w");
 free(debug_print_name);
#endif
 IF_TIME(t_tasks_create = rtclock());
 init_tasks(my_rank,nprocs,_num_threads,tasks_loop0,tasks_loop1,tasks_loop2,tasks_loop3,&_num_tasks_to_execute,&_num_tasks_to_unpack);
 IF_TIME(t_tasks_create = rtclock() - t_tasks_create);
#ifdef __DYNSCHEDULER_MORE_DEBUG_PRINT
 std::vector<int> prev_firing_count_B_0;
 std::vector<int> prev_firing_count_B_1;
 std::vector<int> prev_firing_count_B_2;
 std::vector<int> prev_firing_count_B_3;
 std::vector<int> prev_firing_count_x_0;
 std::vector<int> prev_firing_count_x_1;
 std::vector<int> prev_firing_count_x_2;
 std::vector<int> prev_firing_count_x_3;
 std::vector<int> prev_firing_count_w_0;
 std::vector<int> prev_firing_count_w_1;
 std::vector<int> prev_firing_count_w_2;
 std::vector<int> prev_firing_count_w_3;
 for (__p=0; __p<__MAX_NUM_RECVS; __p++) {
  prev_firing_count_B_0.push_back(-1);
  prev_firing_count_B_1.push_back(-1);
  prev_firing_count_B_2.push_back(-1);
  prev_firing_count_B_3.push_back(-1);
  prev_firing_count_x_0.push_back(-1);
  prev_firing_count_x_1.push_back(-1);
  prev_firing_count_x_2.push_back(-1);
  prev_firing_count_x_3.push_back(-1);
  prev_firing_count_w_0.push_back(-1);
  prev_firing_count_w_1.push_back(-1);
  prev_firing_count_w_2.push_back(-1);
  prev_firing_count_w_3.push_back(-1);
 }
#endif
#pragma omp parallel private(__tid)
{
 __tid = omp_get_thread_num();
 int __flag;
 int __loop;
 int task_id1,task_id2,task_id3;
#pragma omp single nowait
{ /* receive thread */
 __tid_receiver = __tid;
 IF_TIME(t_unpack_start = rtclock());
 for (__p=0; __p<__MAX_NUM_RECVS; __p++) {
  MPI_Irecv(recv_buf_B[__p], max_num_elements_B, MPI_DOUBLE, MPI_ANY_SOURCE, 13123, MPI_COMM_WORLD, &recv_reqs_B[__p]);
  MPI_Irecv(recv_buf_x[__p], max_num_elements_x, MPI_DOUBLE, MPI_ANY_SOURCE, 12123, MPI_COMM_WORLD, &recv_reqs_x[__p]);
  MPI_Irecv(recv_buf_w[__p], max_num_elements_w, MPI_DOUBLE, MPI_ANY_SOURCE, 7123, MPI_COMM_WORLD, &recv_reqs_w[__p]);
 }
 IF_TIME(t_unpack += rtclock() - t_unpack_start);
#ifdef __DYNSCHEDULER_DEBUG_PRINT
 fprintf(__debug_print_fp, "node %d recv thread %d num tasks to unpack %ld max recv buffers %d B max elements %d x max elements %d w max elements %d\n", my_rank, __tid, _num_tasks_to_unpack, __MAX_NUM_RECVS, max_num_elements_B, max_num_elements_x, max_num_elements_w);
 IF_DEBUG_FLUSH(fflush(__debug_print_fp));
#endif
 IF_TIME(t_wait_unpack_start = rtclock());
 while(_num_tasks_to_unpack > 0) {
  for (__p=0; (__p<recv_buf_B.size()) || (__p<recv_buf_x.size()) || (__p<recv_buf_w.size()); __p++) {
  if (__p<recv_buf_B.size()) {
   if (recv_reqs_B[__p] != MPI_REQUEST_NULL) {
    MPI_Test(&recv_reqs_B[__p],&recv_completed_B[__p],recv_stats_B);
#ifdef __DYNSCHEDULER_DEBUG_PRINT
    if (recv_completed_B[__p]) {
     __loop = (int)recv_buf_B[__p][0];
     task_id1 = (int)recv_buf_B[__p][1];
     task_id2 = (int)recv_buf_B[__p][2];
     task_id3 = (int)recv_buf_B[__p][3];
     int size;
     MPI_Get_count(recv_stats_B, MPI_DOUBLE, &size);
     fprintf(__debug_print_fp, "received B node %d from node %d buf %d size %d loop %d task %d %d %d\n", my_rank, recv_stats_B->MPI_SOURCE, __p, size, __loop, task_id1, task_id2, task_id3);
     IF_DEBUG_FLUSH(fflush(__debug_print_fp));
    }
#endif
   }
   if (recv_completed_B[__p]) {
    __loop = (int)recv_buf_B[__p][0];
    task_id1 = (int)recv_buf_B[__p][1];
    task_id2 = (int)recv_buf_B[__p][2];
    task_id3 = (int)recv_buf_B[__p][3];
    switch(__loop){
     case 0:
#ifdef __DYNSCHEDULER_MORE_DEBUG_PRINT
      if (prev_firing_count_B_0[__p] != tasks_loop0[(task_id2-lb_tasks_loop0_dim0)*max_num_tasks_loop0_dim1+(task_id3-lb_tasks_loop0_dim1)+0]) {
       prev_firing_count_B_0[__p] = tasks_loop0[(task_id2-lb_tasks_loop0_dim0)*max_num_tasks_loop0_dim1+(task_id3-lb_tasks_loop0_dim1)+0];
       fprintf(__debug_print_fp, "waiting to unpack B node %d loop %d task %d %d %d count %d\n", my_rank, __loop, task_id1, task_id2, task_id3, tasks_loop0[(task_id2-lb_tasks_loop0_dim0)*max_num_tasks_loop0_dim1+(task_id3-lb_tasks_loop0_dim1)+0]);
       IF_DEBUG_FLUSH(fflush(__debug_print_fp));
      }
#endif
      if (tasks_loop0[(task_id2-lb_tasks_loop0_dim0)*max_num_tasks_loop0_dim1+(task_id3-lb_tasks_loop0_dim1)+0] <= 1) {
       IF_TIME(t_wait_unpack += rtclock() - t_wait_unpack_start);
       IF_TIME(t_unpack_start = rtclock());
       if (tasks_loop0[(task_id2-lb_tasks_loop0_dim0)*max_num_tasks_loop0_dim1+(task_id3-lb_tasks_loop0_dim1)+0] == 1) {
        tasks_loop0[(task_id2-lb_tasks_loop0_dim0)*max_num_tasks_loop0_dim1+(task_id3-lb_tasks_loop0_dim1)+0] -= !is_receiver_B_0(task_id1,task_id2,task_id3,N,n,my_rank,nprocs);
       }
       unpack_buf_B_0(my_rank,nprocs, B_trans, B_size_1, buff_mang_B,recv_buf_B[__p]);
       tasks_loop0[(task_id2-lb_tasks_loop0_dim0)*max_num_tasks_loop0_dim1+(task_id3-lb_tasks_loop0_dim1)+0]--;
       recv_completed_B[__p] = 0;
       MPI_Irecv(recv_buf_B[__p], max_num_elements_B, MPI_DOUBLE, MPI_ANY_SOURCE, 13123, MPI_COMM_WORLD, &recv_reqs_B[__p]);
#ifdef __DYNSCHEDULER_MORE_DEBUG_PRINT
       prev_firing_count_B_0[__p] = -1;
#endif
#ifdef __DYNSCHEDULER_DEBUG_PRINT
       fprintf(__debug_print_fp, "unpacked B node %d loop %d task %d %d %d count %d\n", my_rank, __loop, task_id1, task_id2, task_id3, tasks_loop0[(task_id2-lb_tasks_loop0_dim0)*max_num_tasks_loop0_dim1+(task_id3-lb_tasks_loop0_dim1)+0]);
       IF_DEBUG_FLUSH(fflush(__debug_print_fp));
#endif
       if (tasks_loop0[(task_id2-lb_tasks_loop0_dim0)*max_num_tasks_loop0_dim1+(task_id3-lb_tasks_loop0_dim1)+0] == 0) {
        remote_update_dep_tasks_0(task_id1,task_id2,task_id3,N,n,my_rank,nprocs,_num_threads,tasks_loop0,tasks_loop1,tasks_loop2,tasks_loop3);
        _num_tasks_to_unpack--;
#ifdef __DYNSCHEDULER_DEBUG_PRINT
        fprintf(__debug_print_fp, "remote updated B node %d loop %d task %d %d %d num tasks left to unpack %ld\n", my_rank, __loop, task_id1, task_id2, task_id3, _num_tasks_to_unpack);
        IF_DEBUG_FLUSH(fflush(__debug_print_fp));
#endif
       }
       IF_TIME(t_unpack += rtclock() - t_unpack_start);
       IF_TIME(t_wait_unpack_start = rtclock());
      }
      break;
     default:
      fprintf(stderr, "ERROR: B data cannot be sent by a task of loop %d\n", __loop);
      assert(0);
    }
   }
  }
  if (__p<recv_buf_x.size()) {
   if (recv_reqs_x[__p] != MPI_REQUEST_NULL) {
    MPI_Test(&recv_reqs_x[__p],&recv_completed_x[__p],recv_stats_x);
#ifdef __DYNSCHEDULER_DEBUG_PRINT
    if (recv_completed_x[__p]) {
     __loop = (int)recv_buf_x[__p][0];
     task_id1 = (int)recv_buf_x[__p][1];
     task_id2 = (int)recv_buf_x[__p][2];
     task_id3 = (int)recv_buf_x[__p][3];
     int size;
     MPI_Get_count(recv_stats_x, MPI_DOUBLE, &size);
     fprintf(__debug_print_fp, "received x node %d from node %d buf %d size %d loop %d task %d %d %d\n", my_rank, recv_stats_x->MPI_SOURCE, __p, size, __loop, task_id1, task_id2, task_id3);
     IF_DEBUG_FLUSH(fflush(__debug_print_fp));
    }
#endif
   }
   if (recv_completed_x[__p]) {
    __loop = (int)recv_buf_x[__p][0];
    task_id1 = (int)recv_buf_x[__p][1];
    task_id2 = (int)recv_buf_x[__p][2];
    task_id3 = (int)recv_buf_x[__p][3];
    switch(__loop){
     case 1:
#ifdef __DYNSCHEDULER_MORE_DEBUG_PRINT
      if (prev_firing_count_x_1[__p] != tasks_loop1[(task_id2-lb_tasks_loop1_dim0)*max_num_tasks_loop1_dim1+(task_id3-lb_tasks_loop1_dim1)+0]) {
       prev_firing_count_x_1[__p] = tasks_loop1[(task_id2-lb_tasks_loop1_dim0)*max_num_tasks_loop1_dim1+(task_id3-lb_tasks_loop1_dim1)+0];
       fprintf(__debug_print_fp, "waiting to unpack x node %d loop %d task %d %d %d count %d\n", my_rank, __loop, task_id1, task_id2, task_id3, tasks_loop1[(task_id2-lb_tasks_loop1_dim0)*max_num_tasks_loop1_dim1+(task_id3-lb_tasks_loop1_dim1)+0]);
       IF_DEBUG_FLUSH(fflush(__debug_print_fp));
      }
#endif
      if (tasks_loop1[(task_id2-lb_tasks_loop1_dim0)*max_num_tasks_loop1_dim1+(task_id3-lb_tasks_loop1_dim1)+0] <= 1) {
       IF_TIME(t_wait_unpack += rtclock() - t_wait_unpack_start);
       IF_TIME(t_unpack_start = rtclock());
       if (tasks_loop1[(task_id2-lb_tasks_loop1_dim0)*max_num_tasks_loop1_dim1+(task_id3-lb_tasks_loop1_dim1)+0] == 1) {
        tasks_loop1[(task_id2-lb_tasks_loop1_dim0)*max_num_tasks_loop1_dim1+(task_id3-lb_tasks_loop1_dim1)+0] -= !is_receiver_x_1(task_id1,task_id2,task_id3,N,n,my_rank,nprocs);
       }
       unpack_buf_x_1(my_rank,nprocs, x_trans, buff_mang_x,recv_buf_x[__p]);
       tasks_loop1[(task_id2-lb_tasks_loop1_dim0)*max_num_tasks_loop1_dim1+(task_id3-lb_tasks_loop1_dim1)+0]--;
       recv_completed_x[__p] = 0;
       MPI_Irecv(recv_buf_x[__p], max_num_elements_x, MPI_DOUBLE, MPI_ANY_SOURCE, 12123, MPI_COMM_WORLD, &recv_reqs_x[__p]);
#ifdef __DYNSCHEDULER_MORE_DEBUG_PRINT
       prev_firing_count_x_1[__p] = -1;
#endif
#ifdef __DYNSCHEDULER_DEBUG_PRINT
       fprintf(__debug_print_fp, "unpacked x node %d loop %d task %d %d %d count %d\n", my_rank, __loop, task_id1, task_id2, task_id3, tasks_loop1[(task_id2-lb_tasks_loop1_dim0)*max_num_tasks_loop1_dim1+(task_id3-lb_tasks_loop1_dim1)+0]);
       IF_DEBUG_FLUSH(fflush(__debug_print_fp));
#endif
       if (tasks_loop1[(task_id2-lb_tasks_loop1_dim0)*max_num_tasks_loop1_dim1+(task_id3-lb_tasks_loop1_dim1)+0] == 0) {
        remote_update_dep_tasks_1(task_id1,task_id2,task_id3,N,n,my_rank,nprocs,_num_threads,tasks_loop0,tasks_loop1,tasks_loop2,tasks_loop3);
        _num_tasks_to_unpack--;
#ifdef __DYNSCHEDULER_DEBUG_PRINT
        fprintf(__debug_print_fp, "remote updated x node %d loop %d task %d %d %d num tasks left to unpack %ld\n", my_rank, __loop, task_id1, task_id2, task_id3, _num_tasks_to_unpack);
        IF_DEBUG_FLUSH(fflush(__debug_print_fp));
#endif
       }
       IF_TIME(t_unpack += rtclock() - t_unpack_start);
       IF_TIME(t_wait_unpack_start = rtclock());
      }
      break;
     case 2:
#ifdef __DYNSCHEDULER_MORE_DEBUG_PRINT
      if (prev_firing_count_x_2[__p] != tasks_loop2[(task_id2-lb_tasks_loop2_dim0)+0]) {
       prev_firing_count_x_2[__p] = tasks_loop2[(task_id2-lb_tasks_loop2_dim0)+0];
       fprintf(__debug_print_fp, "waiting to unpack x node %d loop %d task %d %d %d count %d\n", my_rank, __loop, task_id1, task_id2, task_id3, tasks_loop2[(task_id2-lb_tasks_loop2_dim0)+0]);
       IF_DEBUG_FLUSH(fflush(__debug_print_fp));
      }
#endif
      if (tasks_loop2[(task_id2-lb_tasks_loop2_dim0)+0] <= 1) {
       IF_TIME(t_wait_unpack += rtclock() - t_wait_unpack_start);
       IF_TIME(t_unpack_start = rtclock());
       if (tasks_loop2[(task_id2-lb_tasks_loop2_dim0)+0] == 1) {
        tasks_loop2[(task_id2-lb_tasks_loop2_dim0)+0] -= !is_receiver_x_2(task_id1,task_id2,N,n,my_rank,nprocs);
       }
       unpack_buf_x_2(my_rank,nprocs, x_trans, buff_mang_x,recv_buf_x[__p]);
       tasks_loop2[(task_id2-lb_tasks_loop2_dim0)+0]--;
       recv_completed_x[__p] = 0;
       MPI_Irecv(recv_buf_x[__p], max_num_elements_x, MPI_DOUBLE, MPI_ANY_SOURCE, 12123, MPI_COMM_WORLD, &recv_reqs_x[__p]);
#ifdef __DYNSCHEDULER_MORE_DEBUG_PRINT
       prev_firing_count_x_2[__p] = -1;
#endif
#ifdef __DYNSCHEDULER_DEBUG_PRINT
       fprintf(__debug_print_fp, "unpacked x node %d loop %d task %d %d %d count %d\n", my_rank, __loop, task_id1, task_id2, task_id3, tasks_loop2[(task_id2-lb_tasks_loop2_dim0)+0]);
       IF_DEBUG_FLUSH(fflush(__debug_print_fp));
#endif
       if (tasks_loop2[(task_id2-lb_tasks_loop2_dim0)+0] == 0) {
        remote_update_dep_tasks_2(task_id1,task_id2,N,n,my_rank,nprocs,_num_threads,tasks_loop0,tasks_loop1,tasks_loop2,tasks_loop3);
        _num_tasks_to_unpack--;
#ifdef __DYNSCHEDULER_DEBUG_PRINT
        fprintf(__debug_print_fp, "remote updated x node %d loop %d task %d %d %d num tasks left to unpack %ld\n", my_rank, __loop, task_id1, task_id2, task_id3, _num_tasks_to_unpack);
        IF_DEBUG_FLUSH(fflush(__debug_print_fp));
#endif
       }
       IF_TIME(t_unpack += rtclock() - t_unpack_start);
       IF_TIME(t_wait_unpack_start = rtclock());
      }
      break;
     default:
      fprintf(stderr, "ERROR: x data cannot be sent by a task of loop %d\n", __loop);
      assert(0);
    }
   }
  }
  if (__p<recv_buf_w.size()) {
   if (recv_reqs_w[__p] != MPI_REQUEST_NULL) {
    MPI_Test(&recv_reqs_w[__p],&recv_completed_w[__p],recv_stats_w);
#ifdef __DYNSCHEDULER_DEBUG_PRINT
    if (recv_completed_w[__p]) {
     __loop = (int)recv_buf_w[__p][0];
     task_id1 = (int)recv_buf_w[__p][1];
     task_id2 = (int)recv_buf_w[__p][2];
     task_id3 = (int)recv_buf_w[__p][3];
     int size;
     MPI_Get_count(recv_stats_w, MPI_DOUBLE, &size);
     fprintf(__debug_print_fp, "received w node %d from node %d buf %d size %d loop %d task %d %d %d\n", my_rank, recv_stats_w->MPI_SOURCE, __p, size, __loop, task_id1, task_id2, task_id3);
     IF_DEBUG_FLUSH(fflush(__debug_print_fp));
    }
#endif
   }
   if (recv_completed_w[__p]) {
    __loop = (int)recv_buf_w[__p][0];
    task_id1 = (int)recv_buf_w[__p][1];
    task_id2 = (int)recv_buf_w[__p][2];
    task_id3 = (int)recv_buf_w[__p][3];
    switch(__loop){
     case 3:
#ifdef __DYNSCHEDULER_MORE_DEBUG_PRINT
      if (prev_firing_count_w_3[__p] != tasks_loop3[(task_id2-lb_tasks_loop3_dim0)*max_num_tasks_loop3_dim1+(task_id3-lb_tasks_loop3_dim1)+0]) {
       prev_firing_count_w_3[__p] = tasks_loop3[(task_id2-lb_tasks_loop3_dim0)*max_num_tasks_loop3_dim1+(task_id3-lb_tasks_loop3_dim1)+0];
       fprintf(__debug_print_fp, "waiting to unpack w node %d loop %d task %d %d %d count %d\n", my_rank, __loop, task_id1, task_id2, task_id3, tasks_loop3[(task_id2-lb_tasks_loop3_dim0)*max_num_tasks_loop3_dim1+(task_id3-lb_tasks_loop3_dim1)+0]);
       IF_DEBUG_FLUSH(fflush(__debug_print_fp));
      }
#endif
      if (tasks_loop3[(task_id2-lb_tasks_loop3_dim0)*max_num_tasks_loop3_dim1+(task_id3-lb_tasks_loop3_dim1)+0] <= 1) {
       IF_TIME(t_wait_unpack += rtclock() - t_wait_unpack_start);
       IF_TIME(t_unpack_start = rtclock());
       if (tasks_loop3[(task_id2-lb_tasks_loop3_dim0)*max_num_tasks_loop3_dim1+(task_id3-lb_tasks_loop3_dim1)+0] == 1) {
        tasks_loop3[(task_id2-lb_tasks_loop3_dim0)*max_num_tasks_loop3_dim1+(task_id3-lb_tasks_loop3_dim1)+0] -= !is_receiver_w_3(task_id1,task_id2,task_id3,N,n,my_rank,nprocs);
       }
       unpack_buf_w_3(my_rank,nprocs, w_trans, buff_mang_w,recv_buf_w[__p]);
       tasks_loop3[(task_id2-lb_tasks_loop3_dim0)*max_num_tasks_loop3_dim1+(task_id3-lb_tasks_loop3_dim1)+0]--;
       recv_completed_w[__p] = 0;
       MPI_Irecv(recv_buf_w[__p], max_num_elements_w, MPI_DOUBLE, MPI_ANY_SOURCE, 7123, MPI_COMM_WORLD, &recv_reqs_w[__p]);
#ifdef __DYNSCHEDULER_MORE_DEBUG_PRINT
       prev_firing_count_w_3[__p] = -1;
#endif
#ifdef __DYNSCHEDULER_DEBUG_PRINT
       fprintf(__debug_print_fp, "unpacked w node %d loop %d task %d %d %d count %d\n", my_rank, __loop, task_id1, task_id2, task_id3, tasks_loop3[(task_id2-lb_tasks_loop3_dim0)*max_num_tasks_loop3_dim1+(task_id3-lb_tasks_loop3_dim1)+0]);
       IF_DEBUG_FLUSH(fflush(__debug_print_fp));
#endif
       if (tasks_loop3[(task_id2-lb_tasks_loop3_dim0)*max_num_tasks_loop3_dim1+(task_id3-lb_tasks_loop3_dim1)+0] == 0) {
        remote_update_dep_tasks_3(task_id1,task_id2,task_id3,N,n,my_rank,nprocs,_num_threads,tasks_loop0,tasks_loop1,tasks_loop2,tasks_loop3);
        _num_tasks_to_unpack--;
#ifdef __DYNSCHEDULER_DEBUG_PRINT
        fprintf(__debug_print_fp, "remote updated w node %d loop %d task %d %d %d num tasks left to unpack %ld\n", my_rank, __loop, task_id1, task_id2, task_id3, _num_tasks_to_unpack);
        IF_DEBUG_FLUSH(fflush(__debug_print_fp));
#endif
       }
       IF_TIME(t_unpack += rtclock() - t_unpack_start);
       IF_TIME(t_wait_unpack_start = rtclock());
      }
      break;
     default:
      fprintf(stderr, "ERROR: w data cannot be sent by a task of loop %d\n", __loop);
      assert(0);
    }
   }
  }
  }
  for (__p=0; __p<recv_buf_B.size(); __p++) {
   if (recv_reqs_B[__p] != MPI_REQUEST_NULL) break;
  }
  if (__p == recv_buf_B.size()) {
   for (__p=0; __p<nprocs; __p++) {
    double *buf = NULL;
    size_t buf_size = 0;
    buf = (double *) polyrt_max_alloc(buf, sizeof(double)*max_num_elements_B, &buf_size);
    recv_buf_B.push_back(buf);
    recv_reqs_B.push_back(MPI_REQUEST_NULL);
    recv_completed_B.push_back(0);
#ifdef __DYNSCHEDULER_MORE_DEBUG_PRINT
    prev_firing_count_B_0.push_back(-1);
    prev_firing_count_B_1.push_back(-1);
    prev_firing_count_B_2.push_back(-1);
    prev_firing_count_B_3.push_back(-1);
#endif
    MPI_Irecv(buf, max_num_elements_B, MPI_DOUBLE, MPI_ANY_SOURCE, 13123, MPI_COMM_WORLD, &recv_reqs_B[recv_reqs_B.size()-1]);
   }
  }
  for (__p=0; __p<recv_buf_x.size(); __p++) {
   if (recv_reqs_x[__p] != MPI_REQUEST_NULL) break;
  }
  if (__p == recv_buf_x.size()) {
   for (__p=0; __p<nprocs; __p++) {
    double *buf = NULL;
    size_t buf_size = 0;
    buf = (double *) polyrt_max_alloc(buf, sizeof(double)*max_num_elements_x, &buf_size);
    recv_buf_x.push_back(buf);
    recv_reqs_x.push_back(MPI_REQUEST_NULL);
    recv_completed_x.push_back(0);
#ifdef __DYNSCHEDULER_MORE_DEBUG_PRINT
    prev_firing_count_x_0.push_back(-1);
    prev_firing_count_x_1.push_back(-1);
    prev_firing_count_x_2.push_back(-1);
    prev_firing_count_x_3.push_back(-1);
#endif
    MPI_Irecv(buf, max_num_elements_x, MPI_DOUBLE, MPI_ANY_SOURCE, 12123, MPI_COMM_WORLD, &recv_reqs_x[recv_reqs_x.size()-1]);
   }
  }
  for (__p=0; __p<recv_buf_w.size(); __p++) {
   if (recv_reqs_w[__p] != MPI_REQUEST_NULL) break;
  }
  if (__p == recv_buf_w.size()) {
   for (__p=0; __p<nprocs; __p++) {
    double *buf = NULL;
    size_t buf_size = 0;
    buf = (double *) polyrt_max_alloc(buf, sizeof(double)*max_num_elements_w, &buf_size);
    recv_buf_w.push_back(buf);
    recv_reqs_w.push_back(MPI_REQUEST_NULL);
    recv_completed_w.push_back(0);
#ifdef __DYNSCHEDULER_MORE_DEBUG_PRINT
    prev_firing_count_w_0.push_back(-1);
    prev_firing_count_w_1.push_back(-1);
    prev_firing_count_w_2.push_back(-1);
    prev_firing_count_w_3.push_back(-1);
#endif
    MPI_Irecv(buf, max_num_elements_w, MPI_DOUBLE, MPI_ANY_SOURCE, 7123, MPI_COMM_WORLD, &recv_reqs_w[recv_reqs_w.size()-1]);
   }
  }
 }
 IF_TIME(t_unpack_start = rtclock());
 for (__p=0; __p<recv_buf_B.size(); __p++) {
  if (recv_reqs_B[__p] != MPI_REQUEST_NULL) MPI_Request_free(&recv_reqs_B[__p]);
 }
 for (__p=0; __p<recv_buf_x.size(); __p++) {
  if (recv_reqs_x[__p] != MPI_REQUEST_NULL) MPI_Request_free(&recv_reqs_x[__p]);
 }
 for (__p=0; __p<recv_buf_w.size(); __p++) {
  if (recv_reqs_w[__p] != MPI_REQUEST_NULL) MPI_Request_free(&recv_reqs_w[__p]);
 }
 IF_TIME(t_unpack += rtclock() - t_unpack_start);
#ifdef __DYNSCHEDULER_DEBUG_PRINT
 fprintf(__debug_print_fp, "node %d receive done\n", my_rank);
 IF_DEBUG_FLUSH(fflush(__debug_print_fp));
#endif
}
{ /* compute-pack-send threads */
#ifdef __DYNSCHEDULER_DEBUG_PRINT
 fprintf(__debug_print_fp, "node %d compute thread %d num tasks to execute %ld\n", my_rank, __tid, _num_tasks_to_execute);
 IF_DEBUG_FLUSH(fflush(__debug_print_fp));
#endif
 IF_TIME(t_wait_comp_start[__tid] = rtclock());
 while (_num_tasks_to_execute > 0) {
  __Task task;
  __flag = pqueue.try_pop(task);
  if(__flag==0) { IF_TIME(waiting_pops[__tid]++); continue; }
  __loop = task.loop;
  task_id1 = task.id1;
  task_id2 = task.id2;
  task_id3 = task.id3;
#ifdef __DYNSCHEDULER_DEBUG_PRINT
  long int captured;
#pragma omp atomic capture
  captured = --_num_tasks_to_execute;
  fprintf(__debug_print_fp, "computing node %d thread %d loop %d task %d %d %d ready tasks %lu\n", my_rank, __tid, __loop, task_id1, task_id2, task_id3, pqueue.size());
  IF_DEBUG_FLUSH(fflush(__debug_print_fp));
#else
#pragma omp atomic
  _num_tasks_to_execute--;
#endif
  switch(__loop){
   case 0:
    IF_TIME(t_wait_comp[__tid] += rtclock() - t_wait_comp_start[__tid]);
    IF_TIME(t_data_mang_start[__tid] = rtclock());
    data_tile_alloc_0(task_id1,task_id2,task_id3, B_trans, B_size_1, buff_mang_B, A_trans, A_size_1, buff_mang_A, u1_trans, buff_mang_u1, v1_trans, buff_mang_v1, u2_trans, buff_mang_u2, v2_trans, buff_mang_v2);
    IF_TIME(t_data_mang[__tid] += rtclock() - t_data_mang_start[__tid]);
    IF_TIME(t_comp_start[__tid] = rtclock());
    compute_task_0(task_id1,task_id2,task_id3, B_trans, B_size_1, buff_mang_B, A_trans, A_size_1, buff_mang_A, u1_trans, buff_mang_u1, v1_trans, buff_mang_v1, u2_trans, buff_mang_u2, v2_trans, buff_mang_v2);
    IF_TIME(t_comp[__tid] += rtclock() - t_comp_start[__tid]);
    IF_TIME(t_data_mang_start[__tid] = rtclock());
    data_tile_ref_count_update_0(task_id1,task_id2,task_id3, B_trans, B_size_1, buff_mang_B, A_trans, A_size_1, buff_mang_A, u1_trans, buff_mang_u1, v1_trans, buff_mang_v1, u2_trans, buff_mang_u2, v2_trans, buff_mang_v2);
    IF_TIME(t_data_mang[__tid] += rtclock() - t_data_mang_start[__tid]);
    IF_TIME(t_pack_start[__tid] = rtclock());
    pack_send_0(my_rank,nprocs,task_id1,task_id2,task_id3, B_trans, B_size_1, buff_mang_B,send_buf_B[__tid],send_reqs_B[__tid],send_stats_B[__tid], A_trans, A_size_1, buff_mang_A,send_buf_A[__tid],send_reqs_A[__tid],send_stats_A[__tid], u1_trans, buff_mang_u1,send_buf_u1[__tid],send_reqs_u1[__tid],send_stats_u1[__tid], v1_trans, buff_mang_v1,send_buf_v1[__tid],send_reqs_v1[__tid],send_stats_v1[__tid], u2_trans, buff_mang_u2,send_buf_u2[__tid],send_reqs_u2[__tid],send_stats_u2[__tid], v2_trans, buff_mang_v2,send_buf_v2[__tid],send_reqs_v2[__tid],send_stats_v2[__tid],__MAX_NUM_SENDS,&__total_count[__tid]);
    IF_TIME(t_pack[__tid] += rtclock() - t_pack_start[__tid]);
    IF_TIME(t_tasks_manage_start[__tid] = rtclock());
    local_update_dep_tasks_0(task_id1,task_id2,task_id3,N,n,my_rank,nprocs,_num_threads,tasks_loop0,tasks_loop1,tasks_loop2,tasks_loop3);
    IF_TIME(t_tasks_manage[__tid] += rtclock() - t_tasks_manage_start[__tid]);
    IF_TIME(t_wait_comp_start[__tid] = rtclock());
    break;
   case 1:
    IF_TIME(t_wait_comp[__tid] += rtclock() - t_wait_comp_start[__tid]);
    IF_TIME(t_data_mang_start[__tid] = rtclock());
    data_tile_alloc_1(task_id1,task_id2,task_id3, x_trans, buff_mang_x, B_trans, B_size_1, buff_mang_B, y_trans, buff_mang_y);
    IF_TIME(t_data_mang[__tid] += rtclock() - t_data_mang_start[__tid]);
    IF_TIME(t_comp_start[__tid] = rtclock());
    compute_task_1(task_id1,task_id2,task_id3, x_trans, buff_mang_x, B_trans, B_size_1, buff_mang_B, y_trans, buff_mang_y);
    IF_TIME(t_comp[__tid] += rtclock() - t_comp_start[__tid]);
    IF_TIME(t_data_mang_start[__tid] = rtclock());
    data_tile_ref_count_update_1(task_id1,task_id2,task_id3, x_trans, buff_mang_x, B_trans, B_size_1, buff_mang_B, y_trans, buff_mang_y);
    IF_TIME(t_data_mang[__tid] += rtclock() - t_data_mang_start[__tid]);
    IF_TIME(t_pack_start[__tid] = rtclock());
    pack_send_1(my_rank,nprocs,task_id1,task_id2,task_id3, x_trans, buff_mang_x,send_buf_x[__tid],send_reqs_x[__tid],send_stats_x[__tid], B_trans, B_size_1, buff_mang_B,send_buf_B[__tid],send_reqs_B[__tid],send_stats_B[__tid], y_trans, buff_mang_y,send_buf_y[__tid],send_reqs_y[__tid],send_stats_y[__tid],__MAX_NUM_SENDS,&__total_count[__tid]);
    IF_TIME(t_pack[__tid] += rtclock() - t_pack_start[__tid]);
    IF_TIME(t_tasks_manage_start[__tid] = rtclock());
    local_update_dep_tasks_1(task_id1,task_id2,task_id3,N,n,my_rank,nprocs,_num_threads,tasks_loop0,tasks_loop1,tasks_loop2,tasks_loop3);
    IF_TIME(t_tasks_manage[__tid] += rtclock() - t_tasks_manage_start[__tid]);
    IF_TIME(t_wait_comp_start[__tid] = rtclock());
    break;
   case 2:
    IF_TIME(t_wait_comp[__tid] += rtclock() - t_wait_comp_start[__tid]);
    IF_TIME(t_data_mang_start[__tid] = rtclock());
    data_tile_alloc_2(task_id1,task_id2, x_trans, buff_mang_x, z_trans, buff_mang_z);
    IF_TIME(t_data_mang[__tid] += rtclock() - t_data_mang_start[__tid]);
    IF_TIME(t_comp_start[__tid] = rtclock());
    compute_task_2(task_id1,task_id2, x_trans, buff_mang_x, z_trans, buff_mang_z);
    IF_TIME(t_comp[__tid] += rtclock() - t_comp_start[__tid]);
    IF_TIME(t_data_mang_start[__tid] = rtclock());
    data_tile_ref_count_update_2(task_id1,task_id2, x_trans, buff_mang_x, z_trans, buff_mang_z);
    IF_TIME(t_data_mang[__tid] += rtclock() - t_data_mang_start[__tid]);
    IF_TIME(t_pack_start[__tid] = rtclock());
    pack_send_2(my_rank,nprocs,task_id1,task_id2, x_trans, buff_mang_x,send_buf_x[__tid],send_reqs_x[__tid],send_stats_x[__tid], z_trans, buff_mang_z,send_buf_z[__tid],send_reqs_z[__tid],send_stats_z[__tid],__MAX_NUM_SENDS,&__total_count[__tid]);
    IF_TIME(t_pack[__tid] += rtclock() - t_pack_start[__tid]);
    IF_TIME(t_tasks_manage_start[__tid] = rtclock());
    local_update_dep_tasks_2(task_id1,task_id2,N,n,my_rank,nprocs,_num_threads,tasks_loop0,tasks_loop1,tasks_loop2,tasks_loop3);
    IF_TIME(t_tasks_manage[__tid] += rtclock() - t_tasks_manage_start[__tid]);
    IF_TIME(t_wait_comp_start[__tid] = rtclock());
    break;
   case 3:
    IF_TIME(t_wait_comp[__tid] += rtclock() - t_wait_comp_start[__tid]);
    IF_TIME(t_data_mang_start[__tid] = rtclock());
    data_tile_alloc_3(task_id1,task_id2,task_id3, w_trans, buff_mang_w, B_trans, B_size_1, buff_mang_B, x_trans, buff_mang_x);
    IF_TIME(t_data_mang[__tid] += rtclock() - t_data_mang_start[__tid]);
    IF_TIME(t_comp_start[__tid] = rtclock());
    compute_task_3(task_id1,task_id2,task_id3, w_trans, buff_mang_w, B_trans, B_size_1, buff_mang_B, x_trans, buff_mang_x);
    IF_TIME(t_comp[__tid] += rtclock() - t_comp_start[__tid]);
    IF_TIME(t_data_mang_start[__tid] = rtclock());
    data_tile_ref_count_update_3(task_id1,task_id2,task_id3, w_trans, buff_mang_w, B_trans, B_size_1, buff_mang_B, x_trans, buff_mang_x);
    IF_TIME(t_data_mang[__tid] += rtclock() - t_data_mang_start[__tid]);
    IF_TIME(t_pack_start[__tid] = rtclock());
    pack_send_3(my_rank,nprocs,task_id1,task_id2,task_id3, w_trans, buff_mang_w,send_buf_w[__tid],send_reqs_w[__tid],send_stats_w[__tid], B_trans, B_size_1, buff_mang_B,send_buf_B[__tid],send_reqs_B[__tid],send_stats_B[__tid], x_trans, buff_mang_x,send_buf_x[__tid],send_reqs_x[__tid],send_stats_x[__tid],__MAX_NUM_SENDS,&__total_count[__tid]);
    IF_TIME(t_pack[__tid] += rtclock() - t_pack_start[__tid]);
    IF_TIME(t_tasks_manage_start[__tid] = rtclock());
    local_update_dep_tasks_3(task_id1,task_id2,task_id3,N,n,my_rank,nprocs,_num_threads,tasks_loop0,tasks_loop1,tasks_loop2,tasks_loop3);
    IF_TIME(t_tasks_manage[__tid] += rtclock() - t_tasks_manage_start[__tid]);
    IF_TIME(t_wait_comp_start[__tid] = rtclock());
    break;
   default:
    assert(0);
  }
#ifdef __DYNSCHEDULER_MORE_DEBUG_PRINT
  fprintf(__debug_print_fp, "computed node %d thread %d loop %d task %d %d %d tasks left %ld\n", my_rank, __tid, __loop, task_id1, task_id2, task_id3, captured);
  IF_DEBUG_FLUSH(fflush(__debug_print_fp));
#endif
 }
#ifdef __DYNSCHEDULER_DEBUG_PRINT
 fprintf(__debug_print_fp, "node %d compute thread %d done\n", my_rank, __tid);
 IF_DEBUG_FLUSH(fflush(__debug_print_fp));
#endif
}
}
#ifdef __DYNSCHEDULER_DEBUG_PRINT
 fclose(__debug_print_fp);
#endif
/*End of dynamic scheduling code*/
 IF_TIME(t_writeout = rtclock());
 
#ifndef USE_LOCAL_ARRAYS
 if (fopen(".test", "r")) {
 copy_back_writeout_0(my_rank, nprocs, B_trans, B_size_1, buff_mang_B, x_trans, buff_mang_x, w_trans, buff_mang_w);
write_out(my_rank,nprocs,lw_buf_B,lw_recv_buf_B,displs_lw_B,lw_buf_x,lw_recv_buf_x,displs_lw_x,lw_buf_w,lw_recv_buf_w,displs_lw_w, B_trans, B_size_1, buff_mang_B, x_trans, buff_mang_x, w_trans, buff_mang_w);
} 
#endif
 IF_TIME(t_writeout = rtclock() - t_writeout);
free_buffer_mang(buff_mang_B);
free_buffer_mang(buff_mang_A);
free_buffer_mang(buff_mang_u1);
free_buffer_mang(buff_mang_v1);
free_buffer_mang(buff_mang_u2);
free_buffer_mang(buff_mang_v2);
free_buffer_mang(buff_mang_x);
free_buffer_mang(buff_mang_y);
free_buffer_mang(buff_mang_z);
free_buffer_mang(buff_mang_w);
#ifdef TIME
 IF_TIME(t_local = rtclock() - t_local - t_writeout);
 for (__tid=0; __tid<_num_threads; __tid++) __total_count_threads += __total_count[__tid];
 waiting_pops_threads = waiting_pops[0];
                    for (__tid=1; __tid<_num_threads; __tid++) if (waiting_pops_threads < waiting_pops[__tid]) waiting_pops_threads = waiting_pops[__tid];
 t_tasks_manage_threads = t_tasks_manage[0];
                    for (__tid=1; __tid<_num_threads; __tid++) if (t_tasks_manage_threads < t_tasks_manage[__tid]) t_tasks_manage_threads = t_tasks_manage[__tid];
 if (__tid_receiver != 0) t_comp_threads_min = t_comp[0];
                    else t_comp_threads_min = t_comp[1];
                    for (__tid=1; __tid<_num_threads; __tid++) if ((__tid != __tid_receiver) && (t_comp_threads_min > t_comp[__tid])) t_comp_threads_min = t_comp[__tid];
 t_comp_threads = t_comp[0];
                    for (__tid=1; __tid<_num_threads; __tid++) if (t_comp_threads < t_comp[__tid]) t_comp_threads = t_comp[__tid];
 t_wait_comp_threads = t_wait_comp[0];
                    for (__tid=1; __tid<_num_threads; __tid++) if (t_wait_comp_threads < t_wait_comp[__tid]) t_wait_comp_threads = t_wait_comp[__tid];
 t_data_mang_threads = t_data_mang[0];
      for (__tid=1; __tid<_num_threads; __tid++) if (t_data_mang_threads < t_data_mang[__tid]) t_data_mang_threads = t_data_mang[__tid];
 t_pack_threads = t_pack[0];
                    for (__tid=1; __tid<_num_threads; __tid++) if (t_pack_threads < t_pack[__tid]) t_pack_threads = t_pack[__tid];
 char *buffer = (char *)malloc(4096 * sizeof(char));
 strcpy(buffer, "");
#ifdef __DETAILED_TIME_REPORT
 sprintf(buffer+strlen(buffer), "Node %d\n", my_rank);
 sprintf(buffer+strlen(buffer), "Write-out time: %lf\n", t_writeout);
 sprintf(buffer+strlen(buffer), "Values communicated: %0.0lf\n", __total_count_threads);
 sprintf(buffer+strlen(buffer), "Data init time: %lf\n", t_data_init);
 sprintf(buffer+strlen(buffer), "Minimum computation time across all threads: %lf\n", t_comp_threads_min);
 sprintf(buffer+strlen(buffer), "Maximum computation time across all threads: %lf\n", t_comp_threads);
 sprintf(buffer+strlen(buffer), "'Waiting for computation' time: %lf\n", t_wait_comp_threads);
 sprintf(buffer+strlen(buffer), "Data mang time: %lf\n", t_data_mang_threads);
 sprintf(buffer+strlen(buffer), "Tasks creation time: %lf\n", t_tasks_create);
 sprintf(buffer+strlen(buffer), "Tasks management time: %lf\n", t_tasks_manage_threads);
 sprintf(buffer+strlen(buffer), "Packing time: %lf\n", t_pack_threads);
 sprintf(buffer+strlen(buffer), "Unpacking time: %lf\n", t_unpack);
 sprintf(buffer+strlen(buffer), "'Waiting for unpacking' time: %lf\n", t_wait_unpack);
 sprintf(buffer+strlen(buffer), "Sum of split times: %lf\n", t_comp_threads + t_wait_comp_threads + t_tasks_create + t_tasks_manage_threads + t_pack_threads);
 sprintf(buffer+strlen(buffer), "Total time minus write-out time: %lf\n", t_local);
 sprintf(buffer+strlen(buffer), "-------");
 fprintf(stdout, "%s\n", buffer);
#endif
 MPI_Reduce(&__total_count_threads, &__total_count_all, 1, MPI_DOUBLE, MPI_SUM, 0, MPI_COMM_WORLD);
 MPI_Reduce(&waiting_pops_threads, &waiting_pops_global, 1, MPI_INT, MPI_MAX, 0, MPI_COMM_WORLD);
 MPI_Reduce(&t_comp_threads_min, &t_globalcomp_min, 1, MPI_DOUBLE, MPI_MIN, 0, MPI_COMM_WORLD);
 MPI_Reduce(&t_comp_threads, &t_globalcomp, 1, MPI_DOUBLE, MPI_MAX, 0, MPI_COMM_WORLD);
 MPI_Reduce(&t_wait_comp_threads, &t_globalwait_comp, 1, MPI_DOUBLE, MPI_MAX, 0, MPI_COMM_WORLD);
 MPI_Reduce(&t_tasks_create, &t_globaltasks_create, 1, MPI_DOUBLE, MPI_MAX, 0, MPI_COMM_WORLD);
 MPI_Reduce(&t_tasks_manage_threads, &t_globaltasks_manage, 1, MPI_DOUBLE, MPI_MAX, 0, MPI_COMM_WORLD);
 MPI_Reduce(&t_pack_threads, &t_globalpack, 1, MPI_DOUBLE, MPI_MAX, 0, MPI_COMM_WORLD);
 MPI_Reduce(&t_unpack, &t_globalunpack, 1, MPI_DOUBLE, MPI_MAX, 0, MPI_COMM_WORLD);
 MPI_Reduce(&t_wait_unpack, &t_globalwait_unpack, 1, MPI_DOUBLE, MPI_MAX, 0, MPI_COMM_WORLD);
 MPI_Reduce(&t_local, &t_global, 1, MPI_DOUBLE, MPI_MAX, 0, MPI_COMM_WORLD);
 if (my_rank==0) {
  strcpy(buffer, "SUMMARY\n");
  sprintf(buffer+strlen(buffer), "Write-out time spent in master node: %0.6lf s\n", t_writeout);
  sprintf(buffer+strlen(buffer), "Total communication volume across all nodes: %0.6lf GB\n", __total_count_all*8/(1024*1024*1024));
  sprintf(buffer+strlen(buffer), "Maximum computation time spent across all nodes: %0.6lf s\n", t_globalcomp);
  sprintf(buffer+strlen(buffer), "Minimum computation time spent across all nodes: %0.6lf s\n", t_globalcomp_min);
  sprintf(buffer+strlen(buffer), "Maximum 'waiting for computation' time spent across all nodes: %0.6lf s\n", t_globalwait_comp);
  sprintf(buffer+strlen(buffer), "Maximum number of waiting/unsuccessful POPs across all nodes: %d\n", waiting_pops_global);
  sprintf(buffer+strlen(buffer), "Maximum tasks creation time spent across all nodes: %0.6lf s\n", t_globaltasks_create);
  sprintf(buffer+strlen(buffer), "Maximum tasks management time spent across all nodes: %0.6lf s\n", t_globaltasks_manage);
  sprintf(buffer+strlen(buffer), "Maximum packing time spent across all nodes: %0.6lf s\n", t_globalpack);
  sprintf(buffer+strlen(buffer), "Maximum unpacking time spent across all nodes: %0.6lf s\n", t_globalunpack);
  sprintf(buffer+strlen(buffer), "Maximum 'waiting for unpacking' time spent across all nodes: %0.6lf s\n", t_globalwait_unpack);
  sprintf(buffer+strlen(buffer), "Maximum total time spent across all nodes: %0.6lf s\n", t_global);
  sprintf(buffer+strlen(buffer), "Maximum data init time spent across all nodes: %0.6lf s\n", t_globaldata_init);
  sprintf(buffer+strlen(buffer), "-------");
  fprintf(stdout, "%s\n", buffer);
 }
#endif
 MPI_Barrier(MPI_COMM_WORLD);
 MPI_Finalize();

    IF_TIME(t_end = rtclock());
    IF_TIME(fprintf(stdout, "%0.6lfs\n", t_end - t_start));

    if (fopen(".test", "r")) {
#ifdef __MPI
        if (my_rank == 0) {
            print_array();
        }
#else
        print_array();
#endif
    }

    return 0;
}


void init_tasks(int my_rank,int nprocs,int num_threads,int *tasks_loop0,int *tasks_loop1,int *tasks_loop2,int *tasks_loop3,long int *p_num_tasks_to_execute,long int *p_num_tasks_to_unpack)
{
int remote_dep_tasks, local_dep_tasks, affinity;
long int _num_tasks_to_execute = 0;
long int _num_tasks_to_unpack = 0;





  int t1, t2, t3, t4;

 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;

/* Start of CLooG code */
if (n >= 1) {
  lbp=0;
  ubp=floord(n-1,128);
#pragma omp parallel for private(remote_dep_tasks, local_dep_tasks, affinity, t4) reduction(+:_num_tasks_to_execute,_num_tasks_to_unpack)
  for (t3=lbp;t3<=ubp;t3++) {
    for (t4=0;t4<=floord(n-1,128);t4++) {
      if (pi_0(0,t3,t4,N,n,nprocs) == my_rank) { _num_tasks_to_execute++;tasks_loop0[(t3-lb_tasks_loop0_dim0)*max_num_tasks_loop0_dim1+(t4-lb_tasks_loop0_dim1)+0] = get_num_remote_src_tasks_0(0,t3,t4,N,n,my_rank,nprocs) + get_num_local_src_tasks_0(0,t3,t4,N,n,my_rank,nprocs);if (tasks_loop0[(t3-lb_tasks_loop0_dim0)*max_num_tasks_loop0_dim1+(t4-lb_tasks_loop0_dim1)+0] == 0) { remote_dep_tasks = get_num_remote_dep_tasks_0(0,t3,t4,N,n,my_rank,nprocs); local_dep_tasks = get_num_local_dep_tasks_0(0,t3,t4,N,n,my_rank,nprocs); if (__is_multi_partitioned[0] || __is_block_cyclic[0]) affinity = -1; else affinity = pi_threads_0(0,t3,t4,N,n,my_rank,nprocs,num_threads); __Task task(0,0,t3,t4,remote_dep_tasks,local_dep_tasks,affinity); pqueue.push(task); IF_DYNSCHEDULER_MORE_DEBUG_PRINT( fprintf(__debug_print_fp, "added node %d loop 0 task %d %d %d                 remote_dep_tasks %d local_dep_tasks %d ready tasks %lu\n", my_rank, 0,t3,t4, remote_dep_tasks, local_dep_tasks, pqueue.size())); IF_DEBUG_FLUSH(fflush(__debug_print_fp)); }} else if (is_receiver_0(0,t3,t4,N,n,my_rank,nprocs)) { _num_tasks_to_unpack++;tasks_loop0[(t3-lb_tasks_loop0_dim0)*max_num_tasks_loop0_dim1+(t4-lb_tasks_loop0_dim1)+0] = 1 + get_num_local_src_tasks_0(0,t3,t4,N,n,my_rank,nprocs);};
    }
  }
  lbp=0;
  ubp=floord(n-1,128);
#pragma omp parallel for private(remote_dep_tasks, local_dep_tasks, affinity, t4) reduction(+:_num_tasks_to_execute,_num_tasks_to_unpack)
  for (t3=lbp;t3<=ubp;t3++) {
    for (t4=0;t4<=floord(n-1,128);t4++) {
      if (pi_1(1,t3,t4,N,n,nprocs) == my_rank) { _num_tasks_to_execute++;tasks_loop1[(t3-lb_tasks_loop1_dim0)*max_num_tasks_loop1_dim1+(t4-lb_tasks_loop1_dim1)+0] = get_num_remote_src_tasks_1(1,t3,t4,N,n,my_rank,nprocs) + get_num_local_src_tasks_1(1,t3,t4,N,n,my_rank,nprocs);if (tasks_loop1[(t3-lb_tasks_loop1_dim0)*max_num_tasks_loop1_dim1+(t4-lb_tasks_loop1_dim1)+0] == 0) { remote_dep_tasks = get_num_remote_dep_tasks_1(1,t3,t4,N,n,my_rank,nprocs); local_dep_tasks = get_num_local_dep_tasks_1(1,t3,t4,N,n,my_rank,nprocs); if (__is_multi_partitioned[1] || __is_block_cyclic[1]) affinity = -1; else affinity = pi_threads_1(1,t3,t4,N,n,my_rank,nprocs,num_threads); __Task task(1,1,t3,t4,remote_dep_tasks,local_dep_tasks,affinity); pqueue.push(task); IF_DYNSCHEDULER_MORE_DEBUG_PRINT( fprintf(__debug_print_fp, "added node %d loop 1 task %d %d %d                 remote_dep_tasks %d local_dep_tasks %d ready tasks %lu\n", my_rank, 1,t3,t4, remote_dep_tasks, local_dep_tasks, pqueue.size())); IF_DEBUG_FLUSH(fflush(__debug_print_fp)); }} else if (is_receiver_1(1,t3,t4,N,n,my_rank,nprocs)) { _num_tasks_to_unpack++;tasks_loop1[(t3-lb_tasks_loop1_dim0)*max_num_tasks_loop1_dim1+(t4-lb_tasks_loop1_dim1)+0] = 1 + get_num_local_src_tasks_1(1,t3,t4,N,n,my_rank,nprocs);};
    }
  }
  lbp=0;
  ubp=floord(n-1,128);
#pragma omp parallel for private(remote_dep_tasks, local_dep_tasks, affinity, t4) reduction(+:_num_tasks_to_execute,_num_tasks_to_unpack)
  for (t3=lbp;t3<=ubp;t3++) {
    if (pi_2(2,t3,N,n,nprocs) == my_rank) { _num_tasks_to_execute++;tasks_loop2[(t3-lb_tasks_loop2_dim0)+0] = get_num_remote_src_tasks_2(2,t3,N,n,my_rank,nprocs) + get_num_local_src_tasks_2(2,t3,N,n,my_rank,nprocs);if (tasks_loop2[(t3-lb_tasks_loop2_dim0)+0] == 0) { remote_dep_tasks = get_num_remote_dep_tasks_2(2,t3,N,n,my_rank,nprocs); local_dep_tasks = get_num_local_dep_tasks_2(2,t3,N,n,my_rank,nprocs); if (__is_multi_partitioned[2] || __is_block_cyclic[2]) affinity = -1; else affinity = pi_threads_2(2,t3,N,n,my_rank,nprocs,num_threads); __Task task(2,2,t3,remote_dep_tasks,local_dep_tasks,affinity); pqueue.push(task); IF_DYNSCHEDULER_MORE_DEBUG_PRINT( fprintf(__debug_print_fp, "added node %d loop 2 task %d %d                 remote_dep_tasks %d local_dep_tasks %d ready tasks %lu\n", my_rank, 2,t3, remote_dep_tasks, local_dep_tasks, pqueue.size())); IF_DEBUG_FLUSH(fflush(__debug_print_fp)); }} else if (is_receiver_2(2,t3,N,n,my_rank,nprocs)) { _num_tasks_to_unpack++;tasks_loop2[(t3-lb_tasks_loop2_dim0)+0] = 1 + get_num_local_src_tasks_2(2,t3,N,n,my_rank,nprocs);};
  }
  lbp=0;
  ubp=floord(n-1,128);
#pragma omp parallel for private(remote_dep_tasks, local_dep_tasks, affinity, t4) reduction(+:_num_tasks_to_execute,_num_tasks_to_unpack)
  for (t3=lbp;t3<=ubp;t3++) {
    for (t4=0;t4<=floord(n-1,128);t4++) {
      if (pi_3(3,t3,t4,N,n,nprocs) == my_rank) { _num_tasks_to_execute++;tasks_loop3[(t3-lb_tasks_loop3_dim0)*max_num_tasks_loop3_dim1+(t4-lb_tasks_loop3_dim1)+0] = get_num_remote_src_tasks_3(3,t3,t4,N,n,my_rank,nprocs) + get_num_local_src_tasks_3(3,t3,t4,N,n,my_rank,nprocs);if (tasks_loop3[(t3-lb_tasks_loop3_dim0)*max_num_tasks_loop3_dim1+(t4-lb_tasks_loop3_dim1)+0] == 0) { remote_dep_tasks = get_num_remote_dep_tasks_3(3,t3,t4,N,n,my_rank,nprocs); local_dep_tasks = get_num_local_dep_tasks_3(3,t3,t4,N,n,my_rank,nprocs); if (__is_multi_partitioned[3] || __is_block_cyclic[3]) affinity = -1; else affinity = pi_threads_3(3,t3,t4,N,n,my_rank,nprocs,num_threads); __Task task(3,3,t3,t4,remote_dep_tasks,local_dep_tasks,affinity); pqueue.push(task); IF_DYNSCHEDULER_MORE_DEBUG_PRINT( fprintf(__debug_print_fp, "added node %d loop 3 task %d %d %d                 remote_dep_tasks %d local_dep_tasks %d ready tasks %lu\n", my_rank, 3,t3,t4, remote_dep_tasks, local_dep_tasks, pqueue.size())); IF_DEBUG_FLUSH(fflush(__debug_print_fp)); }} else if (is_receiver_3(3,t3,t4,N,n,my_rank,nprocs)) { _num_tasks_to_unpack++;tasks_loop3[(t3-lb_tasks_loop3_dim0)*max_num_tasks_loop3_dim1+(t4-lb_tasks_loop3_dim1)+0] = 1 + get_num_local_src_tasks_3(3,t3,t4,N,n,my_rank,nprocs);};
    }
  }
}
/* End of CLooG code */




*p_num_tasks_to_execute = _num_tasks_to_execute;
*p_num_tasks_to_unpack = _num_tasks_to_unpack;
}

void data_tile_ref_count_init_B_0(int ts1,int ts2,int ts3, double** B_trans,int B_size_1, BufferManager *buff_mang_B)
{


  int t1, t2, t3;

 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;

/* Start of CLooG code */
if ((ts1 == 0) && (ts2 >= 0) && (128*ts2 <= n-1) && (ts3 >= 0) && (128*ts3 <= n-1)) {
  increment_ref_count(buff_mang_B, (ts3)* ( B_size_1 ) + (ts2));;
}
/* End of CLooG code */

}

void data_tile_ref_count_init_A_0(int ts1,int ts2,int ts3, double** A_trans,int A_size_1, BufferManager *buff_mang_A)
{


  int t1, t2, t3;

 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;

/* Start of CLooG code */
if ((ts1 == 0) && (ts2 >= 0) && (128*ts2 <= n-1) && (ts3 >= 0) && (128*ts3 <= n-1)) {
  increment_ref_count(buff_mang_A, (ts3)* ( A_size_1 ) + (ts2));;
}
/* End of CLooG code */

}

void data_tile_ref_count_init_u1_0(int ts1,int ts2,int ts3, double** u1_trans, BufferManager *buff_mang_u1)
{


  int t1, t2;

 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;

/* Start of CLooG code */
if ((ts1 == 0) && (ts2 >= 0) && (128*ts2 <= n-1) && (ts3 >= 0) && (128*ts3 <= n-1)) {
  increment_ref_count(buff_mang_u1, (ts3));;
}
/* End of CLooG code */

}

void data_tile_ref_count_init_v1_0(int ts1,int ts2,int ts3, double** v1_trans, BufferManager *buff_mang_v1)
{


  int t1, t2;

 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;

/* Start of CLooG code */
if ((ts1 == 0) && (ts2 >= 0) && (128*ts2 <= n-1) && (ts3 >= 0) && (128*ts3 <= n-1)) {
  increment_ref_count(buff_mang_v1, (ts2));;
}
/* End of CLooG code */

}

void data_tile_ref_count_init_u2_0(int ts1,int ts2,int ts3, double** u2_trans, BufferManager *buff_mang_u2)
{


  int t1, t2;

 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;

/* Start of CLooG code */
if ((ts1 == 0) && (ts2 >= 0) && (128*ts2 <= n-1) && (ts3 >= 0) && (128*ts3 <= n-1)) {
  increment_ref_count(buff_mang_u2, (ts3));;
}
/* End of CLooG code */

}

void data_tile_ref_count_init_v2_0(int ts1,int ts2,int ts3, double** v2_trans, BufferManager *buff_mang_v2)
{


  int t1, t2;

 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;

/* Start of CLooG code */
if ((ts1 == 0) && (ts2 >= 0) && (128*ts2 <= n-1) && (ts3 >= 0) && (128*ts3 <= n-1)) {
  increment_ref_count(buff_mang_v2, (ts2));;
}
/* End of CLooG code */

}

void read_in_data_tile_alloc_A_0(int ts1,int ts2,int ts3, double** A_trans,int A_size_1, BufferManager *buff_mang_A)
{


  int t1, t2, t3;

 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;

/* Start of CLooG code */
if ((ts1 == 0) && (ts2 >= 0) && (128*ts2 <= n-1) && (ts3 >= 0) && (128*ts3 <= n-1)) {
  if(A_trans[(ts3) * ((A_size_1)) + (ts2)] == NULL) A_trans[(ts3)* ( A_size_1 ) + (ts2)] = (double *) buffer_alloc_atomic((ts3)* ( A_size_1 ) + (ts2), 128*128 * sizeof(double), buff_mang_A);;
}
/* End of CLooG code */

}

void read_in_copy_A_0(int ts1,int ts2,int ts3, double** A_trans,int A_size_1, BufferManager *buff_mang_A)
{


  int t1, t2, t3;

 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;

/* Start of CLooG code */
if ((ts1 == 0) && (ts2 >= 0) && (128*ts2 <= n-1) && (ts3 >= 0) && (128*ts3 <= n-1)) {
  for (t2=128*ts3;t2<=min(n-1,128*ts3+127);t2++) {
    for (t3=128*ts2;t3<=min(n-1,128*ts2+127);t3++) {
      
#ifndef USE_LOCAL_ARRAYS
 A_trans[((1)*( t2 ))/128* ( A_size_1 ) + ((1)*( t3 ))/128][((1)*( t2 ))%128* (128) + ((1)*( t3 ))%128] = A[t2][t3]; 
#else
 A_trans[((1)*( t2 ))/128* ( A_size_1 ) + ((1)*( t3 ))/128][((1)*( t2 ))%128* (128) + ((1)*( t3 ))%128] = ((double)(t2* t3))/N; 
#endif
;
    }
  }
}
/* End of CLooG code */

}

void read_in_data_tile_alloc_u1_0(int ts1,int ts2,int ts3, double** u1_trans, BufferManager *buff_mang_u1)
{


  int t1, t2;

 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;

/* Start of CLooG code */
if ((ts1 == 0) && (ts2 >= 0) && (128*ts2 <= n-1) && (ts3 >= 0) && (128*ts3 <= n-1)) {
  if(u1_trans[(ts3)] == NULL) u1_trans[(ts3)] = (double *) buffer_alloc_atomic((ts3), 128 * sizeof(double), buff_mang_u1);;
}
/* End of CLooG code */

}

void read_in_copy_u1_0(int ts1,int ts2,int ts3, double** u1_trans, BufferManager *buff_mang_u1)
{


  int t1, t2;

 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;

/* Start of CLooG code */
if ((ts1 == 0) && (ts2 >= 0) && (128*ts2 <= n-1) && (ts3 >= 0) && (128*ts3 <= n-1)) {
  for (t2=128*ts3;t2<=min(n-1,128*ts3+127);t2++) {
    
#ifndef USE_LOCAL_ARRAYS
 u1_trans[((1)*( t2 ))/128][((1)*( t2 ))%128] = u1[t2]; 
#else
 u1_trans[((1)*( t2 ))/128][((1)*( t2 ))%128] = t2; 
#endif
;
  }
}
/* End of CLooG code */

}

void read_in_data_tile_alloc_v1_0(int ts1,int ts2,int ts3, double** v1_trans, BufferManager *buff_mang_v1)
{


  int t1, t2;

 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;

/* Start of CLooG code */
if ((ts1 == 0) && (ts2 >= 0) && (128*ts2 <= n-1) && (ts3 >= 0) && (128*ts3 <= n-1)) {
  if(v1_trans[(ts2)] == NULL) v1_trans[(ts2)] = (double *) buffer_alloc_atomic((ts2), 128 * sizeof(double), buff_mang_v1);;
}
/* End of CLooG code */

}

void read_in_copy_v1_0(int ts1,int ts2,int ts3, double** v1_trans, BufferManager *buff_mang_v1)
{


  int t1, t2;

 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;

/* Start of CLooG code */
if ((ts1 == 0) && (ts2 >= 0) && (128*ts2 <= n-1) && (ts3 >= 0) && (128*ts3 <= n-1)) {
  for (t2=128*ts2;t2<=min(n-1,128*ts2+127);t2++) {
    
#ifndef USE_LOCAL_ARRAYS
 v1_trans[((1)*( t2 ))/128][((1)*( t2 ))%128] = v1[t2]; 
#else
 v1_trans[((1)*( t2 ))/128][((1)*( t2 ))%128] = (t2+1)/N/4.0; 
#endif
;
  }
}
/* End of CLooG code */

}

void read_in_data_tile_alloc_u2_0(int ts1,int ts2,int ts3, double** u2_trans, BufferManager *buff_mang_u2)
{


  int t1, t2;

 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;

/* Start of CLooG code */
if ((ts1 == 0) && (ts2 >= 0) && (128*ts2 <= n-1) && (ts3 >= 0) && (128*ts3 <= n-1)) {
  if(u2_trans[(ts3)] == NULL) u2_trans[(ts3)] = (double *) buffer_alloc_atomic((ts3), 128 * sizeof(double), buff_mang_u2);;
}
/* End of CLooG code */

}

void read_in_copy_u2_0(int ts1,int ts2,int ts3, double** u2_trans, BufferManager *buff_mang_u2)
{


  int t1, t2;

 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;

/* Start of CLooG code */
if ((ts1 == 0) && (ts2 >= 0) && (128*ts2 <= n-1) && (ts3 >= 0) && (128*ts3 <= n-1)) {
  for (t2=128*ts3;t2<=min(n-1,128*ts3+127);t2++) {
    
#ifndef USE_LOCAL_ARRAYS
 u2_trans[((1)*( t2 ))/128][((1)*( t2 ))%128] = u2[t2]; 
#else
 u2_trans[((1)*( t2 ))/128][((1)*( t2 ))%128] = (t2+1)/N/2.0; 
#endif
;
  }
}
/* End of CLooG code */

}

void read_in_data_tile_alloc_v2_0(int ts1,int ts2,int ts3, double** v2_trans, BufferManager *buff_mang_v2)
{


  int t1, t2;

 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;

/* Start of CLooG code */
if ((ts1 == 0) && (ts2 >= 0) && (128*ts2 <= n-1) && (ts3 >= 0) && (128*ts3 <= n-1)) {
  if(v2_trans[(ts2)] == NULL) v2_trans[(ts2)] = (double *) buffer_alloc_atomic((ts2), 128 * sizeof(double), buff_mang_v2);;
}
/* End of CLooG code */

}

void read_in_copy_v2_0(int ts1,int ts2,int ts3, double** v2_trans, BufferManager *buff_mang_v2)
{


  int t1, t2;

 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;

/* Start of CLooG code */
if ((ts1 == 0) && (ts2 >= 0) && (128*ts2 <= n-1) && (ts3 >= 0) && (128*ts3 <= n-1)) {
  for (t2=128*ts2;t2<=min(n-1,128*ts2+127);t2++) {
    
#ifndef USE_LOCAL_ARRAYS
 v2_trans[((1)*( t2 ))/128][((1)*( t2 ))%128] = v2[t2]; 
#else
 v2_trans[((1)*( t2 ))/128][((1)*( t2 ))%128] = (t2+1)/N/6.0; 
#endif
;
  }
}
/* End of CLooG code */

}

void data_tile_ref_count_init_x_1(int ts1,int ts2,int ts3, double** x_trans, BufferManager *buff_mang_x)
{


  int t1, t2;

 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;

/* Start of CLooG code */
if ((ts1 == 1) && (ts2 >= 0) && (128*ts2 <= n-1) && (ts3 >= 0) && (128*ts3 <= n-1)) {
  increment_ref_count(buff_mang_x, (ts2));;
}
/* End of CLooG code */

}

void data_tile_ref_count_init_B_1(int ts1,int ts2,int ts3, double** B_trans,int B_size_1, BufferManager *buff_mang_B)
{


  int t1, t2, t3;

 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;

/* Start of CLooG code */
if ((ts1 == 1) && (ts2 >= 0) && (128*ts2 <= n-1) && (ts3 >= 0) && (128*ts3 <= n-1)) {
  increment_ref_count(buff_mang_B, (ts3)* ( B_size_1 ) + (ts2));;
}
/* End of CLooG code */

}

void data_tile_ref_count_init_y_1(int ts1,int ts2,int ts3, double** y_trans, BufferManager *buff_mang_y)
{


  int t1, t2;

 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;

/* Start of CLooG code */
if ((ts1 == 1) && (ts2 >= 0) && (128*ts2 <= n-1) && (ts3 >= 0) && (128*ts3 <= n-1)) {
  increment_ref_count(buff_mang_y, (ts3));;
}
/* End of CLooG code */

}

void read_in_data_tile_alloc_x_1(int ts1,int ts2,int ts3, double** x_trans, BufferManager *buff_mang_x)
{


  int t1, t2;

 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;

/* Start of CLooG code */
if ((ts1 == 1) && (ts2 >= 0) && (128*ts2 <= n-1) && (ts3 >= 0) && (128*ts3 <= n-1)) {
  if(x_trans[(ts2)] == NULL) x_trans[(ts2)] = (double *) buffer_alloc_atomic((ts2), 128 * sizeof(double), buff_mang_x);;
}
/* End of CLooG code */

}

void read_in_copy_x_1(int ts1,int ts2,int ts3, double** x_trans, BufferManager *buff_mang_x)
{


  int t1, t2;

 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;

/* Start of CLooG code */
if ((ts1 == 1) && (ts2 >= 0) && (128*ts2 <= n-1) && (ts3 >= 0) && (128*ts3 <= n-1)) {
  for (t2=128*ts2;t2<=min(n-1,128*ts2+127);t2++) {
    
#ifndef USE_LOCAL_ARRAYS
 x_trans[((1)*( t2 ))/128][((1)*( t2 ))%128] = x[t2]; 
#else
 x_trans[((1)*( t2 ))/128][((1)*( t2 ))%128] = 0.0; 
#endif
;
  }
}
/* End of CLooG code */

}

void read_in_data_tile_alloc_B_1(int ts1,int ts2,int ts3, double** B_trans,int B_size_1, BufferManager *buff_mang_B)
{


  int t1, t2, t3;

 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;

/* Start of CLooG code */
if ((ts1 == 1) && (ts2 >= 0) && (128*ts2 <= n-1) && (ts3 >= 0) && (128*ts3 <= n-1)) {
  if(B_trans[(ts3) * ((B_size_1)) + (ts2)] == NULL) B_trans[(ts3)* ( B_size_1 ) + (ts2)] = (double *) buffer_alloc_atomic((ts3)* ( B_size_1 ) + (ts2), 128*128 * sizeof(double), buff_mang_B);;
}
/* End of CLooG code */

}

void read_in_copy_B_1(int ts1,int ts2,int ts3, double** B_trans,int B_size_1, BufferManager *buff_mang_B)
{


  int t1, t2, t3;

 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;

/* Start of CLooG code */
if ((ts1 == 1) && (ts2 >= 0) && (128*ts2 <= n-1) && (ts3 >= 0) && (128*ts3 <= n-1)) {
  for (t2=128*ts3;t2<=min(n-1,128*ts3+127);t2++) {
    for (t3=128*ts2;t3<=min(n-1,128*ts2+127);t3++) {
      
#ifndef USE_LOCAL_ARRAYS
 B_trans[((1)*( t2 ))/128* ( B_size_1 ) + ((1)*( t3 ))/128][((1)*( t2 ))%128* (128) + ((1)*( t3 ))%128] = B[t2][t3]; 
#else
 B_trans[((1)*( t2 ))/128* ( B_size_1 ) + ((1)*( t3 ))/128][((1)*( t2 ))%128* (128) + ((1)*( t3 ))%128] = 0.0; 
#endif
;
    }
  }
}
/* End of CLooG code */

}

void read_in_data_tile_alloc_y_1(int ts1,int ts2,int ts3, double** y_trans, BufferManager *buff_mang_y)
{


  int t1, t2;

 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;

/* Start of CLooG code */
if ((ts1 == 1) && (ts2 >= 0) && (128*ts2 <= n-1) && (ts3 >= 0) && (128*ts3 <= n-1)) {
  if(y_trans[(ts3)] == NULL) y_trans[(ts3)] = (double *) buffer_alloc_atomic((ts3), 128 * sizeof(double), buff_mang_y);;
}
/* End of CLooG code */

}

void read_in_copy_y_1(int ts1,int ts2,int ts3, double** y_trans, BufferManager *buff_mang_y)
{


  int t1, t2;

 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;

/* Start of CLooG code */
if ((ts1 == 1) && (ts2 >= 0) && (128*ts2 <= n-1) && (ts3 >= 0) && (128*ts3 <= n-1)) {
  for (t2=128*ts3;t2<=min(n-1,128*ts3+127);t2++) {
    
#ifndef USE_LOCAL_ARRAYS
 y_trans[((1)*( t2 ))/128][((1)*( t2 ))%128] = y[t2]; 
#else
 y_trans[((1)*( t2 ))/128][((1)*( t2 ))%128] = (t2+1)/N/8.0; 
#endif
;
  }
}
/* End of CLooG code */

}

void data_tile_ref_count_init_x_2(int ts1,int ts2, double** x_trans, BufferManager *buff_mang_x)
{


  int t1, t2;

 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;

/* Start of CLooG code */
if ((ts1 == 2) && (ts2 >= 0) && (128*ts2 <= n-1)) {
  increment_ref_count(buff_mang_x, (ts2));;
}
/* End of CLooG code */

}

void data_tile_ref_count_init_z_2(int ts1,int ts2, double** z_trans, BufferManager *buff_mang_z)
{


  int t1, t2;

 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;

/* Start of CLooG code */
if ((ts1 == 2) && (ts2 >= 0) && (128*ts2 <= n-1)) {
  increment_ref_count(buff_mang_z, (ts2));;
}
/* End of CLooG code */

}

void read_in_data_tile_alloc_x_2(int ts1,int ts2, double** x_trans, BufferManager *buff_mang_x)
{


  int t1, t2;

 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;

/* Start of CLooG code */
if ((ts1 == 2) && (ts2 >= 0) && (128*ts2 <= n-1)) {
  if(x_trans[(ts2)] == NULL) x_trans[(ts2)] = (double *) buffer_alloc_atomic((ts2), 128 * sizeof(double), buff_mang_x);;
}
/* End of CLooG code */

}

void read_in_copy_x_2(int ts1,int ts2, double** x_trans, BufferManager *buff_mang_x)
{


  int t1, t2;

 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;

/* Start of CLooG code */
if ((ts1 == 2) && (ts2 >= 0) && (128*ts2 <= n-1)) {
  for (t2=128*ts2;t2<=min(n-1,128*ts2+127);t2++) {
    
#ifndef USE_LOCAL_ARRAYS
 x_trans[((1)*( t2 ))/128][((1)*( t2 ))%128] = x[t2]; 
#else
 x_trans[((1)*( t2 ))/128][((1)*( t2 ))%128] = 0.0; 
#endif
;
  }
}
/* End of CLooG code */

}

void read_in_data_tile_alloc_z_2(int ts1,int ts2, double** z_trans, BufferManager *buff_mang_z)
{


  int t1, t2;

 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;

/* Start of CLooG code */
if ((ts1 == 2) && (ts2 >= 0) && (128*ts2 <= n-1)) {
  if(z_trans[(ts2)] == NULL) z_trans[(ts2)] = (double *) buffer_alloc_atomic((ts2), 128 * sizeof(double), buff_mang_z);;
}
/* End of CLooG code */

}

void read_in_copy_z_2(int ts1,int ts2, double** z_trans, BufferManager *buff_mang_z)
{


  int t1, t2;

 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;

/* Start of CLooG code */
if ((ts1 == 2) && (ts2 >= 0) && (128*ts2 <= n-1)) {
  for (t2=128*ts2;t2<=min(n-1,128*ts2+127);t2++) {
    
#ifndef USE_LOCAL_ARRAYS
 z_trans[((1)*( t2 ))/128][((1)*( t2 ))%128] = z[t2]; 
#else
 z_trans[((1)*( t2 ))/128][((1)*( t2 ))%128] = (t2+1)/N/9.0; 
#endif
;
  }
}
/* End of CLooG code */

}

void data_tile_ref_count_init_w_3(int ts1,int ts2,int ts3, double** w_trans, BufferManager *buff_mang_w)
{


  int t1, t2;

 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;

/* Start of CLooG code */
if ((ts1 == 3) && (ts2 >= 0) && (128*ts2 <= n-1) && (ts3 >= 0) && (128*ts3 <= n-1)) {
  increment_ref_count(buff_mang_w, (ts3));;
}
/* End of CLooG code */

}

void data_tile_ref_count_init_B_3(int ts1,int ts2,int ts3, double** B_trans,int B_size_1, BufferManager *buff_mang_B)
{


  int t1, t2, t3;

 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;

/* Start of CLooG code */
if ((ts1 == 3) && (ts2 >= 0) && (128*ts2 <= n-1) && (ts3 >= 0) && (128*ts3 <= n-1)) {
  increment_ref_count(buff_mang_B, (ts3)* ( B_size_1 ) + (ts2));;
}
/* End of CLooG code */

}

void data_tile_ref_count_init_x_3(int ts1,int ts2,int ts3, double** x_trans, BufferManager *buff_mang_x)
{


  int t1, t2;

 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;

/* Start of CLooG code */
if ((ts1 == 3) && (ts2 >= 0) && (128*ts2 <= n-1) && (ts3 >= 0) && (128*ts3 <= n-1)) {
  increment_ref_count(buff_mang_x, (ts2));;
}
/* End of CLooG code */

}

void read_in_data_tile_alloc_w_3(int ts1,int ts2,int ts3, double** w_trans, BufferManager *buff_mang_w)
{


  int t1, t2;

 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;

/* Start of CLooG code */
if ((ts1 == 3) && (ts2 >= 0) && (128*ts2 <= n-1) && (ts3 >= 0) && (128*ts3 <= n-1)) {
  if(w_trans[(ts3)] == NULL) w_trans[(ts3)] = (double *) buffer_alloc_atomic((ts3), 128 * sizeof(double), buff_mang_w);;
}
/* End of CLooG code */

}

void read_in_copy_w_3(int ts1,int ts2,int ts3, double** w_trans, BufferManager *buff_mang_w)
{


  int t1, t2;

 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;

/* Start of CLooG code */
if ((ts1 == 3) && (ts2 >= 0) && (128*ts2 <= n-1) && (ts3 >= 0) && (128*ts3 <= n-1)) {
  for (t2=128*ts3;t2<=min(n-1,128*ts3+127);t2++) {
    
#ifndef USE_LOCAL_ARRAYS
 w_trans[((1)*( t2 ))/128][((1)*( t2 ))%128] = w[t2]; 
#else
 w_trans[((1)*( t2 ))/128][((1)*( t2 ))%128] = 0.0; 
#endif
;
  }
}
/* End of CLooG code */

}

void read_in_data_tile_alloc_B_3(int ts1,int ts2,int ts3, double** B_trans,int B_size_1, BufferManager *buff_mang_B)
{


  int t1, t2, t3;

 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;

/* Start of CLooG code */
if ((ts1 == 3) && (ts2 >= 0) && (128*ts2 <= n-1) && (ts3 >= 0) && (128*ts3 <= n-1)) {
  if(B_trans[(ts3) * ((B_size_1)) + (ts2)] == NULL) B_trans[(ts3)* ( B_size_1 ) + (ts2)] = (double *) buffer_alloc_atomic((ts3)* ( B_size_1 ) + (ts2), 128*128 * sizeof(double), buff_mang_B);;
}
/* End of CLooG code */

}

void read_in_copy_B_3(int ts1,int ts2,int ts3, double** B_trans,int B_size_1, BufferManager *buff_mang_B)
{


  int t1, t2, t3;

 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;

/* Start of CLooG code */
if ((ts1 == 3) && (ts2 >= 0) && (128*ts2 <= n-1) && (ts3 >= 0) && (128*ts3 <= n-1)) {
  for (t2=128*ts3;t2<=min(n-1,128*ts3+127);t2++) {
    for (t3=128*ts2;t3<=min(n-1,128*ts2+127);t3++) {
      
#ifndef USE_LOCAL_ARRAYS
 B_trans[((1)*( t2 ))/128* ( B_size_1 ) + ((1)*( t3 ))/128][((1)*( t2 ))%128* (128) + ((1)*( t3 ))%128] = B[t2][t3]; 
#else
 B_trans[((1)*( t2 ))/128* ( B_size_1 ) + ((1)*( t3 ))/128][((1)*( t2 ))%128* (128) + ((1)*( t3 ))%128] = 0.0; 
#endif
;
    }
  }
}
/* End of CLooG code */

}

void read_in_data_tile_alloc_x_3(int ts1,int ts2,int ts3, double** x_trans, BufferManager *buff_mang_x)
{


  int t1, t2;

 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;

/* Start of CLooG code */
if ((ts1 == 3) && (ts2 >= 0) && (128*ts2 <= n-1) && (ts3 >= 0) && (128*ts3 <= n-1)) {
  if(x_trans[(ts2)] == NULL) x_trans[(ts2)] = (double *) buffer_alloc_atomic((ts2), 128 * sizeof(double), buff_mang_x);;
}
/* End of CLooG code */

}

void read_in_copy_x_3(int ts1,int ts2,int ts3, double** x_trans, BufferManager *buff_mang_x)
{


  int t1, t2;

 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;

/* Start of CLooG code */
if ((ts1 == 3) && (ts2 >= 0) && (128*ts2 <= n-1) && (ts3 >= 0) && (128*ts3 <= n-1)) {
  for (t2=128*ts2;t2<=min(n-1,128*ts2+127);t2++) {
    
#ifndef USE_LOCAL_ARRAYS
 x_trans[((1)*( t2 ))/128][((1)*( t2 ))%128] = x[t2]; 
#else
 x_trans[((1)*( t2 ))/128][((1)*( t2 ))%128] = 0.0; 
#endif
;
  }
}
/* End of CLooG code */

}

void data_tile_init_0(int my_rank, int nprocs, double** B_trans,int B_size_1, BufferManager *buff_mang_B, double** A_trans,int A_size_1, BufferManager *buff_mang_A, double** u1_trans, BufferManager *buff_mang_u1, double** v1_trans, BufferManager *buff_mang_v1, double** u2_trans, BufferManager *buff_mang_u2, double** v2_trans, BufferManager *buff_mang_v2, double** x_trans, BufferManager *buff_mang_x, double** y_trans, BufferManager *buff_mang_y, double** z_trans, BufferManager *buff_mang_z, double** w_trans, BufferManager *buff_mang_w)
{
  int t1, t2, t3, t4;
 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;
/* Start of CLooG code */
if (n >= 1) {
  for (t2=0;t2<=floord(n-1,128);t2++) {
    w_trans[(t2)] = NULL;;
  }
  for (t2=0;t2<=floord(n-1,128);t2++) {
    z_trans[(t2)] = NULL;;
  }
  for (t2=0;t2<=floord(n-1,128);t2++) {
    y_trans[(t2)] = NULL;;
  }
  for (t2=0;t2<=floord(n-1,128);t2++) {
    x_trans[(t2)] = NULL;;
  }
  for (t2=0;t2<=floord(n-1,128);t2++) {
    v2_trans[(t2)] = NULL;;
  }
  for (t2=0;t2<=floord(n-1,128);t2++) {
    u2_trans[(t2)] = NULL;;
  }
  for (t2=0;t2<=floord(n-1,128);t2++) {
    v1_trans[(t2)] = NULL;;
  }
  for (t2=0;t2<=floord(n-1,128);t2++) {
    u1_trans[(t2)] = NULL;;
  }
  for (t2=0;t2<=floord(n-1,128);t2++) {
    for (t3=0;t3<=floord(n-1,128);t3++) {
      A_trans[(t2)* ( A_size_1 ) + (t3)] = NULL;;
    }
  }
  for (t2=0;t2<=floord(n-1,128);t2++) {
    for (t3=0;t3<=floord(n-1,128);t3++) {
      B_trans[(t2)* ( B_size_1 ) + (t3)] = NULL;;
    }
  }
  for (t3=0;t3<=floord(n-1,128);t3++) {
    for (t4=0;t4<=floord(n-1,128);t4++) {
      if(pi_0(0,t3,t4,N,n,nprocs) == my_rank) data_tile_ref_count_init_A_0(0, t3, t4, A_trans, A_size_1,buff_mang_A);;
    }
  }
  for (t3=0;t3<=floord(n-1,128);t3++) {
    for (t4=0;t4<=floord(n-1,128);t4++) {
      if(pi_0(0,t3,t4,N,n,nprocs) == my_rank) data_tile_ref_count_init_u1_0(0, t3, t4, u1_trans,buff_mang_u1);;
    }
  }
  for (t3=0;t3<=floord(n-1,128);t3++) {
    for (t4=0;t4<=floord(n-1,128);t4++) {
      if(pi_0(0,t3,t4,N,n,nprocs) == my_rank) data_tile_ref_count_init_v1_0(0, t3, t4, v1_trans,buff_mang_v1);;
    }
  }
  for (t3=0;t3<=floord(n-1,128);t3++) {
    for (t4=0;t4<=floord(n-1,128);t4++) {
      if(pi_0(0,t3,t4,N,n,nprocs) == my_rank) data_tile_ref_count_init_u2_0(0, t3, t4, u2_trans,buff_mang_u2);;
    }
  }
  for (t3=0;t3<=floord(n-1,128);t3++) {
    for (t4=0;t4<=floord(n-1,128);t4++) {
      if(pi_0(0,t3,t4,N,n,nprocs) == my_rank) data_tile_ref_count_init_v2_0(0, t3, t4, v2_trans,buff_mang_v2);;
    }
  }
  for (t3=0;t3<=floord(n-1,128);t3++) {
    for (t4=0;t4<=floord(n-1,128);t4++) {
      if(pi_0(0,t3,t4,N,n,nprocs) == my_rank) data_tile_ref_count_init_B_0(0, t3, t4, B_trans, B_size_1,buff_mang_B);;
    }
  }
  for (t3=0;t3<=floord(n-1,128);t3++) {
    for (t4=0;t4<=floord(n-1,128);t4++) {
      if(pi_0(0,t3,t4,N,n,nprocs) == my_rank) data_tile_ref_count_init_B_0(0, t3, t4, B_trans, B_size_1,buff_mang_B);;
    }
  }
  for (t3=0;t3<=floord(n-1,128);t3++) {
    for (t4=0;t4<=floord(n-1,128);t4++) {
      if(pi_0(0,t3,t4,N,n,nprocs) == my_rank){ read_in_data_tile_alloc_A_0(0, t3, t4, A_trans, A_size_1,buff_mang_A);read_in_copy_A_0(0, t3, t4, A_trans, A_size_1,buff_mang_A);};
    }
  }
  for (t3=0;t3<=floord(n-1,128);t3++) {
    for (t4=0;t4<=floord(n-1,128);t4++) {
      if(pi_0(0,t3,t4,N,n,nprocs) == my_rank){ read_in_data_tile_alloc_u1_0(0, t3, t4, u1_trans,buff_mang_u1);read_in_copy_u1_0(0, t3, t4, u1_trans,buff_mang_u1);};
    }
  }
  for (t3=0;t3<=floord(n-1,128);t3++) {
    for (t4=0;t4<=floord(n-1,128);t4++) {
      if(pi_0(0,t3,t4,N,n,nprocs) == my_rank){ read_in_data_tile_alloc_v1_0(0, t3, t4, v1_trans,buff_mang_v1);read_in_copy_v1_0(0, t3, t4, v1_trans,buff_mang_v1);};
    }
  }
  for (t3=0;t3<=floord(n-1,128);t3++) {
    for (t4=0;t4<=floord(n-1,128);t4++) {
      if(pi_0(0,t3,t4,N,n,nprocs) == my_rank){ read_in_data_tile_alloc_u2_0(0, t3, t4, u2_trans,buff_mang_u2);read_in_copy_u2_0(0, t3, t4, u2_trans,buff_mang_u2);};
    }
  }
  for (t3=0;t3<=floord(n-1,128);t3++) {
    for (t4=0;t4<=floord(n-1,128);t4++) {
      if(pi_0(0,t3,t4,N,n,nprocs) == my_rank){ read_in_data_tile_alloc_v2_0(0, t3, t4, v2_trans,buff_mang_v2);read_in_copy_v2_0(0, t3, t4, v2_trans,buff_mang_v2);};
    }
  }
  for (t3=0;t3<=floord(n-1,128);t3++) {
    for (t4=0;t4<=floord(n-1,128);t4++) {
      if(pi_1(1,t3,t4,N,n,nprocs) == my_rank) data_tile_ref_count_init_x_1(1, t3, t4, x_trans,buff_mang_x);;
    }
  }
  for (t3=0;t3<=floord(n-1,128);t3++) {
    for (t4=0;t4<=floord(n-1,128);t4++) {
      if(pi_1(1,t3,t4,N,n,nprocs) == my_rank) data_tile_ref_count_init_B_1(1, t3, t4, B_trans, B_size_1,buff_mang_B);;
    }
  }
  for (t3=0;t3<=floord(n-1,128);t3++) {
    for (t4=0;t4<=floord(n-1,128);t4++) {
      if(pi_1(1,t3,t4,N,n,nprocs) == my_rank) data_tile_ref_count_init_y_1(1, t3, t4, y_trans,buff_mang_y);;
    }
  }
  for (t3=0;t3<=floord(n-1,128);t3++) {
    for (t4=0;t4<=floord(n-1,128);t4++) {
      if(pi_1(1,t3,t4,N,n,nprocs) == my_rank) data_tile_ref_count_init_x_1(1, t3, t4, x_trans,buff_mang_x);;
    }
  }
  for (t3=0;t3<=floord(n-1,128);t3++) {
    for (t4=0;t4<=floord(n-1,128);t4++) {
      if(pi_1(1,t3,t4,N,n,nprocs) == my_rank){ read_in_data_tile_alloc_x_1(1, t3, t4, x_trans,buff_mang_x);read_in_copy_x_1(1, t3, t4, x_trans,buff_mang_x);};
    }
  }
  for (t3=0;t3<=floord(n-1,128);t3++) {
    for (t4=0;t4<=floord(n-1,128);t4++) {
      if(pi_1(1,t3,t4,N,n,nprocs) == my_rank){ read_in_data_tile_alloc_B_1(1, t3, t4, B_trans, B_size_1,buff_mang_B);read_in_copy_B_1(1, t3, t4, B_trans, B_size_1,buff_mang_B);};
    }
  }
  for (t3=0;t3<=floord(n-1,128);t3++) {
    for (t4=0;t4<=floord(n-1,128);t4++) {
      if(pi_1(1,t3,t4,N,n,nprocs) == my_rank){ read_in_data_tile_alloc_y_1(1, t3, t4, y_trans,buff_mang_y);read_in_copy_y_1(1, t3, t4, y_trans,buff_mang_y);};
    }
  }
  for (t3=0;t3<=floord(n-1,128);t3++) {
    if(pi_2(2,t3,N,n,nprocs) == my_rank) data_tile_ref_count_init_x_2(2, t3, x_trans,buff_mang_x);;
  }
  for (t3=0;t3<=floord(n-1,128);t3++) {
    if(pi_2(2,t3,N,n,nprocs) == my_rank) data_tile_ref_count_init_z_2(2, t3, z_trans,buff_mang_z);;
  }
  for (t3=0;t3<=floord(n-1,128);t3++) {
    if(pi_2(2,t3,N,n,nprocs) == my_rank) data_tile_ref_count_init_x_2(2, t3, x_trans,buff_mang_x);;
  }
  for (t3=0;t3<=floord(n-1,128);t3++) {
    if(pi_2(2,t3,N,n,nprocs) == my_rank){ read_in_data_tile_alloc_x_2(2, t3, x_trans,buff_mang_x);read_in_copy_x_2(2, t3, x_trans,buff_mang_x);};
  }
  for (t3=0;t3<=floord(n-1,128);t3++) {
    if(pi_2(2,t3,N,n,nprocs) == my_rank){ read_in_data_tile_alloc_z_2(2, t3, z_trans,buff_mang_z);read_in_copy_z_2(2, t3, z_trans,buff_mang_z);};
  }
  for (t3=0;t3<=floord(n-1,128);t3++) {
    for (t4=0;t4<=floord(n-1,128);t4++) {
      if(pi_3(3,t3,t4,N,n,nprocs) == my_rank) data_tile_ref_count_init_w_3(3, t3, t4, w_trans,buff_mang_w);;
    }
  }
  for (t3=0;t3<=floord(n-1,128);t3++) {
    for (t4=0;t4<=floord(n-1,128);t4++) {
      if(pi_3(3,t3,t4,N,n,nprocs) == my_rank) data_tile_ref_count_init_B_3(3, t3, t4, B_trans, B_size_1,buff_mang_B);;
    }
  }
  for (t3=0;t3<=floord(n-1,128);t3++) {
    for (t4=0;t4<=floord(n-1,128);t4++) {
      if(pi_3(3,t3,t4,N,n,nprocs) == my_rank) data_tile_ref_count_init_x_3(3, t3, t4, x_trans,buff_mang_x);;
    }
  }
  for (t3=0;t3<=floord(n-1,128);t3++) {
    for (t4=0;t4<=floord(n-1,128);t4++) {
      if(pi_3(3,t3,t4,N,n,nprocs) == my_rank) data_tile_ref_count_init_w_3(3, t3, t4, w_trans,buff_mang_w);;
    }
  }
  for (t3=0;t3<=floord(n-1,128);t3++) {
    for (t4=0;t4<=floord(n-1,128);t4++) {
      if(pi_3(3,t3,t4,N,n,nprocs) == my_rank){ read_in_data_tile_alloc_w_3(3, t3, t4, w_trans,buff_mang_w);read_in_copy_w_3(3, t3, t4, w_trans,buff_mang_w);};
    }
  }
  for (t3=0;t3<=floord(n-1,128);t3++) {
    for (t4=0;t4<=floord(n-1,128);t4++) {
      if(pi_3(3,t3,t4,N,n,nprocs) == my_rank){ read_in_data_tile_alloc_B_3(3, t3, t4, B_trans, B_size_1,buff_mang_B);read_in_copy_B_3(3, t3, t4, B_trans, B_size_1,buff_mang_B);};
    }
  }
  for (t3=0;t3<=floord(n-1,128);t3++) {
    for (t4=0;t4<=floord(n-1,128);t4++) {
      if(pi_3(3,t3,t4,N,n,nprocs) == my_rank){ read_in_data_tile_alloc_x_3(3, t3, t4, x_trans,buff_mang_x);read_in_copy_x_3(3, t3, t4, x_trans,buff_mang_x);};
    }
  }
}
/* End of CLooG code */
}
int pack_B_1_0(int ts1,int ts2,int ts3,double *send_buf_B,int send_counts_B, double** B_trans,int B_size_1, BufferManager *buff_mang_B){
int recv_proc;
  int t1, t2, t3, t4;
 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;
/* Start of CLooG code */
if ((ts1 == 0) && (ts2 >= 0) && (128*ts2 <= n-1) && (ts3 >= 0) && (128*ts3 <= n-1)) {
  if(B_trans[(ts3) * ((B_size_1)) + (ts2)] == NULL) B_trans[(ts3)* ( B_size_1 ) + (ts2)] = (double *) buffer_alloc_atomic((ts3)* ( B_size_1 ) + (ts2), 128*128 * sizeof(double), buff_mang_B);;
  for (t1=128*ts3+2;t1<=min(n+1,128*ts3+129);t1++) {
    for (t2=128*ts2;t2<=min(n-1,128*ts2+127);t2++) {
      send_buf_B[send_counts_B++] = B_trans[((1)*( t1-2 ))/128* ( B_size_1 ) + ((1)*( t2 ))/128][((1)*( t1-2 ))%128* (128) + ((1)*( t2 ))%128];
    }
  }
}
/* End of CLooG code */
return send_counts_B;
}
int unpack_B_1_0(int ts1,int ts2,int ts3,double *recv_buf_B,int curr_displs_B, double** B_trans,int B_size_1, BufferManager *buff_mang_B){
int recv_proc;
  int t1, t2, t3, t4;
 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;
/* Start of CLooG code */
if ((ts1 == 0) && (ts2 >= 0) && (128*ts2 <= n-1) && (ts3 >= 0) && (128*ts3 <= n-1)) {
  if(B_trans[(ts3) * ((B_size_1)) + (ts2)] == NULL) B_trans[(ts3)* ( B_size_1 ) + (ts2)] = (double *) buffer_alloc_atomic((ts3)* ( B_size_1 ) + (ts2), 128*128 * sizeof(double), buff_mang_B);;
  for (t1=128*ts3+2;t1<=min(n+1,128*ts3+129);t1++) {
    for (t2=128*ts2;t2<=min(n-1,128*ts2+127);t2++) {
      B_trans[((1)*( t1-2 ))/128* ( B_size_1 ) + ((1)*( t2 ))/128][((1)*( t1-2 ))%128* (128) + ((1)*( t2 ))%128] = recv_buf_B[curr_displs_B++];
    }
  }
}
/* End of CLooG code */
return curr_displs_B;
}
void data_tile_copy_back_B_0(int ts1,int ts2,int ts3, double** B_trans,int B_size_1, BufferManager *buff_mang_B)
{
#ifndef USE_LOCAL_ARRAYS
  int t1, t2, t3;
 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;
/* Start of CLooG code */
if ((ts1 == 0) && (ts2 >= 0) && (128*ts2 <= n-1) && (ts3 >= 0) && (128*ts3 <= n-1)) {
  for (t2=128*ts3;t2<=min(n-1,128*ts3+127);t2++) {
    for (t3=128*ts2;t3<=min(n-1,128*ts2+127);t3++) {
      B[t2][t3] = B_trans[((1)*( t2 ))/128* ( B_size_1 ) + ((1)*( t3 ))/128][((1)*( t2 ))%128* (128) + ((1)*( t3 ))%128];;
    }
  }
}
/* End of CLooG code */
#endif
}
int writeout_pack_B_0(int ts1,int ts2,int ts3,double *lw_buf_B,int lw_count_B, double** B_trans,int B_size_1, BufferManager *buff_mang_B){
int recv_proc;
  int t1, t2, t3, t4;
 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;
/* Start of CLooG code */
if ((ts1 == 0) && (ts2 >= 0) && (128*ts2 <= n-1) && (ts3 >= 0) && (128*ts3 <= n-1)) {
  if(B_trans[(ts3) * ((B_size_1)) + (ts2)] == NULL) B_trans[(ts3)* ( B_size_1 ) + (ts2)] = (double *) buffer_alloc_atomic((ts3)* ( B_size_1 ) + (ts2), 128*128 * sizeof(double), buff_mang_B);;
  for (t1=128*ts3+2;t1<=min(n+1,128*ts3+129);t1++) {
    for (t2=128*ts2;t2<=min(n-1,128*ts2+127);t2++) {
      lw_buf_B[lw_count_B++] = B_trans[((1)*( t1-2 ))/128* ( B_size_1 ) + ((1)*( t2 ))/128][((1)*( t1-2 ))%128* (128) + ((1)*( t2 ))%128];
    }
  }
}
/* End of CLooG code */
return lw_count_B;
}
int writeout_unpack_B_0(int ts1,int ts2,int ts3,double *lw_recv_buf_B,int displs_lw_B, double** B_trans,int B_size_1, BufferManager *buff_mang_B){
int recv_proc;
  int t1, t2, t3;
 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;
/* Start of CLooG code */
if ((ts1 == 0) && (ts2 >= 0) && (128*ts2 <= n-1) && (ts3 >= 0) && (128*ts3 <= n-1)) {
  for (t1=128*ts3;t1<=min(n-1,128*ts3+127);t1++) {
    for (t2=128*ts2;t2<=min(n-1,128*ts2+127);t2++) {
      
#ifndef USE_LOCAL_ARRAYS
 B[t1][t2] = lw_recv_buf_B[displs_lw_B++]; 
#endif
;
    }
  }
}
/* End of CLooG code */
return displs_lw_B;
}
void pack_send_0(int my_rank,int nprocs,int t1,int t2,int t3, double** B_trans,int B_size_1, BufferManager *buff_mang_B,double **send_buf_B,MPI_Request *send_reqs_B,MPI_Status *send_stats_B, double** A_trans,int A_size_1, BufferManager *buff_mang_A,double **send_buf_A,MPI_Request *send_reqs_A,MPI_Status *send_stats_A, double** u1_trans, BufferManager *buff_mang_u1,double **send_buf_u1,MPI_Request *send_reqs_u1,MPI_Status *send_stats_u1, double** v1_trans, BufferManager *buff_mang_v1,double **send_buf_v1,MPI_Request *send_reqs_v1,MPI_Status *send_stats_v1, double** u2_trans, BufferManager *buff_mang_u2,double **send_buf_u2,MPI_Request *send_reqs_u2,MPI_Status *send_stats_u2, double** v2_trans, BufferManager *buff_mang_v2,double **send_buf_v2,MPI_Request *send_reqs_v2,MPI_Status *send_stats_v2,int max_num_sends,double *p_total_count)
{
int __p, receiver_list[nprocs];
int send_counts_A[nprocs];
for (__p=0;__p<nprocs;__p++) send_counts_A[__p] = 0;
int index_A[nprocs];
int current_index_A = 0;
int send_counts_u1[nprocs];
for (__p=0;__p<nprocs;__p++) send_counts_u1[__p] = 0;
int index_u1[nprocs];
int current_index_u1 = 0;
int send_counts_v1[nprocs];
for (__p=0;__p<nprocs;__p++) send_counts_v1[__p] = 0;
int index_v1[nprocs];
int current_index_v1 = 0;
int send_counts_u2[nprocs];
for (__p=0;__p<nprocs;__p++) send_counts_u2[__p] = 0;
int index_u2[nprocs];
int current_index_u2 = 0;
int send_counts_v2[nprocs];
for (__p=0;__p<nprocs;__p++) send_counts_v2[__p] = 0;
int index_v2[nprocs];
int current_index_v2 = 0;
int send_counts_B[nprocs];
for (__p=0;__p<nprocs;__p++) send_counts_B[__p] = 0;
int index_B[nprocs];
int current_index_B = 0;
int _i;
double __total_count = 0;
;
    for (__p=0; __p<nprocs; __p++) { if(send_counts_A[__p] > 0) {send_buf_A[index_A[__p]][0] = 0;send_buf_A[index_A[__p]][1] = t1;send_buf_A[index_A[__p]][2] = t2;send_buf_A[index_A[__p]][3] = t3; IF_DYNSCHEDULER_DEBUG_PRINT( fprintf(__debug_print_fp, "sending node %d loop 0 task %d %d %d 		        size %d to node %d\n", my_rank, t1,t2,t3, send_counts_A[__p], __p)); IF_DEBUG_FLUSH(fflush(__debug_print_fp));IF_TIME(__total_count += send_counts_A[__p]); MPI_Isend(send_buf_A[index_A[__p]], send_counts_A[__p], MPI_DOUBLE, __p, 14123, MPI_COMM_WORLD, &send_reqs_A[index_A[__p]]); send_counts_A[__p] = 0;}};
;
    for (__p=0; __p<nprocs; __p++) { if(send_counts_u1[__p] > 0) {send_buf_u1[index_u1[__p]][0] = 0;send_buf_u1[index_u1[__p]][1] = t1;send_buf_u1[index_u1[__p]][2] = t2;send_buf_u1[index_u1[__p]][3] = t3; IF_DYNSCHEDULER_DEBUG_PRINT( fprintf(__debug_print_fp, "sending node %d loop 0 task %d %d %d 		        size %d to node %d\n", my_rank, t1,t2,t3, send_counts_u1[__p], __p)); IF_DEBUG_FLUSH(fflush(__debug_print_fp));IF_TIME(__total_count += send_counts_u1[__p]); MPI_Isend(send_buf_u1[index_u1[__p]], send_counts_u1[__p], MPI_DOUBLE, __p, 11123, MPI_COMM_WORLD, &send_reqs_u1[index_u1[__p]]); send_counts_u1[__p] = 0;}};
;
    for (__p=0; __p<nprocs; __p++) { if(send_counts_v1[__p] > 0) {send_buf_v1[index_v1[__p]][0] = 0;send_buf_v1[index_v1[__p]][1] = t1;send_buf_v1[index_v1[__p]][2] = t2;send_buf_v1[index_v1[__p]][3] = t3; IF_DYNSCHEDULER_DEBUG_PRINT( fprintf(__debug_print_fp, "sending node %d loop 0 task %d %d %d 		        size %d to node %d\n", my_rank, t1,t2,t3, send_counts_v1[__p], __p)); IF_DEBUG_FLUSH(fflush(__debug_print_fp));IF_TIME(__total_count += send_counts_v1[__p]); MPI_Isend(send_buf_v1[index_v1[__p]], send_counts_v1[__p], MPI_DOUBLE, __p, 8123, MPI_COMM_WORLD, &send_reqs_v1[index_v1[__p]]); send_counts_v1[__p] = 0;}};
;
    for (__p=0; __p<nprocs; __p++) { if(send_counts_u2[__p] > 0) {send_buf_u2[index_u2[__p]][0] = 0;send_buf_u2[index_u2[__p]][1] = t1;send_buf_u2[index_u2[__p]][2] = t2;send_buf_u2[index_u2[__p]][3] = t3; IF_DYNSCHEDULER_DEBUG_PRINT( fprintf(__debug_print_fp, "sending node %d loop 0 task %d %d %d 		        size %d to node %d\n", my_rank, t1,t2,t3, send_counts_u2[__p], __p)); IF_DEBUG_FLUSH(fflush(__debug_print_fp));IF_TIME(__total_count += send_counts_u2[__p]); MPI_Isend(send_buf_u2[index_u2[__p]], send_counts_u2[__p], MPI_DOUBLE, __p, 10123, MPI_COMM_WORLD, &send_reqs_u2[index_u2[__p]]); send_counts_u2[__p] = 0;}};
;
    for (__p=0; __p<nprocs; __p++) { if(send_counts_v2[__p] > 0) {send_buf_v2[index_v2[__p]][0] = 0;send_buf_v2[index_v2[__p]][1] = t1;send_buf_v2[index_v2[__p]][2] = t2;send_buf_v2[index_v2[__p]][3] = t3; IF_DYNSCHEDULER_DEBUG_PRINT( fprintf(__debug_print_fp, "sending node %d loop 0 task %d %d %d 		        size %d to node %d\n", my_rank, t1,t2,t3, send_counts_v2[__p], __p)); IF_DEBUG_FLUSH(fflush(__debug_print_fp));IF_TIME(__total_count += send_counts_v2[__p]); MPI_Isend(send_buf_v2[index_v2[__p]], send_counts_v2[__p], MPI_DOUBLE, __p, 9123, MPI_COMM_WORLD, &send_reqs_v2[index_v2[__p]]); send_counts_v2[__p] = 0;}};
for (__p=0; __p<nprocs; __p++) { receiver_list[__p] = 0; } sigma_B_1_0(t1,t2,t3,N,n, my_rank, nprocs, receiver_list); for (__p=0; __p<nprocs; __p++) { if (receiver_list[__p] != 0) { if (send_counts_B[__p] == 0) { for (;current_index_B<max_num_sends;current_index_B++) { if (send_reqs_B[current_index_B] == MPI_REQUEST_NULL) { _i = current_index_B; break; } } if (current_index_B<max_num_sends) { current_index_B++; } else { MPI_Waitany(max_num_sends, send_reqs_B, &_i, &send_stats_B[__p]); assert(_i != MPI_UNDEFINED); } index_B[__p] = _i; send_counts_B[__p] = 4; }send_counts_B[__p] = pack_B_1_0(t1,t2,t3, send_buf_B[index_B[__p]], send_counts_B[__p], B_trans, B_size_1,buff_mang_B); } };
    for (__p=0; __p<nprocs; __p++) { if(send_counts_B[__p] > 0) {send_buf_B[index_B[__p]][0] = 0;send_buf_B[index_B[__p]][1] = t1;send_buf_B[index_B[__p]][2] = t2;send_buf_B[index_B[__p]][3] = t3; IF_DYNSCHEDULER_DEBUG_PRINT( fprintf(__debug_print_fp, "sending node %d loop 0 task %d %d %d 		        size %d to node %d\n", my_rank, t1,t2,t3, send_counts_B[__p], __p)); IF_DEBUG_FLUSH(fflush(__debug_print_fp));IF_TIME(__total_count += send_counts_B[__p]); MPI_Isend(send_buf_B[index_B[__p]], send_counts_B[__p], MPI_DOUBLE, __p, 13123, MPI_COMM_WORLD, &send_reqs_B[index_B[__p]]); send_counts_B[__p] = 0;}};
*p_total_count += __total_count;
}
void unpack_buf_A_0(int my_rank,int nprocs, double** A_trans,int A_size_1, BufferManager *buff_mang_A,double *recv_buf_A)
{
int curr_displs_A;
curr_displs_A = 4; ;
}
void unpack_buf_u1_0(int my_rank,int nprocs, double** u1_trans, BufferManager *buff_mang_u1,double *recv_buf_u1)
{
int curr_displs_u1;
curr_displs_u1 = 4; ;
}
void unpack_buf_v1_0(int my_rank,int nprocs, double** v1_trans, BufferManager *buff_mang_v1,double *recv_buf_v1)
{
int curr_displs_v1;
curr_displs_v1 = 4; ;
}
void unpack_buf_u2_0(int my_rank,int nprocs, double** u2_trans, BufferManager *buff_mang_u2,double *recv_buf_u2)
{
int curr_displs_u2;
curr_displs_u2 = 4; ;
}
void unpack_buf_v2_0(int my_rank,int nprocs, double** v2_trans, BufferManager *buff_mang_v2,double *recv_buf_v2)
{
int curr_displs_v2;
curr_displs_v2 = 4; ;
}
void unpack_buf_B_0(int my_rank,int nprocs, double** B_trans,int B_size_1, BufferManager *buff_mang_B,double *recv_buf_B)
{
int curr_displs_B;
curr_displs_B = 4; if (is_receiver_B_1_0((int)recv_buf_B[1],(int)recv_buf_B[2],(int)recv_buf_B[3],N,n,my_rank,nprocs) != 0) { curr_displs_B = unpack_B_1_0((int)recv_buf_B[1],(int)recv_buf_B[2],(int)recv_buf_B[3],recv_buf_B,curr_displs_B, B_trans, B_size_1,buff_mang_B); };
}
int pack_x_1_1(int ts1,int ts2,int ts3,double *send_buf_x,int send_counts_x, double** x_trans, BufferManager *buff_mang_x){
int recv_proc;
  int t1, t2, t3, t4;
 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;
/* Start of CLooG code */
if ((ts1 == 1) && (ts2 >= ts3+1) && (128*ts2 <= n-1) && (ts3 >= 0)) {
  if(x_trans[(ts2)] == NULL) x_trans[(ts2)] = (double *) buffer_alloc_atomic((ts2), 128 * sizeof(double), buff_mang_x);;
  for (t1=128*ts2+2;t1<=min(n+1,128*ts2+129);t1++) {
    send_buf_x[send_counts_x++] = x_trans[((1)*( t1-2 ))/128][((1)*( t1-2 ))%128];
  }
}
/* End of CLooG code */
return send_counts_x;
}
int unpack_x_1_1(int ts1,int ts2,int ts3,double *recv_buf_x,int curr_displs_x, double** x_trans, BufferManager *buff_mang_x){
int recv_proc;
  int t1, t2, t3, t4;
 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;
/* Start of CLooG code */
if ((ts1 == 1) && (ts2 >= ts3+1) && (128*ts2 <= n-1) && (ts3 >= 0)) {
  if(x_trans[(ts2)] == NULL) x_trans[(ts2)] = (double *) buffer_alloc_atomic((ts2), 128 * sizeof(double), buff_mang_x);;
  for (t1=128*ts2+2;t1<=min(n+1,128*ts2+129);t1++) {
    x_trans[((1)*( t1-2 ))/128][((1)*( t1-2 ))%128] = recv_buf_x[curr_displs_x++];
  }
}
/* End of CLooG code */
return curr_displs_x;
}
int pack_x_2_1(int ts1,int ts2,int ts3,double *send_buf_x,int send_counts_x, double** x_trans, BufferManager *buff_mang_x){
int recv_proc;
  int t1, t2, t3, t4;
 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;
/* Start of CLooG code */
if ((ts1 == 1) && (ts2 >= 0) && (ts2 == ts3) && (128*ts2 <= n-1)) {
  if(x_trans[(ts2)] == NULL) x_trans[(ts2)] = (double *) buffer_alloc_atomic((ts2), 128 * sizeof(double), buff_mang_x);;
  for (t1=128*ts2+2;t1<=min(n+1,128*ts2+129);t1++) {
    send_buf_x[send_counts_x++] = x_trans[((1)*( t1-2 ))/128][((1)*( t1-2 ))%128];
  }
}
/* End of CLooG code */
return send_counts_x;
}
int unpack_x_2_1(int ts1,int ts2,int ts3,double *recv_buf_x,int curr_displs_x, double** x_trans, BufferManager *buff_mang_x){
int recv_proc;
  int t1, t2, t3, t4;
 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;
/* Start of CLooG code */
if ((ts1 == 1) && (ts2 >= 0) && (ts2 == ts3) && (128*ts2 <= n-1)) {
  if(x_trans[(ts2)] == NULL) x_trans[(ts2)] = (double *) buffer_alloc_atomic((ts2), 128 * sizeof(double), buff_mang_x);;
  for (t1=128*ts2+2;t1<=min(n+1,128*ts2+129);t1++) {
    x_trans[((1)*( t1-2 ))/128][((1)*( t1-2 ))%128] = recv_buf_x[curr_displs_x++];
  }
}
/* End of CLooG code */
return curr_displs_x;
}
int pack_x_3_1(int ts1,int ts2,int ts3,double *send_buf_x,int send_counts_x, double** x_trans, BufferManager *buff_mang_x){
int recv_proc;
  int t1, t2, t3, t4;
 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;
/* Start of CLooG code */
if ((ts1 == 1) && (ts2 >= 0) && (ts2 <= ts3-1) && (128*ts3 <= n-1)) {
  if(x_trans[(ts2)] == NULL) x_trans[(ts2)] = (double *) buffer_alloc_atomic((ts2), 128 * sizeof(double), buff_mang_x);;
  for (t1=128*ts2+2;t1<=128*ts2+129;t1++) {
    send_buf_x[send_counts_x++] = x_trans[((1)*( t1-2 ))/128][((1)*( t1-2 ))%128];
  }
}
/* End of CLooG code */
return send_counts_x;
}
int unpack_x_3_1(int ts1,int ts2,int ts3,double *recv_buf_x,int curr_displs_x, double** x_trans, BufferManager *buff_mang_x){
int recv_proc;
  int t1, t2, t3, t4;
 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;
/* Start of CLooG code */
if ((ts1 == 1) && (ts2 >= 0) && (ts2 <= ts3-1) && (128*ts3 <= n-1)) {
  if(x_trans[(ts2)] == NULL) x_trans[(ts2)] = (double *) buffer_alloc_atomic((ts2), 128 * sizeof(double), buff_mang_x);;
  for (t1=128*ts2+2;t1<=128*ts2+129;t1++) {
    x_trans[((1)*( t1-2 ))/128][((1)*( t1-2 ))%128] = recv_buf_x[curr_displs_x++];
  }
}
/* End of CLooG code */
return curr_displs_x;
}
void data_tile_copy_back_x_1(int ts1,int ts2,int ts3, double** x_trans, BufferManager *buff_mang_x)
{
#ifndef USE_LOCAL_ARRAYS
  int t1, t2;
 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;
/* Start of CLooG code */
if ((ts1 == 1) && (ts2 >= 0) && (128*ts2 <= n-1) && (ts3 >= 0) && (128*ts3 <= n-1)) {
  for (t2=128*ts2;t2<=min(n-1,128*ts2+127);t2++) {
    x[t2] = x_trans[((1)*( t2 ))/128][((1)*( t2 ))%128];;
  }
}
/* End of CLooG code */
#endif
}
int writeout_pack_x_1(int ts1,int ts2,int ts3,double *lw_buf_x,int lw_count_x, double** x_trans, BufferManager *buff_mang_x){
int recv_proc;
  int t1, t2, t3, t4;
 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;
/* Start of CLooG code */
/* End of CLooG code */
return lw_count_x;
}
int writeout_unpack_x_1(int ts1,int ts2,int ts3,double *lw_recv_buf_x,int displs_lw_x, double** x_trans, BufferManager *buff_mang_x){
int recv_proc;
  int t1, t2, t3;
 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;
/* Start of CLooG code */
/* End of CLooG code */
return displs_lw_x;
}
void pack_send_1(int my_rank,int nprocs,int t1,int t2,int t3, double** x_trans, BufferManager *buff_mang_x,double **send_buf_x,MPI_Request *send_reqs_x,MPI_Status *send_stats_x, double** B_trans,int B_size_1, BufferManager *buff_mang_B,double **send_buf_B,MPI_Request *send_reqs_B,MPI_Status *send_stats_B, double** y_trans, BufferManager *buff_mang_y,double **send_buf_y,MPI_Request *send_reqs_y,MPI_Status *send_stats_y,int max_num_sends,double *p_total_count)
{
int __p, receiver_list[nprocs];
int send_counts_x[nprocs];
for (__p=0;__p<nprocs;__p++) send_counts_x[__p] = 0;
int index_x[nprocs];
int current_index_x = 0;
int send_counts_B[nprocs];
for (__p=0;__p<nprocs;__p++) send_counts_B[__p] = 0;
int index_B[nprocs];
int current_index_B = 0;
int send_counts_y[nprocs];
for (__p=0;__p<nprocs;__p++) send_counts_y[__p] = 0;
int index_y[nprocs];
int current_index_y = 0;
int _i;
double __total_count = 0;
for (__p=0; __p<nprocs; __p++) { receiver_list[__p] = 0; } sigma_x_1_1(t1,t2,t3,N,n, my_rank, nprocs, receiver_list); for (__p=0; __p<nprocs; __p++) { if (receiver_list[__p] != 0) { if (send_counts_x[__p] == 0) { for (;current_index_x<max_num_sends;current_index_x++) { if (send_reqs_x[current_index_x] == MPI_REQUEST_NULL) { _i = current_index_x; break; } } if (current_index_x<max_num_sends) { current_index_x++; } else { MPI_Waitany(max_num_sends, send_reqs_x, &_i, &send_stats_x[__p]); assert(_i != MPI_UNDEFINED); } index_x[__p] = _i; send_counts_x[__p] = 4; }send_counts_x[__p] = pack_x_1_1(t1,t2,t3, send_buf_x[index_x[__p]], send_counts_x[__p], x_trans,buff_mang_x); } }for (__p=0; __p<nprocs; __p++) { receiver_list[__p] = 0; } sigma_x_2_1(t1,t2,t3,N,n, my_rank, nprocs, receiver_list); for (__p=0; __p<nprocs; __p++) { if (receiver_list[__p] != 0) { if (send_counts_x[__p] == 0) { for (;current_index_x<max_num_sends;current_index_x++) { if (send_reqs_x[current_index_x] == MPI_REQUEST_NULL) { _i = current_index_x; break; } } if (current_index_x<max_num_sends) { current_index_x++; } else { MPI_Waitany(max_num_sends, send_reqs_x, &_i, &send_stats_x[__p]); assert(_i != MPI_UNDEFINED); } index_x[__p] = _i; send_counts_x[__p] = 4; }send_counts_x[__p] = pack_x_2_1(t1,t2,t3, send_buf_x[index_x[__p]], send_counts_x[__p], x_trans,buff_mang_x); } }for (__p=0; __p<nprocs; __p++) { receiver_list[__p] = 0; } sigma_x_3_1(t1,t2,t3,N,n, my_rank, nprocs, receiver_list); for (__p=0; __p<nprocs; __p++) { if (receiver_list[__p] != 0) { if (send_counts_x[__p] == 0) { for (;current_index_x<max_num_sends;current_index_x++) { if (send_reqs_x[current_index_x] == MPI_REQUEST_NULL) { _i = current_index_x; break; } } if (current_index_x<max_num_sends) { current_index_x++; } else { MPI_Waitany(max_num_sends, send_reqs_x, &_i, &send_stats_x[__p]); assert(_i != MPI_UNDEFINED); } index_x[__p] = _i; send_counts_x[__p] = 4; }send_counts_x[__p] = pack_x_3_1(t1,t2,t3, send_buf_x[index_x[__p]], send_counts_x[__p], x_trans,buff_mang_x); } };
    for (__p=0; __p<nprocs; __p++) { if(send_counts_x[__p] > 0) {send_buf_x[index_x[__p]][0] = 1;send_buf_x[index_x[__p]][1] = t1;send_buf_x[index_x[__p]][2] = t2;send_buf_x[index_x[__p]][3] = t3; IF_DYNSCHEDULER_DEBUG_PRINT( fprintf(__debug_print_fp, "sending node %d loop 1 task %d %d %d 		        size %d to node %d\n", my_rank, t1,t2,t3, send_counts_x[__p], __p)); IF_DEBUG_FLUSH(fflush(__debug_print_fp));IF_TIME(__total_count += send_counts_x[__p]); MPI_Isend(send_buf_x[index_x[__p]], send_counts_x[__p], MPI_DOUBLE, __p, 12123, MPI_COMM_WORLD, &send_reqs_x[index_x[__p]]); send_counts_x[__p] = 0;}};
;
    for (__p=0; __p<nprocs; __p++) { if(send_counts_B[__p] > 0) {send_buf_B[index_B[__p]][0] = 1;send_buf_B[index_B[__p]][1] = t1;send_buf_B[index_B[__p]][2] = t2;send_buf_B[index_B[__p]][3] = t3; IF_DYNSCHEDULER_DEBUG_PRINT( fprintf(__debug_print_fp, "sending node %d loop 1 task %d %d %d 		        size %d to node %d\n", my_rank, t1,t2,t3, send_counts_B[__p], __p)); IF_DEBUG_FLUSH(fflush(__debug_print_fp));IF_TIME(__total_count += send_counts_B[__p]); MPI_Isend(send_buf_B[index_B[__p]], send_counts_B[__p], MPI_DOUBLE, __p, 13123, MPI_COMM_WORLD, &send_reqs_B[index_B[__p]]); send_counts_B[__p] = 0;}};
;
    for (__p=0; __p<nprocs; __p++) { if(send_counts_y[__p] > 0) {send_buf_y[index_y[__p]][0] = 1;send_buf_y[index_y[__p]][1] = t1;send_buf_y[index_y[__p]][2] = t2;send_buf_y[index_y[__p]][3] = t3; IF_DYNSCHEDULER_DEBUG_PRINT( fprintf(__debug_print_fp, "sending node %d loop 1 task %d %d %d 		        size %d to node %d\n", my_rank, t1,t2,t3, send_counts_y[__p], __p)); IF_DEBUG_FLUSH(fflush(__debug_print_fp));IF_TIME(__total_count += send_counts_y[__p]); MPI_Isend(send_buf_y[index_y[__p]], send_counts_y[__p], MPI_DOUBLE, __p, 6123, MPI_COMM_WORLD, &send_reqs_y[index_y[__p]]); send_counts_y[__p] = 0;}};
*p_total_count += __total_count;
}
void unpack_buf_x_1(int my_rank,int nprocs, double** x_trans, BufferManager *buff_mang_x,double *recv_buf_x)
{
int curr_displs_x;
curr_displs_x = 4; if (is_receiver_x_1_1((int)recv_buf_x[1],(int)recv_buf_x[2],(int)recv_buf_x[3],N,n,my_rank,nprocs) != 0) { curr_displs_x = unpack_x_1_1((int)recv_buf_x[1],(int)recv_buf_x[2],(int)recv_buf_x[3],recv_buf_x,curr_displs_x, x_trans,buff_mang_x); }if (is_receiver_x_2_1((int)recv_buf_x[1],(int)recv_buf_x[2],(int)recv_buf_x[3],N,n,my_rank,nprocs) != 0) { curr_displs_x = unpack_x_2_1((int)recv_buf_x[1],(int)recv_buf_x[2],(int)recv_buf_x[3],recv_buf_x,curr_displs_x, x_trans,buff_mang_x); }if (is_receiver_x_3_1((int)recv_buf_x[1],(int)recv_buf_x[2],(int)recv_buf_x[3],N,n,my_rank,nprocs) != 0) { curr_displs_x = unpack_x_3_1((int)recv_buf_x[1],(int)recv_buf_x[2],(int)recv_buf_x[3],recv_buf_x,curr_displs_x, x_trans,buff_mang_x); };
}
void unpack_buf_B_1(int my_rank,int nprocs, double** B_trans,int B_size_1, BufferManager *buff_mang_B,double *recv_buf_B)
{
int curr_displs_B;
curr_displs_B = 4; ;
}
void unpack_buf_y_1(int my_rank,int nprocs, double** y_trans, BufferManager *buff_mang_y,double *recv_buf_y)
{
int curr_displs_y;
curr_displs_y = 4; ;
}
int pack_x_1_2(int ts1,int ts2,double *send_buf_x,int send_counts_x, double** x_trans, BufferManager *buff_mang_x){
int recv_proc;
  int t1, t2, t3;
 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;
/* Start of CLooG code */
if ((ts1 == 2) && (ts2 >= 0) && (128*ts2 <= n-1)) {
  if(x_trans[(ts2)] == NULL) x_trans[(ts2)] = (double *) buffer_alloc_atomic((ts2), 128 * sizeof(double), buff_mang_x);;
  for (t1=128*ts2+2;t1<=min(n+1,128*ts2+129);t1++) {
    send_buf_x[send_counts_x++] = x_trans[((1)*( t1-2 ))/128][((1)*( t1-2 ))%128];
  }
}
/* End of CLooG code */
return send_counts_x;
}
int unpack_x_1_2(int ts1,int ts2,double *recv_buf_x,int curr_displs_x, double** x_trans, BufferManager *buff_mang_x){
int recv_proc;
  int t1, t2, t3;
 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;
/* Start of CLooG code */
if ((ts1 == 2) && (ts2 >= 0) && (128*ts2 <= n-1)) {
  if(x_trans[(ts2)] == NULL) x_trans[(ts2)] = (double *) buffer_alloc_atomic((ts2), 128 * sizeof(double), buff_mang_x);;
  for (t1=128*ts2+2;t1<=min(n+1,128*ts2+129);t1++) {
    x_trans[((1)*( t1-2 ))/128][((1)*( t1-2 ))%128] = recv_buf_x[curr_displs_x++];
  }
}
/* End of CLooG code */
return curr_displs_x;
}
void data_tile_copy_back_x_2(int ts1,int ts2, double** x_trans, BufferManager *buff_mang_x)
{
#ifndef USE_LOCAL_ARRAYS
  int t1, t2;
 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;
/* Start of CLooG code */
if ((ts1 == 2) && (ts2 >= 0) && (128*ts2 <= n-1)) {
  for (t2=128*ts2;t2<=min(n-1,128*ts2+127);t2++) {
    x[t2] = x_trans[((1)*( t2 ))/128][((1)*( t2 ))%128];;
  }
}
/* End of CLooG code */
#endif
}
int writeout_pack_x_2(int ts1,int ts2,double *lw_buf_x,int lw_count_x, double** x_trans, BufferManager *buff_mang_x){
int recv_proc;
  int t1, t2, t3;
 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;
/* Start of CLooG code */
if ((ts1 == 2) && (ts2 >= 0) && (128*ts2 <= n-1)) {
  if(x_trans[(ts2)] == NULL) x_trans[(ts2)] = (double *) buffer_alloc_atomic((ts2), 128 * sizeof(double), buff_mang_x);;
  for (t1=128*ts2+2;t1<=min(n+1,128*ts2+129);t1++) {
    lw_buf_x[lw_count_x++] = x_trans[((1)*( t1-2 ))/128][((1)*( t1-2 ))%128];
  }
}
/* End of CLooG code */
return lw_count_x;
}
int writeout_unpack_x_2(int ts1,int ts2,double *lw_recv_buf_x,int displs_lw_x, double** x_trans, BufferManager *buff_mang_x){
int recv_proc;
  int t1, t2;
 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;
/* Start of CLooG code */
if ((ts1 == 2) && (ts2 >= 0) && (128*ts2 <= n-1)) {
  for (t1=128*ts2;t1<=min(n-1,128*ts2+127);t1++) {
    
#ifndef USE_LOCAL_ARRAYS
 x[t1] = lw_recv_buf_x[displs_lw_x++]; 
#endif
;
  }
}
/* End of CLooG code */
return displs_lw_x;
}
void pack_send_2(int my_rank,int nprocs,int t1,int t2, double** x_trans, BufferManager *buff_mang_x,double **send_buf_x,MPI_Request *send_reqs_x,MPI_Status *send_stats_x, double** z_trans, BufferManager *buff_mang_z,double **send_buf_z,MPI_Request *send_reqs_z,MPI_Status *send_stats_z,int max_num_sends,double *p_total_count)
{
int __p, receiver_list[nprocs];
int send_counts_x[nprocs];
for (__p=0;__p<nprocs;__p++) send_counts_x[__p] = 0;
int index_x[nprocs];
int current_index_x = 0;
int send_counts_z[nprocs];
for (__p=0;__p<nprocs;__p++) send_counts_z[__p] = 0;
int index_z[nprocs];
int current_index_z = 0;
int _i;
double __total_count = 0;
for (__p=0; __p<nprocs; __p++) { receiver_list[__p] = 0; } sigma_x_1_2(t1,t2,N,n, my_rank, nprocs, receiver_list); for (__p=0; __p<nprocs; __p++) { if (receiver_list[__p] != 0) { if (send_counts_x[__p] == 0) { for (;current_index_x<max_num_sends;current_index_x++) { if (send_reqs_x[current_index_x] == MPI_REQUEST_NULL) { _i = current_index_x; break; } } if (current_index_x<max_num_sends) { current_index_x++; } else { MPI_Waitany(max_num_sends, send_reqs_x, &_i, &send_stats_x[__p]); assert(_i != MPI_UNDEFINED); } index_x[__p] = _i; send_counts_x[__p] = 4; }send_counts_x[__p] = pack_x_1_2(t1,t2, send_buf_x[index_x[__p]], send_counts_x[__p], x_trans,buff_mang_x); } };
    for (__p=0; __p<nprocs; __p++) { if(send_counts_x[__p] > 0) {send_buf_x[index_x[__p]][0] = 2;send_buf_x[index_x[__p]][1] = t1;send_buf_x[index_x[__p]][2] = t2; IF_DYNSCHEDULER_DEBUG_PRINT( fprintf(__debug_print_fp, "sending node %d loop 2 task %d %d 		        size %d to node %d\n", my_rank, t1,t2, send_counts_x[__p], __p)); IF_DEBUG_FLUSH(fflush(__debug_print_fp));IF_TIME(__total_count += send_counts_x[__p]); MPI_Isend(send_buf_x[index_x[__p]], send_counts_x[__p], MPI_DOUBLE, __p, 12123, MPI_COMM_WORLD, &send_reqs_x[index_x[__p]]); send_counts_x[__p] = 0;}};
;
    for (__p=0; __p<nprocs; __p++) { if(send_counts_z[__p] > 0) {send_buf_z[index_z[__p]][0] = 2;send_buf_z[index_z[__p]][1] = t1;send_buf_z[index_z[__p]][2] = t2; IF_DYNSCHEDULER_DEBUG_PRINT( fprintf(__debug_print_fp, "sending node %d loop 2 task %d %d 		        size %d to node %d\n", my_rank, t1,t2, send_counts_z[__p], __p)); IF_DEBUG_FLUSH(fflush(__debug_print_fp));IF_TIME(__total_count += send_counts_z[__p]); MPI_Isend(send_buf_z[index_z[__p]], send_counts_z[__p], MPI_DOUBLE, __p, 5123, MPI_COMM_WORLD, &send_reqs_z[index_z[__p]]); send_counts_z[__p] = 0;}};
*p_total_count += __total_count;
}
void unpack_buf_x_2(int my_rank,int nprocs, double** x_trans, BufferManager *buff_mang_x,double *recv_buf_x)
{
int curr_displs_x;
curr_displs_x = 4; if (is_receiver_x_1_2((int)recv_buf_x[1],(int)recv_buf_x[2],N,n,my_rank,nprocs) != 0) { curr_displs_x = unpack_x_1_2((int)recv_buf_x[1],(int)recv_buf_x[2],recv_buf_x,curr_displs_x, x_trans,buff_mang_x); };
}
void unpack_buf_z_2(int my_rank,int nprocs, double** z_trans, BufferManager *buff_mang_z,double *recv_buf_z)
{
int curr_displs_z;
curr_displs_z = 4; ;
}
int pack_w_1_3(int ts1,int ts2,int ts3,double *send_buf_w,int send_counts_w, double** w_trans, BufferManager *buff_mang_w){
int recv_proc;
  int t1, t2, t3, t4;
 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;
/* Start of CLooG code */
if ((ts1 == 3) && (ts2 >= 0) && (128*ts2 <= n-129) && (ts3 >= 0) && (128*ts3 <= n-1)) {
  if(w_trans[(ts3)] == NULL) w_trans[(ts3)] = (double *) buffer_alloc_atomic((ts3), 128 * sizeof(double), buff_mang_w);;
  for (t1=128*ts3+2;t1<=min(n+1,128*ts3+129);t1++) {
    send_buf_w[send_counts_w++] = w_trans[((1)*( t1-2 ))/128][((1)*( t1-2 ))%128];
  }
}
/* End of CLooG code */
return send_counts_w;
}
int unpack_w_1_3(int ts1,int ts2,int ts3,double *recv_buf_w,int curr_displs_w, double** w_trans, BufferManager *buff_mang_w){
int recv_proc;
  int t1, t2, t3, t4;
 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;
/* Start of CLooG code */
if ((ts1 == 3) && (ts2 >= 0) && (128*ts2 <= n-129) && (ts3 >= 0) && (128*ts3 <= n-1)) {
  if(w_trans[(ts3)] == NULL) w_trans[(ts3)] = (double *) buffer_alloc_atomic((ts3), 128 * sizeof(double), buff_mang_w);;
  for (t1=128*ts3+2;t1<=min(n+1,128*ts3+129);t1++) {
    w_trans[((1)*( t1-2 ))/128][((1)*( t1-2 ))%128] = recv_buf_w[curr_displs_w++];
  }
}
/* End of CLooG code */
return curr_displs_w;
}
void data_tile_copy_back_w_3(int ts1,int ts2,int ts3, double** w_trans, BufferManager *buff_mang_w)
{
#ifndef USE_LOCAL_ARRAYS
  int t1, t2;
 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;
/* Start of CLooG code */
if ((ts1 == 3) && (ts2 >= 0) && (128*ts2 <= n-1) && (ts3 >= 0) && (128*ts3 <= n-1)) {
  for (t2=128*ts3;t2<=min(n-1,128*ts3+127);t2++) {
    w[t2] = w_trans[((1)*( t2 ))/128][((1)*( t2 ))%128];;
  }
}
/* End of CLooG code */
#endif
}
int writeout_pack_w_3(int ts1,int ts2,int ts3,double *lw_buf_w,int lw_count_w, double** w_trans, BufferManager *buff_mang_w){
int recv_proc;
  int t1, t2, t3, t4;
 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;
/* Start of CLooG code */
if ((ts1 == 3) && (128*ts2 >= n-128) && (128*ts2 <= n-1) && (ts3 >= 0) && (128*ts3 <= n-1)) {
  if(w_trans[(ts3)] == NULL) w_trans[(ts3)] = (double *) buffer_alloc_atomic((ts3), 128 * sizeof(double), buff_mang_w);;
  for (t1=128*ts3+2;t1<=min(n+1,128*ts3+129);t1++) {
    lw_buf_w[lw_count_w++] = w_trans[((1)*( t1-2 ))/128][((1)*( t1-2 ))%128];
  }
}
/* End of CLooG code */
return lw_count_w;
}
int writeout_unpack_w_3(int ts1,int ts2,int ts3,double *lw_recv_buf_w,int displs_lw_w, double** w_trans, BufferManager *buff_mang_w){
int recv_proc;
  int t1, t2, t3;
 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;
/* Start of CLooG code */
if ((ts1 == 3) && (128*ts2 >= n-128) && (128*ts2 <= n-1) && (ts3 >= 0) && (128*ts3 <= n-1)) {
  for (t1=128*ts3;t1<=min(n-1,128*ts3+127);t1++) {
    
#ifndef USE_LOCAL_ARRAYS
 w[t1] = lw_recv_buf_w[displs_lw_w++]; 
#endif
;
  }
}
/* End of CLooG code */
return displs_lw_w;
}
void pack_send_3(int my_rank,int nprocs,int t1,int t2,int t3, double** w_trans, BufferManager *buff_mang_w,double **send_buf_w,MPI_Request *send_reqs_w,MPI_Status *send_stats_w, double** B_trans,int B_size_1, BufferManager *buff_mang_B,double **send_buf_B,MPI_Request *send_reqs_B,MPI_Status *send_stats_B, double** x_trans, BufferManager *buff_mang_x,double **send_buf_x,MPI_Request *send_reqs_x,MPI_Status *send_stats_x,int max_num_sends,double *p_total_count)
{
int __p, receiver_list[nprocs];
int send_counts_w[nprocs];
for (__p=0;__p<nprocs;__p++) send_counts_w[__p] = 0;
int index_w[nprocs];
int current_index_w = 0;
int send_counts_B[nprocs];
for (__p=0;__p<nprocs;__p++) send_counts_B[__p] = 0;
int index_B[nprocs];
int current_index_B = 0;
int send_counts_x[nprocs];
for (__p=0;__p<nprocs;__p++) send_counts_x[__p] = 0;
int index_x[nprocs];
int current_index_x = 0;
int _i;
double __total_count = 0;
for (__p=0; __p<nprocs; __p++) { receiver_list[__p] = 0; } sigma_w_1_3(t1,t2,t3,N,n, my_rank, nprocs, receiver_list); for (__p=0; __p<nprocs; __p++) { if (receiver_list[__p] != 0) { if (send_counts_w[__p] == 0) { for (;current_index_w<max_num_sends;current_index_w++) { if (send_reqs_w[current_index_w] == MPI_REQUEST_NULL) { _i = current_index_w; break; } } if (current_index_w<max_num_sends) { current_index_w++; } else { MPI_Waitany(max_num_sends, send_reqs_w, &_i, &send_stats_w[__p]); assert(_i != MPI_UNDEFINED); } index_w[__p] = _i; send_counts_w[__p] = 4; }send_counts_w[__p] = pack_w_1_3(t1,t2,t3, send_buf_w[index_w[__p]], send_counts_w[__p], w_trans,buff_mang_w); } };
    for (__p=0; __p<nprocs; __p++) { if(send_counts_w[__p] > 0) {send_buf_w[index_w[__p]][0] = 3;send_buf_w[index_w[__p]][1] = t1;send_buf_w[index_w[__p]][2] = t2;send_buf_w[index_w[__p]][3] = t3; IF_DYNSCHEDULER_DEBUG_PRINT( fprintf(__debug_print_fp, "sending node %d loop 3 task %d %d %d 		        size %d to node %d\n", my_rank, t1,t2,t3, send_counts_w[__p], __p)); IF_DEBUG_FLUSH(fflush(__debug_print_fp));IF_TIME(__total_count += send_counts_w[__p]); MPI_Isend(send_buf_w[index_w[__p]], send_counts_w[__p], MPI_DOUBLE, __p, 7123, MPI_COMM_WORLD, &send_reqs_w[index_w[__p]]); send_counts_w[__p] = 0;}};
;
    for (__p=0; __p<nprocs; __p++) { if(send_counts_B[__p] > 0) {send_buf_B[index_B[__p]][0] = 3;send_buf_B[index_B[__p]][1] = t1;send_buf_B[index_B[__p]][2] = t2;send_buf_B[index_B[__p]][3] = t3; IF_DYNSCHEDULER_DEBUG_PRINT( fprintf(__debug_print_fp, "sending node %d loop 3 task %d %d %d 		        size %d to node %d\n", my_rank, t1,t2,t3, send_counts_B[__p], __p)); IF_DEBUG_FLUSH(fflush(__debug_print_fp));IF_TIME(__total_count += send_counts_B[__p]); MPI_Isend(send_buf_B[index_B[__p]], send_counts_B[__p], MPI_DOUBLE, __p, 13123, MPI_COMM_WORLD, &send_reqs_B[index_B[__p]]); send_counts_B[__p] = 0;}};
;
    for (__p=0; __p<nprocs; __p++) { if(send_counts_x[__p] > 0) {send_buf_x[index_x[__p]][0] = 3;send_buf_x[index_x[__p]][1] = t1;send_buf_x[index_x[__p]][2] = t2;send_buf_x[index_x[__p]][3] = t3; IF_DYNSCHEDULER_DEBUG_PRINT( fprintf(__debug_print_fp, "sending node %d loop 3 task %d %d %d 		        size %d to node %d\n", my_rank, t1,t2,t3, send_counts_x[__p], __p)); IF_DEBUG_FLUSH(fflush(__debug_print_fp));IF_TIME(__total_count += send_counts_x[__p]); MPI_Isend(send_buf_x[index_x[__p]], send_counts_x[__p], MPI_DOUBLE, __p, 12123, MPI_COMM_WORLD, &send_reqs_x[index_x[__p]]); send_counts_x[__p] = 0;}};
*p_total_count += __total_count;
}
void unpack_buf_w_3(int my_rank,int nprocs, double** w_trans, BufferManager *buff_mang_w,double *recv_buf_w)
{
int curr_displs_w;
curr_displs_w = 4; if (is_receiver_w_1_3((int)recv_buf_w[1],(int)recv_buf_w[2],(int)recv_buf_w[3],N,n,my_rank,nprocs) != 0) { curr_displs_w = unpack_w_1_3((int)recv_buf_w[1],(int)recv_buf_w[2],(int)recv_buf_w[3],recv_buf_w,curr_displs_w, w_trans,buff_mang_w); };
}
void unpack_buf_B_3(int my_rank,int nprocs, double** B_trans,int B_size_1, BufferManager *buff_mang_B,double *recv_buf_B)
{
int curr_displs_B;
curr_displs_B = 4; ;
}
void unpack_buf_x_3(int my_rank,int nprocs, double** x_trans, BufferManager *buff_mang_x,double *recv_buf_x)
{
int curr_displs_x;
curr_displs_x = 4; ;
}
void copy_back_writeout_0(int my_rank, int nprocs, double** B_trans,int B_size_1, BufferManager *buff_mang_B, double** x_trans, BufferManager *buff_mang_x, double** w_trans, BufferManager *buff_mang_w)
{
#ifndef USE_LOCAL_ARRAYS
  int t1, t2, t3, t4;
 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;
/* Start of CLooG code */
if (n >= 1) {
  for (t3=0;t3<=floord(n-1,128);t3++) {
    for (t4=0;t4<=floord(n-1,128);t4++) {
      if(pi_0(0,t3,t4,N,n,nprocs) == my_rank) data_tile_copy_back_B_0(0,t3,t4, B_trans, B_size_1,buff_mang_B);;
    }
  }
  for (t3=0;t3<=floord(n-1,128);t3++) {
    for (t4=0;t4<=floord(n-1,128);t4++) {
      if(pi_1(1,t3,t4,N,n,nprocs) == my_rank) data_tile_copy_back_x_1(1,t3,t4, x_trans,buff_mang_x);;
    }
  }
  for (t3=0;t3<=floord(n-1,128);t3++) {
    if(pi_2(2,t3,N,n,nprocs) == my_rank) data_tile_copy_back_x_2(2,t3, x_trans,buff_mang_x);;
  }
  for (t3=0;t3<=floord(n-1,128);t3++) {
    for (t4=0;t4<=floord(n-1,128);t4++) {
      if(pi_3(3,t3,t4,N,n,nprocs) == my_rank) data_tile_copy_back_w_3(3,t3,t4, w_trans,buff_mang_w);;
    }
  }
}
/* End of CLooG code */
#endif
}
void write_out(int my_rank,int nprocs,double *lw_buf_B,double *lw_recv_buf_B,int *displs_lw_B,double *lw_buf_x,double *lw_recv_buf_x,int *displs_lw_x,double *lw_buf_w,double *lw_recv_buf_w,int *displs_lw_w, double** B_trans,int B_size_1, BufferManager *buff_mang_B, double** x_trans, BufferManager *buff_mang_x, double** w_trans, BufferManager *buff_mang_w)
{
#ifndef USE_LOCAL_ARRAYS
int __p, proc;
int _lb_dist, _ub_dist, lbd_t1, ubd_t1, lbd_t2, ubd_t2, lbd_t3, ubd_t3, lbd_t4, ubd_t4, lbd_t5, ubd_t5, lbd_t6, ubd_t6, lbd_t7, ubd_t7, lbd_t8, ubd_t8, lbd_t9, ubd_t9, lbd_t10, ubd_t10;
int lw_recv_counts_B[nprocs];
int curr_displs_lw_B[nprocs];
int lw_count_B = 0;
int lw_recv_counts_x[nprocs];
int curr_displs_lw_x[nprocs];
int lw_count_x = 0;
int lw_recv_counts_w[nprocs];
int curr_displs_lw_w[nprocs];
int lw_count_w = 0;
  int t1, t2, t3, t4, t5, t6, t7;
 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;
/* Start of CLooG code */
if (n >= 1) {
  for (t5=0;t5<=floord(n-1,128);t5++) {
    for (t7=0;t7<=floord(n-1,128);t7++) {
      if (pi_0(0,t5,t7,N,n,nprocs) == my_rank)lw_count_B = writeout_pack_B_0(0,t5,t7,lw_buf_B,lw_count_B, B_trans, B_size_1,buff_mang_B);;
    }
  }
  assert((nprocs == 1) || (lw_count_B <= displs_lw_B[1])); MPI_Gather(&lw_count_B, 1, MPI_INT, lw_recv_counts_B, 1, MPI_INT, 0, MPI_COMM_WORLD);MPI_Gatherv(lw_buf_B, lw_count_B, MPI_DOUBLE, lw_recv_buf_B, lw_recv_counts_B, displs_lw_B, MPI_DOUBLE, 0, MPI_COMM_WORLD); lw_count_B = 0; for (__p=0; __p<nprocs; __p++) curr_displs_lw_B[__p] = displs_lw_B[__p];;
  if (my_rank == 0) {
    for (t5=0;t5<=floord(n-1,128);t5++) {
      for (t7=0;t7<=floord(n-1,128);t7++) {
        proc = pi_0(0, t5, t7, N, n, nprocs);if ((my_rank != proc) && (lw_recv_counts_B[proc] > 0)) {curr_displs_lw_B[proc] = writeout_unpack_B_0(0,t5,t7,lw_recv_buf_B,curr_displs_lw_B[proc], B_trans, B_size_1,buff_mang_B);};
      }
    }
  }
  for (t5=0;t5<=floord(n-1,128);t5++) {
    for (t7=0;t7<=floord(n-1,128);t7++) {
      if (pi_1(1,t5,t7,N,n,nprocs) == my_rank)lw_count_x = writeout_pack_x_1(1,t5,t7,lw_buf_x,lw_count_x, x_trans,buff_mang_x);;
    }
  }
  assert((nprocs == 1) || (lw_count_x <= displs_lw_x[1])); MPI_Gather(&lw_count_x, 1, MPI_INT, lw_recv_counts_x, 1, MPI_INT, 0, MPI_COMM_WORLD);MPI_Gatherv(lw_buf_x, lw_count_x, MPI_DOUBLE, lw_recv_buf_x, lw_recv_counts_x, displs_lw_x, MPI_DOUBLE, 0, MPI_COMM_WORLD); lw_count_x = 0; for (__p=0; __p<nprocs; __p++) curr_displs_lw_x[__p] = displs_lw_x[__p];;
  if (my_rank == 0) {
    for (t5=0;t5<=floord(n-1,128);t5++) {
      for (t7=0;t7<=floord(n-1,128);t7++) {
        proc = pi_1(1, t5, t7, N, n, nprocs);if ((my_rank != proc) && (lw_recv_counts_x[proc] > 0)) {curr_displs_lw_x[proc] = writeout_unpack_x_1(1,t5,t7,lw_recv_buf_x,curr_displs_lw_x[proc], x_trans,buff_mang_x);};
      }
    }
  }
  for (t5=0;t5<=floord(n-1,128);t5++) {
    if (pi_2(2,t5,N,n,nprocs) == my_rank)lw_count_x = writeout_pack_x_2(2,t5,lw_buf_x,lw_count_x, x_trans,buff_mang_x);;
  }
  assert((nprocs == 1) || (lw_count_x <= displs_lw_x[1])); MPI_Gather(&lw_count_x, 1, MPI_INT, lw_recv_counts_x, 1, MPI_INT, 0, MPI_COMM_WORLD);MPI_Gatherv(lw_buf_x, lw_count_x, MPI_DOUBLE, lw_recv_buf_x, lw_recv_counts_x, displs_lw_x, MPI_DOUBLE, 0, MPI_COMM_WORLD); lw_count_x = 0; for (__p=0; __p<nprocs; __p++) curr_displs_lw_x[__p] = displs_lw_x[__p];;
  if (my_rank == 0) {
    for (t5=0;t5<=floord(n-1,128);t5++) {
      proc = pi_2(2, t5, N, n, nprocs);if ((my_rank != proc) && (lw_recv_counts_x[proc] > 0)) {curr_displs_lw_x[proc] = writeout_unpack_x_2(2,t5,lw_recv_buf_x,curr_displs_lw_x[proc], x_trans,buff_mang_x);};
    }
  }
  for (t5=0;t5<=floord(n-1,128);t5++) {
    for (t7=0;t7<=floord(n-1,128);t7++) {
      if (pi_3(3,t5,t7,N,n,nprocs) == my_rank)lw_count_w = writeout_pack_w_3(3,t5,t7,lw_buf_w,lw_count_w, w_trans,buff_mang_w);;
    }
    assert((nprocs == 1) || (lw_count_w <= displs_lw_w[1])); MPI_Gather(&lw_count_w, 1, MPI_INT, lw_recv_counts_w, 1, MPI_INT, 0, MPI_COMM_WORLD);MPI_Gatherv(lw_buf_w, lw_count_w, MPI_DOUBLE, lw_recv_buf_w, lw_recv_counts_w, displs_lw_w, MPI_DOUBLE, 0, MPI_COMM_WORLD); lw_count_w = 0; for (__p=0; __p<nprocs; __p++) curr_displs_lw_w[__p] = displs_lw_w[__p];;
    if (my_rank == 0) {
      for (t7=0;t7<=floord(n-1,128);t7++) {
        proc = pi_3(3, t5, t7, N, n, nprocs);if ((my_rank != proc) && (lw_recv_counts_w[proc] > 0)) {curr_displs_lw_w[proc] = writeout_unpack_w_3(3,t5,t7,lw_recv_buf_w,curr_displs_lw_w[proc], w_trans,buff_mang_w);};
      }
    }
  }
}
/* End of CLooG code */
#endif
}
void data_tile_alloc_0(int ts1,int ts2,int ts3, double** B_trans,int B_size_1, BufferManager *buff_mang_B, double** A_trans,int A_size_1, BufferManager *buff_mang_A, double** u1_trans, BufferManager *buff_mang_u1, double** v1_trans, BufferManager *buff_mang_v1, double** u2_trans, BufferManager *buff_mang_u2, double** v2_trans, BufferManager *buff_mang_v2)
{
  int t1, t2, t3;
 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;
/* Start of CLooG code */
if ((ts1 == 0) && (ts2 >= 0) && (128*ts2 <= n-1) && (ts3 >= 0) && (128*ts3 <= n-1)) {
  if(B_trans[(ts3) * ((B_size_1)) + (ts2)] == NULL) B_trans[(ts3)* ( B_size_1 ) + (ts2)] = (double *) buffer_alloc_atomic((ts3)* ( B_size_1 ) + (ts2), 128*128 * sizeof(double), buff_mang_B);;
  if(A_trans[(ts3) * ((A_size_1)) + (ts2)] == NULL) A_trans[(ts3)* ( A_size_1 ) + (ts2)] = (double *) buffer_alloc_atomic((ts3)* ( A_size_1 ) + (ts2), 128*128 * sizeof(double), buff_mang_A);;
  if(u1_trans[(ts3)] == NULL) u1_trans[(ts3)] = (double *) buffer_alloc_atomic((ts3), 128 * sizeof(double), buff_mang_u1);;
  if(v1_trans[(ts2)] == NULL) v1_trans[(ts2)] = (double *) buffer_alloc_atomic((ts2), 128 * sizeof(double), buff_mang_v1);;
  if(u2_trans[(ts3)] == NULL) u2_trans[(ts3)] = (double *) buffer_alloc_atomic((ts3), 128 * sizeof(double), buff_mang_u2);;
  if(v2_trans[(ts2)] == NULL) v2_trans[(ts2)] = (double *) buffer_alloc_atomic((ts2), 128 * sizeof(double), buff_mang_v2);;
}
/* End of CLooG code */
}
void data_tile_ref_count_update_0(int ts1,int ts2,int ts3, double** B_trans,int B_size_1, BufferManager *buff_mang_B, double** A_trans,int A_size_1, BufferManager *buff_mang_A, double** u1_trans, BufferManager *buff_mang_u1, double** v1_trans, BufferManager *buff_mang_v1, double** u2_trans, BufferManager *buff_mang_u2, double** v2_trans, BufferManager *buff_mang_v2)
{
  int t1, t2, t3;
 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;
/* Start of CLooG code */
if ((ts1 == 0) && (ts2 >= 0) && (128*ts2 <= n-1) && (ts3 >= 0) && (128*ts3 <= n-1)) {
  decrement_ref_count(buff_mang_B, (ts3)* ( B_size_1 ) + (ts2));;
  decrement_ref_count(buff_mang_A, (ts3)* ( A_size_1 ) + (ts2));;
  decrement_ref_count(buff_mang_u1, (ts3));;
  decrement_ref_count(buff_mang_v1, (ts2));;
  decrement_ref_count(buff_mang_u2, (ts3));;
  decrement_ref_count(buff_mang_v2, (ts2));;
}
/* End of CLooG code */
}
void compute_task_0(int t1,int t2,int t3, double** B_trans,int B_size_1, BufferManager *buff_mang_B, double** A_trans,int A_size_1, BufferManager *buff_mang_A, double** u1_trans, BufferManager *buff_mang_u1, double** v1_trans, BufferManager *buff_mang_v1, double** u2_trans, BufferManager *buff_mang_u2, double** v2_trans, BufferManager *buff_mang_v2)
{
int i_j_0_div = 0;int i_j_0_mod = 0;double *B_i_j_0 = NULL;double *A_i_j_0 = NULL;int i_0_div = 0;int i_0_mod = 0;double *u1_i_0 = NULL;int j_0_div = 0;int j_0_mod = 0;double *v1_j_0 = NULL;double *u2_i_0 = NULL;double *v2_j_0 = NULL; int t4, t5, t6, t7, t8, t9, t10;
 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;
/* Start of CLooG code */
for (t4=128*t3;t4<=min(n-1,128*t3+127);t4++) {
  lbv=128*t2;
  ubv=min(n-1,128*t2+127);
  #if (defined __DATA_DIST_DECLS || defined USE_LOCAL_ARRAYS)
  i_j_0_div = ((1)*( t4 ))/128* ( B_size_1 ) + ((1)*( lbv ))/128;i_j_0_mod = ((1)*( t4 ))%128* (128) + ((1)*( lbv ))%128;B_i_j_0 = B_trans[i_j_0_div];A_i_j_0 = A_trans[i_j_0_div];i_0_div = ((1)*( t4 ))/128;i_0_mod = ((1)*( t4 ))%128;u1_i_0 = u1_trans[i_0_div];j_0_div = ((1)*( lbv ))/128;j_0_mod = ((1)*( lbv ))%128;v1_j_0 = v1_trans[j_0_div];u2_i_0 = u2_trans[i_0_div];v2_j_0 = v2_trans[j_0_div];;
  #endif
#pragma ivdep
#pragma vector always
  for (t10=lbv;t10<=ubv;t10++) {
    B_i_j_0[i_j_0_mod + t10 - lbv]=A_i_j_0[i_j_0_mod + t10 - lbv]+u1_i_0[i_0_mod]*v1_j_0[j_0_mod + t10 - lbv]+u2_i_0[i_0_mod]*v2_j_0[j_0_mod + t10 - lbv];;
  }
}
/* End of CLooG code */
}
void data_tile_alloc_1(int ts1,int ts2,int ts3, double** x_trans, BufferManager *buff_mang_x, double** B_trans,int B_size_1, BufferManager *buff_mang_B, double** y_trans, BufferManager *buff_mang_y)
{
  int t1, t2, t3;
 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;
/* Start of CLooG code */
if ((ts1 == 1) && (ts2 >= 0) && (128*ts2 <= n-1) && (ts3 >= 0) && (128*ts3 <= n-1)) {
  if(x_trans[(ts2)] == NULL) x_trans[(ts2)] = (double *) buffer_alloc_atomic((ts2), 128 * sizeof(double), buff_mang_x);;
  if(B_trans[(ts3) * ((B_size_1)) + (ts2)] == NULL) B_trans[(ts3)* ( B_size_1 ) + (ts2)] = (double *) buffer_alloc_atomic((ts3)* ( B_size_1 ) + (ts2), 128*128 * sizeof(double), buff_mang_B);;
  if(y_trans[(ts3)] == NULL) y_trans[(ts3)] = (double *) buffer_alloc_atomic((ts3), 128 * sizeof(double), buff_mang_y);;
}
/* End of CLooG code */
}
void data_tile_ref_count_update_1(int ts1,int ts2,int ts3, double** x_trans, BufferManager *buff_mang_x, double** B_trans,int B_size_1, BufferManager *buff_mang_B, double** y_trans, BufferManager *buff_mang_y)
{
  int t1, t2, t3;
 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;
/* Start of CLooG code */
if ((ts1 == 1) && (ts2 >= 0) && (128*ts2 <= n-1) && (ts3 >= 0) && (128*ts3 <= n-1)) {
  decrement_ref_count(buff_mang_x, (ts2));;
  decrement_ref_count(buff_mang_B, (ts3)* ( B_size_1 ) + (ts2));;
  decrement_ref_count(buff_mang_y, (ts3));;
}
/* End of CLooG code */
}
void compute_task_1(int t1,int t2,int t3, double** x_trans, BufferManager *buff_mang_x, double** B_trans,int B_size_1, BufferManager *buff_mang_B, double** y_trans, BufferManager *buff_mang_y)
{
int i_0_div = 0;int i_0_mod = 0;double *x_i_0 = NULL;int j_i_0_div = 0;int j_i_0_mod = 0;double *B_j_i_0 = NULL;int j_0_div = 0;int j_0_mod = 0;double *y_j_0 = NULL; int t4, t5, t6, t7, t8, t9, t10;
 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;
/* Start of CLooG code */
for (t4=128*t3;t4<=min(n-1,128*t3+127);t4++) {
  lbv=128*t2;
  ubv=min(n-1,128*t2+127);
  #if (defined __DATA_DIST_DECLS || defined USE_LOCAL_ARRAYS)
  i_0_div = ((1)*( lbv ))/128;i_0_mod = ((1)*( lbv ))%128;x_i_0 = x_trans[i_0_div];j_i_0_div = ((1)*( t4 ))/128* ( B_size_1 ) + ((1)*( lbv ))/128;j_i_0_mod = ((1)*( t4 ))%128* (128) + ((1)*( lbv ))%128;B_j_i_0 = B_trans[j_i_0_div];j_0_div = ((1)*( t4 ))/128;j_0_mod = ((1)*( t4 ))%128;y_j_0 = y_trans[j_0_div];;
  #endif
#pragma ivdep
#pragma vector always
  for (t10=lbv;t10<=ubv;t10++) {
    x_i_0[i_0_mod + t10 - lbv]=x_i_0[i_0_mod + t10 - lbv]+beta*B_j_i_0[j_i_0_mod + t10 - lbv]*y_j_0[j_0_mod];;
  }
}
/* End of CLooG code */
}
void data_tile_alloc_2(int ts1,int ts2, double** x_trans, BufferManager *buff_mang_x, double** z_trans, BufferManager *buff_mang_z)
{
  int t1, t2;
 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;
/* Start of CLooG code */
if ((ts1 == 2) && (ts2 >= 0) && (128*ts2 <= n-1)) {
  if(x_trans[(ts2)] == NULL) x_trans[(ts2)] = (double *) buffer_alloc_atomic((ts2), 128 * sizeof(double), buff_mang_x);;
  if(z_trans[(ts2)] == NULL) z_trans[(ts2)] = (double *) buffer_alloc_atomic((ts2), 128 * sizeof(double), buff_mang_z);;
}
/* End of CLooG code */
}
void data_tile_ref_count_update_2(int ts1,int ts2, double** x_trans, BufferManager *buff_mang_x, double** z_trans, BufferManager *buff_mang_z)
{
  int t1, t2;
 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;
/* Start of CLooG code */
if ((ts1 == 2) && (ts2 >= 0) && (128*ts2 <= n-1)) {
  decrement_ref_count(buff_mang_x, (ts2));;
  decrement_ref_count(buff_mang_z, (ts2));;
}
/* End of CLooG code */
}
void compute_task_2(int t1,int t2, double** x_trans, BufferManager *buff_mang_x, double** z_trans, BufferManager *buff_mang_z)
{
int i_0_div = 0;int i_0_mod = 0;double *x_i_0 = NULL;double *z_i_0 = NULL; int t3, t4, t5, t6, t7, t8, t9, t10;
 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;
/* Start of CLooG code */
lbv=128*t2;
ubv=min(n-1,128*t2+127);
#if (defined __DATA_DIST_DECLS || defined USE_LOCAL_ARRAYS)
i_0_div = ((1)*( lbv ))/128;i_0_mod = ((1)*( lbv ))%128;x_i_0 = x_trans[i_0_div];z_i_0 = z_trans[i_0_div];;
#endif
#pragma ivdep
#pragma vector always
for (t10=lbv;t10<=ubv;t10++) {
  x_i_0[i_0_mod + t10 - lbv]=x_i_0[i_0_mod + t10 - lbv]+z_i_0[i_0_mod + t10 - lbv];;
}
/* End of CLooG code */
}
void data_tile_alloc_3(int ts1,int ts2,int ts3, double** w_trans, BufferManager *buff_mang_w, double** B_trans,int B_size_1, BufferManager *buff_mang_B, double** x_trans, BufferManager *buff_mang_x)
{
  int t1, t2, t3;
 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;
/* Start of CLooG code */
if ((ts1 == 3) && (ts2 >= 0) && (128*ts2 <= n-1) && (ts3 >= 0) && (128*ts3 <= n-1)) {
  if(w_trans[(ts3)] == NULL) w_trans[(ts3)] = (double *) buffer_alloc_atomic((ts3), 128 * sizeof(double), buff_mang_w);;
  if(B_trans[(ts3) * ((B_size_1)) + (ts2)] == NULL) B_trans[(ts3)* ( B_size_1 ) + (ts2)] = (double *) buffer_alloc_atomic((ts3)* ( B_size_1 ) + (ts2), 128*128 * sizeof(double), buff_mang_B);;
  if(x_trans[(ts2)] == NULL) x_trans[(ts2)] = (double *) buffer_alloc_atomic((ts2), 128 * sizeof(double), buff_mang_x);;
}
/* End of CLooG code */
}
void data_tile_ref_count_update_3(int ts1,int ts2,int ts3, double** w_trans, BufferManager *buff_mang_w, double** B_trans,int B_size_1, BufferManager *buff_mang_B, double** x_trans, BufferManager *buff_mang_x)
{
  int t1, t2, t3;
 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;
/* Start of CLooG code */
if ((ts1 == 3) && (ts2 >= 0) && (128*ts2 <= n-1) && (ts3 >= 0) && (128*ts3 <= n-1)) {
  decrement_ref_count(buff_mang_w, (ts3));;
  decrement_ref_count(buff_mang_B, (ts3)* ( B_size_1 ) + (ts2));;
  decrement_ref_count(buff_mang_x, (ts2));;
}
/* End of CLooG code */
}
void compute_task_3(int t1,int t2,int t3, double** w_trans, BufferManager *buff_mang_w, double** B_trans,int B_size_1, BufferManager *buff_mang_B, double** x_trans, BufferManager *buff_mang_x)
{
int i_0_div = 0;int i_0_mod = 0;double *w_i_0 = NULL;int i_j_0_div = 0;int i_j_0_mod = 0;double *B_i_j_0 = NULL;int j_0_div = 0;int j_0_mod = 0;double *x_j_0 = NULL; int t4, t5, t6, t7, t8, t9, t10;
 int lb, ub, lbp, ubp, lbd, ubd, lb2, ub2;
 register int lbv, ubv;
/* Start of CLooG code */
for (t4=128*t2;t4<=min(n-1,128*t2+127);t4++) {
  lbv=128*t3;
  ubv=min(n-1,128*t3+127);
  #if (defined __DATA_DIST_DECLS || defined USE_LOCAL_ARRAYS)
  i_0_div = ((1)*( lbv ))/128;i_0_mod = ((1)*( lbv ))%128;w_i_0 = w_trans[i_0_div];i_j_0_div = ((1)*( lbv ))/128* ( B_size_1 ) + ((1)*( t4 ))/128;i_j_0_mod = ((1)*( lbv ))%128* (128) + ((1)*( t4 ))%128;B_i_j_0 = B_trans[i_j_0_div];j_0_div = ((1)*( t4 ))/128;j_0_mod = ((1)*( t4 ))%128;x_j_0 = x_trans[j_0_div];;
  #endif
#pragma ivdep
#pragma vector always
  for (t5=lbv;t5<=ubv;t5++) {
    w_i_0[i_0_mod + t5 - lbv]=w_i_0[i_0_mod + t5 - lbv]+alpha*B_i_j_0[i_j_0_mod + (t5 - lbv) * 128]*x_j_0[j_0_mod];;
  }
}
/* End of CLooG code */
}
